import logging
from enum import Enum

import authenta.authenta_key_lib_interface as authenta_key_lib
from authenta.authenta_key_lib_interface import key_lib_close
import authenta.authenta_key_mgmt_interface as key_lib_key_mgmt

class DeviceType(Enum):
  """Authenta device type
  """
  NAXOS = 1
  INDIGO = 2

class AuthentaDevice:
  """ Base class for Authenta devices
  """
  def __init__(self, endpoint_platform, logger=None, user_defined_key_driver_data=None, user_defined_key_driver_dev_info_dict=None):
    self.endpoint_platform = endpoint_platform
    self.logger = logger or logging.getLogger(__name__)
    self.device_type = None
    self._uid = None
    self._keylib_context = None
    self._key_driver_data = user_defined_key_driver_data
    self._key_driver_dev_info_dict = user_defined_key_driver_dev_info_dict
    self._root_signing_key_externally_set = False
    self._session_key_externally_set = False

    if not self.logger.hasHandlers():
      # avoid exception in case no other handler is added
      self.logger.addHandler(logging.NullHandler())

  def __enter__(self):
    self.endpoint_platform.acquire_device(self.device_type)

  def __exit__(self, exception_type, exception_value, traceback):
    self.endpoint_platform.release_device(self.device_type)
    self.logger.info("exiting AuthentaDevice")

    # to surpress exception stack trace - return True
    return False

  def __del__(self):
    # body of destructor
    if authenta_key_lib is not None and self._keylib_context is not None:
      authenta_key_lib.key_lib_close(self._keylib_context)
    self._keylib_context = None

  @property
  def uid(self):
    return self._uid

  def initialize_internal_properties(self):
    self._uid = None
    self.logger.info('Clearing key library keys')
    if authenta_key_lib is not None and self._keylib_context is not None:
      authenta_key_lib.key_lib_clear_keys(self._keylib_context)


  def _open_key_driver(self):
      self.logger.info('Initializing key library context')
      if self._key_driver_dev_info_dict is None:
            self._key_driver_dev_info_dict = { }

      self._keylib_context = authenta_key_lib.key_lib_open(self.uid.hex(), self._key_driver_dev_info_dict, self._key_driver_data)
      if self._keylib_context is None:
          self.logger.error('Failed to open key library context! Command signing and verification will not be available!')
  

  def set_hmac_root_signing_key(self, key: bytes):
        """
        This function exposes the key driver to an application
        that wants to take responsibilty for setting the keys
        rather than leaving that top the SDK.

        This function sets the HMAC root signing key. This
        could be the SRK or CRK. The session key will be
        derived using this key (unless device.set_hmac_session_key
        is called).  Deriving the session key will still be a
        function handled in by the SDK as necessary.

        The UID must be set on this device before calling this function

        Arguments
          param[in] key The root signing key; this is a bytes object understood the key library
          implementation. It can be the raw key or a key identifier meaningful to the
          driver.
        
        Return:
          True upon success,
          False otherwise
       """
        ret = authenta_key_lib.set_hmac_root_signing_key(self._keylib_context, key)
        if ret:
              self._root_signing_key_externally_set = True

        return ret

  def set_hmac_session_key(self, key: bytes):
        """
        This function exposes the key driver to an application
        that wants to take responsibilty for setting the keys
        rather than leaving that top the SDK.

        This function sets the HMAC session key. Calls to derive
        the session key will not be made inside the SDK.

        The UID must be set on this device before calling this function

        Arguments
          param[in] key The root signing key; this is a bytes object understood the key library
          implementation. It can be the raw key or a key identifier meaningful to the
          driver.
        
        Return:
          True upon success,
          False otherwise
       """
        ret = authenta_key_lib.set_hmac_session_key(self._keylib_context, key)
        if ret:
              self._session_key_externally_set = True

        return ret

  def open_key_driver_explicit(self):
      """
      If an application is only using atomic functions it MUST call
      this function to call open on the key driver explicitly.

      This function requires that _uid be set to the attached device's
      UID (retrived from request_uid command) BEFORE making calling this
      function

      If the key library was opened already this function is a no-op, 
      (determined by testing if the internal variable _keylib_context 
      equals None).
      """
      if self._keylib_context is None:
          self._open_key_driver()
      
  def close_key_driver_explicit(self):
      """
      If an application is only using atomic functions it MUST call
      this function to close the key driver library when done with it.

      if the key driver is not open this is a no-op
      """
      if self._keylib_context is not None:
          authenta_key_lib.key_lib_close(self._keylib_context)
          self._keylib_context = None
        
        
