import os
import ctypes as ct
import json

import logging

from authenta.ctype_utils import *

# from enum import Enum

# class AuthentaKeyUsageID(Enum):
#     SRK = 1
#     CRK = 2
#     SSesK = 3
#     CSesK = 4


# class AuthentaKeyType(Enum):
    

# class AuthentaKeyInfo:
#     usage = None
#     type = None
    
_file = 'authenta_key_driver.so'

_path = os.path.join(os.path.dirname(__file__), _file)
KEY_LIB = ct.cdll.LoadLibrary(_path)

def key_lib_open(uid, device_info_dictionary,  user_defined_data=None):
    """
    * key_lib_open
    * 
    * This should always be called before using the key library and returns a context that must
    * be used in subsequent calls to the library.
    * The retuned context may hold state and is associated with a particular device. This
    * open function must be called once for every different device instance that an application will 
    * use.
    * 
    * Performs any required library and context initialization.
    * 
    * @param[in] uid UID (hex string representation) of the device associated with this open instance of
    *                the key library.
    * 
    * @param[in] device_info_dictionary Dictionary containing properties of the associated device.
    *                                   The contents will include at least:
    *                                   device_type: [Naxos | Indigo]
    *                                   device_version: [hex_string representing version returned by device,
    *                                                    e.g., '010005' ]
    * 
    * @param[in] user_defined_data Data that may be used by a particular implementation of the
    *                              key driver.
    * 
    * @return A "context" object that will be a parameter to all subsequent library function calls. 
    *         The returned context is associated with a single device. For additional devices
    *         call this function to get a new context for each additional device instance.
    *         
    *         None on failure
    """
    KEY_LIB.key_lib_open.argtypes = \
        (ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte))
    KEY_LIB.key_lib_open.restype = ct.c_int

    if  isinstance(uid, bytes):
        uid = uid.hex()
    ct_uid = str2c_str(uid)
    dev_info = json.dumps(device_info_dictionary)
    ct_dev_dict = str2c_str(dev_info)

    data = user_defined_data if user_defined_data != None else b''
    data_len = len(user_defined_data) if user_defined_data != None else 0
    ct_user_defined_data = str2c_buff(data, data_len)

    context = b'\x00'*8
    ct_context = str2c_buff(context, len(context))

    rc = KEY_LIB.key_lib_open(ct_uid, ct_dev_dict, ct_user_defined_data, ct_context)

    if rc == 0:
        #return context
        return bytes(ct_context)
    else:
        return None


def key_lib_close(context):
    """
    TBD
    """
    KEY_LIB.key_lib_close.argtypes = ([ct.POINTER(ct.c_ubyte)])
    KEY_LIB.key_lib_close.restype = ct.c_int

    ct_context = str2c_buff(context, len(context))

    rc = KEY_LIB.key_lib_close(ct_context)


def key_lib_clear_keys(context):
    """
    * key_lib_clear_context
    * 
    * Clears an key state kept in the key library.
    * 
    * return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))

    KEY_LIB.key_lib_clear_keys.argtypes = \
        ([ct.POINTER(ct.c_ubyte)])
    KEY_LIB.key_lib_clear_keys.restype = ct.c_int
    rc = KEY_LIB.key_lib_clear_keys(ct_context)

    return  rc == 0


def set_srk_as_hmac_root_signing_key(context, force: bool=False):
    """
    return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_force = bool2c_bool(force)


    KEY_LIB.set_srk_as_hmac_root_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_srk_as_hmac_root_signing_key.restype = ct.c_int

    rc = KEY_LIB.set_srk_as_hmac_root_signing_key(ct_context, ct_force)

    if rc == 0:
        return True
    else:
        return False


def set_crk_as_hmac_root_signing_key(context, force: bool=False):
    """
    set the client root key as the signing session key

    return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_force = bool2c_bool(force)

    KEY_LIB.set_crk_as_hmac_root_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_crk_as_hmac_root_signing_key.restype = ct.c_int

    rc = KEY_LIB.set_crk_as_hmac_root_signing_key(ct_context, ct_force)

    return rc == 0

def set_csk_as_hmac_session_signing_key(context, force: bool=False):
    """
    set the client session key as the signing session key

    return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_force = bool2c_bool(force)

    KEY_LIB.set_csk_as_hmac_session_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_csk_as_hmac_session_signing_key.restype = ct.c_int

    rc = KEY_LIB.set_csk_as_hmac_session_signing_key(ct_context, ct_force)

    return rc == 0

def set_ecdsa_verification_key(context, force: bool=False):
    """
    return True on Success, False on failure
    """
    KEY_LIB.set_ecdsa_verification_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_ecdsa_verification_key.restype = ct.c_int

    ct_context = str2c_buff(context, len(context))
    ct_force = bool2c_bool(force)


    rc = KEY_LIB.set_ecdsa_verification_key(ct_context, ct_force)

    return rc == 0

def set_ecdsa_signing_key(context, force: bool=False):
    """
    return True on Success, False on failure
    """
    KEY_LIB.set_ecdsa_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_ecdsa_signing_key.restype = ct.c_int

    ct_context = str2c_buff(context, len(context))
    ct_force = bool2c_bool(force)


    rc = KEY_LIB.set_ecdsa_signing_key(ct_context, ct_force)

    return rc == 0


def set_hmac_root_signing_key(context, key: bytes):
    """
    Add a HMAC signing key to the key library context.
    This function can be used to allow an app to set
    a signinging key in the key lib explicitely rather than relying on
    the SDK to call the set_srk_as_hmac_root_signing_key
    or set_crk_as_hmac_root_signing_key functions.
    Also can be used for setting CRK.

    Arguments:
        context[in]   context object returned by key_lib_open()
        key[in]       bytes object that is meaningful to the key driver
                      implementation.
                      it could be the raw key value (bytes) of the key
                      or a key identifier understood by the key driver
                      
    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_key = str2c_buff(key, len(key))
    ct_key_len = ct.c_int(len(key))

    KEY_LIB.set_hmac_root_signing_key.argtypes = (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int)
    KEY_LIB.set_hmac_root_signing_key.restype = ct.c_int

    ret = KEY_LIB.set_hmac_root_signing_key(ct_context, ct_key, ct_key_len)

    return bool(ret == 0)

def set_hmac_session_key(context, key: bytes):
    """
    Add a session key to the key library context

    This function can be used to allow an app to set
    a session key in the key lib explicitely rather than relying on
    the SDK to call the derive_hmac_session_ke functions.
    Also can be used for setting CSK.

    Arguments:
        context[in]   context object returned by key_lib_open()
        key[in]       bytes object that is meaningful to the key driver
                      implementation.
                      it could be the raw key value (bytes) of the key
                      or a key identifier understood by the key driver

    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_key = str2c_buff(key, len(key))
    ct_key_len = ct.c_int(len(key))

    KEY_LIB.set_hmac_session_key.argtypes = (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int)
    KEY_LIB.set_hmac_session_key.restype = ct.c_int

    ret = KEY_LIB.set_hmac_session_key(ct_context, ct_key, ct_key_len)

    return bool(ret == 0)

def derive_hmac_session_key(context, powerOnAmtc0: bytes, force: bool=False):
    """
    return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_po_amtc0 = str2c_buff(powerOnAmtc0, len(powerOnAmtc0))
    ct_po_amtc0_len = ct.c_int(len(powerOnAmtc0))
    ct_force = bool2c_bool(force)

    KEY_LIB.derive_hmac_session_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.c_bool)
    KEY_LIB.derive_hmac_session_key.restype = ct.c_int

    rc = KEY_LIB.derive_hmac_session_key(ct_context, ct_po_amtc0, ct_po_amtc0_len, ct_force)

    return rc == 0

def derive_ecdh_session_key(context, ecdh_pub_key: bytes, powerOnAmtc0: bytes, force: bool=False):
    """
    return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_ecdh_pub_key = str2c_buff(ecdh_pub_key, len(ecdh_pub_key))
    ct_po_amtc0 = str2c_buff(powerOnAmtc0, len(powerOnAmtc0))
    ct_po_amtc0_len = ct.c_int(len(powerOnAmtc0))
    ct_force = bool2c_bool(force)

    KEY_LIB.derive_ecdh_session_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.c_bool)
    KEY_LIB.derive_ecdh_session_key.restype = ct.c_int

    rc = KEY_LIB.derive_ecdh_session_key(ct_context, ct_ecdh_pub_key, ct_po_amtc0, ct_po_amtc0_len, ct_force)

    return rc == 0

def gen_root_sha256_hmac(context, message: bytes):
    """
    return 32 byte HMAC Digest (bytes) on Success, None on failure
    """
    digest = b''
    ct_context = str2c_buff(context, len(context))
    ct_message_len = int2c_int(len(message))
    ct_message = str2c_buff(message, len(message))
    ct_digest = str2c_buff(digest, 32)
 
    KEY_LIB.gen_root_sha256_hmac.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte))
    KEY_LIB.gen_root_sha256_hmac.restype = ct.c_int


    rc = KEY_LIB.gen_root_sha256_hmac(ct_context, ct_message, ct_message_len, ct_digest)

    if rc == 0:
        return bytes(ct_digest)
    else:
        return None

def gen_session_sha256_hmac(context, message: bytes):
    """
    return 32 byte HMAC Digest (bytes) on Success, None on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_message_len = int2c_int(len(message))
    ct_message = str2c_buff(message, len(message))
    digest = b''
    ct_digest = str2c_buff(digest, 32)
 
    KEY_LIB.gen_session_sha256_hmac.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte))
    KEY_LIB.gen_session_sha256_hmac.restype = ct.c_int


    rc = KEY_LIB.gen_session_sha256_hmac(ct_context, ct_message, ct_message_len, ct_digest)

    if rc == 0:
        return bytes(ct_digest)
    else:
        return None


def generate_ecdsa_signature(context, message: bytes):
    """
    return 64 byte ECDSA signature (bytes) on Success, None on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_message = str2c_buff(message, len(message))
    ct_message_len = int2c_int(len(message))
    sig = b''
    ct_sig = str2c_buff(sig, 64)
 
    KEY_LIB.generate_ecdsa_signature.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte))
    KEY_LIB.generate_ecdsa_signature.restype = ct.c_int


    rc = KEY_LIB.generate_ecdsa_signature(ct_context, ct_message, ct_message_len, ct_sig)

    if rc == 0:
        return bytes(ct_sig)
    else:
        return None


def generate_kts_ecdsa_signature(context, message: bytes):
    """
    return on success: 64 byte ECDSA signature (bytes) signed using the preset KTS ECDSA key, 
           on failure: None 
    """
    ct_context = str2c_buff(context, len(context))
    ct_message = str2c_buff(message, len(message))
    ct_message_len = int2c_int(len(message))
    sig = b''
    ct_sig = str2c_buff(sig, 64)
 
    KEY_LIB.generate_kts_ecdsa_signature.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte))
    KEY_LIB.generate_kts_ecdsa_signature.restype = ct.c_int


    rc = KEY_LIB.generate_kts_ecdsa_signature(ct_context, ct_message, ct_message_len, ct_sig)

    if rc == 0:
        return bytes(ct_sig)
    else:
        return None

def gen_kts_sha256_hmac(context, message: bytes):
    """
    return on success: 32 byte HMAC Digest (bytes) signed using the preset KTS HMAC key, 
           on failure: None 
    """
    digest = b''
    ct_context = str2c_buff(context, len(context))
    ct_message_len = int2c_int(len(message))
    ct_message = str2c_buff(message, len(message))
    ct_digest = str2c_buff(digest, 32)
 
    KEY_LIB.gen_kts_sha256_hmac.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte))
    KEY_LIB.gen_kts_sha256_hmac.restype = ct.c_int


    rc = KEY_LIB.gen_kts_sha256_hmac(ct_context, ct_message, ct_message_len, ct_digest)

    if rc == 0:
        return bytes(ct_digest)
    else:
        return None

def verify_root_sha256_hmac(context, msg: bytes, sig: bytes):
    """
    return True if the signature is verified, False if not, None on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_msg = str2c_buff(msg, len(msg))
    ct_msg_len = ct.c_int(len(msg))
    ct_sig = str2c_buff(sig, len(sig))
    ct_verified = ct.c_bool(False)

    KEY_LIB.verify_root_sha256_hmac.argtypes = (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_bool))
    KEY_LIB.verify_root_sha256_hmac.restype = ct.c_int
    ret_code = KEY_LIB.verify_root_sha256_hmac(ct_context, ct_msg, ct_msg_len, ct_sig, ct_verified)

    verified = None
    if ret_code == 0:
        verified = bool(ct_verified.value)

    return verified

def verify_session_sha256_hmac(context, msg: bytes, sig: bytes):
    """
    return True if the signature is verified, False if not, None on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_msg = str2c_buff(msg, len(msg))
    ct_msg_len = ct.c_int(len(msg))
    ct_sig = str2c_buff(sig, len(sig))
    ct_verified = ct.c_bool(False)

    KEY_LIB.verify_session_sha256_hmac.argtypes = (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_bool))
    KEY_LIB.verify_session_sha256_hmac.restype = ct.c_int
    ret_code = KEY_LIB.verify_session_sha256_hmac(ct_context, ct_msg, ct_msg_len, ct_sig, ct_verified)

    verified = None
    if ret_code == 0:
        verified = bool(ct_verified.value)

    return verified

def verify_ecdsa_signature(context, msg: bytes, sig: bytes):
    """
    return True if the signature is verified, False if not, None on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_msg = str2c_buff(msg, len(msg))
    ct_msg_len = ct.c_int(len(msg))
    ct_sig = str2c_buff(sig, len(sig))
    ct_verified = ct.c_bool(False)

    KEY_LIB.verify_ecdsa_signature.argtypes = (ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_bool))
    KEY_LIB.verify_ecdsa_signature.restype = ct.c_int
    ret_code = KEY_LIB.verify_ecdsa_signature(ct_context, ct_msg, ct_msg_len, ct_sig, ct_verified)

    verified = None
    if ret_code == 0:
        verified = bool(ct_verified.value)

    return verified


def set_kts_ecdsa_signing_key(context, slot_num: int, force: bool=False):
    """ 
        set the KTS resident key for ECDSA signing associated with the UID
        return True on Success, False on failure
    """ 
    ct_context = str2c_buff(context, len(context))
    ct_slot_num = int2c_int(slot_num)
    ct_force = bool2c_bool(force)

    KEY_LIB.set_kts_ecdsa_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_int, ct.c_bool)
    KEY_LIB.set_kts_ecdsa_signing_key.restype = ct.c_int

    rc = KEY_LIB.set_kts_ecdsa_signing_key(ct_context, ct_slot_num, ct_force)

    return rc == 0

def set_kts_hmac_signing_key(context, slot_num: int, ecdh_pub_key: bytes, force: bool=False):
    """ 
        set the KTS resident key for HMAC signing associated with the UID
        return True on Success, False on failure
    """
    ct_context = str2c_buff(context, len(context))
    ct_slot_num = int2c_int(slot_num)
    ct_ecdh_pub_key = str2c_buff(ecdh_pub_key, len(ecdh_pub_key))
    ct_force = bool2c_bool(force)

    KEY_LIB.set_kts_hmac_signing_key.argtypes = \
        (ct.POINTER(ct.c_ubyte), ct.c_int, ct.POINTER(ct.c_ubyte), ct.c_bool)
    KEY_LIB.set_kts_hmac_signing_key.restype = ct.c_int

    rc = KEY_LIB.set_kts_hmac_signing_key(ct_context, ct_slot_num, ct_ecdh_pub_key, ct_force)

    return rc == 0
