from ctypes import c_char_p
import os
import ctypes as ct
import logging

from authenta.ctype_utils import *

_file = 'authenta_key_driver.so'

_path = os.path.join(os.path.dirname(__file__), _file)
KEY_LIB = ct.cdll.LoadLibrary(_path)

KeyMgmt_KeyName_SRK = "srk"
KeyMgmt_KeyName_CRK = "crk"
KeyMgmt_KeyName_CSK = "tmp_client_session_key"
KeyMgmt_KeyName_ECDSA_PRV  = "ecdsa_prv"
KeyMgmt_KeyName_ECDSA_PUB  = "ecdsa_pub"
KeyMgmt_KeyName_ECDSA_DS_PuK = "ECDSA_DS_PuK"
KeyMgmt_KeyName_ECDH_SR_PRV  = "ecdh_prv"
KeyMgmt_KeyName_Session  = "session"

def delete_key(uid: str, key_name: str) -> bool:
    """
    Delete key with the input UID and key name from the key repository

    Args:
        uid[in]       UID to assocate with the new key
        key_name[in]  key_name of new key, i.e. type of key, for example "srk", "crk", etc.

    Return:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)

    KEY_LIB.delete_key_by_key_type.argtypes = (ct.c_char_p,)
    KEY_LIB.delete_key_by_key_type.restype = ct.c_int

    ret = KEY_LIB.delete_key_by_key_type(ct_uid, ct_key_name)

    return bool(ret == 0)



def add_key(uid: str, key_name: str, key: bytes, metadata: bytes, is_private_key: bool=True, allow_replace: bool=False) -> bool:
    """
    Add a new hmac key (uid, key_name, key) and generate handle for the new key

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  key_name of new key, i.e. type of key, for example "srk", "crk", etc.
        key[in]       raw key value (bytes) of new key
        metadata[in]  addition data to be associated with the key (opaque to the key library)
        is_private_key True if the key should be stored as a private key (default)
                       False if the key should be stored as a public key
        allow_replace  True if the key existing key with matching uid and key_name should be replaced
                       False if the function should leave the data unchanged and fail if there is a 
                       matching uid and key_name (default)

    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    ct_key = str2c_str(key.hex())
    the_metadata = metadata.hex() if metadata is not None else ''
    ct_metadata = str2c_str(the_metadata)
    ct_is_priv = bool2c_bool(is_private_key)
    ct_allow_replace = bool2c_bool(allow_replace)

    KEY_LIB.add_nonvolatile_based_key.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool)
    KEY_LIB.add_nonvolatile_based_key.restype = ct.c_int

    ret = KEY_LIB.add_nonvolatile_based_key(ct_uid, ct_key_name, ct_key, ct_metadata, ct_is_priv, ct_allow_replace)

    return bool(ret == 0)

def copy_key(uid: str, key_name: str, copy_name: str) -> bool:
    """
    Make a copy an private or symetric key. Store using the same UID with the input key name

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  name of key to copy
        copy_name[in] name of new copy

    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    ct_copy_name = str2c_str(copy_name)

    KEY_LIB.copy_private_key.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p)
    KEY_LIB.copy_private_key.restype = ct.c_int

    ret = KEY_LIB.copy_private_key(ct_uid, ct_key_name, ct_copy_name)

    return bool(ret == 0)

def export_private_key_in_the_clear(uid: str, key_name: str) -> bytes:
    """
    Return the value of a private or symetric key associated with the input UID and key name

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  name of key to copcopy_name: stry

    Returns:
        The hex string value of the key upon success,
        None if no key with the input uid and name is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    key = bytes(32)
    ct_key = bytes2c_buff(key, 32)
    ct_key_len = ct.c_int(32)

    KEY_LIB.get_key.argtypes = (ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_int))
    KEY_LIB.get_key.restype = ct.c_int

    ret = KEY_LIB.get_key(ct_uid, ct_key_name, ct_key, ct.pointer(ct_key_len))

    if ret == 0:
        return bytes(ct_key)
    else:
        return None

def read_public_key(uid: str, key_name: str) -> bytes:
    """
    Return the value of a public key associated with the input UID and key name

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  name of key to copcopy_name: stry

    Returns:
        The hex string value of the key upon success,
        None if no key with the input uid and name is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    key = bytes(64)
    ct_key = bytes2c_buff(key, 64)
    ct_key_len = ct.c_int(64)

    KEY_LIB.get_key.argtypes = (ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_ubyte), ct.POINTER(ct.c_int))
    KEY_LIB.get_key.restype = ct.c_int

    ret = KEY_LIB.get_key(ct_uid, ct_key_name, ct_key, ct.pointer(ct_key_len))

    if ret == 0:
        return bytes(ct_key)
    else:
        return None

def export_priv_key_as_hex(uid: str, key_name: str) -> str:
    """
    Return the value of a private or symetric key associated with the input UID and key name

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  name of key to copcopy_name: stry

    Returns:
        The hex string value of the key upon success,
        None if no key with the input uid and name is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    key = str('')
    ct_key_len = int2c_int(128)
    ct_key = str2c_str(key)


    KEY_LIB.export_prv_key_in_clear_as_hex.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_int))
    KEY_LIB.export_prv_key_in_clear_as_hex.restype = ct.c_int

    ret = KEY_LIB.export_prv_key_in_clear_as_hex(ct_uid, ct_key_name, ct_key, ct.pointer(ct_key_len))

    if ret == 0:
        return ct_key.value.decode() if ct_key.value is not None else None
    else:
        return None

def read_public_key_as_hex(uid: str, key_name: str) -> str:
    """
    Return the value of a public key associated with the input UID and key name

    Arguments:
        uid[in]       UID to assocate with the new key
        key_name[in]  name of key to copcopy_name: stry

    Returns:
        The hex string value of the key upon success,
        None if no key with the input uid and name is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key_name = str2c_str(key_name)
    key = str('')
    ct_key_len = int2c_int(129)
    ct_key = str2c_str(key)


    KEY_LIB.read_public_key_as_hex.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.POINTER(ct.c_int))
    KEY_LIB.read_public_key_as_hex.restype = ct.c_int

    ret = KEY_LIB.read_public_key_as_hex(ct_uid, ct_key_name, ct_key, ct.pointer(ct_key_len))

    if ret == 0:
        return ct_key.value.decode() if ct_key.value is not None else None
    else:
        return None

def set_new_srk_key_data_for_uid(uid: str, key_handle: str)->bool:
    """
    Replace the srk key data for the given UID with the
    key content associated with the given key handle.

    Note that the existing key data for the srk key of the
    input UID is overwritten with the key data associated
    with the input handle.  The entry for the input handle
    remains unchanged.

    Args:
        uid[in] the uid of the srk key to replace.
                an entry with this UID and key type 'srk'
                must exist in the key repository.

        key_handle[in] handle to the new key data
                       this handle must refer to an existing
                       key in the key repository

    Return  True on success
            False on failure
    """
    KEY_LIB.set_new_srk_key_data_for_uid.argtypes = \
        (ct.c_char_p, ct.c_char_p)
    KEY_LIB.set_new_srk_key_data_for_uid.restype = ct.c_int

    ct_uid = str2c_str(uid)
    ct_handle = str2c_str(key_handle)

    rc = KEY_LIB.set_new_srk_key_data_for_uid(ct_uid, ct_handle)

    return bool(rc == 0)

def delete_crk(uid:str) -> bool:
    """
    Delete the CRK from the key repository

    Args:
        uid[in]       uid to of the CRK to delete

    Return:
        True upon success,
        False if no crk key for the associated UID is found or any other failure
    """
    ct_uid = str2c_str(uid)

    KEY_LIB.delete_crk.argtypes = ([ct.c_char_p])
    KEY_LIB.delete_crk.restype = ct.c_int

    ret = KEY_LIB.delete_crk(ct_uid)

    return bool(ret == 0)



def add_crk(uid: str, key: bytes, metadata: bytes=None) -> bool:
    """
    Add a CRK to the key library

    Arguments:
        uid[in]       uid to associated the CRK with
        key[in]       raw key value (bytes) of new key
        metadata[in]  addition data to be associated with the key (opaque to the key library)

    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key = str2c_str(key.hex())
    the_metadata = metadata.hex() if metadata is not None else ''
    ct_metadata = str2c_str(the_metadata)

    KEY_LIB.add_crk.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p)
    KEY_LIB.add_crk.restype = ct.c_int

    ret = KEY_LIB.add_crk(ct_uid, ct_key, ct_metadata)

    return bool(ret == 0)

def add_csk(uid: str, key: bytes, metadata: bytes=None) -> bool:
    """
    Add a CSK to the key library

    Arguments:
        uid[in]       uid to associated the CSK with
        key[in]       raw key value (bytes) of new key
        metadata[in]  addition data to be associated with the key (opaque to the key library)

    Returns:
        True upon success,
        False if no key with the input handle is found or any other failure
    """
    ct_uid = str2c_str(uid)
    ct_key = str2c_str(key.hex())
    the_metadata = metadata.hex() if metadata is not None else ''
    ct_metadata = str2c_str(the_metadata)

    KEY_LIB.add_csk.argtypes = (ct.c_char_p, ct.c_char_p, ct.c_char_p)
    KEY_LIB.add_csk.restype = ct.c_int

    ret = KEY_LIB.add_csk(ct_uid, ct_key, ct_metadata)

    return bool(ret == 0)

def contains_crk(uid: str) -> bool:
    """
    Search for CRK in the key library

    Arguments:
        uid[in]   uid associated with the CSK being searched for

    Returns:
        True of a CRK is found,
        False otherwise
    """
    ct_uid = str2c_str(uid)

    KEY_LIB.contains_crk.argtypes = ([ ct.c_char_p ])
    KEY_LIB.contains_crk.restype = ct.c_int

    ret = KEY_LIB.contains_crk(ct_uid)

    return bool(ret != 0)

def contains_csk(uid: str) -> bool:
    """
    Search for CSK in the key library

    Arguments:
        uid[in]   uid associated with the CSK being searched for

    Returns:
        True of a CSK is found,
        False otherwise
    """
    ct_uid = str2c_str(uid)

    KEY_LIB.contains_csk.argtypes = ([ ct.c_char_p ])
    KEY_LIB.contains_csk.restype = ct.c_int

    ret = KEY_LIB.contains_csk(ct_uid)

    return bool(ret != 0)

def contains_key(uid: str, keyname: str) -> bool:
    """
    Search for a key associated with the input uid and keyname in the key library

    Arguments:
        uid[in] uid  uid associated with the CSK being searched for
        uid[in] keyname  keyname associated with the CSK being searched for

    Returns:
        True of a key matching input parameters is found,
        False otherwise
    """
    ct_uid = str2c_str(uid)
    ct_keyname = str2c_str(keyname)

    KEY_LIB.contains_key.argtypes = (ct.c_char_p, ct.c_char_p)
    KEY_LIB.contains_key.restype = ct.c_int

    ret = KEY_LIB.contains_key(ct_uid, ct_keyname)

    return bool(ret != 0)
