from random import random, shuffle
from time import sleep, time
from typing import Tuple, Optional
import sys
import os
import logging
import authenta.naxos.client_svc as naxos_client_svc
import authenta.indigo.client_svc as indigo_client_svc
import authenta.platform_driver_pb2 as pb2
from .endpoint_platform import EndpointPlatform, PlatformReturnType, EpPlatformRetCode
from .authenta_device import DeviceType


class GrpcPlatform(EndpointPlatform):
  """ GRPC implementation of endpoint platform
  """

  _naxos_subopcode_to_grpc_function = {
      0x34: naxos_client_svc.auth_modify,
      0x35: naxos_client_svc.provision_crtm,
      0x36: naxos_client_svc.deprovision_crtm,
      0x37: naxos_client_svc.auth_lock,
      0x38: naxos_client_svc.crtm_register_read,
      0x3C: naxos_client_svc.crtm_register_write,
      0x3A: naxos_client_svc.flash_extend,
      0x39: naxos_client_svc.replace_srk,
      0x3B: naxos_client_svc.acrtm_update_start,
      0x3D: naxos_client_svc.acrtm_update_finish,
      0x3E: naxos_client_svc.request_uid,
      0x3F: naxos_client_svc.measure,
      0x03: naxos_client_svc.request_monotonic_counter
  }

  _indigo_subopcode_to_grpc_function = {
      0x3E: indigo_client_svc.request_uid,
      0x35: indigo_client_svc.provision_crtm,
      0x36: indigo_client_svc.deprovision_crtm,
      0x37: indigo_client_svc.apa_modify,
      0x3A: indigo_client_svc.flash_extend,
      0x3B: indigo_client_svc.acrtm_update_start,
      0x3D: indigo_client_svc.acrtm_update_finish,
      0x38: indigo_client_svc.crtm_register_read,
      0x3C: indigo_client_svc.crtm_register_write,
      0x39: indigo_client_svc.replace_srk,
      0x34: indigo_client_svc.manage_kts,
      0x3F: indigo_client_svc.measure,
      0x40: indigo_client_svc.pcr_extend,
      0x43: indigo_client_svc.create_signature,
      0x44: indigo_client_svc.verify_signature,
      0x03: indigo_client_svc.request_monotonic_counter
  }

  def __init__(self, device_driver_path, ip_addr, port="50051", logger=None):
    self.device_id = None # this could be the file descriptor
    self.ip_addr = ip_addr
    self.port = port
    self.device_driver_path = device_driver_path
    self._context = None
    self.device_acquired = None
    self.logger = logger or logging.getLogger(__name__)
    if not self.logger.hasHandlers():
      # avoid exception in case no other handler is added
      self.logger.addHandler(logging.NullHandler())

  @property
  def context(self):
    return self._context

  def _grpc_call_authenta_naxos(self, function, request):
    """ Perform a Naxos GRPC function call to send request and return response
    """

    cmd_obj = pb2.naxos_content(bytes = request, nb_bytes = len(request), fd = self.device_id)
    grpc_msg_ret = function(self.context, cmd_obj)
    return grpc_msg_ret

  def _grpc_call_authenta_indigo(self, function, request):
    """ Perform a Indigo GRPC function call to send request and return response
    """

    cmd_obj = pb2.indigo_content(bytes = request, nb_bytes = len(request), fd = self.device_id)
    grpc_msg_ret = function(self.context, cmd_obj)
    return grpc_msg_ret

  def exec_on_device(self, device_type: DeviceType, cmd, cmd_name=None):
    """ Excute Authenta command on device

        e.g., 9b ...  or 96
        On the endpoint, GRPC server just execute the comamnd and return status

        Returns:
          PlatformReturnType
    """
    if cmd_name:
      nm = cmd_name
    else:
      nm = ''

    self.logger.info('sending command {}: 0x{}'.format(nm, cmd.hex()))
    grpc_msg_ret = None

    if device_type == DeviceType.NAXOS:
      opcode = cmd[0]
      if opcode == 0x9b:
        sub_opcode = cmd[1]
        func = self._naxos_subopcode_to_grpc_function.get(sub_opcode, None)
        if func is not None:
          grpc_msg_ret = self._grpc_call_authenta_naxos(func, cmd)
        else:
          self.logger.error('Unrecognized sub-opcode {} for authenta command'.format(sub_opcode))
      elif opcode == 0x96:
        # read gprr
        grpc_msg_ret = naxos_client_svc.read_general_purpose_readout_register(self.context, self.device_id)
      else:
        self.logger.error('Unknown command {}'.format(cmd))

    elif device_type == DeviceType.INDIGO:
      opcode = cmd[-2]
      if opcode == 0x9b:
        sub_opcode = cmd[-1]
        func = self._indigo_subopcode_to_grpc_function.get(sub_opcode, None)
        if func is not None:
          grpc_msg_ret = self._grpc_call_authenta_indigo(func, cmd)
        else:
          self.logger.error('Unrecognized sub-opcode {} for authenta command'.format(sub_opcode))
      else:
        self.logger.error('Unknown command {}'.format(cmd.hex()))

    else:
      self.logger.error('Unknown device type {}'.format(device_type))

    if grpc_msg_ret is not None:
      ret_obj = self._toPlatformReturnType(grpc_msg_ret)
    else:
      ret_obj = PlatformReturnType(EpPlatformRetCode.EP_ERROR)

    return ret_obj

  # Non-Authenta Platform Commands

  def acquire_device(self, device_type: DeviceType):
    """ Open the device and get device id (This could be file descriptor)
        On the endpoint, the device should be locked
    """

    device_lock_timeout = 5
    accumulator         = 0.0
    now                 = time()
    st_time             = now
    
    client_svc = None
    if device_type == DeviceType.NAXOS:
      client_svc = naxos_client_svc
    elif device_type == DeviceType.INDIGO:
      client_svc = indigo_client_svc
    else:
      self.logger.error('Unknown device type {}'.format(device_type))
      return self

    try:
      # connect to RPC Server
      ip = self.ip_addr + ":" + self.port
      self._context  = client_svc.t_authenta_context(ip)
      self.device = client_svc.authenta_device_open(self.context, self.device_driver_path)

      if  not self.device.status:
        self.logger.info("Opened filehandle to device")
        self.device_id = self.device.fd # file descriptor
      else:
        self.logger.error("Failed to Open device!")
        raise FileNotFoundError

      while True:
        # lock the device
        self.device_acquired = (client_svc.authenta_device_acquire(self.context, self.device_id).status == 0)

        if self.device_acquired:
          self.logger.info("gRPCPlatform; device authenta_device_acquire: success" )
          break
        elif (now-st_time)>device_lock_timeout:
          self.logger.error("gRPCPlatform; device authenta_device_acquire: FAILED" )
          raise PermissionError
        else:
          self.logger.warning("device busy...")
          accumulator = accumulator + random()
          sleep(accumulator)
          now = time()

      self.logger.info("RPC endpoint connection established: enpoint ip_addr:{}:{}".format(self.ip_addr, self.port))

    except FileNotFoundError as e:
      self.logger.error("RPC Connection Failed: ip_addr: {}:{}".format(self.ip_addr, self.port))
      #self.logger.exception(e)
      print("\n\n\n")
      print("error: RPC Connection Failed: ip_addr: {}:{}".format(self.ip_addr, self.port))
      #print(e)
      print("\n\n\n")
    except PermissionError as e:
      pass

    return self

  def release_device(self, device_type: DeviceType):
    """ Close the device based on the device id

        On the endpoint, the device should be unlocked
    """

    client_svc = None
    if device_type == DeviceType.NAXOS:
      client_svc = naxos_client_svc
    elif device_type == DeviceType.INDIGO:
      client_svc = indigo_client_svc
    else:
      self.logger.error('Unknown device type {}'.format(device_type))
      return

    if self.device_acquired:
      self.device_acquired = client_svc.authenta_device_release(self.context, self.device_id).status == 0
      self.logger.info("gRPCPlatfrom: endpoint device released")

  def is_device_acquired(self):
    return self.device_acquired is not None and self.device_acquired

  # Non-Authenta Device Commands

  def naxos_page_program(self, address: int, data: bytes):
    """ Issue Page_Program command

        Args:
          address (int)
          data (bytes)

        Returns:
          bool: True on success or False on failure
    """
    addr_content = pb2.flash_address(length = len(data), address = address, fd = self.device_id)
    cmd = pb2.flash_write_content(address = addr_content, data = data)
    cmd_rtn = naxos_client_svc.page_program(self.context, cmd)
    ret = not cmd_rtn.status
    self.logger.info("Page program ret:{}".format(ret))
    return ret

  def naxos_fast_read(self, address: int):
    """ Issue Fast_Read command

        Args:
          address (int)

        Returns:
          (True, response) on success or (False, None) on failure
    """
    self.logger.info("address: {}".format(address))
    cmd = pb2.flash_address(length=2048, address=address, fd=self.device_id)
    cmd_rtn = naxos_client_svc.fast_read(self.context, cmd)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return: {}".format(ret))
    return ret, response

  def naxos_dyb_write(self, lockbit: int, address: int):
    """ Issue DYB_Write command

        Args:
          lockbit (int)
          address (int)

        Returns:
          bool: True on success or False on failure
    """
    if lockbit != 0 and lockbit != 1:
      self.logger.error("Lockbit was not boolean")
      return False

    addr_content = pb2.flash_address(length=1, address=address, fd=self.device_id)
    cmd = pb2.flash_lock_config(address=addr_content, lockbit=lockbit)
    cmd_rtn = naxos_client_svc.dyb_write(self.context, cmd)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_dyb_read(self, address: int):
    """ Issue DYB_Read command

        Args:
          address (int)

        Returns:
          (True, response) on success or (False, None) on failure
    """
    cmd = pb2.flash_address(length=1, address=address, fd=self.device_id)
    cmd_rtn = naxos_client_svc.dyb_read(self.context, cmd)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return: {}".format(ret))
    return ret, response

  def naxos_ppb_program(self, address: int):
    """ Issue PPB_Program command

        Args:
          address (int)

        Returns:
          bool: True on success or False on failure
    """
    self.logger.info("addr: {}".format(address))
    cmd = pb2.flash_address(length = 1, address = address, fd = self.device_id)
    cmd_rtn = naxos_client_svc.ppb_program(self.context, cmd)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_ppb_read(self, address: int):
    """ Issue PPB_Read command

        Args:
          address (int)

        Returns:
          (True, response) on success or (False, None) on failure
    """
    self.logger.info("addr: {}".format(address))
    cmd = pb2.flash_address(length = 1, address = address, fd = self.device_id)
    cmd_rtn = naxos_client_svc.ppb_read(self.context, cmd)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return: {}".format(ret))
    return ret, response

  def naxos_ppb_lock_bit_read(self):
    """ Issue PPB_Lock_Bit_Read command

        Returns:
          (True, response) on success or (False, None) on failure
    """
    cmd_rtn = naxos_client_svc.ppb_lock_bit_read(self.context, self.device_id)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return: {}".format(ret))
    return ret, response

  def naxos_subsector_erase(self, address: int):
    """ Issue PPB_Lock_Bit_Read command

        Args:
          address (int)

        Returns:
          bool: True on success or False on failure
    """
    cmd = pb2.flash_address(length = 0, address = address, fd = self.device_id)
    cmd_rtn = naxos_client_svc.subsector_erase(self.context, cmd)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_ppb_erase(self):
    """ Issue PPB_Erase command

        Returns:
          bool: True on success or False on failure
    """
    cmd_rtn = naxos_client_svc.ppb_erase(self.context, self.device_id)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_write_enable(self):
    """ Issue Write_Enable command

        Returns:
          bool: True on success or False on failure
    """
    cmd_rtn = naxos_client_svc.write_enable(self.context, self.device_id)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_write_disable(self):
    """ Issue Write_Disable command

        Returns:
          bool: True on success or False on failure
    """
    cmd_rtn = naxos_client_svc.write_disable(self.context, self.device_id)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_ppb_lock_bit_write(self):
    """ Issue PPB_Lock_Bit_Write command

        Returns:
          bool: True on success or False on failure
    """
    cmd_rtn = naxos_client_svc.ppb_lock_bit_write(self.context, self.device_id)
    ret = not cmd_rtn.status
    self.logger.info("return: {}".format(ret))
    return ret

  def naxos_read_status_register(self):
    """ Issue Read_Status_Register command

        Returns:
          (True, response) on success or (False, None) on failure
    """
    cmd_rtn = naxos_client_svc.read_status_register(self.context, self.device_id)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return: {}".format(ret))
    return ret, response

  def indigo_read_multiple_block(self, address: int, length: int) -> Tuple[bool, Optional[bytes]]:
    """
    Read_Multiple_Block (read data from flash)
    Return (True, response) on success or (False, None) on failure
    """
    self.logger.info("address: {}; length: {}".format(address, length))
    cmd = pb2.flash_address(length=length, address=address, fd=self.device_id)
    cmd_rtn = indigo_client_svc.read_multiple_block(self._context, cmd)
    ret = not cmd_rtn.status
    response = cmd_rtn.return_buf[:cmd_rtn.buffer_size] if ret else None
    self.logger.info("return:{}".format(ret))
    return ret, response

  def indigo_write_multiple_block(self, address: int, data: bytes) -> bool:
    """
    Write_Multiple_Block (write data to flash)
    Return True on success or False on failure
    """
    self.logger.info("address: {}; data (first 16 btyes): {}".format(address, data[:16].hex()))
    data_sz = len(data)
    addr_content = pb2.flash_address(length=data_sz, address=address, fd=self.device_id)
    cmd = pb2.flash_write_content(address=addr_content, data=data)
    cmd_rtn = indigo_client_svc.write_multiple_block(self._context, cmd)
    ret = not cmd_rtn.status
    self.logger.info("return:{}".format(ret))
    return ret

  def _toPlatformReturnType(self, grpc_message_ret):
    """ Convert a message_return type to a PlatformRetrunType
    """
    if isinstance(grpc_message_ret, pb2.message_return):
      buf = grpc_message_ret.return_buf
    else:
      buf = None # it is a 'platform_driver_pb2.status_type_msg' - no buffer field

    if grpc_message_ret.status == 0:
      status = EpPlatformRetCode.EP_OK
    else:
      status = EpPlatformRetCode.EP_ERROR

    return PlatformReturnType(status, buf)
