#! /usr/bin/env python3

"""
The Python implementation of the Indigo platform_driver GRPC client.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import grpc
import authenta.platform_driver_pb2 as platform_driver_pb2
import authenta.platform_driver_pb2_grpc as platform_driver_pb2_grpc

# device context class
class t_authenta_context():
    def __init__(self, ip_address):
        self.ip_address = ip_address

# empty class
class empty():
    def __init(self):
        pass

""" 
send rpc call and get response
@param[in] server_addr  Server address
@param[in] request      Request message to send
@param[in] func_name    RPC function name
return response message on success or None on RPC failure
"""
def rpc_call(server_addr, request, func_name):
    response = None

    try:
        with grpc.insecure_channel(server_addr) as channel:
            stub = platform_driver_pb2_grpc.IndigoStub(channel)

            # get rpc function based on function name
            func = getattr(stub, func_name)
            response = func(request)
    except:
        # rpc failure
        print('Error: rpc failure when calling ' + func_name)
        response = empty()

        if func_name == "Authenta_detect_devices":
            response.listsize = 0
        else:
            response.status = 1

    return response

"""
Update return_type of device context based on RPC response
@param[in] context      device context
@param[in] response     response from RPC call
"""
def update_dev_context_return_type(context, response):    
    context.return_type = response.status

"""
Update out_buf of device context based on RPC response
@param[in] context      device context
@param[in] response     response from RPC call
"""
def update_dev_context_out_buf(context, response):    
    context.out_buf = response

"""
Update large_out_buf of device context based on RPC response
@param[in] context      device context
@param[in] response     response from RPC call
"""
def update_dev_context_large_out_buf(context, response):    

    context.large_out_buf = response

"""
Update descriptor of device context based on RPC response
@param[in] context      device context
@param[in] response     response from RPC call
"""
def update_dev_context_descriptor(context, response):    
    context.descriptor = response

"""
Update devices_list of device context based on RPC response
@param[in] context      device context
@param[in] response     response from RPC call
"""
def update_dev_context_devices_list(context, response):    
    context.devices_list = response


# RPC client implementations

# device_descriptor Authenta_device_open(string_msg)
def authenta_device_open(context, path: str):
    request = platform_driver_pb2.string_msg(value = path)
    response = rpc_call(context.ip_address, request, "Authenta_device_open")
    update_dev_context_descriptor(context, response)
    return response

# status_type_msg Authenta_device_close(int_msg)
def authenta_device_close(context, fd: int):
    request = platform_driver_pb2.int_msg(value = fd)
    response = rpc_call(context.ip_address, request, "Authenta_device_close")
    update_dev_context_return_type(context, response)
    return response

# status_type_msg Acquire_Device(int_msg)
def authenta_device_acquire(context, fd: int):
    request = platform_driver_pb2.int_msg(value = fd)
    response = rpc_call(context.ip_address, request, "Acquire_Device")
    update_dev_context_return_type(context, response)
    return response

# status_type_msg Release_Device(int_msg)
def authenta_device_release(context, fd: int):
    request = platform_driver_pb2.int_msg(value = fd)
    response = rpc_call(context.ip_address, request, "Release_Device")
    update_dev_context_return_type(context, response)
    return response

# list_info Authenta_detect_devices(void_msg)
def authenta_detect_devices(context):
    request = platform_driver_pb2.void_msg()
    response = rpc_call(context.ip_address, request, "Authenta_detect_devices")
    update_dev_context_devices_list(context, response)
    return response

# message_return Request_UID(indigo_content)
def request_uid(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Request_UID")    
    update_dev_context_return_type(context, response)
    return response

# message_return Provision_CRTM(indigo_content)
def provision_crtm(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Provision_CRTM")
    update_dev_context_return_type(context, response)
    return response

# message_return Deprovision_CRTM(indigo_content)
def deprovision_crtm(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Deprovision_CRTM")
    update_dev_context_return_type(context, response)
    return response

# message_return APA_Modify(indigo_content)
def apa_modify(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "APA_Modify")
    update_dev_context_return_type(context, response)
    return response

# message_return Flash_Extend(indigo_content)
def flash_extend(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Flash_Extend")
    update_dev_context_return_type(context, response)
    return response

# message_return ACRTM_Update_Start(indigo_content)
def acrtm_update_start(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "ACRTM_Update_Start")
    update_dev_context_return_type(context, response)
    return response

# message_return ACRTM_Update_Finish(indigo_content)
def acrtm_update_finish(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "ACRTM_Update_Finish")
    update_dev_context_return_type(context, response)
    return response

# message_return CRTM_Register_Read(indigo_content)
def crtm_register_read(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "CRTM_Register_Read")
    update_dev_context_return_type(context, response)
    return response

# message_return CRTM_Register_Write(indigo_content)
def crtm_register_write(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "CRTM_Register_Write")
    update_dev_context_return_type(context, response)
    return response

# message_return Replace_SRK(indigo_content)
def replace_srk(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Replace_SRK")
    update_dev_context_return_type(context, response)
    return response

# message_return Manage_KTS(indigo_content)
def manage_kts(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Manage_KTS")
    update_dev_context_return_type(context, response)
    return response

# message_return Measure(indigo_content)
def measure(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Measure")
    update_dev_context_return_type(context, response)
    return response

# message_return PCR_Extend(indigo_content)
def pcr_extend(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "PCR_Extend")
    update_dev_context_return_type(context, response)
    return response

# message_return CreateSignature(indigo_content)
def create_signature(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "CreateSignature")
    update_dev_context_return_type(context, response)
    return response

# message_return VerifySignature(indigo_content)
def verify_signature(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "VerifySignature")
    update_dev_context_return_type(context, response)
    return response

# message_return Request_Monotonic_Counter(indigo_content)
def request_monotonic_counter(context, cmd: platform_driver_pb2.indigo_content):
    response = rpc_call(context.ip_address, cmd, "Request_Monotonic_Counter")
    update_dev_context_return_type(context, response)
    return response

# message_return Read_Multiple_Block(flash_address)
def read_multiple_block(context, address: platform_driver_pb2.flash_address):
    response = rpc_call(context.ip_address, address, "Read_Multiple_Block")
    update_dev_context_return_type(context, response)
    return response

# message_return Write_Multiple_Block(flash_write_content)
def write_multiple_block(context, content: platform_driver_pb2.flash_write_content):
    response = rpc_call(context.ip_address, content, "Write_Multiple_Block")
    update_dev_context_return_type(context, response)
    return response
