from ..cmd_error_code import CmdErrorCode

class IndigoCmdErrorCode(CmdErrorCode):
    """
        Indigo command error parsing implementation

        The error message is set based on the error specified in the TRD for the
        given ESR and command name.
        An ESR of 0x00 indicates ther error was from a non-Authenta command or
        failure happened before an ESR was returned. The error message is determined
        by the non_authenta_err_msg parameter.

        Args:
            esr: The ESR returned by the device (0x80 always indicates successful
                 command completion)
            cmd_name: The name of the command that failed
            non_authenta_err_msg: Caller's error message. Optional, only used
                                  for non-Authenta errors
    """

    def __init__(self, esr: bytes, cmd_name: str, non_authenta_err_msg="Unknown Error"):
        """
        Translate esr to error message based on command name
        """
        self._esr = esr
        self._cmd_name = cmd_name
        self._valid_response = None # bool
        self._sig_matched = None # bool
        self._nonce_matched = None # bool
        self._last_opcode_matched = None # bool
        self._esr_bit = None
        self._error_msg = self._msg_lookup(non_authenta_err_msg)

    def _msg_lookup(self, non_authenta_err_msg):
        if self._esr == b'\x80\x00':
            return 'OK'
        elif self._esr == b'\x00\x00':
            return non_authenta_err_msg

        bits = [int(i) for i in bin(int.from_bytes(self.esr, byteorder="big"))[::-1][:-2]]
        ones = [index for index in range(len(bits)) if bits[index] == 1]

        # if len(ones) > 1:
        #     # special ESR (e.g., 0x35, 0x3B, and 0x37 in Auth_Modify)
        #     self._esr_bit = int.from_bytes(self._esr, byteorder="big")
        # else:
        #     self._esr_bit = ones[0]

        self._esr_bit = ones[0]

        cmd_err_bit_dict = type(self).error_map.get(self.cmd_name)
        if cmd_err_bit_dict:
            ret_msg = cmd_err_bit_dict.get(self._esr_bit)
            base_msg = type(self).std_error_map.get(self._esr_bit)

            if base_msg is not None:
                if ret_msg is None:
                    ret_msg = base_msg
                else:
                    ret_msg = base_msg + '; or ' + ret_msg

        else:
            ret_msg = '{} not found'.format(self._cmd_name)
        return ret_msg

    @property
    def esr(self):
        return self._esr

    @property
    def esr_success(self):
        return self.esr == b'\x00\x80' # Successful Completion

    @property
    def error_msg(self):
        return self._error_msg

    @property
    def esr_bit(self):
        return self._esr_bit if self._esr_bit < 8 else None #UPDATE

    @property
    def esr_bit_msg(self):
        if self._esr_bit is None:
            return 'Non-Authenta error'

        if self.esr_bit is None:
            return 'ESR: ' + self._esr.hex()
        else:
            return 'ESR bit: ' + str(self.esr_bit)

    @property
    def cmd_name(self):
        return self._cmd_name

    @property
    def valid_response(self):
        return self._valid_response

    def set_valid_response(self, valid: bool):
        self._valid_response = valid

    @property
    def signature_matched(self):
        return self._sig_matched

    def set_signature_matched(self, matched: bool):
        self._sig_matched = matched

    @property
    def nonce_matched(self):
        return self._nonce_matched

    @property
    def esr_and_response_success(self):
        return self.esr_success and self._valid_response

    def set_nonce_matched(self, matched: bool):
        self._nonce_matched = matched

    @property
    def last_successful_opcode_matched(self):
        return self._last_opcode_matched

    def set_successful_opcode_matched(self, matched: bool):
        self._last_opcode_matched = matched

    # Standard errors map (for all commands)
    std_error_map = {
        2: 'Bad signature or incorrect payload size or Opcode out of range',
        4: 'Counter data mismatch',
        6: 'PUF failed to extract correctly'
    }

    # Individual function errors map
    error_map = dict.fromkeys(['request_uid'], {
        5: 'Failed internal UID integrity check'
    })
    error_map.update(dict.fromkeys(['provision_crtm'], {
        1: 'Already provisioned',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['deprovision_crtm'], {
        3: 'Deprovisioning is not allowed at this time',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['apa_modify'], {
        3: 'The requested key does not have rights to perform a modify',
        4: 'Used/referenced keys are not valid or do not exist',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['acrtm_update_start'], {
        5: 'Option #1 and CRTM_Range_Invalid set; '
           'or Option #0 and Recovery_Mode set; '
           'or Option #1 and Auto_Measure_Activate cleared; '
           'or Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['acrtm_update_finish'], {
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['crtm_register_read'], {
        2: 'CRTM_Register_Address out of range (> 0xFFF); '
           'or volatile read to other than CRTM_Addr & CRTM_Size out of range',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['crtm_register_write'], {
        2: 'CRTM_Update_Start was not given when writing to nonvolatile registers '
           '(CRTM_Range_Invalid bit cleared); '
           'or CRTM_Register_Address out of range (> 0xFFF); '
           'or address out of range for a volatile write; ',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['flash_extend'], {
        4: 'CRTM_Range_Invalid or Recovery_Mode is set',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['replace_srk'], {
        2: 'First new customer key, ECDSA must be first; '
           'or Option #2/#3, device is not in RioT Mode or access to the active '
           'UDS is not allowed',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['manage_kts'], {
        2: 'First key uploaded to KTS, ECDSA must be first',
        4: 'Command signing key does not exist or is not valid',
        5: 'Flash program failure or other error condition programming internal '
           'flash bit flags, AMTC0, etc.'
    }))
    error_map.update(dict.fromkeys(['measure'], {
        2: 'Data frame count does not match DataProvided; '
           'or Partition_AltSource_Addr, Partition_Dest_Addr, or '
           'Partition_KeySource_Addr is outside of active partition',
        5: 'Command option relative to Riot Mode or prior events is not allowed; '
           'or Option #0 and Recovery_Mode bit set; '
           'or Option #1/#2, UDS is not accessible; '
           'or Option #1/#2, Config[1:0] == `11b; '
           'or Command output to flash is write protected'
    }))
    error_map.update(dict.fromkeys(['pcr_extend'], {
        2: 'PCRAddr does not point to a valid PCR'
    }))
    error_map.update(dict.fromkeys(['create_signature', 'verify_signature'], {
        2: 'Partition_Dest_Addr, Partition_Source_Addr, Partition_KeySource_Addr, '
           'or Partition_Message_Addr outside active partition',
        4: 'Command signing key does not exist; '
           'or ECDSA_DID_PrK does not exist',
        5: 'Command authorizer does not have access to signing or verification key; '
           'or SignatureDestination is to write-protected WPG or Partition_Message_Addr '
           'is unreadable read-protected WPG'
    }))
    error_map.update(dict.fromkeys(['request_monotonic_counter'], {
        2: 'A command signature should not be included when requesting AMTC0 or AMTC1'
    }))
