import logging
from logging import error

from random import _urandom as nonce
from typing import NoReturn, Tuple
from collections import namedtuple
from hashlib import sha256
from ..endpoint_platform import PlatformReturnType, EpPlatformRetCode
from .cmd_error_code import IndigoCmdErrorCode
from ..authenta_device import AuthentaDevice, DeviceType
from .defs_h import *

import authenta.authenta_key_lib_interface as authenta_key_lib


class IndigoDevice(AuthentaDevice):
    _INDIGO_SIGNED_COMMAND_OFFSET_IDX = (224)

    def __init__(self, endpoint_platform, logger=logging.getLogger(__name__), user_defined_key_driver_data=None, user_defined_key_driver_dev_info_dict=None):
        super().__init__(endpoint_platform, logger, user_defined_key_driver_data, user_defined_key_driver_dev_info_dict)
        self.device_type = DeviceType.INDIGO
        self.initialize_internal_properties()
        if self._key_driver_dev_info_dict is None:
            self._key_driver_dev_info_dict = { }
        self._key_driver_dev_info_dict['device_type'] = 'INDIGO'
        self._key_driver_dev_info_dict['device_version'] = '1.0'


    def initialize_internal_properties(self):
        super().initialize_internal_properties()
        self._amtc0 = None
        self._boot_amtc0 = None
        self._amtc1 = None
        self._provision_sts_struct = None
        self._dev_ecdsa_public_key = None
        self._ecdh_public_key = None
        self._uid = None
        self._revision = None
        self._extended_crtm_status_struct = None
        self._device_config_volatile_struct = None
        self._device_config_struct = None
        self._authenta_initial_regs_parsed = None
        self._kts_hmac_signing_key_slots = [ None, None, None, None, None, None, None, None ]
        self._kts_ecdsa_signing_key_slots = [ None, None, None, None, None, None, None, None ]
        self._client_keys_set = False
        self._ecdsa_signing_key_set = False
        self._ecdh_session_key_set = False
 
    @property
    def revision(self):
        return self._revision

    @property
    def amtc0(self):
        return self._amtc0

    @property
    def boot_amtc0(self):
        return self._boot_amtc0

    @property
    def extended_crtm_status_struct(self):
        return self._extended_crtm_status_struct

    @property
    def provision_sts_struct(self):
        return self._provision_sts_struct

    @property
    def device_config_struct(self):
        return self._device_config_struct

    @property
    def device_config_volatile_struct(self):
        return self._device_config_volatile_struct

    """
    Meta Functions
    """

    def meta_request_uid(self, config: bytes = b"\x00\x00", nonce: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get UID from Indigo Device and check response received

            Send atomic_request_uid command
            Check value returned: ESR status, nonce and signature values+

            Args:
                config: specifies the command’s functionality.
                        Default to a all bits clear since the 
                        TRD specifies this as the required value

                nonce: Optional A 16-bytes nonce to use in the Request_UID
                       command sent to the Naxos device. If not provided, one will be
                       randomely generated.

            Returns:
                error_code: IndigoCmdErrorCode
                response: Command response bytes, None if no response received

            Raises:
                TypeError: If config is not of type bytes
        """
        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        # Atomic Request UID
        response, sent_nonce = self.atomic_request_uid(config, nonce)
        if(response.status != EpPlatformRetCode.EP_OK):
            return IndigoCmdErrorCode(b'\x00\x00', 'request_uid', 'request_uid command failed'), None

        # parse response. 
        request_uid_parsed_response = self._request_uid_response_bytes(response.buffer)

        # Note that a config value indicating the response is 
        # signed using ECDSA is used, since that is always true for request_uid
        error_code = self._handle_request_uid_provision_response('request_uid', response.buffer, sent_nonce, True)

        # test to see if command returned good status
        if error_code.esr_success:
            # parse provision_sts bytes
            provision_sts_lst = bin(int.from_bytes(request_uid_parsed_response['Device Status'], 'big')).lstrip('0b')
            provision_sts_lst = ((16 - len(provision_sts_lst))*'0') + provision_sts_lst
            provision_sts_lst_int = [int(d) for d in provision_sts_lst]
            self._provision_sts_struct = self.PROVISION_STS_STRUCT(*provision_sts_lst_int)

            self._uid = request_uid_parsed_response['Data']['UID']
            self._revision = request_uid_parsed_response['Data']['Revision']
            self._dev_ecdsa_public_key = request_uid_parsed_response['Data']['ECDSA_DS_PuK'][::-1]
            # if error_code.signature_matched and resp_ecdsa_pub != self._dev_ecdsa_public_key:
            #     self.logger.error('ECDSA_DS_PuK returned from request_uid does NOT match device specific ECDSA public key in key store (yet request_uid response sig was verified)!')

            self.logger.debug("==== Request_UID ====")
            self.logger.debug("UID:      {}".format(self._uid.hex()))
            self.logger.debug("Revision: 0x{}".format(self._revision.hex()))
            self.logger.debug("Provision_sts:")
            self.logger.debug("    Provisioned:          {}".format(self.provision_sts_struct.Provisioned))
            self.logger.debug("    SRK_Replaced:         {}".format(self.provision_sts_struct.SRK_Replaced))
            self.logger.debug("    Remotely_Provisioned: {}".format(self.provision_sts_struct.Remotely_Provisioned))
            self.logger.debug("    HSM_Present:          {}".format(self.provision_sts_struct.HSM_Present))
            self.logger.debug("    RIOT_Mode:            {}".format(self.provision_sts_struct.RIOT_Mode))
            self.logger.debug("    UDS_Replaced:         {}".format(self.provision_sts_struct.UDS_Replaced))
            self.logger.debug("    UDS_Disabled:         {}".format(self.provision_sts_struct.UDS_Disabled))
            self.logger.debug("    Flash_PCR0_Reset:     {}".format(self.provision_sts_struct.Flash_PCR0_Reset))
            self.logger.debug("    Server_Deprovision:   {}".format(self.provision_sts_struct.Server_Deprovision))
            self.logger.debug("    PCR_Reset:            {}".format(self.provision_sts_struct.PCR_Reset))
            self.logger.debug("    Deprovision_Start:    {}".format(self.provision_sts_struct.Deprovision_Start))
            self.logger.debug("ECDSA_DS_PuK (from resp): {}".format(self._dev_ecdsa_public_key.hex()))
            
        return error_code, response.buffer

    def meta_acrtm_update_start(self, config: bytes, amtc0_in: bytes = None)\
                                -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            A-CRTM Update Start - initiates the start of virtual CRTM registers or Recovery Partition updates

            Checks for info needed to call the command
            Sends atomic_acrtm_update_start command

            Args:
                config: specifies the command’s functionality
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if amtc0_in is not None and not isinstance(amtc0_in, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('amtc0_in', type(amtc0_in)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_start', "Missing required data to run command: config value"), None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_acrtm_update_start', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        sign_w_session_key = self._is_hmac_sig(config)
        if sign_w_session_key:
            if self.provision_sts_struct is None:
                error_code, _ = self.meta_request_uid()
                if not error_code.esr_success:
                    self.logger.error('Failed attempt to retrieve provision \
                        status by calling request_uid on the device')
                    return error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        if not self._prepare_signing_key(config):
            return IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_start', "Failed to find or generate signing key!"), None

        # Atomic A-CRTM Update Start
        response = self.atomic_acrtm_update_start(config, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            return IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_start', 'acrtm_update_start command failed'), None

        error_code = self._meta_handle_atomic_command_response("acrtm_update_start", config, response.buffer, True)

        return error_code, response.buffer

    def meta_acrtm_update_finish(self, config: bytes, amtc0_in: bytes = None)\
                                 -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get information from Indigo Device and check response received

            Send atomic_acrtm_update_finish command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If amtc0_in is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_finish', "Missing required data to run command: config value"), None

        # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_acrtm_update_finish', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        if not self._prepare_signing_key(config):
            return IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_finish', "Failed to find or generate signing key!"), None

        # Atomic ACRTM Update Finish
        response = self.atomic_acrtm_update_finish(config, amtc0_in)

        if response.status != EpPlatformRetCode.EP_OK:
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'acrtm_update_finish', 'acrtm_update_finish command failed')
            return error_code, None

        response_bytes = response.buffer

        #read response
        error_code = self._meta_handle_atomic_command_response('acrtm_update_finish', config, response_bytes, True)

        return error_code, response_bytes

    def meta_replace_srk(self, config: bytes, public_key: bytes, nonce: bytes = None, amtc0_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Send command to Replace Device SRK and check response received

            Send atomic_replace_srk command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                public_key: new ECDSA or ECDH uncompressed public key to be programmed in the
                            targeted public key table slot.
                nonce:  Optional A 16-byte nonce to use in the Provision_CRTM
                        command sent sent to the Naxos device. If not provided, one will
                        be randomly generated.
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error-code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)
                sent_nonce: nonce sent to  teh device with the call to Repace_SRK
                active_key_updated: boolean indicating if the active key or keys has/have been changed
                                 there are 3 options:
                                    None:  active keys have not been changed (don't change active key in the key library)
                                    False: the active key that was modified in this command has been changed
                                           (change that key in the key libray - if this cmd used ECDH signing, and the
                                           ECDH key changed, the key libray must be updated before calling 
                                           verify_replace_srk_response)
                                    True:  both ECDSA and ECDH keys active keys may have been changed (change both in
                                           the key library, if this cmd used ECDH signing, and the
                                           ECDH key changed, the key libray must be updated before calling 
                                           verify_replace_srk_response)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If public_key is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(public_key, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('public_key', type(public_key)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'replace_srk', "Missing required data to run command: config"), None, None, None

        if public_key is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'replace_srk', "Missing required data to run command: public_key"), None, None, None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_replace_srk', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None, None, None

        # make sure the signing key is available before calling the atomic function to send the command
        self._prepare_signing_key(config)

        orig_srk_replaced = self._provision_sts_struct.SRK_Replaced

        response, sent_nonce = self.atomic_replace_srk(config, public_key, nonce, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'replace_srk', 'replace_srk command failed')
            return error_code, None, sent_nonce, None

        response_bytes = response.buffer

        parsed_response = self._parse_response_bytes(response_bytes)

        ESR = parsed_response['ESR']

        error_code = IndigoCmdErrorCode(ESR, 'replace_srk')
        returned_nonce = parsed_response['Nonce']
        nonce_matched = (sent_nonce == returned_nonce)
        if not nonce_matched:
            self.logger.error("Response validation failed nonce matching!!! sent nonce= 0x{}; returned nonce={}".format(sent_nonce, returned_nonce))
        error_code.set_nonce_matched(nonce_matched)
        error_code.set_valid_response(error_code.esr_success and nonce_matched)

        use_updated_key = False
        if error_code.valid_response:
            cmd_option = None
            ecdh_changed = (config[1] & 0x04) == 0 # bit 2 is clear
            cmd_option = (config[1] & 0x03)

            if cmd_option == 3 or cmd_option == 2 or (orig_srk_replaced == 0 and cmd_option == 1) or (orig_srk_replaced == 0 and not ecdh_changed):
                # the following cases don't cause a change in active keys
                # 1. change UDS
                # 2. revert while Micron factory keys were already in use
                # 3. change ECDSA while Micron factory ECDH is in use
                use_updated_key = None
            elif orig_srk_replaced == 0 and ecdh_changed:
                # if ecdh changes while factory keys are in use can cause active EDSA and ECDH to change
                # note changing ECDH before ECDSA while factory keys are in use will return a device error
                use_updated_key = True

        return error_code, response_bytes, sent_nonce, use_updated_key

    def verify_replace_srk_response(self, config: bytes, response: bytes, sent_nonce: bytes, keys_updated: bool):
        """
            Verify the response to a call to replace_srk

            This verification is typically done as part of the meta_XXX  
            command, but in the case of replace_srk is done seperately.

            This gives the application an oppportuninty to change the
            keys in the key library if necessary.

            Handle atomic command response

            Args:
                command_name: atomic commmand name
                conifg:       the config avalue sent with the command
                response:     the command response to be verified
                sent_nonce:   the nonce sent with the command
 
            Returns:
                error_code: IndigoCmdErrorCode
        """
        replace_srk_parsed_response = self._provision_crtm_response_bytes(response)
        ecdh_public_key = replace_srk_parsed_response["Data"]["ECDH_PV_PuK"]
        
        if self._is_hmac_sig(config) and keys_updated is not None:
            # load new hmac key
            pwrOnCount = replace_srk_parsed_response['Monotonic Count'][2:6] + b'\x00\x00'
            rc = authenta_key_lib.derive_ecdh_session_key(self._keylib_context, ecdh_public_key, pwrOnCount, True)
            if not rc:
                self.logger.error('Attempt to set ECDH session key failed!')

        ret_code = self._meta_handle_atomic_command_response('replace_srk', config, response, True, sent_nonce)

        if ret_code.valid_response:
            self._ecdh_public_key = ecdh_public_key
            ecdh_changed = (config[1] & 0x04) == 0 # bit 2 is clear

            # get the boot_amtc0 so the session key can be derived in subsequent commands
            read_status_err, _ = self._read_extended_crtm_status()
            if not read_status_err.esr_success:
                self.logger.info('Failed attempting to read extended crtm status after successful replace_srk')
                self.logger.info('_read_extended_crtm_status error message: ' + read_status_err.error_msg)

            # for keys_updated None means no keys changed
            #                  False means the new modified key is active
            #                  True means both keys may have changed (only possible for ECDH change)
            if keys_updated is not None:
                if ecdh_changed:
                    # if the _ecdh_public_key is None then we know the HMAC session key
                    # has not been generated, so no need to worry about it being stale,
                    # so don't bother updating it.
                    if self._ecdh_public_key is not None:
                        rc = authenta_key_lib.derive_ecdh_session_key(self._keylib_context, self._ecdh_public_key, self._boot_amtc0)
                        if not rc:
                            self.logger.error('Attempt to set ECDH session key failed!')
                            
                    if keys_updated:
                        # the ecdsa signing key may have been updated even if this command
                        # changed the ECDH key, so let's reset it
                        rc = authenta_key_lib.set_ecdsa_signing_key(self._keylib_context, True)
                        if rc:
                            self._ecdsa_signing_key_set = True
                        else:
                            self.logger.error('Attempt to set ECDSA key failed!')
                else:
                    rc = authenta_key_lib.set_ecdsa_signing_key(self._keylib_context, True)
                    if rc:
                        self._ecdsa_signing_key_set = True
                    else:
                        self.logger.error('Attempt to set ECDSA key failed!')

        elif keys_updated is not None:
            authenta_key_lib.key_lib_clear_keys(self._keylib_context)
        
        return ret_code

    def meta_crtm_register_write(self, config: bytes, crtm_addr: int, data: bytes, mask: bytes,
                                 amtc0_in: bytes = None)->Tuple[IndigoCmdErrorCode, bytes]:
        """
            Write the bit-alterable CRTM registers

            Checks for info needed to call the command
            Send atomic_crtm_register_write command

            Args:
                config: specifies the command’s functionality
                crtm_addr: address to write to in the virtual CRTM register space
                data: data to write to the virtual CRTM register space
                mask: mask to write to the virtual CRTM register space
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If crtm_addr is not of type int
                TypeError: If data is not of type bytes
                TypeError: If mask is not of type bytes
                TypeError: If amtc0_in is not of type bytes
                TypeError: If sign is not of type bool
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(crtm_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('crtm_addr', type(crtm_addr)))

        if not isinstance(data, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('data', type(data)))

        if not isinstance(mask, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('mask', type(mask)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_write', \
                                          "Missing required data to run command: config value"), None

        if crtm_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_write', \
                                          "Missing required data to run command: crtm_addr"), None

        if data is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_write', \
                                          "Missing required data to run command: data to write"), None

        if mask is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_write', \
                                          "Missing required data to run command: mask"), None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_crtm_register_write', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        # Get the signing key even if command is not signed (i.e. even if it is a volatile write)
        # because it will still need to be verified
        self._prepare_signing_key(config)

        error_code = None

        self.logger.debug('config=0x{}; crtm_addr=0x{:x}; data: {}; mask: {}'.format(config.hex(), crtm_addr, data.hex(), mask.hex()))

        response = self.atomic_crtm_register_write(config, crtm_addr, data, mask, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_write', 'crtm_register_write command failed')
            return error_code, None

        response_bytes = response.buffer

        #read response
        error_code = self._meta_handle_atomic_command_response('crtm_register_write', config, response_bytes, True)

        return error_code, response_bytes

    def meta_flash_extend(self, config: bytes, amtc0_in: bytes = None)\
                          -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Calculates the CRTM digest and either extend it into Flash_PCR0 or directly write
            it into Flash_DigestNV ro Flash_DigestNV_Partial

            Checks for info needed to call the command
            Send atomic_flash_extend command

            Args:
                config: specifies the command’s functionality
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)
        """
        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'flash_extend',
                                        "Missing requried data to run command: config value"), None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_flash_extend', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        self._prepare_signing_key(config)
        
        # sign the command for Option #1 or #3 (Options #0 and #2 are not signed)
        if(config[1] & 1 == 1):
            sign = True
        else:
            sign = False

        response = self.atomic_flash_extend(config, sign, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'flash_extend', 'flash_extend command failed')
            return error_code, None

        response_bytes = response.buffer

        error_code = self._meta_handle_atomic_command_response("flash_extend", config, response_bytes, True)

        return error_code, response_bytes

    def meta_crtm_register_read(self, config: bytes, crtm_addr: int, verify_resp_sig: bool=True, amtc0_in: bytes=None)\
                                -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Read 256 aligned bytes of the virtual CRTM registers

            Checks for info needed to call the command
            Send atomic_crtm_register_read command
            Check value returned: ESR status and crtm_addr

            Args:
                config: specifies the command’s functionality
                crtm_addr: address to be read from the virtual CRTM register space
                verify_resp_sig: In some cases a key won't be available until after register_read is done
                                 if this flag is False, don't verify the signature on the response here
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If crtm_addr is not of type int
                TypeError: If amtc0_in is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(crtm_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('crtm_addr', type(config)))

        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_read', "Missing required data to run command config value")
        if crtm_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_read', "Missing required data to run command: CRTM Address")

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_crtm_register_read', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        self.logger.debug('config=0x{}; crtm_addr=0x{:x}'.format(config.hex(), crtm_addr))

        # register_read is not a signed command, but we sill need to prepare key to verify response
        self._prepare_signing_key(config)

        response = self.atomic_crtm_register_read(config, crtm_addr, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'crtm_register_read', 'crtm_register_read command failed')
            return error_code, None

        response_bytes = response.buffer

        #read response
        error_code = self._meta_handle_atomic_command_response('crtm_register_read', config, response_bytes, verify_resp_sig)

        if not verify_resp_sig:
            return error_code, response_bytes

        # test to see if command returned good status
        if error_code._valid_response:
            crtm_addr_mask = crtm_addr & 0xFFFFFF00

            # if signature_checks and last_auth_cmd_matches:
            # check if this is the register holding boot_amtc0
            if crtm_addr_mask == CRTM_ADDRESS_BLOCK_INITIAL_REGS:
                # set boot_amtc0
                reg_0_parsed_response = self._crtm_register_0_read_parsed_response(response_bytes)
                self._boot_amtc0 = reg_0_parsed_response['Data']['PwrOnCountNV'] # 4 bytes
                self._boot_amtc0 = self._boot_amtc0 + b'\x00\x00'  # the 4 high order bytes of 6 are returned 
                self._amtc1 = reg_0_parsed_response['Data']['A_MTC1'] # 4 bytes
                # only provide the high 32 bits non-volatile value
                self._amtc1 = self._amtc1[2:4] + b'\x00\x00'
                self._authenta_initial_regs_parsed = reg_0_parsed_response

                self._set_device_config(reg_0_parsed_response['Data']['Device_Config'])
                self._set_device_config_volatile(reg_0_parsed_response['Data']['Device_Config_Volatile'])

                self.logger.debug("==== CRTM_Register_Read Address 0x{:02x} ====".format(crtm_addr_mask))
                self.logger.debug("PowerOnCnt: 0x{}".format(self._boot_amtc0.hex()))
                self.logger.debug("AMTC-1: 0x{}".format(self._amtc1.hex()))
                self.logger.debug("Device_Config Reg: 0x{}".format(reg_0_parsed_response['Data']['Device_Config'].hex())) 
                self.logger.debug(self.DEVICE_CONFIG_FORM.format(**self.device_config_struct._asdict()))
                self.logger.debug("Device_Config_Volatile Reg: 0x{}".format(reg_0_parsed_response['Data']['Device_Config_Volatile'].hex())) 
                self.logger.debug(self.DEVICE_CONFIG_VOLATILE_FORM.format(**self.device_config_volatile_struct._asdict()))
            else:
                if crtm_addr_mask == CRTM_ADDRESS_BLOCK_PUB_KEYS:
                    # this address holds ECDSA public key which is used to derive the session key - let's store it
                    generic_resp = self._parse_response_bytes(response_bytes)
                    resp_data = generic_resp['Data'][::-1]
                    resp_data = {
                        'ECDSA_DS_PuK': resp_data[0:64],
                        'ECDH_PV_PuK': resp_data[64:128],
                        'ECDSA_DID_PuK': resp_data[128:192],
                        'Reserved': resp_data[192:256]
                    }
                    self._ecdh_public_key = resp_data['ECDH_PV_PuK']
                    # resp_ecdsa_pub = resp_data['ECDSA_DS_PuK']
                    # if resp_ecdsa_pub != self._dev_ecdsa_public_key:
                    #     self.logger.error('ECDSA_DS_PuK returned from crtm_register_read does NOT match device specific ECDSA public key in key store (yet request_id response sig was verified)!')

                    self._ecdsa_did_public_key = resp_data['ECDSA_DID_PuK']
                    self._dev_ecdsa_public_key = resp_data['ECDSA_DS_PuK']

                    self.logger.debug("==== CRTM_Register_Read Address 0x{:02x} ====".format(crtm_addr_mask))
                    self.logger.debug("ECDSA_DS_PuK: {}".format(self._dev_ecdsa_public_key.hex()))
                    self.logger.debug("ECDH_PV_PuK: {}".format(self._ecdh_public_key.hex()))
                    self.logger.debug("ECDSA_DID_PuK: {}".format(self._ecdsa_did_public_key.hex()))

        return error_code, response_bytes

    def meta_pcr_extend(self, config: bytes, address: int, input_digest: bytes, amtc0_in: bytes = None)\
                        -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get information from Indigo Device and check response received

            Send atomic_pcr_extend command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                address: specifies which PCR to extend
                input_digest: used to extend with the referenced PCR
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If amtc0_in is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(address, int):
            raise TypeError('{} should be of type int and not {}'.format('address', type(address)))

        if not isinstance(input_digest, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('input_digest', type(input_digest)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'pcr_extend', "Missing required data to run command: config value"), None
        if address is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'pcr_extend', "Missing required data to run command: address"), None
        if input_digest is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'pcr_extend', "Missing required data to run command: data to write"), None

        if len(input_digest) != 32:
            raise ValueError("input_digest is not 32 bytes")

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_pcr_extend', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        self._prepare_signing_key(config)

        response = self.atomic_pcr_extend(config, address, input_digest, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'pcr_extend', 'pcr_extend command failed')
            return error_code, None

        response_bytes = response.buffer

        #read response
        error_code = self._meta_handle_atomic_command_response('pcr_extend', config, response_bytes, True)

        return error_code, response_bytes

    def meta_manage_kts(self, config: bytes, byte_addr: int, input_key: bytes, amtc0_in: bytes = None)\
                        -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Send command to Manage KTS and check response received

            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                input_key: new ECDSA or ECDH uncompressed public key or ECDSA private key to be
                programmed in the targeted public key table slot
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error-code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If byte_addr is not of type int
                TypeError: If input_key is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(byte_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('byte_addr', type(byte_addr)))

        if not isinstance(input_key, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('input_key', type(input_key)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'manage_kts', "Missing required data to run command: config"), None

        if byte_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'manage_kts', "Missing required data to run command: byte_addr"), None

        if input_key is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'manage_kts', "Missing required data to run command: input_key"), None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_manage_kts', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        if not self._get_signing_key_with_kts_option(config, byte_addr):
            return IndigoCmdErrorCode(b'\x00\x00', 'manage_kts', "Failed to find signing key! Aborting."), None

        response, sent_nonce = self.atomic_manage_kts(config, byte_addr, input_key, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'manage_kts', 'manage_kts command failed')
            return error_code, None

        response_bytes = response.buffer

        # manage_kts_parsed_response = self._manage_kts_response_bytes(response_bytes)

        # TODO perhaps, if this is Manage_KTS option #1, the returned public key should 
        #      be added to the KEY LIBRARY?

        if self._is_keyRef_bit_set(config):          
		    # this command may have updated the ECDSA PUK, so fetch it again
            ruid_rc, _  = self.meta_request_uid()
            if not ruid_rc.valid_response:
                self.logger.error("request_uid FAILED after executing Manage_KTS command! This and subsequent functions should be suspect!")

        #read response
        error_code = self._meta_handle_atomic_command_response('manage_kts', config, response_bytes, True, sent_nonce)

        return error_code, response_bytes

    def meta_measure(self, config: bytes, partition_alt_src_addr: int, partition_key_src_addr: int,
                     partition_dest_addr: int, input_digest: bytes = None,
                     amtc0_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get information from Indigo Device and check response received

            Send atomic_measure command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                partition_alt_src_addr: address in the active partition that may be used as data 
                    to include into the command measurement
                partition_key_src_addr: flash at the 512byte base address in the active partition 
                    that may be used as the source for a private signing key
                partition_dest_addr: address within the active partition that may be used to store 
                    the signature result of the command
                input_digest: data that can be provided by the host as an input message to sign
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If partition_alt_src_addr is not of type int
                TypeError: If partition_key_src_addr is not of type int
                TypeError: If partition_dest_addr is not of type int
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(partition_alt_src_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_alt_src_addr', type(partition_alt_src_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(partition_dest_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_dest_addr', type(partition_dest_addr)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'measure', "Missing required data to run command: config"), None
        if partition_alt_src_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'measure', "Missing required data to run command: partition_src_addr"), None
        if partition_key_src_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'measure', "Missing required data to run command: partition_key_addr"), None
        if partition_dest_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'measure', "Missing required data to run command: partition_dest_addr"), None

        if input_digest is not None:
            if len(input_digest) != 32:
                raise ValueError("input_digest is not 32 bytes")
        else:
            input_digest = b"\x00" * 32

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_measure', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        # measure command signature is only be present if MeasureDestination is set
        # but we still need to call preparing key to verify response
        self._prepare_signing_key(config)

        response = self.atomic_measure(config, partition_alt_src_addr, partition_key_src_addr, partition_dest_addr, input_digest, amtc0_in)
        
        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'measure', 'measure command failed')
            return error_code, None

        response_bytes = response.buffer

        error_code = self._meta_handle_atomic_command_response('measure', config, response_bytes, True)

        return error_code, response_bytes

    def meta_provision_crtm(self, config: bytes, nonce: bytes=None, amtc0_in: bytes=None)\
                            -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Provision_CRTM - activate the A-CRTM features in the Indigo device

            Checks for info needed to call the command
            Sends atomic_provision_crtm command

            Args:
                config: specifies the command’s functionality
                nonce:  Optional A 16-byte nonce to use in the Provision_CRTM
                        command sent sent to the Naxos device. If not provided, one will
                        be randomly generated.
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)
        """       

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'provision_crtm', "Missing \
                                          required data to run command: config value"), None

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            error_code = self._get_amtc0_if_not_set('meta_provision_crtm', amtc0_in)
            if not error_code.esr_success:
                return error_code, None

        response, sent_nonce = self.atomic_provision_crtm(config, nonce, amtc0_in)

        if response.status is not EpPlatformRetCode.EP_OK:
            return IndigoCmdErrorCode(b'\x00\x00', 'meta_provision_crtm'), None

        response_bytes = response.buffer

        error_code = self._handle_request_uid_provision_response('provision_crtm', response_bytes, sent_nonce, True)

        if error_code._valid_response:
            provision_crtm_parsed_response = self._provision_crtm_response_bytes(response_bytes)
            self._ecdh_public_key = provision_crtm_parsed_response["Data"]["ECDH_PV_PuK"]
            self._provision_sts_struct = None  # reset provision status because the previous value is no longer valid
            self._set_extended_crtm_status(provision_crtm_parsed_response['Device Status'])

        return error_code, response_bytes

    def meta_deprovision_crtm(self, config: bytes, amtc1_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Deprovision_CRTM - deactivate the A-CRTM features in the Indigo device and destroy
            certain keys/data when the device is provisioned

            Checks for info needed to call the command
            Sends atomic_deprovision_crtm command

            Args:
                config: specifies the command’s functionality
                amtc1_in: specifies the current A-MTC1 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)
        """

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'deprovision_crtm', "Missing \
                                          required data to run command: config value"), None

        if self.provision_sts_struct is None:
            error_code, _ = self.meta_request_uid()
            if not error_code.esr_success:
                self.logger.error('Failed attempt to retrieve provision \
                      status by calling request_uid on the device')
                return error_code, None

        if not self.provision_sts_struct.Provisioned:
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'deprovision_crtm', \
                                                'deprovision_crtm failed; device is not provisioned')
            return error_code, None

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        # check if boot_amtc0 and amtc1 have been retrieved, and if not, retrieve them by calling _read_extended_crtm_status
        if (self._amtc1 is None and amtc1_in is None) or self._boot_amtc0 is None:
            error_code, _ = self._read_extended_crtm_status()
            if error_code.esr_success:
                self.logger.info('Retrieved BOOT_AMTC0 value amd AMTC1 value')
            else:
                self.logger.error('Failed attempting to retrieve BOOT_AMTC0 from device')
                return error_code, None

        if self._provision_sts_struct.Remotely_Provisioned:
            self._prepare_signing_key(config)
        else:
            # There is an exception from the standard
            # response signature rules for deprovision; sometimes the
            # response is signed w/ a different key than the command...
            #
            # If the config[1] (Allow_SRK) is set the response is signed by ECDSA_DS_PrK,
            # otherwise it is signed by the session key
            if (config[1] & 0x02) == 0:
                # better get it (since in the case of deprovision
                # the command and response are signed w/ different keys)
                # and better get it before deprovisioning, because once
                # the device is deprovisioned, its too late to query the
                # device for the key
                self._prepare_signing_key(b'\x00\x00') # pass in a config indicating "sign-with-session"

        response = self.atomic_deprovision_crtm(config, amtc1_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'deprovision_crtm', 'deprovision_crtm command failed')
            return error_code, None

        response_bytes = response.buffer

        error_code = self._meta_handle_atomic_command_response('deprovision_crtm', config, response_bytes, True)

        if error_code.valid_response:
            # destroy and clear values #TODO
            self.initialize_internal_properties()

        return error_code, response_bytes

    def meta_apa_modify(self, config: bytes, wpg_addr: int, apa_data: bytes, key_slot: int,
                        amtc0_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Used to change and read the Authenta protection attributes at the chosen
            write protect group (WPG)

            Checks for info needed to call the command
            Sends atomic_apa_modify command

            Args:
                config: specifies the command’s functionality
                wpg_addr: address of the write protect group
                key_slot: if KeyRef config bit (config[14]) is not set, this 
                          argument is be ignored.
                          if KeyRef config bit is set, then this argument must 
                          hold a key slot in the public key table associated with
                          current value of PublicKeySlotPointer in the APA_Data 
                          for the input wpg_addr. It must be a valid key slot, between
                          3 and 10 inclusive.
                apa_data: a data sub-packet that can be used to directly replace the 
                          existing APA_Data structure for the chosen write protect group
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                error-code: IndigoCmdErrorCode
                response_bytes: PlatformReturnType (status and command response received from Indigo Device)
        """
        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(wpg_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('wpg_addr', type(wpg_addr)))

        if not isinstance(apa_data, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('apa_data', type(apa_data)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify',
                                        "Missing required data to run command: config value"), None
        
        if wpg_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify', 
                                        "Missing required data to run command: WPG address"), None

        if apa_data is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify',
                                        "Missing required data to run command: APA Data"), None
        if len(apa_data) != 80:
            return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify',
                                        "APA Data field should be size 80 but found {}".format(len(apa_data))), None
        

        if self._is_keyRef_bit_set(config):
            if key_slot is None or key_slot < 0 or key_slot > 7:
                return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify', \
                                          "invalid Public Key Table keySlot, must correspond to the "
                                          "PublicKeySlotPointer found in APA Data: found {}"
                                          ", must be between 3 and 10!".format(key_slot)), None

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_apa_modify', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None

        # make sure the signing key is available before calling the atomic function to send the command    
        if not self._get_signing_key_with_kts_option(config, key_slot):
            return IndigoCmdErrorCode(b'\x00\x00', 'apa_modify', "Failed to find signing key! Aborting."), None

        response = self.atomic_apa_modify(config, wpg_addr = wpg_addr, apa_data = apa_data, key_slot = key_slot, amtc0_in = amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'apa_modify', 'apa_modify command failed')
            return error_code, None

        response_bytes = response.buffer

        apa_modify_parsed_response = self._apa_modify_response_bytes(response_bytes)

        error_code = self._meta_handle_atomic_command_response("apa_modify", config, response_bytes, True)

        return error_code, response_bytes

    def meta_create_signature(self, config: bytes, byte_addr: int, partition_msg_addr: int,
                              partition_key_src_addr: int, partition_dest_addr: int, 
                              input_digest: bytes = None, amtc0_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get information from Indigo Device and check response received

            Send atomic_create_signature command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                partition_dest_addr: specifies the 512 byte base address within the active partition
                    that may be used to store the signature result of the CreateSignature command
                partition_key_src_addr: specifies flash at the 512 byte base address in the active
                    partition that may be used as the source for a private signing key
                partition_msg_addr: specifies the 512 byte base address in the active partition that
                    may be used as the source of a 32 byte message
                input_digest: 256 bytes of data that can be provided by the host as an input message
                    to sign
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If byte_addr is not of type int
                TypeError: If partition_dest_addr is not of type int
                TypeError: If partition_key_src_addr is not of type int
                TypeError: If partition_msg_addr is not of type int
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))
        
        if not isinstance(byte_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('byte_addr', type(byte_addr)))

        if not isinstance(partition_dest_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_dest_addr', type(partition_dest_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(partition_msg_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_msg_addr', type(partition_msg_addr)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Missing required data to run command: config"), None

        if partition_dest_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Missing required data to run command: partition_dest_addr"), None

        if partition_key_src_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Missing required data to run command: partition_key_src_addr"), None

        if partition_msg_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Missing required data to run command: partition_msg_addr"), None

        if input_digest is not None:
            if len(input_digest) != 32:
                raise ValueError("input_digest is not 32 bytes")
        else:
            input_digest = b"\x00" * 32

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_create_signature', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        if not self._get_signing_key_with_kts_option(config, byte_addr):
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Failed to find signing key! Aborting."), None

        response = self.atomic_create_signature(config, byte_addr, partition_msg_addr, partition_key_src_addr, partition_dest_addr, input_digest, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'create_signature', 'create_signature command failed')
            return error_code, None

        response_bytes = response.buffer

        error_code = self._meta_handle_atomic_command_response('create_signature', config, response_bytes, True)

        return error_code, response_bytes

    def meta_verify_signature(self, config: bytes, byte_addr: int, partition_msg_addr: int,
                              partition_key_src_addr: int, partition_src_addr: int,
                              input_digest: bytes = None, amtc0_in: bytes = None) -> Tuple[IndigoCmdErrorCode, bytes]:
        """
            Get information from Indigo Device and check response received

            Send atomic_create_signature command
            Send _meta_handle_atomic_command_response to Check values from response returned

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                partition_src_addr: specifies the 512 byte base address within the active partition
                    that may be used as the source of the signature to verify
                partition_key_src_addr: specifies flash at the 512 byte base address in the active
                    partition that may be used as the source for a public verification key
                partition_msg_addr: specifies the 512 byte base address in the active partition that
                    may be used as the source of a 32 byte message.
                input_digest: 256 bytes of data that can be provided by the host as an input message
                    to sign
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                error_code: IndigoCmdErrorCode
                response: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If partition_src_addr is not of type int
                TypeError: If partition_key_src_addr is not of type int
                TypeError: If partition_msg_addr is not of type int
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(partition_src_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_src_addr', type(partition_src_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(partition_msg_addr, int):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('partition_msg_addr', type(partition_msg_addr)))

        # make sure we have the info needed to call the function
        if config is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'verify_signature', "Missing required data to run command: config"), None

        if partition_src_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'verify_signature', "Missing required data to run command: partition_src_addr"), None

        if partition_key_src_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'verify_signature', "Missing required data to run command: partition_key_src_addr"), None

        if partition_msg_addr is None:
            return IndigoCmdErrorCode(b'\x00\x00', 'verify_signature', "Missing required data to run command: partition_msg_addr"), None

        if input_digest is not None:
            if len(input_digest) != 32:
                raise ValueError("input_digest is not 32 bytes")
        else:
            input_digest = b'\x00' * 32

        # check if amtc0 has been retrieved
        # if not, retrieve it by calling request_uid
        if amtc0_in is None:
            amtc0_error_code = self._get_amtc0_if_not_set('meta_verify_signature', amtc0_in)
            if not amtc0_error_code.esr_success:
                return amtc0_error_code, None

        # make sure the signing key is available before calling the atomic function to send the command
        if not self._get_signing_key_with_kts_option(config, byte_addr):
            return IndigoCmdErrorCode(b'\x00\x00', 'create_signature', "Failed to find signing key! Aborting."), None

        response = self.atomic_verify_signature(config, byte_addr, partition_msg_addr, partition_key_src_addr, partition_src_addr, input_digest, amtc0_in)

        if(response.status != EpPlatformRetCode.EP_OK):
            error_code = IndigoCmdErrorCode(b'\x00\x00', 'verify_signature', 'verify_signature command failed')
            return error_code, None

        response_bytes = response.buffer

        error_code = self._meta_handle_atomic_command_response('verify_signature', config, response_bytes, True)

        return error_code, response_bytes


    """
    Non-Authenta commands
    """

    def meta_read_multiple_block(self, address: int, length: int) -> Tuple[bool, bytes]:
        """
        Read_Multiple_Block (read data from flash)
        Return (True, response) on success or (False, None) on failure
        """

        length_len = len(bytes(length))

        if (length_len % 512) != 0:
            raise ValueError("Length size is not a multiple of 512")

        ret, response =  self.endpoint_platform.indigo_read_multiple_block(address, length)

        return ret, response

    def meta_write_multiple_block(self, address: int, data: bytes) -> bool:
        """
        Write_Multiple_Block (write data to flash)
        Return True on success or False on failure
        """
        
        data_len = len(data)

        if (data_len % 512) != 0:
            raise ValueError("Data size is not a multiple of 512")

        response = self.endpoint_platform.indigo_write_multiple_block(address, data)

        return response


    """
    Atomic Functions
    """

    def atomic_request_uid(self, config: bytes, nonce_in: bytes=None) -> Tuple[PlatformReturnType, bytes]:
        """
            Request_UID command implementation

            Creates the command string using the _generic_pack_command and send the command to request
            information specific to the device

            Args:
                config: specifies the command’s functionality

                nonce_in: Optional 16-byte bytes object holding a nonce to use
                    in the Request_UID command sent to the Naxos device. If no nonce
                    is provided, one will be generated in this function and returned.

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
                a_nonce: nonce sent with request command

            Raises:
                TypeError: If config is not of type bytes
        """

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if nonce_in is not None:
            if not isinstance(nonce_in, bytes):
                self.logger.error('input argument is of wrong type; nonce must be of type bytes!')
                raise TypeError('{} should be of type bytes and not {}'.format('nonce_in', type(nonce_in)))
            if len(nonce_in) != 16:
                self.logger.error('input argument is of wrong length; nonce must be 16 bytes but found {}!'.format(len(nonce_in)))
                raise TypeError('{} must be exactly 2 bytes, but found length {}'.format('nonce_in', len(nonce_in)))
            a_nonce = nonce_in
        else:
            a_nonce = nonce(16)

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x3E', config, a_nonce=a_nonce)

        # send command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

    def atomic_flash_extend(self, config: bytes, sign: bool, amtc0_in: bytes = None) -> PlatformReturnType:
        """
            Flash_Extend command implementation

            Creates the command string using the _generic_pack_command and sends the command

            Args:
                config: specifies the command’s functionality
                sign: indicates whether the command is to be signed
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0;\
                Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        cmd_str = self._generic_pack_command(b'\x9B\x3A', config, monotonic_counter = use_amtc0)

        use_hmac_session_key = self._is_hmac_sig(config)

        # sign the command - on config[1] is 0 this is non-signed
        sig = ((config[1] & 0x01) != 0) 
        if sig:
            cmd_str = self._sign_cmd(cmd_str, use_hmac_session_key)

        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_acrtm_update_start(self, config: bytes, amtc0_in: bytes = None) -> PlatformReturnType:
        """
            A-CRTM Update Start implementation

            Creates the command string using the _generic_pack_command and sends the command to
            initiate the start of virtual CRTM registers

            Args:
                config: specifies the command’s functionality
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                response: PlatfromReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0;\
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        use_hmac_session_key = self._is_hmac_sig(config)

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x3B', config, monotonic_counter = use_amtc0)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, use_hmac_session_key)
        
        # send the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_acrtm_update_finish(self, config: bytes, amtc0_in: bytes) -> PlatformReturnType:
        """
            ACRTM_Update_Finish command implementation

            Creates the command string using the _generic_pack_command and send the command
            Indicate either the finish of CRTM register updates or
            the finish of Recovery Partition updates or Recovery Mode

            Args:
                config: specifies the command’s functionality
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If amtc0_in is not of type bytes
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        use_hmac_session_key = self._is_hmac_sig(config)

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x3D', config, monotonic_counter = use_amtc0)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, use_hmac_session_key)

        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_crtm_register_write(self, config: bytes, crtm_addr: int, data: bytes, mask: bytes,
                                   amtc0_in: bytes = None)->PlatformReturnType:
        """
            CRTM_Register_Write command implementation

            Write the bit-alterable CRTM registers. This command may be issued zero to n times and
            alter the volatile or nonvolatile state of select registers. The command operates on up
            to 128 byte addresses at a time, with individual bit mask control on each register
            location.

            Args:
                config: specifies the command’s functionality
                crtm_addr: address to be read from the virtual CRTM register space
                data: 128 bytes of data to write to the virtual CRTM register space
                mask: data mask to write to the virtual CRTM register space
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If crtm_addr is not of type int
                TypeError: If data is not of type bytes
                TypeError: If mask is not of type bytes
                TypeError: If amtc0_in is not of type bytes
                TypeError: If sign is not of type bool
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if (config is None):
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if (crtm_addr is None):
            self.logger.error("Missing required data to run command: CRTM Address")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if (data is None):
            self.logger.error("Missing required data to run command: data to write")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if (mask is None):
            self.logger.error("Missing required data to run command: mask")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if (use_amtc0 is None):
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(crtm_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('crtm_addr', type(crtm_addr)))

        if not isinstance(data, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('data', type(data)))

        if not isinstance(mask, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('mask', type(mask)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if (config[1] % 2 != 1):
            use_session_key = self._is_hmac_sig(config)
        
        data = self._create_data_buffer(data, crtm_addr)

        mask = self._create_mask_buffer(mask, crtm_addr)

        data = mask + data

        # pack command
        cmd_str = self._generic_pack_command(b'\x9B\x3C', config, address = crtm_addr.to_bytes(4, 'big'),
                                            data = data, monotonic_counter = use_amtc0)

        # if bit 0 of the config is set it indicates a volatile write if set - no signature needed
        if (config[1] % 2 != 1):
            # sign the command
            cmd_str = self._sign_cmd(cmd_str, use_session_key)

        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_crtm_register_read(self, config: bytes, crtm_addr: int,
                                  amtc0_in: bytes = None) -> PlatformReturnType:
        """
            CRTM_Register_Read command implementation

            Read 256 aligned bytes of the virtual CRTM registers

            Args:
                config: specifies the command’s functionality
                crtm_addr: address to be read from the virtual CRTM register space
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)

            Raises:
                TypeError: If config is not of type bytes
                TypeError: If crtm_addr is not of type int
                TypeError: If amtc0_in is not of type bytes
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if crtm_addr is None:
            self.logger.error("Missing required data to run command: CRTM Address")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(crtm_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('crtm_addr', type(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x38', config, address = crtm_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0)

        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_pcr_extend(self, config: bytes, address: int, input_digest: bytes,
                          amtc0_in: bytes = None) -> PlatformReturnType:
        """
            PCR Extend command implementation

            Used to modify one of eight selected general purpose PCR registers
            Creates the command string using the _generic_pack_command and send the command

            Args:
                config: specifies the command’s functionality
                address: specifies which PCR address to extend
                input_digest: used to extend with the referenced PCR
                amtc0_in: current AMTC0 value in the flash device

            Return:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if address is None:
            self.logger.error("Missing required data to run command: address value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if input_digest is None:
            self.logger.error("Missing required data to run command: input_digest value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(address, int):
            raise TypeError('{} should be of type int and not {}'.format('address', type(address)))

        if not isinstance(input_digest, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('input_digest', type(input_digest)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        use_session_key = self._is_hmac_sig(config)

        cmd_str = self._generic_pack_command(b'\x9B\x40', config, data=input_digest, address=address.to_bytes(4, 'big'),
                                            monotonic_counter=use_amtc0)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, use_session_key)

        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_provision_crtm(self, config: bytes, nonce_in: bytes = None, amtc0_in: bytes = None) \
                              -> Tuple[PlatformReturnType, bytes]:
        """
            Provision_CRTM implementation

            Creates the command string using the _generic_pack_command and sends the command to
            activate the A-CRTM features in the Indigo Device

            Args:
                config: specifies the command’s functionality
                nonce_in: Optional 16-byte bytes object holding a nonce to use
                    in the Request_UID command sent to the Naxos device. If no nonce
                    is provided, one will be generated in this function and returned.
                amtc0_in: specifies the current A-MTC0 value in the flash device

            Returns:
                response: PlatformReturnType (status and command response received from Indigo Device)
                a_nonce: nonce sent with request command
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if len(config) != 2:
            raise TypeError('{} must be exactly 2 bytes, but found length {}'.format('config', len(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))
        
        if nonce_in is not None:
            if not isinstance(nonce_in, bytes):
                self.logger.error('input argument is of wrong type; nonce must be of type bytes!')
                raise TypeError('{} should be of type bytes and not {}'.format('nonce_in', type(nonce_in)))
            if len(nonce_in) != 16: 
                raise TypeError('{} must be exactly 16 bytes, but found length {}'.format('nonce_in', len(nonce_in)))
            a_nonce = nonce_in
        else:
            a_nonce = nonce(16)

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x35', config, monotonic_counter = use_amtc0,
                                            a_nonce = a_nonce)

        # sign command using ECDSA signature if provision is remote provision
        # no signature is included for local provision (0x01 representing b'00000001')
        if (config[1] & 0x01) != 0:
            # remote provision: ECDSA signature with SRK
            authenta_key_lib.set_ecdsa_signing_key(self._keylib_context)
            authenta_key_lib.set_ecdsa_verification_key(self._keylib_context)

            cmd_str = self._sign_cmd(cmd_str, False) 

        # send the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

    def atomic_deprovision_crtm(self, config: bytes, amtc1_in: bytes = None) -> PlatformReturnType:
        """
            Deprovision_CRTM implementation

            Creates the command string using the _generic_pack_command and sends the command to
            deactivate the A-CRTM features in the Indigo Device

            Args:
                config: specifies the command’s functionality
                amtc1_in: specifies the current A-MTC1 value in the flash device

            Returns:
                response: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc1 for debug
        if amtc1_in:
            use_amtc1 = amtc1_in
        else:
            use_amtc1 = self._amtc1

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(use_amtc1, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc1', type(use_amtc1)))

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if not self.provision_sts_struct:
            # need this information to determine if the command must be signed
            self.logger.error("Missing required data to run command: device provision" \
                  "status; run _read_extended_crtm_status before deprovision_crtm")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)


        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x36', config, monotonic_counter=use_amtc1)

        # generate command signature
        if self.provision_sts_struct.Remotely_Provisioned:
            use_session_key = self._is_hmac_sig(config)

            # sign the command
            cmd_str = self._sign_cmd(cmd_str, use_session_key) 
        elif self.provision_sts_struct.HSM_Present: # if no HSM, deprovision is not signed
            # check config[1] bit, (Allow_SRK)...
            # if it is set, use SRK, otherwise use CRK
            if (config[1] & 0x02) == 0: 
                # Allow_SRK not set
                cmd_str = self._sign_cmd(cmd_str, True, False)
            else:
                # Allow_SRK is set
                if (config[0] & 0x80) == 0: 
                    # SigType (config[15]) is not set, invalid config
                    self.logger.error("Invalid config value; Config[15] not set (HMAC sig) and config[0] (Allow_SRK) is set!")
                    return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

                cmd_str = self._sign_cmd(cmd_str, False)
 
        # send the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_apa_modify(self, config: bytes, wpg_addr: int, apa_data: bytes,
                          key_slot: int, amtc0_in: bytes) -> PlatformReturnType:
        """
            APA_Modify command implementation

            Change and read the Authenta protection attributes of the chosen write protect group.

            Args:
                config: specifies the command’s functionality
                wpg_addr: address to be read from the virtual CRTM register space
                key_slot: if KeyRef config bit (config[14]) is not set, this 
                          argument will be ignored.
                          if KeyRef config bit is set, then this argument must 
                          hold a key slot in the public key table associated with
                          current value of PublicKeySlotPointer in the APA_Data 
                          for the input wpg_addr. It must be a valid key slot, between
                          3 and 10 inclusive.
                apa_data: data sub-packet that can be used to directly replace the existing
                amtc0_in: current AMTC0 value in the flash device
                    APA_Data structure

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(wpg_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('wpg_addr', type(wpg_addr)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(apa_data, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('apa_data', type(apa_data)))

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if wpg_addr is None:
            self.logger.error("Missing required data to run command: WPG Address")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if apa_data is None:
            self.logger.error("Missing required data to run command: APA Data")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)


        if self._is_keyRef_bit_set(config):
            if key_slot is None or key_slot < 3 or key_slot > 10:
                self.logger.error("Invalid Public Key Slot input!")
                return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        # make sure the signing key is set - if the signing using
        # HMAC then the ECDH Public Key from the device must
        # have been set before calling this function. Otherwise
        # it would have to be fetched now which would change the
        # AMTC0 value. So it's to late to fetch now.
        if not self._is_hmac_sig(config) or self._ecdh_public_key is not None:
           self._get_signing_key_with_kts_option(config, key_slot)
        use_kts_key = self._is_keyRef_bit_set(config)
        
        # pack command
        cmd_str = self._generic_pack_command(b'\x9B\x37', config, address = wpg_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0, data = apa_data)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, self._is_hmac_sig(config), use_kts_slot_key=use_kts_key)

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_replace_srk(self, config: bytes, public_key: bytes, nonce_in: bytes = None, amtc0_in: bytes = None)\
                           -> Tuple[PlatformReturnType, bytes]:
        """
            Replace SRK command implementation

            Used in a remotely provisioned device to change the host ECDSA and ECDH server public keys
            or the active Riot Unique Device Secret, PUF_UDS or customer UDS, when in Riot Mode

            Args:
                config: specifies the command’s functionality
                public_key: specifies the new ECDSA or ECDH uncompressed public key
                nonce_in: Optional 16-byte bytes object holding a nonce to use
                    in the Request_UID command sent to the Naxos device. If no nonce
                    is provided, one will be generated in this function and returned.
                amtc0_in: current AMTC0 value in the flash device

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
                a_nonce: nonce sent with request command
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(public_key, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format("public_key", type(public_key)))

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        if public_key is None:
            self.logger.error("Missing required data to run command: public_key")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        use_session_key = self._is_hmac_sig(config)

        if nonce_in is not None:
            if not isinstance(nonce_in, bytes):
                self.logger.error('input argument is of wrong type; nonce must be of type bytes!')
                raise TypeError('{} should be of type bytes and not {}'.format('nonce_in', type(nonce_in)))
            if len(nonce_in) != 16: 
                raise TypeError('{} must be exactly 16 bytes, but found length {}'.format('nonce_in', len(nonce_in)))
            a_nonce = nonce_in
        else:
            a_nonce = nonce(16)

        # pack command
        cmd_str = self._generic_pack_command(b'\x9B\x39', config, monotonic_counter = use_amtc0,
                                            a_nonce = a_nonce, data = public_key)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, use_session_key)

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

    def atomic_manage_kts(self, config: bytes, byte_addr: int, input_key: bytes,
                          amtc0_in: bytes = None) -> Tuple[PlatformReturnType, bytes]:
        """
            Manage KTS command implementation

            Manage the keys in Public Key Table slots #3 through #10

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                amtc0_in: current AMTC0 value in the flash device
                input_key: specifies the new ECDSA or ECDH uncompressed public key or ECDSA private
                    key to be programmed in the targeted public slot

            Returns:
                cmd_str: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(byte_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('byte_addr', type(byte_addr)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(input_key, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('input_key', type(input_key)))

        # make sure we ahve the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        if byte_addr is None:
            self.logger.error("Missing required data to run command: Byte Address")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        if input_key is None:
            self.logger.error("Missing required data to run command: Input Key")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None

        # make sure the signing key is set - if the signing using
        # HMAC then the ECDH Public Key from the device must
        # have been set before calling this function. Otherwise
        # it would have to be fetched now which would change the
        # AMTC0 value. So it's to late to fetch now.
        if not self._is_hmac_sig(config) or self._ecdh_public_key is not None:
           self._get_signing_key_with_kts_option(config, byte_addr)
        use_kts_key = self._is_keyRef_bit_set(config)

        a_nonce = nonce(16)

        # pack command
        cmd_str = self._generic_pack_command(b'\x9B\x34', config,
                                            address = byte_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0, a_nonce = a_nonce,
                                            data = input_key)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, self._is_hmac_sig(config), use_kts_slot_key=use_kts_key)

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

    def atomic_measure(self, config: bytes, partition_alt_src_addr: int,
                       partition_key_src_addr: int, partition_dest_addr: int,
                       input_digest: bytes = None, amtc0_in: bytes = None) -> PlatformReturnType:
        """
            Measure command implementation

            A general purpose Indigo service used to create a SHA-256 based digest or HMAC based on a
            variety of selectable inputs. When Indigo is configured in Riot Mode, this command may also
            be used to generate a Compound Device Identity key (HMAC_CDI) and the Asymmetric Device ID
            key pair.

            Args:
                config: specifies the command’s functionality
                partition_alt_src_addr: specifies the 512byte base address in the active partition
                    that may be used as data to include into the command measurement
                amtc0_in: specifies the current A-MTC0 value in the flash device
                partition_dest_addr: specifies the 512byte base address within the active partition
                    that may be used to store the signature result of the command
                partition_key_src_addr: specifies flash at the 512byte base address in the active
                    partition that may be used as the source for a private signing key
                input_digest: 256 bytes of data that can be provided by the host as an input message
                    to sign
                partition_message_addr: specifies the 512byte base address in the active partition
                    that may be used by the device owner or the flash WPG owner

            Returns:
                response: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(partition_alt_src_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_alt_src_addr', type(partition_alt_src_addr)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(partition_dest_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_dest_addr', type(partition_dest_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(input_digest, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('input_digest', type(input_digest)))

        # make sure we have the info we need to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_alt_src_addr is None:
            self.logger.error("Missing required data to run command: partition_alt_src_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; "
                  "Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)
        
        if partition_dest_addr is None:
            self.logger.error("Missing required data to run command: partition_dest_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_key_src_addr is None:
            self.logger.error("Missing required data to run command: partition_key_src_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if input_digest is None:
            self.logger.error("Missing required data to run command: input_digest")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        # pack data
        data = input_digest + partition_key_src_addr.to_bytes(4, 'big') + partition_dest_addr.to_bytes(4, 'big')

        # pack the command
        # TODO unsure how to handle extra data
        cmd_str = self._generic_pack_command(b'\x9B\x3F', config,
                                            address = partition_alt_src_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0, data = data)

        # sign the command if MeasureDestination (config[7]) is set
        if (config[1] & 0x80) != 0:
            cmd_str = self._sign_cmd(cmd_str, self._is_hmac_sig(config))

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_create_signature(self, config: bytes, byte_addr: int, partition_msg_addr: int,
                                partition_key_src_addr: int, partition_dest_addr: int,
                                input_digest: bytes, amtc0_in: bytes) -> PlatformReturnType:
        """
            Create Signature implementation

            Allows an authorized party (device owner or the holder of any key table slot's private key)
            to create an elliptic curve digital signature of a message using an ECDSA private key and
            also returns its public key

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                amtc0_in: specifies the current A-MTC0 value in the flash device
                partition_dest_addr: specifies the 512byte base address within the active partition
                    that may be used to store the signature result of the command
                partition_key_src_addr: specifies flash at the 512byte base address in the active
                    partition that may be used as the source for a private signing key
                input_digest: 256bytes of data that can be provided by the host as an input message
                    to sign
                partition_msg_addr: specifies the 512byte base address in the active partition
                    that may be used by the device owner or the flash WPG owner

            Returns:
                response: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(byte_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('byte_addr', type(byte_addr)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(partition_dest_addr, int):
            raise TypeError('{} shoule be of type int and not {}'
                            .format('partition_dest_addr', type(partition_dest_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(input_digest, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('input_digest', type(input_digest)))

        if not isinstance(partition_msg_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_msg_addr', type(partition_msg_addr)))

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if byte_addr is None:
            self.logger.error("Missing required data to run command: byte_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_dest_addr is None:
            self.logger.error("Missing required data to run command: partition_dest_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_key_src_addr is None:
            self.logger.error("Missing required data to run command: partition_key_src_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if input_digest is None:
            self.logger.error("Missing required data to run command: input_digest")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_msg_addr is None:
            self.logger.error("Missing required data to run command: partition_msg_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        # make sure the signing key is set - if the signing using
        # HMAC then the ECDH Public Key from the device must
        # have been set before calling this function. Otherwise
        # it would have to be fetched now which would change the
        # AMTC0 value. So it's to late to fetch now.
        if not self._is_hmac_sig(config) or self._ecdh_public_key is not None:
           self._get_signing_key_with_kts_option(config, byte_addr)
        use_kts_key = self._is_keyRef_bit_set(config)

        # pack the data
        data = partition_msg_addr.to_bytes(4, 'big') + input_digest + partition_key_src_addr.to_bytes(4, 'big') + partition_dest_addr.to_bytes(4, 'big')

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x43', config, address = byte_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0, data = data)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, self._is_hmac_sig(config), use_kts_slot_key=use_kts_key)

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

    def atomic_verify_signature(self, config: bytes, byte_addr: int, partition_msg_addr: int,
                                partition_key_src_addr: int, partition_src_addr: int,
                                input_digest: bytes = None, amtc0_in: bytes = None) -> PlatformReturnType:
        """
            Verify Signature implementation

            Allows an elliptic curve digital signature to be verified. The signature being verified,
            the message, and the public key must be in readable flash (if flash is specified as a
            source for any of these parameters)

            Args:
                config: specifies the command’s functionality
                byte_addr: indicates which key table slot is being addressed
                amtc0_in: specifies the current A-MTC0 value in the flash device
                partition_src_addr: specifies the 512byte base address within the active partition
                    that may be used to store the signature result of the command
                partition_key_src_addr: specifies flash at the 512byte base address in the active
                    partition that may be used as the source for a private signing key
                input_digest: 256bytes of data that can be provided by the host as an input message
                    to sign
                partition_msg_addr: specifies the 512byte base address in the active partition
                    that may be used by the device owner or the flash WPG owner

            Returns:
                response: PlatformReturnType (status and command response received from Indigo Device)
        """
        # use the input amtc0 for debug
        if amtc0_in:
            use_amtc0 = amtc0_in
        else:
            use_amtc0 = self.amtc0

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(byte_addr, int):
            raise TypeError('{} should be of type int and not {}'.format('byte_addr', type(byte_addr)))

        if not isinstance(use_amtc0, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('use_amtc0', type(use_amtc0)))

        if not isinstance(partition_src_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_src_addr', type(partition_src_addr)))

        if not isinstance(partition_key_src_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_key_src_addr', type(partition_key_src_addr)))

        if not isinstance(input_digest, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('input_digest', type(input_digest)))

        if not isinstance(partition_msg_addr, int):
            raise TypeError('{} should be of type int and not {}'
                            .format('partition_msg_addr', type(partition_msg_addr)))

        # make sure we have the info needed to call the function
        if config is None:
            self.logger.error("Missing required data to run command: config value")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if byte_addr is None:
            self.logger.error("Missing required data to run command: byte_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if use_amtc0 is None:
            self.logger.error("Missing required data to run command: amtc0; \
                  Run request_uid before running this command")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_src_addr is None:
            self.logger.error("Missing required data to run command: partition_src_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_key_src_addr is None:
            self.logger.error("Missing required data to run command: partition_key_src_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if input_digest is None:
            self.logger.error("Missing required data to run command: input_digest")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        if partition_msg_addr is None:
            self.logger.error("Missing required data to run command: partition_msg_addr")
            return PlatformReturnType(EpPlatformRetCode.EP_ERROR)

        # make sure the signing key is set - if the signing using
        # HMAC then the ECDH Public Key from the device must
        # have been set before calling this function. Otherwise
        # it would have to be fetched now which would change the
        # AMTC0 value. So it's to late to fetch now.
        if not self._is_hmac_sig(config) or self._ecdh_public_key is not None:
           self._get_signing_key_with_kts_option(config, byte_addr)
        use_kts_key = self._is_keyRef_bit_set(config)

        # pack the data
        data = partition_msg_addr.to_bytes(4, 'big') + input_digest + partition_key_src_addr.to_bytes(4, 'big') + partition_src_addr.to_bytes(4, 'big')

        # pack the command
        cmd_str = self._generic_pack_command(b'\x9B\x44', config, address = byte_addr.to_bytes(4, 'big'),
                                            monotonic_counter = use_amtc0, data = data)

        # sign the command
        cmd_str = self._sign_cmd(cmd_str, self._is_hmac_sig(config), use_kts_slot_key=use_kts_key)

        # execute the command
        return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)


    """
    Utility Functions
    """

    def _generic_pack_command(self, request: bytes, config: bytes = b'\0'*2,
                             address: bytes = b'\0'*4, monotonic_counter: bytes = b'\0'*8,
                             a_nonce: bytes = b'\0'*16, data: bytes = b'\0'*256,
                             mac_ecdsa_s: bytes = b'\0'*32, ecdsa_r: bytes = b'\0'*32,
                             stuff_bytes: bytes = b'\0'*160) -> bytes:
        """
            “Generic” A-CRTM request command implementation

            Pack all arguments into command string

            Args:
                request: command request id
                config: specifies the command’s functionality
                address: register address
                monotonic_counter: count value will increment after each command
                a_nonce: specifies a random value to be used by the flash device 
                         when generating the ECC signature for the command’s results.
                         if no nonce is supplied it defaults to a radomly 
                         generated nonce in this function and returned
                data: data to be written
                mac_ecdsa_s: device signature
                ecdsa_r: device signature
                stuff_bytes:

            Returns:
                cmd_str: command response received from Indigo Device

            Raises:
                TypeError: If arguments are not of type bytes
                ValueError: If cmd_str length is not equal to 512
        """

        if not isinstance(request, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('request', type(request)))

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('config', type(config)))

        if not isinstance(address, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('address', type(address)))

        if not isinstance(monotonic_counter, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('monotonic_counter', type(monotonic_counter)))

        if not isinstance(a_nonce, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('a_nonce', type(a_nonce)))

        if not isinstance(data, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('data', type(data)))

        if not isinstance(mac_ecdsa_s, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('mac_ecdsa_s', type(mac_ecdsa_s)))

        if not isinstance(ecdsa_r, bytes):
            raise TypeError('{} should be of type bytes and not {}'.format('ecdsa_r', type(ecdsa_r)))

        if not isinstance(stuff_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('stuff_bytes', type(stuff_bytes)))

        if len(monotonic_counter) < 8:
            monotonic_counter = ((8 - len(monotonic_counter)) * b'\x00') + monotonic_counter

        if len(data) < 256:
            data = ((256 - len(data))*b'\x00') + data

        cmd_str = (stuff_bytes + ecdsa_r + mac_ecdsa_s + data + a_nonce + monotonic_counter
                   + address + config + request)

        # check length, raise error
        if len(cmd_str) != 512:
            self.logger.exception("Indigo command string MUST be exactly 512 bytes - found {}".format(len(cmd_str)))
            raise ValueError("Command string is not 512 bytes")

        return cmd_str

    @staticmethod
    def _parse_response_bytes(response_bytes: bytes):
        """
            Creates a dictionary with the command response received from Indigo Device

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                parsed_bytes: response_bytes dictionary

            Raises:
                TypeError: If response_bytes is not of type bytes
                ValueError: If response_bytes length is not equal to 512
        """

        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError('Response Bytes buffer is not 512 bytes')

        parsed_bytes = {
            'Response': response_bytes[510:],
            'Device Status': response_bytes[508:510],
            'ESR': response_bytes[506:508],
            'Monotonic Count': response_bytes[498:506],
            'Nonce': response_bytes[482:498],
            'Data': response_bytes[226:482],
            'MAC/ECDSA(s)': response_bytes[194:226],
            'ECDSA(r)': response_bytes[162:194],
            'Stuff Bytes': response_bytes[0:162]
        }

        return parsed_bytes

    @staticmethod
    def _request_uid_parse(parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_byte

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: data dictionary
        """

        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        data = parsed_data['Data']
        parsed_data['Data'] = {
            'Zeros': data[0:172],
            'UID': data[172:188],
            'Revision': data[188:192],
            'ECDSA_DS_PuK': data[192:256]
        }

        return parsed_data

    @staticmethod
    def _provision_response_parse(parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_byte

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: data dictionary
        """

        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        data = parsed_data['Data']
        parsed_data['Data'] = {
            'Zeros': data[0:208],
            'ECDH_PV_PuK': data[208:240],
            'UID': data[240:256]
        }

        return parsed_data

    def _request_uid_response_bytes(self, response_bytes: bytes):
        """
            Gets a nested dictionary of response_bytes values

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                request_uid_response: nested dictionary of response_bytes values
        """
        # assert(len(response_bytes) == 512)

        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError("Response_bytes is not 512 bytes")

        generic_parse = self._parse_response_bytes(response_bytes)
        request_uid_response = self._request_uid_parse(generic_parse)

        return request_uid_response

    @staticmethod
    def _crtm_register_0_read_parse(parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_byte

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: data dictionary
        """

        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        # The 4-byte data structures are treated as integers and stored stored in little-endian,
        # here the byte order has to be reversed to big-endian.
        data = parsed_data['Data'][::-1]
        parsed_data['Data'] = {
            'Flash_PCR0': data[0x0:0x20],
            'CRTM_Addr0': data[0x20:0x24][::-1],
            'CRTM_Size0': data[0x24:0x28][::-1],
            'CRTM_Addr1': data[0x28:0x2C][::-1],
            'CRTM_Size1': data[0x2C:0x30][::-1],
            'CRTM_Addr2': data[0x30:0x34][::-1],
            'CRTM_Size2': data[0x34:0x38][::-1],
            'CRTM_Addr3': data[0x38:0x3C][::-1],
            'CRTM_Size3': data[0x3C:0x40][::-1],
            'CRTM_Addr4': data[0x40:0x44][::-1],
            'CRTM_Size4': data[0x44:0x48][::-1],
            'CRTM_Addr5': data[0x48:0x4C][::-1],
            'CRTM_Size5': data[0x4C:0x50][::-1],
            'CRTM_Addr6': data[0x50:0x54][::-1],
            'CRTM_Size6': data[0x54:0x58][::-1],
            'CRTM_Addr7': data[0x58:0x5C][::-1],
            'CRTM_Size7': data[0x5C:0x60][::-1],
            'Flash_DigestNV_Partial': data[0x60:0x80],
            'Flash_DigestNV': data[0x80:0xA0],
            'Device_Config': data[0xA0:0xA4][::-1],
            'Provision_Sts': data[0xA4:0xA8][::-1],
            'CRTM_Max_AutoMeasure_Size': data[0xA8:0xAC][::-1],
            'A_MTC1': data[0xAC:0xB0][::-1],
            'KTR': data[0xB0:0xD0],
            'EEPROM_Attributes': data[0xD0:0xD4][::-1],
            'PwrOnCountNV': data[0xD4:0xD8][::-1],
            'Device_Config_Volatile': data[0xD8:0xDC][::-1],
            'Reserved': data[0xDC:0xE0],
            'KTR1': data[0xE0:0x100]
        }

        return parsed_data

    def _crtm_register_0_read_parsed_response(self, response_bytes: bytes):
        """
            Gets a nested dictionary of response_bytes values

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                crtm_register_read_response: nested dictionary of response_bytes values
        """

        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError("Response_bytes is not 512 bytes")

        generic_parse = self._parse_response_bytes(response_bytes)
        crtm_register_read_response = self._crtm_register_0_read_parse(generic_parse)

        return crtm_register_read_response

    @staticmethod
    def _manage_kts_parse(parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_byte

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: data dictionary
        """

        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        data = parsed_data['Data']
        parsed_data['Data'] = {
            'Zeros': data[0:176],
            'OutputKey': data[176:240],
            'UID': data[240:256]
        }

        return parsed_data

    def _manage_kts_response_bytes(self, response_bytes: bytes):
        """
            Gets a nested dictionary of response_bytes values

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                manage_kts_response: nested dictionary of response_bytes values
        """
        # assert(len(response_bytes) == 512)

        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError("Response_bytes is not 512 bytes")

        generic_parse = self._parse_response_bytes(response_bytes)
        manage_kts_response = self._manage_kts_parse(generic_parse)

        return manage_kts_response

    def _set_extended_crtm_status(self, device_status_response: bytes):
        """
            Creates Extended CRTM Status Dictionary

            Creates an Extended CRTM Status dictionary with the Device Status value
            from response

            Args:
                device_status_response: device status response received from Indigo Device
        """
        #convert extended crtm status from bytes to binary
        extended_crtm_status_bin = bin(int.from_bytes(device_status_response, 'big')).lstrip('0b')
        extended_crtm_status_bin = ((16 - len(extended_crtm_status_bin))*'0') + extended_crtm_status_bin

        #convert extended crtm status from string to int
        extended_crtm_status_lst_int = [int(d, 16) for d in extended_crtm_status_bin]

        extended_crtm_status_lst = []

        extended_crtm_status_lst.append(0)
        extended_crtm_status_lst.extend(extended_crtm_status_lst_int[9:12])
        extended_crtm_status_lst.append(0)
        extended_crtm_status_lst.extend(extended_crtm_status_lst_int[13:])

        self._extended_crtm_status_struct = self.EXTENDED_CRTM_STATUS_STRUCT(*extended_crtm_status_lst)

    def _read_extended_crtm_status(self):
        error_code, _ = self.meta_crtm_register_read(b'\x80\x00', 0x00)
        return error_code, self._extended_crtm_status_struct

    def _set_device_config_volatile(self, device_config_volatile: bytes):
        """
            Creates Device Config Volatile Dictionary

            Creates a Device Config Volatile dictionary with the " Device_Config_Volatile"
            value from response to register_read

            Args:
                device_config_volatile: device status response received from Indigo Device
        """

        #convert device config volatile from bytes to binary
        device_config_volatile_bin = bin(int.from_bytes(device_config_volatile, 'big')).lstrip('0b')
        device_config_volatile_bin = ((32 - len(device_config_volatile_bin))*'0') + device_config_volatile_bin

        #convert device config volatile from string to int
        device_config_volatile_int = [int(d, 32) for d in device_config_volatile_bin]

        device_config_volatile_list = []

        device_config_volatile_list.append(0)
        device_config_volatile_list.extend(device_config_volatile_int[27:30])
        device_config_volatile_list.append(0)

        self._device_config_volatile_struct = self.CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(*device_config_volatile_list)


    def _set_device_config(self, device_config: bytes):
        """
            Creates a Device Config dictionary with the ' Device_Config' value
            from response to register_read of the initial registers

            Args:
                device_config: "Device_Config" field in the register_read 
                               response received from Indigo Device
        """

        #convert device config from bytes to binary
        device_config_bin = bin(int.from_bytes(device_config, 'big')).lstrip('0b')
        device_config_bin = ((32 - len(device_config_bin))*'0') + device_config_bin

        #convert device config from string to int
        device_config_int = [int(d, 32) for d in device_config_bin]

        device_config_list = []

        device_config_list.append(0)
        device_config_list.extend(device_config_int[20:32])
        # device_config_list.append(0)

        self._device_config_struct = self.CRTM_REGISTER_DEVICE_CFG_STRUCT(*device_config_list)
        

    def set_CRK_Read_Disable_bit(self, use_ecdsa_sig: bool)->IndigoCmdErrorCode:
        """
        Set the CRK_Read_Diable bit in the Device_Config register (Device_Config[2])

        ARGS:
            use_ecdsa_sig  if True sign commands to the device using ECDSA signature
                           if False sign commands to the device using HMAC signature
        """
        if use_ecdsa_sig:
            config = b'\x80\x00'
        else:
            config = b'\x00\x00'

        if self._authenta_initial_regs_parsed is None:
            device_config = None
            rc, resp = self.meta_crtm_register_read(config, CRTM_ADDRESS_BLOCK_INITIAL_REGS)
            if not rc.valid_response:
                return rc

            reg_0_parsed_response = self._crtm_register_0_read_parsed_response(resp)
        else:
            reg_0_parsed_response = self._authenta_initial_regs_parsed

        device_config = reg_0_parsed_response['Data']['Device_Config']

        device_config = device_config[0:3] + (device_config[3] | 0x04).to_bytes(1, 'big')
        device_config = device_config[::-1]

        # device_config is at register-address 0xA0 
        dc_addr = 0xA0
        data = b'\x00' * 128
        data = data[0:dc_addr-0x80] + device_config + data[(dc_addr+4)-0x80:]
        data = data[::-1]
        mask = b'\xFF' * 128
        mask = mask[0:dc_addr-0x80] + b'\x00\x00\x00\x00' + mask[(dc_addr+4)-0x80:]
        mask = mask[::-1]

        rc, _ = self.meta_acrtm_update_start(config)
        if not rc.valid_response:
            return rc

        rc, _ = self.meta_crtm_register_write(config, 0x80, data, mask)
        if not rc.valid_response:
            self.meta_acrtm_update_finish(config)
            return rc

        rc_finish, _ = self.meta_acrtm_update_finish(config)
        if not rc_finish.valid_response:
            return rc

        return rc

    def set_CSK_Read_Disable_bit(self, use_ecdsa_sig: bool)->IndigoCmdErrorCode:
        """
        Set the CSK_Read_Disable bit in the Device_Config_Volatile register 
        (Device_Config_Volatile[2])

        ARGS:
            use_ecdsa_sig  if True sign commands to the device using ECDSA signature
                           if False sign commands to the device using HMAC signature
        """
        if use_ecdsa_sig:
            config = b'\x80\x00'
        else:
            config = b'\x00\x00'

        if self._authenta_initial_regs_parsed is None:
            device_config_vol = None
            rc, resp = self.meta_crtm_register_read(config, CRTM_ADDRESS_BLOCK_INITIAL_REGS)
            if not rc.valid_response:
                return rc

            reg_0_parsed_response = self._crtm_register_0_read_parsed_response(resp)
        else:
            reg_0_parsed_response = self._authenta_initial_regs_parsed

        device_config_vol = reg_0_parsed_response['Data']['Device_Config_Volatile']

        device_config_vol = device_config_vol[0:3] + (device_config_vol[3] | 0x04).to_bytes(1, 'big')
        device_config_vol = device_config_vol[::-1]
        
        # device_config_volatile is at register-address 0xD8
        dcv_addr = 0xD8
        data = b'\x00' * 128
        data = data[0:dcv_addr-0x80] + device_config_vol + data[(dcv_addr+4)-0x80:]
        data = data[::-1]
        mask = b'\xFF' * 128
        mask = mask[0:dcv_addr-0x80] + b'\x00\x00\x00\x00' + mask[(dcv_addr+4)-0x80:]
        mask = mask[::-1]

        rc, _ = self.meta_acrtm_update_start(config)
        if not rc.valid_response:
            return rc

        rc, _ = self.meta_crtm_register_write(config, 0x80, data, mask)
        if not rc.valid_response:
            self.meta_acrtm_update_finish(config)
            return rc

        rc_finish, _ = self.meta_acrtm_update_finish(config)
        if not rc_finish.valid_response:
            return rc

        return rc


    def _handle_request_uid_provision_response(self, command_name: str, response: bytes, sent_nonce: bytes, do_response_verification: bool = True):
        """
            Handle atomic command response to request_uid and provision

            Check values from the response received from Indigo device

            Args:
                command_name: atomic commmand name
                signed_with_session_key: boolean indicating if the caommand was signed with the
                                         HMAC session key (True) or the ECDSA key (False)
                response: atomic command id
                do_response_verification: True if response_verification is needed, False otherwise.
                verification_key: If not None, use this key to verify the response signature (used when using a KTS resident key)

            Returns:
                error_code: IndigoCmdErrorCode
                response_bytes: command response bytes received from Indigo Device

            Raises:
                TypeError: If command_name is not of type string
                TypeError: If response_bytes is not of type bytes
                TypeError: If do_response_verification is not of type boolean
        """

        if not isinstance(command_name, str):
            raise TypeError('{} should be of type string and not {}'
                            .format('command_name', type(command_name)))

        if not isinstance(command_name, str):
            raise TypeError('{} should be of type string and not {}'
                            .format('command_name', type(command_name)))

        if not isinstance(response, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response)))

        if not isinstance(do_response_verification, bool):
            raise TypeError('{} should be of type bool and not {}'
                            .format('do_response_verification', type(do_response_verification)))

        parsed_response = self._request_uid_response_bytes(response)

        ESR = parsed_response['ESR']

        error_code = IndigoCmdErrorCode(ESR, command_name)
        resp_valid = error_code.esr_success

        if not do_response_verification:
            if error_code.esr_success:
                self._amtc0 = parsed_response['Monotonic Count']
            return error_code

        if error_code.esr_success:
            error_code._error_msg = 'ESR OK'
            self._uid = parsed_response["Data"]["UID"]
            self._revision = parsed_response['Data']['Revision']

            # now that we have the UID open the key driver if it hasn't been opened yet
            if self._keylib_context is None:
                self._key_driver_dev_info_dict['device_revision'] = self._revision.hex()               
                self._open_key_driver();

            # set the (trusted) ECDSA verfication key which must be installed in the key library
            authenta_key_lib.set_ecdsa_verification_key(self._keylib_context)

            # set AMTC0 here in case verifying the response
            # requires sending a command
            self._amtc0 = parsed_response['Monotonic Count']

            signature_valid = self._response_ecdsa_sig_is_valid(response, command_name)
            error_code.set_signature_matched(signature_valid)
            if not signature_valid:
                self.logger.error("signature verification failed!")
                error_code._error_msg += '; signature verification failed'
            
            resp_valid = resp_valid and signature_valid
            nonce_matched = parsed_response['Nonce'] == sent_nonce
            error_code.set_nonce_matched(nonce_matched)
            if not nonce_matched:
                self.logger.error("nonce match failure! sent nonce= 0x{}; returned nonce={}!". format(sent_nonce.hex(), parsed_response['Nonce'].hex()))
                resp_valid = False
                error_code._error_msg += '; nonce mismatch'
            # if the response is not valid now, we are in a strange
            # situation that the ESR indicates success but the response
            # indicates that it is not valid, so not sure if the 
            # AMTC0 in the response is correct!
            # solution here is to clear the AMTC0 and the next command
            # will have to refetch it.
            #
            # COMMENTING OUT FIX:
            # The above is correct, but this solution can cause a problem
            # when this call is "embedded" in some other action, for example,
            # in a meta_XXX function, when fetching information to determine
            # the session, etc. and we assume the amtc0 is set without checking.
            # So, for now we will not reset _amtc0 and if it is wrong than
            # the chip will cause an error because of that - this is not terrible
            # because we're in a unexpected or bad situation anyway with the
            # response not signed correctly.
            # TODO: Figure out the best cource of action here...
            #
            #if not resp_valid:
            #    self._amtc0 = None
        else:
            self.logger.error("command returned ESR value: {}; Err Msg: {}".format(ESR, error_code.error_msg))

        error_code.set_valid_response(resp_valid)
        if resp_valid:
            error_code._error_msg = 'OK'

        return error_code

    def _meta_handle_atomic_command_response(self, command_name: str, config: bytes, response: bytes, do_response_verification: bool = True, sent_nonce: bytes = None):
        """
            Handle atomic command response

            Check values from the response received from Indigo device

            Args:
                command_name: atomic commmand name
                signed_with_session_key: boolean indicating if the caommand was signed with the
                                         HMAC session key (True) or the ECDSA key (False)
                response: atomic command id
                do_response_verification: True if response_verification is needed, False otherwise.
                verification_key: If not None, use this key to verify the response signature (used when using a KTS resident key)

            Returns:
                error_code: IndigoCmdErrorCode
                response_bytes: command response bytes received from Indigo Device

            Raises:
                TypeError: If command_name is not of type string
                TypeError: If response_bytes is not of type bytes
                TypeError: If do_response_verification is not of type boolean
        """

        if not isinstance(command_name, str):
            raise TypeError('{} should be of type string and not {}'
                            .format('command_name', type(command_name)))

        if not isinstance(config, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('config', type(config)))

        if not len(config) == 2:
            raise ValueError('{} should be of type bytes of length 2 and and not be length {}'
                            .format('config', len(config)))

        if not isinstance(response, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response)))

        if not isinstance(do_response_verification, bool):
            raise TypeError('{} should be of type bool and not {}'
                            .format('do_response_verification', type(do_response_verification)))

        if command_name != 'deprovision_crtm' or self.provision_sts_struct.Remotely_Provisioned:
            signed_with_session_key = self._is_hmac_sig(config)
        else:
            # There is an exception from the standard
            # response signature rules for deprovision_crtm on a locally
            # provisioned Indigo device; sometimes the response
            # is signed w/ a different key than the command...
            #
            # If the config[1] (Allow_SRK) is set the response is signed by ECDSA_DS_PrK,
            # otherwise it is signed by the session key
            signed_with_session_key = ((config[1] & 0x02) == 0)

        generic_parse = self._parse_response_bytes(response)

        ESR = generic_parse['ESR']

        error_code = IndigoCmdErrorCode(ESR, command_name)
        resp_valid = error_code.esr_success

        if not do_response_verification:
            if error_code.esr_success:
                self._amtc0 = generic_parse['Monotonic Count']
            return error_code

        if error_code.esr_success:
            error_code._error_msg = 'ESR OK'
            # set AMTC0 here in case verifying the response
            # requires sending a command
            self._amtc0 = generic_parse['Monotonic Count']

            if signed_with_session_key:
                signature_valid = self._response_hmac_sig_is_valid(response, command_name)
            else:
                signature_valid = self._response_ecdsa_sig_is_valid(response, command_name)
            
            error_code.set_signature_matched(signature_valid)
            if not signature_valid:
                self.logger.error("signature verification failed!")
                error_code._error_msg += '; signature verification failed'
            
            resp_valid = resp_valid and signature_valid

            if sent_nonce is not None:
                nonce_matched = generic_parse['Nonce'] == sent_nonce
                error_code.set_nonce_matched(nonce_matched)
                if not nonce_matched:
                    self.logger.error("nonce match failure! sent nonce= 0x{}; returned nonce={}!". format(sent_nonce.hex(), generic_parse['Nonce'].hex()))
                    resp_valid = False
                    error_code._error_msg += '; nonce mismatch'
            # if the response is not valid now, we are in a strange
            # situation that the ESR indicates success but the response
            # indicates that it is not valid, so not sure if the 
            # AMTC0 in the response is correct!
            # solution here is to clear the AMTC0 and the next command
            # will have to refetch it.
            #
            # COMMENTING OUT FIX:
            # The above is correct, but this solution can cause a problem
            # when this call is "embedded" in some other action, for example,
            # in a meta_XXX function, when fetching information to determine
            # the session, etc. and we assume the amtc0 is set without checking.
            # So, for now we will not reset _amtc0 and if it is wrong than
            # the chip will cause an error because of that - this is not terrible
            # because we're in a unexpected or bad situation anyway with the
            # response not signed correctly.
            # TODO: Figure out the best cource of action here...
            #
            #if not resp_valid:
            #    self._amtc0 = None
        else:
            self.logger.error("command returned ESR value: {}; Err Msg: {}".format(ESR, error_code.error_msg))

        error_code.set_valid_response(resp_valid)
        if resp_valid:
            error_code._error_msg = 'OK'

        # request_uid does not
        # return extended_crtm_status and always has to handle its own verification because the response
        # structure is non-standard
        if error_code.esr_success and command_name not in ('request_uid'):
                self._set_extended_crtm_status(generic_parse['Device Status'])

        return error_code

    def _get_amtc0_if_not_set(self, caller_name: str, amtc0_test_input: bytes = None):
        """
            Gets the A-MTC0 if it is not already set

            Calls meta_request_uid to get the A-MTC0

            Args:
                caller_name: name of the function calling this command
                amtc0_test_input:

            Returns:
                error_code: IndigoCmdErrorCode

            Raises:
                TypeError: If argument is not of type string
        """
        if not isinstance(caller_name, str):
            raise TypeError('{} should be of type string and not {}'
                            .format('caller name', type(caller_name)))

        # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
        if not amtc0_test_input and not self.amtc0:
            error_code, _ = self.meta_request_uid()
            if error_code.esr_success:
                self.logger.info('Retrieved AMTC0 value')
            else:
                self.logger.error('Failed attempting to retrieve AMTC0 from device')
        else:
            error_code = IndigoCmdErrorCode(b'\x00\x80', caller_name, "OK")
        return error_code

    def _get_boot_amtc0_if_not_set(self, caller_name: str):
        """
            Gets the Boot A-MTC0 if it is not already set

            Calls _read_extended_crtm_status to retrieve the boot_amtc0

            Args:
                caller_name: name of the function calling this command

            Returns:
                error_code: IndigoCmdErrorCode

            Raises:
                TypeError: If argument is not of type string
        """
        if not isinstance(caller_name, str):
            raise TypeError('{} should be of type string and not {}'.format('caller name', type(caller_name)))

        # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling
        # _read_extended_crtm_status (if device is Provisioned)
        if not self._boot_amtc0:
            if self.provision_sts_struct is None:
                error_code, _ = self.meta_request_uid()
                if not error_code.esr_success:
                    self.logger.error('Call to request_uid failed while attempting \
                          to retrieve provision status')
                    error_code = IndigoCmdErrorCode(b'\x00\x00', caller_name, "Call to \
                                                        request_uid failed while attempting \
                                                        to retrieve provision status")
                    return error_code

            if self.provision_sts_struct.Provisioned:
                error_code, _ = self._read_extended_crtm_status()
                if error_code.esr_success:
                    self.logger.info('Retrieved BOOT_AMTC0 value')
                else:
                    self.logger.error('Failed attempting to retrieve BOOT_AMTC0 from device')
            else:
                self.logger.error("Device is not provisioned; Cannot retrieve BOOT_AMTC0")
                error_code = IndigoCmdErrorCode(b'\x00\x00', caller_name, "Device is not "\
                                                    "provisioned; Cannot retrieve BOOT_AMTC0")
        else:
            error_code = IndigoCmdErrorCode(b'\x00\x80', caller_name, "OK")

        return error_code

    def _provision_crtm_response_bytes(self, response_bytes: bytes):
        """
            Gets a nested dictionary of response_bytes values

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                provision_crtm_response: nested dictionary of response_bytes values
        """
        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError("Response_bytes is not 512 bytes")

        generic_parse = self._parse_response_bytes(response_bytes)
        provision_crtm_response = self._provision_crtm_parse(generic_parse)

        return provision_crtm_response

    def _provision_crtm_parse(self, parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_byte

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: data dictionary
        """
        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        data = parsed_data['Data']
        parsed_data['Data'] = {
            'Zeros': data[0:208],
            'ECDH_PV_PuK': data[176:240][::-1],
            'UID': data[240:256]
        }

        return parsed_data

    def _apa_modify_response_bytes(self, response_bytes: bytes):
        """
            Gets a nested dictionary of response_bytes values

            Args:
                response_bytes: command response received from Indigo Device

            Returns:
                apa_modify_response: nested dictionary of response_bytes values
        """

        if not isinstance(response_bytes, bytes):
            raise TypeError('{} should be of type bytes and not {}'
                            .format('response_bytes', type(response_bytes)))

        if len(response_bytes) != 512:
            raise ValueError("Response_bytes is not 512 bytes")

        generic_parse = self._parse_response_bytes(response_bytes)
        apa_modify_response = self._apa_modify_parse(generic_parse)

        return apa_modify_response

    def _apa_modify_parse(self, parsed_data: dict):
        """
            Creates a dictionary with the Data from _parse_response_bytes

            Args:
                parsed_data: response_bytes dictionary

            Returns:
                parsed_data: dictionary
        """

        if not isinstance(parsed_data, dict):
            raise TypeError('{} should be of type dict and not {}'
                            .format('parsed_data', type(parsed_data)))

        data = parsed_data['Data']
        parsed_data['Data'] = {
            'Zeros': data[0:172],
            'APA_DataResponse': data[172:252],
            'WPG_Addr': data[252:256]
        }

        return parsed_data

    def _create_data_buffer(self, data: bytes, crtm_addr: int):
        """
        Creates a 128 byte data buffer

            Args:
                data: data to be written to the register
                crtm_addr: address of the CRTM register space

            Returns:
                data: 128 byte data buffer
        """
        if len(data) < 128:
            data = data + b'\x00'*(crtm_addr % 128)
            data = ((128 - len(data))*b'\x00') + data

        return data

    def _create_mask_buffer(self, mask: bytes, crtm_addr: int):
        """
        Creates a 128 byte mask buffer

            Args:
                mask: mask to be used with the data to be written to the register
                crtm_addr: address of the CRTM register space

            Returns:
                mask: 128 byte mask buffer
        """    
        if len(mask) < 128:
            mask = mask + b'\xFF'*(crtm_addr % 128)
            mask = ((128 - len(mask))*b'\xFF') + mask

        return mask


    """
    Signing Functions
    """


    @staticmethod
    def _replace_ecdsa_key_in_cmd_str(cmd: bytes, ecdsa_val: bytes) -> bytes:
        """
        replace the ECDSA(r) value within the input command 
        with the input ecdsa value and return the resulting bytes object

        Args:
            cmd:       input command bytes in which the ECDS(r) is be be replaced
                       MUST be 512 bytes long
            ecdsa_val: the bytes to set bytes to replace the field with)
                       MUST be 64 bytes long


        Return:
            the resulting bytes object with after replacement
        """
        assert len(cmd) == 512 and len(ecdsa_val) == 64
        if len(cmd) != 512:
            raise ValueError("[IndigoDevice.replace_ecdsa_r_withing _cmd] Invalid input command length: Must be 512 found {}!".format(len(cmd)))
        if len(ecdsa_val) != 64:
            raise ValueError("[IndigoDevice.replace_ecdsa_r_withing _cmd] Invalid input ECDSA(r) length: Must be 64 found {}!".format(len(cmd)))

        return cmd[:512-352] + ecdsa_val + cmd[512-288:]

    @staticmethod
    def _replace_hmac_key_in_cmd_str(cmd: bytes, mac_val: bytes) -> bytes:
        """
        replace the MAC/ECDSA(s) value within the input command 
        with the input ecdsa value and return the resulting bytes object

        Args:
            cmd:       input command bytes in which the MAC or ECDSA(s) is be be replaced
                       MUST be 512 bytes long
            ecdsa_val: the bytes to set bytes to replace the field with
                       MUST be 32 bytes long


        Return:
            the resulting bytes object with after replacement
        """
        assert len(cmd) == 512 and len(mac_val) == 32
        if len(cmd) != 512:
            raise ValueError("[IndigoDevice.replace_ecdsa_r_withing _cmd] Invalid input command length: Must be 512 found {}!".format(len(cmd)))
        if len(mac_val) != 32:
            raise ValueError("[IndigoDevice.replace_ecdsa_r_withing _cmd] Invalid input ECDSA(r) length: Must be 32 found {}!".format(len(cmd)))
 
        return cmd[:192] + mac_val + cmd[224:]

    def _is_hmac_sig(self, config: bytes) -> bool:
        return (config[0] & 0x80) == 0

    def _is_keyRef_bit_set(self, config: bytes) -> bool:
        return not ((config[0] & 0x40) == 0)

    def _fetch_ecdh_pub_key(self):
        err_code, _ = self.meta_crtm_register_read(b'\x80\x00', CRTM_ADDRESS_BLOCK_PUB_KEYS)
        return err_code

    def _sign_cmd(self, cmd, hmac_sig: bool, use_session_key: bool=True, use_kts_slot_key: bool=False):
        """
        PARAMETERS:
            cmd             the full command including the bytes to sign and 
                            placeholder bytes for sig
            hmac_sig        True if it should be signed using HMAC, False if ECDSA
            use_session_key Ignored if hmac_sig is False; Otherwise True if sign
                            with session key, False if sign with root key
            use_kts_slot_key True if signing should be with a key set in a kts slot, False otherwise
        Return
            Full command to be sent to the device including signature
        """
        signed_cmd = cmd
        data_to_sign = cmd[self._INDIGO_SIGNED_COMMAND_OFFSET_IDX:]


        if hmac_sig:
            if use_kts_slot_key:
                signature = authenta_key_lib.gen_kts_sha256_hmac(self._keylib_context, data_to_sign)
            elif use_session_key:
                signature = authenta_key_lib.gen_session_sha256_hmac(self._keylib_context, data_to_sign)
            else:
                signature = authenta_key_lib.gen_root_sha256_hmac(self._keylib_context, data_to_sign)
            
            if signature is not None:
                signed_cmd = self._replace_hmac_key_in_cmd_str(cmd[:512], signature) + cmd[512:]
        else:
            if use_kts_slot_key:
                signature = authenta_key_lib.generate_kts_ecdsa_signature(self._keylib_context, data_to_sign)
            else:
                signature = authenta_key_lib.generate_ecdsa_signature(self._keylib_context, data_to_sign)

            if signature is not None:
                signed_cmd = self._replace_ecdsa_key_in_cmd_str(cmd[:512], signature) + cmd[512:]

        return signed_cmd
 

    def read_KTR_value(self)->Tuple[IndigoCmdErrorCode, bytes]:
        """
        THIS FUNCTION IS ONLY TO BE USED WHEN LOCALLY PROVISIONED!!!

        Convenience function to read the KTR1 value from the device by 
        calling meta_crtm_register_read and return KTR1 value as a bytes

        Note: Upon success _device_config_struct and _device_config_volatile_struct
              are populated
        Return a tuple, with first value the error code of the call to 
        meta_crtm_register_read, and the seconde the 32 byte KTR1 value
        If the error code indicated failure, the value of the next value 
        in the tuple is undefined

        NOTE: Register read is sent in this function and its response signature is NOT verified
        """
        assert(not self._provision_sts_struct.Remotely_Provisioned)

        ktr_val = None
        rc, resp = self.meta_crtm_register_read(b'\x80\00', CRTM_ADDRESS_BLOCK_INITIAL_REGS, verify_resp_sig=False)
        if rc.esr_success:
            reg_0_parsed_response = self._crtm_register_0_read_parsed_response(resp)
            ktr_val = reg_0_parsed_response['Data']['KTR']

            self._set_device_config(reg_0_parsed_response['Data']['Device_Config'])
            self._set_device_config_volatile(reg_0_parsed_response['Data']['Device_Config_Volatile'])

        return rc, ktr_val

    def _set_client_keys(self):
        """
        Set the CRK and/or CSK based on the provision state of the attached device
        This function may call meta_crtm_register_read (without validating the signature)
        """

        # if it had been done already return True
        if self._client_keys_set:
            return True

        # if not locally provisioned return false
        if not self._provision_sts_struct.Provisioned or self._provision_sts_struct.Remotely_Provisioned:
            # client root key and client session keys are
            # only relevant for localy provisioned devices
            self.logger.debug("FAIL; Device MUST be Locally Provisioned to set CRK and/or CSK!")
            return False
        
        if self.provision_sts_struct.HSM_Present:
            if not self._root_signing_key_externally_set:
                self.logger.info("Setting CRK as root signing key in key driver")
                if not authenta_key_lib.set_crk_as_hmac_root_signing_key(self._keylib_context):
                    self.logger.warning("Failed to set Client Root Key; No CRK was found by the key library for device with UID {}".format(self.uid.hex()))
                    return False

            # now that the create the session key
            if not self._session_key_externally_set:
                self.logger.info("Setting deriving session key from CRK in key driver")
                if not authenta_key_lib.derive_ecdh_session_key(self._keylib_context, self._ecdh_public_key, self._boot_amtc0):
                    self.logger.warning("Failed to set CSK for UID {}".format(self.uid.hex()))
                    return False
        else:
            # No TPM or HSM available so the Session key should be set in the key library
            if not self._session_key_externally_set:
                self.logger.info("Setting CSK as session signing key in key driver")
                if not authenta_key_lib.set_csk_as_hmac_session_signing_key(self._keylib_context):
                    self.logger.warning("Failed to set Client Session Key; No CSK was found by the key library for device with UID {}".format(self.uid.hex()))
                    return False

        self._client_keys_set = True
        return True
                                

    def _prepare_signing_key(self, config: bytes, allow_device_interaction=True) -> bool:
        """
        return the correct signing key based on the input config value and the state of the device

        Args:
            config: A two byte config value used as a bit array. 
                    If config[15] is set return the SRK or CRK handle,
                    if not return the session key

        Return:
            Return True on success
                   False on failure
        """
        rc = False

        if not self._is_hmac_sig(config):
            if self._ecdsa_signing_key_set:
                return True
                
            # use the (default) ecdsa key in the key library
            rc = authenta_key_lib.set_ecdsa_signing_key(self._keylib_context)
            if not rc:
                self.logger.warning("Attempt to set the ECDSA signing key failed!")
            if not authenta_key_lib.set_ecdsa_verification_key(self._keylib_context):
                self.logger.warning("Attempt to set the ECDSA verification key failed!")

            self._ecdsa_signing_key_set = True            
        else:
            # use session key

            if self._ecdh_session_key_set:
                return True

            # generate the session key
            if self._uid is None or self.provision_sts_struct is None:
                if allow_device_interaction:
                    rc, _ = self.meta_request_uid()
                    if not rc.esr_success:
                        self.logger.error('Unable to determine correct HMAC signing key; attempt to retrieve provision_sts_struct by calling meta_request_uid Failed: ' + rc.error_msg)
                        return False
                else:
                    self.logger.error('Unable to determine correct HMAC signing key; UID and provision_sts_struct MUST be set before calling prepare_signing_key with device iteraction set to false')
                    return False

            if(self.provision_sts_struct.Provisioned):
                if(self.provision_sts_struct.Remotely_Provisioned):
                        
                    # make sure we have boot-AMTC0 (PowerOnAmtc0)
                    if not allow_device_interaction and self._boot_amtc0 is None:
                        self.logger.error('Unable to derive ECDH session key - missing PowerOnCount')
                        return False

                    error_code = self._get_boot_amtc0_if_not_set('')
                    if not error_code.esr_success or self._boot_amtc0 is None:
                        return False

                    # make sure we have the device ECDH public key (ECDH_PV_PuK) from the device                    
                    if self._ecdh_public_key is None:
                        if not allow_device_interaction:
                            self.logger.error('Unable to derive ECDH session key - missing device ECDH public key (ECDH_PV_PuK)')
                            return False

                        err_code = self._fetch_ecdh_pub_key()
                        if not err_code._valid_response:
                            self.logger.error('Cannot create session key due to failure to generate shared secret! Failed to fetch ECDH_PV_PuK from device')
                            return False

                    if self._ecdh_public_key is None:
                        self.logger.error('INTERNAL Cannot create session key! Failed to set ECDH_PV_PuK after crtm_register_read; Unable to generate ECDH shared secret!')
                        return False

                    # now set the session key in the key library
                    rc = authenta_key_lib.derive_ecdh_session_key(self._keylib_context, self._ecdh_public_key, self._boot_amtc0)
                    if rc:
                        self._ecdh_session_key_set = True
                    else:
                        self.logger.error('Attempt to set ECDH session key failed!')
                else:
                    # if a session key needs to be derived then we need the PowerOnCount
                    # and the device's ECDH public key
                    if self.provision_sts_struct.HSM_Present:
                        # make sure we have boot-AMTC0 (PowerOnAmtc0)
                        if not allow_device_interaction and self._boot_amtc0 is None:
                            self.logger.error('Unable to derive ECDH session key - missing PowerOnCount')
                            return False

                        error_code = self._get_boot_amtc0_if_not_set('')
                        if not error_code.esr_success or self._boot_amtc0 is None:
                            return False

                        # make sure we have the device ECDH public key (ECDH_PV_PuK) from the device                    
                        if self._ecdh_public_key is None:
                            if not allow_device_interaction:
                                self.logger.error('Unable to derive ECDH session key - missing device ECDH public key (ECDH_PV_PuK)')
                                return False

                            err_code = self._fetch_ecdh_pub_key()
                            if not err_code._valid_response:
                                self.logger.error('Cannot create session key due to failure to generate shared secret! Failed to fetch ECDH_PV_PuK from device')
                                return False

                        if self._ecdh_public_key is None:
                            self.logger.error('INTERNAL Cannot create session key! Failed to set ECDH_PV_PuK after crtm_register_read; Unable to generate ECDH shared secret!')
                            return False

                    rc = self._set_client_keys()
                    if rc:
                        self._ecdh_session_key_set = True
                    else:
                        self.logger.error('Attempt to set client key, (CRK or CSK), failed!')
                        
        return rc

    def _response_ecdsa_sig_is_valid(self, response: bytes, cmd_name: str) -> bool:
        # not sure if this should be done - it was probably done already
        # if not authenta_key_lib.set_ecdsa_verification_key(self._keylib_context):
        #    self.logger.info("Failed to set ECDSA verification key in key library!")

        AUTHENTA_RESP_DATA_WO_KEY_LEN = 286
        parsed_response = self._parse_response_bytes(response)
        signed_data = response[-AUTHENTA_RESP_DATA_WO_KEY_LEN:]
        returned_sig = parsed_response["ECDSA(r)"] + parsed_response["MAC/ECDSA(s)"] 

        return authenta_key_lib.verify_ecdsa_signature(self._keylib_context, signed_data, returned_sig)

    def _response_hmac_sig_is_valid(self, response: bytes, cmd_name: str) -> bool:
        # the _session_key should have been calculated already for the
        # command signature, if it is not set there is something wrong
        #self._prepare_signing_key(b'\x00\x00')

        AUTHENTA_RESP_DATA_WO_KEY_LEN = 286
        parsed_response = self._parse_response_bytes(response)
        signed_data = response[-AUTHENTA_RESP_DATA_WO_KEY_LEN:]
        returned_sig = parsed_response["MAC/ECDSA(s)"]

        return authenta_key_lib.verify_session_sha256_hmac(self._keylib_context, signed_data, returned_sig)

    def _get_kts_slot_key_name(self, is_pub_key: bool, config: bytes, slot: int)->str:
        """
        return the keyname used in the key library based on the 
        input slot and the sigType value in the input config.
        if is_pub_key is True return the public key name, if not
        return the private key name
        ARGS:
            is_pub_key True if the key is public, False if private
            config     config value with the sigType (bit 15) of the key
            slot       the slot number - must be between 3 and 10 inclusive
        """
        if self._is_hmac_sig(config):
            ecdh_ecdsa = "ecdh"
        else:
            ecdh_ecdsa = "ecdsa"

        if is_pub_key:
            pub_prv = "pub"
        else:
            pub_prv = "prv"

        return "{}_{}_slot_{}".format(ecdh_ecdsa, pub_prv, slot)

    def _get_signing_key_with_kts_option(self, config: bytes, slot: int)->bool:
        rc = True
        sign_with_hmac = self._is_hmac_sig(config)
        use_kts_key = self._is_keyRef_bit_set(config)

        if use_kts_key:
            if sign_with_hmac:
                # make sure we have the ECDH pubilc key from the device
                # and if not get it

                if self._ecdh_public_key is None:
                    err_code = self._fetch_ecdh_pub_key()
                    if not err_code._valid_response:
                        self.logger.error('Failed to fetch ECDH_PV_PuK from device; Unable to generate ECDH shared secret!')
                        return False

                if self._ecdh_public_key is None:
                    self.logger.error('INTERNAL failed to set ECDH_PV_PuK after crtm_register_read; Unable to generate ECDH shared secret!')
                    return False

                # now that we know we have ECDH public key from the device set the KTS signing key

                rc = authenta_key_lib.set_kts_hmac_signing_key(self._keylib_context, slot, self._ecdh_public_key)
            else: # sign with ECDSA
                rc = authenta_key_lib.set_kts_ecdsa_signing_key(self._keylib_context, slot)
        else:
            # use the standard funtion to get the crrect signing key
            rc = self._prepare_signing_key(config)

        return rc

    PROVISION_STS_STRUCT = namedtuple('PROVISION_STS_STRUCT',
                                    ["More_Reserved", "More_Reserved1", "More_Reserved2",
                                    "More_Reserved3", "Deprovision_Start", "PCR_Reset",
                                    "Server_Deprovision", "Flash_PCR0_Reset", "UDS_Disabled",
                                    "UDS_Replaced", "RIOT_Mode", "HSM_Present",
                                    "Remotely_Provisioned", "Reserved", "SRK_Replaced",
                                    "Provisioned"]
                                    )

    EXTENDED_CRTM_STATUS_STRUCT = namedtuple('EXTENDED_CRTM_STATUS_STRUCT',
                                            ['Ext_CRTM_Reserved_1', 'Recovery_Mismatch', 'Partial_Measurement',
                                            'Recovery_Mode', 'Ext_CRTM_Reserved_0', 'CRTM_PCR0_Invalid',
                                            'CRTM_Range_Invalid', 'CRTM_Mismatch']
                                            )

    CRTM_REGISTER_DEVICE_CFG_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_STRUCT',
                                                    ["DEV_CFG_RESERVED_5", "Auth_Boot_Partition_Enable_2",
                                                    "Auth_Boot_Partition_Enable_1", "Auth_Boot_Partition_Enable_0",
                                                    "Recovery_Partition_Enable", 
                                                    "DEV_CFG_RESERVED_4", "DEV_CFG_RESERVED_3", "DEV_CFG_RESERVED_2", 
                                                    "DEV_CFG_RESERVED_1", "DEV_CFG_RESERVED_0", 
                                                    "CRK_Read_Disable", "Recovery_Mode_Enable",
                                                    "Auto_Measure_Activate"]
                                                    )

    CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT',
                                                    ["DEV_CFG_VOL_RESERVED_1", "AllowBootReadProtect",
                                                    "Disable_Deprovision", "CSK_Read_Disable",
                                                    "DEV_CFG_VOL_RESERVED_0"]
                                                    )
    PROVISION_STS_FORM = \
'''
Provision_sts:
   Provisioned:           {Provisioned}
   SRK_Replaced:          {SRK_Replaced} 
   Check_Engine:          {Check_Engine}
   Remotely_Provisioned:  {Remotely_Provisioned}
   TPM_Supported:         {TPM_Supported}
   RIOT_Mode:             {RIOT_Mode}
   UDS_Replaced:          {UDS_Replaced}
   UDS_Disabled:          {UDS_Disabled}
'''


    DEVICE_CONFIG_FORM = \
'''
Device_Config bits: 
    Auto Measure Activate:      {Auto_Measure_Activate}
    Recovery Mode Enable:       {Recovery_Mode_Enable} 
    Client Root Key Disabled:   {CRK_Read_Disable}
    Recovery Partition Enable:  {Recovery_Partition_Enable}
    Auth_Boot_Partition_Enable: {Auth_Boot_Partition_Enable_0}{Auth_Boot_Partition_Enable_1}{Auth_Boot_Partition_Enable_2}

'''

    DEVICE_CONFIG_VOLATILE_FORM = \
'''
Device_Config_Volatile bits: 
    CSK_Read_Disable:         {CSK_Read_Disable}
    Disable_Deprovision:      {Disable_Deprovision} 
    AllowBootReadProtect:     {AllowBootReadProtect}
'''


    DEVICE_STATUS_FORM = \
'''
================================================================================
Current Micron Authenta Flash Memory Device Configuration Settings
================================================================================

  UID:      {UID}
  Revision: {Revision}

  Provision_sts attributes:
    Provisioned:            {Provisioned}
    SRK_Replaced:           {SRK_Replaced}
    Remotely_Provisioned:   {Remotely_Provisioned}
    HSM_Present:            {HSM_Present}
    RIOT_Mode:              {RIOT_Mode}
    UDS_Replaced:           {UDS_Replaced}
    UDS_Disabled:           {UDS_Disabled}
    Flash_PCR0_Reset:       {Flash_PCR0_Reset}
    Server_Deprovision:     {Server_Deprovision}
    PCR_Reset:              {PCR_Reset}
    Deprovision_Start:      {Deprovision_Start}

  Extended_CRTM_Status:
    CRTM_Mismatch:          {CRTM_Mismatch}
    CRTM_Range_Invalid:     {CRTM_Range_Invalid}
    CRTM_PCR0_Invalid:      {CRTM_PCR0_Invalid}
    Recovery_Mode:          {Recovery_Mode}
    Partial_Measurement:    {Partial_Measurement}
    Recovery_Mismatch:      {Recovery_Mismatch}

  CRTM Register attributes:
    Power On Mono Count:            {PwrOnCountNV}
    AMTC-1:                         {A_MTC1}

    Device_Config Reg:              {Device_Config}
    Device_Config bits:
      Auto Measure Activated:       {Auto_Measure_Activate}
      Recovery Mode Enabled:        {Recovery_Mode_Enable}
      Client Root Key Readable:     {CRK_Read_Disable}
      Recovery Partition Enabled:   {Recovery_Partition_Enable}
      Auth_Boot_Partition_Enable_0: {Auth_Boot_Partition_Enable_0}
      Auth_Boot_Partition_Enable_1: {Auth_Boot_Partition_Enable_1}
      Auth_Boot_Partition_Enable_2: {Auth_Boot_Partition_Enable_2}

    Device_Config_Volatile Reg:     {Device_Config_Volatile}
    Device_Config_Volatile bits:
      Client Session Key Readable:  {CSK_Read_Disable}
      Deprovision Disabled:         {Disable_Deprovision}
      Allow Boot Read Protect:      {AllowBootReadProtect}

    Defined CRTM Non-Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes): {CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes): {CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes): {CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes): {CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes): {CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes): {CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes): {CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes): {CRTM_Size7}
    
    Defined CRTM Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes): {CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes): {CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes): {CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes): {CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes): {CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes): {CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes): {CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes): {CRTM_Size7}

    MaxAutoMeasureSize:     {CRTM_Max_AutoMeasure_Size}

    Flash Digest Values:
    PCR0:       {Flash_PCR0}
    NV:         {Flash_DigestNV}
    NV Partial: {Flash_DigestNV_Partial}

    Scratchpad Buffer:
        {Customer_RAM}
'''
