#! /usr/bin/env python3

"""
Common functions used across the sample applications and tests.
"""

import ctypes as ct
import hmac
from .defs_h import *
from .cmd_error_code import IndigoCmdErrorCode

class CRTM_Registers():
    """
    CRTM register addresses
    """
    Flash_PCR0              = 0x00
    CRTM_Range_Start_Addr_0 = 0x20
    CRTM_Range_Size_0       = 0x24
    CRTM_Range_Start_Addr_1 = 0x28
    CRTM_Range_Size_1       = 0x2C
    CRTM_Range_Start_Addr_2 = 0x30
    CRTM_Range_Size_2       = 0x34
    CRTM_Range_Start_Addr_3 = 0x38
    CRTM_Range_Size_3       = 0x3C
    CRTM_Range_Start_Addr_4 = 0x40
    CRTM_Range_Size_4       = 0x44
    CRTM_Range_Start_Addr_5 = 0x48
    CRTM_Range_Size_5       = 0x4C
    CRTM_Range_Start_Addr_6 = 0x50
    CRTM_Range_Size_6       = 0x54
    CRTM_Range_Start_Addr_7 = 0x58
    CRTM_Range_Size_7       = 0x5C
    Flash_DigestNV_Partial  = 0x60
    Flash_DigestNV          = 0x80
    Device_Config           = 0xA0
    Provision_Sts           = 0xA4
    CRTM_Max_AutoMeasure_Size = 0xA8
    A_MTC1                  = 0xAC
    KTR_Base                = 0xB0
    EEPROM_Attributes       = 0xD0
    Power_On_CountNV        = 0xD4
    Device_Config_Volatile  = 0xD8
    KTR1_Base               = 0xE0
    Customer_EEPROM         = 0x200
    Customer_RAM            = 0x300
    Public_Key_Table_Slots1_2 = 0x400
    Public_Key_Table_Slots3_4 = 0x500
    Public_Key_Table_Slots5_6 = 0x600
    Public_Key_Table_Slots7_8 = 0x700
    Public_Key_Table_Slots9_10 = 0x800
    ECDSA_DS_PuK            = 0x900
    ECDSA_PV_PuK            = 0x940
    ECDSA_DID_PuK           = 0x980
    Device_Cert_Low         = 0xA00
    Device_Cert_High        = 0xB00
    PCR0_PCR7               = 0xC00

CRTM_Address_Ranges = {
    0: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_0,
        "size": CRTM_Registers.CRTM_Range_Size_0
    },
    1: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_1,
        "size": CRTM_Registers.CRTM_Range_Size_1
    },
    2: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_2,
        "size": CRTM_Registers.CRTM_Range_Size_2
    },
    3: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_3,
        "size": CRTM_Registers.CRTM_Range_Size_3
    },
    4: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_4,
        "size": CRTM_Registers.CRTM_Range_Size_4
    },
    5: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_5,
        "size": CRTM_Registers.CRTM_Range_Size_5
    },
    6: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_6,
        "size": CRTM_Registers.CRTM_Range_Size_6
    },
    7: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_7,
        "size": CRTM_Registers.CRTM_Range_Size_7
    }
}

def enable_automeasure(device):
    """
    Enable automeasure
    return True on success or False on failure
    """
    # read the current device config
    err, response = device.meta_crtm_register_read(b'\x80\x00', CRTM_Registers.Device_Config)
    if not err.esr_and_response_success:
        return False

    device_config = response[318:322]
    device_config_int = int.from_bytes(device_config, byteorder="big")
    data_int = device_config_int | 0x00000001
    data = data_int.to_bytes(4, byteorder="big")
    data = b'\x00'*92+data+b'\x00'*32
    mask = b'\x00'*92+b'\x00'*4+b'\x00'*32

    err, _ = device.meta_acrtm_update_start(b'\x80\x00')
    if not err.esr_and_response_success:
        return False
    
    err, _ = device.meta_crtm_register_write(b'\x80\x00', CRTM_Registers.Device_Config, data, mask)
    if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b'\x80\x00')
        return False

    err, _ = device.meta_acrtm_update_finish(b'\x80\x00')
    if not err.esr_and_response_success:
        return False

    return True

def enable_recovery_partition(device):
    """
    Enable recovery partition
    return True on success or False on failure
    """
    # read the current device config
    err, response = device.meta_crtm_register_read(b'\x80\x00', CRTM_Registers.Device_Config)
    if not err.esr_and_response_success:
        return False

    device_config = response[318:322]
    device_config_int = int.from_bytes(device_config, byteorder="big")
    data_int = device_config_int | 0x00000100
    data = data_int.to_bytes(4, byteorder="big")
    data = b'\x00'*92+data+b'\x00'*32
    mask = b'\x00'*92+b'\x00'*4+b'\x00'*32

    err, _ = device.meta_acrtm_update_start(b'\x80\x00')
    if not err.esr_and_response_success:
        return False
    
    err, _ = device.meta_crtm_register_write(b'\x80\x00', CRTM_Registers.Device_Config, data, mask)
    if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b'\x80\x00')
        return False

    err, _ = device.meta_acrtm_update_finish(b'\x80\x00')
    if not err.esr_and_response_success:
        return False

    return True

def disable_automeasure_and_recovery_partition(device):
    """
    Disable automeasure
    return True on success or False on failure
    """
    # read the current device config
    err, response = device.meta_crtm_register_read(b'\x80\x00', CRTM_Registers.Device_Config)
    if not err.esr_and_response_success:
        return False

    device_config = response[318:322]
    device_config_int = int.from_bytes(device_config, byteorder="big")
    data_int = device_config_int & 0xFFFFFEFE
    data = data_int.to_bytes(4, byteorder="big")
    data = b'\x00'*92+data+b'\x00'*32
    mask = b'\x01'*92+b'\x00'*4+b'\x01'*32

    err, _ = device.meta_acrtm_update_start(b'\x80\x00')
    if not err.esr_and_response_success:
        return False

    err, _ = device.meta_crtm_register_write(b'\x80\x00', CRTM_Registers.Device_Config, data, mask)
    if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b'\x80\x00')
        return False

    err, _ = device.meta_acrtm_update_finish(b'\x80\x00')
    if not err.esr_and_response_success:
        return False

    return True

def set_measurement_range(device, config:bytes, range_list:list):
    """
    Set a CRTM Address Measurement Range Value (start address and length)
    return True on success or False on failure
    """

    # Call Update_Start Option #0 to indicate an update to the CRTM Registers
    err, _ = device.meta_acrtm_update_start(config)
    if not err.esr_success:
        return False

    data = bytearray(b'\x00'*128)
    mask = bytearray(b'\xFF'*128)

    range_id_list = [x[0] for x in range_list]
    data_to_write = [x[1:3] for x in range_list]
    for range_id in range(0, len(range_id_list)):
        addr = CRTM_Address_Ranges[range_id_list[range_id]]["addr"]
        crtm_addr_bytes = data_to_write[range_id][0].to_bytes(4, byteorder='big')[::-1]
        crtm_addr_bytes += data_to_write[range_id][1].to_bytes(4, byteorder='big')[::-1]

        data[addr:(addr+8)] = crtm_addr_bytes
        mask[addr:(addr+8)] = b'\x00'*8
    
    data = bytes(data)
    mask = bytes(mask)

    err, _ = device.meta_crtm_register_write(config, CRTM_ADDRESS_BLOCK_INITIAL_REGS, data[::-1], mask[::-1])
    if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(config)
        return False

    # Call Update_Finish Option #0 to complete the writes to the CRTM registers
    err, _ = device.meta_acrtm_update_finish(config)
    if not err.esr_success:
        return False

    return True


def set_CRTM_Max_AutoMeasure_Size(device, config:bytes, size:int):
    """
    Set CRTM_Max_AutoMeasure_Size to true
    return True on success or False on failure
    """
    # Call Update_Start Option #0 to indicate an update to the CRTM Registers
    err, _ = device.meta_acrtm_update_start(config)
    if not err.esr_success:
        return False

    crtm_addr = CRTM_Registers.CRTM_Max_AutoMeasure_Size
    data = size.to_bytes(4,'big')
    mask = b'\x00'*4
    data = device._create_data_buffer(data, crtm_addr)
    mask = device._create_mask_buffer(mask, crtm_addr)

    err, _ = device.meta_crtm_register_write(config, crtm_addr, data, mask)
    if not err.esr_success:
        return False

    # Call Update_Finish Option #0 to complete the writes to the CRTM registers
    err, _ = device.meta_acrtm_update_finish(config)
    if not err.esr_success:
        return False

    return True

def log_status(device):
    err, _ = device.meta_request_uid()
    if not err.esr_success:
        print('ERROR request_uid: ' + err.error_msg)
        exit(1)

    print("\nDevice Status:")
    print('CRTM_Mismatch: {}'.format(device.extended_crtm_status_struct.CRTM_Mismatch))
    print('CRTM_Range_Invalid: {}'.format(device.extended_crtm_status_struct.CRTM_Range_Invalid))
    print('CRTM_PCR0_Invalid: {}'.format(device.extended_crtm_status_struct.CRTM_PCR0_Invalid))
    print('Recovery_Mode: {}'.format(device.extended_crtm_status_struct.Recovery_Mode))
    print('Partial_Measurement: {}'.format(device.extended_crtm_status_struct.Partial_Measurement))
    print('Recovery_Mismatch: {}'.format(device.extended_crtm_status_struct.Recovery_Mismatch))

    return True
