""" The Python implementation of the platform_driver GRPC client.
    (C) 2019 Micron Technology, Inc. All rights reserved.
"""

import grpc
import authenta.platform_driver_pb2 as platform_driver_pb2
import authenta.platform_driver_pb2_grpc as platform_driver_pb2_grpc

# device context class
class t_authenta_context:
  def __init__(self, ip_address):
    self.ip_address = ip_address

# empty class
class empty:
  def __init__(self):
    pass

def rpc_call(server_addr, request, func_name):
  """ Send rpc call and get response
      @param[in] server_addr  Server address
      @param[in] request      Request message to send
      @param[in] func_name    RPC function name
      return response message on success or None on RPC failure
  """
  response = None

  try:
    with grpc.insecure_channel(server_addr) as channel:
      stub = platform_driver_pb2_grpc.NaxosStub(channel)

      # get rpc function based on function name
      func = getattr(stub, func_name)
      response = func(request)
  except:
    # rpc failure
    print('Error: rpc failure when calling ' + func_name)
    response = empty()

    if func_name == "Authenta_detect_devices":
      response.listsize = 0
    else:
      response.status = 1

  return response

def update_dev_context_return_type(context, response):
  """ Update return_type of device context based on RPC response
      @param[in] context      device context
      @param[in] response     response from RPC call
  """
  context.return_type = response.status

def update_dev_context_out_buf(context, response):
  """ Update out_buf of device context based on RPC response
      @param[in] context      device context
      @param[in] response     response from RPC call
  """
  context.out_buf = response

def update_dev_context_large_out_buf(context, response):
  """ Update large_out_buf of device context based on RPC response
      @param[in] context      device context
      @param[in] response     response from RPC call
  """

  context.large_out_buf = response

def update_dev_context_descriptor(context, response):    
  """ Update descriptor of device context based on RPC response
      @param[in] context      device context
      @param[in] response     response from RPC call
  """
  context.descriptor = response

def update_dev_context_devices_list(context, response):
  """ Update devices_list of device context based on RPC response
      @param[in] context      device context
      @param[in] response     response from RPC call
  """
  context.devices_list = response


# RPC client implementations

# device_descriptor Authenta_device_open(string_msg)
def authenta_device_open(context, path: str):
  request = platform_driver_pb2.string_msg(value = path)
  response = rpc_call(context.ip_address, request, "Authenta_device_open")
  update_dev_context_descriptor(context, response)

  return response

# status_type_msg Authenta_device_close(int_msg)
def authenta_device_close(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Authenta_device_close")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Authenta_device_acquire(int_msg)
def authenta_device_acquire(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Authenta_device_acquire")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Authenta_device_release(int_msg)
def authenta_device_release(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Authenta_device_release")
  update_dev_context_return_type(context, response)

  return response

# list_info Authenta_detect_devices(void_msg)
def authenta_detect_devices(context):
  request = platform_driver_pb2.void_msg()
  response = rpc_call(context.ip_address, request, "Authenta_detect_devices")
  update_dev_context_devices_list(context, response)

  return response

# status_type_msg Write_Enable(int_msg)
def write_enable(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Write_Enable")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Write_Disable(int_msg)
def write_disable(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Write_Disable")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Request_UID(naxos_content)
def request_uid(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Request_UID")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Auth_Modify(naxos_content)
def auth_modify(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Auth_Modify")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Provision_CRTM(naxos_content)
def provision_crtm(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Provision_CRTM")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Deprovision_CRTM(naxos_content)
def deprovision_crtm(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Deprovision_CRTM")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Auth_lock(naxos_content)
def auth_lock(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Auth_lock")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg CRTM_Register_Read(naxos_content)
def crtm_register_read(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "CRTM_Register_Read")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg CRTM_Register_Write(naxos_content)
def crtm_register_write(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "CRTM_Register_Write")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Flash_extend(naxos_content)
def flash_extend(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Flash_extend")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Replace_SRK(naxos_content)
def replace_srk(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Replace_SRK")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg ACRTM_Update_Start(naxos_content)
def acrtm_update_start(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "ACRTM_Update_Start")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg ACRTM_Update_Finish(naxos_content)
def acrtm_update_finish(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "ACRTM_Update_Finish")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Measure(naxos_content)
def measure(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Measure")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg Request_Monotonic_Counter(naxos_content)
def request_monotonic_counter(context, cmd: platform_driver_pb2.naxos_content):
  response = rpc_call(context.ip_address, cmd, "Request_Monotonic_Counter")
  update_dev_context_return_type(context, response)

  return response

# message_return Read_General_Purpose_Readout_Register(int_msg)
def read_general_purpose_readout_register(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Read_General_Purpose_Readout_Register")    
  update_dev_context_out_buf(context, response)

  return response

# message_return Fast_Read(flash_address)
def fast_read(context, addr: platform_driver_pb2.flash_address):
  response = rpc_call(context.ip_address, addr, "Fast_Read")
  update_dev_context_large_out_buf(context, response)

  return response

# message_return Page_Program(flash_write_content)
def page_program(context, content: platform_driver_pb2.flash_write_content):
  response = rpc_call(context.ip_address, content, "Page_Program")
  update_dev_context_out_buf(context, response)

  return response
    
# status_type_msg DYB_Write(flash_lock_config)
def dyb_write(context, lock_conf: platform_driver_pb2.flash_lock_config):
  response = rpc_call(context.ip_address, lock_conf, "DYB_Write")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg PPB_Program(flash_address)
def ppb_program(context, addr: platform_driver_pb2.flash_address):
  response = rpc_call(context.ip_address, addr, "PPB_Program")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg PPB_Erase(int_msg)
def ppb_erase(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "PPB_Erase")
  update_dev_context_return_type(context, response)

  return response

# status_type_msg PPB_Lock_Bit_Write(int_msg)
def ppb_lock_bit_write(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "PPB_Lock_Bit_Write")
  update_dev_context_return_type(context, response)

  return response

# message_return DYB_Read(flash_address)
def dyb_read(context, addr: platform_driver_pb2.flash_address):
  response = rpc_call(context.ip_address, addr, "DYB_Read")
  update_dev_context_out_buf(context, response)

  return response

# message_return PPB_Read(flash_address)
def ppb_read(context, addr: platform_driver_pb2.flash_address):
  response = rpc_call(context.ip_address, addr, "PPB_Read")
  update_dev_context_out_buf(context, response)

  return response

# message_return PPB_Lock_Bit_Read(int_msg)
def ppb_lock_bit_read(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "PPB_Lock_Bit_Read")
  update_dev_context_out_buf(context, response)

  return response

# message_return Password_Read(int_msg)
def password_read(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Password_Read")
  update_dev_context_out_buf(context, response)

  return response

# message_return Read_Status_register(int_msg)
def read_status_register(context, fd: int):
  request = platform_driver_pb2.int_msg(value=fd)
  response = rpc_call(context.ip_address, request, "Read_Status_register")
  update_dev_context_out_buf(context, response)

  return response

# status_type_msg Subsector_Erase(flash_address)
def subsector_erase(context, addr: platform_driver_pb2.flash_address):
  response = rpc_call(context.ip_address, addr, "Subsector_Erase")
  update_dev_context_return_type(context, response)

  return response
