from ..cmd_error_code import CmdErrorCode

class NaxosCmdErrorCode(CmdErrorCode):
  """ Naxos command error parsing implementation.

      The error message is set based on the error specified in the TRD for the
      given ESR and command name.

      An ESR of 0x00 indicates that the error was from a non-Authenta command,
      or a failure happened before an ESR was returned. The error message is
      determined by the non_authenta_err_msg parameter.

      Args:
        esr (bytes): The ESR returned by the device or 0x00 if the failure
          happened before retrieving the ESR from the device, or the command is
          not an Authenta command. 0x80 always indicates success, even for
          non-Authenta commands.
        cmd_name (str): The name of the command that failed.
        non_authenta_err_msg: Caller's error message. Optional, only used for
          non-Authenta errors, i.e. when ESR is 0x00, otherwise it is ignored.
  """
  def __init__(self, esr: bytes, cmd_name: str, non_authenta_err_msg: str="Unknown Error"):
    self._esr = esr
    self._cmd_name = cmd_name
    self._valid_response = None
    self._sig_matched = None
    self._nonce_matched = None
    self._last_opcode_matched = None
    self._esr_bit = None
    self._error_msg = self._msg_lookup(non_authenta_err_msg)

  def _msg_lookup(self, non_authenta_err_msg):
    if self._esr == b'\x80':
      return 'OK'
    elif self._esr == b'\x00':
      return non_authenta_err_msg

    bits = [int(i) for i in bin(int.from_bytes(self.esr, byteorder="big"))[::-1][:-2]]
    ones = [index for index in range(len(bits)) if bits[index] == 1]

    if len(ones) > 1:
      # special ESR (e.g., 0x35, 0x3B, and 0x37 in Auth_Modify)
      self._esr_bit = int.from_bytes(self._esr, byteorder="big")
    else:
      self._esr_bit = ones[0]

    cmd_err_bit_dict = type(self).error_map.get(self.cmd_name)
    if cmd_err_bit_dict:
      ret_msg = cmd_err_bit_dict.get(self._esr_bit)
      base_msg = type(self).std_error_map.get(self._esr_bit)

      if base_msg is not None:
        if ret_msg is None:
          ret_msg = base_msg
        else:
          ret_msg = base_msg + '; or ' + ret_msg

      if ret_msg is None:
        ret_msg = 'Unkonwn error'

    else:
      ret_msg = '{} not found'.format(self._cmd_name)

    return ret_msg

  @property
  def esr(self):
    return self._esr

  @property
  def esr_success(self):
    return self.esr == b'\x80'

  @property
  def error_msg(self):
    return self._error_msg

  @property
  def esr_bit(self):
    """
        Returns:
          ESR bit position (int)
          None for special ESR (e.g., 0x35, 0x3B, and 0x37 in Auth_Modify)
    """
    return self._esr_bit if self._esr_bit < 8 else None

  @property
  def esr_bit_msg(self):
    """
        Returns:
          str: ESR bit position message (string) in the format "ESR bit: ?"
            if special ESR (e.g., 0x35, 0x3B, and 0x37 in Auth_Modify),
            return string in the format: "ESR: ?"
    """
    if self._esr_bit is None:
      return 'Non-Authenta error'

    if self.esr_bit is None:
      return 'ESR: ' + self._esr.hex()
    else:
      return 'ESR bit: ' + str(self.esr_bit)

  @property
  def cmd_name(self):
    return self._cmd_name

  @property
  def valid_response(self):
    return self._valid_response

  def set_valid_response(self, valid: bool):
    self._valid_response = valid

  @property
  def signature_matched(self):
    return self._sig_matched

  def set_signature_matched(self, matched: bool):
    self._sig_matched = matched

  @property
  def nonce_matched(self):
    return self._nonce_matched

  @property
  def esr_and_response_success(self):
    return self.esr_success and self._valid_response

  def set_nonce_matched(self, matched: bool):
    self._nonce_matched = matched

  @property
  def last_opcode_matched(self):
    return self._last_opcode_matched

  def set_last_opcode_matched(self, matched: bool):
    self._last_opcode_matched = matched

  # Starndard errors map (for all command)
  std_error_map = {
      2: 'Bad signature or incorrect payload size or Opcode out of range; or reached the limit of 256 successive non-signed Authenta calls to the device',
      4: 'Counter data mismatch'
  }

  # Individual function errors map
  error_map = dict.fromkeys(['request_uid'], {
      2: 'Failed internal UID inegrity check'
  })
  error_map.update(dict.fromkeys(['provision_crtm'], {
      1: 'Already provisioned',
      5: 'Flash program failure'
  }))
  error_map.update(dict.fromkeys(['deprovision_crtm'], {
      5: 'Flash program or other error condition programming internal flash bit flags, AMTC0, etc.; or locally provisioned without TPM and Disable_Deprovision is set'
  }))
  error_map.update(dict.fromkeys(['acrtm_update_start'], {
      5: 'Option# 1 and CRTM_Range_Invalid set; '
         'or Option #0 and Recovery_Mode set; '
         'or Option #1 and Auto_Measure_Activate cleared; '
         'or Flash program or other error condition programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['auth_lock', 'acrtm_update_finish'], {
      5: 'Flash program or other error condition programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['crtm_register_read'], {
      2: 'crtm_register_address out of range (> 0xFFF); or Config[0] is set and command is not reading any of the CRTM_Addr or CRTM_Size registers',
      5: 'Flash program or other error condition programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['crtm_register_write'], {
      2: 'crtm_update_start was not given when writting to nonvolatile registers; '
         'or crtm_register_address out of range (> 0xFFF); '
         'or device is provisioned for local with TPM and writing to KTR prior to setting CRK_Read_Disable bit; '
         'or EEPROM_Modify bit is clear and writting to EEPROM register; '
         'or Config[0] is set and command is not writting to any of the CRTM_Addr or CRTM_Size registers; '
         'or device is provisioned for local without TPM and writing to KTR prior to settting CSK_Read_Disable bit',
      5: 'Flash program or other error condition programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['flash_extend'], {
      4: 'crtm_range_invalid or recovery_mode set; '
         'or a size of 0x0 is read in the CRTM_Size[0] register; '
         'or CRTM_Addr[0] is outside of the main array address space',
      5: 'Flash program failure programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['replace_srk'], {
      2: 'Device is locally provisioned; '
         'or Option #2/#3 and device is not in RioT mode/UDS is not allowed to access',
      5: 'Not all 8 KTR registers have been written; '
         'or Option #1/#3 and KTR not equal to 0x00...00; '
         'or Flash program or other error condition while programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['measure'], {
      2: 'Payload size incorrect for optional parameters; '
         'or EEPROM address is invlaid based on Config settings; '
         'or device is provisoned for local with TPM and destination is set to KTR and issuing Measure command prior to setting CRK_Read_Disable bit; '
         'or device is provisoned for local without TPM and destination is set to KTR and issuing Measure command prior to setting CSK_Read_Disable bit',
      5: 'Command option relative to riot_mode and prior events not allowed; '
         'or Option #0 and in Recovery_Mode; '
         'or Config[1:0] == `11b; '
         'or EEPROM_Modify bit is clear and writing measurement result to EEPROM register; '
         'or Flash program or other error condition while programming internal flash bit flags, AMTC0, etc.'
  }))
  error_map.update(dict.fromkeys(['auth_modify'], {
      2: 'Source_Addr/Dest_Addr/Size is invalid based on Config settings; '
         'or Option #1 and Size exceeds 0x10000; '
         'or Option #1 and Source_Addr + Size - 1 exceeds the main array address space of the device; '
         'or Option #2 and Size exceeds 0x10000 (for active AP block specified by APBlockAddr matching Dest_Addr)',
      5: 'Device is in Recovery Mode; '
         'or Falsh program or other error condition while programming internal flash bit flags, AMTC0, etc.',
      0x35: 'Attempt to modify non-AM blocks',
      0x37: 'Option #0 and device is in Recovery Block update process and Source_Addr == Recovery_Base and Recovery Block is configured as AM Block and CreateAPBlock is set',
      0x3B: 'Option #0 and CreateAPBlock is set and no more AP blocks may be created'
  }))
  error_map.update(dict.fromkeys(['request_monotonic_counter'], {}))
