from random import _urandom as nonce
import hmac
import logging
from hashlib import sha256
from authenta.authenta_key_lib_interface import set_srk_as_hmac_root_signing_key

import authenta.naxos.utils as sdk2_utils
from .defs_h import *
import authenta.authenta_key_lib_interface as authenta_key_lib
import authenta.authenta_key_mgmt_interface as key_lib_key_mgmt

from ..authenta_device import AuthentaDevice, DeviceType
from ..endpoint_platform import PlatformReturnType, EpPlatformRetCode
from .cmd_error_code import NaxosCmdErrorCode

NAXOS_REGISTER_BLOCK_ADDR_WITH_BOOT_AMT0 = 0xD0


class NaxosDevice(AuthentaDevice):
  """ Base class for Naxos device.

      Define common internal data structure here (e.g., crtm registers)
      Define common functions here
  """
  _cmd_error_code = NaxosCmdErrorCode

  def __init__(self, endpoint_platform, logger=logging.getLogger(__name__), user_defined_key_driver_data=None, user_defined_key_driver_dev_info_dict=None):
    super().__init__(endpoint_platform, logger, user_defined_key_driver_data, user_defined_key_driver_dev_info_dict)
    self.device_type = DeviceType.NAXOS
    self.initialize_internal_properties()
    self._server_session_key
    if self._key_driver_dev_info_dict is None:
        self._key_driver_dev_info_dict = { }
    self._key_driver_dev_info_dict['device_type'] = 'NAXOS'
    self._key_driver_dev_info_dict['device_version'] = 'UNKNOWN'

  def __enter__(self):
    super().__enter__()
    # initializing structures is needed because we don't know
    # if the underlying hardware device has been rebooted or not
    # so we cant trust boot_amtc0, amt1, etc.
    self.initialize_internal_properties()

    # read client session/root key if needed
    err, *_ = self.meta_request_uid()
    if not err.esr_success:
      self.logger.error('meta_request_uid Failed: %s', err.error_msg)
      return


  def initialize_internal_properties(self):
    """ Set class variables to None

        If this class is not being used outside a 'with' statement this MUST
        be called whenever the hardware device this device object has been
        rebooted.
    """
    super().initialize_internal_properties()
    self._amtc0 = None
    self._boot_amtc0 = None
    self._amtc1 = None
    self._provision_sts_struct = None
    self._extended_crtm_status_struct = None
    self._device_config_volatile_struct = None
    self._revision = None
    self._CRTM_reg_read_response_lst = None
    self._crtm_reg_data_bytes = None
    self._crtm_reg_data_table = None
    self._server_session_key = None
    self._client_keys_set = False

  @staticmethod
  def binary_str2attribute_lst(bstr, idx_lst):
    segment_lst = list()
    for segment in idx_lst:
      segment_lst.append(bstr[:segment])
      bstr = bstr[segment:]

    return segment_lst

  @staticmethod
  def reg_bytes2struct(reg_hex_str, reg_struct_def):
    bin_len = len(reg_hex_str) * 4
    bit_str = '{:b}'.format(int(reg_hex_str, 16))
    bit_str = '0' * (bin_len - len(bit_str)) + bit_str
    binstr_lst = NaxosDevice.binary_str2attribute_lst(bit_str, reg_struct_def)

    return binstr_lst

  @property
  def uid(self):
    return sdk2_utils.change_word_order(self._uid) if self._uid is not None else None

  @property
  def uid_as_str(self):
    return self.uid.hex()

  @property
  def raw_uid(self):
    return self._uid

  @property
  def revision(self):
    return self._revision

  @property
  def amtc0(self):
    return self._amtc0

  @property
  def boot_amtc0(self):
    return self._boot_amtc0

  @property
  def amtc1(self):
    return self._amtc1

  @property
  def device_config_volatile(self):
    return self._device_config_volatile_struct

  @property
  def server_session_key(self):
    return self._server_session_key

  @property
  def provision_sts_struct(self):
    return self._provision_sts_struct

  @property
  def extended_crtm_status_struct(self):
    return self._extended_crtm_status_struct

  @property
  def CRTM_reg_read_response_lst(self):
    return self._CRTM_reg_read_response_lst

  @property
  def crtm_reg_data_bytes(self):
    return self._crtm_reg_data_bytes

  @property
  def crtm_reg_data_table(self):
    return self._crtm_reg_data_table

  @property
  def Device_CFG(self):
    if self.crtm_reg_data_table is not None:
      reg_bytes = self.crtm_reg_data_table['Device_Config']
      reg_bit_attr = self.reg_bytes2struct(reg_bytes, self.CRTM_Reg_Device_CFG_struct)
      reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
      return self.CRTM_REGISTER_DEVICE_CFG_STRUCT(*reg_bit_attr_int)

    return None

  @property
  def Device_CFG_Volatile(self):
    if self.crtm_reg_data_table is not None:
      reg_bytes = self.crtm_reg_data_table['Device_Config_Volatile']
      reg_bit_attr = self.reg_bytes2struct(reg_bytes, self.CRTM_Reg_Device_CFG_Vol_struct)
      reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
      return self.CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(*reg_bit_attr_int)

  @property
  def keylib_context(self):
    return self._keylib_context

    return None

  def get_device_revision(self)->bytes:
    """ Retrieve the Authenta device revision.

        Return the device's revision returned from calling meta_request_uid
        on the device. This can (and is meant to be) called before calling
        __init__(), i.e. without using Python 'with', and it can be called
        on the NaxosDevice class.

        The idea of this function is that a NaxosDevice can be created
        before knowing what the revision of the attached device is, and this
        function will be used to get the revision. Then the proper class
        associated with the revision can be instantiated.

        If the platform device has not been acquired before calling this
        function, it will be acquired and released in this function. If the
        platform device has already been acquired, this function will not
        attempt to acquire the device.

        Returns:
          the (3 byte) device revision returned in the request_uid command,
            or None if retrieving it from the device fails.
    """
    if self._revision is not None:
      return self._revision

    if self.endpoint_platform is None:
      return None

    try:
      acquired_device_here = False
      if not self.endpoint_platform.is_device_acquired():
        self.endpoint_platform.acquire_device(self.device_type)
        acquired_device_here = True

      err, *_ = self.meta_request_uid()
      if not err.esr_success:
        self.logger.info('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
    finally:
      if acquired_device_here:
        self.endpoint_platform.release_device(self.device_type)

    return self.revision

  def _sign_cmd(self, cmd_str, sign_with_session_key=True, sign_with_client_root_key=False, additional_data=None):
    """ Sign a command.

        This function assumes self.provision_sts_struct is set. If it is not
        set before callng this command, meta_request_uid MUST be called
        before calling this function.

        Arguments:
          cmd_str                    the command string to be signed
          sign_with_session_key      this is true of all commands except for provision, and sometimes deprovision
          sign_with_client_root_key  this is false for all commands except sometimes deprovision
          additional_data            additional data to be signed along with commands - used in repalce_srk

        Returns:
          bytes: command signature
    """
    key_lib_sig = None
    str_to_sign = (cmd_str + additional_data) if additional_data else cmd_str

    self.logger.info('CMD bytes to be signed:{}'.format(str_to_sign.hex()))

    if sign_with_session_key:
      # true for most commands
      assert(self.provision_sts_struct is not None)

      if self.provision_sts_struct.Remotely_Provisioned:
        # remote provision
        # derive session key (if necessary)
        assert(self._boot_amtc0 is not None)
        self.logger.info('for session key')
        self.logger.info('boot_time_amtc0:{}'.format(self._boot_amtc0.hex()))

        if not self._session_key_externally_set:
          if not self._root_signing_key_externally_set:
            kl_call_ok = authenta_key_lib.set_srk_as_hmac_root_signing_key(self._keylib_context)
            if not kl_call_ok:
                self.logger.error('Failed attempting to set SRK as signing key in Authenta Key Library')

          kl_call_ok = authenta_key_lib.derive_hmac_session_key(self._keylib_context, self.boot_amtc0)
          if not kl_call_ok:
            self.logger.error('Failed attempting to set Session key in Authenta Key Library')

      else:
        # local provision
        if not self._set_client_keys():
          self.logger.error('Failed attempting to set Client Session Key as signing key in Authenta Key Library')

      key_lib_sig = authenta_key_lib.gen_session_sha256_hmac(self._keylib_context, str_to_sign)
    else:
      # provision and sometimes deprovision are signed with the SRK or CRK directly
      if not sign_with_client_root_key:
        # typical case...
        # sign with server root Key
        if not self._root_signing_key_externally_set:
          kl_call_ok = authenta_key_lib.set_srk_as_hmac_root_signing_key(self._keylib_context)
          if not kl_call_ok:
            self.logger.error('Failed attempting to set SRK as signing key in Authenta Key Library')
          else:
            # in the  case of deprovison, sometimes we can sign with SRK on a locally provisioned
            # device, so make sure we don't think client keys are set
            self._client_keys_set = False

      elif not self._root_signing_key_externally_set and not self._client_keys_set: # if not previously set
        # this is a special case for deprovision of locally provisioned
        # and SRK signature was explicitly requested in the config
        kl_call_ok = authenta_key_lib.set_crk_as_hmac_root_signing_key(self._keylib_context)
        if not kl_call_ok:
          self.logger.error('Failed attempting to set CRK as signing key in Authenta Key Library')

      key_lib_sig = authenta_key_lib.gen_root_sha256_hmac(self._keylib_context, str_to_sign)
    if key_lib_sig is None:
      self.logger.warning('Failed to create signature in key library')
      key_lib_sig = b''
    else:
      self.logger.debug('Signature of Input Bytes:{}'.format(key_lib_sig.hex()))

    return key_lib_sig

  def _response_sig_is_valid(self, response: bytes, sign_with_session_key: bool=False):
    """ Verify response signature.

        Returns:
          bool: True on success or False on failure
    """
    signing_str = response[:32]
    response_sig = self.get_response_signature(response)

    if sign_with_session_key:
      rc = self.get_boot_amtc0_if_not_set('_response_sig_is_valid')
      if not rc.esr_success:
        self.logger.warning('Unable to verify response signature: ' + rc.error_msg)
        return False

      if self.provision_sts_struct is None:
        rc, *_ = self.meta_request_uid()
        if not rc.esr_success:
          self.logger.warning('Unable to verify response signature; attempt to retrieve provision_sts_struct by calling meta_request_uid Failed: ' + rc.error_msg)
          return False

    if self.provision_sts_struct.Remotely_Provisioned or not self._provision_sts_struct.Provisioned:
      # remote provision
      #
      # (or if it is not provisioned and the command was 
      # Request_UID - the only signed response for a 
      # non-provisioned device and the only time this function 
      # would called for a non-provisioned device)
      # On a non-provisioned device Request_UID is always signed with the SRK

      # commenting the following line out because this shouldn't be required now,
      # this would have been done before signing the command...
      if not (sign_with_session_key and self._session_key_externally_set):
        if not self._root_signing_key_externally_set:
          if authenta_key_lib.set_srk_as_hmac_root_signing_key(self._keylib_context):
            self.logger.info('set SRK as signing key')
          else:
            self.logger.error('Failed to set signing key (SRK)!')
            return False

      if sign_with_session_key and not self._session_key_externally_set:
        if authenta_key_lib.derive_hmac_session_key(self._keylib_context, self._boot_amtc0):
          self.logger.debug('successful call to authenta_key_lib.derive_hmac_session_key')
        else:
          self.logger.error('Failed to generate session key while attempting to  verify command response!')
          return False
    else:
      # local provision

      if not self._set_client_keys():
        self.logger.error('Failed to set signing key CRK or CSK!')
        return False

    if sign_with_session_key:
      verified = authenta_key_lib.verify_session_sha256_hmac(self._keylib_context, signing_str, response_sig)
    else:
      # signed with server root Key
      verified = authenta_key_lib.verify_root_sha256_hmac(self._keylib_context, signing_str, response_sig)

    if verified is None:
      self.logger.warning('Failed to verify response signature in key library')
      verified = False
      
    self.logger.info("signature verification result: {}".format(verified))
    return verified

  # Atomic Commands

  def atomic_request_uid(self, config: bytes, nonce_in: bytes=None):
    """ Request_UID command implementation.

        Update internal data (e.g., amtc0)

        Args:
          config (bytes)
          nonce (bytes): Optional 4-byte bytes object holding a nonce to use
            in the Request_UID command sent to the Naxos device. If no nonce
            is provided, one will be generated in this function and returned.

        Returns:
          PlatformReturnType, the nonce (bytes) sent with the command
    """

    if nonce_in is not None:
      if not isinstance(nonce_in, bytes):
        self.logger.error('input argument is of wrong type; nonce must be of type bytes!')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
      if len(nonce_in) != 4: 
        self.logger.error('input argument is of wrong length; nonce a bytes of length 4!')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
      a_nonce = nonce_in
    else:
      a_nonce = bytes.fromhex(nonce(4).hex())
    # format {opcode}{sub_opcode}{config}{nonce}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3E') + config + a_nonce

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

  def atomic_read_gprr(self):
      """ Read General Purpose Response Register implementation

          Returns:
            PlatformReturnType
      """
      # format {opcode}
      cmd_str = bytes.fromhex('96')

      return self.endpoint_platform.exec_on_device(self.device_type, cmd_str, 'read_gprr')

  def atomic_crtm_register_read(self, config: bytes, crtm_addr: int, amtc0_in: bytes=None):
    """ CRTM_Register_Read implementation

        Args:
          config (bytes)
          crtm_addr (int)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if crtm_addr is None:
      self.logger.error('Missing required data to run command: CRTM Address')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}{crtm_addr}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('38') + config + use_amtc0 + crtm_addr.to_bytes(2, byteorder='big')

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_crtm_register_write(self, config: bytes, crtm_addr: int, data: bytes, sign: bool, amtc0_in: bytes=None):
    """ CRTM_Register_Write implementation

        Args:
          config (bytes)
          crtm_addr (int): adrees of where the data it to be written
          data (bytes): data to be written. Exactly 4 bytes
          sign (bool): If True, the command will be signed - otherwise, it
            won't. This should indicate whether the register being written to
            is volatile. Set to True for nonvolatile registers, False for
            volatile registers.
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if crtm_addr is None:
      self.logger.error('Missing required data to run command: CRTM Address')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if data is None:
      self.logger.error('Missing required data to run command: no input data to write')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}{crtm_address}{data}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3C') + config + use_amtc0 + crtm_addr.to_bytes(2, byteorder='big') + data[0:4]

    # bit 0 on of the config indicates a volaltile write - no signature needed
    if sign:
      if self.provision_sts_struct is None: # needed for signing
        self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

      if not self.boot_amtc0: # needed for signing
        self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
      sig = self._sign_cmd(cmd_str)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_request_monotonic_counter(self, addr: int):
    """ Request_Monotonic_Counter command implementation

        Args:
          addr (int)

        Returns:
          PlatformReturnType
    """
    reserved = b'\x00'
    tag = bytes([
      0x55, 0x55, 0x55, 0x55, 0x00, 0x00, 0x00, 0x00, 0xaa, 0xaa, 0xaa, 0xaa
    ])

    # format {opcode}{sub_opcode}{addr}{reserved}{tag}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('03') + addr.to_bytes(1, byteorder='big') + reserved + tag

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_acrtm_update_start(self, config: bytes, amtc0_in: bytes=None):
    """ ACRTM_Update_Start command implementation

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """
    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.boot_amtc0 is None: #needed for signing
      self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.provision_sts_struct is None: # needed for signing
      self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3B') + config + use_amtc0

    # sign the command
    sig = self._sign_cmd(cmd_str)
    cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_acrtm_update_finish(self, config: bytes, amtc0_in: bytes=None):
    """ ACRTM_Update_Finsih command implementation.

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.boot_amtc0 is None: #needed for signing
      self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.provision_sts_struct is None: # needed for signing
      self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3D') + config + use_amtc0

    # sign the command
    sig = self._sign_cmd(cmd_str)
    cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_provision_crtm(self, config: bytes, nonce_in: bytes, amtc0_in: bytes=None, signature_func=None):
    """ Provision_CRTM command implementation.

        Args:
          config (bytes)
          nonce (bytes): Optional 4-byte bytes object holding a nonce to use in
            the Request_UID command sent to the Naxos device. If no nonce is
            provided, one will be generated in this function, and returned.

          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType, the nonce (bytes) sent with the command
    """
    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')

    if nonce_in is not None:
      if not isinstance(nonce_in, bytes):
        self.logger.error('input argument is of wrong type; nonce must be of type bytes!')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None
      if len(nonce_in) != 4: 
        self.logger.error('input argument is of wrong length; nonce a bytes of length 4!')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR), None
      a_nonce = nonce_in
    else:
      a_nonce = bytes.fromhex(nonce(4).hex())
    # format {opcode}{sub_opcode}{config}{nonce}{amtc_0}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('35') + config + a_nonce + use_amtc0

    if signature_func is not None:
      cmd_str = cmd_str + signature_func(self, cmd_str)

    elif (config[0] & 0b00000001):
      sig = self._sign_cmd(cmd_str, False)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

  def atomic_deprovision_crtm(self, config: bytes, amtc1_in: bytes=None, signature_func=None):
    """ Deprovision_CRTM command implementation

        Args:
          config (bytes)
          amtc1_in (int): ONLY FOR DEBUG
          signature_func: the function to use to generate the command signature.
                          the function interface is
                            func_name(device: AuthentaDevice, command: str)->bytes
                            
                            where "device" is this object's "self" variable and
                                  "command" is command to sign formatted as a hex string
                                  and the functions returns the signature as a bytes object

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc1_in:
      use_amtc1 = amtc1_in
    else:
      use_amtc1 = self.amtc1

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if not self.provision_sts_struct:
      # need this information to determince if the command must be signed
      self.logger.error('Missing required data to run command: device provision status; run read_extended_crtm_status before deprovision_crtm')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if not use_amtc1:
      self.logger.error('Missing required data to run command: amtc0; Run another command (e.g. read_extended_crtm_status) before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_1}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('36') + config + use_amtc1

    if signature_func is not None:
          cmd_str = cmd_str + signature_func(self, cmd_str)

    # check if the command must be signed. It does not have to
    # be if the device is locally provisioned without TPM
    elif self.provision_sts_struct.Remotely_Provisioned or self.provision_sts_struct.TPM_Supported:
      sign_with_client_root_key = False

      # if the device is locally provisioned check the Allow_SRK flag (config[1]).
      # if this bit is not set, sign the command using the client root key;
      # else: sign command using the server root key
      if not self.provision_sts_struct.Remotely_Provisioned and not (config[0] & 0b00000010):
        sign_with_client_root_key = True

      sig = self._sign_cmd(cmd_str, sign_with_session_key=False, sign_with_client_root_key=sign_with_client_root_key)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_auth_lock(self, amtc0_in: bytes=None):
    """ Auth_Lock command implementation

        Args:
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """
    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.boot_amtc0 is None: #needed for signing
      self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if self.provision_sts_struct is None: # needed for signing
      self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('37') + b'\x00' + use_amtc0

    # sign the command
    sig = self._sign_cmd(cmd_str)
    cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_flash_extend(self, config: bytes, amtc0_in: bytes=None):
    """ Flash_Extend command implementation

        Args:
          config (bytes)
          amtc0 (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """
    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode}{sub_opcode}{config}{amtc_0}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3A') + config + use_amtc0

    # sign the command for options #1 (config[2]:[0]=b'01') and #3 (config[2]:[0]=b'11'),
    # but not for options #0 (config[2]:[0]=b'00') or #2 (config[2]:[0]=b'10')
    if ((config[0] & 0b00000101) == 1) or ((config[0] & 0b00000101) == 5): # option #0 or option #3
      if not self.boot_amtc0: #needed for signing
        self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
      if self.provision_sts_struct is None: # needed for signing
        self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
      sig = self._sign_cmd(cmd_str)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)

  def atomic_measure(self, config: bytes, eeprom_alt_src_addr: int, eeprom_key_src_addr: int,
                      eeprom_dest_addr: int, in_digest: bytes=None, amtc0_in: bytes=None):
    """ Measure command implementation construct command

        Args:
          config (bytes): length must be exactly 2 bytes
          eeprom_alt_src_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used as data to include into the
            command measurement.
          eeprom_key_src_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used as a key when performing an
            HMAC calculation.
          eeprom_dest_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used to store the result of a
            Measure command.
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error('Missing required data to run command: config value')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if eeprom_alt_src_addr is None or eeprom_alt_src_addr > 0xFFFF:
      self.logger.error('Missing required data to run command: eeprom_src_addr; or eeprom_alt_src_addr value too large to fit in 2 bytes')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if eeprom_key_src_addr is None or eeprom_key_src_addr > 0xFFFF:
      self.logger.error('Missing required data to run command: eeprom_key_addr; or eeprom_key_src_addr value too large to fit in 2 bytes')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if eeprom_dest_addr is None or eeprom_dest_addr > 0xFFFF:
      self.logger.error('Missing required data to run command: eeprom_dest_addr; or eeprom_dest_addr value too large to fit in 2 bytes')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if in_digest and len(in_digest) != 32:
      self.logger.error('in_digest value should be 32 bytes')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error('Missing required data to run command: amtc0; Run request_uid before running this command')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode[1]}{sub_opcode[1]}{config[2]}{amtc_0[4]}{eeprom_altSource_Addr[2]}{eeprom_keySource_addr[2]}{eeprom_dest_addr[2]}{input_digest[32]}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('3F') + config + use_amtc0 + eeprom_alt_src_addr.to_bytes(2, byteorder='big') + eeprom_key_src_addr.to_bytes(2, byteorder='big') + eeprom_dest_addr.to_bytes(2, byteorder='big') 

    #if bit 10 of config is set, include digest in measurement, and don't sign
    if(config[1] & (1 << 6)):
      if not in_digest:
        self.logger.error('Missing in_digest while bit 10 of digest is set')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
      cmd_str = cmd_str + in_digest

    if(config[1] & (1 << 6) and config[1] & (1 << 7)):
      self.logger.error('Badly formed config, option can only have either bit 6 or 7 set but not both')
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    if(config[1] & 1 << 7):  # bit 7 is set
      if not self.boot_amtc0: #needed for signing
        self.logger.error('Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

      if self.provision_sts_struct is None: # needed for signing
        self.logger.error('Missing required data to sign command: provision_sts_struct; To retrieve provision_sts_struct run request_uid')
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

      sig = self._sign_cmd(cmd_str)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)


  # Meta Commands

  def meta_request_uid(self, config: bytes=b'\x00', nonce: bytes=None):
    """ Read and return UID, AMTC0, and provision_sts. 

        On success, the associated class variables will be populated with 
        the returned results.

        If the sts_provision register check_engine flag (provision_sts[2])
        returned with the request_uid response is set, request_uid will be called
        again in this function. (See sts_provision check_engine bit in the TRD).
        Request_uid will be called again only if the response is seccesfully 
        validated, includingn signature validation, and will be called an
        additional 2 times even if the check_engine flag continues to be set in
        the response.

        If provision_sts[2] (check_engine) bit is set, it 'indicates that
        internal service is needed' and 'a 'dummy' request_UID command must be
        issued'. See section 3.4.3.1, 'Provision_Sts[2]' check_engine' in the
        TRD. To avoid the possibility of an infinite loop here, we run
        request_uid an 2 additional times, even if the check_engine flag
        continues to return as set - there is probably something else wrong if
        that happens.

        Args:
          config (bytes): Optional

          nonce (bytes): Optional A 4-bytes nonce to use in the Request_UID
            command sent to the Naxos device. If not provided, one will be
            randomely generated.

        Returns:
          Tuple:
            NaxosCmdErrorCode,
            response (bytes) - may equal None on failure),
            nonce    (4-byte bytes) - nonce sent with the command
    """
    times_through_while_loop = 0
    run_again_check_engine_set = True
    while run_again_check_engine_set:
      run_again_check_engine_set = False
      error_code = None
      atomic_ret_obj, sent_nonce = self.atomic_request_uid(config, nonce)
      if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
        error_code = self._cmd_error_code(b'00', 'request_uid', 'request_uid failed before setting GPRR')
        return error_code, None, sent_nonce

      error_code, uid_response_bytes = self.meta_read_gprr('request_uid', 0x3E, False)

      # test to see if command returned good status
      if error_code.esr_success:
        self.logger.info('Data-in:{}'.format(uid_response_bytes.hex()))

        uid_respones_bytes_struct = self._get_uid_response_struct(uid_response_bytes)
        ret_nonce = self._get_uid_attr('Nonce', uid_respones_bytes_struct)

        # data below is needed in _response_sig_is_valid() 
        # parse provision_sts byte
        b_lst = (list('{:08b}'.format(uid_response_bytes[3])))
        b_lst_int = [int(d) for d in b_lst]
        self._provision_sts_struct = self.PROVISION_STS_STRUCT(*b_lst_int)

        self._uid = self._get_uid_attr('UID', uid_respones_bytes_struct)
        self._revision = self._get_uid_attr('Revision', uid_respones_bytes_struct)

        # now that we have the UID open the key driver if it hasn't been opened yet
        if self._keylib_context is None:
          self._key_driver_dev_info_dict['device_revision'] = self._revision.hex()             
          self._open_key_driver()

        # verify that the response is valid
        signature_checks = self._response_sig_is_valid(uid_response_bytes)

        nonce_matched = (sent_nonce == ret_nonce)

        if not signature_checks:
          self.logger.warning('Response validation failed signature check!!!')
        if not nonce_matched:
          self.logger.warning('Response validation failed nonce matching!!!')

        error_code.set_signature_matched(signature_checks)
        error_code.set_nonce_matched(nonce_matched)
        error_code.set_valid_response(signature_checks and nonce_matched)

        # if the check_engine flag is set run request_uid again.
        # BUT, ONLY if the signature and other checks are valid
        # and therefore trust the response
        if self._provision_sts_struct.Check_Engine == 1 \
          and error_code.valid_response \
          and times_through_while_loop < 2:
          run_again_check_engine_set = True
          times_through_while_loop += 1
          self.logger.info('porvision_sts[2], check_engine, is set; resubmitting request_UID to device')

        self.logger.debug('Request_UID info:')
        self.logger.debug('Device UID:{}'.format(self.uid))
        self.logger.debug('Revision:  {}'.format(self.revision.hex()))
        self.logger.debug('A-MTC0:    0x{}'.format(self.amtc0.hex()))
        self.logger.debug('Nonce:     {}'.format(ret_nonce.hex()))
        self.logger.debug('Signature: {}'.format(self._get_uid_attr('Signature', uid_respones_bytes_struct).hex()))
      else:
        self.logger.error('Command returned error:0x{}'.format(error_code.esr.hex()))

    return error_code, uid_response_bytes, sent_nonce

  def meta_read_gprr(self, command_name: str, sub_opcode: int, do_response_verification: bool=True):
    """ Read and return GPRR data.

        Args:
          command_name (str): Command name in lower case
          sub_opcode (int)
          do_response_verification (bool)

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    atomic_response = self.atomic_read_gprr()
    if atomic_response.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'read_gprr', 'failed to read GPRR')
      return error_code, None

    response_bytes = atomic_response.buffer
    ESR = response_bytes[0]

    error_code = self._cmd_error_code(ESR.to_bytes(1, byteorder='big'), command_name)

    if ESR == 0x80:
      self.logger.info(command_name + ' Data-in:{}'.format(response_bytes.hex()))

      # if signature_checks and last_auth_cmd_matches:
      if command_name != 'request_monotonic_counter':
        self.set_amtc0(response_bytes)
      else:
        self._amtc0 = response_bytes[0x0D:0x11]

      # request_uid and request_monotonic_counter does not return extended_crtm_status and 
      # always has to handle their own verification because the response
      # structure is non-standard
      if command_name not in ['request_uid', 'request_monotonic_counter']:
        self.set_amtc1(response_bytes)
        self.set_extended_crtm_status(response_bytes[1:3])
        if command_name is 'provision_crtm':
              self._boot_amtc0 = self._amtc0
        if do_response_verification:
          self._do_response_verification_checks(response_bytes, sub_opcode, error_code)

    else:
      self.logger.warning(command_name + ' command returned error:0x{:02X}'.format(ESR))

    return error_code, response_bytes

  def meta_crtm_register_read(self, config: bytes, crtm_addr: int, amtc0_in: bytes=None, verify_sig: bool=True):
    """ Read and return a 16 byte register from the Authenta chip.

        If this function is called the register containing the powerOnCounter
        (Boot AMTC0, address 0x0D) the boot_amtc0 of this object will be updated
        to the value in the read register.

        Args:
          config (bytes)
          crtm_addr (int)
          amtc0_in (int): ONLY FOR DEBUG
          verify_sig (bool)

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """

    if crtm_addr is None:
      return self._cmd_error_code(b'\x00', 'crtm_register_read', 'Missing required data to run command: CRTM Address')

    # check if amtc0 has been read, and if not, load it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_crtm_register_read', amtc0_in)
    if not rc.esr_success:
      return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_crtm_register_read(config, crtm_addr, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'crtm_register_read', 'crtm_register_read failed before setting GPRR')
      return error_code, None

    # if this is reading the boot_amtc0 and it not been set yet,
    # we would need to set boot_amtc0 to verify the response signature.
    # To avoid sending us into a loop if boot_amtc0 is not set
    # don't do the verification meta_read_gprr. Set the boot_amtc0
    # first, then do the check, and finally set it back if response
    # verification fails.
    verify_resp_in_get_gprr = True
    if crtm_addr == NAXOS_REGISTER_BLOCK_ADDR_WITH_BOOT_AMT0 or not verify_sig:
      verify_resp_in_get_gprr = False

    error_code, response_bytes = self.meta_read_gprr('crtm_register_read', 0x38, verify_resp_in_get_gprr)

    # test to see if command returned good status
    if error_code.esr_success:
      # set new boot_amtc0
      if crtm_addr == NAXOS_REGISTER_BLOCK_ADDR_WITH_BOOT_AMT0:
        register_data = self.get_response_payload(response_bytes)

        # format of register data is 
        # EEPROM_Attr | PwrOnCount | Device_Config_Volatile | Reserved 
        # 4 bytes each                
        self._boot_amtc0 = register_data[4:8] # 4 bytes
        self.logger.info('PowerOn A-MTC0 set to: {}'.format(self._boot_amtc0.hex()))

        # if locally provisioned with no TPM then 
        if not self.provision_sts_struct.Remotely_Provisioned and not self.provision_sts_struct.TPM_Supported:
          self.set_device_config_volatile(response_bytes[24:28])

        # verify response signagure if not verifed in meta_read_gpr
        if not verify_resp_in_get_gprr:
          self._do_response_verification_checks(response_bytes, 0x38, error_code)

    return error_code, response_bytes

  def meta_crtm_register_write(self, config: bytes, crtm_addr: int, data: bytes, sign: bool, amtc0_in: bytes=None):
    """ Write a 4 byte register to the Authenta chip.

        Args:
          config (bytes)
          crtm_addr (int): Address of where the data is to be written.
          data (bytes): Data to be written. Must be exactly 4 bytes.
          sign (bool): If True, the command will be signed. Otherwise, it
            won't. This should indicate whether the register being written
            to is volatile or not. Set to True for nonvolatile registers,
            False for volatile registers.
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """

    # make sure we have the info needed to call the function
    paramter_err = None
    if config is None:
      paramter_err = self._cmd_error_code(b'\x00', 'crtm_register_write', 'Missing required data to run command: config value')
    elif crtm_addr is None:
      paramter_err = self._cmd_error_code(b'\x00', 'crtm_register_write', 'Missing required data to run command: CRTM Address')
    elif data is None:
      paramter_err = self._cmd_error_code(b'\x00', 'crtm_register_write', 'Missing required data to run command: missing input data to write')

    if paramter_err:
      return paramter_err, None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_crtm_register_write', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling read_extended_crtm_status 
    rc = self.get_boot_amtc0_if_not_set('meta_crtm_register_write')
    if not rc.esr_success:
      return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_crtm_register_write(config, crtm_addr, data, sign, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'crtm_register_write', 'crtm_register_write failed before setting GPRR')
      return error_code, None

    gprr_error_code, response_bytes = self.meta_read_gprr('crtm_register_write', 0x3C)

    return gprr_error_code, response_bytes

  def meta_request_monotonic_counter(self, addr: int):
    """ Read and return monotonic counter value from the Authenta chip.

        Args:
          addr (int): 0x80 for reading amtc0 or 0x81 for reading amtc1

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    if addr is None:
      return self._cmd_error_code(b'\x00', 'request_monotonic_counter', 'Missing required data to run command: Address')

    error_code = None
    atomic_ret_obj = self.atomic_request_monotonic_counter(addr)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'request_monotonic_counter', 'request_monotonic_counter failed before setting GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr('request_monotonic_counter', 0x03, False)

    # test to see if command returned good status
    if error_code.esr_success:
      # addr 0x80 will return amtc0 value
      # addr 0x81 will return amtc0 and amtc1 value
      if addr == 0x81:
        amtc1 = response_bytes[17:21]
      else:
        amtc1 = None

      # verify that the response if valid
      # addr: 0x80: signature = HMAC(Tag || A-MTC0) Ksn(*)
      # addr: 0x81: signature = HMAC(Tag || A-MTC0 || A-MTC1) Ksn(*)
      if addr != 0x81:
        signing_str = response_bytes[1:17]
        response_sig = response_bytes[17:49]
      else:
        signing_str = response_bytes[1:21]
        response_sig = response_bytes[21:53]

      signature_checks = False
      continue_flag = True

      if self.provision_sts_struct is None:
        error_code, *_ = self.meta_request_uid()
        if not error_code.esr_success:
          self.logger.warning('Unable to verify response signature; attempt to retrieve provision_sts_struct by calling meta_request_uid Failed: ' + error_code.error_msg)
          continue_flag = False

      if continue_flag:
        ret = self.get_boot_amtc0_if_not_set('meta_request_monotonic_counter')
        if not ret.esr_success:
          self.logger.warning('Unable to get boot amtc0: ' + ret.error_msg)
          continue_flag = False

      if continue_flag:
        if not self.provision_sts_struct.Remotely_Provisioned:
          # local provision
          self._set_client_keys()

        verified = authenta_key_lib.verify_session_sha256_hmac(self._keylib_context, signing_str, response_sig)
        if verified is None:
          self.logger.warning('Failed to verify response signature in key library')
          verified = False
            
        signature_checks = verified

      if not signature_checks:
        self.logger.warning('response validation failed signature check!!!')
      else:
        if amtc1 is not None:
          self._amtc1 = amtc1

      error_code.set_signature_matched(signature_checks)
      error_code.set_valid_response(signature_checks)

    else:
      self.logger.warning('Command returned error:0x{}'.format(error_code.esr.hex()))

    return error_code, response_bytes

  def meta_acrtm_update_start(self, config: bytes, amtc0_in: bytes=None):
    """ Start ACRTM update mode

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """

    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'acrtm_update_start', 'Missing required data to run command: config value'), None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_acrtm_update_start', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling read_extended_crtm_status 
    rc = self.get_boot_amtc0_if_not_set('meta_acrtm_update_start')
    if not rc.esr_success:
      return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_acrtm_update_start(config, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'acrtm_update_start', 'acrtm_update_start failed before setting GPRR')
      return error_code, None

    gprr_error_code, response_bytes = self.meta_read_gprr('acrtm_update_start', 0x3b)

    return gprr_error_code, response_bytes

  def meta_acrtm_update_finish(self, config: bytes, amtc0_in: bytes=None):
    """ End ACRTM update mode

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """

    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'acrtm_update_finish', 'Missing required data to run command: config value'), None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_acrtm_update_finish', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # check if boot_amtc0 has been retrieved, and if not, retrieve it by callilng read_extended_crtm_status 
    rc = self.get_boot_amtc0_if_not_set('meta_acrtm_update_finish')
    if not rc.esr_success:
      return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_acrtm_update_finish(config, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'acrtm_update_finish', 'acrtm_update_finish failed before setting GPRR')
      return error_code, None

    gprr_error_code, response_bytes = self.meta_read_gprr('acrtm_update_finish', 0x3D)

    return gprr_error_code, response_bytes

  def meta_provision_crtm(self, config: bytes, nonce: bytes=None, amtc0_in: bytes=None, provision_signature_func=None):
    """ Provision Authenta device.

        If the provision_crtm command is successful this function will call
        read_extended_crtm_status to the retrieve and reset the boot_amtc0
        value needed for rederiving the session key.

        NOTE: When doing a local provision, signature verification of the response
        will not be done in this function. Instead the caller MUST call
        verify_provision_crtm_response_for_local_provision() after calling
        this function. This is becuase the new CRK must be read from the
        KTR register and made available to the key driver for signature
        verification to be performed.
        BEFORE calling the verify function read the new key from KTR and
        add it to the key repository as the CRK.

        Args:
          config (bytes)
          nonce (bytes): Optional A 4-byte nonce to use in the Request_UID
            command sent sent to the Naxos device. If not provided, one will
            be randomly generated.
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure), nonce sent to the device
    """

    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'provision_crtm', 'Missing required data to run command: config value'), None, None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_provision_crtm', amtc0_in)
    if not rc.esr_success:
      return rc, None, None

    error_code = None
    atomic_ret_obj, sent_nonce = self.atomic_provision_crtm(config, nonce, amtc0_in, signature_func=provision_signature_func)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'provision_crtm', 'provision_crtm failed before setting GPRR')
      return error_code, None, sent_nonce

    error_code, response_bytes = self.meta_read_gprr('provision_crtm', 0x35, False)

    if error_code.esr_success:
      # make sure the nonce matches. The returned nonce is bytes 3-7 of the response
      ret_nonce = response_bytes[4:8]
      nonce_matched = (sent_nonce == ret_nonce)

      if not nonce_matched:
        self.logger.warning('Response validation failed nonce matching!!!')

      error_code.set_nonce_matched(nonce_matched)

      # if the device was just locally provisioned we cannot verify the
      # the response until the calling application reads the client or
      # session root key from the KTR and makes it accessable to the key
      # driver, so return with verifying. The application will be
      # required to call verify_provision_crtm_response_for_local_provision.
      if (config[0] & 0x01) == 0:
        # locally provisioned
        return error_code, response_bytes, sent_nonce

      # assume response is valid and set status struct so we can verify status
      # (otherwise the device will seem like it's not provisoined and verification will fail)
      # set it back if verification fails
      # old_amtc0 = self.amtc0
      # old_extended_stat = self.extended_crtm_status_struct
      
      err, *_ = self.meta_request_uid()
      if not err.esr_success:
        self.logger.error('meta_request_uid Failed: ' + err.error_msg)
        return error_code, response_bytes

      self._do_response_verification_checks(response_bytes, 0x35, error_code)

      error_code.set_valid_response(error_code.valid_response and nonce_matched)

      # if the response is valid, update internal data
      # if error_code.valid_response:
      # get the boot_amtc0 so the session key can be derived in subsequent commands
      read_status_err, _ = self.read_extended_crtm_status()
      if not read_status_err.esr_success:
        self.logger.info('Failed attempting to read extended crtm status after successful provision_crtm')
        self.logger.info('read_extended_crtm_status error message: %s: %s', read_status_err.cmd_name, read_status_err.error_msg)

    return error_code, response_bytes, sent_nonce

  def verify_provision_crtm_response_for_local_provision(self, response_bytes: bytes, sent_nonce: bytes):
    ESR = response_bytes[0]

    error_code = self._cmd_error_code(ESR.to_bytes(1, byteorder='big'), "provision_crtm")

    if ESR == 0x80:
      # assume response is valid and set status struct so we can verify status
      # (otherwise the device will seem like it's not provisoined and verification will fail)
      # set it back if verification fails
      # old_amtc0 = self.amtc0
      # old_extended_stat = self.extended_crtm_status_struct
      
      err, *_ = self.meta_request_uid()
      if not err.esr_success:
        self.logger.error('meta_request_uid Failed: ' + err.error_msg)
        return error_code, response_bytes

      if self.provision_sts_struct.Remotely_Provisioned:
        if not self._root_signing_key_externally_set:
          authenta_key_lib.set_srk_as_hmac_root_signing_key(self.keylib_context)
      else:
        self._set_client_keys()

      self._do_response_verification_checks(response_bytes, 0x35, error_code)

      # make sure the nonce matches. The returned nonce is bytes 3-7 of the response
      ret_nonce = response_bytes[4:8]
      nonce_matched = (sent_nonce == ret_nonce)

      if not nonce_matched:
        self.logger.warning('Response validation failed nonce matching!!!')

      error_code.set_nonce_matched(nonce_matched)
      error_code.set_valid_response(error_code.valid_response and nonce_matched)

      # if the response is valid, update internal data
      # if error_code.valid_response:
      # get the boot_amtc0 so the session key can be derived in subsequent commands
      read_status_err, _ = self.read_extended_crtm_status()
      if not read_status_err.esr_success:
        self.logger.info('Failed attempting to read extended crtm status after successful provision_crtm')
        self.logger.info('read_extended_crtm_status error message: %s: %s', read_status_err.cmd_name, read_status_err.error_msg)

    return error_code

  def meta_deprovision_crtm(self, config: bytes, amtc1_in: bytes=None, deprovision_signature_func=None):
    """ Deprovision a previously provisioned Authenta device

        Args:
          config (bytes)
          amtc1_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'deprovision_crtm', 'Missing required data to run command: config value'), None

    if self.provision_sts_struct is None:
      rc, *_ = self.meta_request_uid()
      if not rc.esr_success:
        self.logger.error('Failed attempt to retrieve provision status by calling request_uid on the device')
        return rc, None

    if not self.provision_sts_struct.Provisioned:
      error_code = self._cmd_error_code(b'\x00', 'deprovision_crtm', 'deprovision_crtm failed; device is not provisioned')
      return error_code, None

    # check if boot_amtc0 and amtc1 have been retrieved, and if not, retrieve them by calling read_extended_crtm_status 
    elif self._amtc1 is None or self._boot_amtc0 is None:
      rc, _ = self.read_extended_crtm_status()
      if rc.esr_success:
        self.logger.info('Retrieved BOOT_AMTC0 value amd AMTC1 value')
      else:
        self.logger.error('Failed attempting to retrieve BOOT_AMTC0 from device')
        return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_deprovision_crtm(config, amtc1_in, signature_func=deprovision_signature_func)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'deprovision_crtm', 'deprovision_crtm failed before retrieving GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr('deprovision_crtm', 0x36, False)

    if error_code.esr_success:
      sign_with_session_key = True

      # special case: if Allow_SRK bit set when local provion with TPM, use server root key to verify signature
      if not self.provision_sts_struct.Remotely_Provisioned and self.provision_sts_struct.TPM_Supported \
          and (config[0] & 0b00000010):
          sign_with_session_key = False

      self._do_response_verification_checks(response_bytes, 0x36, error_code, sign_with_session_key)

      # delete old CRK in Key XML if response validation succeeds for local provision with TPM
      if not self.provision_sts_struct.Remotely_Provisioned and self.provision_sts_struct.TPM_Supported \
        and error_code.esr_and_response_success:
        self.logger.info("Note: A locally provisioned device has been deprovisioned; Remove the Client Root Key from the Key Libarary if one is present.")

      self.initialize_internal_properties()

    return error_code, response_bytes

  def meta_auth_lock(self, amtc0_in: bytes=None):
    """ Cryptographically unlock the Authenta device to allow writing to
        non-volatile-locked areas

        Args:
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_auth_lock', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # check if boot_amtc0 has been retrieved, and if not, retrieve it by read_extended_crtm_status 
    rc = self.get_boot_amtc0_if_not_set('meta_auth_lock')
    if not rc.esr_success:
      return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_auth_lock(amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'auth_lock', 'auth_lock failed before retrieving GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr('auth_lock', 0x37)

    return error_code, response_bytes

  def meta_flash_extend(self, config: bytes, amtc0_in: bytes=None):
    """ Issue Flash_Extend command and get command repsonse

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'flash_extend', 'Missing required data to run command: config value'), None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_flash_extend', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # see if this function will be signed
    if ((config[0] & 0b00000101) == 1) or ((config[0] & 0b00000101) == 5): # option #0 or option #3
      # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling read_extended_crtm_status 
      rc = self.get_boot_amtc0_if_not_set('meta_flash_extend')
      if not rc.esr_success:
        return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_flash_extend(config, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'flash_extend', 'flash_extend failed before retrieving GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr('flash_extend', 0x3A)

    return error_code, response_bytes

  def meta_measure(self, config: bytes, eeprom_alt_src_addr: int, eeprom_key_src_addr: int,
                    eeprom_dest_addr: int, in_digest: bytes=None, amtc0_in: bytes=None):
    """ Issue Measure command and get command response

        Args:
          config (bytes): length must be exactly 2 bytes
          eeprom_alt_src_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used as data to include into the
            command measurement.
          eeprom_key_src_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used as a key when performing an HMAC
            calculation.
          eeprom_dest_addr (int): The base address of one of eight 32-byte
            EEPROM locations that may be used to store the result of a Measure
            command.
          amtc0_in (int): ONLY FOR DEBUG

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'measure', 'Missing required data to run command: config'), None
    if eeprom_alt_src_addr is None or eeprom_alt_src_addr > 0xFFFF:
      return self._cmd_error_code(b'\x00', 'measure', 'Missing required data to run command: eeprom_src_addr; or eeprom_alt_src_addr value too large to fit in 2 bytes'), None
    if eeprom_key_src_addr is None or eeprom_key_src_addr > 0xFFFF:
      return self._cmd_error_code(b'\x00', 'measure', 'Missing required data to run command: eeprom_key_addr; or eeprom_key_src_addr value too large to fit in 2 bytes'), None
    if eeprom_dest_addr is None or eeprom_dest_addr > 0xFFFF:
      return self._cmd_error_code(b'\x00', 'measure', 'Missing required data to run command: eeprom_dest_addr; or eeprom_dest_addr value too large to fit in 2 bytes'), None
    if in_digest and len(in_digest) != 32:
      return self._cmd_error_code(b'\x00', 'measure', 'Invalid parameter: in_digest; value should be 32 bytes'), None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_measure', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # see if this function will be signed
    if config[1] & 1 << 7:
      # check if boot_amtc0 has been retrieved, and if not, retrieve it by
      # calling read_extended_crtm_status
      rc = self.get_boot_amtc0_if_not_set('meta_measure')
      if not rc.esr_success:
        return rc, None

    error_code = None
    atomic_ret_obj = self.atomic_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr, in_digest, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'measure', 'measure failed before retrieving GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr('measure', 0x3F)

    return error_code, response_bytes

  # Non-Authenta device functions

  def meta_page_program(self, address: int, data: bytes):
    """ Issue Page_Program command

        Args:
          address (int): address on the NaxosDevice to write data to
          data (bytes): data to write to input address

        Returns:
          bool
    """

    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before a Page_Program command.')
      return ret

    ret = self.endpoint_platform.naxos_page_program(address, data)
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure following a Page_Program command.')
      return ret

    return ret

  def meta_fast_read(self, address: int):
    """ Issue Fast_Read command

        Args:
          address (int): address on the NaxosDevice to read from

        Returns:
          bool, bytes (response - may equal None on failure)
    """
    ret, response = self.endpoint_platform.naxos_fast_read(address)
    if not ret:
      self.logger.error('Command failure.')

    return ret, response

  def meta_dyb_write(self, lockbit: int, address: int):
    """ Issue DYB_Write command

        Args:
          address (int): address on the NaxosDevice to set lockbit on
          lockbit (int): value to set lockbit to; MUST be 0 or 1

        Returns:
          bool
    """

    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before dyb_write command.')
      return ret

    ret = self.endpoint_platform.naxos_dyb_write(lockbit, address)
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure following a dyb_write command.')
      return ret

    return ret

  def meta_dyb_read(self, address: int):
    """ Issue DYB_Read command

        Args:
          address (int): address on the NaxosDevice to read lockbit on

        Returns:
          bool, bytes (response - may equal None on failure)
    """

    ret = self.endpoint_platform.naxos_dyb_read(address)
    if not ret:
      self.logger.error('Command failure.')

    return ret

  def meta_ppb_program(self, address: int):
    """ Issue PPB_Program command

        Args:
          address (int): address on the Naxos device to set the PPB bit

        Returns:
          bool
    """

    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before a ppb_program command.')
      return ret

    ret = self.endpoint_platform.naxos_ppb_program(address)
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure following a ppb_program command.')
      return ret

    return ret

  def meta_ppb_erase(self):
    """ Issue PPB_Erase command to Naxos device.

        Clear PPB bit, unlock the non-volatile bits on the Naxos Device

        Returns:
          bool
    """

    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before a ppb_program command.')
      return ret

    ret = self.endpoint_platform.naxos_ppb_erase()
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure before a ppb_erase command.')
      return ret

    return ret

  def meta_ppb_read(self, address: int):
    """ Issue PPB_Erase command

        Args:
          address (int): address on the Naxos device to read lock bit om

        Returns:
          bool, bytes (response - may equal None on failure)
    """
    ret, response = self.endpoint_platform.naxos_ppb_read(address)
    if not ret:
      self.logger.error('Command failure.')

    return ret, response

  def meta_ppb_lock_bit_read(self):
    """ Issue PPB_Lock_Bit_Read command

        Returns:
          bool, bytes (response - may equal None on failure)
    """
    ret, response = self.endpoint_platform.naxos_ppb_lock_bit_read()
    if not ret:
      self.logger.error('Command failure.')

    return ret, response

  def meta_subsector_erase(self, address: int):
    """ Issue Subsector_Erase command to Naxos device.

        Args:
          address (int): address on the Naxos device to erase

        Returns:
          bool
    """
    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before a subsector erase command.')
      return ret

    ret = self.endpoint_platform.naxos_subsector_erase(address)
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure following a subsector erase command.')
      return ret

    return ret

  def meta_ppb_lock_bit_write(self):
    """ Issue PPB_Lock_Bit_Write command to Naxos device.

        Returns:
          bool
    """

    ret = self.endpoint_platform.naxos_write_enable()
    if not ret:
      self.logger.error('Write_Enable failure before a ppb_lock_bit_write command.')
      return ret

    ret = self.endpoint_platform.naxos_ppb_lock_bit_write()
    if not ret:
      self.logger.error('Command failure.')
      return ret

    ret = self.endpoint_platform.naxos_write_disable()
    if not ret:
      self.logger.error('Write_Disable failure following a ppb_lock_bit_write command.')
      return ret

    return ret

  def meta_read_status_register(self):
    """ Issue Read_Status_Register command to Naxos device

        Returns:
          bool: True on success, false on failure
          bytes: Response - may equal None on failure
    """
    ret, response = self.endpoint_platform.naxos_read_status_register()
    self.logger.debug('return={}, response={}'.format(ret, response.hex()))
    if not ret:
      self.logger.error('Command failure.')

    return ret, response

  # Utility functions

  def read_ktr(self, verify_sig: bool=True):
    """ Read data from KTR (assume device is provisioned)

        Args:
          verify_sig (bool): Optional; flag indicating whether to verifiy
            response signature

        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    data = None

    err, KTR_read_resp = self.meta_crtm_register_read(b'\00', sdk2_utils.CRTM_Registers.KTR_Base, verify_sig=verify_sig)
    if not err.esr_success:
      self.logger.error('reading least-significant-bits KTR register while attempting to read Cliennt Session Key: ' + err.error_msg)
      return err, data

    LSB_KTR = sdk2_utils.change_word_order(KTR_read_resp[16:32])

    err, KTR_read_resp = self.meta_crtm_register_read(b'\00', sdk2_utils.CRTM_Registers.KTR_Base + 16, verify_sig=verify_sig)
    if not err.esr_success:
      self.logger.error('reading most-significant-bits KTR register while attempting to read Cliennt Session Key: ' + err.error_msg)
      return err, data

    MSB_KTR = sdk2_utils.change_word_order(KTR_read_resp[16:32])

    data = MSB_KTR + LSB_KTR
    return err, data

  def _set_client_keys(self):
    """
    set client session key and client root key if applicable

    NOTE: before calling this function the following internal
          structures MUST be true:
            the device must Locally Provisioned
            self._provision_sts_struct
            self._boot_amtc0 - ONLY if provisioned TPM_Supported is True

    Return True if key(s) are set,
           False otherwise
    """
    if self._client_keys_set:
          return True
    kset = False
    rc = False
    if self.provision_sts_struct.Provisioned and not self.provision_sts_struct.Remotely_Provisioned:
      if self.provision_sts_struct.TPM_Supported:
        # the client root key (CRK) must have been set 
        # key repository by the application so it could 
        # be read by the key driver
        # The application would have read this from the device's
        # KTR register after provisioning and saved it to the key repo,,.
        # Note, it can be read until the CRK_Read_Disable bit is set
        if not self._root_signing_key_externally_set:
          rc = authenta_key_lib.set_crk_as_hmac_root_signing_key(self._keylib_context, True)
          if rc:
            # # _boot_amtc0 MUST NOT be None - see NOTE in function description
            # if self._boot_amtc0 is None:                
            #   self.logger.error("Failed to generate Client Session Key in key driver! PowerOnCount not set!")
            #   return False

            if not self._session_key_externally_set and self.boot_amtc0 is not None:
              rc = authenta_key_lib.derive_hmac_session_key(self._keylib_context, self._boot_amtc0, True)
              if not rc:
                self.logger.error("Failed to generate Client Session Key in key driver!")
              kset = rc
        else:
          self.logger.error("Client Root Key not found by the key driver, no signing key found!")
      else:
        # the client session key (CSK) must have been set 
        # key repository by the application so it could 
        # be read by the key driver
        # The application would have read this from the device's
        # KTR register befofe attempting aa signed command
        # Note, it can be read until the CSK_Read_Disable bit is set
        rc = authenta_key_lib.set_csk_as_hmac_session_signing_key(self._keylib_context)
        if not rc:
            self.logger.error("Client Root Key not found by the key driver, no signing key found!")
        kset = rc

    self._client_keys_set = kset

    return rc


  def _do_response_verification_checks(self, response_bytes: bytes, error_code: NaxosCmdErrorCode, use_session_key: bool=True):
    """ Verify signature and last command, and record the results in the
        input error_code object.

        If verifcation passes set AMTC0,

        Args:
          response_bytes: The response returned from read_gprr
          error_code: The object to record verification results in
    """
    # verify response validity
    signature_checks = self._response_sig_is_valid(response_bytes, use_session_key)

    if not signature_checks:
      self.logger.warning('Response validation failed signature check!!!')

    error_code.set_signature_matched(signature_checks)
    error_code.set_valid_response(signature_checks)

  def register_read_all(self):
    crtm_response_bytes = list()

    CRTM_REG_SZ = 0x360
    CRTM_REG_READ_BLK_SZ = 0x10

    for addr in range(0, CRTM_REG_SZ, CRTM_REG_READ_BLK_SZ):
      err, data = self.meta_crtm_register_read(config=b'\x00', crtm_addr=addr)

      if not err.esr_success:
        break

      crtm_response_bytes.append(data)

    self._CRTM_reg_read_response_lst = crtm_response_bytes
    return err

  def build_crtm_reg_data_byte_table(self) -> None:
    row_vals = [row[0x10:0x20] for row in self._CRTM_reg_read_response_lst]
    self._crtm_reg_data_bytes = b''.join(row_vals)

  def build_crtm_reg_attr_table(self) -> None:
    crtm_reg_data_table = dict()

    for key, value in self.ACRTM_REGISTERS.items():
      block_index, index, block_size = value.split(',')

      try:
        block_index = int(block_index, 16)

        start_index = block_index + int(index)
        end_index = start_index + int(block_size)
        datum = self._crtm_reg_data_bytes[start_index:end_index]

        crtm_reg_data_table[key] = datum.hex() if isinstance(datum, bytes) else datum

      except ValueError:
        continue

    self._crtm_reg_data_table = crtm_reg_data_table

  def set_amtc0(self, response: bytes):
    self._amtc0 = response[0x8:0x0c] # 4 bytes

  def set_amtc1(self, response: bytes):
    self._amtc1 = response[0x0C:0x10] # 4 bytes

  def set_extended_crtm_status(self, reg_data: bytes):
    reg_bit_attr = self.reg_bytes2struct(reg_data.hex(), self.Extended_Crtm_Status_struct)
    reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
    self._extended_crtm_status_struct = self.EXTENDED_CRTM_STATUS_STRUCT(*reg_bit_attr_int)

  def set_server_session_key(self, key: bytes):
    self._server_session_key = key

  @staticmethod
  def get_response_payload(response: bytes):
    return response[0x10:0x20] # 16 bytes

  @staticmethod
  def get_response_signature(response: bytes):
    return sdk2_utils.change_word_order(response[32: 64])

  def _get_uid_response_struct(self, uid_response_bytes):
    attr_list = NaxosDevice.binary_str2attribute_lst(uid_response_bytes, self.UID_ATTR_IDX)
    uid_response_bytes_struct = self.REQUEST_UID_STRUCT._make(attr_list)
    return uid_response_bytes_struct

  def _get_uid_attr(self, attr, uid_response_bytes_struct):
    # attributes as kept as byte strings in _uid_response_bytes_struct
    # retrieve the attr from struct and return it as byte string
    attr_val = uid_response_bytes_struct._asdict()[attr]
    return attr_val

  def read_extended_crtm_status(self):
    # this could be any other command except for request_uid
    error_code, _ = self.meta_crtm_register_read(b'\x00', NAXOS_REGISTER_BLOCK_ADDR_WITH_BOOT_AMT0)
    return error_code, self._extended_crtm_status_struct

  def get_amtc0_if_not_set(self, caller_name: str, amtc0_test_input: bytes=None):
    """ Check if amtc0 has been retrieved, and if not, retrieve it by calling
        request_uid.
    """
    if not amtc0_test_input and not self.amtc0:
      rc, *_ = self.meta_request_uid()
      if rc.esr_success:
        self.logger.info(caller_name + ': retrieved AMTC0 value')
      else:
        self.logger.warning(caller_name + ': Failed attempting to retrieve AMTC0 from device')
    else:
      rc = self._cmd_error_code(b'\x80', caller_name, 'OK')
    return rc

  def get_boot_amtc0_if_not_set(self, caller_name: str):
    """ This function checks both boot_amtc0 and provision_sts and if either
        has not been set, it will retrieve them by calling
        read_extended_crtm_status or meta_request_uid respectively.
    """
    if self.provision_sts_struct is None:
      rc, *_ = self.meta_request_uid()
      if not rc.esr_success:
        self.logger.warning(caller_name + ': Call to request_uid failed while attempting to retrieve provision status')
        rc = self._cmd_error_code(b'\x00', caller_name, 'Call to request_uid failed while attempting to retrieve provision status')
        return rc

    # check if amtc0 has been retrieved, and if not, retrieve it by calling 
    if not self._boot_amtc0:
      if self.provision_sts_struct.Provisioned:
        rc, _ = self.read_extended_crtm_status()
        if rc.esr_success:
          self.logger.info(caller_name + ': retrieved BOOT_AMTC0 value')
        else:
          self.logger.warning(caller_name + ': Failed attempting to retrieve BOOT_AMTC0 from device')
      else:
        self.logger.warning(caller_name + ': Device is not provisioned; Cannot retrieve BOOT_AMTC0')
        rc = self._cmd_error_code(b'\x00', caller_name, 'Device is not provisioned; Cannot retrieve BOOT_AMTC0')
    else:
      rc = self._cmd_error_code(b'\x80', caller_name, 'OK')

    return rc

  # The following functions for TEST ONLY
  def test_only_set_amtc0(self, test_amtc0: bytes):
    """ For creating negative test we expose setting the AMTC0 value.

        Normally amtc0 is retrieved and set when it is returned in response to
          commands.
    """
    self._amtc0 = test_amtc0

  def test_only_set_amtc1(self, test_amtc1: bytes):
    """ For creating negative test we expose setting the AMTC1 value.

        Normally amtc1 is retrieved and set when it is returned in response to
          commands.
    """
    self._amtc1 = test_amtc1

  def test_only_set_boot_amtc0(self, test_boot_amtc0: bytes):
    """ For creating negative test we expose setting the BOOT_AMTC0 value
        Normally boot_amtc0 is retrieved from the device by calling 
        read_extended_crtm_status or by reading address 0xD0 with meta_register_read
    """
    self._boot_amtc0 = test_boot_amtc0

  def set_device_config_volatile(self, reg_data: bytes):
    # Device_Config_Volatile
    #      only low-order bits [3]:[2] are used, the rest, (the other 30 bits), are reserved
    reg_bit_attr = self.reg_bytes2struct(reg_data.hex(), self.CRTM_Reg_Device_CFG_Vol_struct)
    reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
    self._device_config_volatile_struct = self.CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(*reg_bit_attr_int)

  # ---------------------------------------------
  # python struct (namedtuples) defs
  # ---------------------------------------------
  REQUEST_UID_STRUCT = namedtuple('REQUEST_UID_STRUCT',
    ['ESR', 'RSVD0', 'Provision_Sts', 'Nonce', 'A_MTCO', 'Revision', 'DupSig',
      'UID', 'Signature'])

  UID_ATTR_IDX = REQUEST_UID_STRUCT(ESR=1, RSVD0=2, Provision_Sts=1, Nonce=4,
    A_MTCO=4, Revision=3, DupSig=1, UID=16, Signature=32)

  # ACRTM_[Provision, Replace_SRK]
  PROVISION_STS_STRUCT = namedtuple('PROVISION_STS_STRUCT',
    ['UDS_Disabled', 'UDS_Replaced', 'RIOT_Mode', 'TPM_Supported',
    'Remotely_Provisioned', 'Check_Engine', 'SRK_Replaced',
    'Provisioned'])

  EXTENDED_CRTM_STATUS_STRUCT = namedtuple('EXTENDED_CRTM_STATUS_STRUCT',
    ['Reserved_2', 'LACMD', 'Reserved_1', 'Recovery_Mismatch',
      'Partial_Measurement', 'Recovery_Block', 'Auth_Unlock',
      'CRTM_PCR0_Invalid', 'CRTM_Range_Invalid', 'CRTM_Mismatch'])

  Extended_Crtm_Status_struct = EXTENDED_CRTM_STATUS_STRUCT(Reserved_2=4,
    LACMD=4, Reserved_1=1, Recovery_Mismatch=1, Partial_Measurement=1,
    Recovery_Block=1, Auth_Unlock=1, CRTM_PCR0_Invalid=1,
    CRTM_Range_Invalid=1, CRTM_Mismatch=1)

  COMMON_SRK_STATUS_STRUCT = namedtuple('COMMON_SRK_STATUS_STRUCT',
    ['ESR', 'Extended_CRTM_Status', 'Dup_Sig', 'Nonce', 'A_MTCO', 'A_MTC1',
     'UID', 'Signature'])

  COMMON_SRK_IDX = COMMON_SRK_STATUS_STRUCT(ESR=1, Extended_CRTM_Status=2,
    Dup_Sig=1, Nonce=4, A_MTCO=4, A_MTC1=4, UID=16, Signature=32)

  # ACRTM_Update_[Start, Finish], Measure, Flash Extend, Auth_Lock,
  # Deprovision_CRTM
  GENERIC_ACRTM_CMD_RESULTS = namedtuple('GENERIC_ACRTM_CMD_RESULTS',
    ['ESR', 'Extended_CRTM_Status', 'RSVD0', 'Dup_Sig', 'A_MTCO', 'A_MTC1',
     'RSVD1', 'Signature'])

  GENERIC_ACRTM_IDX = GENERIC_ACRTM_CMD_RESULTS(ESR=1, Extended_CRTM_Status=2,
    RSVD0=4, Dup_Sig=1, A_MTCO=4, A_MTC1=4, RSVD1=16, Signature=32)

  CRTM_Register_RW_STRUCT = namedtuple('CRTM_Register_RW_RESULTS',
    ['ESR', 'Extended_CRTM_Status', 'RSVD0', 'CRTM_Register_Addr',
    'RSVD1','Dup_Sig', 'AMTC0',  'AMTC1', 'Reg_Data', 'Signature'])

  CRTM_Register_RW_IDX = CRTM_Register_RW_STRUCT(ESR=1,
    Extended_CRTM_Status=2, RSVD0=1, CRTM_Register_Addr=2,
    RSVD1=1, Dup_Sig=1, AMTC0=4, AMTC1 =4, Reg_Data=16,
    Signature=32)

  CRTM_REGISTER_DEVICE_CFG_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_STRUCT',
    ['DEV_CFG_RESERVED', 'AMRecovery', 'CRK_Read_Disable', 
    'Recovery_Mode_Enabled', 'Auto_Measure_Active'])

  CRTM_Reg_Device_CFG_struct = CRTM_REGISTER_DEVICE_CFG_STRUCT(
    DEV_CFG_RESERVED=28, AMRecovery=1, CRK_Read_Disable=1,
    Recovery_Mode_Enabled=1, Auto_Measure_Active=1)

  CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT',
    ['DEV_CFG_VOL_RESERVED_1', 'Disable_Deprovision', 'CSK_Read_Disable', 
     'DEV_CFG_VOL_RESERVED_0'])

  CRTM_Reg_Device_CFG_Vol_struct = CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(
    DEV_CFG_VOL_RESERVED_1=28, Disable_Deprovision=1, CSK_Read_Disable=1,
    DEV_CFG_VOL_RESERVED_0=2)

  # output templates
  PROVISION_STS_FORM = """
Provision_sts:
    UDS_Disabled:          {UDS_Disabled}
    UDS_Replaced:          {UDS_Replaced}
    RIOT_Mode_Enabled:     {RIOT_Mode}
    TPM_Supported:         {TPM_Supported}
    Remotely_Provisioned:  {Remotely_Provisioned}
    Check_Engine:          {Check_Engine}
    SRK_Replaced:          {SRK_Replaced} 
    Provisioned:           {Provisioned}
"""

  DEVICE_STATUS_FORM = """
=============================================================================
DEMO: Checking Micron Authenta Flash Memory Device status.
====================================================================
        Current Config Settings
====================================================================

  Provision_sts attributes:
    Provisioned:          {Provisioned}
    SRK_Replaced:         {SRK_Replaced}   
    Check_Engine:         {Check_Engine}
    Remotely_Provisioned: {Remotely_Provisioned}
    TPM_Supported:        {TPM_Supported}
    RIOT_Mode_Enabled:    {RIOT_Mode}
    UDS_Replaced:         {UDS_Replaced}
    UDS_Disabled:         {UDS_Disabled}

  Extended_CRTM_STATUS:
    CRTM_Mismatch:        {CRTM_Mismatch}
    CRTM_Range_Invalid:   {CRTM_Range_Invalid}
    CRTM_PCR0_Invalid:    {CRTM_PCR0_Invalid}
    Auth_Unlock:          {Auth_Unlock}
    Recovery_Block:       {Recovery_Block}
    Partial_Measurement:  {Partial_Measurement}
    Recovery_Mismatch:    {Recovery_Mismatch}
    Last_Auth_Cmd (lower nibble):  {LACMD}

  CRTM Register attributes:
    PwrOnCount:                 {PwrOnCount}
    
    Device_Config Reg:          {Device_Config} 
    Device_Config bits: 
      Auto Measure:             {Auto_Measure_Active}
      Recovery Mode Enable:     {Recovery_Mode_Enabled}
      Client Root Key Readable: {CRK_Read_Disable}
      AMRecovery:               {AMRecovery}

    Device_Config_Volatile Reg: {Device_Config_Volatile} 
    Device_Config_Volatile bits: 
      Disable_Deprovision:      {Disable_Deprovision} 
      CSK_Read_Disable:         {CSK_Read_Disable}

    Defined CRTM Non-Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes):{CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes):{CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes):{CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes):{CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes):{CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes):{CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes):{CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes):{CRTM_Size7}

    Defined CRTM Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes):{CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes):{CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes):{CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes):{CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes):{CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes):{CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes):{CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes):{CRTM_Size7}

    MaxAutoMeasureSize:   {CRTM_Max_AutoMeasure_Size}

    Recovery Base Addr:   {Recovery_Base_Addr0}

    Flash Digest Values:
    PCR0:
      {flash_PCR0L}
      {flash_PCR0H}

    NV:
      {Flash_DigestNVL}
      {Flash_DigestNVH}

    NV Partial:
      {Flash_Digest_NV_PartialL}
      {Flash_Digest_NV_PartialH}


    Scratchpad Buffer Size: {Customer_RAM_sz}

    Scratchpad Buffer:
      {Customer_RAM}
"""

  ACRTM_REGISTERS = {
      'var_lst_decode': 'req_idx, byte_count, byte_len',
      'flash_PCR0L': '0x00,00,16',
      'flash_PCR0H': '0x10,00,16',
      'CRTM_Addr0': '0x20,00,04',
      'CRTM_Size0': '0x20,04,04',
      'CRTM_Addr1': '0x20,08,04',
      'CRTM_Size1': '0x20,12,04',
      'CRTM_Addr2': '0x30,00,04',
      'CRTM_Size2': '0x30,04,04',
      'CRTM_Addr3': '0x30,08,04',
      'CRTM_Size3': '0x30,12,04',
      'CRTM_Addr4': '0x40,00,04',
      'CRTM_Size4': '0x40,04,04',
      'CRTM_Addr5': '0x40,08,04',
      'CRTM_Size5': '0x40,12,04',
      'CRTM_Addr6': '0x50,00,04',
      'CRTM_Size6': '0x50,04,04',
      'CRTM_Addr7': '0x50,08,04',
      'CRTM_Size7': '0x50,12,04',
      'Flash_Digest_NV_PartialL' : '0x60,00,16',
      'Flash_Digest_NV_PartialH' : '0x70,00,16',
      'Flash_DigestNVL' : '0x80,00,16',
      'Flash_DigestNVH' : '0x90,00,16',
      'Device_Config' : '0xa0,00,04',
      'Recovery_Base_Addr0' : '0xa0,04,04',
      'CRTM_Max_AutoMeasure_Size' : '0xa0,08,04',
      'Reserved_0' : '0xa0,12,04',
      'KTRL' : '0xb0,00,16',
      'KTRH' : '0xc0,00,16',
      'EEPROM_Attributes' : '0xd0,00,04',
      'PwrOnCount' : '0xd0,04,04',
      'Device_Config_Volatile' : '0xd0,08,04',
      'Reserved_1' : '0xd0,12,04',
      'Reserved_2' : '0xe0,00,144',
      'AppendConfig0' : '0x170,00,04',
      'AppendConfig1' : '0x170,04,04',
      'AppendConfig2' : '0x170,08,04',
      'AppendConfig3' : '0x170,12,04',
      'AMBitmap' : '0x180,00,32',
      'Reserved_3' : '0x1A0,00,96',
      'Customer_EEPROM' : '0x200,00,256',
      'Customer_RAM' : '0x300,00,96'
  }
