import hmac
import sys
import os
import logging

from hashlib import sha256
from random import _urandom as nonce
import authenta.naxos.utils as sdk2_utils

from .defs_h import *
from ..endpoint_platform import EndpointPlatform, PlatformReturnType, EpPlatformRetCode
from .device import NaxosDevice, NAXOS_REGISTER_BLOCK_ADDR_WITH_BOOT_AMT0
from .cmd_error_code import NaxosCmdErrorCode
from .cmd_error_code_1_5 import NaxosCmdErrorCode_1_5
import authenta.authenta_key_lib_interface as authenta_key_lib


class NaxosDevice_1_5(NaxosDevice):
  """ Naxos 1.5 implementation
      Make changes based on NaxosDevice
  """
  _cmd_error_code = NaxosCmdErrorCode_1_5

  def __init__(self, endpoint_platform, logger=logging.getLogger(__name__), user_defined_key_driver_data=None, user_defined_key_driver_dev_info_dict=None):
    super().__init__(endpoint_platform, logger, user_defined_key_driver_data, user_defined_key_driver_dev_info_dict)

  def initialize_internal_properties(self):
    """ Set class variables to None

        If this class is not being used outside a "with" statement this MUST
        be called whenever the hardware device has been rebooted
    """
    super().initialize_internal_properties()
    if self._key_driver_dev_info_dict is None:
        self._key_driver_dev_info_dict = { }
    self._key_driver_dev_info_dict['device_version'] = '1.5'

  @property
  def Device_CFG(self):
    if self.crtm_reg_data_table is not None:
      reg_bytes = self.crtm_reg_data_table['Device_Config']
      reg_bit_attr = self.reg_bytes2struct(reg_bytes, self.CRTM_Reg_Device_CFG_struct)
      reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
      return self.CRTM_REGISTER_DEVICE_CFG_STRUCT(*reg_bit_attr_int)

    return None

  @property
  def Device_CFG_Volatile(self):
    if self.crtm_reg_data_table is not None:
      reg_bytes = self.crtm_reg_data_table["Device_Config_Volatile"]
      reg_bit_attr = self.reg_bytes2struct(reg_bytes, self.CRTM_Reg_Device_CFG_Vol_struct)
      reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
      return self.CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(*reg_bit_attr_int)

    return None

  @property
  def device_config_volatile(self):
    return self._device_config_volatile_struct

  def get_response_signature(self, response: bytes):
    return sdk2_utils.change_word_order(response[32: 64])

  # Atomic Commands

  def atomic_auth_modify(self, config: bytes, source_addr: int, destination_addr: int, size: int, source_content_digest: bytes=None, amtc0_in: bytes=None):
    """ Auth_Modify command implementation

        Args:
          config (bytes): 2 bytes
          source_addr (int)
          destination_addr (int)
          size (int)
          source_content_digest (bytes): (optional) digest of data to be
            programed for options #1 and #3
          amtc0 (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None or source_addr is None or destination_addr is None or size is None:
      self.logger.error("Missing required data to run command: Function Paramaters: config ,source_addr, destination_addr, size")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if len(config) != 2:
      self.logger.error("Invalid length for paramter config; it must be 2 bytes.")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if size > 0xFFFFFF: 
      # size will be stored in 3 bytes, so it cannot be larger than this value
      self.logger.error("Invalid value for parameter size. It will sent as 3 bytes; Max value 0xFFFFFF.")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
    if use_amtc0 is None:
      self.logger.error(" Missing required data to run command: amtc0; Run request_uid before running this command")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

    # format {opcode[1]}{sub_opcode[1]}{config[2]}{amtc_0[4]}{source_addr[4]}{destination_addr[4]}{size[3]}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('34') + config + use_amtc0 + source_addr.to_bytes(4, byteorder="big") + destination_addr.to_bytes(4, byteorder="big") + size.to_bytes(3, byteorder='big')

    # don't sign if command option #2 (config[1]:[0]=b'10')
    if (config[1] & 0b00000011) != 2:     # if not option #2
      if not self.boot_amtc0: #needed for signing
        self.logger.error("Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status")
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)
      # sign /w Session Key
      # self.logger.info("boot_amtc0:{}".format(self.boot_amtc0.hex()))
      if (config[1] & 0b00000011) == 0:
        # config[1:0] == 00b (option #0)
        # HMAC(cmd_str)Ksn(*)
        sig = self._sign_cmd(cmd_str)
      else:
        if source_content_digest is None:
          self.logger.error("Missing required data to run command using option #1/#3: source_content_digest")
          return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None)

        # config[1:0] == 01b (option #1/#3)
        # HMAC(cmd_str || SHA256[Source Content])Ksn(*)
        sig = self._sign_cmd(cmd_str, additional_data=source_content_digest)
      cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str)


  def atomic_replace_srk(self, config: bytes, additional_data: bytes=None, nonce_in: bytes=None, amtc0_in: bytes=None, signature_func=None):
    """ Replace_SRK command implementation

        Args:
          config (bytes)
          additional_data (bytes): hash of the key(s) required based on 
            replace_srk option (for options #0 and #1, hash should include the 
            new SRK key. for option #2, hash should include the new and old SRK 
            keys)
          nonce_in (bytes): Optional 4-byte bytes object holding a nonce to
            use in the Request_UID command sent to the Naxos device. If no
            nonce is supplied, one will be generated in this function and
            returned.
          amtc0 (int): ONLY FOR DEBUG

        Returns:
          PlatformReturnType
          The nonce sent with the command
    """

    # use the input amtc0 for debug
    if amtc0_in:
      use_amtc0 = amtc0_in
    else:
      use_amtc0 = self.amtc0

    # make sure we have the info needed to call the function
    if config is None:
      self.logger.error("Missing required data to run command: config value")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
    if use_amtc0 is None:
      self.logger.error("Missing required data to run command: amtc0; Run request_uid before running this command")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
    if self.boot_amtc0 is None: #needed for signing
      self.logger.error("Missing required data to sign command: boot_amtc0; To retrieve boot_amtc0 run read_extended_crtm_status")
      return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None

    if nonce_in is not None:
      if not isinstance(nonce_in, bytes):
        self.logger.error("input argument is of wrong type; nonce must be of type bytes!")
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
      if len(nonce_in) != 4: 
        self.logger.error("input argument is of wrong length; nonce a bytes of length 4!")
        return PlatformReturnType(EpPlatformRetCode.EP_ERROR, None), None
      a_nonce = nonce_in
    else:
      a_nonce = bytes.fromhex(nonce(4).hex())

    # format {opcode[1]}{sub_opcode[1]}{config[2]}{amtc_0[4]}{reserved[1]}{nonce[4]}
    cmd_str = bytes.fromhex(ACRTM) + bytes.fromhex('39') + config + use_amtc0 + b'\x00' + a_nonce
    config_int = int.from_bytes(config, byteorder='big')

    if signature_func is not None:
          sig = signature_func(self, cmd_str)

    elif config_int & 0x03 == 0x00:
      # replace SRK with a third-party one
      # HMAC(cmd_str || SHA256(new Kr(s)))Ksn(s)
      sig = self._sign_cmd(cmd_str, True, additional_data=additional_data)

    elif config_int & 0x03 == 0x01 or config_int & 0x03 == 0x03:
      # revert back to Micron SRK/UDS
      # HMAC(cmd_str || SHA256(0x00..00))Ksn(s)

      data_to_hash = bytes([
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
          0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
      ])
      additional_data = sha256(data_to_hash).digest()
      sig = self._sign_cmd(cmd_str, additional_data=additional_data)

    elif config_int & 0x03 == 0x02:
      # replace UDS with a third-party one
      # HMAC(cmd_str || SHA256(existing UDS || new UDS))Ksn(s)

      sig = self._sign_cmd(cmd_str, True, additional_data=additional_data)
    else:
      # should never get here
      sig = b''

    cmd_str = cmd_str + sig

    return self.endpoint_platform.exec_on_device(self.device_type, cmd_str), a_nonce

  # Meta Commands

  def meta_request_uid(self, config: bytes = b"\x00", nonce: bytes=None):
    """ Read and return UID

        Args:
          config (bytes): Optional
          nonce (bytes): Optional A 4-byte nonce to use in the Request_UID
            command sent sent to the Naxos device. If provided, one will be
            randomly generated and returned.
        Returns:
          NaxosCmdErrorCode
          response (bytes): may equal None on failure
          nonce(4-byte bytes): nonce sent with the command
    """
    error_code = None
    atomic_ret_obj, sent_nonce = self.atomic_request_uid(config, nonce)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'00', 'request_uid', 'request_uid failed before setting GPRR')
      return error_code, None, sent_nonce

    error_code, uid_response_bytes = self.meta_read_gprr("request_uid", 0x3E, False)

    # test to see if command returned good status
    if error_code.esr_success:
      self.logger.info("Data-in:{}".format(uid_response_bytes.hex()))

      uid_respones_bytes_struct = self._get_uid_response_struct(uid_response_bytes)
      ret_nonce = self._get_uid_attr('Nonce', uid_respones_bytes_struct)

      # data below is needed in _response_sig_is_valid()
      # parse provision_sts byte
      b_lst = (list("{:08b}".format(uid_response_bytes[3])))
      b_lst_int = [int(d) for d in b_lst]
      self._provision_sts_struct = self.PROVISION_STS_STRUCT(*b_lst_int)

      self._uid = self._get_uid_attr('UID', uid_respones_bytes_struct)
      self._revision = self._get_uid_attr('Revision', uid_respones_bytes_struct)

      # now that we have the UID open the key driver if it hasn't been opened yet
      if self._keylib_context is None:
        self._key_driver_dev_info_dict['device_revision'] = self._revision.hex()             
        self._open_key_driver()


      # verify that the response is valid
      if self._provision_sts_struct.ID_Key_Select and self._provision_sts_struct.Provisioned:
        # response is signed using active session key (server or client)
        signature_checks = self._response_sig_is_valid(uid_response_bytes, \
          sign_with_session_key=True)
      else:
        # response is signed using server root key
        signing_str = uid_response_bytes[:32]
        response_sig = self.get_response_signature(uid_response_bytes)

        if self._root_signing_key_externally_set or authenta_key_lib.set_srk_as_hmac_root_signing_key(self._keylib_context):
          # in case client key was previously set, make sure
          # it gets set again if necessary
          self._client_keys_set = False
           
          self.logger.info('set SRK as signing key for request_uid')
          verified = authenta_key_lib.verify_root_sha256_hmac(self._keylib_context, signing_str, response_sig)

          if verified is None:
            self.logger.warning('Failed to verify response signature in key library')
            verified = False

        else:
          self.logger.info('Failed set SRK as signing key for request_uid')
          verified = False

        signature_checks = verified

      nonce_matched = (sent_nonce == ret_nonce)

      if not signature_checks:
        self.logger.warning("response validation failed signature check!!!")
      if not nonce_matched:
        self.logger.warning("response validation failed nonce matching!!!")

      error_code.set_signature_matched(signature_checks)
      error_code.set_nonce_matched(nonce_matched)
      error_code.set_valid_response(signature_checks and nonce_matched)
    else:
      self.logger.error("Command returned error: 0x{}".format(error_code.esr.hex()))

    return error_code, uid_response_bytes, sent_nonce

  def meta_auth_modify(self, config: bytes, source_addr: int, destination_addr: int, size: int,
    source_content: bytes=None, source_content_digest: bytes=None, amtc0_in: bytes=None):
    """ Issue Auth_Modify command and get command repsonse

        Args:
          config (bytes): 2 bytes
          source_addr (int)
          destination_addr (int)
          size (int)
          source_content (bytes): (OPTIONAL) data to be programed for
            options #1 and #3
          source_content_digest (bytes): (OPTIONAL) digest of data to be
            programed for options #1 and #3
          amtc0_in (int): ONLY FOR DEBUG
        Returns:
          NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    # make sure we have the info needed to call the function
    if config is None is None or source_addr is None or destination_addr is None or size is None:
      return self._cmd_error_code(b'\x00', 'auth_modify', "Missing required data to run command: Function Paramaters: config ,source_addr, destination_addr, size"), None
    if len(config) != 2:
      return self._cmd_error_code(b'\x00', 'auth_modify', "Invalid length for paramter config; it must be 2 bytes."), None
    if size > 0xFFFFFF: 
      # size will be stored in 3 bytes, so it cannot be larger than this value
      return self._cmd_error_code(b'\x00', 'auth_modify', "Invalid value for parameter size. It will sent as 3 bytes; Max value 0xFFFFFF."), None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_auth_modify', amtc0_in)
    if not rc.esr_success:
      return rc, None

    # see if this function will be signed
    if (config[0] & 0b00000011) != 2:     # if not option #2
      # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling read_extended_crtm_status 
      rc = self.get_boot_amtc0_if_not_set('meta_auth_modify')
      if not rc.esr_success:
        return rc, None

    if source_content is not None and source_content_digest is None:
      source_content_digest = sha256(source_content).digest()

    error_code = None
    atomic_ret_obj = self.atomic_auth_modify(config, source_addr, destination_addr, size, source_content_digest, amtc0_in)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'auth_modify', 'auth_modify failed before retrieving GPRR')
      return error_code, None

    error_code, response_bytes = self.meta_read_gprr("auth_modify", 0x34)

    return error_code, response_bytes

  def meta_replace_srk(self, config: bytes, additional_data: bytes=None, nonce: bytes=None, amtc0_in: bytes=None, replace_srk_signature_func=None):
    """ Issue Replace_SRK command and get command repsonse

        Args:
          config (bytes)
          additional_data (bytes): hash of the key(s) required based on 
            replace_srk option (for options #0 and #1, hash should include the 
            new SRK key. for option #2, hash should include the new and old SRK 
            keys)
          nonce (bytes): Optional A 4-byte nonce to use in the Request_UID
            command sent sent to the Naxos device. If not provided, one will
            be randomly generated and returned.
          param amtc0_in (int): ONLY FOR DEBUG
        Returns:
          NaxosCmdErrorCode
          response (bytes): may equal None on failure
          nonce (4-byte bytes): nonce sent with the command
    """
    # make sure we have the info needed to call the function
    if config is None:
      return self._cmd_error_code(b'\x00', 'replace_srk', "Missing required data to run command: config"), None, None

    # check if amtc0 has been retrieved, and if not, retrieve it by calling request_uid
    rc = self.get_amtc0_if_not_set('meta_replace_srk', amtc0_in)
    if not rc.esr_success:
      return rc, None, None

    # check if boot_amtc0 has been retrieved, and if not, retrieve it by calling read_extended_crtm_status 
    rc = self.get_boot_amtc0_if_not_set('meta_replace_srk')
    if not rc.esr_success:
      return rc, None, None

    error_code = None
    atomic_ret_obj, sent_nonce = self.atomic_replace_srk(config, additional_data, nonce, amtc0_in, replace_srk_signature_func)
    if atomic_ret_obj.status != EpPlatformRetCode.EP_OK:
      error_code = self._cmd_error_code(b'\x00', 'replace_srk', 'replace_srk failed before retrieving GPRR')
      return error_code, None, sent_nonce

    error_code, response_bytes = self.meta_read_gprr("replace_srk", 0x39, False)

    return error_code, response_bytes, sent_nonce

  def meta_crtm_register_read(self, config: bytes, crtm_addr: int, amtc0_in: bytes=None, verify_sig: bool=True):
    """ Read and return a 16-byte register from the Authenta chip

        If this function is called, the register containing the powerOnCounter
        (Boot AMTC0, address 0x0D) the boot_amtc0 of this object will be
        updated to the value in the read register.

        Args:
          config (bytes)
          amtc0_in (int): ONLY FOR DEBUG
          verify_sig (bool)

        Returns:
            NaxosCmdErrorCode, bytes (response - may equal None on failure)
    """
    # call parent class where this function is already implemented.
    error_code, response_bytes = super().meta_crtm_register_read(config, crtm_addr, amtc0_in, verify_sig)

    return error_code, response_bytes

  # Utility functions

  def _do_response_verification_checks(self, response_bytes: bytes, sub_opcode: int, error_code: NaxosCmdErrorCode, use_session_key: bool=True):
    """ Verify signature and last command, and record the results in the
        input error_code object. If verifcation passes set AMTC0.

        Args:
          response_bytes: the response returned from read_gprr
          sub_opcode: sub_opcode of the command that generated the gprr
          error_code: The object to record verification results in
    """

    # do the parent class verification checks and then add additional 1.5 test
    super()._do_response_verification_checks(response_bytes, error_code, use_session_key)

    # verify  that the response is associated with the command sent
    last_auth_cmd_matches = None

    # commands, other than request_uid, return extended crtm status in their response
    # so set extended_crtm_status here
    self.set_extended_crtm_status(response_bytes[1:3])
    last_cmd_bits = self._extended_crtm_status_struct.LACMD # least-significant 4 bits of the subopcode of the last successful command
    last_auth_cmd_matches =  last_cmd_bits == (sub_opcode & 0b00001111)

    if not last_auth_cmd_matches:
      self.logger.warning("response validation failed last successful Authenta command check!!!")

    error_code.set_last_opcode_matched(last_auth_cmd_matches)
    error_code.set_valid_response(error_code.valid_response and last_auth_cmd_matches)

  def verify_replace_srk_response(self, response_bytes: bytes, sent_nonce: bytes, error_code: NaxosCmdErrorCode, sign_with_session_key: bool=False, force: bool=False):
    # make sure the nonce matches. The returned nonce is bytes 3-7 of the response
    ret_nonce = response_bytes[4:8]
    nonce_matched = (sent_nonce == ret_nonce)

    if not nonce_matched:
      self.logger.warning('Response validation failed nonce matching!!!')

    error_code.set_nonce_matched(nonce_matched)

    # Run signature verification checks
    signing_str = response_bytes[:32]
    response_sig = self.get_response_signature(response_bytes)

    if sign_with_session_key:
      rc = self.get_boot_amtc0_if_not_set('_response_sig_is_valid')
      if not rc.esr_success:
        self.logger.warning('Unable to verify response signature: ' + rc.error_msg)
        return False

      if self.provision_sts_struct is None:
        rc, *_ = self.meta_request_uid()
        if not rc.esr_success:
          self.logger.warning('Unable to verify response signature; attempt to retrieve provision_sts_struct by calling meta_request_uid Failed: ' + rc.error_msg)
          return False

    if self.provision_sts_struct.Remotely_Provisioned or not self._provision_sts_struct.Provisioned:
      # remote provision
      #
      # (or if it is not provisioned and the command was 
      # Request_UID - the only signed response for a 
      # non-provisioned device and the only time this function 
      # would called for a non-provisioned device)
      # On a non-provisioned device Request_UId is always signed with the SRK

      # commenting the following line out because this shouldn't be required now,
      # this wybeen done before signing the command...
      if not self._root_signing_key_externally_set:
        if authenta_key_lib.set_srk_as_hmac_root_signing_key(self._keylib_context, force=force):
          self.logger.info('set SRK as signing key')
        else:
          self.logger.error('Failed to set signing key (SRK)!')
          return False
    else:
      # local provision
      self.logger.error('The SRK cannot be replaced while the device is locally provisioned')
      return False

    if sign_with_session_key:
      if self._server_session_key is None and not self._session_key_externally_set:
        if authenta_key_lib.derive_hmac_session_key(self._keylib_context, self._boot_amtc0, force=force):
          self.logger.debug('successful call to authenta_key_lib.derive_hmac_session_key')
        else:
          self.logger.error('Failed to generate session key while attempting to  verify command response!')
          return False

      verified = authenta_key_lib.verify_session_sha256_hmac(self._keylib_context, signing_str, response_sig)
    else:
      # signed with server root Key
      verified = authenta_key_lib.verify_root_sha256_hmac(self._keylib_context, signing_str, response_sig)

    if verified is None:
      self.logger.warning('Failed to verify response signature in key library')
      verified = False

    signature_checks = verified
    # signature_checks = self._response_sig_is_valid(response_bytes, sign_with_session_key=sign_with_session_key, force=True)
    if not signature_checks:
      self.logger.warning('Response validation failed signature check!!!')

    error_code.set_signature_matched(signature_checks)
    error_code.set_valid_response(signature_checks)

    # verify  that the response is associated with the command sent
    last_auth_cmd_matches = None

    # commands, other than request_uid, return extended crtm status in their response
    # so set extended_crtm_status here
    self.set_extended_crtm_status(response_bytes[1:3])
    last_cmd_bits = self._extended_crtm_status_struct.LACMD # least-significant 4 bits of the subopcode of the last successful command
    last_auth_cmd_matches =  last_cmd_bits == (0x39 & 0b00001111)

    if not last_auth_cmd_matches:
      self.logger.warning("response validation failed last successful Authenta command check!!!")

    error_code.set_last_opcode_matched(last_auth_cmd_matches)
    error_code.set_valid_response(error_code.valid_response and last_auth_cmd_matches)

  def set_device_config_volatile(self, reg_data: bytes):
    # Device_Config_Volatile
    # Only low-order bits [7]:[3]:[2] are used, the rest, (the other 29 bits), are reserved
    reg_bit_attr = self.reg_bytes2struct(reg_data.hex(), self.CRTM_Reg_Device_CFG_Vol_struct)
    reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
    self._device_config_volatile_struct = self.CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(*reg_bit_attr_int)

  def set_extended_crtm_status(self, reg_data: bytes):
    reg_bit_attr = self.reg_bytes2struct(reg_data.hex(), self.Extended_Crtm_Status_struct)
    reg_bit_attr_int = [int(s, 2) for s in reg_bit_attr]
    self._extended_crtm_status_struct = self.EXTENDED_CRTM_STATUS_STRUCT(*reg_bit_attr_int)



  # ---------------------------------------------
  # python struct (namedtuples) defs
  # ---------------------------------------------
  PROVISION_STS_STRUCT = namedtuple('PROVISION_STS_STRUCT',
    ["ID_Key_Select", "UDS_Replaced", "RIOT_Mode", "TPM_Supported",
     "Remotely_Provisioned", "Check_Engine", "SRK_Replaced",
     "Provisioned"])

  EXTENDED_CRTM_STATUS_STRUCT = namedtuple('EXTENDED_CRTM_STATUS_STRUCT',
    ["KMS_Backup", "UDS_Disabled", "Reserved", "Virtual_A24", "LACMD",
     "Tamper_Detect", "Recovery_Mismatch", "Partial_Measurement",
     "Recovery_Block", "Auth_Unlock", "CRTM_PCR0_Invalid",
     "CRTM_Range_Invalid", "CRTM_Mismatch"])

  Extended_Crtm_Status_struct = EXTENDED_CRTM_STATUS_STRUCT(
    KMS_Backup=1, UDS_Disabled=1, Reserved=1, Virtual_A24=1, LACMD=4,
    Tamper_Detect=1, Recovery_Mismatch=1, Partial_Measurement=1,
    Recovery_Block=1, Auth_Unlock=1, CRTM_PCR0_Invalid=1,
    CRTM_Range_Invalid=1, CRTM_Mismatch=1)

  CRTM_REGISTER_DEVICE_CFG_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_STRUCT',
    ["DEV_CFG_RESERVED_1", "ID_Key_Select", "Half_Device_Swap_Enable",
     "Half_Device_Swap_Active", "DEV_CFG_RESERVED_0", "AMRecovery",
     "CRK_Read_Disable", "Recovery_Mode_Enabled", "Auto_Measure_Active"])

  CRTM_Reg_Device_CFG_struct = CRTM_REGISTER_DEVICE_CFG_STRUCT(
    DEV_CFG_RESERVED_1=24, ID_Key_Select=1, Half_Device_Swap_Enable=1,
    Half_Device_Swap_Active=1, DEV_CFG_RESERVED_0=1, AMRecovery=1,
    CRK_Read_Disable=1, Recovery_Mode_Enabled=1, Auto_Measure_Active=1)

  CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT = namedtuple('CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT',
    ["DEV_CFG_VOL_RESERVED_2", "Clear_Tamper", "DEV_CFG_VOL_RESERVED_1",
     "Disable_Deprovision", "CSK_Read_Disable", "DEV_CFG_VOL_RESERVED_0"])

  CRTM_Reg_Device_CFG_Vol_struct = CRTM_REGISTER_DEVICE_CFG_VOL_STRUCT(
    DEV_CFG_VOL_RESERVED_2=24, Clear_Tamper=1, DEV_CFG_VOL_RESERVED_1=3,
    Disable_Deprovision=1, CSK_Read_Disable=1, DEV_CFG_VOL_RESERVED_0=2)

    # output templates
  PROVISION_STS_FORM = '''
Provision_sts:
    ID_Key_Select:         {ID_Key_Select}
    UDS_Replaced:          {UDS_Replaced}
    RIOT_Mode_enabled:     {RIOT_Mode}
    TPM_Supported:         {TPM_Supported}
    Remotely_Provisioned:  {Remotely_Provisioned}
    Check_Engine:          {Check_Engine}
    SRK_Replaced:          {SRK_Replaced} 
    Provisioned:           {Provisioned}
'''

  DEVICE_STATUS_FORM = '''
=============================================================================
DEMO: Checking Micron Authenta Flash Memory Device status.
====================================================================
        Current Config Settings
====================================================================

  Provision_sts attributes:
    Provisioned:          {Provisioned}
    SRK_Replaced:         {SRK_Replaced}
    Check_Engine:         {Check_Engine}
    Remotely_Provisioned: {Remotely_Provisioned}
    TPM_Supported:        {TPM_Supported}
    RIOT_Mode_enabled:    {RIOT_Mode}
    UDS_Replaced:         {UDS_Replaced}
    ID_Key_Select:        {ID_Key_Select}

  Extended_CRTM_STATUS:
    CRTM_Mismatch:        {CRTM_Mismatch}
    CRTM_Range_Invalid:   {CRTM_Range_Invalid}
    CRTM_PCR0_Invalid:    {CRTM_PCR0_Invalid}
    Auth_Unlock:          {Auth_Unlock}
    Recovery_Block:       {Recovery_Block}
    Partial_Measurement:  {Partial_Measurement}
    Recovery_Mismatch:    {Recovery_Mismatch}
    Tamper_Detect:        {Tamper_Detect}
    Virtual_A24:          {Virtual_A24}
    UDS_Disabled:         {UDS_Disabled}
    KMS_Backup:           {KMS_Backup}
    Last_Auth_Cmd (lower nibble):  {LACMD}

  CRTM Register attributes:
    PwrOnCount:                 {PwrOnCount}

    Device_Config Reg:          {Device_Config} 
    Device_Config bits: 
      Auto Measure:             {Auto_Measure_Active}
      Recovery Mode Enable:     {Recovery_Mode_Enabled}
      Client Root Key Readable: {CRK_Read_Disable}
      AMRecovery:               {AMRecovery}
      Half_Device_Swap_Active:  {Half_Device_Swap_Active}
      Half_Device_Swap_Enable:  {Half_Device_Swap_Enable}
      ID_Key_Select:            {ID_Key_Select}

    Device_Config_Volatile Reg: {Device_Config_Volatile} 
    Device_Config_Volatile bits: 
      Disable_Deprovision:      {Disable_Deprovision} 
      CSK_Read_Disable:         {CSK_Read_Disable}
      Clear_Tamper:             {Clear_Tamper}

    Defined CRTM Non-Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes):{CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes):{CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes):{CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes):{CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes):{CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes):{CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes):{CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes):{CRTM_Size7}

    Defined CRTM Volatile Address Ranges:
      CRTM Address 0) Start Address: {CRTM_Addr0}; Size (in bytes):{CRTM_Size0}
      CRTM Address 1) Start Address: {CRTM_Addr1}; Size (in bytes):{CRTM_Size1}
      CRTM Address 2) Start Address: {CRTM_Addr2}; Size (in bytes):{CRTM_Size2}
      CRTM Address 3) Start Address: {CRTM_Addr3}; Size (in bytes):{CRTM_Size3}
      CRTM Address 4) Start Address: {CRTM_Addr4}; Size (in bytes):{CRTM_Size4}
      CRTM Address 5) Start Address: {CRTM_Addr5}; Size (in bytes):{CRTM_Size5}
      CRTM Address 6) Start Address: {CRTM_Addr6}; Size (in bytes):{CRTM_Size6}
      CRTM Address 7) Start Address: {CRTM_Addr7}; Size (in bytes):{CRTM_Size7}

    MaxAutoMeasureSize:   {CRTM_Max_AutoMeasure_Size}

    Recovery Base Addr:   {Recovery_Base_Addr0}

    Flash Digest Values:
    PCR0:       
      {flash_PCR0L}
      {flash_PCR0H}

    NV:
      {Flash_DigestNVL}
      {Flash_DigestNVH}

    NV Partial:
      {Flash_Digest_NV_PartialL}
      {Flash_Digest_NV_PartialH}

    Scratchpad Buffer Size: {Customer_RAM_sz}

    Scratchpad Buffer:
      {Customer_RAM}

    EEPROM Attributes: {EEPROM_Attributes}

    EEPROM Buffer:
      {Customer_EEPROM}
'''

  ACRTM_REGISTERS = {
      "var_lst_decode": "req_idx, byte_count, byte_len",
      "flash_PCR0L": "0x00,00,16",
      "flash_PCR0H": "0x10,00,16",
      "CRTM_Addr0": "0x20,00,04",
      "CRTM_Size0": "0x20,04,04",
      "CRTM_Addr1": "0x20,08,04",
      "CRTM_Size1": "0x20,12,04",
      "CRTM_Addr2": "0x30,00,04",
      "CRTM_Size2": "0x30,04,04",
      "CRTM_Addr3": "0x30,08,04",
      "CRTM_Size3": "0x30,12,04",
      "CRTM_Addr4": "0x40,00,04",
      "CRTM_Size4": "0x40,04,04",
      "CRTM_Addr5": "0x40,08,04",
      "CRTM_Size5": "0x40,12,04",
      "CRTM_Addr6": "0x50,00,04",
      "CRTM_Size6": "0x50,04,04",
      "CRTM_Addr7": "0x50,08,04",
      "CRTM_Size7": "0x50,12,04",
      "Flash_Digest_NV_PartialL": "0x60,00,16",
      "Flash_Digest_NV_PartialH": "0x70,00,16",
      "Flash_DigestNVL": "0x80,00,16",
      "Flash_DigestNVH": "0x90,00,16",
      "Device_Config": "0xa0,00,04",
      "Recovery_Base_Addr0": "0xa0,04,04",
      "CRTM_Max_AutoMeasure_Size": "0xa0,08,04",
      "Reserved_0": "0xa0,12,04",
      "KTRL": "0xb0,00,16",
      "KTRH": "0xc0,00,16",
      "EEPROM_Attributes" : "0xd0,00,04",
      "PwrOnCount": "0xd0,04,04",
      "Device_Config_Volatile": "0xd0,08,04",
      "Reserved_1": "0xd0,12,04",
      "Reserved_2": "0xe0,00,144",
      "AppendConfig0": "0x170,00,04",
      "AppendConfig1": "0x170,04,04",
      "AppendConfig2": "0x170,08,04",
      "AppendConfig3": "0x170,12,04",
      "AMBitmap": "0x180,00,32",
      "Reserved_3": "0x1A0,00,96",
      "Customer_EEPROM": "0x200,00,256",
      "Customer_RAM": "0x300,00,96"
  }
