"""
Common functions used accross the sample applications and tests.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import ctypes as ct
import hmac
from .cmd_error_code import NaxosCmdErrorCode
import authenta.authenta_key_lib_interface as authenta_key_lib

class CRTM_Registers():
    """ CRTM register addresses
    """
    Flash_PCR0_Low_Addr     = 0x00
    Flash_PCR0_High_Addr    = 0x10
    CRTM_Range_Start_Addr_0 = 0x20
    CRTM_Range_Size_0       = 0x24
    CRTM_Range_Start_Addr_1 = 0x28
    CRTM_Range_Size_1       = 0x2C
    CRTM_Range_Start_Addr_2 = 0x30
    CRTM_Range_Size_2       = 0x34
    CRTM_Range_Start_Addr_3 = 0x38
    CRTM_Range_Size_3       = 0x3C
    CRTM_Range_Start_Addr_4 = 0x40
    CRTM_Range_Size_4       = 0x44
    CRTM_Range_Start_Addr_5 = 0x48
    CRTM_Range_Size_5       = 0x4C
    CRTM_Range_Start_Addr_6 = 0x50
    CRTM_Range_Size_6       = 0x54
    CRTM_Range_Start_Addr_7 = 0x58
    CRTM_Range_Size_7       = 0x5C
    Flash_DigestNV_Partial_L = 0x60
    Flash_DigestNV_Partial_H = 0x70
    Flash_DigestNV_L        = 0x80
    Flash_DigestNV_H        = 0x90
    Device_Config           = 0xA0
    Recovery_Block_Base_Addr = 0xA4
    CRTM_Max_AutoMeasure_Size = 0xA8
    KTR_Base                = 0xB0
    EEPROM_Attributes       = 0xD0
    Power_On_Count          = 0xD4
    Device_Config_Volatile  = 0xD8
    Append_Config_0         = 0x170
    Append_Config_1         = 0x174
    Append_Config_2         = 0x178
    Append_Config_3         = 0x17C
    Auth_Mod_Block_Config   = 0x180
    Customer_EEPROM         = 0x200
    Customer_RAM            = 0x300

CRTM_Adress_Ranges = {
    0: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_0,
        "size": CRTM_Registers.CRTM_Range_Size_0
    },
    1: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_1,
        "size": CRTM_Registers.CRTM_Range_Size_1
    },
    2: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_2,
        "size": CRTM_Registers.CRTM_Range_Size_2
    },
    3: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_3,
        "size": CRTM_Registers.CRTM_Range_Size_3
    },
    4: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_4,
        "size": CRTM_Registers.CRTM_Range_Size_4
    },
    5: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_5,
        "size": CRTM_Registers.CRTM_Range_Size_5
    },
    6: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_6,
        "size": CRTM_Registers.CRTM_Range_Size_6
    },
    7: {
        "addr": CRTM_Registers.CRTM_Range_Start_Addr_7,
        "size": CRTM_Registers.CRTM_Range_Size_7
    }
}


def warn_on_response_verification_fail(err: NaxosCmdErrorCode):
  if err.valid_response is not None and not err.valid_response:
    # Notify that we could not verify the response
    print('WARN: ' + err.cmd_name + ': Response verification failure!')
    if err.signature_matched is not None and not err.signature_matched:
      print('WARN: response from Authenta command ' + err.cmd_name + ' failed signature verification!')
    if err.nonce_matched is not None and not err.nonce_matched:
      print('WARN: response from Authenta command ' + err.cmd_name + ' failed NONCE match!')
    if err.last_opcode_matched is not None and not err.last_opcode_matched:
      print('WARN: response from Authenta command ' + err.cmd_name + ' failed Sub-OpCode match!')


def change_word_order(data):
  """ Change word order of response data from a Naxos 1.4 device

      This function serves as a utility for fixing the word order of data from
      a Naxos 1.4 device. It expects input bytes read from a 1.4 device, and
      outputs bytes with words in the correct order.

      Args:
        `data`: A byte-encoded value, of which to modify the word order.

      Returns:
        A byte-encoded value, with word order swapped in keeping with 1.4
        rules.
  """
  data = bytearray(data)
  data_len = len(data)
  assert(data_len % 8 == 0)

  for i in range(0, data_len // 2, 4):
    off = data_len - i
    data[i:(i+4)], data[(off-4):off] = data[(off-4):off], data[i:(i+4)]

  return bytes(data)

def get_dev_info(device):
  """ Get device uid, current amtc0, power-on amtc0, and provision status

      Returns:
        (uid, amtc0, boot_amtc0, provision_sts) on success or None on failure
  """
  # Request_UID
  err, *_ = device.meta_request_uid()
  if not err.esr_and_response_success:
    return None

  uid = device.uid
  amtc0 = device.amtc0
  provision_sts = device.provision_sts_struct

  if provision_sts.Provisioned != 0:
    # get power-on MTC0 value
    # CRTM_Register_Read read 16 bytes at one time
    # read 0xD0 instead of 0xD4
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.EEPROM_Attributes)
    if not err.esr_and_response_success:
      return None
    boot_amtc0 = response[20:24]
  else:
    boot_amtc0 = None

  return uid, amtc0, boot_amtc0, provision_sts

def check_srk(device):
  """ Check whether SRK is the correct key

      Returns:
        bool: True if correct or False if not or error occurs
  """
  # Request_UID
  err, response, *_ = device.meta_request_uid()
  if not err.esr_and_response_success:
    return False

  signature = device.get_response_signature(response)
  data = response[:32]

  # check signature
  if not device._root_signing_key_externally_set:
    if authenta_key_lib.set_srk_as_hmac_root_signing_key(device.keylib_context):
      device.logger.info('set SRK as signing key')
    else:
      device.logger.error('Failed to set signing key (SRK)!')
      return False

  verified = authenta_key_lib.verify_root_sha256_hmac(device.keylib_context, data, signature)
  if not verified:
    device.logger.error("SRK in the key xml file is not correct")

  return verified

def set_crtm_addr_range_reg(device, config, range_id, crtm_addr, crtm_size):
  """ Set a CRTM Address Measurement Range Value (start address and length)

      Returns:
        bool: True on success or False on failure
  """
  sign = True if config == b"\x00" else False

  addr = CRTM_Adress_Ranges[range_id]["addr"]
  data = crtm_addr.to_bytes(4, byteorder='big')
  err, _ = device.meta_crtm_register_write(config, addr, data, sign)
  if not err.esr_and_response_success:
    return False

  addr = CRTM_Adress_Ranges[range_id]["size"]
  data = crtm_size.to_bytes(4, byteorder='big')
  err, _ = device.meta_crtm_register_write(config, addr, data, sign)
  if not err.esr_and_response_success:
    return False

  return True

def get_flash_digest(device, addr: int):
  """ Get flash digest value

      Returns:
        digest on success or None on failure
  """
  digest = None
  config = b"\x00"

  err, response = device.meta_crtm_register_read(config, addr)
  if not err.esr_and_response_success:
    device.logger.error("Failed to read data from CRTM register 0x%X" %(addr))
    return digest

  digest_l = change_word_order(response[16:32])

  addr += 16
  err, response = device.meta_crtm_register_read(config, addr)
  if not err.esr_and_response_success:
    device.logger.error("Failed to read data from CRTM register 0x%X" %(addr))
    return digest

  digest_h = change_word_order(response[16:32])

  digest = b"".join((digest_h, digest_l))
  assert(len(digest) == 32)

  return digest

def write_data_reverse(device, config, addr, data, sign):
  """ Write data to CRTM registers in reverse word order.

      Returns:
        bool: True on success or False on failure
  """
  data_len = len(data)
  assert(data_len % 4 == 0)

  for i in range(0, data_len, 4):
    # CRTM_Register_Write
    data_to_write = data[i:(i + 4)]
    addr_to_write = addr + data_len - 4 - i

    err, _ = device.meta_crtm_register_write(config, addr_to_write, data_to_write, sign)
    if not err.esr_and_response_success:
      return False

  return True

def write_data_to_block(device, addr: int, data: bytes):
  """ Write data to a legacy block with the following steps:
      (1) unlock the volatile lock of the legacy block;
      (2) erase the content;
      (3) write data to it;
      (4) lock the volatile lock of the legacy block

      Returns:
        bool: True on success or False on failure
  """
  # unlock the legacy block to write data to
  ret = device.meta_dyb_write(False, addr)
  if not ret:
    device.logger.error("Failed to unlock the volatile lock of legacy block 0x%08X" %(addr))
    return False

  # erase the legacy block
  ret = device.meta_subsector_erase(addr)
  if not ret:
    device.meta_dyb_write(True, addr)
    device.logger.error("Failed to erase legacy block 0x%08X" %(addr))
    return False

  # write data to the legacy block
  # Page_Proram can write at most 256-byte to flash
  max_write_size = 256

  for start in range(0, len(data), max_write_size):
    if len(data) - start <  max_write_size:
      write_size = len(data) - start
    else:
      write_size = max_write_size

    end = start + write_size
    ret = device.meta_page_program(addr + start, data[start:end])
    if not ret:
      device.meta_dyb_write(True, addr)
      device.logger.error("Failed to write data to legacy block 0x%08X" %(addr))
      return False

  # lock the legacy block
  ret = device.meta_dyb_write(True, addr)
  if not ret:
    device.logger.error("Failed to lock the volatile lock of legacy block 0x%08X" %(addr))
    return False

  return True

def enable_automeasure(device):
  """ Enable automeasure

      Returns:
        bool: True on success or Flase on failure
  """
  # read current device config
  err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
  if not err.esr_and_response_success:
    return False

  device_config = response[16:20]
  device_config_int = int.from_bytes(device_config, byteorder="big")
  data_int = device_config_int | 0x00000001
  data = data_int.to_bytes(4, byteorder="big")

  err, _ = device.meta_acrtm_update_start(b"\x00")
  if not err.esr_and_response_success:
    return False

  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    return False

  err, _ = device.meta_acrtm_update_finish(b"\x00")
  if not err.esr_and_response_success:
    return False

  return True

def disable_automeasure(device):
  """ Disable automeasure

      Returns:
        bool: True on success or Flase on failure
  """
  # read current device config
  err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
  if not err.esr_and_response_success:
    return False

  device_config = response[16:20]
  device_config_int = int.from_bytes(device_config, byteorder="big")
  data_int = device_config_int & 0xFFFFFFFE
  data = data_int.to_bytes(4, byteorder="big")

  err, _ = device.meta_acrtm_update_start(b"\x00")
  if not err.esr_and_response_success:
    return False

  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    return False

  err, _ = device.meta_acrtm_update_finish(b"\x00")
  if not err.esr_and_response_success:
    return False

  return True

def configure_am(device, addr, to_am):
  """ Configure a block in the main array address space to AM/legacy block
      Set to_am to True if configuring to AM block
      Set to_am to False if configuring to legacy block

      Returns:
        bool: True on success or False on failure
  """
  am_blcok_size = 0x10000
  am_block_num = 256
  am_block_num_half = int(am_block_num / 2)

  block_num = int(addr / am_blcok_size)
  if block_num < am_block_num_half:
    # low ambitmap register
    ambitmap_base_addr = CRTM_Registers.Auth_Mod_Block_Config
  else:
    # high ambitmap register
    ambitmap_base_addr = CRTM_Registers.Auth_Mod_Block_Config + 0x10
    block_num -= am_block_num_half

  ambitmap_addr = ambitmap_base_addr + 4 * int(block_num / 32)

  # read AMBitmap
  err, response = device.meta_crtm_register_read(b"\x00", ambitmap_base_addr)
  if not err.esr_and_response_success:
    return False

  ambitmap_data_read = response[16:32]

  # write AMBitmap data
  start = ambitmap_addr - ambitmap_base_addr
  if to_am:
    # configure to AM block
    ambitmap_data_write = int(ambitmap_data_read[start:(start + 4)].hex(), 16) & (~(0x1 << (block_num % 32)))
  else:
    # convert back to legacy block
    ambitmap_data_write = int(ambitmap_data_read[start:(start + 4)].hex(), 16) | (0x1 << (block_num % 32))

  err, _ = device.meta_acrtm_update_start(b"\x00")
  if not err.esr_and_response_success:
    return False

  err, response = device.meta_crtm_register_write(b"\x00", ambitmap_addr, ambitmap_data_write.to_bytes(4, byteorder='big'), True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    return False

  err, _ = device.meta_acrtm_update_finish(b"\x00")
  if not err.esr_and_response_success:
    return False

  # read AMBitmap to verify
  err, response = device.meta_crtm_register_read(b"\x00", ambitmap_base_addr)
  if not err.esr_and_response_success:
    return False

  ambitmap_data_read = response[16:32]

  if int(ambitmap_data_read[start:(start + 4)].hex(), 16) != ambitmap_data_write:
    return False

  return True
