/***
 * 
 * @file authenta_key_driver.c
 * 
 * @brief Combined Naxos Indigo version of the key library (and crypto lib interface) implementation
 * 
 * @copyright (C) 2020 Micron Technology, Inc. All rights reserved.
 * 
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>

#if !defined(MBEDTLS_CONFIG_FILE)
#include "mbedtls/config.h"
#else
#include MBEDTLS_CONFIG_FILE
#endif

#if defined(MBEDTLS_PLATFORM_C)
  #include "mbedtls/platform.h"
#else
  #include <stdio.h>
  #include <stdlib.h>
  #define mbedtls_fprintf         fprintf
  #define mbedtls_printf          printf
  #define MBEDTLS_EXIT_SUCCESS    EXIT_SUCCESS
  #define MBEDTLS_EXIT_FAILURE    EXIT_FAILURE
#endif /* MBEDTLS_PLATFORM_C */

#include "mbedtls/aes.h"
#include "mbedtls/md.h"
#include "mbedtls/sha256.h" /* SHA-256 only */
#include "easy_ecc/ecc.h"

#include "authenta_key_management_utils.h"
#include "authenta_key_driver.h"

#define MAX_KTS_KEY_NAME_LEN 32

static void _clear_context_keys(KeyLibraryContext context)
{
    if(context->hmac_root_key != NULL) {
        memset(context->hmac_root_key, 0, AUTHENTA_HMAC_KEY_LEN);
        free(context->hmac_root_key);
    }
    if(context->hmac_session_key != NULL) {
        memset(context->hmac_session_key, 0, AUTHENTA_HMAC_KEY_LEN);
        free(context->hmac_session_key);
    }
    if(context->ecdh_pub_key != NULL) {
        memset(context->ecdh_pub_key, 0, AUTHENTA_ECDSA_PUBLIC_KEY_LEN);
        free(context->ecdh_pub_key);
    }
    if(context->ecdsa_signing_key != NULL) {
        memset(context->ecdsa_signing_key, 0, AUTHENTA_ECDSA_PRIVATE_KEY_LEN);
        free(context->ecdsa_signing_key);
    }
    if(context->ecdsa_verification_key != NULL) {
        memset(context->ecdsa_verification_key, 0, AUTHENTA_ECDSA_PRIVATE_KEY_LEN+1);
        free(context->ecdsa_verification_key);
    }
    if(context->uid != NULL) {
        free(context->uid);
    }
    memset(context, 0, sizeof(KeyLibraryContext_struct));
}

static void get_kts_ecdsa_slot_key_name(uint32_t slot_num, char *key_name)
{
    assert(key_name != NULL);
    sprintf(key_name, "ecdsa_prv_slot_%i", slot_num);
}

static void get_kts_ecdh_slot_key_name(uint32_t slot_num, char *key_name)
{
    assert(key_name != NULL);
    sprintf(key_name, "ecdh_prv_slot_%i", slot_num);
}

static void bytesToHex(const uint8_t* bytes_arr, uint32_t bytes_arr_len, char* hex_arr, uint32_t hex_arr_len)
{
    assert(bytes_arr != NULL && 
           hex_arr != NULL && 
           hex_arr_len >= (2*bytes_arr_len));

    for(int i = 0; i < bytes_arr_len; i++)
    {
        sprintf(&hex_arr[i*2], "%02X", bytes_arr[i]);
    }
}

/**
 * key_lib_open
 * 
 * This should always be called before using the key library and returns a context that must
 * be used in subsequent calls to the library.
 * The retuned context holds state and is associated with a particular device.  There this
 * open function must be called once for every deferent device instance that an application will 
 * use.
 * 
 * Performs any rquired library and context initialization.
 * 
 * @param[in] uid UID of the device associated with this open instance of the key library.
 * 
 * @param[in] device_info_dictionary JSON formatted set of name-value pairsproperties of the associated device.
 *                                   The contents will include at least:
 *                                   device_type: [Naxos | Indigo]
 *                                   device_version: [hex_string representing version returned by device,
 *                                                    e.g., '010005' ]
 *                                   Not used in this implementation.
 * 
 * @param[in] user_defined_data this data that may be used by a particular implementation of the
 *                              key driver.
 * 
 * @param[out] context_buf object that will be a parameter passed to all subsequent library function calls. 
 *                         The returned context is associated with a single device. For additional devices
 *                         call this function to get a new context for each additional device instance.
 * 
 * @return KeyLibrary_OK      success
 *         KeyLibrary_MEMORY  failed allocating memory
 *         KeyLibrary_BAD_PARAMETER no uid passed in
 * . 
 **/
KeyLibraryErrorCode key_lib_open(
    const char * const uid, 
    const char * const device_info_json, 
    const uint8_t * const user_defined_data, 
    uint8_t* context_buf)
{
    if(_KDI_LOG_TRACE_) {
        printf("[key_lib_open] Entered\n");
    }

    if(NULL == uid || strlen(uid) <= 0)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    KeyLibraryContext context = (KeyLibraryContext_struct*) malloc(sizeof(KeyLibraryContext_struct));
    if(context == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[key_lib_open] ERROR allocating memory for for context object FAILED (uid %s)!\n", uid);
        }
        return KeyLibrary_MEMORY;
    }

    memset(context, 0, sizeof(KeyLibraryContext_struct));
    context->uid = (uint8_t*) malloc(strlen(uid) + 1);
    if(context->uid == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[key_lib_open] ERROR allocating memory for uid %s FAILED!\n", uid);
        }
        return KeyLibrary_MEMORY;
    }
    strcpy((char*)context->uid, uid);

    memcpy(context_buf, (unsigned char*)&context, sizeof(context));

    return KeyLibrary_OK;
}

/**
 * key_lib_close
 * 
 * This should always be called after a use of a context is completed.
 * 
 * The memeory, information, and state associated with the input context will
 * be freed and use of a context after passing it to this function is undefined and
 * may cause fatal runtime execution error.
 * 
 * Use of the library with another context can continue.
 * 
 * @param[in] context the context to be closed.
 * 
 **/
KeyLibraryErrorCode key_lib_close(uint8_t* context_buf)
{
    if(_KDI_LOG_TRACE_) {
        printf("[key_lib_close] Entered\n");
    }

    if(context_buf != NULL)
    {
        KeyLibraryContext context = *((KeyLibraryContext*)context_buf);
        if(_KDI_LOG_TRACE_) {
            printf("[key_lib_close] Freeing allocated memory for context structure\n");
        }

        _clear_context_keys(context);
        free(context);
    }

    return KeyLibrary_OK;
}


/**
 * key_lib_clear_keys
 * 
 * Reset any key state (i.e. remove any cached keys from  the context)
 * 
 * @param context[in]  the context object returned from key_lib_open
 * 
 * @return KeyLibrary_OK success
 *         KeyLibrary_BAD_PARAMETER if the input context or uid are NULL or invalid
 **/
KeyLibraryErrorCode key_lib_clear_keys(const uint8_t* context_buf)
{
    if(_KDI_LOG_TRACE_) {
        printf("[key_lib_clear_keys] Entered\n");
    }

    if(context_buf == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf( "[key_lib_clear_keys] ERROR Bad input (context_buf is NULL)!\n");
        }
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    // save the uid if its there
    uint8_t* uid = NULL;
    if(context->uid)
    {
        uid = context->uid;
        context->uid = NULL;
    }
    
    _clear_context_keys(context);
    memset(context, 0, sizeof(KeyLibraryContext_struct));

    if(uid)
    {
        context->uid = uid;
    }

    return KeyLibrary_OK;
}


KeyLibraryErrorCode set_srk_as_hmac_root_signing_key(const uint8_t* context_buf, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->uid == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(force || context->hmac_root_key == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_srk_as_hmac_root_signing_key] Fetching key from file...\n");
        }
        if(NULL == context->hmac_root_key)
        {
            context->hmac_root_key = malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->hmac_root_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_srk_as_hmac_root_signing_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        uint32_t key_len = AUTHENTA_HMAC_KEY_LEN;
        uint32_t rc = get_key(context->uid, KeyName_SRK, context->hmac_root_key, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_srk_as_hmac_root_signing_key] key not found for %s/%s\n", context->uid, KeyName_SRK);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_HMAC_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[128];
                bytesToHex(context->hmac_root_key, 32, out_hex, sizeof(out_hex));
                printf("[set_srk_as_hmac_root_signing_key] key found, but wrong length! (%s)\n", context->hmac_root_key);
            }
            memset(context->hmac_root_key, 0, AUTHENTA_HMAC_KEY_LEN);
            free(context->hmac_root_key);
            context->hmac_root_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->hmac_root_key, 32, out_hex, sizeof(out_hex));
        printf("[set_srk_as_hmac_root_signing_key] HMAC Root key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}

KeyLibraryErrorCode set_crk_as_hmac_root_signing_key(const uint8_t* context_buf, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(force || context->hmac_root_key == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_crk_as_hmac_root_signing_key] Fetching key from file...\n");
        }

        if(NULL == context->hmac_root_key)
        {
            context->hmac_root_key = malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->hmac_root_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_crk_as_hmac_root_signing_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        uint32_t key_len = AUTHENTA_HMAC_KEY_LEN;
        uint32_t rc = get_key(context->uid, KeyName_CRK, context->hmac_root_key, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_crk_as_hmac_root_signing_key] key not found for %s/%s\n", context->uid, KeyName_CRK);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_HMAC_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[128];
                bytesToHex(context->hmac_root_key, 32, out_hex, sizeof(out_hex));
                printf("[set_crk_as_hmac_root_signing_key] key found, but wrong length! (%s)\n", context->hmac_root_key);
            }
            memset(context->hmac_root_key, 0, AUTHENTA_HMAC_KEY_LEN);
            free(context->hmac_root_key);
            context->hmac_root_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->hmac_root_key, 32, out_hex, sizeof(out_hex));
        printf("[set_crk_as_hmac_root_signing_key] HMAC Root key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}


KeyLibraryErrorCode set_csk_as_hmac_session_signing_key(const uint8_t* context_buf, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(force || context->hmac_session_key == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_crk_as_hmac_root_signing_key] Fetching key from file...\n");
        }

        if(NULL == context->hmac_session_key)
        {
            context->hmac_session_key = malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->hmac_session_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_crk_as_hmac_root_signing_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        uint32_t key_len = AUTHENTA_HMAC_KEY_LEN;
        uint32_t rc = get_key(context->uid, KeyName_CSK, context->hmac_session_key, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_crk_as_hmac_root_signing_key] key not found for %s/%s\n", context->uid, KeyName_CSK);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_HMAC_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[128];
                bytesToHex(context->hmac_session_key, 32, out_hex, sizeof(out_hex));
                printf("[set_crk_as_hmac_root_signing_key] key found, but wrong length! (%s)\n", context->hmac_session_key);
            }
            memset(context->hmac_session_key, 0, AUTHENTA_HMAC_KEY_LEN);
            free(context->hmac_session_key);
            context->hmac_session_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->hmac_session_key, 32, out_hex, sizeof(out_hex));
        printf("[set_crk_as_hmac_root_signing_key] HMAC Root key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}


KeyLibraryErrorCode set_ecdsa_signing_key(const uint8_t* context_buf, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->uid == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(force || context->ecdsa_signing_key == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_ecdsa_signing_key] Fetching key from file...\n");
        }
        if(NULL == context->ecdsa_signing_key)
        {
            context->ecdsa_signing_key = malloc(AUTHENTA_ECDSA_PRIVATE_KEY_LEN);
            if(context->ecdsa_signing_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_ecdsa_signing_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        uint32_t key_len = AUTHENTA_ECDSA_PRIVATE_KEY_LEN;
        uint32_t rc = get_key(context->uid, KeyName_ECDSA_PRV, context->ecdsa_signing_key, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_ecdsa_signing_key] key not found for %s/%s\n", context->uid, KeyName_ECDSA_PRV);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_ECDSA_PRIVATE_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[128];
                bytesToHex(context->ecdsa_signing_key, 64, out_hex, sizeof(out_hex));
                printf("[set_ecdsa_signing_key] key found, but wrong length! (%s)\n", context->ecdsa_signing_key);
            }
            memset(context->ecdsa_signing_key, 0, key_len);
            free(context->ecdsa_signing_key);
            context->ecdsa_signing_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[AUTHENTA_ECDSA_PRIVATE_KEY_LEN*2+1];
        bytesToHex(context->ecdsa_signing_key, AUTHENTA_ECDSA_PRIVATE_KEY_LEN, out_hex, sizeof(out_hex));
        printf("[set_ecdsa_signing_key] ECDSA signing key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}


/**
 * Set the HMAC root key to the input key
 * 
 * This is typically used for setting the HMAC key to the CRK read from the KTR
 * 
 **/
KeyLibraryErrorCode set_hmac_root_signing_key(const uint8_t* context_buf, const char* key, uint32_t key_len)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(key == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(key_len < AUTHENTA_HMAC_KEY_LEN)
    {
        return KeyLibrary_INVALID_KEY;
    }

    if(_KDI_LOG_TRACE_) {
        printf("[set_hmac_root_signing_key] setting HMAC signing key to input key.\n");
    }

    if(context->hmac_root_key == NULL)
    {
        context->hmac_root_key = malloc(key_len);
        if(context->hmac_root_key == NULL)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_hmac_root_signing_key] malloc Failure!\n");
            }
            return KeyLibrary_MEMORY;
        }
    }
    
    memcpy(context->hmac_root_key, key, key_len);

    return KeyLibrary_OK;   
}

/**
 * Set the HMAC session key to the input key
 * 
 * This is typically used for setting the HMAC session key to the CSK read from the KTR.
 **/
KeyLibraryErrorCode set_hmac_session_key(const uint8_t* context_buf, const char* key, uint32_t key_len)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(key == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(key_len < AUTHENTA_HMAC_KEY_LEN)
    {
        return KeyLibrary_INVALID_KEY;
    }

    if(_KDI_LOG_TRACE_) {
        printf("[set_hmac_root_signing_key] setting HMAC signing key to input key.\n");
    }

    if(context->hmac_session_key == NULL)
    {
        context->hmac_session_key = malloc(key_len);
        if(context->hmac_session_key == NULL)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_hmac_root_signing_key] malloc Failure!\n");
            }
            return KeyLibrary_MEMORY;
        }
    }

    memcpy(context->hmac_session_key, key, key_len);

    return KeyLibrary_OK;   
}

KeyLibraryErrorCode set_ecdsa_verification_key(const uint8_t* context_buf, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->uid == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(force || context->ecdsa_verification_key == NULL)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_ecdsa_verification_key] Fetching key from file...\n");
        }
        if(context->ecdsa_verification_key == NULL)
        {
            context->ecdsa_verification_key = malloc(AUTHENTA_ECDSA_PUBLIC_KEY_LEN+1);
            if(context->ecdsa_verification_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_ecdsa_verification_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        context->ecdsa_verification_key[0] = '\04';
        uint32_t key_len = AUTHENTA_ECDSA_PUBLIC_KEY_LEN;        
        uint32_t rc = get_key(context->uid, KeyName_ECDSA_DS_PuK, context->ecdsa_verification_key+1, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_ecdsa_verification_key] key not found for %s/%s\n", context->uid, KeyName_ECDSA_PUB);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_ECDSA_PUBLIC_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[129];
                bytesToHex(context->ecdsa_verification_key+1, 64, out_hex, sizeof(out_hex));
                printf("[set_ecdsa_verification_key] key found, but wrong length! (%s)\n", context->ecdsa_verification_key);
            }
            memset(context->ecdsa_verification_key, 0, key_len+1);
            free(context->ecdsa_verification_key);
            context->ecdsa_verification_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[AUTHENTA_ECDSA_PUBLIC_KEY_LEN*2+1];
        bytesToHex(context->ecdsa_verification_key+1, AUTHENTA_ECDSA_PUBLIC_KEY_LEN, out_hex, sizeof(out_hex));
        printf("[set_ecdsa_verification_key] HMAC Verification key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}


KeyLibraryErrorCode _sha256_hmac(const uint8_t *key, int key_len, const uint8_t *msg_buf, const int msg_size, uint8_t *digest)
{
    assert(key != NULL && msg_buf != NULL && digest != NULL);

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(key, key_len, out_hex, sizeof(out_hex));
        printf("[_sha256_hmac] HMAC Using key: 0x%s\n", out_hex);
    }

    uint8_t tmp[32];
    mbedtls_md_context_t sha_ctx;
    mbedtls_md_init( &sha_ctx );
    int rc_mbed = mbedtls_md_setup( &sha_ctx, mbedtls_md_info_from_type( MBEDTLS_MD_SHA256 ), 1 );
    if(rc_mbed != 0)
    {
        mbedtls_printf( "  ! mbedtls_md_setup() returned -0x%04x\n", -rc_mbed );
        return KeyLibrary_CRYPTO;
    }

    rc_mbed = mbedtls_md_hmac_starts( &sha_ctx, key, key_len);
    if(rc_mbed != 0)
    {
        mbedtls_printf( "  ! mbedtls_md_hmac_starts() returned -0x%04x\n", -rc_mbed );
        return KeyLibrary_CRYPTO;
    }
    int n;
    for(int offset = 0; offset < msg_size; offset += sizeof(tmp))
    {
        n = ( msg_size - offset > sizeof(tmp) ) ? sizeof(tmp) : (int) ( msg_size - offset );

        memcpy(tmp, &msg_buf[offset], n);
        rc_mbed = mbedtls_md_hmac_update( &sha_ctx, tmp, n);
        if(rc_mbed != 0)
        {
            mbedtls_printf( "  ! mbedtls_md_hmac_update() returned -0x%04x\n", -rc_mbed );
            return KeyLibrary_CRYPTO;
        }
    }
    rc_mbed = mbedtls_md_hmac_finish( &sha_ctx, digest);
    if(rc_mbed != 0)
    {
        mbedtls_printf( "  ! mbedtls_md_hmac_finish() returned -0x%04x\n", -rc_mbed );
        return KeyLibrary_CRYPTO;
    }

    if(_KDI_LOG_TRACE_) {
        char dig_hex[65];
        bytesToHex(digest, 32, dig_hex, sizeof(dig_hex));
        printf("[_sha256_hmac] digest: %s\n", dig_hex);
    }

    return KeyLibrary_OK;
}


/*
 *
 * @param context_buf[in]  the object returned by authneta_keylib_context_open
 * @param powerOnAmtc0[in] MUST be an exactly 4-byte buffer
 * @param force[in] generate the key again even if it has already been generated already previously
 * 
 * @return success: KeyLibrary_OK
 */
KeyLibraryErrorCode derive_hmac_session_key(const uint8_t* context_buf, const uint8_t * powerOnAmtc0, uint32_t powerOnAmtc0_len, bool force)
{
    KeyLibraryErrorCode rc = KeyLibrary_OK;

    if(context_buf == NULL || powerOnAmtc0 == NULL || powerOnAmtc0_len == 0)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(powerOnAmtc0_len > sizeof(context->pwr_on_amtc0))
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(context->hmac_root_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    if(_KDI_LOG_TRACE_) {
        printf("[derive_hmac_session_key]      input powerOnAmtc0=0x");
        for(int i=0; i < powerOnAmtc0_len; i++) {
            printf("%02x", powerOnAmtc0[i]);
        }
        printf("\n[derive_hmac_session_key] context->pwr_on_amtc0=0x");
        for(int i=0; i < context->pwr_on_amtc0_len; i++) {
            printf("%02x", context->pwr_on_amtc0[i]);
        }
        printf("\n");
    }


    if(force || context->hmac_session_key == NULL || powerOnAmtc0_len != context->pwr_on_amtc0_len ||
       memcmp(powerOnAmtc0, context->pwr_on_amtc0, context->pwr_on_amtc0_len))
    {
        if(_KDI_LOG_TRACE_) {
            printf("[derive_hmac_session_key] generating session key...\n");
        }
        if(context->hmac_session_key == NULL)
        {
            context->hmac_session_key = (uint8_t*) malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->hmac_session_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[derive_hmac_session_key] ERROR Allocating memory for the session key failed!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }

        rc = _sha256_hmac(context->hmac_root_key, 
                          AUTHENTA_HMAC_KEY_LEN, 
                          powerOnAmtc0, 
                          powerOnAmtc0_len, 
                          context->hmac_session_key);
        if(rc == KeyLibrary_OK)
        {
            context->pwr_on_amtc0_len = powerOnAmtc0_len;
            memcpy(context->pwr_on_amtc0, powerOnAmtc0, powerOnAmtc0_len);
        }
        else
        {
            if(_KDI_LOG_TRACE_) {
                printf("[derive_hmac_session_key] ERROR HMAC failed while generating session key! (freeing allocated memory for session key)\n");
            }
            // clean up
            free(context->hmac_session_key);
            context->hmac_session_key = NULL;
        }
        
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->hmac_session_key, 32, out_hex, sizeof(out_hex));
        printf("[derive_hmac_session_key] Session Key: %s\n", out_hex);
    }


    return rc;
}


/*
 *
 * @param context_buf[in]  the object returned by authneta_keylib_context_open
 * @param ecdh_pub_key[in] MUST be an exactly AUTHENTA_ECDH_PUBLIC_KEY_LEN length (64-byte) buffer
 * @param powerOnAmtc0[in] the amtc0 value at poweron to be used on creating the session key as
 *                         as defined in the Authenta specifications
 * @param powerOnAmtc0_len[in] the length of the powerOnAmtc0 buffer
 * @param force[in]        if true derive the session key even if there is one stored already,
 *                         if false the implementation is allowed to reuse one it is holding
 *                         as long as the inputs match those used when the current one was derived
 * 
 * @return success: KeyLibrary_OK
 */
KeyLibraryErrorCode derive_ecdh_session_key(
    const uint8_t* context_buf, 
    const uint8_t * ecdh_pub_key, 
    const uint8_t * powerOnAmtc0, 
    uint32_t powerOnAmtc0_len,
    bool force)
{
    if(_KDI_LOG_TRACE_) {
        printf("[derive_ecdh_session_key] entered\n");
    }

    KeyLibraryErrorCode rc = KeyLibrary_OK;

    if(powerOnAmtc0 == NULL || ecdh_pub_key == NULL || context_buf == NULL || powerOnAmtc0_len == 0)
    {
        if(_KDI_LOG_TRACE_) {
            if (powerOnAmtc0_len == 0) {
                printf("[derive_ecdh_session_key] returning Bad Paramter; powerOnAmtc0_len may not be 0!\n");
            }
            else {
                printf(
                  "[derive_ecdh_session_key] returning Bad Paramter the following pointers may not be NULL: powerOnAmtc0=%p; ecdh_pub_key=%p, context_buf=%p\n",
                  powerOnAmtc0, ecdh_pub_key, context_buf);
            }
        }
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(powerOnAmtc0_len > sizeof(context->pwr_on_amtc0))
    {
        return KeyLibrary_BAD_PARAMETER;
    }


    // if(_KDI_LOG_TRACE_) {
    //     printf("[derive_ecdh_session_key] context->pwr_on_amtc0=%i: input powerOnAmtc0=%i\n", *((int*)context->pwr_on_amtc0), *((int*) powerOnAmtc0));
    // }

    if(force || context->hmac_session_key == NULL || powerOnAmtc0_len != context->pwr_on_amtc0_len 
       || memcmp(context->pwr_on_amtc0, powerOnAmtc0, powerOnAmtc0_len)
       || context->ecdh_pub_key == NULL || memcmp(context->ecdh_pub_key+1, ecdh_pub_key, AUTHENTA_ECDH_PUBLIC_KEY_LEN))
    {
        if(_KDI_LOG_TRACE_) {
            printf("[derive_ecdh_session_key] generating session key...\n");
        }

        if(context->ecdh_pub_key == NULL)
        {
            context->ecdh_pub_key = malloc(AUTHENTA_ECDH_PUBLIC_KEY_LEN+1);
            if(context->ecdh_pub_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[derive_ecdh_session_key] ERROR Allocating memory for the ECDH public key failed!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        context->ecdh_pub_key[0] = '\04';
        memcpy(context->ecdh_pub_key+1, ecdh_pub_key, AUTHENTA_ECDH_PUBLIC_KEY_LEN);

        /* get the ECDH private key from the key store */
        uint8_t ecdh_prv_key[AUTHENTA_ECDH_PRIVATE_KEY_LEN];
        uint32_t ecdh_prv_key_len=AUTHENTA_ECDH_PRIVATE_KEY_LEN;
        uint32_t rc = get_key(context->uid, KeyName_ECDH_SR_PRV, ecdh_prv_key, &ecdh_prv_key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[derive_ecdh_session_key] key not found for %s/%s\n", context->uid, KeyName_ECDH_SR_PRV);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(ecdh_prv_key_len != AUTHENTA_ECDH_PRIVATE_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[derive_ecdh_session_key] key found, but wrong length! (%i)\n", ecdh_prv_key_len);
            }
            return KeyLibrary_INVALID_KEY;
        }

        if(_KDI_LOG_TRACE_)
        {
            char out_hex[256];
            bytesToHex(&(context->ecdh_pub_key[1]), AUTHENTA_ECDH_PUBLIC_KEY_LEN, out_hex, sizeof(out_hex));
            printf("[derive_ecdh_session_key] ecdh_pub_key: %s\n", out_hex);
            bytesToHex(ecdh_prv_key, AUTHENTA_ECDH_PRIVATE_KEY_LEN, out_hex, sizeof(out_hex));
            printf("[derive_ecdh_session_key] ecdh_prv_key: %s\n", out_hex);
        }

        uint8_t shared_secret[32] = { 0 };
        if(0 == ecdh_shared_secret(context->ecdh_pub_key, ecdh_prv_key, shared_secret))
        {
            if(_KDI_LOG_TRACE_)
            {
                printf("[derive_ecdh_session_key] Failed to generate shared secret: ecdh library error\n");
            }
            return KeyLibrary_CRYPTO;
        }

        if(_KDI_LOG_TRACE_)
        {
            char out_hex[256];
            bytesToHex(shared_secret, 32, out_hex, sizeof(out_hex));
            printf("[derive_ecdh_session_key] shared secret: %s\n", out_hex);
        }

        if(context->hmac_session_key == NULL)
        {
            context->hmac_session_key = (uint8_t*) malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->hmac_session_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[derive_ecdh_session_key] ERROR Allocating memory for the session key failed!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }

        if(_KDI_LOG_TRACE_) {
            printf("[derive_ecdh_session_key] generating session key by HMACing powerOnAmtc0=%i\n", *((int*) powerOnAmtc0));
            char out_hex[32];
            bytesToHex(powerOnAmtc0, powerOnAmtc0_len, out_hex, sizeof(out_hex));
            printf("[derive_ecdh_session_key] powerOnAmtc0: %s\n", out_hex);
        }

        rc = _sha256_hmac(shared_secret, 
                          AUTHENTA_HMAC_KEY_LEN, 
                          powerOnAmtc0, 
                          powerOnAmtc0_len, 
                          context->hmac_session_key);
        if(rc == KeyLibrary_OK)
        {
            context->pwr_on_amtc0_len = powerOnAmtc0_len;
            memcpy(context->pwr_on_amtc0, powerOnAmtc0, powerOnAmtc0_len);
        }
        else
        {
            if(_KDI_LOG_TRACE_) {
                printf("[derive_ecdh_session_key] ERROR HMAC failed while generating session key! (freeing allocated memory for session key)\n");
            }
            // clean up
            free(context->hmac_session_key);
            context->hmac_session_key = NULL;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->hmac_session_key, 32, out_hex, sizeof(out_hex));
        printf("[derive_ecdh_session_key] Session Key: %s\n", out_hex);
    }


    return rc;
}


KeyLibraryErrorCode gen_root_sha256_hmac(
    const uint8_t* context_buf, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* digest)
{
    if(_KDI_LOG_TRACE_) {
        printf("[gen_root_sha256_hmac] enter\n");
    }
    if(context_buf == NULL || msg_buf == NULL || digest == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->hmac_root_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    return _sha256_hmac(context->hmac_root_key, AUTHENTA_HMAC_KEY_LEN,
                        msg_buf, msg_len,
                        digest);
}


KeyLibraryErrorCode gen_session_sha256_hmac(
    const uint8_t* context_buf, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* digest)
{
    if(_KDI_LOG_TRACE_) {
        printf("[gen_session_sha256_hmac] enter\n");
    }
    if(context_buf == NULL || msg_buf == NULL || digest == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->hmac_session_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }
    return _sha256_hmac(context->hmac_session_key, AUTHENTA_HMAC_KEY_LEN,
                        msg_buf, msg_len,
                        digest);
}

KeyLibraryErrorCode gen_kts_sha256_hmac(
    const uint8_t* context_buf, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* digest)
{
    if(_KDI_LOG_TRACE_) {
        printf("[gen_kts_sha256_hmac] enter\n");
    }
    if(context_buf == NULL || msg_buf == NULL || digest == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->kts_hmac_signing_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    return _sha256_hmac(context->kts_hmac_signing_key, AUTHENTA_HMAC_KEY_LEN,
                        msg_buf, msg_len,
                        digest);
}



KeyLibraryErrorCode generate_ecdsa_signature
(
    const uint8_t* context_buf, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* signature)
{
    if(_KDI_LOG_TRACE_) {
        printf("[generate_ecdsa_signature] enter\n");
    }
    if(context_buf == NULL || msg_buf == NULL || signature == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->ecdsa_signing_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

	//calculate sha256
	uint8_t msg_hash[32] = { 0 };
	mbedtls_sha256_ret(msg_buf, msg_len, msg_hash, 0);

    if(_KDI_LOG_TRACE_)
    {
        printf("[generate_ecdsa_signature] generated sha256 hash of message\n");
        printf("msg_hash: 0x");
        for(int i=0; i < sizeof(msg_hash); i++)
        {
            printf("%02x", msg_hash[i]);
        }
        printf("\n");
    }

	int ret = ecdsa_sign(context->ecdsa_signing_key, msg_hash, signature);	

    
    if(_KDI_LOG_TRACE_)
    {
        if(ret == 0)
        {
            printf("[generate_ecdsa_signature] ERROR ecdsa_sign failed!");
        }
        else
        {
            printf("[generate_ecdsa_signature] generated ecdsa signature\n");
            uint32_t sig_len = AUTHENTA_ECDSA_PRIVATE_KEY_LEN * 2;
            printf("signature: 0x");
            for(int i=0; i < sig_len; i++)
            {
                printf("%02x", signature[i]);
            }
            printf("\n");
        }
    }

    if(ret == 0)
    {
        return KeyLibrary_CRYPTO;
    }
    else
    {
        return KeyLibrary_OK;
    }
}

KeyLibraryErrorCode generate_kts_ecdsa_signature
(
    const uint8_t* context_buf, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* signature)
{
    if(_KDI_LOG_TRACE_) {
        printf("[generate_kts_ecdsa_signature] enter\n");
    }
    if(context_buf == NULL || msg_buf == NULL || signature == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->kts_ecdsa_signing_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

	//calculate sha256
	uint8_t msg_hash[32] = { 0 };
	mbedtls_sha256_ret(msg_buf, msg_len, msg_hash, 0);

    if(_KDI_LOG_TRACE_)
    {
        printf("[generate_kts_ecdsa_signature] generated sha256 hash of message\n");
        printf("msg_hash: 0x");
        for(int i=0; i < sizeof(msg_hash); i++)
        {
            printf("%02x", msg_hash[i]);
        }
        printf("\n");
    }

	int ret = ecdsa_sign(context->kts_ecdsa_signing_key, msg_hash, signature);	

    
    if(_KDI_LOG_TRACE_)
    {
        if(ret == 0)
        {
            printf("[generate_kts_ecdsa_signature] ERROR ecdsa_sign failed!");
        }
        else
        {
            printf("[generate_kts_ecdsa_signature] generated ecdsa signature\n");
            uint32_t sig_len = AUTHENTA_ECDSA_PRIVATE_KEY_LEN * 2;
            printf("signature: 0x");
            for(int i=0; i < sig_len; i++)
            {
                printf("%02x", signature[i]);
            }
            printf("\n");
        }
    }

    if(ret == 0)
    {
        return KeyLibrary_CRYPTO;
    }
    else
    {
        return KeyLibrary_OK;
    }
}

static
KeyLibraryErrorCode _verify_sha256_hmac(
    const uint8_t * key, 
    const uint8_t * msg, 
    uint32_t msg_len, 
    const uint8_t * sig, 
    bool * verified)
{
    assert(key != NULL && msg != NULL && sig != NULL && verified != NULL);

    uint8_t calculated_digest[32];
    KeyLibraryErrorCode rc;
    rc = _sha256_hmac(key, AUTHENTA_HMAC_KEY_LEN,
                      msg, msg_len,
                      calculated_digest);
    if(rc != KeyLibrary_OK)
    {
        if(_KDI_LOG_TRACE_)
        {
            printf("[_verify_sha256_hmac] ERROR _sha256_hmac failed!");
        }
        return rc;
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[256];
        bytesToHex(key, 32, out_hex, sizeof(out_hex));
        printf("[_verify_sha256_hmac] key: 0x%s\n", out_hex);
        uint32_t msg_str_len = msg_len * 2 + 1;
        char *msg_str = (char*) malloc(msg_str_len);
        bytesToHex(msg, msg_len, msg_str, msg_str_len);
        printf("[_verify_sha256_hmac] msg: 0x%s\n", msg_str);
        free(msg_str);

        char in_sig_hex[65];
        bytesToHex(sig, 32, in_sig_hex, sizeof(in_sig_hex));
        printf("[_verify_sha256_hmac] input signature:       0x%s\n", in_sig_hex);
        char calc_sig_hex[65];
        bytesToHex(calculated_digest, 32, calc_sig_hex, sizeof(calc_sig_hex));
        printf("[_verify_sha256_hmac] calculated signature:  0x%s\n", calc_sig_hex);
    }

    *verified = !memcmp(calculated_digest, sig, 32);

    return KeyLibrary_OK;    
}


KeyLibraryErrorCode verify_root_sha256_hmac(
    const uint8_t* context_buf, 
    const uint8_t * msg, 
    uint32_t msg_len, 
    const uint8_t * sig, 
    bool * verified)
{
    if(context_buf == NULL || msg == NULL || sig == NULL || verified == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->hmac_root_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    return _verify_sha256_hmac(
                    context->hmac_root_key,
                    msg, 
                    msg_len,
                    sig,
                    verified);
}

KeyLibraryErrorCode verify_session_sha256_hmac(
    const uint8_t* context_buf, 
    const uint8_t * msg, 
    uint32_t msg_len, 
    const uint8_t * sig, 
    bool * verified)
{
    if(context_buf == NULL || msg == NULL || sig == NULL || verified == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->hmac_session_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    return _verify_sha256_hmac(
                    context->hmac_session_key,
                    msg, 
                    msg_len,
                    sig,
                    verified);
}


KeyLibraryErrorCode verify_ecdsa_signature(
    const uint8_t* context_buf, 
    const uint8_t * msg, 
    uint32_t msg_len, 
    const uint8_t * sig, 
    bool * verified)
{
    if(context_buf == NULL || msg == NULL || sig == NULL || verified == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->ecdsa_verification_key == NULL)
    {
        return KeyLibrary_KEY_NOT_FOUND;
    }

    // get the sha256 hash of the message to verify
	uint8_t msg_hash[32] = { 0 };
	mbedtls_sha256_ret(msg, msg_len, msg_hash, 0);

    if(_KDI_LOG_TRACE_)
    {
        printf("[generate_ecdsa_signature] generated sha256 hash of message\n");
        printf("msg_hash: 0x");
        for(int i=0; i < sizeof(msg_hash); i++)
        {
            printf("%02x", msg_hash[i]);
        }
        printf("\n");
    }

    *verified = ecdsa_verify(context->ecdsa_verification_key,msg_hash, sig);

    return KeyLibrary_OK;
}

/**
 * set the KTS ECDSA private key associated with the UID
 */
KeyLibraryErrorCode set_kts_ecdsa_signing_key(const uint8_t* context_buf, uint32_t slot_id, bool force)
{
    if(context_buf == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(context->uid == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }

    if(force || context->kts_ecdsa_signing_key == NULL || context->kts_slot != slot_id)
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_ecdsa_kts_signing_key] Fetching key from file...\n");
        }

        if(NULL == context->kts_ecdsa_signing_key)
        {
            context->kts_ecdsa_signing_key = malloc(AUTHENTA_ECDSA_PRIVATE_KEY_LEN);
            if(context->kts_ecdsa_signing_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_ecdsa_kts_signing_key] malloc Failure!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        uint32_t key_len = AUTHENTA_ECDSA_PRIVATE_KEY_LEN;
        char keyname[MAX_KTS_KEY_NAME_LEN];
        get_kts_ecdsa_slot_key_name(slot_id, keyname);
        uint32_t rc = get_key(context->uid, keyname, context->kts_ecdsa_signing_key, &key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_ecdsa_kts_signing_key] key not found for %s/%s\n", context->uid, KeyName_ECDSA_PRV);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(key_len != AUTHENTA_ECDSA_PRIVATE_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                char out_hex[128];
                bytesToHex(context->kts_ecdsa_signing_key, 64, out_hex, sizeof(out_hex));
                printf("[set_ecdsa_kts_signing_key] key found, but wrong length! (%s)\n", context->kts_ecdsa_signing_key);
            }
            memset(context->kts_ecdsa_signing_key, 0, key_len);
            free(context->kts_ecdsa_signing_key);
            context->kts_ecdsa_signing_key = NULL;
            return KeyLibrary_INVALID_KEY;
        }
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[AUTHENTA_ECDSA_PRIVATE_KEY_LEN*2+1];
        bytesToHex(context->kts_ecdsa_signing_key, AUTHENTA_ECDSA_PRIVATE_KEY_LEN, out_hex, sizeof(out_hex));
        printf("[set_ecdsa_kts_signing_key] HMAC Root key: 0x%s\n", out_hex);
    }

    return KeyLibrary_OK;
}

/**
 * set the KTS resident key for HMAC signing associated with the UID
 */
KeyLibraryErrorCode set_kts_hmac_signing_key(const uint8_t* context_buf, uint32_t slot_id, const uint8_t * ecdh_pub_key, bool force)
{
    if(_KDI_LOG_TRACE_) {
        printf("[set_kts_hmac_signing_key] entered\n");
    }

    KeyLibraryErrorCode rc = KeyLibrary_OK;

    if(context_buf == NULL || ecdh_pub_key == NULL)
    {
        return KeyLibrary_BAD_PARAMETER;
    }
    KeyLibraryContext context = *((KeyLibraryContext*)context_buf);

    if(force || context->kts_hmac_signing_key == NULL || slot_id != context->kts_slot
       || context->ecdh_pub_key == NULL || memcmp(context->ecdh_pub_key+1, ecdh_pub_key, AUTHENTA_ECDH_PUBLIC_KEY_LEN))
    {
        if(_KDI_LOG_TRACE_) {
            printf("[set_kts_hmac_signing_key] generating session key...\n");
        }

        if(context->ecdh_pub_key == NULL)
        {
            context->ecdh_pub_key = malloc(AUTHENTA_ECDH_PUBLIC_KEY_LEN+1);
            if(context->ecdh_pub_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_kts_hmac_signing_key] ERROR Allocating memory for the ECDH public key failed!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        context->ecdh_pub_key[0] = '\04';
        memcpy(context->ecdh_pub_key+1, ecdh_pub_key, AUTHENTA_ECDH_PUBLIC_KEY_LEN);

        /* get the ECDH private key from the key store */
        uint8_t kts_ecdh_prv_key[AUTHENTA_ECDH_PRIVATE_KEY_LEN];
        uint32_t kts_ecdh_prv_key_len=AUTHENTA_ECDH_PRIVATE_KEY_LEN;
        char keyname[MAX_KTS_KEY_NAME_LEN];
        get_kts_ecdh_slot_key_name(slot_id, keyname);
        uint32_t rc = get_key(context->uid, keyname, kts_ecdh_prv_key, &kts_ecdh_prv_key_len);
        if(rc  != 0)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_kts_hmac_signing_key] key not found for %s/%s\n", context->uid, keyname);
            }
            return KeyLibrary_KEY_NOT_FOUND;
        }
        if(kts_ecdh_prv_key_len != AUTHENTA_ECDH_PRIVATE_KEY_LEN)
        {
            if(_KDI_LOG_TRACE_) {
                printf("[set_kts_hmac_signing_key] key found, but wrong length! (%i)\n", kts_ecdh_prv_key_len);
            }
            return KeyLibrary_INVALID_KEY;
        }

        if(context->kts_hmac_signing_key == NULL)
        {
            context->kts_hmac_signing_key = (uint8_t*) malloc(AUTHENTA_HMAC_KEY_LEN);
            if(context->kts_hmac_signing_key == NULL)
            {
                if(_KDI_LOG_TRACE_) {
                    printf("[set_kts_hmac_signing_key] ERROR Allocating memory for the session key failed!\n");
                }
                return KeyLibrary_MEMORY;
            }
        }
        
        if(0 == ecdh_shared_secret(context->ecdh_pub_key, kts_ecdh_prv_key, context->kts_hmac_signing_key))
        {
            return KeyLibrary_CRYPTO;
        }
        
        context->kts_slot = slot_id;
    }

    if(_KDI_LOG_TRACE_) {
        char out_hex[128];
        bytesToHex(context->kts_hmac_signing_key, 32, out_hex, sizeof(out_hex));
        printf("[set_kts_hmac_signing_key] Session Key: %s\n", out_hex);
    }


    return rc;
}
