#ifndef _ATHENTA_KEY_DRIVER_H_
#define _ATHENTA_KEY_DRIVER_H_

/***
 * 
 * @file authenta_key_driver.h
 * 
 * @brief Key library (and crypto lib interface) header file
 * 
 * @copyright (C) 2020 Micron Technology, Inc. All rights reserved.
 * 
 **/

#include <stdint.h>
#include <stdbool.h>

#define AUTHENTA_HMAC_KEY_LEN 32
#define AUTHENTA_ECDSA_PRIVATE_KEY_LEN 32
#define AUTHENTA_ECDSA_PUBLIC_KEY_LEN 64
#define AUTHENTA_ECDH_PUBLIC_KEY_LEN 64
#define AUTHENTA_ECDH_PRIVATE_KEY_LEN 32

typedef enum {
    KeyLibrary_OK = 0,
    KeyLibrary_ERROR,
    KeyLibrary_BAD_PARAMETER,
    KeyLibrary_MEMORY,
    KeyLibrary_KEY_NOT_FOUND,
    KeyLibrary_INVALID_KEY,
    KeyLibrary_CRYPTO
} KeyLibraryErrorCode;

typedef enum {
    SRK = 1,
    CRK = 2,
    SESSION = 3,
} KeyLibraryKeyType;

typedef struct  {
    uint8_t *uid;
    uint8_t  pwr_on_amtc0[6];
    uint32_t pwr_on_amtc0_len;
    uint8_t *hmac_root_key;
    uint8_t *hmac_session_key;
    uint8_t *ecdsa_signing_key;
    uint8_t *ecdsa_verification_key;
    uint8_t *ecdh_pub_key;
    uint32_t kts_slot;
    uint8_t *kts_hmac_signing_key;
    uint8_t *kts_ecdsa_signing_key;
} 
KeyLibraryContext_struct;

typedef KeyLibraryContext_struct *KeyLibraryContext;

KeyLibraryErrorCode key_lib_open(const char * const uid, const char * const device_info_json, const uint8_t * const user_defined_data, uint8_t* contextp);
KeyLibraryErrorCode key_lib_close(uint8_t* context);

KeyLibraryErrorCode key_lib_clear_keys(const uint8_t* context_buf);


/**
 * Set the key designated in the key store as the 
 * Server Root Key as the HMAC root signing key
 **/
KeyLibraryErrorCode set_srk_as_hmac_root_signing_key(const uint8_t* context, bool force);

/**
 * Set the key designated in the key store as the 
 * Client Root Key as the HMAC root signing key
 **/
KeyLibraryErrorCode set_crk_as_hmac_root_signing_key(const uint8_t* context, bool force);

/**
 * Set the key designated in the key store as the 
 * Client Root Key as the HMAC root signing key
 **/
KeyLibraryErrorCode set_csk_as_hmac_session_signing_key(const uint8_t* context, bool force);

/**
 * Set the HMAC root key to the input key
 * 
 * This is typically used for setting the HMAC key to the CRK read from the KTR
 * 
 **/
KeyLibraryErrorCode set_hmac_root_signing_key(const uint8_t* context, const char* key, uint32_t key_len);

/**
 * Set the HMAC session key to the input key
 * 
 * This is typically used for setting the HMAC session key to the CSK read from the KTR.
 **/
KeyLibraryErrorCode set_hmac_session_key(const uint8_t* context, const char* key, uint32_t key_len);

/**
 * Genarate the session key using the HMAC root signing key and the input powerOnAMTC0
 * and designate that as the HMAC session signing key.
 * The HMAC root signing key must have been set before calling this function.
 * 
 * Note: This functions may be implemnented to use a session key previosuly
 * generated without generating a new one if force paramater = FALSE
 */
KeyLibraryErrorCode derive_hmac_session_key(const uint8_t* context, const uint8_t *powerOnAmtc0, uint32_t powerOnAmtc0_len, bool force);

/**
 * Genarate the session key using a shared seceret and the input powerOnAMTC0
 * and designate that as the HMAC session signing key.
 * The HMAC root signing key must have been set before calling this function.
 * 
 * Note: This functions may be implemnented to use a session key previosuly
 * generated without generating a new one if force paramater = FALSE
 */
KeyLibraryErrorCode derive_ecdh_session_key(const uint8_t* context_buf, const uint8_t * ecdh_pub_key, const uint8_t * powerOnAmtc0, uint32_t powerOnAmtc0_len, bool force);

/**
 * set the ECDSA private key associated with the UID
 */
KeyLibraryErrorCode set_ecdsa_signing_key(const uint8_t* context, bool force);

/**
 * set the ECDSA private public key associated with the UID
 */
KeyLibraryErrorCode set_ecdsa_verification_key(const uint8_t* context, bool force);


/**
 * set the KTS resident key for ECDSA signing associated with the UID
 */
KeyLibraryErrorCode set_kts_ecdsa_signing_key(const uint8_t* context, uint32_t slot_id, bool force);

/**
 * set the KTS resident key for HMAC signing associated with the UID
 */
KeyLibraryErrorCode set_kts_hmac_signing_key(const uint8_t* context, uint32_t slot_id, const uint8_t * ecdh_pub_key, bool force);



/**  gen_sha256_hmac
 *
 * Computes an hmac for the data in the io_buf argument
 * using the srk key associated with the input UID and
 * places the result in the same buffer. The byte-len of the
 * returned digest is returned in the io_buf_size argument.
 * 
 * @param[in] context object returned from open call. Opaque outside of the key library but internally.
 *                    Functions to set the uid and the HMAC signing key must have been called using this
 *                    context before calling this function.
 *  
 * @param[in,out] io_buf buffer initially containing the data to sign 
 *                       and will contain the HMAC signature on
 *                       successful return 
 * 
 * @param[in,out] io_buf_size should initially contain the length of the  
 *                       data to sign in io_buf, and will contain the 
 *                       length of the resulting HMAC signature on
 *                       successful return 
 *
 * @return  0 success
 *          1 failure
 */
KeyLibraryErrorCode gen_root_sha256_hmac(
    const uint8_t* context, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* digest);

/** gen_session_hmac (gen_server_session_hmac_from_uid)
 *
 * First generate the Server Session Key by getting the HMAC of
 * input AMTC0 using the Server Root Key (SRK) associated with the input UID 
 * as the key. Then use that Server Session Key as the key for
 * generating HMAC signature of the data input in i0_buf.
 * Return the signature in the same io_buf.
 *
 * The length of the input data in io_buf will initially be in
 * the contents of io_buf_size.  On successful
 * Computes an hmac for the data in the io_buf argument
 * using the crk key associated with the input UID and
 * places the result in the same buffer. The byte-len of the
 * returned digest is returned in the io_buf_size argument.
 * 
 * @param[in] context object returned from open call. Opaque outside of the key library but internally.
 *                    Functions to set the uid and the HMAC sessionsigning key must have been called using this
 *                    context before calling this function.
 * 
 * @param[in,out] io_buf buffer initially containing the data to sign 
 *                       and will contain the HMAC signature on
 *                       successful return
 *
 * @param[in,out] io_buf_size should initially contain the length of the  
 *                       data to sign in io_buf, and will contain the 
 *                       length of the resulting HMAC signature on
 *                       successful return 
 * 
 * @param[in]  amtc0    The A-MTC0 value to use in generating the session key - 
 *                      i.e., the data used when deriving the session key
 *
 * @return  0 success
 *          1 failure
 */
KeyLibraryErrorCode gen_session_sha256_hmac(
    const uint8_t* context, 
    const uint8_t * msg_buf, 
    uint32_t msg_len, 
    uint8_t* digest);


/** verify_srk_hmac_from_uid
*
* verfy the input HMAC signature was generated using SRK
* associated with the input UID.
* 
* If the input signature (sig) matches the signature derived from
* HMAC-ing the input message (msg) with the SRK associated with the
* input UID (uid), set the contents of output argument (verified)
* to TRUE, otherwise set it to FALSE.
*
* @param[in] context object returned from open call. Opaque outside of the key library but internally.
*                    Functions to set the uid and the HMAC signing key must have been called using this
*                    context before calling this function.
* @param[in]  msg      pointer to the signed data
* @param[in]  msg_size length of the signed data
* @param[in]  sig      the HMAC signature of signed data
* @param[out] verified pointer to result, if the function returns success
*                      the content will be TRUE if signature matches, FALSE if not
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
KeyLibraryErrorCode verify_root_sha256_hmac(const uint8_t* context, const uint8_t * msg, uint32_t msg_len, const uint8_t * sig, bool * verified);


/** verify_client_session_hmac_from_uid
*
* verfy the input HMAC signature was generated using a session key
* that is derived by an HMAC that uses the CRK associated with the 
* input UID as the HMAC key, and the input AMTC0 as the HMAC data.
* 
* If the input signature (sig) matches, set the contents of output 
* argument (verified) to TRUE, otherwise set it to FALSE.
*
 * @param[in] context object returned from open call. Opaque outside of the key library but internally.
 *                    Functions to set the uid and the HMAC session signing key must have been called using this
 *                    context before calling this function.
* @param[in]  msg      pointer to the signed data
* @param[in]  msg_size length of the signed data
* @param[in]  sig      the HMAC signature of signed data
* @param[in]  amtc0    The A-MTC0 value to use in generating the session key - 
                       that is the data use in deriving the session key
* @param[out] verified pointer to result, if the function returns success
*                      the content will be TRUE if signature matches, FALSE if not
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
KeyLibraryErrorCode verify_session_sha256_hmac(const uint8_t* context, const uint8_t * msg, uint32_t msg_len, const uint8_t * sig, bool * verified);


/**
* Sign an input message using the ECDSA key referenced by the input handle
*
 * @param[in] context object returned from open call. Opaque outside of the key library but internally.
 *                    Functions to set the uid and the ECDSA signing key must have been called using this
 *                    context before calling this function.
* @param[in]  msg       pointer to the data to sign
* @param[in]  msg_len   length of the data to sign
* @param[out] signature the buffer to the ecdsa signature
*                       buffer MUST be at least 64 bytes (twice as long a the AUTHENTA_ECDSA_PRIVATE_KEY_LEN)
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
KeyLibraryErrorCode generate_ecdsa_signature(const uint8_t* context, const uint8_t * msg, uint32_t msg_len, uint8_t * signature);

/**
* verify an ecdsa signature
*
 * @param[in] context object returned from open call. Opaque outside of the key library but internally.
 *                    Functions to set the uid and the ECDSA verification key must have been called using this
 *                    context before calling this function.
* @param[in]  msg      pointer to the signed data
* @param[in]  msg_size length of the signed data
* @param[in]  sig      the ecdsa signature of signed data
* @param[out] verified pointer to result, if the function returns success
*                      the content will be TRUE if signature matches, FALSE if not
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
KeyLibraryErrorCode verify_ecdsa_signature(const uint8_t* context, const uint8_t * msg, uint32_t msg_lem, const uint8_t * sig, bool * verified);


#endif /* _ATHENTA_KEY_DRIVER_H_ */