/***
 * 
 * @file authenta_key_management_utils.c
 * 
 * @brief Authenta key library key management functions implementation
 * 
 * @copyright (C) 2020 Micron Technology, Inc. All rights reserved.
 * 
 **/

/******
 * 
 * In this implementation keys are stored in an XML formatted file (in the clear).
 * 
 * An example of the expected XML structure is below:
 *  
 *   <key_content>
 *     <uid>1032547698badcfe1f32547698baaa55</uid>
 *     <key_type>hmac</key_type>
 *     <key_name>srk</key_name>
 *     <key>d5f55d8a267ad5daa3e93e30d804bb70993aa705c908a82d850e7e495b01566a</key>  
 *     <handle>1138</handle>
 *   </key_content>
 * 
 *   <key_content>
 *     <uid>a89bc41465ae78813497b51d1f2b32dc</uid>
 *     <key_type>ecdh</key_type>
 *     <key_name>ecdh_sr_prk</key_name>
 *     <key_pub_x>2b2c0aeb03441c8880a8cc3cb715c0f6014d3022b76812c6ac37d6b57ad057c6</key_pub_x> 
 *     <key_pub_y>1c287b9342da3c10d1a38694c1bd4351e0a5f0f55979af97d98070d85e9b6ace</key_pub_y>
 *     <key_prv>bf5962409acc3c370d52f9d173c7f0ddab637bc373e69462a3f438c098ca5570</key_prv>
 *     <handle>1148</handle>
 *   </key_content>
 * 
 ******/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h> 

#include <libxml/parser.h>
#include <libxml/tree.h>

// #include "easy_ecc/ecc.h"
// #include "hmac_sha/hmac_sha2.h"

#include "authenta_key_management_utils.h"

#include "assert.h"

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER; 

static xmlChar *getRawKeyFromNode(xmlNode * rootNode);
static xmlNode *find_node_from_input_root(
          xmlNode * rootNode, 
    const xmlChar * uid, 
    const xmlChar * handle, 
    const xmlChar * key_name, 
    const xmlChar * key);
/**
* Get the handle value for new key
*/
static uint32_t get_new_key_handle(char * handle_out_buff, const uint32_t handle_out_buff_size);



static xmlNodePtr tmpTree_rootNode =  NULL;
static xmlNodePtr getTmpTreeRootNode()
{
    if(tmpTree_rootNode == NULL)
    {
        tmpTree_rootNode =  xmlNewNode(NULL, BAD_CAST "tmp_tree_root");
    }

    return tmpTree_rootNode;
}

static void clearTmpTreeRootNode()
{
    if(tmpTree_rootNode != NULL)
    {
        xmlFreeNode(tmpTree_rootNode);
        tmpTree_rootNode =  xmlNewNode(NULL, BAD_CAST "tmp_tree_root");
    }
}

static void char2bytes(xmlChar ** inArray, int inArraySize){
    if(!inArray || inArraySize <= 0){
        return;
    }
    int i;
    for(i=0; i<inArraySize; i++){
        if((*inArray)[i] >= '0' && (*inArray)[i] <= '9')
            (*inArray)[i] -= '0';
        else if((*inArray)[i] >= 'a' && (*inArray)[i] <= 'f'){
            (*inArray)[i] -= 'a';
            (*inArray)[i] += 10;
        }
        else if((*inArray)[i] >= 'A' && (*inArray)[i] <= 'F'){
            (*inArray)[i] -= 'A';
            (*inArray)[i] += 10;
        }

        if(i%2 == 1)
            (*inArray)[i/2] = ((*inArray)[i-1] << 4) + (*inArray)[i];
    }
    /*printf("\n");
    for(i=0; i<inArraySize/2; i++){
        printf("0x%x ",(*inArray)[i]);
    }
    printf("\n");
    */
}

static void bytesToHex(const uint8_t* bytes_arr, uint32_t bytes_arr_len, char* hex_arr, uint32_t hex_arr_len)
{
    assert(bytes_arr != NULL && 
           hex_arr != NULL && 
           hex_arr_len >= (2*bytes_arr_len));

    for(int i = 0; i < bytes_arr_len; i++)
    {
        sprintf(&hex_arr[i*2], "%02X", bytes_arr[i]);
    }
}


/** \fn uint8_t CharsToNumber( char ch1, char ch2 )
* \brief Convert 2 hex chars in a byte number
* \param ch1 High Nibble as char in hex
* \param ch2 Low Nibble as char in hex
* \return converted value
*/
static uint8_t CharsToNumber( char ch1, char ch2 )
{
    uint8_t val = 0;
    if ( ch1 >= 'A' && ch1 <= 'F' ) 
    {
        val = (ch1 - 'A') + 10;
    }
    else if ( ch1 >= 'a' && ch1 <= 'f' ) 
    {
        val = (ch1 - 'a') + 10;
    }
    else if ( ch1 >= '0' && ch1 <= '9' ) 
    {
        val = (ch1 - '0') ;
    }
    val <<= 4;
    if ( ch2 >= 'A' && ch2 <= 'F' ) 
    {
        val += (ch2 - 'A') + 10;
    }
    else if ( ch2 >= 'a' && ch2 <= 'f' ) 
    {
        val += (ch2 - 'a') + 10;
    }
    else if ( ch2 >= '0' && ch2 <= '9' ) 
    {
        val += (ch2 - '0') ;
    }
    return val;
}

/** \fn getPublicKeyFromString
* \brief Convert a 128 hex chars string in a public key 
* \param[out] key Output key (64 bytes)
* \param[in] string input 128 byte string vector
* \return 0 or error code
*/
uint32_t GetPublicKeyFromString( uint8_t* key, const char* string )
{
    if(_KDI_LOG_TRACE_)
    {
        printf("GetPublicKeyFromString(0x%lx, '%s')\n", (unsigned long) key, string);
    }
    uint8_t index;
    uint8_t val;
    bool string_completed = false;
    const char* current = string;
    for ( index =0; index < 64; index++ )
    {
        if ( string_completed == true )
        {
            val = 0;
        }
        else
        {
            char ch1 = *current;
            char ch2 = *(current +1 );
            if ( ch1 == 0 )
            {
                string_completed = true;
                ch2 = 0; 
            }
            if ( ch2 == 0 )
            {
                string_completed = true;
            }
            val = CharsToNumber( ch1, ch2 );
            current += 2;
        }
        key[ index ] = val; 
    }

    return 0;
}



/** \fn getPrivateKeyFromString
* \brief Convert a 64 hex chars string in a private key 
* \param key Output key (32 bytes)
* \param string input 64 byte string vector
* \return 0 or error code
*/
uint32_t GetPrivateKeyFromString( uint8_t* key, const char* string )
{
    if(_KDI_LOG_TRACE_)
    {
        printf("GetPrivateKeyFromString(0x%lx, '%s')\n", (unsigned long) key, string);
    }
    uint8_t index;
    uint8_t val;
    bool string_completed = false;
    const char* current = string;
    for ( index =0; index < 32; index++ )
    {
        if ( string_completed == true )
        {
            val = 0;
        }
        else
        {
            char ch1 = *current;
            char ch2 = *(current +1 );
            if ( ch1 == 0 )
            {
                string_completed = true;
                ch2 = 0; 
            }
            if ( ch2 == 0 )
            {
                string_completed = true;
            }
            val = CharsToNumber( ch1, ch2 );
            current += 2;
        }
        key[ index ] = val; 
    }

    return 0;
}


static xmlChar *getRawKeyFromNode(xmlNode * rootNode)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("getRawKeyFromNode(0x%lx)\n", (unsigned long) rootNode);
    }
    xmlNode * child_node = rootNode->children;
    while(child_node){
        if(!xmlStrcmp(child_node->name, XML_NodeName_KEY)){
            if(NULL != child_node->children) {
                return child_node->children->content; // remove 0x, clean up later
            }
        }
        child_node = child_node->next;
    }
    return NULL;
}

static xmlNode *
find_node_from_input_root(xmlNode * rootNode, const xmlChar * uid, const xmlChar * handle, const xmlChar * key_name, const xmlChar * key)
{
    if(rootNode == NULL){
        return NULL;
    }
    if(_KDI_LOG_TRACE_)
    {
        printf("find_node_from_input_root(0x%lx, '%s', '%s', '%s', '%s')\n", (unsigned long) rootNode, uid, handle, key_name, key);
    }
/*    if(handle)
        printf("find_node called with handle %s\n", handle);
    if(uid)
        printf("find_node called with uid %s\n", uid);
*/
    xmlNode * cur_node = NULL;
    xmlNode * child_node;
    xmlNode * res_node;

    for (cur_node = rootNode; cur_node; cur_node = cur_node->next) {
        if (cur_node->type == XML_ELEMENT_NODE && !xmlStrcmp(cur_node->name, XML_NodeName_Key_Content)) {
            child_node = cur_node->children;
            while(child_node){
                if(uid    && !xmlStrcmp(child_node->name, XML_NodeName_UID) && !xmlStrcmp(child_node->children->content, uid)){
                    if (key_name == NULL) {
                        return cur_node;
                    }
                    else {
                        // check key_name
                        child_node = child_node->next;
                        while (child_node) {
                            if(!xmlStrcmp(child_node->name, XML_NodeName_key_Name) && !xmlStrcmp(child_node->children->content, key_name)){
                                if(_KDI_LOG_TRACE_)
                                {
                                    printf("[find_node_from_input_root] Node found\n");
                                }
                                return cur_node;
                            }
                            child_node = child_node->next;
                        }
                        // finish iterating all child nodes
                        break;
                    }
                }
                else if(handle && !xmlStrcmp(child_node->name, XML_NodeName_Handle) && !xmlStrcmp(child_node->children->content, handle)){
                    return cur_node;
                }
                else if(key    && !xmlStrcmp(child_node->name, XML_NodeName_KEY) && !xmlStrcmp(child_node->children->content, key)){
                    return cur_node;
                }
                child_node = child_node->next;
            }
        }
        else {
            res_node = find_node_from_input_root(cur_node->children, uid, handle, key_name, key);
            if(res_node){
                return res_node;
            }
        }
    }
    return NULL;
}

static xmlNode *
find_node(xmlChar * uid, xmlChar * handle, xmlChar * key_name, xmlChar * key)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("find_node('%s', '%s', '%s', '%s')\n", uid, handle, key_name, key);
    }
    assert(uid != NULL || handle != NULL || key != NULL);

    // look in the temporary ram based key tree
    xmlNodePtr root_node = getTmpTreeRootNode();
    xmlNode * node = find_node_from_input_root(root_node, uid, handle, key_name, key);

    if(node == NULL)
    {
        /* handle wasn't found in the volatile key tree, check the non-volatile (file-based) tree */
        const char * key_xml_files[2] = { KEY_XML_FILE_PRV, KEY_XML_FILE_PUB };

        // check in both key files
        for (int i = 0; i < 2; i++)
        {
            xmlDoc *doc = doc = xmlReadFile(key_xml_files[i], NULL, 0);
            if (doc == NULL) 
            {
                printf("error: could not parse file %s\n", key_xml_files[i]);
                continue;
            }

            /*Get the root element node */
            root_node = xmlDocGetRootElement(doc);

            node = find_node_from_input_root(root_node, uid, handle, key_name, key);
            if(node != NULL)
            {
                break;
            }
            
            xmlFreeDoc(doc);
            xmlCleanupParser();
        }
    }

    return node;
}

/**
 * Internal function used by get_key() to convert hex string to bytes.
 * 
 * param[in]  keyContent a string of hex characters representing a key.
 *                       Typically a private key will have 64 hex charackters (32 bytes)
 *                       and a public key will have 128 hex characters (64 bytes - x|y)
 * 
 * param[out] raw_key  the raw byte version of the input version (keyContent)
 * 
 * param[in]  raw_key_buf_len the buffer length of raw_key. If the buffer is not large enough
 *                            to hold the raw key, half the length of keyContent, return fail.
 * 
 * return 0 on success, 1 on fail
 */
uint32_t _convert_hex_char_key_to_bytes(const char *keyContent, uint8_t * raw_key, uint32_t raw_key_buf_len)
{
    uint32_t rc = 1;

    uint32_t key_len_hex = strlen((const char*)keyContent);
    // if(_KDI_LOG_TRACE_)
    // {
    //     printf("converting hex char key with length %d to bytes\n",strlen(keyContent));
    // }

    uint32_t key_len_bytes = (uint32_t) ((key_len_hex+1) / 2);
    if(key_len_bytes > raw_key_buf_len)
    {
        if(_KDI_LOG_TRACE_)
        {
            printf("[_convert_hex_char_key_to_bytes] WARN raw_key buffer of size %d not large enough to fit key: 0x%s\n", 
            raw_key_buf_len, keyContent);
        }

        return 1;
    }

    if(key_len_bytes <= raw_key_buf_len)
    {
        if(key_len_bytes == 32)
        {
            rc = GetPrivateKeyFromString(raw_key, (const char *)keyContent);
            if(_KDI_LOG_TRACE_)
            {
                char key_str[64];
                bytesToHex(raw_key, key_len_bytes, key_str, sizeof(key_str));
                printf("raw priv signing key: 0x%s\n", key_str);
            }
        }
        else if (key_len_bytes == 64)
        {
            rc = GetPublicKeyFromString(raw_key, (const char *)keyContent);
            if(_KDI_LOG_TRACE_)
            {
                char key_str[128];
                bytesToHex(raw_key, key_len_bytes, key_str, sizeof(key_str));
                printf("raw public key: 0x%s\n", key_str);
            }
        }
        else
        {
            // using a temp key buffer since the func char2bytes copies in-place
            char tmp_key[key_len_hex];
            memcpy(tmp_key, keyContent, key_len_hex);
            char2bytes((xmlChar ** )&tmp_key, key_len_hex);
            memcpy(raw_key, tmp_key, key_len_bytes);
            rc = 0;
            if(_KDI_LOG_TRACE_)
            {
                char key_str[128];
                bytesToHex(raw_key, key_len_bytes, key_str, sizeof(key_str));
                printf("raw key of len %d: 0x%s\n", key_len_bytes, key_str);
            }
        }
    }

    return rc;      
}


/**
* Get the handle value for new key
*/
static uint32_t get_new_key_handle(char * handle_out_buff, const uint32_t handle_out_buff_size)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("get_new_key_handle(0x%lx, %d)\n", (unsigned long) handle_out_buff, handle_out_buff_size);
    }
    assert(handle_out_buff != NULL);

    // iterate all nodes to find the maximum key handle value
    // new key handle value will be maximum value + 1
    uint32_t max_handle_value = 0;

    const char * key_xml_files[2] = { KEY_XML_FILE_PUB, KEY_XML_FILE_PRV };

    for (int i = 0; i < 2; i++)
    {
        xmlDoc *doc = NULL;
        doc = xmlReadFile(key_xml_files[i], NULL, 0);
        if (doc != NULL)
        {
            /*Get the root element node */
            xmlNode * root_element = xmlDocGetRootElement(doc);

            // iterate through the file based keys
            for (xmlNode * cur_node = root_element->children; cur_node != NULL; cur_node = cur_node->next)
            {
                if (cur_node->type == XML_ELEMENT_NODE && !xmlStrcmp(cur_node->name, (const xmlChar *) XML_NodeName_Key_Content))
                {
                    for (xmlNode * child_node = cur_node->children; child_node != NULL; child_node = child_node->next)
                    {
                        if(!xmlStrcmp(child_node->name, XML_NodeName_Handle))
                        {
                            uint32_t handle_value = atoi((const char *) child_node->children->content);

                            if (handle_value > max_handle_value)
                            {
                                max_handle_value = handle_value;
                            }
                        }
                    }
                }
            }

            xmlFreeDoc(doc);
            xmlCleanupParser();
        }
    }

    // iterate through the temporary ram based keys as well
    xmlNodePtr temp_root_node = getTmpTreeRootNode();
    for (xmlNode * cur_node = temp_root_node->children; cur_node != NULL; cur_node = cur_node->next)
    {
        if (cur_node->type == XML_ELEMENT_NODE && !xmlStrcmp(cur_node->name, (const xmlChar *) XML_NodeName_Key_Content))
        {
            for (xmlNode * child_node = cur_node->children; child_node != NULL; child_node = child_node->next)
            {
                if(!xmlStrcmp(child_node->name, XML_NodeName_Handle))
                {
                    uint32_t handle_value = atoi((const char *) child_node->children->content);

                    if (handle_value > max_handle_value)
                    {
                        max_handle_value = handle_value;
                    }
                }
            }
        }
    }

    char tmp_handle[AUTHENTA_KEY_DRIVER_MAX_HANDLE_LEN] = { 0 };
    sprintf(tmp_handle, "%i", max_handle_value + 1);

    if (strlen(tmp_handle) + 1 > handle_out_buff_size)
    {
        printf("error: size of output buffer for handle is not large enough\n");
        return 1;
    }

    memset(handle_out_buff, 0, handle_out_buff_size);
    strcpy(handle_out_buff, tmp_handle);

    return 0;
}



/**
* Get handle for a key based on input UID and key name
*
* @param[in]  uid               UID of the key to return
*                               May NOT be NULL
* @param[in]  key_name          which key associated with input UID to return (srk, ecdh_sr_prk, session)
*                               May NOT be NULL
* @param[out] metadata_out_buff output buffer to place a NULL terinated char string holding metadata associated with the key
*                               NULL indicates not to return metadata
* @param[out] metadata_out_buff_size length of metadata buffer. Must be large enough to hold NULL terminated metadata.
*
* @param[out] handle_out_buff   output buffer to hold the new key handle;
*                               May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
* 
* @return Success 0
*         Fail    1  (input uid, key_name, or handle_buffer is null, key file not found, uid not found) 
*/
uint32_t get_key_handle(
    const char * uid,
    const char * key_name,
    char * metadata_out_buff,
    unsigned int metadata_out_buff_size,
    char * handle_out_buff,
    unsigned int handle_out_buff_size)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("get_key_handle('%s', '%s', 0x%lx, %d, 0x%lx, %d')\n", uid, key_name,
        (unsigned long) metadata_out_buff, metadata_out_buff_size, (unsigned long) handle_out_buff, handle_out_buff_size);
    }
    uint32_t status = 1;
    xmlNode *root_element = NULL;

    if (uid == NULL || handle_out_buff == NULL)
    {
        return status;
    }

    const char * key_xml_files[2] = { KEY_XML_FILE_PRV, KEY_XML_FILE_PUB };

    // check in both key files
    for (int i = 0; i < 2; i++)
    {
        xmlDoc *doc = doc = xmlReadFile(key_xml_files[i], NULL, 0);

        if (doc == NULL) 
        {
            printf("[get_key_handle] error: could not parse file %s\n", key_xml_files[i]);
            continue;
        }

        /*Get the root element node */
        root_element = xmlDocGetRootElement(doc);

        xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
        if(node != NULL)
        {
            xmlNode * child_node = node->children;

            while (child_node)
            {
                if(!xmlStrcmp(child_node->name, XML_NodeName_Handle))
                {
                    const char * handle_str = (const char *) child_node->children->content;
                    if(_KDI_LOG_TRACE_)
                    {
                        printf("[get_key_handle] found handle: %s\n", handle_str);
                    }

                    if (handle_out_buff_size < strlen(handle_str) + 1)
                    {
                        printf("[get_key_handle] error: Handle output buffer size is not large enough to hold the handle string\n");
                    }
                    else
                    {
                        status = 0;
                        memset(handle_out_buff, 0, handle_out_buff_size);
                        strcpy(handle_out_buff, handle_str);
                    }
                }

                if(metadata_out_buff != NULL && !xmlStrcmp(child_node->name, XML_NodeName_metadata))
                {
                    const char * metadata_str = (const char *) child_node->children->content;
                    if(_KDI_LOG_TRACE_)
                    {
                        printf("[get_key_handle] found metadata: %s\n", metadata_str);
                    }

                    if (metadata_out_buff_size < strlen(metadata_str) + 1)
                    {
                        printf("[get_key_handle] error: Metadata output buffer size is not large enough to hold the metadata string\n");
                    }
                    else
                    {
                        status = 0;
                        memset(metadata_out_buff, 0, metadata_out_buff_size);
                        strcpy(metadata_out_buff, metadata_str);
                        break;
                    }
                }
                
                child_node = child_node->next;
            }  
        }

        xmlFreeDoc(doc);
        xmlCleanupParser();
    }

    if(status)
    {
        printf("error: can't find key\n");
    }
    
    return status;
}


/**
* Get the (raw) key based on input UID and key name
*
* @param[in]  uid               UID of the key to return
*                               May NOT be NULL
* @param[in]  key_name          which key associated with input UID to return (srk, ecdh_sr_prk, session)
*                               May NOT be NULL
*
* @param[out] raw_key bytes containing the raw key; MUST NOT be NULL; 
*                     Call will fail if not large enough to hold the raw key
*                     May NOT be NULL.
* @param[in,out] raw_key_buf_len pointer to len of input buffer, 
*                                on succesful return will hold the length
*                                of the returned key
*                                May NOT be NULL.
* 
* @return Success 0
*         Fail    1  (input uid, key_name, or handle_buffer is null, key file not found, uid not found) 
*/
uint32_t get_key(
    const unsigned char * uid,
    const char * key_name,
    uint8_t * key, 
    uint32_t *key_buf_len_p)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("get_key('%s', '%s', 0x%lx, %d')\n", uid, key_name,
                (unsigned long) key, *key_buf_len_p);
    }
    uint32_t status = 1;
    xmlNode *root_element = NULL;

    if (uid == NULL || key == NULL)
    {
        return status;
    }

    const char * key_xml_files[2] = { KEY_XML_FILE_PRV, KEY_XML_FILE_PUB };

    // check in both key files
    for (int i = 0; i < 2; i++)
    {
        pthread_mutex_lock(&lock); 
        xmlDoc *doc = doc = xmlReadFile(key_xml_files[i], NULL, 0);
        pthread_mutex_unlock(&lock);   

        if(_KDI_LOG_TRACE_)
        {
            printf("[get_key] Looking for key in file %s\n", key_xml_files[i]);
        }

        if (doc == NULL) 
        {
            printf("[get_key] error: could not parse file %s\n", key_xml_files[i]);
            continue;
        }

        /*Get the root element node */
        root_element = xmlDocGetRootElement(doc);

        xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
        if(node != NULL)
        {
            const xmlChar * keyContent = getRawKeyFromNode(node);
            if(keyContent != NULL && strlen((const char*)keyContent) > 0)
            {
                if(_KDI_LOG_TRACE_)
                {
                    printf("[get_key] found key for uid/key-name %s/%s: %s in key tree\n", uid, key_name, keyContent);
                }
                uint32_t key_len = (uint32_t) (strlen((const char*)keyContent)+1)/2;
                if(key_len <= *key_buf_len_p)
                {
                    status = _convert_hex_char_key_to_bytes((const char*)keyContent, key, *key_buf_len_p);
                }
                else if(_KDI_LOG_TRACE_)
                {
                    printf("[get_key] key output buffer of size %d not large enough to hold key of size %d for uid %s/key-name %s\n", *key_buf_len_p, key_len, uid, key_name);
                }
                *key_buf_len_p = key_len;
                break;
            }
            else {
                printf("[get_key] Failed to extract raw key for uid/key-name %s/%s\n", uid, key_name);
            }
        }
    
        xmlFreeDoc(doc);
        xmlCleanupParser();
    }

    if (status)
    {
        printf("[get_key] error: can't find key\n");
    }

    return status;
}

/**
* Add a new temprary key (handle, key) and generate handle for the new key
* 
* @param[in]  key_hex          pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[out] handle_out_buff  output buffer to hold the new key handle;
*                              May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
*
* @return 0 on success
*         1 on failure
*/
uint32_t add_volatile_key(
    const char * key_hex,
    char * handle_out_buff,
    unsigned int handle_out_buff_size)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("add_volatile_key('%s', 0x%lx, %d)\n", key_hex, (unsigned long) handle_out_buff, handle_out_buff_size);
    }
    assert(key_hex != NULL && handle_out_buff != NULL);
    
    xmlNode *root_element = NULL;
    uint32_t status = 1;
    if (key_hex == NULL || handle_out_buff == NULL)
    {
        return status;
    }

    // if (strlen(key_hex) != 64)
    // {
    //     printf("[add_volatile_key] error: size of input parameter (e.g., uid or key) is not valid\n");
    //     return status;
    // }

    /*Get the root element node */
    root_element = getTmpTreeRootNode();
    if(root_element == NULL)
    {
        printf("[add_volatile_key] Failed to retrieve Temporary Key Tree root node!\n");
        return status;
    }

    // generate new key handle
    if (get_new_key_handle(handle_out_buff, handle_out_buff_size))
    {
        printf("[add_volatile_key] error: could not get new key handle\n");
    }
    else
    {
        // create a new node
        xmlNode * new_node = xmlNewNode(NULL, (xmlChar *) XML_NodeName_Key_Content);
        xmlNewChild(new_node, NULL, (xmlChar *) XML_NodeName_Handle, (xmlChar *) handle_out_buff);
        xmlNewChild(new_node, NULL, (xmlChar *) XML_NodeName_KEY, (xmlChar *) key_hex);

        // add new node to root
        xmlAddChild(root_element, new_node);
        status = 0;
    }

    return status;
}

/**
* Add a new hmac key (uid, key_name, key) to the key file and generate handle for the new key
*
* @param[in]  uid              pointer to hex ascii representation of the UID to associate the key with
*                              May NOT be NULL
* @param[in]  key_name         pointer to the type of key, e.g. "crk" or "session"
*                              May NOT be NULL
* @param[in]  key              pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[in]  metadata         NULL terinated char string holding metadata to associate with the key
*                              NULL or empty string indicates no metadata to be stored
* @param[in]  is_private_key   if true the input is stored as a private key, 
*                              if false it is stored as a public key
* @param[in]  allow_replace    if true an existing key with a matching UID and name will be replaced
*                              if false it if a key with a matching UID and name exists it will not
*                              be changed and this function will fail
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_nonvolatile_based_key(
    const char * uid,
    const char * key_name,
    const char * key,
    const char * metadata,
    const bool is_private_key,
    const bool allow_replace)
{
    uint32_t status = 1;
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    const char * key_xml_file = NULL;

    if (uid == NULL || key_name == NULL || key == NULL)
    {
        return status;
    }

    if (strlen(uid) != 32 || strlen(key) != 64)
    {
        printf("[add_nonvolatile_based_key] error: size of input parameter (e.g., uid or key) is not valid\n");
        return status;
    }

    if (is_private_key)
    {
        key_xml_file = KEY_XML_FILE_PRV;
    }
    else
    {
        key_xml_file = KEY_XML_FILE_PUB;
    }

    pthread_mutex_lock(&lock);

    doc = xmlReadFile(key_xml_file, NULL, 0);

    if (doc == NULL)
    {
        pthread_mutex_unlock(&lock);
        printf("[add_nonvolatile_based_key] error: could not parse file %s\n", key_xml_file);
        return status;
    }

    /*Get the root element node */
    root_element = xmlDocGetRootElement(doc);
    if (root_element == NULL)
    {
        pthread_mutex_unlock(&lock);
        printf("[add_nonvolatile_based_key] error: could retrieve root element after parsing file %s\n", key_xml_file);
        return status;
    }


    xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
    if(node == NULL)
    {
        char handle[128]; // somewhat arbitrary 128 byte max handle length

        // generate new key handle
        if (get_new_key_handle(handle, sizeof(handle)))
        {
            printf("[add_nonvolatile_based_key] error: could not get new key handle\n");
        }
        else
        {
            // create a new node
            xmlNode * new_node = xmlNewNode(NULL, xmlStrdup((xmlChar *) XML_NodeName_Key_Content));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_UID), xmlStrdup((xmlChar *) uid));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_key_Name), xmlStrdup((xmlChar *) key_name));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_Handle), xmlStrdup((xmlChar *) handle));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_KEY), xmlStrdup((xmlChar *) key));
            if(metadata != NULL && strlen(metadata) > 0)
            {
                xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_metadata), xmlStrdup((xmlChar *) metadata));
            }

            // add new node to root
            xmlAddChild(root_element, new_node);
            status = 0;
        }
    }
    else if(allow_replace)
    {
        bool key_set = FALSE;
        bool metadata_set = false;
        xmlNode * childIterator = node->children;
        while(childIterator){
            if(!xmlStrcmp(childIterator->name, (const unsigned char *) XML_NodeName_KEY))
            {
                xmlNodeSetContent(childIterator->children, (const unsigned char *) key);
                key_set = TRUE;
            }
            else if(metadata != NULL && !xmlStrcmp(childIterator->name, (const unsigned char *) XML_NodeName_metadata))
            {
                xmlNodeSetContent(childIterator->children, (const unsigned char *) metadata);
                metadata_set = TRUE;
            }
            childIterator = childIterator->next;
        }
        
        if(!key_set)
        {
            xmlNewChild(node, NULL, xmlStrdup((xmlChar *) XML_NodeName_KEY), xmlStrdup((xmlChar *) key));
        }

        if(metadata && !metadata_set && strlen(metadata) > 0)
        {
            xmlNewChild(node, NULL, xmlStrdup((xmlChar *) XML_NodeName_metadata), xmlStrdup((xmlChar *) metadata));
        }
    }
    else
    {
        printf("error: Cannot add key for uid 0x%s and key name %s! Key with same uid and key_name already exists\n", uid, key_name);
    }
    
    xmlSaveFile(key_xml_file, doc);

    pthread_mutex_unlock(&lock);

    xmlFreeDoc(doc);
    xmlCleanupParser();

    // sleep to give OS time so next time the 
    // file is read the data is the latest
    sleep(5);

    return status;
}


/**
* Add a new temporary hmac key (uid, key_name, key) to the memory based volatile key tree
* and generate handle for the new key.
* If a key associated with the same UID and key_name already exists remove it
* and add then add the key.
*
* @param[in]  uid              pointer to hex ascii representation of the UID to associate the key with
*                              May NOT be NULL
* @param[in]  key_name         pointer to the type of key, e.g. "crk" or "session"
*                              May NOT be NULL
* @param[in]  key              pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[in]  metadata         NULL terinated char string holding metadata to associate with the key
*                              NULL or empty string indicates no metadata to be stored
* @param[out] handle_out_buff  output buffer to hold the new key handle;
*                              May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_volatile_based_key(
    const char * uid,
    const char * key_name,
    const char * key,
    const char * metadata,
    char * handle_out_buff,
    unsigned int handle_out_buff_size)
{
    uint32_t status = 1;
    xmlNode *root_element = NULL;

    if (uid == NULL || key_name == NULL || key == NULL || handle_out_buff == NULL)
    {
        return status;
    }

    if (strlen(uid) != 32 || strlen(key) != 64)
    {
        printf("[add_nonvolatile_based_key] error: size of input parameter (e.g., uid or key) is not valid\n");
        return status;
    }

    /*Get the root element node */
    root_element = getTmpTreeRootNode();
    if (root_element == NULL)
    {
        printf("[add_nonvolatile_based_key] error: could retrieve root element\n");
        return status;
    }

    xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
    if(node != NULL)
    {
        delete_key_by_key_type(uid, key_name);
        node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
    }

    if(node == NULL)
    {
        // generate new key handle
        if (get_new_key_handle(handle_out_buff, handle_out_buff_size))
        {
            printf("[add_nonvolatile_based_key] error: could not get new key handle\n");
        }
        else
        {
            // create a new node
            xmlNode * new_node = xmlNewNode(NULL, xmlStrdup((xmlChar *) XML_NodeName_Key_Content));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_UID), xmlStrdup((xmlChar *) uid));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_key_Name), xmlStrdup((xmlChar *) key_name));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_Handle), xmlStrdup((xmlChar *) handle_out_buff));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_KEY), xmlStrdup((xmlChar *) key));
            if(metadata != NULL && strlen(metadata) > 0)
            {
                xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_metadata), xmlStrdup((xmlChar *) metadata));
            }

            // add new node to root
            xmlAddChild(root_element, new_node);
            status = 0;
        }
    }
    else
    {
        printf("error: Cannot add key for uid 0x%s and key name %s! Key with same uid and key_name already exists\n", uid, key_name);
    }
    
    xmlCleanupParser();

    return status;
}


/**
* Delete key by handle
*
* First the volatile key tree is check, then the 
* non-volatile (file-based) tree is searched. (Only 
* (one key is deleted).
*
* @param[in] handle  handle of key to be deleted
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
uint32_t delete_key_by_handle(const char * handle)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("delete_key_by_handle('%s')\n", handle);
    }
    uint32_t status = 1;

    assert(handle != NULL);

    if (handle == NULL)
    {
        return status;
    }

    /* first check the temp keys */
    xmlNode * node = find_node_from_input_root(getTmpTreeRootNode(), NULL, (xmlChar*) handle, NULL, NULL);

    if(node == NULL)
    {
        /* handle wasn't found in the volatile key tree, check the non-volatile (file-based) tree */
        
        const char * key_xml_files[2] = { KEY_XML_FILE_PUB, KEY_XML_FILE_PRV };

        for (int i = 0; i < 2; i++)
        {
            pthread_mutex_lock(&lock);
            xmlDoc *doc = doc = xmlReadFile(key_xml_files[i], NULL, 0);

            if (doc == NULL) 
            {
                printf("error: could not parse file %s\n", key_xml_files[i]);
                continue;
            }

            /*Get the root element node */
            xmlNode * root_element = xmlDocGetRootElement(doc);
            if (root_element == NULL) 
            {
                printf("warn: missing data in key file %s\n", key_xml_files[i]);
            }
            else
            {
                node = find_node_from_input_root(root_element, NULL, (xmlChar*) handle, NULL, NULL);
            }

            if(node != NULL)
            {
                // delete node
                xmlUnlinkNode(node);
                xmlFreeNode(node);                
                xmlSaveFile(key_xml_files[i], doc);

                pthread_mutex_unlock(&lock);

                xmlFreeDoc(doc);
                status = 0;
                break;
            }

            pthread_mutex_unlock(&lock);
            xmlFreeDoc(doc);        
            xmlCleanupParser();
        }

        if(node == NULL)
        {
            printf("error: can't find key with handle %s\n", handle);
        }   
    }
    else
    {
        // delete node
        xmlUnlinkNode(node);
        xmlFreeNode(node);
        status = 0;
    }

    return status;
}



/**
* Delete key by key tpe and UID
*
* First the volatile key tree is checked, then the 
* non-volatile (file-based) tree is searched. (Only 
* (one key is deleted).
*
* @param[in] uid  the uid associated with the key
* @param[in] key_name  the name of the key
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
uint32_t delete_key_by_key_type(
    const char * uid,
    const char * key_name)
{
    if(_KDI_LOG_TRACE_)
    {
        printf("delete_key_by_key_type('%s', '%s')\n", uid, key_name);
    }
    uint32_t status = 1;

    assert(uid != NULL &&  key_name != NULL);

    if (uid == NULL || key_name == NULL)
    {
        return status;
    }


    /*Get the root element node */
    xmlNode * node = find_node_from_input_root(getTmpTreeRootNode(), (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
    if(node == NULL)
    {
        /* handle wasn't found in the volatile key tree, check the non-volatile (file-based) tree */     

        const char * key_xml_files[2] = { KEY_XML_FILE_PRV, KEY_XML_FILE_PUB };

        // check in both key files
        for (int i = 0; i < 2; i++)
        {
            pthread_mutex_lock(&lock);
            xmlDoc *doc = doc = xmlReadFile(key_xml_files[i], NULL, 0);

            if (doc == NULL) 
            {
                printf("[delete_key_by_key_type] error: could not parse file %s\n", key_xml_files[i]);
                pthread_mutex_unlock(&lock);
                continue;
            }

            /*Get the root element node */
            xmlNode * root_element = xmlDocGetRootElement(doc);
            if (root_element == NULL) 
            {
                printf("[delete_key_by_key_type] warn: missing data in key file %s\n", key_xml_files[i]);
            }
            else
            {
                node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
            }

            if(node != NULL)
            {
                // delete node
                xmlUnlinkNode(node);
                xmlFreeNode(node);
                xmlSaveFile(key_xml_files[i], doc);

                pthread_mutex_unlock(&lock);
                
                status = 0;

                if(_KDI_LOG_TRACE_)
                {
                    printf("[delete_key_by_key_type] Deleted %s key with uid %s\n", key_name, uid);
                }

            }
            else
            {
                pthread_mutex_unlock(&lock);
            }

            xmlFreeDoc(doc);
            xmlCleanupParser();

            if (status == 0) {
                break;
            }
        }
    }

    if(_KDI_LOG_TRACE_ && status)
    {
        printf("[delete_key_by_key_type] Did not find %s key with uid %s to delete\n", key_name, uid);
    }

    return status;
}


/**
* Make a copy an hmac private or symetric key. Store using the same UID with the input key name
*
* @param[in]  uid          ascii representation of the UID to associate the key with
*                          May NOT be NULL
* @param[in]  key_name     key name of the key to make a copy of
*                          May NOT be NULL
* @param[in]  copy_name    key name of the new key
*                          May NOT be NULL
* 
* @return 0 on success
*         1 on failure
*/
uint32_t copy_private_key(
    const char * uid,
    const char * key_name,
    const char * copy_name)
{
    uint32_t status = 1;
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    const char * key_xml_file = KEY_XML_FILE_PRV;

    if (uid == NULL || key_name == NULL || copy_name == NULL)
    {
        return status;
    }

    if (strlen(uid) <= 0)
    {
        printf("[copy_prv_key] error: length of input parameter uid is not valid\n");
        return status;
    }

    pthread_mutex_lock(&lock);

    doc = xmlReadFile(key_xml_file, NULL, 0);
    if (doc == NULL)
    {
        pthread_mutex_unlock(&lock);
        printf("[copy_prv_key] error: could not parse file %s\n", key_xml_file);
        return status;
    }

    /*Get the root element node */
    root_element = xmlDocGetRootElement(doc);
    if (root_element == NULL)
    {
        pthread_mutex_unlock(&lock);
        printf("[copy_prv_key] error: could retrieve root element after parsing file %s\n", key_xml_file);
        xmlFreeDoc(doc);
        xmlCleanupParser();
        return status;
    }

    // make sure this is not a duplicate uid/key-name pair
    xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) copy_name, NULL);
    if(node != NULL)
    {
        pthread_mutex_unlock(&lock);
        printf("[copy_prv_key] error: Cannot add key for uid 0x%s and key name %s! Key with same uid and key_name already exists\n", uid, copy_name);
        xmlFreeDoc(doc);
        xmlCleanupParser();
        return status;
    }

    node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);


    if(node == NULL)
    {
        printf("[copy_prv_key] error: no key with found with  uid=%s; key name=%s\n", uid, key_name);
    }
    else
    {
        const xmlChar * keyContent = getRawKeyFromNode(node);
        if(keyContent == NULL || strlen((const char*)keyContent) <= 0)
        {
            pthread_mutex_unlock(&lock);
            printf("[copy_prv_key] error: Failed to extract key data from key with uid=%s; key name=%s\n", uid, key_name);
            xmlFreeDoc(doc);
            xmlCleanupParser();
            return status;
        }
    
        char handle[128]; // somewhat arbitrary 128 byte max handle length

        // generate new key handle
        if (get_new_key_handle(handle, sizeof(handle)))
        {
            printf("[copy_prv_key] error: could not get new key handle\n");
        }
        else
        {
            // create a new node
            xmlNode * new_node = xmlNewNode(NULL, xmlStrdup((xmlChar *) XML_NodeName_Key_Content));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_UID), xmlStrdup((xmlChar *) uid));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_key_Name), xmlStrdup((xmlChar *) copy_name));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_Handle), xmlStrdup((xmlChar *) handle));
            xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_KEY), xmlStrdup((xmlChar *) keyContent));
            // if(metadata != NULL && strlen(metadata) > 0)
            // {
            //     xmlNewChild(new_node, NULL, xmlStrdup((xmlChar *) XML_NodeName_metadata), xmlStrdup((xmlChar *) metadata));
            // }

            // add new node to root
            xmlAddChild(root_element, new_node);
            status = 0;
        }
    }
    
    xmlSaveFile(key_xml_file, doc);

    pthread_mutex_unlock(&lock);

    xmlFreeDoc(doc);
    xmlCleanupParser();

    return status;
}

/**
* Return the hmac private or symetric key associated with the input UID and key name as a hex string.
*
* @param[in]     uid       ascii representation of the UID to associate the key with
*                          May NOT be NULL
* @param[in]     key_name  key name of the key to make a copy of
*                          May NOT be NULL
* @param[out]    hex_key   buffer to hold the exported key
*                          May NOT be NULL
* @param[in,out] key_len   on input contains the length of output key buffer, on successful
*                          completion contains the length of the exported key (as hex string)
*                          May NOT be NULL
* 
* @return 0 on success
*         1 on failure
*/
static uint32_t _get_key_as_hex(
    const char * calling_function,
    const char * key_store_filename,
    const char * uid,
    const char * key_name,
          char * hex_key,
    uint32_t *key_len)
{
    uint32_t status = 1;
    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;
    const char * key_xml_file = NULL;

    if (uid == NULL || key_name == NULL || hex_key == NULL || key_len == NULL)
    {
        printf("[%s] error: bad argument!\n", calling_function);
        return status;
    }

    pthread_mutex_lock(&lock);
    doc = xmlReadFile(key_store_filename, NULL, 0);
    pthread_mutex_unlock(&lock);

    if (strlen(uid) <= 0)
    {
        printf("[%s] error: length of input parameter uid is not valid\n", calling_function);
        return status;
    }

    if (doc == NULL)
    {
        printf("[%s] error: could not parse file %s\n", calling_function, key_xml_file);
        return status;
    }

    /*Get the root element node */
    root_element = xmlDocGetRootElement(doc);
    if (root_element == NULL)
    {
        printf("[%s] error: could retrieve root element after parsing file %s\n", calling_function, key_xml_file);
        return status;
    }

    xmlNode * node = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) key_name, NULL);
    if(node == NULL)
    {
        printf("[%s] error: no key with found with  uid=%s; key name=%s\n", calling_function, uid, key_name);
        return status;
    }
    else
    {
        const xmlChar * keyContent = getRawKeyFromNode(node);
        if(keyContent == NULL || strlen((const char*)keyContent) <= 0)
        {
            printf("[%s] error: Failed to extract key data from key with uid=%s; key name=%s\n", calling_function, uid, key_name);
            return status;
        }
    
        uint32_t found_key_len = strlen((const char*)keyContent);
        if(found_key_len >= *key_len)
        {
            printf("[%s] error: output buffer not large enough to hold key! uid=%s; key name=%s\n", calling_function, uid, key_name);
            return status;
        }

        strcpy(hex_key, (const char*)keyContent);
        *key_len = found_key_len;
        status = 0;
    }
    
    xmlFreeDoc(doc);
    xmlCleanupParser();

    return status;
}


/**
* Return the hmac private or symetric key associated with the input UID and key name as a hex string.
*
* @param[in]     uid       ascii representation of the UID to associate the key with
*                          May NOT be NULL
* @param[in]     key_name  key name of the key to make a copy of
*                          May NOT be NULL
* @param[out]    hex_key   buffer to hold the exported key
*                          May NOT be NULL
* @param[in,out] key_len   on input contains the length of output key buffer, on successful
*                          completion contains the length of the exported key (as hex string)
*                          May NOT be NULL
* 
* @return 0 on success
*         1 on failure
*/
uint32_t export_prv_key_in_clear_as_hex(
    const char * uid,
    const char * key_name,
          char * hex_key,
    uint32_t *key_len)
{
    return _get_key_as_hex("export_prv_key_in_clear_as_hex", 
                           KEY_XML_FILE_PRV,
                           uid, key_name, hex_key, key_len);
}

/**
* Return the hmac private or symetric key associated with the input UID and key name as a hex string.
*
* @param[in]     uid       ascii representation of the UID to associate the key with
*                          May NOT be NULL
* @param[in]     key_name  key name of the key to make a copy of
*                          May NOT be NULL
* @param[out]    hex_key   buffer to hold the exported key
*                          May NOT be NULL
* @param[in,out] key_len   on input contains the length of output key buffer, on successful
*                          completion contains the length of the exported key (as hex string)
*                          May NOT be NULL
* 
* @return 0 on success
*         1 on failure
*/
uint32_t read_public_key_as_hex(
    const char * uid,
    const char * key_name,
          char * hex_key,
    uint32_t *key_len)
{
    return _get_key_as_hex("read_public_key_as_hex", 
                           KEY_XML_FILE_PUB,
                           uid, key_name, hex_key, key_len);
}

/**
 * Replace the srk key data for the given UID with the
 * key content associated with the given key handle.
 * 
 * Note that the existing key data for the srk key of the
 * input UID is overwritten with the key data associated
 * with the input handle.  The entry for the input handle
 * remains unchanged.
 * 
 * @param[in] uid the uid of the srk key to replace.
 *                an entry with this UID and key type 'srk'
 *                must exist in the key repository.
 * 
 * @param[in] handle handle to the new key data
 *                   this handle must refer to an existing 
 *                   key in the key repository
 *
 * @return  0 success
 *          1 failure
 */
uint32_t set_new_srk_key_data_for_uid(const char * uid, const char * handle)
{
  xmlDoc *doc = NULL;
  xmlNode *root_element = NULL;

  pthread_mutex_lock(&lock);
  
  doc = xmlReadFile(KEY_XML_FILE_PRV, NULL, 0);
  if (doc == NULL) {
      pthread_mutex_unlock(&lock);
      printf("[key_driver.set_new_srk_key_data_for_uid] error: could not parse file %s\n", KEY_XML_FILE_PRV);
      return 1;
  }

  root_element = xmlDocGetRootElement(doc);
  xmlNode * destNode = find_node_from_input_root(root_element, (xmlChar*) uid, NULL, (xmlChar*) "srk", NULL);
  if(destNode == NULL) {
      pthread_mutex_unlock(&lock);
      printf("[key_driver.set_new_srk_key_data_for_uid] error: Failed to find key type 'srk' for UID %s\n", uid);
      return 1;
  }

  xmlNode * keyNode =  find_node_from_input_root(root_element, NULL, (xmlChar*) handle, NULL, NULL);
  if(keyNode == NULL) {
      pthread_mutex_unlock(&lock);
      printf("[key_driver.set_new_srk_key_data_for_uid] error: Failed to find key with handle %s\n", handle);
      return 1;
  }

  xmlNode * childIterator = destNode->children;
  while(childIterator){
      if(!xmlStrcmp(childIterator->name, (const unsigned char *) "key"))
          break;
      childIterator = childIterator->next;
  }
  
  if(childIterator == NULL) {
      pthread_mutex_unlock(&lock);
      printf("[key_driver.set_new_srk_key_data_for_uid] error: Missing key data for handle %s\n", handle);
      return 1;
  }

  xmlChar * keyContent = getRawKeyFromNode(keyNode);
  if(keyContent == NULL) {
      pthread_mutex_unlock(&lock);
      printf("[key_driver.set_new_srk_key_data_for_uid] error: Missing key data field for srk with UID %s\n", uid);
      return 1;
  }

  xmlNodeSetContent(childIterator->children, keyContent);

  xmlSaveFile(KEY_XML_FILE_PRV, doc);

  pthread_mutex_unlock(&lock);

  xmlFreeDoc(doc);
  xmlCleanupParser();
  return 0;
}

uint32_t delete_crk(
    const char* uid)
{
    if(uid == NULL)
    {
        return 1;
    }

    uint32_t rc = delete_key_by_key_type(uid, KeyName_CRK);
    
    return rc;
}

uint32_t add_crk(
    const char* uid,
    const char* key_hex,
    const char* metadata)
{
    if(uid == NULL)
    {
        return 1;
    }
    uint32_t rc = add_nonvolatile_based_key(uid, KeyName_CRK, key_hex, metadata, true, false);
    
    return rc;
}

uint32_t add_csk(
    const char* uid,
    const char* key_hex,
    const char* metadata)
{
    if(uid == NULL)
    {
        return 1;
    }

    uint32_t rc = add_nonvolatile_based_key(uid, KeyName_CSK, key_hex, metadata, true, false);
    
    return rc;
}

bool _prv_keyfile_contains_key(const char *uid, const char *key_name)
{
    assert(uid != NULL && key_name != NULL);

    xmlDoc *doc = NULL;
    xmlNode *root_element = NULL;

    pthread_mutex_lock(&lock);

    doc = xmlReadFile(KEY_XML_FILE_PRV, NULL, 0);

    pthread_mutex_unlock(&lock);

    if (doc == NULL) {
        printf("[key_driver._prv_keyfile_contains_key(uid, key)] error: could not parse file %s\n", KEY_XML_FILE_PRV);
        return false;
    }

    root_element = xmlDocGetRootElement(doc);

    xmlNode * node = find_node_from_input_root(root_element, (const xmlChar *) uid, NULL, (const xmlChar *) key_name, NULL);
    return node != NULL;
}

bool contains_crk(const char *uid)
{
    if(uid == NULL)
    {
        printf("[key_driver.contains_csk] ERROR input uid is NULL");
        return false;
    }

    return _prv_keyfile_contains_key(uid, KeyName_CRK);
}

bool contains_csk(const char *uid)
{
    if(uid == NULL)
    {
        printf("[key_driver.contains_csk] ERROR input uid is NULL");
        return false;
    }

    return _prv_keyfile_contains_key(uid, KeyName_CSK);
}

bool contains_key(const char *uid, const char *keyname)
{
    if(uid == NULL || keyname == NULL)
    {
        printf("[key_driver.contains_csk] ERROR input uid or keyname is NULL");
        return false;
    }

    return _prv_keyfile_contains_key(uid, keyname);
}
