/***
 * 
 * @file authenta_key_management_utils.h
 * 
 * @brief Authenta key library key management functions interface
 * 
 * @copyright (C) 2020 Micron Technology, Inc. All rights reserved.
 * 
 **/
#ifndef XML_KEY_STORAGE_H_
#define XML_KEY_STORAGE_H_


/******
 * 
 * In this implementation keys are stored in an XML formatted file (in the clear).
 * 
 * An example of the expected XML structure is below:
 *  
 *   <key_content>
 *     <uid>1032547698badcfe1f32547698baaa55</uid>
 *     <key_type>hmac</key_type>
 *     <key_name>srk</key_name>
 *     <key>d5f55d8a267ad5daa3e93e30d804bb70993aa705c908a82d850e7e495b01566a</key>  
 *     <handle>1138</handle>
 *   </key_content>
 * 
 *   <key_content>
 *     <uid>a89bc41465ae78813497b51d1f2b32dc</uid>
 *     <key_type>ecdh</key_type>
 *     <key_name>ecdh_sr_prk</key_name>
 *     <key_pub_x>2b2c0aeb03441c8880a8cc3cb715c0f6014d3022b76812c6ac37d6b57ad057c6</key_pub_x> 
 *     <key_pub_y>1c287b9342da3c10d1a38694c1bd4351e0a5f0f55979af97d98070d85e9b6ace</key_pub_y>
 *     <key_prv>bf5962409acc3c370d52f9d173c7f0ddab637bc373e69462a3f438c098ca5570</key_prv>
 *     <handle>1148</handle>
 *   </key_content>
 * 
 ******/

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "assert.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

#ifndef false
#define false FALSE
#endif /* false */

#ifndef true
#define true TRUE
#endif /* true */

#define _KDI_LOG_TRACE_ (false)

#define AUTHENTA_KEY_DRIVER_MAX_HANDLE_LEN 16

static const xmlChar * XML_NodeName_Key_Content = (xmlChar *) "key_content";
static const xmlChar * XML_NodeName_UID = (xmlChar *) "uid";
static const xmlChar * XML_NodeName_KEY = (xmlChar *) "key";
static const xmlChar * XML_NodeName_Handle = (xmlChar *) "handle";
static const xmlChar * XML_NodeName_key_Name = (xmlChar *) "key_name";
static const xmlChar * XML_NodeName_key_Type = (xmlChar *) "key_type";
static const xmlChar * XML_NodeName_key_Pub_X = (xmlChar *) "key_pub_x";
static const xmlChar * XML_NodeName_key_Pub_Y = (xmlChar *) "key_pub_y";
static const xmlChar * XML_NodeName_key_Prv = (xmlChar *) "key_prv";
static const xmlChar * XML_NodeName_metadata = (xmlChar *) "metadata";

#define KEY_NAME_PUB "_pub"
#define KEY_NAME_PRV "_prv"

#define KeyName_SRK ("srk")
#define KeyName_CRK ("crk")
#define KeyName_CSK ("tmp_client_session_key")
#define KeyName_ECDSA_PRV ("ecdsa_prv")
#define KeyName_ECDSA_PUB ("ecdsa_pub")
#define KeyName_ECDSA_DS_PuK ("ECDSA_DS_PuK")
#define KeyName_ECDH_SR_PRV ("ecdh_prv")
#define KeyName_Session ("session")

#define KeyType_HMAC ("hmac")
#define KeyName_ECDH ("ecdh")


// static const char * KeyName_SRK = "srk";
// static const char * KeyName_ECDH_SR_PRV = "ecdh_sr_prv";
// static const char * KeyName_Session = "session";

// static const char * KeyType_HMAC = "hmac";
// static const char * KeyName_ECDH = "ecdh";

#define KEY_XML_FILE_PUB  "key_contents_pub.xml"
#define KEY_XML_FILE_PRV  "key_contents_prv.xml"

/** \fn getPublicKeyFromString
* \brief Convert a 128 hex chars string in a public key 
* \param[out] key Output key (64 bytes)
* \param[in] string input 128 byte string vector
* \return 0 or error code
*/
uint32_t GetPublicKeyFromString( uint8_t* key, const char* string );



/** \fn getPrivateKeyFromString
* \brief Convert a 64 hex chars string in a private key 
* \param key Output key (32 bytes)
* \param string input 64 byte string vector
* \return 0 or error code
*/
uint32_t GetPrivateKeyFromString( uint8_t* key, const char* string );


/**
* Get handle for a key based on input UID and key name
*
* @param[in]  uid               UID of the key to return
*                               May NOT be NULL
* @param[in]  key_name          which key associated with input UID to return (srk, ecdh_sr_prk, session)
*                               May NOT be NULL
* @param[out] metadata_out_buff output buffer to place a NULL terinated char string holding metadata associated with the key
*                               NULL indicates not to return metadata
* @param[out] metadata_out_buff_size length of metadata buffer. Must be large enough to hold NULL terminated metadata.
*
* @param[out] handle_out_buff   output buffer to hold the new key handle;
*                               May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
* 
* @return Success 0
*         Fail    1  (input uid, key_name, or handle_buffer is null, key file not found, uid not found) 
*/
uint32_t get_key_handle(
    const char * uid,
    const char * key_name,
    char * metadata_out_buff,
    unsigned int metadata_out_buff_size,
    char * handle_out_buff,
    unsigned int handle_out_buff_size);

/**
* Get the (raw) key based on input UID and key name
*
* @param[in]  uid               UID of the key to return
*                               May NOT be NULL
* @param[in]  key_name          which key associated with input UID to return (srk, ecdh_sr_prk, session)
*                               May NOT be NULL
* @param[out] metadata_out_buff output buffer to place a NULL terinated char string holding metadata associated with the key
*                               NULL indicates not to return metadata
* @param[out] metadata_out_buff_size length of metadata buffer. Must be large enough to hold NULL terminated metadata.
*
* @param[out] raw_key bytes containing the raw key; MUST NOT be NULL; 
*                     Call will fail if not large enough to hold the raw key
*                     May NOT be NULL.
* @param[in,out] raw_key_buf_len pointer to len of input buffer, 
*                                on succesful return will hold the length
*                                of the returned key
*                                May NOT be NULL.
* 
* @return Success 0
*         Fail    1  (input uid, key_name, or handle_buffer is null, key file not found, uid not found) 
*/
uint32_t get_key(
    const unsigned char * uid,
    const char * key_name,
    uint8_t * key, 
    uint32_t *key_buf_len_p);

/**
* Add a new temprary key (handle, key) and generate handle for the new key
* 
* @param[in]  key_hex          pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[out] handle_out_buff  output buffer to hold the new key handle;
*                              May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
*
* @return 0 on success
*         1 on failure
*/
uint32_t add_volatile_key(
    const char * key_hex,
    char * handle_out_buff,
    unsigned int handle_out_buff_size);

/**
* Add a new hmac key (uid, key_name, key) to the key file and generate handle for the new key
*
* @param[in]  uid              pointer to hex ascii representation of the UID to associate the key with
*                              May NOT be NULL
* @param[in]  key_name         pointer to the type of key, e.g. "crk" or "session"
*                              May NOT be NULL
* @param[in]  key              pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[in]  metadata         NULL terinated char string holding metadata to associate with the key
*                              NULL or empty string indicates no metadata to be stored
* @param[in]  is_private_key   if true the imput is stored as a private key, 
*                              if false it is stored as a public key
* @param[in]  allow_replace    if true an existing key with a matching UID and name will be replaced
*                              if false it if a key with a matching UID and name exists it will not
*                              be changed and this function will fail
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_nonvolatile_based_key(
    const char * uid,
    const char * key_name,
    const char * key,
    const char * metadata,
    const bool is_private_key,
    const bool allow_replace);


/**
* Add a new temporary hmac key (uid, key_name, key) to the memory based volatile key tree
* and generate handle for the new key.
* If a key associated with the same UID and key_name already exists remove it
* and add then add the key.
*
* @param[in]  uid              pointer to hex ascii representation of the UID to associate the key with
*                              May NOT be NULL
* @param[in]  key_name         pointer to the type of key, e.g. "crk" or "session"
*                              May NOT be NULL
* @param[in]  key              pointer to hex ascii representation of the key to add
*                              May NOT be NULL
* @param[in]  metadata         NULL terinated char string holding metadata to associate with the key
*                              NULL or empty string indicates no metadata to be stored
* @param[out] handle_out_buff  output buffer to hold the new key handle;
*                              May NOT be NULL
* @param[in]  handle_out_buff_size size of handle output buf
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_volatile_based_key(
    const char * uid,
    const char * key_name,
    const char * key,
    const char * metadata,
    char * handle_out_buff,
    unsigned int handle_out_buff_size);

/**
* Delete key by handle
*
* First the volatile key tree is check, then the 
* non-volatile (file-based) tree is searched. (Only 
* (one key is deleted).
*
* @param[in] handle  handle of key to be deleted
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
uint32_t delete_key_by_handle(const char * handle);

/**
* Delete key by key tpe and UID
*
* First the volatile key tree is checked, then the 
* non-volatile (file-based) tree is searched. (Only 
* (one key is deleted).
*
* @param[in] uid  the uid associated with the key
* @param[in] key_name  the name of the key
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
uint32_t delete_key_by_key_type(
    const char * uid,
    const char * key_name);


uint32_t copy_private_key(const char * uid, const char * key_name, const char * copy_name);

uint32_t read_public_key_as_hex(
    const char * uid,
    const char * key_name,
          char * hex_key,
    uint32_t *key_len);

uint32_t export_prv_key_in_clear_as_hex(
    const char * uid,
    const char * key_name,
          char * hex_key,
    uint32_t *key_len);

uint32_t set_new_srk_key_data_for_uid(const char * uid, const char * handle);

/**
* Delete the CRK
*
* @param[in] uid  uid in hex string format of the CRK to be deleted
*
* @return Success 0
*         Fail    1  (input handle is null, key file not found, handle not found) 
*/
uint32_t delete_crk(const char* uid);

/**
* Add a new crk to the key file
*
* @param[in] uid         UID in hex string format to associate with the CRK. 
* @param[in] key_hex     pointer to hex ascii representation of the key to add
*                        May NOT be NULL
* @param[in] metadata    NULL terinated char string holding metadata to associate with the key
*                        NULL or empty string indicates no metadata to be stored
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_crk(
    const char* uid,
    const char* key_hex,
    const char* metadata);
    
/**
* Add a new Client Session Key csk to the key file
*
* Note this is meant to be a temporary key
*
* @param[in] uid         UID in hex string format to associate with the CRK. 
* @param[in] key_hex     pointer to hex ascii representation of the key to add
*                        May NOT be NULL
* @param[in] metadata    NULL terinated char string holding metadata to associate with the key
*                        NULL or empty string indicates no metadata to be stored
* 
* @return 0 on success
*         1 on failure
*/
uint32_t add_csk(
    const char* uid,
    const char* key_hex,
    const char* metadata);
    
/**
* Return true if the private key repository contains a CRK for the input UID, false otherwise
*
* @param[in] uid  UID of the associated CRK that is being queried. May NOT be NULL.
*/
bool contains_crk(const char *uid);

/**
* Return true if the private key repository contains a CSK for the input UID, false otherwise
*
* @param[in] uid  UID of the associated CRK that is being queried. May NOT be NULL.
*/
bool contains_csk(const char *uid);

/**
* Return true if the private key repository contains a key for the input UID and keyname, 
* false otherwise
*
* Note this function works for private keys only
*
* @param[in] uid  UID of the associated CRK that is being queried. May NOT be NULL.
* @param[in] keyname  name of the associated key that is being queried. May NOT be NULL.
*/
bool contains_key(const char *uid, const char *keyname);

#endif /* XML_KEY_STORAGE_H_ */