#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import argparse as ap
import logging
from kms_load_test_config import *
import yagmail
from os import path

from datetime import datetime
from kms_latency_profile_decorator import kms_latency_profile, print_prof_data, set_csv_output_filename

# --- Authenta imports ---
from authenta.maps import Envr
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

#import app_common_logging as app_logging

VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_demo.py  <endpoint alias> 
  -- as in --
   $> app_demo.py snor1
  - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_demo.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_demo.py -f mydevices_mapping.json snor1

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''


p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, 
                      description=PROGRAM_DES,
                      epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")


p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

#app_logging.setup_app_logging()
#main_logger = logging.getLogger(__name__)

TIME_BETWEEN_CALLS_SECS = 1
STATUS = True

def get_additional_output_for_getNonceWithMPN(func_name, result):
    success = True if result[1] == 200 else False
    error_message = None if success else result[0]
    status = result[1]

    return f'{success},{error_message},{status}'

def check_status(success_status, error_message, loc_time_start):
    global STATUS
    status_changed = STATUS != success_status

    if status_changed:
        subject = str(success_status)
        if success_status:
            subject = "GetNonce Success"
            contents = f"<div style='font-size:12px'> \
                        <p>Post to GetNonce on KMS Proxy was successful!</p> \
                        <p><b>Timestamp:</b> {loc_time_start}</p></div>"
        else:
            subject = "GetNonce Failure"
            contents = f"<div style='font-size:12px'> \
                        <p>Post to GetNonce on KMS Proxy failed!</p> \
                        <p><b>Timestamp:</b> {loc_time_start}</p> \
                        <p><b>Error message:</b> {error_message}</p></div>"
            
        send_email(subject, contents)

    STATUS = success_status

def send_email(subject, contents):
    try:
        yag = yagmail.SMTP({SENDER_EMAIL:'donotreply@trustiphi.com'}, oauth2_file='/oauth2.py')
        yag.send(to=RECEIVER_EMAIL, subject=subject, contents=contents)
        print("Email sent successfully\n")
    except Exception as ex:
        print("\nERROR: {} Exception thrown in send_email: {}".format(type(ex), ex))
        print("Email was not sent\n")



@kms_latency_profile
def get_nonce_from_kms(connection):
    #
    # get a nonce from the Micron KMS Proxy
    #
    nonce, response, status = MicronKMSProxyAPI.getNonceWithMPN(connection, "MTANXSBB01")

    if nonce:
      print('Nonce retrieved from KMS: {}'.format(nonce))
    else:
        # if nonce is None, response will be in nonce_tag
        print("Failed to receive nonce from KMS: {}\n".format(response))

    if response is None:
        # something is wrong with what was returned by the KMS
        # this should not happen (nonce was returned from KMS and nonce_tag is None)
        print("Failed to receive valid nonce from KMS: missing nonce_tag\n")

    success = True if status == 200 else False
    error_message = None if success else response
    check_status(success, error_message, datetime.now())

    return response, status

def do_get_nonce_test():
  """
  1.	connect to the Micron KMS and get a nonce,
  In an infinite loop:
      2.    write elapsed time information to .csv file

  """

  # setup output CSV
  set_csv_output_filename(FILE_NAME, get_additional_output_for_getNonceWithMPN, "Success,Error Message,Status Code")

  count = 0
  print("Enter CTRL-C to exit...\n")

  while True:
    #
    # Set the password from command line inputs. If neither exists, ask user for password
    #
    if ARGS.tls_certificate_pw:
        cert_password = ARGS.tls_certificate_pw
    elif path.exists(ARGS.tls_certificate_pw_file):
        fp = open(ARGS.tls_certificate_pw_file)
        cert_password = fp.readline()
    else:
        cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")

    # create a connection to the Micron KMS Proxy
    connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)

    get_nonce_from_kms(connection)
    time.sleep(10)
    if INFINITE is False:
      count += 1
      if count > MAX_COUNT:
        break

# program entry point
if __name__ == "__main__":
# call main function
  # sys.exit() acts as a thin wrapper -
  # kind of global exception handler
  try:
    do_get_nonce_test()
  except Exception as ex:
    print("\nERROR: {} Exception thrown in do_get_nonce_test: {}\n".format(type(ex), ex))

  print("Showing Results:")
  print_prof_data()
  exit(0)
