#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import argparse as ap
import logging
import threading

from hashlib import sha256
from os import path
from kms_load_test_config import *
from kms_latency_profile_decorator import print_prof_data, set_csv_output_filename
from kms_load_test_positive_cases import *
from kms_load_test_error_cases import *

import kms_load_test_utils as utils

# --- Authenta imports ---
from authenta.maps import Envr
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
from mock_naxos_device import MockNaxosDevice

import app_common_logging as app_logging


VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
'''

EPILOG = '''


EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_kms_load_test_basic.py  <endpoint alias> 
  -- as in --
   $> app_kms_load_test_basic.py snor1
  - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_kms_load_test_basic.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_kms_load_test_basic.py -f mydevices_mapping.json snor1

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2020

'''


p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, 
                      description=PROGRAM_DES,
                      epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")


p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
logger = logging.getLogger(__name__)


def do_test_loop(test_func, device, loop_times):
  count = 1

  while True:
    #
    # Set the password from command line inputs. If neither exists, ask user for password
    #
    if ARGS.tls_certificate_pw:
        cert_password = ARGS.tls_certificate_pw
    elif path.exists(ARGS.tls_certificate_pw_file):
        fp = open(ARGS.tls_certificate_pw_file)
        cert_password = fp.readline()
    else:
        cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")

    # create a connection to the Micron KMS Proxy
    connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)

    test_func(device, connection)
    time.sleep(SECS_BETWEEN_CALLS)
    if loop_times > 0:
      count += 1
      if count > loop_times:
        break


def start_users():
    num_users = len(ACTION_ITEMS)
    if num_users == 0:
        print('No users to start!')
        return

    # print("Enter CTRL-C to exit...\n")

    logger.info('Number of users: {}'.format(num_users))
    logger.info('Run {} times per user'.format(MAX_COUNT))
    print('Number of users: {}'.format(num_users))
    print('Run {} times per user'.format(MAX_COUNT))

    set_csv_output_filename(FILE_NAME, utils.get_additional_output_for_KMS_functions, "Success,Error Message,Status Code,MPN")

    all_threads = [ ]
    for i in range(num_users):
        # uid = DEVICE_UIDS[i]
        # srk = DEVICE_SRKS[i]
        action_item = ACTION_ITEMS[i]
        uid = action_item[0]
        srk = action_item[1]
        func = func_lookup[action_item[2]]
        device = MockNaxosDevice(uid=uid, srk=bytes.fromhex(srk), amtc0=0x000001d9)

        # start a thread for each one
        # t = threading.Thread(target=perform_get_nonce, args=(device,))
        t = threading.Thread(target=do_test_loop, args=(func, device, MAX_COUNT))
        all_threads.append(t)
        t.start()
        time.sleep(6)

    # now Wait for all of them to finish
    for t in all_threads:
        t.join()



# program entry point
if __name__ == "__main__":
# call main function
  # sys.exit() acts as a thin wrapper -
  # kind of global exception handler
  try:
    start_users()
  except Exception as ex:
    print("\nERROR: {} Exception thrown in start_users: {}\n".format(type(ex), ex))

  print("Showing Results:")
  print_prof_data()
  exit(0)
