#! /usr/bin/env python3

"""
KMS Test setup.
(C) 2020 Micron Technology, Inc. All rights reserved.
"""

import json
import pytest

from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.grpc_platform import GrpcPlatform
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

device_json_file_name = 'devices.json'
ep_alias = 'snor1'
tls_certificate = "../kms_proxy_certs/active_kms_tls_client_cert.pem"
tls_certificate_pw = "Au7_FYtH_9198_o0T_a"

def get_dev_info_from_json(filename, alias):
  """
  Get device info from device json file.
  return (device path, endpoint IP address, endpoint port number, iboot IP address)
  """
  try:
    with open(filename) as f:
      json_data = json.load(f)

    json_data = json_data[alias]

    ip_params = json_data['endpoint_ip'].split(':')
    ep_ip = ip_params[0]
    ep_port = None
    if len(ip_params) > 1:
      ep_port = ip_params[1]

    dev_path = json_data['endpoint_fn']

  except FileNotFoundError:
    print('Error: Failed to open file:{}'.format(filename))
    pytest.exit(1)
  except json.decoder.JSONDecodeError:
    print('Error: Failed to parse json file:{}'.format(filename))
    pytest.exit(1)
  except:
    print('Error: Failed to get device info from json file:{}'.format(filename))
    pytest.exit(1)

  return dev_path, ep_ip, ep_port

@pytest.fixture(scope='session', autouse=True)
def run_setup(request):
  """
  Run test setup before all tests
  """

  # Connect to Naxos Device 
  dev_path, ep_ip, ep_port = get_dev_info_from_json(
    device_json_file_name, ep_alias)

  # save shared data in pytest cache
  request.config.cache.set('dev_path', dev_path)
  request.config.cache.set('ep_ip', ep_ip)
  request.config.cache.set('ep_port', ep_port)

def get_device(request):
  """
  Create NaxosDevice_1_5 object from pytest cache data
  """
  dev_path = request.config.cache.get("dev_path", None)
  ep_ip = request.config.cache.get("ep_ip", None)
  ep_port = request.config.cache.get("ep_port", None)
  
  if ep_port is None:
    ep = GrpcPlatform(dev_path, ep_ip)
  else:
    ep = GrpcPlatform(dev_path, ep_ip, ep_port)

  device = NaxosDevice_1_5(ep)

  return device

def create_connection():
  """
  Create a connection to the Micron KMS Proxy
  """
  connection = MicronKMSProxyAPI.createMTTLS(tls_certificate, tls_certificate_pw)

  return connection
