import os
import time
from pathlib import Path
from functools import wraps
from datetime import datetime
from kms_load_test_config import *

PROF_DATA = {}

additional_csv_fields_func = None
prof_file = None
headers = f'{"Function Name"},{"Elapsed Time"},{"Start Time"},{"End Time"}'

def set_csv_output_filename(filename: str, additional_fields_func=None, additional_headers=None):
    """
    This function sets the name of a CSV file that will be used to output the profiling information
    along with any additional information desired by the caller.

    If the filename is None than no output will be done while profiling.

    This function is also used to set a function that will determine additional information that
    should be output to the CSV file.
    The input function must return a list of strings. These strings
    will be added to the .csv output.  The function must take two input fields,
    the name of the function that is being profiled and the and return value 
    of that function

    Finally this function takes a string containing any additional headers that should go 
    in the output CSV file. The number of headers must be the same as the number of additional
    additional information fields returned by the additional_fields_function.

    Throws any exceptions that are thrown by file open and file write.

    """
    global additional_csv_fields_func
    global prof_file

    if filename is None:
        prof_file = None
        return

    # create file with the input name and the current time concatenated
    fname = filename
    if fname.endswith('.csv'):
        fname = os.path.splitext(fname)[0]
   
    all_headers = headers
    if additional_headers is not None and additional_headers != '':
        all_headers += (',' + additional_headers)

    prof_file = "{}_{}.csv".format(fname, time.strftime("%Y%m%d-%H%M%S", time.localtime()))
    with open(prof_file, 'a') as f:
        f.write(all_headers + '\n')

    print('Created profiling file {}'.format(prof_file))

    # if additional fields are requested in the output set the 
    # function that will return them for each line of output
    additional_csv_fields_func = additional_fields_func


def print_single_prof_data(func_name, idx, additional_output):
    if PROF_DATA[func_name] is not None:
        # print fillowing info to file
        # func_name, elapsed time (ns), start_time, end_time, success, error_message, status
        timing_data = PROF_DATA[func_name][1][idx]

        start_time = datetime.strptime(str(timing_data[1]), '%Y-%m-%d %H:%M:%S.%f')
        end_time = datetime.strptime(str(timing_data[2]), '%Y-%m-%d %H:%M:%S.%f')

        line_out = f'{func_name},{timing_data[0]},{start_time},{end_time},' + additional_output + '\n'

        with open(prof_file, 'a') as f:
            f.write(line_out)
#        print(line_out)

def kms_latency_profile(fn):
    @wraps(fn)
    def with_profiling(*args, **kwargs):
        loc_time_start = datetime.now()
        start_time = time.monotonic()

        ret = fn(*args, **kwargs)

        elapsed_time = time.monotonic() - start_time
        loc_time_end = datetime.now()

        if fn.__name__ not in PROF_DATA:
            PROF_DATA[fn.__name__] = [0, []]
        idx = PROF_DATA[fn.__name__][0]
        timing_data = [elapsed_time, loc_time_start, loc_time_end]
        PROF_DATA[fn.__name__][0] += 1
        PROF_DATA[fn.__name__][1].append(timing_data)

        if additional_csv_fields_func is not None:
            additional_output = additional_csv_fields_func(fn.__name__, ret)
        else:
            additional_output = None

        print_single_prof_data(fn.__name__, idx, additional_output)

        return ret

    return with_profiling

def print_prof_data():
    for fname, data in PROF_DATA.items():
        min_time = min([i[0] for i in data[1]])
        max_time = max([i[0] for i in data[1]])
        avg_time = sum([i[0] for i in data[1]]) / data[0]

        results = f'\nFunction "{fname}" called {data[0]} times'
        print(results)
        min_result = f'Execution time MIN: {min_time:12.3f}'
        print(min_result)
        max_result = f'Execution time MAX: {max_time:12.3f}'
        print(max_result)
        average_result = f'Average: {avg_time:12.3f}\n'
        print(average_result)

        # with open(prof_file, 'a') as f:
        #     f.write(results)

def clear_prof_data():
    global PROF_DATA
    PROF_DATA = {}
