
#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import logging

from kms_load_test_config import *
from kms_latency_profile_decorator import kms_latency_profile
from kms_load_test_utils import *

import app_common_logging as app_logging

VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
'''

EPILOG = '''

AUTHOR:
\tTatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2020

'''

app_logging.setup_app_logging()
logger = logging.getLogger(__name__)


def setup_confirm_device_identity(device, connection):
    #
    # get a nonce from the Micron KMS Proxy
    #
    the_nonce, nonce_tag, status, mpn = get_nonce_from_kms(connection, MPN_CONFIRM_DEVICE_IDENTITY)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY
    if nonce_tag is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    #
    # call Request_UID command on attached device using the nonce
    # retrieved from the Micron KMS Proxy
    #
    err, requid_response, resp_nonce = device.meta_request_uid(b'\x00', nonce_bytes)
    if not err.esr_success:
        logger.error('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'meta_request_uid', MPN_CONFIRM_DEVICE_IDENTITY

    logger.info('Device UID: {}'.format(device.uid_as_str))
    logger.info('Naxos Revision: {}'.format(device.revision.hex()))

    return connection, device, nonce_tag, requid_response


@kms_latency_profile
def confirm_device_identity_no_nonce_tag(device, connection):
    
    connection, device, nonce_tag, requid_response = setup_confirm_device_identity(device, connection)

    nonce_tag = None

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity_no_nonce_tag'] =  confirm_device_identity_no_nonce_tag


@kms_latency_profile
def confirm_device_identity_stale_nonce_tag(device, connection):

    connection, device, nonce_tag_1, requid_response = setup_confirm_device_identity(device, connection)

    #
    # get a nonce from the Micron KMS Proxy again
    #
    the_nonce, nonce_tag_2, status, mpn = get_nonce_from_kms(connection, MPN_CONFIRM_DEVICE_IDENTITY)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag_2))
            return 'get_nonce_from_kms', the_nonce, nonce_tag_2, status, MPN_CONFIRM_DEVICE_IDENTITY
    if nonce_tag_2 is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag_2, status, MPN_CONFIRM_DEVICE_IDENTITY

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag_2, status, MPN_CONFIRM_DEVICE_IDENTITY
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag_1, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity_stale_nonce_tag'] =  confirm_device_identity_stale_nonce_tag

@kms_latency_profile
def confirm_device_identity_invalid_nonce_tag(device, connection):

    connection, device, nonce_tag, requid_response = setup_confirm_device_identity(device, connection)

    nonce_tag = bytes.fromhex(nonce_tag)

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity_invalid_nonce_tag'] =  confirm_device_identity_invalid_nonce_tag


@kms_latency_profile
def confirm_device_identity_signature_failure(device, connection):

    connection, device, nonce_tag, requid_response = setup_confirm_device_identity(device, connection)

    requid_response = requid_response[:32] + b'w\xb7\xa4\x83\xda\xae0J\xcb\xa4\x96\x83\xa0\xe3\xcboE\xe1\x87\xdbWQ\x07d2\x84\xa49\x16\x06\x1a\xe7'

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity_signature_failure'] =  confirm_device_identity_signature_failure

@kms_latency_profile
def confirm_device_identity_signature_missing(device, connection):

    connection, device, nonce_tag, requid_response = setup_confirm_device_identity(device, connection)

    requid_response = requid_response[:32]

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity_signature_missing'] =  confirm_device_identity_signature_missing
