#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import logging

from hashlib import sha256

from kms_load_test_config import *
from kms_latency_profile_decorator import kms_latency_profile
from kms_load_test_utils import *

import app_common_logging as app_logging

# --- Authenta imports ---
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI


VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
'''

EPILOG = '''

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2020

'''

app_logging.setup_app_logging()
logger = logging.getLogger(__name__)


@kms_latency_profile
def kms_replace_srk(device, connection):
    #
    # get a nonce from the Micron KMS Proxy
    #
    the_nonce, nonce_tag, status, mpn = get_nonce_from_kms(connection, MPN_REPLACE_SRK)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REPLACE_SRK
    if nonce_tag is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REPLACE_SRK

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REPLACE_SRK
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    #
    # call Request_UID command on attached device using the nonce
    # retrieved from the Micron KMS Proxy
    #
    err, requid_response, resp_nonce = device.meta_request_uid(b'\x00', nonce_bytes)
    if not err.esr_success:
        logger.error('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'meta_request_uid', MPN_REPLACE_SRK

    logger.info('Device UID: {}'.format(device.uid_as_str))
    logger.info('Naxos Revision: {}'.format(device.revision.hex()))

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_REPLACE_SRK, False)

    if rc == 0:
        return 'verify_request_uid_from_kms', False, err_msg, http_status, MPN_REPLACE_SRK


    # check if the device is provisioned - if not provision it
    #
    #   (Note, if it is provisioned we assume it is remotely provisioned.
    #   If it is possible that the device is locally provisioned, that
    #   should be checked and if locally provisioned, the function should
    #   exit here or deprovision and reprovision remotely.)

    if not device.provision_sts_struct.Provisioned:
        logger.info("Device is not provisioned. Provisioning...")

        # get the uid because it will be needed to verify the deprovision response, 
        # but device data will be cleared internally after successful deprovision
        uid = device.uid_as_str; 
        err, prov_response = device.meta_provision_crtm(b'\x05', nonce=nonce_bytes, provision_signature_func=kms_sign_command_function)
        if not err.esr_success:
            logger.critical('deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
            return 'provision_crtm', False, err.error_msg, 0, MPN_REPLACE_SRK

        logger.info("need to verify the provision command signature")
        verified, rc_msg, status, mpn =  verify_response_signature("verifyProvisionCRTMResponse", prov_response.hex(), connection, MPN_REPLACE_SRK, uid)
        if verified is None:
            logger.critical('KMS verifyProvisionCRTMResponse Failed: {} \n'.format(rc_msg))
            return 'verify_response_signature', False, rc_msg, status, MPN_REPLACE_SRK
        elif(verified):
            logger.info("successfully verified provision_crtm response with the KMS Proxy")
        else:
            logger.info("Signature verification using the KMS Proxy of provision_crtm response Failed!")
    else:
        # provision will supply the PowerOnAMTC0 (PowerOnCount) to the KMS since it is reset with
        # provision and returned with the provision response.
        # If provision is not needed, do a register read of the registers that include the 
        # PowerOnAMTC value and verify with KMS. This to supply the KMS with this value.
        # 
        # The KMS needs the PowerOnCount to derive the session key.
        logger.info("reading register 0xD0 in order to fetch the PowerOnAMTC0 from the Naxos device\n")
        err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
        if not err.esr_success:
            logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            return 'crtm_register_read', False, err.error_msg, status, MPN_REPLACE_SRK

        print("Verifying CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state\n")

        verified, rc_msg, status, mpn =  verify_response_signature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN_REPLACE_SRK, device.uid_as_str)
        if verified is None:
            logger.critical('KMS verifyProvisionCRTMResponse Failed: {} \n'.format(rc_msg))
            return 'verify_response_signature', False, rc_msg, status, MPN_REPLACE_SRK
        elif(verified):
            logger.info("successfully verified crtm_register_read response with the KMS Proxy")
        else:
            logger.info("Signature verification using the KMS Proxy of crtm_register_read response Failed!")

    # 
    # update KTR register
    # 
    # for replace_srk option #0 (replace SRK with new customer SRK) KTR 
    # must be set to hold the new SRK
    #

    # replace the srk with the current srk so it can be done multiple times
    # (TODO: since we are using a mock device this doesn't really matter, 
    # so maybe it is a better test to change the SRK???)
    new_key_bytes = device._srk

    # USING A MOCK DEVICE, SO NO NEED TO ACTUALLY UPDATE THE KTR
    # 
    # Key Transfer Register - 32byte sub-register in CRTM_register
    # used to transfer SRK/UDS keys into Authenta SPI-NOR device.
    # CRTM Register writes are always in 4byte chunks.
    # KTR_IDX_LST is a time ordered list of starting indeces 
    # of those chunks.
    #
    # KTR_IDX_LST = [0xcc,0xc8,0xc4,0xc0,0xbc,0xb8,0xb4,0xb0]

    # config = b'\x00'
    # for reg_num, idx in enumerate(KTR_IDX_LST):
    #     data_blk = new_key_bytes[(reg_num*4):(reg_num+1)*4]

    #     err, _ = device.meta_crtm_register_write(config, idx, data_blk, sign=False)
    #     if not err.esr_success:
    #         print("\n\n")
    #         logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
    #         exit(1)

    print('Replacing the SRK on the device...')
    # if device.set_kms_connection is not None:
    #    device.set_kms_connection(connection)
    # if device.set_mpn is not None:
    #     device.set_mpn(MPN_RECOVER_DEVICE)

    device.set_kms_connection(connection)
    device.set_kms_mpn(MPN_REPLACE_SRK)

    new_key_hash = sha256(new_key_bytes).digest()
    err, repl_srk_resp, sent_nonce = device.meta_replace_srk(b"\x00", new_key_hash, nonce=nonce_bytes, replace_srk_signature_func=kms_sign_command_function)
    if not err.esr_success:
        logger.critical(f'replace_srk: {err.esr_bit_msg}; {err.error_msg}')
        return 'replace_srk', False, err.error_msg, 400, MPN_REPLACE_SRK

    logger.info("The SRK has been replaced with the key with handle " + new_key_bytes.hex())


    print("need to verify the replace_srk command signature")
    verified, rc_msg, status, mpn = verify_response_signature("verifyReplaceSRKResponse", repl_srk_resp.hex(), connection, MPN_REPLACE_SRK, device.uid_as_str)
    if verified is None:
        logger.info('KMS verifyReplaceSRKResponse Failed: {} \n'.format(rc_msg))

    if(verified):
        logger.info("Signature verification using the KMS Proxy of replace_srk succeeded since SRK was replaced with an identical SRK (typically it should fail response FAILED since verification should be with new SRK and pwrOnCount)")
        msg = "OK"
        # logger.info("ENEXPECTED: successfully verified replace_srk response with the KMS Proxy! This is unexpected since the SRK verification must be with the NEW SRK!")
        # msg = "replace_srk signature verification UNEXPECTEDLY succeeded on KMS; Verification SHOULD be with NEW SRK and pwrOnCount!"
    else:
        logger.info("Signature verification using the KMS Proxy of replace_srk response FAILED which is typically okay now should have passed since the SRK was replaced with the identical SRK")
        msg = "Signature verification using the KMS Proxy of replace_srk response FAILED which is typically okay now should have passed since the SRK was replaced with the identical SRK"
        # msg = "OK"
        # logger.info("Signature verification using the KMS Proxy of replace_srk response FAILED as expected (since verification should be with new SRK and pwrOnCount)")

    return 'verify_response_signature', verified, msg, status, MPN_REPLACE_SRK

func_lookup['kms_replace_srk'] =  kms_replace_srk

@kms_latency_profile
def recover_device(device, connection):
    #
    # get a nonce from the Micron KMS Proxy
    #
    the_nonce, nonce_tag, status, mpn = get_nonce_from_kms(connection, MPN_RECOVER_DEVICE)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_RECOVER_DEVICE
    if nonce_tag is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_RECOVER_DEVICE

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_RECOVER_DEVICE
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    #
    # call Request_UID command on attached device using the nonce
    # retrieved from the Micron KMS Proxy
    #
    #device = NaxosDevice(platform)
    err, requid_response, resp_nonce = device.meta_request_uid(b'\x00', nonce_bytes)
    if not err.esr_success:
        logger.error('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'meta_request_uid', MPN_RECOVER_DEVICE

    logger.info('Device UID: {}'.format(device.uid_as_str))
    logger.info('Naxos Revision: {}'.format(device.revision.hex()))

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_RECOVER_DEVICE, False)

    if rc == 0:
        return 'verify_request_uid_from_kms', False, err_msg, http_status, MPN_RECOVER_DEVICE

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    print("reading register 0xD0 in order to fetch the PowerOnAMTC0 from the Naxos device\n")
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    if not err.esr_success:
      print("\n\n")
      logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
      return 'crtm_register_read', False, err.error_msg, 400, MPN_RECOVER_DEVICE

    print("Verifying CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state\n")
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN_RECOVER_DEVICE, device.uid_as_str)
    if verified is None:
      print('KMS verifyCRTMRegisterRWResponse Failed: {}; {} \n'.format(http_status, rc_msg))
      return 'verify_response_signature', False, rc_msg, http_status, MPN_RECOVER_DEVICE

    print('Deprovisioning device...')
    # if device.set_kms_connection is not None:
    #    device.set_kms_connection(connection)
    # if device.set_mpn is not None:
    #     device.set_mpn(MPN_RECOVER_DEVICE)

    device.set_kms_connection(connection)
    device.set_kms_mpn(MPN_RECOVER_DEVICE)

    err, deprov_response = device.meta_deprovision_crtm(b'\x00', None, deprovision_signature_func=kms_sign_command_function)
    if not err.esr_success:
        logger.critical('deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'deprovision', False, err.error_msg, 400, MPN_RECOVER_DEVICE

    print("need to verify the deprovision command signature")
    verified, rc_msg, status, mpn =  verify_response_signature("verifyGenericACRTMResponse", deprov_response.hex(), connection, MPN_RECOVER_DEVICE, device.uid_as_str)
    if verified is None:
        print('KMS verifyGenericACRTMResponse Failed: {} \n'.format(rc_msg))

    if(verified):
      print("\nsuccessfully verified deprovision_crtm response with the KMS Proxy\n")
      msg = "OK"
    else:
      msg = "signature verification failed on KMS"
      print("\nSignature verification using the KMS Proxy of deprovision_crtm response FAILED\n")

    return 'verify_response_signature', verified, msg, status, MPN_RECOVER_DEVICE

func_lookup['recover_device'] =  recover_device

@kms_latency_profile
def confirm_device_identity(device, connection):
    #
    # get a nonce from the Micron KMS Proxy
    #
    the_nonce, nonce_tag, status, mpn = get_nonce_from_kms(connection, MPN_CONFIRM_DEVICE_IDENTITY)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY
    if nonce_tag is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_CONFIRM_DEVICE_IDENTITY
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    #
    # call Request_UID command on attached device using the nonce
    # retrieved from the Micron KMS Proxy
    #
    err, requid_response, resp_nonce = device.meta_request_uid(b'\x00', nonce_bytes)
    if not err.esr_success:
        logger.error('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'meta_request_uid', MPN_CONFIRM_DEVICE_IDENTITY

    logger.info('Device UID: {}'.format(device.uid_as_str))
    logger.info('Naxos Revision: {}'.format(device.revision.hex()))

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, requid_response, MPN_CONFIRM_DEVICE_IDENTITY, False)

    return 'verify_request_uid_from_kms', rc != 0, err_msg, http_status, MPN_CONFIRM_DEVICE_IDENTITY

func_lookup['confirm_device_identity'] =  confirm_device_identity

@kms_latency_profile
def fetch_key_from_kms(device, connection):
    """
    Fetch an SRK from Micron KMS for the attached device
    and store it in the local key repository, as follows...
    1.	get a nonce,
    2.	connect to input Naxos device
    3.  call Request_UID passing the KMS produced nonce
    4.  generate (or get previously generated) RSA key
    5.	uses the Request_UID response and our generated public key to fetch the SRK from the KMS,
    6.	uses the matching private key to decrypt the returned SRK, and

    """

    #
    # get a nonce from the Micron KMS Proxy
    #
    the_nonce, nonce_tag, status, mpn = get_nonce_from_kms(connection, MPN_REQUEST_SRK)
    if the_nonce is None:
            # if nonce is None, error message will be in nonce_tag
            logger.warning("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REQUEST_SRK
    if nonce_tag is None:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            logger.warning("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REQUEST_SRK

    logger.info('nonce received from KMS: {}'.format(the_nonce))

    nonce_bytes = bytes.fromhex(the_nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        logger.error('Nonce retrieved from KMS is too large to use in Naxos Request_UID command!')
        return 'get_nonce_from_kms', the_nonce, nonce_tag, status, MPN_REQUEST_SRK
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    #
    # call Request_UID command on attached device using the nonce
    # retrieved from the Micron KMS Proxy
    #
    #device = NaxosDevice(platform)
    err, response, resp_nonce = device.meta_request_uid(b'\x00', nonce_bytes)
    if not err.esr_success:
        logger.error('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 'meta_request_uid', MPN_REQUEST_SRK

    logger.info('Device UID: {}'.format(device.uid_as_str))
    logger.info('Naxos Revision: {}'.format(device.revision.hex()))

    # use the KMS to confirm that the device is the real device
    rc, err_msg, http_status, mpn = verify_request_uid_from_kms(connection, device, nonce_tag, response, MPN_REQUEST_SRK, False)
    if rc == 0:
        return 'verify_request_uid_from_kms', False, err_msg, http_status, MPN_REQUEST_SRK

    #
    # read the public RSA encryption key from file. This key will be sent to the KMS and
    # the KMS will encrypt the SRK with this key.
    #
    pubkeydata=[]
    try:
        logger.info("opening and reading public key file ./tmp/kms_rsa_encryption_key.pub")
        fp=open("./tmp/kms_rsa_encryption_key.pub", "r")
        for i in fp:
            pubkeydata.append(i.rstrip("\n"))
    except FileNotFoundError as err:
        logger.error("Missing public key file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
        logger.error("Please generate a public/private key pair for the KMS server to encrypt the SRK it sends back by running generate_rsa_encrypt_keys.py")
        return 'encryption_lib', "FileNotFoundError", MPN_REQUEST_SRK
    except OSError as err:
        logger.error("Failed to read public key from file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
        return 'encryption_lib', "OSError", MPN_REQUEST_SRK

    #
    # Read UID from KMS (passing the UID of the attached Naxos
    # device and the response buffer of the call to Request_UID)
    #
    transportKey_enc, iv, tag, SRK_enc, err_msg, status, mpn = get_key(connection, device, the_nonce, ''.join(pubkeydata[1:-1]))

    #
    # if an SRK was returned save it in the local key store
    # associated with the UID of the attached device
    #
    if transportKey_enc is not None:
        logger.info('retrieved SRK for device 0x{} from KMS.'.format(device.uid_as_str))
    else:
        logger.warning('Failed to retrieve SRK for device {} from KMS! {}'.format(device.uid_as_str, err_msg))
 
    return 'get_key', transportKey_enc, iv, tag, SRK_enc, err_msg, status, MPN_REQUEST_SRK

func_lookup['fetch_key_from_kms'] = fetch_key_from_kms
