from collections import namedtuple
import hmac
import sys
import os
import logging

from hashlib import sha256
from random import _urandom as nonce
import authenta.naxos.utils as sdk2_utils
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

#from .defs_h import *
from authenta.naxos.cmd_error_code_1_5 import NaxosCmdErrorCode_1_5


class MockNaxosDevice:
    """
        Tool for testing KMS
        This class will implement a limited subset of NaxosDevice functions
        that are used in KMS testing.  It will be used to simulate a NaxosDevice
        but with canned responses that should be acceptable to the KMS.

        The constructor will take the UID of the device to simulate
        and the SRK to be used in "signing" the commands.

        Implemented functions will be:
            meta_request_uid
    """
    PROVISION_STS_STRUCT = namedtuple('PROVISION_STS_STRUCT',
        ["ID_Key_Select", "UDS_Replaced", "RIOT_Mode", "TPM_Supported",
        "Remotely_Provisioned", "Check_Engine", "SRK_Replaced",
        "Provisioned"])

    _cmd_error_code = NaxosCmdErrorCode_1_5(b'\x80', '', 'OK')

    def __init__(self, uid: str, srk: bytes, amtc0: int, kms_connection=None, logger=logging.getLogger(__name__)):
        self.logger = logger
        self._uid = uid
        self._srk = srk
        self._amtc0 = amtc0 if amtc0 is not None else 1
        self._boot_amtc0 = self._amtc0
        self._amtc1 = self._amtc0
        self._revision = b'\x01\x00\x05'
        self.set_to_provisioned()
        self._session_key = None
        self._generate_session_key()
        self._kms_connection = kms_connection
        self._kms_function_name = ''
        self._kms_mpn = None

    @property
    def uid(self):
        return bytes.fromhex(self._uid) if self._uid is not None else None

    @property
    def revision(self):
        return self._revision

    @property
    def uid_as_str(self):
        return self._uid

    @property
    def amtc0(self):
        return self._amtc0

    @property
    def boot_amtc0(self):
        return self._boot_amtc0

    @property
    def amtc1(self):
        return self._amtc1

    @property
    def provision_sts_struct(self):
        return self._provision_sts_struct

    def set_kms_connection(self, kms_connection):
        self._kms_connection = kms_connection

    def set_kms_mpn(self, mpn):
        """
        This must be set before calling kms_sign_command_function
        """
        self._kms_mpn = mpn

    def set_to_provisioned(self):
        self._provision_sts_struct = self.PROVISION_STS_STRUCT( 0, 0, 0, 0, 1, 0, 1, 1 )

    def set_to_not_provisioned(self):
        self._provision_sts_struct = self.PROVISION_STS_STRUCT( 0, 0, 0, 0, 0, 0, 1, 0 )

    def sign_response_with_srk(self, response_base: bytes)->bytes:            
        myhmac = hmac.new(self._srk, digestmod='sha256')
        myhmac.update(response_base)
        return myhmac.digest()

    def _generate_session_key(self):
        myhmac = hmac.new(self._srk, digestmod='sha256')
        myhmac.update(self._boot_amtc0.to_bytes(4, byteorder="big"))
        self._session_key = myhmac.digest()

    def sign_response_with_session_key(self, response_base: bytes)->bytes:
        if self._session_key is None:
            self._generate_session_key()
        myhmac = hmac.new(self._session_key, digestmod='sha256')
        myhmac.update(response_base)
        return myhmac.digest()

    def kms_sign_command_function(self, device, command: bytes, additional_data: bytes=None):
        """
            This function will send an input command to the Micron KMS Proxy to be signed.
            
            NOTE:
            For this to work the kms_mpn, and _kms_function_name MUST be set before calling this function.
            Also, kms_connection MUST be set, but it is set in the constructor.

            Arguments:
                device A mock (or real) "device" that has the uid_as_str property
                command - the command string to be signed.

            Return:
                the signature as a bytes object
        """
        if self._kms_connection is None:
            self.logger.error("Call to sign command without setting connection to KMS proxy!")
            return b'', "Internal Load-Test Error", 0

        add_data_str = ""
        if additional_data is not None and len(additional_data) > 0:
            add_data_str = additional_data.hex()
        
        sig, rc_msg, status = MicronKMSProxyAPI.GenerateCmdSignature(self._kms_function_name, \
            command.hex(), add_data_str, self._kms_connection, self._kms_mpn, device.uid_as_str)
        if sig is not None:
            sigbytes = bytes.fromhex(sig)
        else:
            self.logger.warning("ERROR: Failed to fetch signature from the KMS Proxy! {}".format(rc_msg))
            sigbytes = b''

        return sigbytes, rc_msg, status, self._kms_mpn


    def meta_request_uid(self, config: bytes=b'\x00', nonce_in: bytes=None):
        if nonce_in is None:
            return NaxosCmdErrorCode_1_5(b'\x00', 'request_uid', non_authenta_err_msg='missing nonce!'), None, None

        # ESR[7:0] | Reserved[15:0] | Provision_Sts[7:0] | nonce[31:0] | A-MTC0[31:0] | Revision[23:0] | DupSig[255:248] | UID[127:0]


        self._amtc0 += 1
        resp = b'\x80' + b'\x00\x00' + b'\x6b' + nonce_in + self.amtc0.to_bytes(4, byteorder="big") + self.revision + b'\x00' + sdk2_utils.change_word_order(self.uid)
        if self._provision_sts_struct.ID_Key_Select == 0:
            # sign with SRK
            sig = self.sign_response_with_srk(resp)
        else:
            sig = self.sign_response_with_session_key(resp)

        resp = resp + sdk2_utils.change_word_order(sig)
        return NaxosCmdErrorCode_1_5(b'\x80', 'request_uid', 'OK'), resp, nonce_in

    def meta_provision_crtm(self, config: bytes, nonce: bytes, amtc0_in: bytes=None, provision_signature_func=None):
        """
            Produce a mock signed response to a Provision_CRTM command on an Authenta device.
            If provision_signature_func argument to this call is set, sign a provision_crtm command
            using that function.
            If provision_signature_func is set, the kms_mpn value MIST be set on this (self) object
            before this call! If not set, error will be returned.

            Arguments:
                config (bytes)
                a_nonce (bytes): A 4-byte nonce to use in the Provision_CRTM command and the
                                 response signature.
                amtc0_in (int): IGNORED (This arg is not intended to be set and will be ignored)

            Returns:
                NaxosCmdErrorCode, bytes (response - may equal None on failure), nonce (input nonce)
        """
 
        if nonce is None:
            return NaxosCmdErrorCode_1_5(b'\x00', 'deprovision_crtm', non_authenta_err_msg='missing nonce!'), None
        a_nonce = nonce

        if amtc0_in is not None:
            the_amtc0 = amtc0_in
        else:
            the_amtc0 = self.amtc0

        if provision_signature_func is not None:
            if self._kms_mpn is None:
                self.logger.error("Cannot sign using KMS Proxy - no MPN set!")
                return NaxosCmdErrorCode_1_5(b'\x00', 'provision_crtm', non_authenta_err_msg='Cannot sign using KMS Proxy - no MPN set!'), None

            self._kms_function_name = "signProvisionCRTMCommand"
            # format {opcode}{sub_opcode}{config}{amtc_1}
            cmd_str = bytes.fromhex('9b35') + config + a_nonce + the_amtc0.to_bytes(4, byteorder="big")
            cmd_sig, *_ = provision_signature_func(self, cmd_str)
            self.logger.info("provision cmd signed w/ kms: {}{}".format(cmd_str.hex(), cmd_sig.hex()))

            if cmd_sig is None or len(cmd_sig) == 0:
                # signing failed, so return bad signature
                return NaxosCmdErrorCode_1_5(b'\x84', 'provision_crtm', 'Bad signature!'), None

        if self._provision_sts_struct.Provisioned:
            # already provisioned, return error
            return NaxosCmdErrorCode_1_5(b'\x82', 'provision_crtm', 'Device is already provisoned'), None

        self.set_to_provisioned()

        # ESR[7:0] |  Extended_CRTM_Status[15:0] | DupSig[255:248] | nonce[31:0] | A-MTC0[31:0] | A-MTC1[31:0] | UID[127:0] 
        self._amtc0 += 1
        resp = b'\x80' + b'\x09\x07' + b'\x00' + a_nonce + self.amtc0.to_bytes(4, byteorder="big") + self.amtc1.to_bytes(4, byteorder="big")  + sdk2_utils.change_word_order(self.uid)
        sig = self.sign_response_with_session_key(resp)
        resp = resp + sdk2_utils.change_word_order(sig)
        return NaxosCmdErrorCode_1_5(b'\x80', 'provision_crtm', 'OK'), resp, a_nonce


    def meta_deprovision_crtm(self, config: bytes, amtc1_in: bytes=None, deprovision_signature_func=None):
        """ Deprovision a previously provisioned Authenta device

            Args:
            config (bytes)
            amtc1_in (int): IGNORED

            Returns:
            NaxosCmdErrorCode, bytes (response - may equal None on failure)
        """

        if deprovision_signature_func is not None:
            if self._kms_mpn is None:
                self.logger.error("Cannot sign using KMS Proxy - no MPN set!")
                return NaxosCmdErrorCode_1_5(b'\x00', 'deprovision_crtm', non_authenta_err_msg='Cannot sign using KMS Proxy - no MPN set!'), None

            self._kms_function_name = "signDeprovisionCRTMCommand"
            # format {opcode}{sub_opcode}{config}{amtc_1}
            cmd_str = bytes.fromhex('9b36') + config + self._amtc1.to_bytes(4, byteorder="big")
            cmd_sig, *_ = deprovision_signature_func(self, cmd_str)
            self.logger.info("deprovision cmd signed w/ kms: {}{}".format(cmd_str.hex(), cmd_sig.hex()))

            if cmd_sig is None or len(cmd_sig) == 0:
                # signing failed, so return bad signature
                return NaxosCmdErrorCode_1_5(b'\x84', 'deprovision_crtm', 'Bad signature!'), None

        self.set_to_not_provisioned()

        # ESR[7:0] |  Extended_CRTM_Status[15:0] | Reserved[0:31] | DupSig[255:248] | A-MTC0[31:0] | A-MTC1[31:0] | Reserved[127:0] 
        self._amtc0 += 1
        resp = b'\x80' + b'\x09\x07' + b'\x00\x00\x00\x00\x00' + self.amtc0.to_bytes(4, byteorder="big") + self.amtc0.to_bytes(4, byteorder="big")  + (b'\x00'*16)
        sig = self.sign_response_with_session_key(resp)
        resp = resp + sdk2_utils.change_word_order(sig)
        return NaxosCmdErrorCode_1_5(b'\x80', 'deprovision_crtm', 'OK'), resp

    def meta_replace_srk(self, config: bytes, additional_data: bytes=None, nonce: bytes=None, amtc0_in: bytes=None, replace_srk_signature_func=None):
        """ Issue Replace_SRK command and get command repsonse

        Args:
          config (bytes)
          additional_data (bytes): hash of the key(s) required based on 
            replace_srk option (for options #0 and #1, hash should include the 
            new SRK key. for option #2, hash should include the new and old SRK 
            keys)
          nonce (bytes): Optional A 4-byte nonce to use in the Request_UID
            command sent sent to the Naxos device. If not provided, one will
            be randomly generated and returned.
          param amtc0_in (int): ONLY FOR DEBUG
        Returns:
          NaxosCmdErrorCode
          response (bytes): may equal None on failure
          nonce (4-byte bytes): nonce sent with the command
        """
        if nonce is None:
            return NaxosCmdErrorCode_1_5(b'\x00', 'replace_srk', non_authenta_err_msg='missing nonce!'), None
        a_nonce = nonce

        if replace_srk_signature_func is not None:
            if self._kms_mpn is None:
                self.logger.error("Cannot sign using KMS Proxy - no MPN set!")
                return NaxosCmdErrorCode_1_5(b'\x00', 'replace_srk', non_authenta_err_msg='Cannot sign using KMS Proxy - no MPN set!'), None, a_nonce

            self._kms_function_name = "signReplaceSRKCommand"
            # format {opcode[1]}{sub_opcode[1]}{config[2]}{amtc_0[4]}{reserved[1]}{nonce[4]}
            cmd_str = bytes.fromhex('9b39') + config + self._amtc0.to_bytes(4, byteorder="big") + b'\x00' + a_nonce

            cmd_sig, *_ = replace_srk_signature_func(self, cmd_str, additional_data)
            self.logger.info("replace_srk cmd signed w/ kms: {}{}".format(cmd_str.hex(), cmd_sig.hex()))

            if cmd_sig is None or len(cmd_sig) == 0:
                # signing failed, so return bad signature
                return NaxosCmdErrorCode_1_5(b'\x84', 'replace_srk', 'Bad signature!'), None, a_nonce

        # ESR[7:0] |  Extended_CRTM_Status[15:0] | DupSig[255:248] | nonce[31:0] | A-MTC0[31:0] | A-MTC1[31:0] | UID[127:0] 
        self._amtc0 += 1
        self._boot_amtc0 = self.amtc0
        self._session_key = None
        resp = b'\x80' + b'\x09\x07' + b'\x00' + a_nonce + self.amtc0.to_bytes(4, byteorder="big") + self.amtc1.to_bytes(4, byteorder="big")  + sdk2_utils.change_word_order(self.uid)
        sig = self.sign_response_with_session_key(resp)
        resp = resp + sdk2_utils.change_word_order(sig)
        return NaxosCmdErrorCode_1_5(b'\x80', 'replace_srk', 'OK'), resp, a_nonce

    def meta_crtm_register_read(self, config: bytes, crtm_addr: int, amtc0_in: bytes=None):
        """ Issue Replace_SRK command and get command repsonse

        Args:
          config (bytes)
          crtm_addr (int)
          amtc0_in (int): IGNORED
        Returns:
          NaxosCmdErrorCode
          response (bytes): may equal None on failure
        """

        # ESR[7:0] |  Extended_CRTM_Status[15:0] | rsvd[7:0] | crtm_register_address[15:0] | 
        #     rsvd[7:0] | DupSig[255:248] | A-MTC0[31:0] | A-MTC1[31:0] | CRTM_Register_value[127:0] 
        #           (FOR addr 0xD0 reg_val -> EEPROM_attr[32b] | PwrOnCount[32b] | device_config_vol[32b] | rsvd[32b])
        d0_reigister_val = (b'\x33'*4) + self.boot_amtc0.to_bytes(4, byteorder='big') + (b'\00'*8)
        self._amtc0 += 1
        resp = b'\x80' + b'\x09\x07' + b'\x00' + crtm_addr.to_bytes(2, byteorder="big") + b'\x00\x00' + self.amtc0.to_bytes(4, byteorder="big") + self.amtc1.to_bytes(4, byteorder="big")  + d0_reigister_val
        sig = self.sign_response_with_session_key(resp)
        resp = resp + sdk2_utils.change_word_order(sig)
        return NaxosCmdErrorCode_1_5(b'\x80', 'crtm_register_read', 'OK'), resp
