"""
Common functions used accross the KMS tests.
(C) 20 Micron Technology, Inc. All rights reserved.
"""
import os
import json
import logging

import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
import app_common_logging as app_logging

from hashlib import sha256
from authenta.naxos.device import NaxosDevice
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.grpc_platform import GrpcPlatform
import authenta.authenta_key_mgmt_interface as key_mgmt

device_json_file_name = 'devices.json'
ep_alias = 'snor1'
sig_connection = None

app_logging.setup_app_logging(False)
logger = logging.getLogger(__name__)

def get_device():
  """
  Create NaxosDevice object
  """

  with open(device_json_file_name) as f:
    json_data = json.load(f)

    json_data = json_data[ep_alias]

    ip_params = json_data['endpoint_ip'].split(':')
    ep_ip = ip_params[0]
    ep_port = None
    if len(ip_params) > 1:
      ep_port = ip_params[1]

    dev_path = json_data['endpoint_fn']
  
  if ep_port is None:
    ep = GrpcPlatform(dev_path, ep_ip)
  else:
    ep = GrpcPlatform(dev_path, ep_ip, ep_port)

  device = NaxosDevice(ep)

  return device


def get_request_uid_response(device, nonce: bytes):
  """
    Make a device.meta_request_uid call outside a 'with' statement.
    I.e. if necessary, the platform.acquire_device() is done here.
  """
  try:
    acquired_device_here = False
    if not device.endpoint_platform.is_device_acquired():
      device.endpoint_platform.acquire_device(device.device_type)
      acquired_device_here = True

    err, response, *_ = device.meta_request_uid(nonce=nonce)
    return err, response
  finally:
    if acquired_device_here:
      device.endpoint_platform.release_device(device.device_type)

def get_nonce_from_kms_proxy(connection, mpn):
  """
  Get a nonce from the Micron KMS Proxy
  """

  nonce, nonce_tag, status = MicronKMSProxyAPI.getNonceWithMPN(connection, mpn)

  if nonce is None or nonce_tag is None:
    if nonce is None:
      # if nonce is None, response will be in nonce_tag
      print("Failed to receive nonce from KMS: {}".format(nonce_tag))
      print("Status code: {}\n".format(status))
    else:
      # something is wrong with what was returned by the KMS
      # this should not happen (nonce was returned from KMS and nonce_tag is None)
      print("Failed to receive valid nonce from KMS: missing nonce_tag\n")
      print("Status code: {}\n".format(status))

    return nonce, nonce_tag

  nonce_bytes = bytes.fromhex(nonce)
  nonce_bytes_len = len(nonce_bytes)
  if nonce_bytes_len > 4:
      print('Nonce retrieved from KMS is large to use in Naxos Request_UID command!')
      return None, None
  padding = 4 - nonce_bytes_len
  if padding > 0:
      nonce_bytes += '\x00' * padding

  return nonce_bytes, nonce_tag


def get_key(UID, Connection, nonce, MPN, keyType, encryptionKey):

  # Defining parts of the HTTP request
  KMSProxyServerBaseAddress = '137.201.240.75'
  KMSProxyServerBaseURL = 'https://' + KMSProxyServerBaseAddress
  request_url = KMSProxyServerBaseURL + '/GetKey'
    
  request_headers = {
    'Content-Type': 'application/json'
  }
  request_body_dict={
    'mpn': MPN,
    'UID': UID,
    'keyType': keyType,
    'nonce': nonce,
    'wrapProtocol': "rsaEncryptedTransportKey",
    'manufacturer': 'Micron'
  }

  data=[]
  if (encryptionKey is not None):
    try:
      fp=open(encryptionKey, "r")
      for i in fp:
        data.append(i.rstrip("\n"))
    except TypeError as err:
      return None
    except FileNotFoundError as err:
      logger.error("Missing public key file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
      logger.error("Please generate a public/private key pair for the KMS server to encrypt the SRK it sends back by running generate_rsa_encrypt_keys.py")
      return None
    except OSError as err:
      logger.error("Failed to read public key from file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
      return None

  if (encryptionKey == "./tmp/illegally_formatted_encryption_key.pub"):
      request_body_dict['wrapKey']=b'\x00'
  else:
      request_body_dict['wrapKey']=''.join(data[1:-1])

  response= MicronKMSProxyAPI.issue_post(Connection, request_url, request_headers, request_body_dict)
  if response is None:
    logger.error("Post to GetKey on KMS Proxy failed!")
    return None, None, None, None, "Post to GetKey on KMS Proxy failed!", 0

  responsedata = response.read()
  if responsedata is not None and len(responsedata) > 0:
      try:
          received = json.loads(if_bytes_make_str(responsedata))
      except json.decoder.JSONDecodeError as x:
          logger.error('Failed to parse response as JSON! Status = {}; code = {}; reason = {}'.format(response.status, response.code, response.reason))
          logger.error('Response:  {}'.format(responsedata))
          return None, None, None, None, 'KMS Returned status {}; Expected JSON in GetKey response instead received: {}'.format(response.status, responsedata). response.status

      logger.info(received)

  else:
      logger.error("Post to GetKey on KMS Proxy returned {} with empty message body!".format(response.status))
      return None, None, None, None, "GetKey on KMS Proxy returned {} with empty message body!".format(response.status), response.status

  if response.status < 200 or response.status >= 300:
      logger.error('Response: status = {}; body = {}'.format(response.status, received))
      return None, None, None, None, 'getKey Response: status = {}; body = {}'.format(response.status, received), response.status
  else:
      logger.info('Response: status = {}; {}'.format(response.status, received))

  return received['encTransportKey'], received['iv'], received['tag'], received['ciphertext'], "OK", response.status

def VerifyRequestUIDResponse(UID, UIDData, Connection, noncetag, MPN, mtcPowerOnValue, request_url, op):
###################### Issue VerifyRequestUIDResponse #############################
  #start = timeit.timeit()
  # Defining parts of the HTTP request
  
  request_headers = {
    'Content-Type': 'application/json'
  }
  request_body_dict={
    'mpn': MPN,
    'UID': UID,
    'op':  op,
    'nonceTag': noncetag,
    'data': UIDData,
    'mtcPowerOnValue': mtcPowerOnValue,
    'manufacturer': 'Micron'
  }

  response = MicronKMSProxyAPI.issue_post(Connection, request_url, request_headers, request_body_dict)
  if response is None:
    logger.error("Post to verifyRequestUIDResponse on KMS Proxy failed!")
    return 0, "Post to verifyRequestUIDResponse on KMS Proxy failed!", 0
  data = response.read()

  if data is not None and len(data) > 0:
    try:
      received = json.loads(MicronKMSProxyAPI.if_bytes_make_str(data))
    except json.decoder.JSONDecodeError as x:
      logger.error('Failed to parse response as JSON! Status = {}; code = {}; reason = {}'.format(response.status, response.code, response.reason))
      logger.error('Response:  {}'.format(data))
      return 0, 'Expected JSON in getNonce response instead received: {}'.format(data), response.status
  else:
    logger.error("Post to verifyRequestUIDResponse on KMS Proxy returned {} with empty message body!".format(response.status))
    return 0, "Post to verifyRequestUIDResponse on KMS Proxy returned {} with empty message body!".format(response.status), response.status

  if response.status < 200 or response.status >= 300:
    if received is not None:
      logger.error('Response: status = {}; body = {}'.format(response.status, received))
      return 0, 'Response: status = {}; body = {}'.format(response.status, received), response.status
    else:
      logger.error('Response: status = {}; code = {}; reason = {}'.format(response.status, response.code, response.reason))
      return 0, 'Response: status = {}; code = {}; reason = {}'.format(response.status, response.code, response.reason), response.status

  return 1, 'OK', response.status

def get_command_sig(operation: str, mpn: str, device: NaxosDevice, command: bytes, additional_data: str="")->bytes:
    """
    send the input command string to the Micron KMS Proxy to get the command signature
    and return the signature returned from the kms_connection

    This is a generic sign command. It will sign the command specified in operation argument
    to this function.  That string must be one of : “signReplaceSRKCommand,” “signProvisionCRTMCommand,”
    or “signDeprovisionCRTMCommand”

    NOTE For Replace_SRK the additional_data argument must contain the hash of the value enter
    into the KTR register (formatted as a hex string). For other commands additional_data should be then
    empty string.

    NOTE sig_connection must be set to a valid connection to the KMS proxy before this 
    function is called!

    ARGS
        operation   Name of the operation; one of “signReplaceSRKCommand”, “signProvisionCRTMCommand”
                    or “signDeprovisionCRTMCommand”
        mpn         The MPN to use
        device      The Naxos Device instance
        command     THe command string to be signed
        additional_data  For Replace_SRK this must be the hash (SHA256) of the contents of KTR
    
    Return the signature as a bytes
    """
    sig, rc_msg, status = MicronKMSProxyAPI.GenerateCmdSignature(operation, \
        command.hex(), additional_data, sig_connection, mpn, device.uid_as_str)
    if sig is not None:
        sigbytes = bytes.fromhex(sig)
    else:
        print("ERROR: Failed to fetch Deprovision_CRTM signature from the KMS Proxy! {}".format(rc_msg))
        print("Status code: {}\n".format(status))
        sigbytes = b''

    return sigbytes

def get_provision_sig(device: NaxosDevice, command: bytes)->bytes:
    """
    NOTE static_vars.new_srk_hash must be set to the hash (SHA256) of the contents of the KTR_IDX_LST
         before calling this function
    """
    return get_command_sig("signProvisionCRTMCommand", "MTANXSBC01", device, command)

def replace_srk_setup(new_srk_name, connection):

  global sig_connection

  sig_connection = connection
  derive_key = None
  new_srk_hash = None
  command = None

  # get the hash of the new SRK - needed by KMS to sign Replace_SRK
  # new_srk_hash = sha256(bytes.fromhex(new_key)).hexdigest()

  nonce_bytes, nonce_tag = get_nonce_from_kms_proxy(connection, "MTANXSBC01")
  if nonce_bytes is None:
    # if nonce is None, error message will be in nonce_tag
    print("Failed to receive  nonce from KMS: {}".format(nonce_tag))
    exit(1)

  # determine which revision of the Naxos device is connecting
  device = get_device()
  err, response = get_request_uid_response(device, nonce_bytes)
  if not err.esr_success:
    print('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
    exit(1)

  # verify the request UID response from the chip

  # if the device is provisioned this function will not provision the device.
  # in that case the Boot A-MTC0 value will have to be set on the KMS in 
  # VerifyRequestUIDResponse. Therefore we will set it here, but that
  # creates a requirment that this Request_UID be the first command on boot.
  # (On the other hand if it is not provisioned we will provision in this 
  # function and the KMS will get the PowerOnAMTC0 value from verifying the 
  # provison.)
  # if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
  #   useAMTC0asPowerOnAMTC0 = 'true'
  # else:
  useAMTC0asPowerOnAMTC0 = 'false'

  rc_int, err, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, "MTANXSBC01", useAMTC0asPowerOnAMTC0)
  if rc_int == 0:
    print("Verify Request_UID response from KMS failed!")
    print("Error message: {}".format(err))
    print("Status code: {}\n".format(status))
    return 1

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  logger.info('Found attached Naxos device revision {}'.format(device.revision.hex()))

  replace_srk_config = b"\x00"

  if dev_rev == 0x010005:
    if derive_key:
      replace_srk_config = b"\x10"
    device = NaxosDevice_1_5(device.endpoint_platform)
  else:
    logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
    exit(1)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      if not err.esr_success:
        print("\n\n")
        logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 2

    # get the uid because it will be needed in multiple places later
    uid = device.uid_as_str; 

    new_key = key_mgmt.export_priv_key_as_hex(uid, new_srk_name)
    if new_key is None:
      logger.critical(f'\nFailed to load new key from key repository: uid=\'{uid}\'; key_name=\'{new_srk_name}\'\n')
      exit(5)
        
    new_key_bytes = bytes.fromhex(new_key)
    new_srk_hash = sha256(new_key_bytes).hexdigest()

    if device.provision_sts_struct.Provisioned and not device.provision_sts_struct.Remotely_Provisioned:
      print("\nThe device is Locally Provisioned.  Deprovision the device and try again.\n")
      logger.info("The device is Locally Provisioned. Aborting.")
      exit(9)

    if device.provision_sts_struct.Provisioned:
      # do a register read that includes the PowerOnAMTC value and
      # verify with KMS to supply the KMS with this value, (which it
      # needs to derive the session key).
      print("reading register 0xD0 in order to fetch the PowerOnAMTC0 from the Naxos device\n")
      err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
      if not err.esr_success:
        print("\n\n")
        logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 1

      print("Verifying CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state\n")
      verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, "MTANXOBG01", uid)
      if verified is None:
        print('KMS verifyCRTMRegisterRWResponse Failed: {}; {} \n'.format(http_status, rc_msg))
        return 1
    else:
      # note that pwrOnCount will be set with the provision and included in provision response
      print("\nThe device is not provisioned.  Provisioning...\n")
      logger.info("Device is not provisioned. Provisioning...")

      err, prov_response, sent_nonce = device.meta_provision_crtm(b'\x05', nonce=nonce_bytes, provision_signature_func=get_provision_sig)
      if not err.esr_success:
        print("\n\n")
        logger.critical('deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 1

      print("need to verify the provision command signature")
      verified, rc_msg, status = MicronKMSProxyAPI.VerifyResponseSignature("verifyProvisionCRTMResponse", prov_response.hex(), connection, "MTANXSBC01", uid)
      if verified is None:
        print('KMS verifyProvisionCRTMResponse Failed: {} \n'.format(rc_msg))
        print("Error message: {}".format(rc_msg))
        print("Status code: {}\n".format(status))
        return 1

      if(verified):
        print("\nsuccessfully verified provision_crtm response with the KMS Proxy\n")
      else:
        print("\nSignature verification using the KMS Proxy of provision_crtm response Failed!\n")

    # update KTR register
    # 
    # for replace_srk option #0 (replace SRK with new customer SRK) KTR 
    # must be set to hold the new SRK

    # Key Transfer Register - 32byte sub-register in CRTM_register
    # used to transfer SRK/UDS keys into Authenta SPI-NOR device.
    # CRTM Register writes are always in 4byte chunks.
    # KTR_IDX_LST is a time ordered list of starting indexes 
    # of those chunks.
    KTR_IDX_LST = [0xcc,0xc8,0xc4,0xc0,0xbc,0xb8,0xb4,0xb0]

    config = b'\x00'
    for reg_num, idx in enumerate(KTR_IDX_LST):
      data_blk = new_key_bytes[(reg_num*4):(reg_num+1)*4]

      err, reg_write_resp = device.meta_crtm_register_write(config, idx, data_blk, sign=False)
      if not err.esr_success:
        logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
        exit(1)

      # verify the request UID response from the chip
      print("\nverify the register_write to the KTR for SRK tranch {} of 7".format(reg_num))
      verified, rc_msg, status = MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", reg_write_resp.hex(), connection, "MTANXSBC01", uid)
      if verified is None:
        print('KMS verifyCRTMRegisterRWResponse Failed: {} \n'.format(rc_msg))
        print("Status code: {}\n".format(status))
        return 1

      if(verified):
        print("successfully verified crtm_register_write response with the KMS Proxy\n")
      else:
        print("Signature verification using the KMS Proxy of crtm_register_write response Failed!\n")

    new_key_hash = sha256(new_key_bytes).digest()

    command = bytes.fromhex('9b') + bytes.fromhex('39') + replace_srk_config + device.amtc0 + b'\x00' + nonce_bytes

  return command, device, new_srk_hash

def get_amtc1(device):
  _, response = device.meta_request_monotonic_counter(0x81)
  amtc1 = response[17:21]

  return amtc1

def if_bytes_make_str(response_data)->str:
    """
    depending on  the Python version response.read() may return bytes or
    str.  If it is a bytes convert and return as a UTF-8 str.
    """
    if(isinstance(response_data, bytes)):
        response_data = response_data.decode('utf-8')
    
    return response_data