#! /usr/bin/env python3

"""
KMS Test for Confirm Device Identity.
(C) 2020 Micron Technology, Inc. All rights reserved.
"""

import pytest
import conftest
import pytest_common_utils as utils
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

@pytest.fixture(scope="module", autouse=True)
def test_confirm_device_identity_setup(request):
  global connection
  global MPN
  global device
  global KMSProxyServerBaseURL

  MPN = "MTANXOBD01"

  KMSProxyServerBaseAddress = '137.201.240.75'
  KMSProxyServerBaseURL = 'https://' + KMSProxyServerBaseAddress

  device = conftest.get_device(request)
  connection = conftest.create_connection()
  yield


def test_confirm_device_identity_Positive(): #230001
  """
  Inputs to verify_request_uid_response API:
    Op is verifyRequestUIDResponse
    Valid UID
    Valid nonceTag
    Valid responseData from Request_UID
    Valid MTC power-on value

  Expected Outputs:
    Success return code
    Success status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
      useAMTC0asPowerOnAMTC0 = 'true'
    else:
      useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"
    
    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 1)
    assert (status == 200)
 
def test_confirm_device_identity_Negative_BadOp(): #230002
  """
  Inputs to verify_request_uid_response API:
    Op field is invalid

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
      useAMTC0asPowerOnAMTC0 = 'true'
    else:
      useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUID"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 401)


def test_confirm_device_identity_Negative_BadRequestURL():
  """
  Inputs to verify_request_uid_response API:
    Request_url field is invalid

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
      useAMTC0asPowerOnAMTC0 = 'true'
    else:
      useAMTC0asPowerOnAMTC0 = 'false'

    op = 'verifyRequestUIDResponse'
    request_url = KMSProxyServerBaseURL + '/VerifyRequestUID'

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 404)

def test_confirm_device_identity_Negative_UnknownUID(): #230003
  """
  Inputs to verify_request_uid_response API:
    UID is unknown to the KMS

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = "1032547698badcfe1f32547698baaa55"
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_IllegallyFormattedUID(): #230004
  """
  Inputs to verify_request_uid_response API:
    UID is formatted as a non-hexadecimal string

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 0)


def test_confirm_device_identity_Negative_MissingUID(): #230005
  """
  Inputs to verify_request_uid_response API:
    UID field is set to None

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = None
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_MissingNonceTag(): #230006
  """
  Inputs to verify_request_uid_response API:
    nonceTag field is set to None

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    nonce_tag = None

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"
    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_StaleNonceTag(): #230007
  """
  Inputs to verify_request_uid_response API:
    nonceTag from a previously used session

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag_1 = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag_1)

    nonce_bytes, nonce_tag_2 = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag_2)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag_1, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_InvalidNonceTag(): #230008
  """
  Inputs to verify_request_uid_response API:
    nonceTag is formatted as a non-hexadecimal string

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    nonce_tag_bytes = bytes.fromhex(nonce_tag)

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag_bytes, MPN, useAMTC0asPowerOnAMTC0,  request_url, op)
    assert (rc_int == 0)
    assert (status == 0)


def test_confirm_device_identity_Positive_MTCPowerOnTrue():
  """
  Inputs to verify_request_uid_response API:
    MTC-Power-On value is set to True

  Expected Outputs:
    Success return code
    Success status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    useAMTC0asPowerOnAMTC0 = 'true'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 1)
    assert (status == 200)


def test_confirm_device_identity_Positive_MTCPowerOnFalse():
  """
  Inputs to verify_request_uid_response API:
    MTC-Power-On value is set to False

  Expected Outputs:
    Success return code
    Success status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 1)
    assert (status == 200)


def test_confirm_device_identity_Negative_MTCPowerOnInvalid(): #230009
  """
  Inputs to verify_request_uid_response API:
    MTC-Power-On value is not valid

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    useAMTC0asPowerOnAMTC0 = 0

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_SignatureFailure(): #230010
  """
  Inputs to verify_request_uid_response API:
    Signature in signed data is not correct

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    response = response[:32] + b'w\xb7\xa4\x83\xda\xae0J\xcb\xa4\x96\x83\xa0\xe3\xcboE\xe1\x87\xdbWQ\x07d2\x84\xa49\x16\x06\x1a\xe7'
    UID = device.uid_as_str
    UIDData = response.hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)


def test_confirm_device_identity_Negative_SignatureMissing(): #230011
  """
  Inputs to verify_request_uid_response API:
    Signature is missing from data

  Expected Outputs:
    Failure error code
    Failure status code
  """
  with device:

    nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, "MTANXOBD01")
    assert (nonce_bytes)
    assert (nonce_tag)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response[:32].hex()

    if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
        useAMTC0asPowerOnAMTC0 = 'true'
    else:
        useAMTC0asPowerOnAMTC0 = 'false'

    request_url = KMSProxyServerBaseURL + '/VerifyRequestUIDResponse'
    op = "verifyRequestUIDResponse"

    rc_int, _, status = utils.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, useAMTC0asPowerOnAMTC0, request_url, op)
    assert (rc_int == 0)
    assert (status == 400)
