#! /usr/bin/env python3

"""
KMS Test for Recover Device.
(C) 2020 Micron Technology, Inc. All rights reserved.
"""

import pytest
import pytest_common_utils as utils
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
import conftest

from authenta.naxos.device import NaxosDevice
from authenta.naxos.device_1_5 import NaxosDevice_1_5

class static_vars:
  UID = None
  Op = None
  command_data = None
  command_data_bytes = None
  validCommand = False

@pytest.fixture(scope="module", autouse=True)
def test_recover_device_setup(request):
  global device
  global connection
  global MPN
  global mtcPowerOnValue

  MPN = "MTANXOBG01"
  mtcPowerOnValue = 'false'
  connection = conftest.create_connection()
  
  yield

def get_deprovision_crtm_sig(device: NaxosDevice, command: bytes)->bytes:
  if static_vars.validCommand == True:
    cmd_str = command.hex()
  else:
    if static_vars.command_data_bytes is not None:
      cmd_str = static_vars.command_data_bytes
    elif static_vars.command_data is None:
      cmd_str = static_vars.command_data
    elif static_vars.command_data is not None:
      cmd_str = static_vars.command_data.hex()

  sig, rc_msg, status = MicronKMSProxyAPI.GenerateCmdSignature(static_vars.Op, \
    cmd_str, "", connection, MPN, static_vars.UID)
  if sig is not None:
    sigbytes = bytes.fromhex(sig)
  else:
    print("ERROR: Failed to fetch Deprovision_CRTM signature from the KMS Proxy! {}".format(rc_msg))
    print("Status code: {}\n".format(status))
    sigbytes = b''

  return sigbytes


def test_recover_device_Negative_UnknownUID(): #240002
  """
  Inputs to generateSignature:
    UID is unknown to the KMS

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = "1032547698badcfe1f32547698baaa55"
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified is None)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    config = b'\x00'

    static_vars.command_data_bytes = None
    static_vars.command_data = bytes.fromhex('9b') + bytes.fromhex('36') + config + amtc1

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_IllegallyFormattedUID(): #240003
  """
  Inputs to generateSignature:
    UID is formatted in some non-valid manner, i.e. non-hexadecimal string

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid  # Illegally formatted UID
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified is None)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    config = b'\x00'

    static_vars.command_data_bytes = None
    static_vars.command_data = bytes.fromhex('9b') + bytes.fromhex('36') + config + amtc1

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 0)


def test_recover_device_Negative_MissingUID(): #240004
  """
  Inputs to generateSignature:
    UID field is empty

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = None  # UID is missing
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified is None)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    config = b'\x00'

    static_vars.command_data_bytes = None
    static_vars.command_data = bytes.fromhex('9b') + bytes.fromhex('36') + config + amtc1

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_BadOp(): #240005
  """
  Inputs to generateSignature:
    Op field is invalid

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTM"  # Invalid Op

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    config = b'\x00'

    static_vars.command_data_bytes = None
    static_vars.command_data = bytes.fromhex('9b') + bytes.fromhex('36') + config + amtc1

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_InvalidCommandData(): #240006
  """
  Inputs to generateSignature:
    Command Data field is invalid

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection
    
    static_vars.command_data_bytes = None
    static_vars.command_data = b'\x00'*12 # Invalid Command Data

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_IllegallyFormattedCommandData(): #240007
  """
  Inputs to generateSignature:
    Command Data field is formatted in an invalid manner, i.e. non-hexadecimal string

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    config = b'\x00'

    static_vars.command_data = None
    static_vars.command_data_bytes = bytes.fromhex('9b') + bytes.fromhex('36') + config + amtc1

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_MissingCommandData(): #240008
  """
  Inputs to generateSignature:
    Command Data field is empty

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    static_vars.command_data_bytes = None
    static_vars.command_data = None # Command Data is missing

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)


def test_recover_device_Negative_MTC1Mismatch(): #240009
  """
  Inputs to generateSignature:
    All fields valid, except that there is a mismatch in MTC1 values between Naxos and the KMS

  Expected Outputs:
    Failure error code
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    # Get A-MTC1
    amtc1 = utils.get_amtc1(device)
    mismatch_amtc1 = int.from_bytes(amtc1, byteorder='big') - 1 # Mismatch A-MTC1
    mismatch_amtc1_bytes = mismatch_amtc1.to_bytes(length=len(amtc1), byteorder='big')
    config = b'\x00'

    static_vars.command_data_bytes = None
    static_vars.command_data = bytes.fromhex('9b') + bytes.fromhex('36') + config + mismatch_amtc1_bytes

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success == False)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified is None)
    assert(status == 400)

def test_recover_device_Positive(): #240001
  """
  Inputs to generateSignature:
    Op is signDeprovisionCRTMCommand
    Valid UID
    Valid command data

  Expected Outputs:
    Command with valid signature
  """

  device = utils.get_device()

  # Get the nonce
  nonce_bytes, nonce_tag = utils.get_nonce_from_kms_proxy(connection, MPN)
  assert (nonce_bytes)
  assert (nonce_tag)

  # Get the UID from device
  err, response = utils.get_request_uid_response(device, nonce_bytes)
  assert (err)
  assert (response)

  # Verify the request UID response from the chip
  rc_int, _, _ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, MPN, mtcPowerOnValue)
  assert (rc_int == 1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  assert(dev_rev == 0x010005)
  
  device = NaxosDevice_1_5(device.endpoint_platform)

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      assert(err.esr_success)
      assert(device.provision_sts_struct.Provisioned)

    static_vars.UID = device.uid_as_str
    static_vars.Op = "signDeprovisionCRTMCommand"

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    assert(err.esr_success)

    # Verify CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, MPN, static_vars.UID)
    assert(verified)

    # Collect info for generateSignature call
    device.use_kms = True
    device.kms_connection = connection

    static_vars.validCommand = True

    # Deprovision the device
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    assert(err.esr_success)

    # verify the deprovision command signature
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", static_vars.UID)
    assert(verified)
    assert(status == 200)


# def test_recover_device_Negative_KeyLookupFailed(): #240010
#   """
#   Inputs to generateSignature:
#     All fields valid, except that the key is missing from the KMS

#   Expected Outputs:
#     Failure error code
#   """
