#! /usr/bin/env python3

"""
KMS Test for Replace Server Root Key.
(C) 2020 Micron Technology, Inc. All rights reserved.
"""

import pytest
import conftest
import pytest_common_utils as utils
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

@pytest.fixture(scope="module", autouse=True)
def test_replace_server_root_key_setup(request):
  global connection
  global MPN
  global new_srk_name

  MPN = 'MTANXSBC01'
  new_srk_name = 'srk'

  connection = conftest.create_connection()

  yield

def test_replace_server_root_key_Positive(): #220001
  """
  Inputs to generateSignature API:
    Op is signReplaceSRKCommand
    Valid command data
		Valid additionData (new key digest value)

  Expected Outputs:
    Command with valid signature
    Success status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig)
  assert(status == 200)

 
def test_replace_server_root_key_Negative_BadOp(): #220002
  """
  Inputs to generateSignature API:
    Op field is invalid

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRK"
  command_data = command.hex()
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 401)


def test_replace_server_root_key_Negative_UnknownUID(): #220003
  """
  Inputs to generateSignature API:
    UID is unknown to the KMS

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, _, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  UID = "1032547698badcfe1f32547698baaa55"

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)


def test_replace_server_root_key_Negative_IllegallyFormattedUID(): #220004
  """
  Inputs to generateSignature API:
    UID is formatted as a non-hexadecimal string

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  UID = ret_device.uid

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 0)


def test_replace_server_root_key_Negative_MissingUID(): #220005
  """
  Inputs to generateSignature API:
    UID field is empty

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, _, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  UID = None

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  

def test_replace_server_root_key_Negative_InvalidCommandData(): #220006
  """
  Inputs to generateSignature API:
    Command Data field is invalid

  Expected Outputs:
    Failure error code
    Failure status code
  """

  _, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = b"\x00"*12
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 0)
  

def test_replace_server_root_key_Negative_IllegallyFormattedCommandData(): #220007
  """
  Inputs to generateSignature API:
    Command Data field is formatted as a non-hexadecimal string

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 0)
  

def test_replace_server_root_key_Negative_MissingCommandData(): #220008
  """
  Inputs to generateSignature API:
    Command Data field empty

  Expected Outputs:
    Failure error code
    Failure status code
  """

  _, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = None
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  

def test_replace_server_root_key_Negative_InvalidAdditionalData(): #220009
  """
  Inputs to generateSignature API:
    Additional Data field is not the correct key value

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  additional_data = 0
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  

def test_replace_server_root_key_Negative_IllegallyFormattedAdditionalData(): #220010
  """
  Inputs to generateSignature API:
    Additional Data field is formatted as a non-hexadecimal string

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  additional_data = bytes.fromhex(additional_data)
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 0)
  

def test_replace_server_root_key_Negative_MissingAdditionalData(): #220011
  """
  Inputs to generateSignature API:
    Additional Data field empty

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  command_data = command.hex()
  additional_data = None
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  

def test_replace_server_root_key_Negative_BadNonce(): #220012
  """
  Inputs to generateSignature API:
    All fields valid, except that there is a mismatch in nonce between Naxos and the KMS

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  nonce_bytes = b'\x12\x12\x12\x12'
  command = command[:8] + nonce_bytes
  command_data = command.hex()
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  

def test_replace_server_root_key_Negative_MTC0Mismatch(): #220013
  """
  Inputs to generateSignature API:
    All fields valid, except that there is a mismatch in MTC0 values between Naxos and the KMS

  Expected Outputs:
    Failure error code
    Failure status code
  """

  command, ret_device, additional_data = utils.replace_srk_setup(new_srk_name, connection)
  
  op = "signReplaceSRKCommand"
  amtc0 = b'\x12\x12\x12\x12'
  command = command[:3] + amtc0 + command[7:]
  command_data = command.hex()
  UID = ret_device.uid_as_str

  sig, _, status = MicronKMSProxyAPI.GenerateCmdSignature(op, command_data, additional_data,
                                                          connection, MPN, UID)
  assert(sig is None)
  assert(status == 400)
  