#! /usr/bin/env python3

"""
KMS Test for Request Server Root Key.
(C) 2020 Micron Technology, Inc. All rights reserved.
"""

import pytest
import time
import pytest_common_utils as utils
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
import conftest

@pytest.fixture(scope="module", autouse=True)
def test_request_server_root_key_setup(request):
  global device
  global connection
  global MPN
  global keyType
  global encryptionKey

  device = conftest.get_device(request)
  connection = conftest.create_connection()

  MPN = "MTANXSBB01"
  keyType = 'SRK'
  encryptionKey = "./tmp/kms_rsa_encryption_key.pub"

  yield


def test_request_server_root_key_Positive(): #210001
  """
  Inputs to getKey:
    Valid UID
    Valid nonce
    key type SRK
    Valid encryption key

  Expected Outputs:
    Valid encrypted key
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    # send the request UID response from the chip to the KMS to be verified
    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')
    assert (rc_int == 1)
    assert (status == 200)

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc)
    assert (iv)
    assert (tag)
    assert (SRK_enc)
    assert (status == 200)



def test_request_server_root_key_Negative_InvalidNonce(): #210002
  """
  Inputs to getKey:
    Invalid (i.e. random) nonce

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    nonce = '\x31\x14\x24\x12'  #Invalid (random) nonce

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)



def test_request_server_root_key_Negative_InvalidMPNforNonce(): #210003
  """
  Inputs to getKey:
    Generate nonce with invalid MPN (i.e. not "MTANXSBB01")

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)   
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, "MTANXSBC01", keyType, encryptionKey) #Invalid MPN
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 401)



def test_request_server_root_key_Negative_StaleNonce(): #210004
  """
  Inputs to getKey:
    Previously used nonce

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce1, nonce_tag1, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce1)
    assert (nonce_tag1)

    nonce2, nonce_tag2, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce2)
    assert (nonce_tag2)

    nonce = nonce1 # Use first nonce received (Stale nonce)
    nonce_tag = nonce_tag1

    nonce_bytes = bytes.fromhex(nonce2)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag2, MPN, 'false')

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)



def test_request_server_root_key_Negative_MissingNonce(): #210005
  """
  Inputs to getKey:
    Nonce field is empty

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    nonce = None # Nonce field is empty

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)



def test_request_server_root_key_Negative_IllegallyFormattedNonce(): #210006
  """
  Inputs to getKey:
    Nonce value formatted in some non-valid manner, i.e. non-hexadecimal string

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    nonce = nonce_bytes # Non-hexadecimal nonce

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 0)


def test_request_server_root_key_Negative_ExpiredNonce(): #210007
  """
  Inputs to getKey:
    All parameters are valid, except wait more than 10 seconds after calling Get Nonce API

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    time.sleep(10) # Wait for 10 seconds before making next function call

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)

# # Test case 210008 ( Bad Product Type ) removed. No way to set the device type

def test_request_server_root_key_Negative_UnknownUID(): #210009
  """
  Inputs to getKey:
    UID is unknown to the KMS

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    UID = "1032547698badcfe1f32547698baaa55"   # Unknown UID

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)


def test_request_server_root_key_Negative_IllegallyFormattedUID(): #210010
  """
  Inputs to getKey:
    UID is formatted in some non-valid manner, i.e. non-hexadecimal string

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    UID = device.uid   # Non-valid UID format

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 0)


def test_request_server_root_key_Negative_MissingUID(): #210011
  """
  Inputs to getKey:
    UID field is empty

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    UID = None

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 500)


def test_request_server_root_key_Negative_InvalidKeyType(): #210012
  """
  Inputs to getKey:
    Key type is not SRK

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    keyType = "FRK"  # Invalid key type

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)



def test_request_server_root_key_Negative_MissingKeyType(): #210013
  """
  Inputs to getKey:
    Key type field is empty

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    keyType = None  # Missing key type

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)


def test_request_server_root_key_Negative_InvalidEncryptionKey(): #210014
  """
  Inputs to getKey:
    Encryption key is not valid

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    encryptionKey = "./tmp/invalid_encryption_key.pub"

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 400)


def test_request_server_root_key_Negative_IllegallyFormattedEncryptionKey(): #210015
  """
  Inputs to getKey:
    Encryption key is formatted in some non-valid manner, i.e non-hexadecimal string

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    encryptionKey = "./tmp/illegally_formatted_encryption_key.pub"

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 0)


def test_request_server_root_key_Negative_MissingEncrytptionKey(): #210016
  """
  Inputs to getKey:
    Encryption key field is empty

  Expected Outputs:
    Failure error code
  """

  with device:

    # Get the nonce
    nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, MPN)
    assert (nonce)
    assert (nonce_tag)

    nonce_bytes = bytes.fromhex(nonce)

    # Get UID from device
    err, response = utils.get_request_uid_response(device, nonce_bytes)
    assert (err)
    assert (response)

    UID = device.uid_as_str
    UIDData = response.hex()

    rc_int, _, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(UID, UIDData, connection, nonce_tag, MPN, 'false')

    encryptionKey = None

    # Get Key
    transportKey_enc, iv, tag, SRK_enc, err_msg, status = utils.get_key(UID, connection, nonce, MPN, keyType, encryptionKey)
    assert (transportKey_enc is None)
    assert (iv is None)
    assert (tag is None)
    assert (SRK_enc is None)
    assert (status == 401)
