///
/// @file platform_driver_server_svc_naxos.cc
///
/// @brief Micron SDK2 GRPC server Naxos services
///
/// @copyright (C) 2019 Micron Technology, Inc. All rights reserved.
///

#include <iostream>
#include <memory>
#include <string>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <mtd/mtd-user.h>

#include <grpcpp/grpcpp.h>

#include "../../platform_driver_server_ll.h"
#include "../platform_driver_server_svc_naxos.h"

#define IOCTL_FAIL (-1)

/* Get basic MTD characteristics info (better to use sysfs) */
#define MEMGETINFO _IOR('M', 1, struct mtd_info_user)

/*#define MEMERASE_BYPASS 0*/

#ifdef MEMERASE_BYPASS
#define NAXOS_SUBSECTOR_ERASE _IOW('M', 37, struct mtd_ioc_transfer) //0x20
#else
/* Erase segment of MTD */
#define MEMERASE _IOW('M', 2, struct erase_info_user)
#endif

/* Lock a chip (for MTD that supports it) */
#define MEMLOCK _IOW('M', 5, struct erase_info_user)
/* Unlock a chip (for MTD that supports it) */
#define MEMUNLOCK _IOW('M', 6, struct erase_info_user)

#define NAXOSGPR _IOR('M', 25, struct mtd_read_gpr)
#define NAXOSCRTM _IOWR('M', 26, struct mtd_crtm)
#define NAXOS_NV_RD_BLK _IOR('M', 27, struct mtd_ioc_transfer)              //0xe2
#define NAXOS_RD_NV_BLK_LLB _IOR('M', 28, struct mtd_ioc_transfer)          //0xa7
#define NAXOS_RD_PASSWD _IOR('M', 29, struct mtd_ioc_transfer)              //0x27
#define NAXOS_LOCK_BLOCK _IOW('M', 30, struct mtd_ioc_transfer)             //0xe5
#define NAXOS_NV_LOCK_BLOCK _IOW('M', 31, struct mtd_ioc_transfer)          //0xe3/0xe4
#define NAXOS_LOCK_GLOBAL_FREEZE_BIT _IOW('M', 32, struct mtd_ioc_transfer) //0xa6
#define NAXOS_V_LOCK_BIT _IOR('M', 33, struct mtd_ioc_transfer)             //0xe8
#define NAXOS_READ_SR _IOR('M', 34, struct mtd_ioc_transfer)                //0x05
#define NAXOS_WRITE_ENABLE _IOW('M', 35, struct mtd_ioc_transfer)           //0x06
#define NAXOS_WRITE_DISABLE _IOW('M', 36, struct mtd_ioc_transfer)          //0x04

#define SIZE_GPR 64
#define SIZE_LARGE_RETURN_BUF 65536
#define MAX_SPI_FLASH_ADDRESS 0xFFFFFF //OD to be revisited

#define GET_FIRST_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n)))
#define GET_SECOND_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 8)))
#define GET_THIRD_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 16)))
#define GET_FOURTH_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 24)))

using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerContext;
using grpc::Status;

using namespace std;
using namespace authenta;

struct mtd_ioc_transfer
{
    uint8_t *tx_buff;
    uint8_t *rx_buff;
    uint32_t buflen;
};

struct mtd_read_gpr
{
    unsigned char opcode4bypass;
    unsigned char *ptr;
};

struct mtd_crtm
{
    unsigned char opcode4bypass;
    unsigned char subopcode4bypass;
    unsigned char payload4bypasslength;
    unsigned char *ptr;
};

static struct mtd_info_user mtd;
static struct device_mt_info *devices_ll_head;
static struct erase_info_user erase_info;
static struct mtd_read_gpr gpr_info;
static struct mtd_crtm crtm_info;
static struct mtd_ioc_transfer mtd_ioc_transfer_info;
static char large_return_buf[SIZE_LARGE_RETURN_BUF];

static status_type authentaCmd(
    const naxos_content *cmd,
    const char *cmdName,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes);

static status_type check_cmd_content(
    const naxos_content *cmd,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes);

static void fill_crtm_request(const naxos_content *cmd);
static void print_crtm_info(void);
static int writeEnable(int fd);


// RPC call implementations

// Authenta_device_open
grpc::Status NaxosServiceImpl::Authenta_device_open(
    grpc::ServerContext *context,
    const string_msg *request,
    device_descriptor *reply)
{
    const char *funcName = "Authenta_device_open";
    int fd = -1;
    const string &path = request->value();
    reply->set_fd(fd);
    reply->set_status(DEVICE_STATUS_ERROR);

    // make sure the inputs are valid
    if (path == "")
    {
        printf("%s: path is not valid\n", funcName);
        return Status::OK;
    }

    fd = open(path.c_str(), O_RDWR);
    if (fd < 0)
    {
        printf("%s: failed opening %s\n", funcName, path.c_str());
        return Status::OK;
    }

    // Read MTD info
    int ret = ioctl(fd, MEMGETINFO, &mtd);
    if (ret == IOCTL_FAIL)
    {
        printf("%s: mtd info not available\n", funcName);
        close(fd);
        return Status::OK;
    }

    reply->set_fd(fd);
    reply->set_status(DEVICE_STATUS_IDLE);
    devices_ll_head = add_head(devices_ll_head, fd, path.c_str());

    return Status::OK;
}

// Authenta_device_close
grpc::Status NaxosServiceImpl::Authenta_device_close(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (close(fd) == -1)
    {
        return Status::OK;
    }

    devices_ll_head = remove_element(devices_ll_head, fd);

    reply->set_status(STATUS_OK);

    return Status::OK;
}


// Authenta_device_acquire
grpc::Status NaxosServiceImpl::Authenta_device_acquire(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Authenta_device_acquire";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    ll_result acq_result = DEVICE_ACQUIRED;

    if (check_control(devices_ll_head, fd))
    {
        printf("%s: Device has already been acquired\n", funcName);
        return Status::OK;
    }
    else
    {
        acq_result = acquire_device(devices_ll_head, fd);

        if (acq_result == DEVICE_BUSY)
        {
            printf("%s: DEVICE_BUSY\n", funcName);
            return Status::OK;
        }
        else if (acq_result == DEVICE_NOT_FOUND)
        {
            printf("%s: DEVICE_NOT_FOUND\n", funcName);
            return Status::OK;
        }
        else if (acq_result != DEVICE_ACQUIRED)
        {
            printf("%s: unknown behavior\n", funcName);
            return Status::OK;
        }
    }
    if (acq_result != DEVICE_ACQUIRED)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Authenta_device_release
grpc::Status NaxosServiceImpl::Authenta_device_release(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Authenta_device_release";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    ll_result rel_result = release_device(devices_ll_head, fd);

    if (rel_result == DEVICE_NOT_OWNED)
    {
        printf("%s: DEVICE_NOT_OWNED\n", funcName);
        return Status::OK;
    }
    else if (rel_result == DEVICE_NOT_FOUND)
    {
        printf("%s: DEVICE_NOT_FOUND\n", funcName);
        return Status::OK;
    }
    else if (rel_result != DEVICE_FREE)
    {
        printf("%s: unknown behavior\n", funcName);
        return Status::OK;
    }
    if (rel_result != DEVICE_FREE)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Authenta_detect_devices
grpc::Status NaxosServiceImpl::Authenta_detect_devices(
    grpc::ServerContext *context,
    const void_msg *request,
    list_info *reply)
{
    //OD implement later
    reply->set_listsize(0);

    return Status::OK;
}

// Write_Enable
grpc::Status NaxosServiceImpl::Write_Enable(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Write_Enable";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    erase_info.length = mtd.erasesize * 16; //OD: necessary?
    erase_info.start = 0;

    int ioctl_ret = writeEnable(fd);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Write_Disable
grpc::Status NaxosServiceImpl::Write_Disable(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Write_Disable";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    erase_info.length = mtd.erasesize * 16; //OD: necessary?
    erase_info.start = 0;

    uint8_t tx_gpr = 0x04;
    uint8_t rx_gpr = 0;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    int ioctl_ret = ioctl(fd, NAXOS_WRITE_DISABLE, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Read_General_Purpose_Readout_Register
grpc::Status NaxosServiceImpl::Read_General_Purpose_Readout_Register(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Read_General_Purpose_Readout_Register";
    int fd = request->value();

    reply->set_buffer_size(0);
    reply->set_status(STATUS_ERROR);

    //printf("%s: got called\n", funcName);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    gpr_info.opcode4bypass = 0x96;
    gpr_info.ptr = (unsigned char *)malloc(SIZE_GPR);
    if (gpr_info.ptr == NULL)
    {
        printf("%s: failed allocating memory for gpr_info\n", funcName);
        return Status::OK;
    }

    //printf("%s: requesting gpr\n", funcName);
    memset(gpr_info.ptr, 0, SIZE_GPR);
    if (ioctl(fd, NAXOSGPR, &gpr_info) < 0)
    {
        free(gpr_info.ptr);
        printf("%s: failed getting gpr\n", funcName);
        return Status::OK;
    }
    //printf("%s: gpr register content:\n", funcName);

    reply->set_return_buf(gpr_info.ptr, SIZE_GPR);
    reply->set_buffer_size(SIZE_GPR);
    reply->set_status(STATUS_OK);

    /*
    for (int i = 0; i < SIZE_GPR; i++)
    {
        printf("0x%x ", gpr_info.ptr[i]);
        if (i % 8 == 7)
        {
            printf("\n");
        }
    }*/
    free(gpr_info.ptr);
    //printf("%s: leaving\n", funcName);

    return Status::OK;
}

// Request_UID
grpc::Status NaxosServiceImpl::Request_UID(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Request_UID", 0x9b, 0x3e, 7));
    return Status::OK;
}

// Auth_Modify
grpc::Status NaxosServiceImpl::Auth_Modify(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Auth_Modify", 0x9b, 0x34, 0));
    return Status::OK;
}

// Provision_CRTM
grpc::Status NaxosServiceImpl::Provision_CRTM(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Provision_CRTM", 0x9b, 0x35, 0));
    return Status::OK;
}

// Deprovision_CRTM
grpc::Status NaxosServiceImpl::Deprovision_CRTM(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)   
{
    reply->set_status(authentaCmd(request, "Deprovision_CRTM", 0x9b, 0x36, 0));
    return Status::OK;
}

// Auth_lock
grpc::Status NaxosServiceImpl::Auth_lock(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply) 
{
    reply->set_status(authentaCmd(request, "Auth_lock", 0x9b, 0x37, 39));
    return Status::OK;
}

// CRTM_Register_Read
grpc::Status NaxosServiceImpl::CRTM_Register_Read(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Read", 0x9b, 0x38, 9));
    return Status::OK;
}

// CRTM_Register_Write
grpc::Status NaxosServiceImpl::CRTM_Register_Write(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Write", 0x9b, 0x3c, 0));
    return Status::OK;
}

// Flash_extend
grpc::Status NaxosServiceImpl::Flash_extend(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Write", 0x9b, 0x3a, 0));
    return Status::OK;
}

// Replace_SRK
grpc::Status NaxosServiceImpl::Replace_SRK(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Replace_SRK", 0x9b, 0x39, 44));
    return Status::OK;
}

// ACRTM_Update_Start
grpc::Status NaxosServiceImpl::ACRTM_Update_Start(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "ACRTM_Update_Start", 0x9b, 0x3b, 39));
    return Status::OK;
}

// ACRTM_Update_Finish
grpc::Status NaxosServiceImpl::ACRTM_Update_Finish(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "ACRTM_Update_Finish", 0x9b, 0x3d, 39));
    return Status::OK;
}

// Measure
grpc::Status NaxosServiceImpl::Measure(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Measure", 0x9b, 0x3f, 0));
    return Status::OK;
}

// Request_Monotonic_Counter
grpc::Status NaxosServiceImpl::Request_Monotonic_Counter(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Request_Monotonic_Counter", 0x9b, 0x03, 0x10));
    return Status::OK;
}

// Fast_Read
grpc::Status NaxosServiceImpl::Fast_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "Fast_Read";
    int fd = request->fd();
    int address = request->address();
    int length = request->length();
    reply->set_status(STATUS_ERROR);

    printf("%s: got called\n", funcName);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address + length > MAX_SPI_FLASH_ADDRESS || length > SIZE_LARGE_RETURN_BUF)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    memset(large_return_buf, '@', sizeof(large_return_buf)); //OD necessary?

    //OD check fd?

    lseek(fd, address, SEEK_SET);
    int ret = read(fd, large_return_buf, length);
    if (ret != length)
    {
        printf("%s: SPI_FLASH_UTIL_ERROR_CODE_FLASH_COMM_FAILURE \n", funcName);
        return Status::OK;
    }
    printf("%s: read command apparently succeeded, size = %d\n", funcName, length);

    for (int i = 0; i < 64; i++)
    {
        printf("0x%x ", large_return_buf[i]);
        if (i % 8 == 7)
        {
            printf("\n");
        }
    }

    reply->set_buffer_size(length);
    reply->set_return_buf(large_return_buf, length);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Page_Program
grpc::Status NaxosServiceImpl::Page_Program(
    grpc::ServerContext *context,
    const flash_write_content *request,
    status_type_msg *reply)
{
    const char *funcName = "Page_Program";
    int fd = request->address().fd();
    int address = request->address().address();
    int length = request->address().length();
    const char *data = request->data().c_str();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    /* printf("%s: sending ", funcName);
    for (int i = 0; i < 5; i++)
    {
        printf("0x%x ", data[i]);
    }
    printf(" to fd %d at address %d and length %d\n", fd, address, length);
    */

    lseek(fd, address, SEEK_SET);
    int ret = write(fd, data, length);
    if (ret != length)
    {
        printf("%s: SPI_FLASH_UTIL_ERROR_CODE_FLASH_COMM_FAILURE \n", funcName);
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// DYB_Write
grpc::Status NaxosServiceImpl::DYB_Write(
    grpc::ServerContext *context,
    const flash_lock_config *request,
    status_type_msg *reply)
{
    const char *funcName = "DYB_Write";
    int fd = request->address().fd();
    int address = request->address().address();
    int lockbit = request->lockbit();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr[5];

    tx_gpr[0] = 0xE5; // Opcode for lock block
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb
    tx_gpr[4] = lockbit ? 0x01 : 0x00;

    /*
    printf("%s: sending ", funcName);
    for (int i = 0; i < 5; i++)
    {
        printf("0x%x ", tx_gpr[i]);
    }
    printf(" to fd %d \n", fd);
    */

    mtd_ioc_transfer_info.buflen = 5;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr[0];
    mtd_ioc_transfer_info.rx_buff = &rx_gpr; //OD: check if this data is needed somewhere
    
    ioctl_ret = ioctl(fd, NAXOS_LOCK_BLOCK, &mtd_ioc_transfer_info);

    //printf("%s: returned %d, rx_gpr = %d \n", funcName, ioctl_ret, rx_gpr);

    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Program
grpc::Status NaxosServiceImpl::PPB_Program(
    grpc::ServerContext *context,
    const flash_address *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Program";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr[5];

    tx_gpr[0] = 0xE3; // Opcode for lock block
    tx_gpr[1] = GET_FOURTH_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[4] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    /*printf("%s: sending ", funcName);
    for (int i = 0; i < 5; i++)
    {
        printf("0x%x ", tx_gpr[i]);
    }
    printf(" to fd %d at address %d and length %d\n", fd, address, length);
    */
    mtd_ioc_transfer_info.buflen = 5;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr[0];
    mtd_ioc_transfer_info.rx_buff = &rx_gpr; //OD: check if needed
    
    ioctl_ret = ioctl(fd, NAXOS_NV_LOCK_BLOCK, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Erase
grpc::Status NaxosServiceImpl::PPB_Erase(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Erase";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0xE4;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_NV_LOCK_BLOCK, &mtd_ioc_transfer_info);
    
    //    printf("%s: received %d response %d ioctl\n", funcName, rx_gpr, ioctl_ret);

    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Lock_Bit_Write
grpc::Status NaxosServiceImpl::PPB_Lock_Bit_Write(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Lock_Bit_Write";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0xA6;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_LOCK_GLOBAL_FREEZE_BIT, &mtd_ioc_transfer_info);

    //printf("%s: received %d response and %d ioctl\n", funcName, rx_gpr, ioctl_ret);

    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// DYB_Read
grpc::Status NaxosServiceImpl::DYB_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "DYB_Read";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr[6];
    uint8_t tx_gpr[4];
    memset(rx_gpr, 0, 6);

    tx_gpr[0] = 0xE8; // Opcode for Read volatile lock bit
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    mtd_ioc_transfer_info.buflen = 4;
    mtd_ioc_transfer_info.tx_buff = tx_gpr;
    mtd_ioc_transfer_info.rx_buff = rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_V_LOCK_BIT, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(rx_gpr, sizeof(rx_gpr));
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Read
grpc::Status NaxosServiceImpl::PPB_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "PPB_Read";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr[5];
    uint8_t tx_gpr[5];
    memset(rx_gpr, 0, 6);

    tx_gpr[0] = 0xE2; // Opcode for Read volatile lock bit
    tx_gpr[1] = GET_FOURTH_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[4] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    /*printf("%s: sending ", funcName);
    for (int i = 0; i < 5; i++)
    {
        printf("0x%x ", tx_gpr[i]);
    }
    printf(" to fd %d at address %d and length %d\n", fd, address, length);
    */

    mtd_ioc_transfer_info.buflen = 5;
    mtd_ioc_transfer_info.tx_buff = tx_gpr;
    mtd_ioc_transfer_info.rx_buff = rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_NV_RD_BLK, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(rx_gpr, 1);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Lock_Bit_Read
grpc::Status NaxosServiceImpl::PPB_Lock_Bit_Read(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "PPB_Lock_Bit_Read";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0xA7;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_RD_NV_BLK_LLB, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr, 1);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Password_Read
grpc::Status NaxosServiceImpl::Password_Read(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Password_Read";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t tx_gpr = 0x27;
    uint8_t rx_gpr[8];
    memset(rx_gpr, 0, 8);

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_READ_SR, &mtd_ioc_transfer_info);
    if (ioctl_ret == -1)
    {
        return Status::OK;
    }

    reply->set_buffer_size(8);
    reply->set_return_buf(&rx_gpr, 8);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Read_Status_register
grpc::Status NaxosServiceImpl::Read_Status_register(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Read_Status_register";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0x05;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    ioctl_ret = ioctl(fd, NAXOS_READ_SR, &mtd_ioc_transfer_info);

    rx_gpr = ioctl_ret;
    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr, 1);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Subsector_Erase
grpc::Status NaxosServiceImpl::Subsector_Erase(
    grpc::ServerContext *context,
    const flash_address *request,
    status_type_msg *reply)
{
    const char *funcName = "Subsector_Erase";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);
    
    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    lseek(fd, address, SEEK_SET);

#ifdef MEMERASE_BYPASS
    uint8_t tx_gpr[4];
    uint8_t rx_gpr;

    tx_gpr[0] = 0x20; // Opcode for subsector erase
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    /*printf("%s: sending ", funcName);
    for (int i = 0; i < 4; i++)
    {
        printf("0x%x ", tx_gpr[i]);
    }
    printf(" to fd %d \n", fd);
    */

    mtd_ioc_transfer_info.buflen = 4;
    mtd_ioc_transfer_info.tx_buff = tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    int ret = ioctl(fd, NAXOS_SUBSECTOR_ERASE, &mtd_ioc_transfer_info); 
    
    // OD: calls spi_nor_erase, which in turn calls fsl_qspi_erase, which in turn calls runcmd
    // with opcode SPINOR_OP_BE_4K, or SPINOR_OP_BE_4K_PMC or SPINOR_OP_SE; SE=0xd8, SSE=0x20,
    // only one is supported at a time; forcing 0x20 by adding SECT_4K to n25q128 entries in spi-nor.c
#else
    erase_info.length = mtd.erasesize * 1;
    erase_info.start = address;
    int ret = ioctl(fd, MEMERASE, &erase_info);
#endif

    if (ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// static functions

// run authenta command
// set expected_nb_bytes to 0 if not want to check command length
static status_type authentaCmd(
    const naxos_content *cmd,
    const char *cmdName,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes)
{
    if (check_cmd_content(
            cmd,
            expected_opcode,
            expected_subopcode,
            expected_nb_bytes) != STATUS_OK)
    {
        printf("%s: command badly formed\n", cmdName);
        return STATUS_ERROR;
    }

    fill_crtm_request(cmd);
    print_crtm_info();
    printf("%s: executing command\n", cmdName);

    if (ioctl(cmd->fd(), NAXOSCRTM, &crtm_info) < 0)
    {
        printf("%s: failed excuting command using fd %d\n",
               cmdName, cmd->fd());
        return STATUS_ERROR;
    }
    else
    {
        printf("%s: sent successfully\n", cmdName);
    }
    return STATUS_OK;
}

// check command opcode, subopcode, and length
// if expected_nb_bytes is set to 0, not check command length
static status_type check_cmd_content(
    const naxos_content *cmd,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes)
{
    int nb_bytes = cmd->nb_bytes();
    const char *bytes = cmd->bytes().c_str();

    if (expected_nb_bytes && nb_bytes != expected_nb_bytes)
    {
        return STATUS_ERROR;
    }
    if (bytes[0] != expected_opcode)
    {
        return STATUS_ERROR;
    }
    if (bytes[1] != expected_subopcode)
    {
        return STATUS_ERROR;
    }

    return STATUS_OK;
}

static void fill_crtm_request(const naxos_content *cmd)
{
    const char *bytes = cmd->bytes().c_str();
    crtm_info.opcode4bypass = bytes[0];
    crtm_info.subopcode4bypass = bytes[1];
    crtm_info.payload4bypasslength = cmd->nb_bytes() - 1;
    crtm_info.ptr = (unsigned char *)&(bytes[1]); //OD: signed to unsigned, fix later?
}

static void print_crtm_info(void)
{
    int i;
    printf("crtm info: ");
    printf("%02x", crtm_info.opcode4bypass);
    for (i = 0; i < crtm_info.payload4bypasslength; i++)
    {
        printf("%02x ", crtm_info.ptr[i]);
    }
    printf("\n");
}

// call NAXOS_WRITE_ENABLE and get return code
static int writeEnable(int fd)
{
    uint8_t tx_gpr = 0x06;
    uint8_t rx_gpr = 0;

    mtd_ioc_transfer_info.buflen = 1;
    mtd_ioc_transfer_info.tx_buff = &tx_gpr;
    mtd_ioc_transfer_info.rx_buff = &rx_gpr;

    return ioctl(fd, NAXOS_WRITE_ENABLE, &mtd_ioc_transfer_info);
}
