///
/// @file platform_driver_server_svc_naxos.cc
///
/// @brief Micron SDK2 GRPC server Naxos services
///
/// @copyright (C) 2019 Micron Technology, Inc. All rights reserved.
///

#include <iostream>
#include <memory>
#include <string>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/spi/spidev.h>

#include <grpcpp/grpcpp.h>

#include "../../platform_driver_server_ll.h"
#include "../platform_driver_server_svc_naxos.h"

/***
 * SPIDEV Definitions
 ***/

// On reads, throw out the first 2 bytes which are falsely read MOSI bytes
#define MOSI_OFFSET_BYTES		2

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

// SPI bits per read/write
static uint8_t const spiBits = 8;

// SPI speed in Hz
static uint32_t const spiSpeed = 1000000;

// SPI mode
static uint8_t const spiMode = 3;

// SPI delay in uS
static uint16_t spiDelay = 0;

/***
 * END SPIDEV Definitions
 ***/

/* The following Naxos read commands may cause changes to the command state, CmdState_t cmdState.
 * See Table 19, Command Set in MT25Q data sheet */
#define		CMD_READ_ID					(uint8_t)(0x9F)
#define		CMD_READ_ID_ALT				(uint8_t)(0x9E)
#define		CMD_AUTHENTA_OP				(uint8_t)(0x9B)
#define		CMD_READ_STATUS_REG			(uint8_t)(0x05)
#define		CMD_READ_FLAG_STATUS_REG	(uint8_t)(0x70)
#define		CMD_READ_GEN_PRP_REG		(uint8_t)(0x96)
#define     CMD_PAGE_PROGRAM            (uint8_t)(0x02)
#define     CMD_READ_MEMORY             (uint8_t)(0x03)
#define     CMD_FAST_READ               (uint8_t)(0x0B)
#define		CMD_OPCODE_NONE				(uint8_t)(0x00)		// Indicate no opcode

/* Authenta Sub-Opcodes */
#define     AUTH_OP_REQ_MTC             (uint8_t)(0x03)
#define     AUTH_OP_AUTH_MODIFY         (uint8_t)(0x34)
#define     AUTH_OP_PROVISION_CRTM      (uint8_t)(0x35)
#define     AUTH_OP_DEPROVISION_CRTM    (uint8_t)(0x36)
#define     AUTH_OP_AUTH_LOCK           (uint8_t)(0x37)
#define     AUTH_OP_CRTM_REG_READ       (uint8_t)(0x38)
#define     AUTH_OP_FLASH_EXTEND        (uint8_t)(0x3A)
#define     AUTH_OP_REPLACE_SRK         (uint8_t)(0x39)
#define     AUTH_OP_ACRTM_UPDATE_START  (uint8_t)(0x3B)
#define     AUTH_OP_CRTM_REG_WRITE      (uint8_t)(0x3C)
#define     AUTH_OP_ACRTM_UPDATE_FIN    (uint8_t)(0x3D)
#define     AUTH_OP_REQ_UID             (uint8_t)(0x3E)
#define     AUTH_OP_CRTM_MEASURE        (uint8_t)(0x3F)

// How many times to read flag status register "busy" bit before giving up
// TODO this should probably be determined from the data sheet, analysis, etc.
#define MAX_FSR_TIMEOUT_RETRIES	10000

// Flag Staus Register Bit-7 indicates: 1=READY, 0=BUSY
#define FLAG_STATUS_READY_MASK	(uint8_t)(1 << 7)

/** State machine to control Naxos command register operations.
 * Use the state to hold-off writing register read commands until
 * the actual read command is issued.
 */
typedef enum
{
	CMD_STATE_RDY_FOR_WRITE = 0,	///< The default state, ready for a "write" command
	CMD_STATE_RDY_FOR_READ,			///< Ready for a "read" command

	NUM_CMD_STATES					///< Total number of command states
} CmdStat_t;

static CmdStat_t cmdState;			///< Command state machine for delayed write+read
static uint8_t saveReadOpcode;		///< The "write" opcode for the "read" command

#define IOCTL_FAIL (-1)

#define SIZE_GPR 64
#define SIZE_LARGE_RETURN_BUF 65536
#define MAX_SPI_FLASH_ADDRESS 0xFFFFFF //OD to be revisited

#define WRITE_IN_PROGRESS_BIT 0x01

#define GET_FIRST_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n)))
#define GET_SECOND_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 8)))
#define GET_THIRD_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 16)))
#define GET_FOURTH_BYTE_OF_UINT32(n) ((uint8_t)((0x000000FF) & (n >> 24)))

using grpc::Server;
using grpc::ServerBuilder;
using grpc::ServerContext;
using grpc::Status;

using namespace std;
using namespace authenta;


static struct device_mt_info *devices_ll_head;
static struct spi_ioc_transfer spi_ioc_transfer_info;
static char large_return_buf[SIZE_LARGE_RETURN_BUF];
static char auth_resp_buf[SIZE_GPR];

static void printLongHex(const uint8_t * const data, const uint32_t len, const uint8_t widthInBytes, const char *title);

static status_type authentaCmd(
    const naxos_content *cmd,
    const char *cmdName,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes);

static status_type check_cmd_content(
    const naxos_content *cmd,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes);

static int writeEnable(int fd);
static int writeDisable(int fd);
static int readFlagStatusRegister(int fd, uint8_t *fsr);

static bool writeCommandIsRead(uint8_t opCode);
static status_type waitForFSR(int fd);

// RPC call implementations

// Authenta_device_open
grpc::Status NaxosServiceImpl::Authenta_device_open(
    grpc::ServerContext *context,
    const string_msg *request,
    device_descriptor *reply)
{
    const char *funcName = "Authenta_device_open";
    int fd = -1;
    const string &path = request->value();
    reply->set_fd(fd);
    reply->set_status(DEVICE_STATUS_ERROR);

    // Setup delayed read state machine
	saveReadOpcode = CMD_OPCODE_NONE;
	cmdState = CMD_STATE_RDY_FOR_WRITE;

    // make sure the inputs are valid
    if (path == "")
    {
        printf("%s: path is not valid\n", funcName);
        return Status::OK;
    }

    fd = open(path.c_str(), O_RDWR);
    if (fd < 0)
    {
        printf("%s: failed opening %s\n", funcName, path.c_str());
        return Status::OK;
    }

	// Set SPI mode
	int ioErr = ioctl(fd, SPI_IOC_WR_MODE, &spiMode);
	if(ioErr == IOCTL_FAIL)
	{
		printf("Error %d. Naxos driver failed to set SPI mode to %d.\r\n", errno, spiMode);
        return Status::OK;
	}
	else
	{
		printf("Set SPI mode to %d\r\n", spiMode);
	}

	// Set SPI bits per word
	ioErr = ioctl(fd, SPI_IOC_WR_BITS_PER_WORD, &spiBits);
	if(ioErr == IOCTL_FAIL)
	{
		printf("Error %d. Naxos driver failed to set bits per word to %d.\r\n", errno, spiBits);
        return Status::OK;
	}
	else
	{
		printf("Set bits per word to %d\r\n", spiBits);
	}

	// Set SPI speed
	ioErr = ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, &spiSpeed);
	if(ioErr == IOCTL_FAIL)
	{
		printf("Error %d. Naxos driver failed to set speed to %d Hz.\r\n", errno, spiSpeed);
        return Status::OK;
	}
	else
	{
		printf("Set speed to %d Hz.\r\n", spiSpeed);
	}

	// Read back all the settings to confirm
	uint8_t spiModeRead = (uint8_t)-1;
	uint8_t spiBitsRead = (uint8_t)-1;
	uint32_t spiSpeedRead = (uint32_t)-1;

	ioctl(fd, SPI_IOC_RD_MODE, &spiModeRead);
	ioctl(fd, SPI_IOC_RD_BITS_PER_WORD, &spiBitsRead);
	ioctl(fd, SPI_IOC_RD_MAX_SPEED_HZ, &spiSpeedRead);

	// Check the read back values against the original
	if( (spiModeRead != spiMode) ||
		(spiBitsRead != spiBits) ||
		(spiSpeedRead != spiSpeed) )
	{
		printf("Naxos driver readback of SPI settings failed.\r\n");
        return Status::OK;
	}

    reply->set_fd(fd);
    reply->set_status(DEVICE_STATUS_IDLE);
    devices_ll_head = add_head(devices_ll_head, fd, path.c_str());

    return Status::OK;
}

// Authenta_device_close
grpc::Status NaxosServiceImpl::Authenta_device_close(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (close(fd) == -1)
    {
        return Status::OK;
    }

    devices_ll_head = remove_element(devices_ll_head, fd);

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Authenta_device_acquire
grpc::Status NaxosServiceImpl::Authenta_device_acquire(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Authenta_device_acquire";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    ll_result acq_result = DEVICE_ACQUIRED;

    if (check_control(devices_ll_head, fd))
    {
        printf("%s: Device has already been acquired\n", funcName);
        return Status::OK;
    }
    else
    {
        acq_result = acquire_device(devices_ll_head, fd);

        if (acq_result == DEVICE_BUSY)
        {
            printf("%s: DEVICE_BUSY\n", funcName);
            return Status::OK;
        }
        else if (acq_result == DEVICE_NOT_FOUND)
        {
            printf("%s: DEVICE_NOT_FOUND\n", funcName);
            return Status::OK;
        }
        else if (acq_result != DEVICE_ACQUIRED)
        {
            printf("%s: unknown behavior\n", funcName);
            return Status::OK;
        }
    }
    if (acq_result != DEVICE_ACQUIRED)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Authenta_device_release
grpc::Status NaxosServiceImpl::Authenta_device_release(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    printf("Release Naxos Device...\n");

    const char *funcName = "Authenta_device_release";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    ll_result rel_result = release_device(devices_ll_head, fd);

    if (rel_result == DEVICE_NOT_OWNED)
    {
        printf("%s: DEVICE_NOT_OWNED\n", funcName);
        return Status::OK;
    }
    else if (rel_result == DEVICE_NOT_FOUND)
    {
        printf("%s: DEVICE_NOT_FOUND\n", funcName);
        return Status::OK;
    }
    else if (rel_result != DEVICE_FREE)
    {
        printf("%s: unknown behavior\n", funcName);
        return Status::OK;
    }
    if (rel_result != DEVICE_FREE)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Authenta_detect_devices
grpc::Status NaxosServiceImpl::Authenta_detect_devices(
    grpc::ServerContext *context,
    const void_msg *request,
    list_info *reply)
{
    //OD implement later
    reply->set_listsize(0);

    return Status::OK;
}

// Write_Enable
grpc::Status NaxosServiceImpl::Write_Enable(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Write_Enable";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    
    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret = writeEnable(fd);
    if (ioctl_ret == IOCTL_FAIL)
    {
        printf("%s: write_enable ioctl call FAILEDr\n", funcName);
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Write_Disable
grpc::Status NaxosServiceImpl::Write_Disable(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "Write_Disable";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret = writeDisable(fd);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Read_General_Purpose_Readout_Register
grpc::Status NaxosServiceImpl::Read_General_Purpose_Readout_Register(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Read_General_Purpose_Readout_Register";
    int fd = request->value();

    reply->set_buffer_size(0);
    reply->set_status(STATUS_ERROR);

    //printf("%s: got called\n", funcName);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    uint8_t tx_gpr[SIZE_GPR + MOSI_OFFSET_BYTES];
    memset(tx_gpr, 0x00, sizeof(tx_gpr));
    tx_gpr[0] = CMD_READ_GEN_PRP_REG;

     // Make a slightly oversized temporary read buffer
    uint8_t retBuf[SIZE_GPR + MOSI_OFFSET_BYTES];
    memset(retBuf, 0x00, sizeof(retBuf));

    spi_ioc_transfer_info.len = SIZE_GPR + MOSI_OFFSET_BYTES;
    spi_ioc_transfer_info.tx_buf = (unsigned long)&tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)retBuf;
    spi_ioc_transfer_info.delay_usecs = spiDelay;
    spi_ioc_transfer_info.speed_hz = spiSpeed;
    spi_ioc_transfer_info.bits_per_word = spiBits;


    //printf("%s: requesting gpr\n", funcName);
    if (ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info) < 0)
    {
        printf("%s: failed getting gpr\n", funcName);
        return Status::OK;
    }
    //printf("%s: gpr register content:\n", funcName);

    reply->set_return_buf(&retBuf[MOSI_OFFSET_BYTES], SIZE_GPR);
    reply->set_buffer_size(SIZE_GPR);
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Request_UID
grpc::Status NaxosServiceImpl::Request_UID(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Request_UID", CMD_AUTHENTA_OP, AUTH_OP_REQ_UID, 7));
    return Status::OK;
}

// Auth_Modify
grpc::Status NaxosServiceImpl::Auth_Modify(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Auth_Modify", CMD_AUTHENTA_OP, AUTH_OP_AUTH_MODIFY, 0));
    return Status::OK;
}

// Provision_CRTM
grpc::Status NaxosServiceImpl::Provision_CRTM(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Provision_CRTM", CMD_AUTHENTA_OP, AUTH_OP_PROVISION_CRTM, 0));
    return Status::OK;
}

// Deprovision_CRTM
grpc::Status NaxosServiceImpl::Deprovision_CRTM(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)   
{
    reply->set_status(authentaCmd(request, "Deprovision_CRTM", CMD_AUTHENTA_OP, AUTH_OP_DEPROVISION_CRTM, 0));
    return Status::OK;
}

// Auth_lock
grpc::Status NaxosServiceImpl::Auth_lock(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply) 
{
    reply->set_status(authentaCmd(request, "Auth_lock", CMD_AUTHENTA_OP, AUTH_OP_AUTH_LOCK, 39));
    return Status::OK;
}

// CRTM_Register_Read
grpc::Status NaxosServiceImpl::CRTM_Register_Read(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Read", CMD_AUTHENTA_OP, AUTH_OP_CRTM_REG_READ, 9));
    return Status::OK;
}

// CRTM_Register_Write
grpc::Status NaxosServiceImpl::CRTM_Register_Write(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Write", CMD_AUTHENTA_OP, AUTH_OP_CRTM_REG_WRITE, 0));
    return Status::OK;
}

// Flash_extend
grpc::Status NaxosServiceImpl::Flash_extend(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "CRTM_Register_Write", CMD_AUTHENTA_OP, AUTH_OP_FLASH_EXTEND, 0));
    return Status::OK;
}

// Replace_SRK
grpc::Status NaxosServiceImpl::Replace_SRK(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Replace_SRK", CMD_AUTHENTA_OP, AUTH_OP_REPLACE_SRK, 44));
    return Status::OK;
}

// ACRTM_Update_Start
grpc::Status NaxosServiceImpl::ACRTM_Update_Start(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "ACRTM_Update_Start", CMD_AUTHENTA_OP, AUTH_OP_ACRTM_UPDATE_START, 39));
    return Status::OK;
}

// ACRTM_Update_Finish
grpc::Status NaxosServiceImpl::ACRTM_Update_Finish(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "ACRTM_Update_Finish", CMD_AUTHENTA_OP, AUTH_OP_ACRTM_UPDATE_FIN, 39));
    return Status::OK;
}

// Measure
grpc::Status NaxosServiceImpl::Measure(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Measure", CMD_AUTHENTA_OP, AUTH_OP_CRTM_MEASURE, 0));
    return Status::OK;
}

// Request_Monotonic_Counter
grpc::Status NaxosServiceImpl::Request_Monotonic_Counter(
    grpc::ServerContext *context,
    const naxos_content *request,
    status_type_msg *reply)
{
    reply->set_status(authentaCmd(request, "Request_Monotonic_Counter", CMD_AUTHENTA_OP, AUTH_OP_REQ_MTC, 0x10));
    return Status::OK;
}

// Fast_Read
grpc::Status NaxosServiceImpl::Fast_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "Fast_Read";
    int fd = request->fd();
    int address = request->address();
    int length = request->length();
    reply->set_status(STATUS_ERROR);

    printf("%s: got called\n", funcName);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address + length > MAX_SPI_FLASH_ADDRESS || length > SIZE_LARGE_RETURN_BUF)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    memset(large_return_buf, '@', sizeof(large_return_buf)); //OD necessary?

    //OD check fd?

	uint32_t bufLen = length + 5;
	uint8_t *tx_gpr = (uint8_t *)malloc(bufLen * sizeof(uint32_t));
	if(tx_gpr == NULL)
	{
		return Status::CANCELLED;
	}

    memset(tx_gpr, 0x00, bufLen * sizeof(uint32_t));

	tx_gpr[0]=CMD_FAST_READ;		// Opcode for Read Fast

	tx_gpr[1]= GET_THIRD_BYTE_OF_UINT32(address);
	tx_gpr[2]= GET_SECOND_BYTE_OF_UINT32(address);
	tx_gpr[3]= GET_FIRST_BYTE_OF_UINT32(address);	// lsb

	// Setup the SPI xfer
    spi_ioc_transfer_info.len = length + 5;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)large_return_buf;

	// Send the SPI command
	int ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
	if (ioctl_ret == IOCTL_FAIL)
	{
        free(tx_gpr);
		return Status::CANCELLED;
	}

    printLongHex((const uint8_t *)&large_return_buf[5], 32 /* length */, 16, "Fast Read Restults");

    reply->set_buffer_size(length);
    reply->set_return_buf(&large_return_buf[5], length);  // Offset 5 bytes for return. Skip SPI TX bytes, 1 dummy.

    reply->set_status(STATUS_OK);

    free(tx_gpr);

    return Status::OK;
}

// Page_Program
grpc::Status NaxosServiceImpl::Page_Program(
    grpc::ServerContext *context,
    const flash_write_content *request,
    status_type_msg *reply)
{
    const char *funcName = "Page_Program";
    int fd = request->address().fd();
    int address = request->address().address();
    int length = request->address().length();
    const char *data = request->data().c_str();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

	uint8_t *tx_gpr = (uint8_t *)malloc((length + 4) * sizeof(uint8_t));
	if(tx_gpr == NULL)
	{
		return Status::CANCELLED;
	}

    memset(tx_gpr, 0x00, (length + 4) * sizeof(uint8_t));

	tx_gpr[0]=CMD_PAGE_PROGRAM;		// Opcode for page program

	tx_gpr[1]= GET_THIRD_BYTE_OF_UINT32(address);	// lsb
	tx_gpr[2]= GET_SECOND_BYTE_OF_UINT32(address);
	tx_gpr[3]= GET_FIRST_BYTE_OF_UINT32(address);

	// write the data into the transfer buffer
	for(int i=4, j=0 ; j<length; i++, j++)
	{
	    tx_gpr[i]=data[j];
	}

    printLongHex((const uint8_t *)data, length, 16, "Page Program Data");

	// Setup the SPI xfer
    spi_ioc_transfer_info.len = length + 4;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)large_return_buf;

	// Send the SPI command
	int ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
	if (ioctl_ret == IOCTL_FAIL)
	{
        free(tx_gpr);
		return Status::CANCELLED;
	}

    free(tx_gpr);

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// DYB_Write
grpc::Status NaxosServiceImpl::DYB_Write(
    grpc::ServerContext *context,
    const flash_lock_config *request,
    status_type_msg *reply)
{
    const char *funcName = "DYB_Write";
    int fd = request->address().fd();
    int address = request->address().address();
    int lockbit = request->lockbit();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[5] = {0x00, };
    uint8_t tx_gpr[5];

    tx_gpr[0] = 0xE5; // Opcode for lock block
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb
    tx_gpr[4] = lockbit ? 0x01 : 0x00;

    spi_ioc_transfer_info.len = 5;
    spi_ioc_transfer_info.tx_buf = (unsigned long)&tx_gpr[0];
    spi_ioc_transfer_info.rx_buf = (unsigned long)rx_gpr; 
    
    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Program
grpc::Status NaxosServiceImpl::PPB_Program(
    grpc::ServerContext *context,
    const flash_address *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Program";
    int fd = request->fd();
    int address = request->address();
    //int length = request->length();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[5] = {0x00, };
    uint8_t tx_gpr[5];

    tx_gpr[0] = 0xE3; // Opcode for lock block
    tx_gpr[1] = GET_FOURTH_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[4] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    /*printf("%s: sending ", funcName);
    for (int i = 0; i < 5; i++)
    {
        printf("0x%x ", tx_gpr[i]);
    }
    printf(" to fd %d at address %d and length %d\n", fd, address, length);
    */
    spi_ioc_transfer_info.len = 5;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)rx_gpr;
    
    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Erase
grpc::Status NaxosServiceImpl::PPB_Erase(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Erase";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0xE4;

    spi_ioc_transfer_info.len = 1;
    spi_ioc_transfer_info.tx_buf = (unsigned long) &tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) &rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Lock_Bit_Write
grpc::Status NaxosServiceImpl::PPB_Lock_Bit_Write(
    grpc::ServerContext *context,
    const int_msg *request,
    status_type_msg *reply)
{
    const char *funcName = "PPB_Lock_Bit_Write";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr = 0;
    uint8_t tx_gpr = 0xA6;

    spi_ioc_transfer_info.len = 1;
    spi_ioc_transfer_info.tx_buf = (unsigned long) &tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) &rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// DYB_Read
grpc::Status NaxosServiceImpl::DYB_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "DYB_Read";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[6];
    uint8_t tx_gpr[6];
    memset(rx_gpr, 0x00, sizeof(rx_gpr));
    memset(tx_gpr, 0x00, sizeof(tx_gpr));

    tx_gpr[0] = 0xE8; // Opcode for Read volatile lock bit
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    spi_ioc_transfer_info.len = 6;
    spi_ioc_transfer_info.tx_buf = (unsigned long) tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr[5], sizeof(rx_gpr));  // Offset 5 bytes for return. Skip SPI TX bytes, 1 dummy.
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Read
grpc::Status NaxosServiceImpl::PPB_Read(
    grpc::ServerContext *context,
    const flash_address *request,
    message_return *reply)
{
    const char *funcName = "PPB_Read";
    int fd = request->fd();
    int address = request->address();
    //int length = request->length();    
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[6];
    uint8_t tx_gpr[6];
    memset(rx_gpr, 0, sizeof(rx_gpr));
    memset(rx_gpr, 0, sizeof(tx_gpr));

    tx_gpr[0] = 0xE2; // Opcode for Read volatile lock bit
    tx_gpr[1] = GET_FOURTH_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[4] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    spi_ioc_transfer_info.len = 6;
    spi_ioc_transfer_info.tx_buf = (unsigned long) tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr[5], 1);   // Offset TX bytes and one dummy
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// PPB_Lock_Bit_Read
grpc::Status NaxosServiceImpl::PPB_Lock_Bit_Read(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "PPB_Lock_Bit_Read";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[3] = {0x00, 0x00, 0x00};
    uint8_t tx_gpr[3] = {0xA7, 0x00, 0x00};

    spi_ioc_transfer_info.len = 3;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr[2], 1);   // Offset TX bytes and 1 dummy
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Password_Read
grpc::Status NaxosServiceImpl::Password_Read(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Password_Read";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t tx_gpr[8+3];
    uint8_t rx_gpr[8+3];
    memset(rx_gpr, 0, sizeof(rx_gpr));
    memset(tx_gpr, 0x00, sizeof(tx_gpr));

    tx_gpr[0] = 0x27;
    tx_gpr[1] = 0x00;

    spi_ioc_transfer_info.len = 8+3;
    spi_ioc_transfer_info.tx_buf = (unsigned long) &tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::CANCELLED;
    }

    reply->set_buffer_size(8);
    reply->set_return_buf(&rx_gpr[2], 8);   // Offset TX byte and 1 dummy
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Read_Status_register
grpc::Status NaxosServiceImpl::Read_Status_register(
    grpc::ServerContext *context,
    const int_msg *request,
    message_return *reply)
{
    const char *funcName = "Read_Status_register";
    int fd = request->value();
    reply->set_status(STATUS_ERROR);
    reply->set_buffer_size(0);

    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    int ioctl_ret;
    uint8_t rx_gpr[3];
    uint8_t tx_gpr[3] = {CMD_READ_STATUS_REG, 0x00, 0x00};

    memset(rx_gpr, 0x00, sizeof(rx_gpr));

    spi_ioc_transfer_info.len = 3;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long)rx_gpr;

    ioctl_ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
    if (ioctl_ret == IOCTL_FAIL)
    {
        return Status::CANCELLED;
    }

    reply->set_buffer_size(1);
    reply->set_return_buf(&rx_gpr[2], 1);   // Offset TX byte and 1 dummy
    reply->set_status(STATUS_OK);

    return Status::OK;
}

// Subsector_Erase
grpc::Status NaxosServiceImpl::Subsector_Erase(
    grpc::ServerContext *context,
    const flash_address *request,
    status_type_msg *reply)
{
    const char *funcName = "Subsector_Erase";
    int fd = request->fd();
    int address = request->address();
    reply->set_status(STATUS_ERROR);
    
    if (!check_control(devices_ll_head, fd))
    {
        printf("%s: device busy\n", funcName);
        return Status::OK;
    }

    if (address > MAX_SPI_FLASH_ADDRESS)
    {
        printf("%s: address overrun\n", funcName);
        return Status::OK;
    }

    // Check flag status register properly
    status_type busyStat = waitForFSR(fd);
    if(STATUS_OK != busyStat)
    {
        return Status::CANCELLED;
    }

    lseek(fd, address, SEEK_SET);

    uint8_t tx_gpr[4];
    uint8_t rx_gpr[4];

    tx_gpr[0] = 0x20; // Opcode for subsector erase
    tx_gpr[1] = GET_THIRD_BYTE_OF_UINT32(address);
    tx_gpr[2] = GET_SECOND_BYTE_OF_UINT32(address);
    tx_gpr[3] = GET_FIRST_BYTE_OF_UINT32(address); // lsb

    spi_ioc_transfer_info.len = 4;
    spi_ioc_transfer_info.tx_buf = (unsigned long) tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) rx_gpr;

    int ret = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info); 
    
    // OD: calls spi_nor_erase, which in turn calls fsl_qspi_erase, which in turn calls runcmd
    // with opcode SPINOR_OP_BE_4K, or SPINOR_OP_BE_4K_PMC or SPINOR_OP_SE; SE=0xd8, SSE=0x20,
    // only one is supported at a time; forcing 0x20 by adding SECT_4K to n25q128 entries in spi-nor.c

    if (ret == IOCTL_FAIL)
    {
        return Status::OK;
    }

    reply->set_status(STATUS_OK);

    return Status::OK;
}

// static functions

// run authenta command
// set expected_nb_bytes to 0 if not want to check command length
static status_type authentaCmd(
    const naxos_content *cmd,
    const char *cmdName,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes)
{
    if (check_cmd_content(
            cmd,
            expected_opcode,
            expected_subopcode,
            expected_nb_bytes) != STATUS_OK)
    {
        printf("%s: command badly formed\n", cmdName);
        return STATUS_ERROR;
    }

    const char *bytes = cmd->bytes().c_str();

	// Check if the write is implying that there will be a subsequent read
	uint8_t opCode = bytes[0];	// Store the potential write opcode
	if(writeCommandIsRead(opCode))
	{
		// This is a "read" command

		// Store the opcode for the real read
		saveReadOpcode = opCode;

		// Set the command state to be ready for the read
		cmdState = CMD_STATE_RDY_FOR_READ;

		// Return without writing
		return STATUS_OK;
	}

    // Check flag status register properly
    status_type busyStat = waitForFSR(cmd->fd());
    if(STATUS_OK != busyStat)
    {
        return STATUS_ERROR;
    }

	// This is an actual write command, not a setup for a read
	// Reset the state
	cmdState = CMD_STATE_RDY_FOR_WRITE;
	saveReadOpcode = CMD_OPCODE_NONE;

    spi_ioc_transfer_info.len = cmd->nb_bytes();
    spi_ioc_transfer_info.tx_buf = (unsigned long) bytes;
    spi_ioc_transfer_info.rx_buf = (unsigned long) auth_resp_buf;

    printf("%s: executing command\n", cmdName);

    if (ioctl(cmd->fd(), SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info) < 0)
    {
        printf("%s: failed excuting command using fd %d\n",
               cmdName, cmd->fd());
        return STATUS_ERROR;
    }
    else
    {
        printf("%s: sent successfully\n", cmdName);
    }
    return STATUS_OK;
}

// check command opcode, subopcode, and length
// if expected_nb_bytes is set to 0, not check command length
static status_type check_cmd_content(
    const naxos_content *cmd,
    char expected_opcode,
    char expected_subopcode,
    int expected_nb_bytes)
{
    int nb_bytes = cmd->nb_bytes();
    const char *bytes = cmd->bytes().c_str();

    if (expected_nb_bytes && nb_bytes != expected_nb_bytes)
    {
        return STATUS_ERROR;
    }
    if (bytes[0] != expected_opcode)
    {
        return STATUS_ERROR;
    }
    if (bytes[1] != expected_subopcode)
    {
        return STATUS_ERROR;
    }

    return STATUS_OK;
}

// call NAXOS_WRITE_ENABLE and get return code
static int writeEnable(int fd)
{
    uint8_t tx_gpr = 0x06;
    uint8_t rx_gpr = 0;

    spi_ioc_transfer_info.len = 1;
    spi_ioc_transfer_info.tx_buf = (unsigned long) &tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) &rx_gpr;
    spi_ioc_transfer_info.delay_usecs = spiDelay;
    spi_ioc_transfer_info.speed_hz = spiSpeed;
    spi_ioc_transfer_info.bits_per_word = spiBits;

    return ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
}

// call NAXOS_WRITE_DISABLE and get return code
static int writeDisable(int fd)
{
    uint8_t tx_gpr = 0x04;
    uint8_t rx_gpr = 0;

    spi_ioc_transfer_info.len = 1;
    spi_ioc_transfer_info.tx_buf = (unsigned long) &tx_gpr;
    spi_ioc_transfer_info.rx_buf = (unsigned long) &rx_gpr;
    spi_ioc_transfer_info.delay_usecs = spiDelay;
    spi_ioc_transfer_info.speed_hz = spiSpeed;
    spi_ioc_transfer_info.bits_per_word = spiBits;

    return ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
}

// Read the flag status register (FSR)
// Result placed in *fsr
// Returns error code TBD
static int readFlagStatusRegister(int fd, uint8_t *fsr)
{
    uint8_t rx[2] = { 0x00, 0x00 };
    uint8_t tx[2] = { CMD_READ_FLAG_STATUS_REG, 0x00 };

    spi_ioc_transfer_info.len = 2;
    spi_ioc_transfer_info.tx_buf = (unsigned long)tx;
    spi_ioc_transfer_info.rx_buf = (unsigned long)rx;
    
    int ioctlErr = ioctl(fd, SPI_IOC_MESSAGE(1), &spi_ioc_transfer_info);
   
    // Flag Status Register is in first returned byte (second SPI xfer)   
    *fsr = rx[1];

    return ioctlErr;
}

/**
 * Check if the op code for a particular command is one which will likely be followed by a read command.
 */
static bool writeCommandIsRead(uint8_t opCode)
{
	bool isRead = false;

	switch(opCode)
	{
	case CMD_READ_ID:
	case CMD_READ_ID_ALT:
	case CMD_READ_STATUS_REG:
	case CMD_READ_FLAG_STATUS_REG:
	case CMD_READ_GEN_PRP_REG:
		isRead = true;
		break;
	default:
		isRead = false;
		break;
	}

	return isRead;
}

/**
 * Continually read the Flag Status Register until Naxos is ready, or too many tries
 *
 * @return STATUS_OK or STATUS_ERROR if too many tries
 */
static status_type waitForFSR(int fd)
{
    status_type waitErr = STATUS_ERROR;

	// Read the FSR many times until we're not busy, or there is an error
	for(uint32_t tries = 0; tries < MAX_FSR_TIMEOUT_RETRIES; tries++)
	{
        // Read the flag status register
        uint8_t fsr;
        int err = readFlagStatusRegister(fd, &fsr);
        if(IOCTL_FAIL == err)
        {
            return STATUS_ERROR;
        }

		// Is it "ready"?
		if( (fsr & FLAG_STATUS_READY_MASK) == FLAG_STATUS_READY_MASK)
		{
			// Set the error code to OK and break out of loop
			waitErr = STATUS_OK;
			break;
		}
	}

	return waitErr;
}

// Print a long string of bytes as hex in neatly-formatted column.
// param[in] data   The data to print
// param[in] len    Number of bytes in data
// param[in] widthInBytes   Number of bytes to display per row (i.e. 0, 8, 16, 23). Will actually display twice this number of characters on screen.
// param[in] title  Title to display. Optional. Set to NULL to skip.
// NOTE. Set width to 0 to use indefinitely row lenghts.
static void printLongHex(const uint8_t * const data, const uint32_t len, const uint8_t widthInBytes, const char *title)
{
    if (NULL != title)
    {
        printf("%s:", title);
    }

    if (0 != widthInBytes)
    {
        printf("\r\n");
    }

    for (uint32_t index = 0; index < len; index++)
    {
        if ((0 != widthInBytes) &&
            (0 == (index % widthInBytes) &&
            (0 != index) ) )
        {
            printf("\r\n");
        }

        printf("%02X", data[index]);
    }

    printf("\r\n");
}
