#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging
from sys import exit
from collections import namedtuple

# --- local imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

from authenta.indigo import utils
import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for getting the Naxos and Indigo devices status in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:
    # commandline using default mapping files
    $> app_get_status.py <device alias> <sig_type>
     -- as in --
    $> app_get_status.py snor1 mmc
     - where 'snor1' is a Naxos device entry in the devices.json file
     - and 'mmc' is a Naxos device entry in the devices.json file


EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure
    9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Tatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2020

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('naxos_device', metavar="<naxos device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('indigo_device', metavar="<indigo device alias>",
               help="endpoint-device alias found in devices.json file.")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():

    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    # Get Naxos device endpoint
    naxos_ep = get_ep(ARGS.naxos_device)
    naxos_device = NaxosDevice(naxos_ep)

    # Get Naxos Device Status
    get_status(naxos_device, 'Authenta SNOR', main_logger)

    print("************************************\n")
    
    # Get Indigo device endpoint
    indigo_ep = get_ep(ARGS.indigo_device)
    indigo_device = IndigoDevice(indigo_ep)

    #Get Indigo Device Status
    get_status(indigo_device, 'Authenta MMC', main_logger)

def get_status(device, device_type, logger):
    with device:

        # Run request_uid
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)

        print("\n{} Device Status:".format(device_type))
        print("UID: {:X}".format(device.uid))
        print("Revision: {:X}".format(device.revision))
        print("provision_sts_struct.Provisioned: {}\n".format(device.provision_sts_struct.Provisioned))

def get_ep(device_name):
    ARGS.ifn.seek(0) # start from the beggining of the file
    envr = Envr(device_name, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]


    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    return ep

if __name__ == "__main__":
    main()