#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.indigo import utils
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo.utils import CRTM_Registers
from authenta import authenta_key_mgmt_interface

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for setting up a slave key (randomly generated by device) and use it to 
   sign data and verify the signature in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_create_random_kts_slave_and_sign.py <device alias> <sig_type>
     -- as in --
    $> app_create_random_kts_slave_and_sign.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and ECDSA is the signature type

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Tatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: {}'.format(err.error_msg))
            exit(2)

        print("\nDevice UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # Check if device is not provisioned. If not provisioned, exit the app.
        if not device.provision_sts_struct.Provisioned:
            print("provision_sts.Provisioned bit unset.")
            print("Device shall be provisioned prior to attempting this app.")
            print("... exiting on error.")
            main_logger.critical("Device is not provisioned.")
            exit(9)

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x48'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x48'

        byte_addr = 0x5
        test_data = b'\x12'*32

        # Upload ECDSA_KTS_PuK
        # (TRD for 0.4) specifies in "3.4.11 Manage_KTS" that "The very first time a keyslot is 
        # populated with a key, the ECDSA_KTS_PuK must be uploaded first".
        key_name = "ecdsa_pub_slot_{}".format(byte_addr)
        ECDSA_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDSA_KTS_PuK_key is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, byte_addr, ECDSA_KTS_PuK_key[::-1])
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded ECDSA public key to key table')

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\xA0'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\xA0'

        # Create a slave key in slot 5
        err, _ = device.meta_manage_kts(config, byte_addr, b"\x00"*64)
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Created ECDSA private key (slave key) in key table')

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x44'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x44'

        # Upload ECDH public key for hmac sigtype option
        key_name = "ecdh_pub_slot_{}".format(byte_addr)
        ECDH_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDH_KTS_PuK_key is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, byte_addr, ECDH_KTS_PuK_key[::-1])
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded ECDH key to key table')

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x85\x10'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x05\x10'

        # Sign data using the slave key just created
        err, _ = device.meta_create_signature(config, byte_addr, 0x0, 0x0, 0x0, test_data)
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('create_signature: {}'.format(err.error_msg))
            exit(1)

        print('Created signature using ECDSA Private key')

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x00'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x00'

        # Read signature from KTR
        err, response = device.meta_crtm_register_read(config, CRTM_Registers.KTR_Base)
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: {}'.format(err.error_msg))
            exit(1)

        reg_0_parsed_response = device._crtm_register_0_read_parsed_response(response)
        ktr_value = reg_0_parsed_response['Data']['KTR']
        print('Signature to verify: {:X}'.format(ktr_value))

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x01'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x01'

        err, response = device.meta_manage_kts(config, byte_addr, b'\x00' * 64)
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        slave_pub_key = response[402:(402 + 64)][::-1]

        print('Got public key associated with the key table slot slave private key')

        # Write slave public key to flash (partition_key_src_addr)
        # because KTR has been occupied by signature;

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x81\x10'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x01\x10'

        partition_key_src_addr = 0xf20000
        data = b'\x00' * (512 - 64) + slave_pub_key

        ret = device.meta_write_multiple_block(partition_key_src_addr, data)
        if not ret:
            print("\n\n")
            main_logger.critical('meta_write_multiple_block failed')
            exit(1)

        print('Written slave public key to flash: {:X}'.format(partition_key_src_addr))

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x81\x38'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x01\x38'

        # Verify signature
        err, response = device.meta_verify_signature(config, 0x0, 0x0, partition_key_src_addr, 0x0, test_data)
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('verify_signature: {}'.format(err.error_msg))
            exit(1)

        if response[226:(226 + 256)] == b'\x00' * 256:
            print('Signature is verified')
        else:
            print('Signature is not verified')

    exit(0)

if __name__ == "__main__":
    main()
