#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.indigo import utils
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo.utils import CRTM_Registers

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for generating DICE key based on measurement in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_establish_dice_cert.py <device alias> <sig_type>
     -- as in --
    $> app_establish_dice_cert.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and ECDSA is the signature type

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Tatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: {}'.format(err.error_msg))
            exit(2)

        print("\nDevice UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # Check if device is not provisioned. If not provisioned, exit the app.
        if not device.provision_sts_struct.Provisioned:
            print("provision_sts.Provisioned bit unset.")
            print("Device shall be provisioned prior to attempting this app.")
            print("... exiting on error.")
            main_logger.critical("Device is not provisioned.")
            exit(9)

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x01'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x01'

        # Create HMAC_CID and Device ID Key (Measure Option #1)
        err, response = device.meta_measure(config, 0x0, 0x0, 0x0, b"\x00"*32)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('measure: {}'.format(err.error_msg))
            exit(1)

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x00'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x00'

        # Read HMAC_CID from KTR
        err, response = device.meta_crtm_register_read(config, 0xB0)
        reg_0_parsed_response = device._crtm_register_0_read_parsed_response(response)
        HMAC_CID = reg_0_parsed_response['Data']['KTR'].hex()

        # Read ECDSA_DID_PuK from CRTM Register
        err, response = device.meta_crtm_register_read(config, 0x980)
        ECDSA_DID_PuK = device._ecdsa_did_public_key.hex()

        # Lock UDS (call measure option #2)
        # Call not implemented due to bug in v.0.4 (UDS cannot be unlock after this command call)

        print("\n\n")
        print('HMAC_CID: {}'.format(HMAC_CID))
        print('ECDSA_DID_PuK: {}'.format(ECDSA_DID_PuK))
        main_logger.info('HMAC_CID: {}'.format(HMAC_CID))
        main_logger.info('ECDSA_DID_PuK: {}'.format(ECDSA_DID_PuK))

    exit(0)

if __name__ == "__main__":
    main()
