#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging
from sys import exit
from collections import namedtuple

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

from authenta.indigo import utils
import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for replacing the SRK secret in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:
    # commandline using default mapping files
    $> app_get_status.py <device alias> <sig_type>
     -- as in --
    $> app_get_status.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'ECDSA' is the signature type


EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure
    9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Run request_uid
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)

        if not device.provision_sts_struct.Provisioned:
            print("\n\n")
            print("UID = " + device.uid.hex())
            print("Revision = " + device.revision.hex())
            print("Device is not provisioned:... Nothing left to do. Exiting...\n")
            main_logger.info("Device is not provisioned.")
            exit(9)

        err, _ = device._read_extended_crtm_status()
        if not err.esr_success:
            main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        # read all of the registers
        crtm_response_bytes = list()

        CRTM_REG_SZ = 0xD00
        CRTM_REG_READ_BLK_SZ = 0x100

        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
        else:
            main_logger.critical('Error: Unrecognized signature type: {}'.format(sig_type))
            exit(1)

        # call a signed function for the case in which we don't have the 
        # correct SRK in the key library we should avoid locking the chip because 
        # of too many consecutive non-signed function calls
        err, _ = device.meta_acrtm_update_finish(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.warning('Call to meta_acrtm_update_finish failed: {}; {}; (non-critical call - continuing...)'.format(err.esr_bit_msg, err.error_msg))


        # Read Register 0
        err, response = device.meta_crtm_register_read(config, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_read: ' + err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        register_response = device._crtm_register_0_read_parsed_response(response)

        # add data from Register 0
        device_attr = register_response['Data']

        # add device provision status bits
        device_attr.update(device.provision_sts_struct._asdict())

        # add device Extended CRTM Flag bits
        device_attr.update(device.extended_crtm_status_struct._asdict())

        # add device config
        device._set_device_config(register_response['Data']['Device_Config'])
        device_attr.update(device.device_config_struct._asdict())

        # add device config volatile
        device._set_device_config_volatile(register_response['Data']['Device_Config_Volatile'])
        device_attr.update(device.device_config_volatile_struct._asdict())

        # add device UID and Revision
        device_attr['UID'] = device.uid.hex()
        device_attr['Revision'] = device.revision.hex()

        # read Customer_RAM from register
        err, response = device.meta_crtm_register_read(config, 0x300)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_read: ' + err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        parsed_response = device._parse_response_bytes(response)
        device_attr['Customer_RAM'] = parsed_response['Data'][0x0:0x60]

        # add CRTM Register Customer RAM Size Data
        device_attr['Customer_RAM_size'] = len(device_attr['Customer_RAM'])

        # make the byte arrays easier to read
        for key, value in device_attr.items():
            try:
                if isinstance(value, bytes):
                    # if the length is 4 bytes display as a integer
                    if len(value) == 4:
                        device_attr[key] = '0x'+value.hex()
                    else:
                        device_attr[key] = value.hex()
            except ValueError:
                continue

        # print device status form to stdout
        print(device.DEVICE_STATUS_FORM.format(**device_attr))



    exit(0)

if __name__ == "__main__":
    main()
