#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility to set a block to be unlocked/read locked using a particular key
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_lock_with_key.py <device alias> <select option> <sig_type>
     -- as in --
    $> app_lock_with_key.py snor1 lock ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'lock' is one of the function options of this app (locks with key)
     - and ECDSA is the signature type


EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID error

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('option', metavar="<select option>", choices=['lock', 'unlock'],
               help="Choose which function to run: 'lock' or 'unlock'")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type
        option = ARGS.option

        # Call Request_UID to get device info
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        # Define some secret data. This can be anything that is 512 Bytes
        SAMPLE_SECRET_DATA = b'\x11'*12 + str.encode("This is the Test Secret Digest") + b'\x11'*470

        # WPG Address
        wpg_addr = 0xf00000

        # key slot will be ignored but must be passed as an argument
        key_slot = 0x0

        if option == 'lock':
            # Write the secret to the data area
            response = device.meta_write_multiple_block(wpg_addr, SAMPLE_SECRET_DATA)
            if not response:
                print("\n\n")
                main_logger.critical('write_multiple_block: failed to write data')
                exit(1)

            print('Written secret data to 0x{:X}'.format(wpg_addr))

            # Read the secret data
            ret, response = device.meta_read_multiple_block(wpg_addr, 512)
            if not ret:
                print("\n\n")
                main_logger.critical('read_multiple_block: failed to read data')
                exit(1)

            print('Read secret data from 0x{:X}'.format(wpg_addr))

            # Read Lock the area by calling APA_Modify
            # Note: Done once to set up the locking 

            # Set up config
            # Config[15] = 0/1, SigType; Config[14] = 0, KeyRef = Host Server Key Only;
            # Config[2] = 1, WriteAPA = write
            if sig_type == "HMAC":
                config = b'\x00\x04'
            elif sig_type == "ECDSA":
                config = b'\x80\x04'

            # Set Authenta Access Rights to read protect the Region; and use the Active Host 
            # Server key to protect the APA_Data Structure
            # AuthentaAccessRights[15] = 0, APAWriteProtect = Unprotected
            # AuthentaAccessRights[14] = 1, APAReadProtect = Protected
            # AuthentaAccessRights[9] = 0, EnableKTSKey = Host Server Key Only
            authenta_access_rights = b'\x40\x00'

            apa_data = b'\x00'*72 + authenta_access_rights[::-1] + b'\x00'*6

            # Call APA_Modify (to lock)
            err, _ = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('apa_modify: ' + err.error_msg)
                exit(1)

            print('Issued APA_Modify to make 0x{:X} read-protected'.format(wpg_addr))

            # Attempt to Read the Secret. This should Fail
            ret, response = device.meta_read_multiple_block(wpg_addr, 512)
            if ret:
                print("\n\n")
                main_logger.critical('read_multiple_block was able to read the Secret. Should Fail.')
                exit(1)

            print('Verified that 0x{:X} is read-protected'.format(wpg_addr))

        elif option == 'unlock':
            # Unlock the Block by calling APA_Modify
            # Note: Done once to unlock

            # Set up the config
            # Config[15] = 0/1, SigType; Config[14] = 0, KeyRef = Host Server Key Only
            # Config[2] = 1, WriteAPA = write
            if sig_type == "HMAC":
                config = b'\x00\x04'
            elif sig_type == "ECDSA":
                config = b'\x80\x04'
            
            # Set Authenta Access Rights to unlock the Region; and use the Active Host 
            # Server key to protect the APA_Data
            # AuthentaAccessRights[15] = 0, APAWriteProtect = Unprotected
            # AuthentaAccessRights[14] = 0, APAReadProtect = Unprotected
            # AuthentaAccessRights[9] = 0, EnableKTSKey = Host Server Key Only
            authenta_access_rights = b'\x00\x00'

            apa_data = b'\00'*72 + authenta_access_rights[::-1] + b'\x00'*6
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('apa_modify: ' + err.error_msg)
                exit(1)

            # wpg_addr and key_slot values are same as above

            # call APA_Modify (to unlock)
            err, _ = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical("apa_modify: " + err.error_msg)
                exit(1)

            print('Issued APA_Modify to make 0x{:X} not read-protected'.format(wpg_addr))

            # Attempt to Read the Secret. This secret should be unlocked 
            ret, response = device.meta_read_multiple_block(wpg_addr, 512)
            if not ret:
                print("\n\n")
                main_logger.critical('read_multiple_block: failed to read')
                exit(1)

            print('Verified that 0x{:X} is not read-protected'.format(wpg_addr))

    exit(0)

if __name__ == "__main__":
    main()
