#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility to set a block to be read locked unless device is in a particular PCR state
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_pcr_state_read_lock.py <device alias> <select option> <sig_type>
     -- as in --
    $> app_pcr_state_read_lock.py snor1 write ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'write' is one of the function options of this app (writes the secret data)
     - and ECDSA is the signature type


EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID error

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('option', metavar="<select option>", choices=['setup', 'write', 'lock', 'read'],
               help="Choose which function to run: 'setup' function, 'write' secret, 'lock' area, or 'read' secret")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")


ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type
        option = ARGS.option

        # Call Request_UID to get device info
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        # Define some secret data
        SAMPLE_SECRET_DATA = b'\x02'*512

        # Define the golden PCR value
        golden_PCR_value = b'\x00'*32

        # WPG Address
        wpg_addr = 0xf80000

        # First Option: Run Set Up Function (command-line: 'setup')
        if option == 'setup':
            # Run APA_Modify to set up the function
            # Note: Done once to set up the locking

            # Config set up
            # Config[15] = 0/1, sig type; Config[14] = 0, KeyRef = Host Server Key only; 
            # Config[2] = 1, WriteAPA = write
            if sig_type == 'HMAC':
                config = b'\x00\x04'
            elif sig_type == 'ECDSA':
                config = b'\x80\x04'

            # ReadPCRCompare is set to the previsously defined golden PCR value
            read_pcr_compare = golden_PCR_value

            # Set the ReadPCRConfig to read protect the Region based on the value of a given 
            # PCR matches the comparison value
            # ReadPCRConfig[15] = 1, PCRReadProtectEnable = Enabled;
            # ReadRCRConfig[8] = 1, PCRReadAccess = Protected;
            # ReadPCRConfig[4] = 0, PCRReadOperator = Not Equal;
            # ReadPCRConfig[2:0] = 001, PCRReadReference. Refers to register PCR1
            read_pcr_config = b'\x81\x01'

            # Set the AuthentaAccessRights to read protect the Region; and use Active Host 
            # Server key to protect the APA_Data structure
            # AuthentaAccessRights[15] = 0, APAWriteProtect = Unprotected;
            # AuthentaAccessRights[14] = 1, APAReadProtect = Protected
            # AuthentaAccessRights[9] = 0, EnableKTSKey = Host Server Key only
            authenta_access_rights = b'\x40\x00'

            # Combine the above elements to get the apa_data
            apa_data = b'\x00'*32 + read_pcr_compare[::-1] + b'\x00'*2 + b'\x00'*2 + read_pcr_config[::-1] + b'\x00'*2 + authenta_access_rights[::-1] + b'\x00'*6

            # key slot will be ignored but must be passed as an argument
            key_slot = 0x0

            err, response = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('apa_modify: ' + err.error_msg)
                exit(1)

            print('Issued APA_Modify to setup PCR read protection for 0x{:X}'.format(wpg_addr))

        # Second Option: Write defined secret to defined secret data area
        elif option == 'write':
            # Write the secret to the data area
            response = device.meta_write_multiple_block(wpg_addr, SAMPLE_SECRET_DATA)
            if not response:
                print("\n\n")
                main_logger.critical('write_multiple_block: failed to write data')
                exit(1)

            print('Written Data to {:X}: {:X}'.format(wpg_addr, SAMPLE_SECRET_DATA))

        # Third Option: Read secret and then lock the data area for read
        elif option == 'lock':
            # Read the secret data
            ret, response = device.meta_read_multiple_block(wpg_addr, 512)
            if not ret:
                print("\n\n")
                main_logger.critical('read_multiple_block')
                exit(1)
            print('Read Data from {:X}: {:X}'.format(wpg_addr, response))

            # Extend PCR1 using PCR_Extend
            # Config
            if sig_type == 'HMAC':
                config = b'\x00\x00'
            elif sig_type == 'ECDSA':
                config = b'\x80\x00'

            # Address of PCR1
            address = 0x1

            # The input digest can be any 32-byte value
            input_digest = b'\x56'*32

            err, response = device.meta_pcr_extend(config, address, input_digest)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('pcr_extend: ' + err.error_msg)
                exit(1)
            
            print('Extended PCR')

            # Note: Once the extend is completed, the secret should be read-locked. 
            #       'Read Secret' should fail if run

        elif option == 'read':
            # Read the secret data (should fail after 'lock')
            ret, response = device.meta_read_multiple_block(wpg_addr, 512)
            if ret:
                print("\n\n")
                main_logger.critical('read_multiple_block should FAIL after lock')
                exit(1)
            print('Read Data from {:X} failed because it is read-locked'.format(wpg_addr))
        
    exit(0)

if __name__ == "__main__":
    main()
