#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility to set a block to be locked unless device is in a particular PCR state
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_pcr_state_write_lock.py <device alias> <sig_type>
     -- as in --
    $> app_pcr_state_write_lock.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and ECDSA is the signature type


EXIT CODES:
  0 : success
  1 : general application error

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")


ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get device info
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        # Get the comparison digest value
        # Note: done once before APA_Modify to get the right PCR0 value, then done on 
        # each reset in preboot.

        # Call measure to create a digest that includes CRTM digest and stores it in KTR
        # Config[15] = 0/1, Config[10] = 0, Config[9:8] = 00, Config[7] = 0, Config[6] = 0, 
        # Config[5:3] = 001, Config[2] = 0, Config[1:0] = 00
        if sig_type == 'HMAC':
            config = b'\x00\x08'
        elif sig_type == 'ECDSA':
            config = b'\x80\x08'
        partition_altsource_addr = 0x0
        partition_keysource_addr = 0x0
        partition_dest_addr = 0x0
        input_digest = b'\x00'*32

        err, _ = device.meta_measure(config, partition_altsource_addr, partition_keysource_addr, partition_dest_addr, input_digest)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('measure: ' + err.error_msg)
            exit(1)

        print('Issued Meausre command to create a digest')

        # Get measurement digest from KTR
        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
        crtm_addr = 0xB0
        err, response = device.meta_crtm_register_read(config, crtm_addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_read: ' + err.error_msg)
            exit(1)
        reg_0_parsed_response = device._crtm_register_0_read_parsed_response(response)
        input_digest = reg_0_parsed_response['Data']['KTR']

        print('Read Measure result from KTR')

        # Extend the value read from KTR into PCR0
        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
        address = 0xC00 # address of PCR0
        err, _ = device.meta_pcr_extend(config, address, input_digest)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('pcr_extend: ' + err.error_msg)
            exit(1)

        print('Extended Measure result to PCR0')

        # Read the value from PCR0 for comparison
        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
        crtm_addr = 0xC00
        err, response = device.meta_crtm_register_read(config, crtm_addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_read: ' + err.error_msg)
            exit(1)
        pcr0_value = response[450:482]
        print('PCR0: {:X}'.format(pcr0_value))

        # Set up attributes for APA_Modify
        # config setup
        # Config[15] = 0/1, HMAC/ECDSA sig;
        # Config[14] = 0, KeyRef = Host Server Key;
        # Config[2] = 1, WriteAPA = write
        if sig_type == 'HMAC':
            config = b'\x00\x04'
        elif sig_type == 'ECDSA':
            config = b'\x80\x04'

        # WPG_Address - test address to be protected
        wpg_addr = 0xf90000

        # WritePCRCompare: use the comparison value read from PCR0
        write_pcr_compare = pcr0_value

        # Set the WritePCRConfig to write protect the Region based on the value of a 
        # given PCR matches the comparison value
        # WritePCRConfig[15] = 1, PCRWriteProtectEnable = Enabled;
        # WritePCRConfig[8] = 1, PCRWriteAccess = Protected
        # WritePCRConfig[4] = 1, PCRWriteOperator = Equal;
        # WritePCRConfig[2:0] = 000, PCRWriteReference. Refers to register PCR0
        write_pcr_config = b'\x81\x10'

        # Set the AuthentaAccessRights to write protect the Region; and use the 
        # Active Host Server key to protect the APA_Data structure
        # AuthentaAccessRights[15] = 1, APAWriteProctect = Protected;
        # AuthentaAccessRights[14] = 0, APAReadProtect = Unprotected
        # AuthentaAccessRights[9] = 0, EnableKTSKey = Host Server Key Only
        authenta_access_rights = b'\x80\x00'

        apa_data = write_pcr_compare[::-1] + b'\x00'*32 + write_pcr_config[::-1] + b'\x00'*2 + b'\x00'*2 + b'\x00'*2 + authenta_access_rights[::-1] + b'\x00'*6

        # key slot will be ignored but must be passed as an argument
        key_slot = 0x0

        # Call APA_Modify
        # Note: Done once to set up the locking
        err, _ = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: ' + err.error_msg)
            exit(1)

        print('Issued APA_Modify to setup PCR write protection for 0x{:X}'.format(wpg_addr))

    exit(0)

if __name__ == "__main__":
    main()
