#! /usr/bin/env python3
import argparse as ap
import sys
import os

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
\t A utility for retrieving current device state of Micron Authenta enabled SPI-NOR device
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_prov_check.py  <endpoint alias> 
    -- as in --
   $> app_prov_check.py snor1
    - where 'snor1' is a device entry in the devices.json file

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<device json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

# p.add_argument("-c", dest='cfn', metavar="<CRTM_Reg_Attributes.json>",
#                default="ACRTM_Registers_v1_4.json", type=ap.FileType('r'),
#                help="CRTM_Reg Attributes json mapping file. See figure 7 in Authenta Addendum for full list of table attributes.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main() -> None:
#    envr = Envr(ARGS.device, ARGS.ifn, ARGS.cfn)
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        err, response = device.meta_request_uid()
        if not err.esr_success:
            print('ERROR request_uid: (ESR=' + err.esr.hex() + ') ' + err.esr_bit_msg + '\n' + err.error_msg)
            exit(2)

        print("Request_UID Response Signature Valid: {} ".format(err._sig_matched))
        print("Request_UID Response Nonce Matched: {} ".format(err._nonce_matched))
        print("\n\nUID: {}".format(device.uid.hex()))
        print("Revision: {}".format(device.revision.hex()))
        request_uid_parsed_response = device._request_uid_response_bytes(response)
        resp_ecdsa_pub = request_uid_parsed_response['Data']['ECDSA_DS_PuK'][::-1]
        if resp_ecdsa_pub is not None:
            print("ECDSA_DS_PuK (from resp): {}".format(resp_ecdsa_pub.hex()))
        else:
            print("Failed to read ECDSA_DS_PuK, (the device specific ECDSA public key), from the device.")

        print("\nProvision Status")
        print("device status Provisioned: {} ".format(device.provision_sts_struct.Provisioned))
        print("device status SRK_Replaced: {} ".format(device.provision_sts_struct.SRK_Replaced))
        print("device status Remotely_Provisioned: {} ".format(device.provision_sts_struct.Remotely_Provisioned))
        print("device status HSM_Present: {} ".format(device.provision_sts_struct.HSM_Present))
        print("device status RIOT_Mode: {} ".format(device.provision_sts_struct.RIOT_Mode))
        print("device status UDS_Replaced: {} \n\n".format(device.provision_sts_struct.UDS_Replaced))


        # if not device.provision_sts_struct.Provisioned:
        #     err, response = device.meta_provision_crtm(b'\x80\x05')
        #     if err.esr_success:
        #         print("Successfully provisioned the device.")
        #     else:
        #         print("Device provisioning FAILED with ESR {}".format(err.esr))
        # else:
        #     err, response = device.meta_deprovision_crtm(b'\x80\x00')
        #     if err.esr_success:
        #         print("Successfully deprovisioned the device.")
        #     else:
        #         print("Device deprovisioning FAILED with ESR {}".format(err.esr))

    exit(0)

if __name__ == "__main__":
    main()
