#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

from authenta.indigo import utils
import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for provisioning a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_provision.py <device alias> <prov_method>
     -- as in --
    $> app_provision.py snor1 remote
     - where 'snor1' is a device entry in the devices.json file
     - and 'remote' is the provisioning method

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : provisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('prov_method', metavar="<prov_method>", choices=['remote', 'local'],
               help="provision for remote use or local use")

# command-line optional arguments
p.add_argument('-hsm', action='store_true', help="Indicates that an HSM is present: ")

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        hsm = ARGS.hsm
        prov_method = ARGS.prov_method

        # Get config value based on command-line arguments
        if prov_method == 'remote':
            # Remotely Provisioned, with Riot mode enabled
            config = b'\x00\x05'
        elif prov_method == 'local':
            if hsm:
                # Locally Provisioned with HSM_Present, with Riot mode enabled
                config = b'\x00\x06'
            else:
                # Locally Provisioned with no HSM_Present, with Riot mode enabled
                config = b'\x00\x04'

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        print("\nDevice UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # Check if device is already provisioned. If already provisioned, exit the app.
        if device.provision_sts_struct.Provisioned:
            print("\nDevice is already provisioned.\n")
            main_logger.critical("Device is already provisioned.")
            exit(9)

        # Provision the device
        err, _ = device.meta_provision_crtm(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('provision_crtm: ' + err.error_msg)
            exit(1)

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        print("\nThe device has been successfully provisioned\n")

    exit(0)

if __name__ == "__main__":
    main()
