#! /usr/bin/env python3
import argparse as ap
from logging import error
import sys
import os
import json
import logging
from sys import exit
from collections import namedtuple

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

from authenta.indigo import utils
import app_common_logging as app_logging

import authenta.authenta_key_mgmt_interface as key_mgmt_lib


VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for replacing the SRK secret in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:
    # commandline using default mapping files
    $> app_replace_srk.py <device alias> <sig_type>
     -- as in --
    $> app_replace_srk.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'ECDSA' is the signature type

    # for test cases where the default devices.json file isn't what is needed
    $> app_replace_srk.py -f mydevices_mapping.json snor1 1141 1143

EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure
    9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        # Check if device is remotely provisioned, SRK can only be replaced while remotely provisioned
        if not device.provision_sts_struct.Provisioned:
            print("\nprovision_sts.Provisioned bit unset.")
            print("Device is not provisioned. It must be Remotely Provisioned to run Replace_SRK.")
            print("Exiting...\n")
            main_logger.critical("Device is not Povisioned.")
            exit(9)
        if not device.provision_sts_struct.Remotely_Provisioned:
            print("\nprovision_sts.Remotely_Provisioned bit unset.")
            print("Device is Locally Provisioned. It must be Remotely Provisioned run Replace_SRK!")
            print("Exiting...\n")
            main_logger.critical("Device is not Remotely Provisioned.")
            exit(9)

        # Config
        # SigType: from command line
        # ECDSA_ECDH#: Replace host ECDSA key (config[2] = `1b)
        # Command option: #0 (config[1:0] = `00b)
        if sig_type == 'HMAC':
            config = b'\x00\x04'
        elif sig_type == 'ECDSA':
            config = b'\x80\x04'
        else:
            # shouldn't be able to reach here
            print('ERROR unrecognized signature type {}!'.format(sig_type))
            exit(1)

        new_ecdsa_pub_key_name = 'new_ecdsa_pub'
        print('reading public key for UID {}; key name {}'.format(device.uid.hex(), new_ecdsa_pub_key_name))
        new_ecdsa_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdsa_pub_key_name)
        if new_ecdsa_pub_key is None:
            main_logger.critical('Failed to read new ECDSA public key using Key Management Library: UID: {}; Key Name: {}'.format(device.uid.hex(), new_ecdsa_pub_key_name))
            exit(1)

        err, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdsa_pub_key[::-1])
        if not err.valid_response:
            main_logger.critical('replace_srk: ' + err.esr_bit_msg + ' - ' + err.error_msg)
            exit(1)

        ecdsa_keys_changed = False
        if active_key_updated is not None:
            # the active ecdsa key has been changed
            print("updating keys before replace_srk response validation")
            ecdsa_keys_changed = copy_key(device, 'ecdsa_prv', 'previous_ecdsa_prv', 'new_ecdsa_prv')
            if not ecdsa_keys_changed:
                print("ERROR changing keys. Continuing to attempt validation, but expecting error!")


        ecdsa_resp_err = device.verify_replace_srk_response(config, response, sent_nonce, active_key_updated)
        if not ecdsa_resp_err.valid_response:
            main_logger.critical('replace_srk response: ' + ecdsa_resp_err.esr_bit_msg + ' - ' + ecdsa_resp_err.error_msg)
            if ecdsa_keys_changed:
                print("changing back to old ECDSA private key due to replace_srk response validation failure")
                ecdsa_keys_changed = copy_key(device, 'previous_ecdsa_prv', 'tmp_ecdsa_prv', 'ecdsa_prv')
            exit(2)

        # Config
        # SigType: from command line
        # ECDSA_ECDH#: Replace host ECDH key (config[2] = `0b)
        # Command option: #0 (config[1:0] = `00b)
        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
        else:
            # shouldn't be able to reach here
            print('ERROR unrecognized signature type {}!'.format(sig_type))
            exit(1)

        new_ecdh_pub_key_name = 'new_ecdh_pub'
        print('reading public key for UID {}; key name {}'.format(device.uid.hex(), new_ecdh_pub_key_name))
        new_ecdh_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdh_pub_key_name)
        if new_ecdh_pub_key is None:
            main_logger.critical('Failed to read new ECDSA public key using Key Management Library: UID: {}; Key Name: {}'.format(device.uid.hex(), new_ecdh_pub_key_name))
            exit(1)

        err, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdh_pub_key[::-1])
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('replace_srk: ' + err.esr_bit_msg + ' - ' + err.error_msg)
            exit(1)

        ecdh_keys_changed = False
        if active_key_updated is not None:
            print("updating keys before (ECDH) replace_srk response validation")
            ecdh_keys_changed = copy_key(device, 'ecdh_prv', 'previous_ecdh_prv', 'new_ecdh_prv')
            if not ecdh_keys_changed:
                print("ERROR changing keys. Continuing to attempt validation, but expecting error!")
            

        ecdh_resp_err = device.verify_replace_srk_response(config, response, sent_nonce, active_key_updated)
        if not ecdh_resp_err.valid_response:
            main_logger.critical('replace_srk response: ' + ecdh_resp_err.esr_bit_msg + ' - ' + ecdh_resp_err.error_msg)
            if ecdh_keys_changed:
                print("changing back to old ECDH private key due to replace_srk response validation failure")
                ecdsa_keys_changed = copy_key(device, 'previous_ecdh_prv', 'tmp_ecdh_prv', 'ecdh_prv')
            exit(3)

        # at this point of new if the ECDSA key has not already been changed,
        # change them now.
        if active_key_updated is not None and active_key_updated and not ecdsa_keys_changed:
            print("changing ECDSA key in key library")
            ecdsa_keys_changed = copy_key(device, 'ecdsa_prv', 'previous_ecdsa_prv', 'new_ecdsa_prv')
            if not ecdsa_keys_changed:
                print("ERROR changing ECDSA key. Comtinuing to attempt validation, but expecting error!")

    # clear keys by doing with device again
    with device:
        # Verify SRK by running ACRTM_Update_Start and ACRTM_Update_Finish
        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'
            
        err, _ = device.meta_acrtm_update_start(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("There was an error running Update_Start")
            exit(1)
        err, _ = device.meta_acrtm_update_finish(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("There was an error running Update_Finish")
            exit(1)

        # Note:
        # If you want to run any command after running replace_srk option #0, make sure to
        # update key_contents_prv.xml file with the new private keys (ecdsa_prv and ecdh_prv)
        # If you want to use Micron host keys, use replace_srk option #1 to revert those
        # keys (see app_revert_factory_srk.py for example)

        print("\nThe customer public key in slot #2 has been added or replaced with a new public key of the same type.")

    exit(0)

def copy_key(device, dest_key_name, intermediate_key_name, source_key_name):
    """
    copy key_name to intermediate_key_name
    copy new_key_name to key_name
    
    delete existing intermediate_key_name if there is one
    """
    ecdsa_keys_changed = False
    if not key_mgmt_lib.delete_key(device.uid.hex(), intermediate_key_name):
        print("Failed to delete key '{}' for UID '{}' ".format(intermediate_key_name, device.uid.hex()))

    if key_mgmt_lib.copy_key(device.uid.hex(), dest_key_name, intermediate_key_name):
        print("copied key '{}' to {} for UID '{}' ".format(dest_key_name, intermediate_key_name, device.uid.hex()))
        if key_mgmt_lib.delete_key(device.uid.hex(), dest_key_name):
            print("deleted key '{}' for UID '{}' ".format(dest_key_name, device.uid.hex()))
            if key_mgmt_lib.copy_key(device.uid.hex(), source_key_name, dest_key_name):
                print("copied key '{}' to '{}' for UID '{}' ".format(source_key_name, dest_key_name, device.uid.hex()))
                ecdsa_keys_changed = True
            else:
                print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' ".format(source_key_name, dest_key_name, device.uid.hex()))
        else:
            print("ERROR! Failed to delete key '{}' for UID '{}' ".format(dest_key_name, device.uid.hex()))
    else:
        print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' ".format(dest_key_name, intermediate_key_name, device.uid.hex()))

    return ecdsa_keys_changed

if __name__ == "__main__":
    main()
