#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging
from sys import exit
from collections import namedtuple

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

from authenta.indigo import utils
import app_common_logging as app_logging

import authenta.authenta_key_mgmt_interface as key_mgmt_lib

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for reverting the current SRK in a device to Micron factory SRK.
'''

EPILOG = '''

EXAMPLE CMDLNS:
    # commandline using default mapping files
    $> app_revert_factory_srk.py <device alias> <sig_type>
     -- as in --
    $> app_revert_factory_srk.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'ECDSA is the signature type

    # for test cases where the default devices.json file isn't what is needed
    $> app_replace_srk.py -f mydevices_mapping.json snor1 1141 1143

EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure
    9 : unprovisioned device or Micron key already active

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument('-fact_ecdsa', dest='fact_keyname_ecdsa', metavar="<Micron-factory-ECDSA-keyname>",
               default="factory_ecdsa_prv",
               help="the name associated with Micron factory key for this device in the key repository.")

p.add_argument('-fact_hmac', dest='fact_keyname_hmac', metavar="<Micron-factory-HMAC-keyname>",
               default="factory_ecdh_prv",
               help="the name associated with Micron factory key for this device in the key repository.")

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
              default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type
        fact_keyname_ecdsa = ARGS.fact_keyname_ecdsa
        fact_keyname_hmac = ARGS.fact_keyname_hmac

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        # Check if device is remotely provisioned, SRK can only be replaced while remotely provisioned
        if not device.provision_sts_struct.Provisioned and not device.provision_sts_struct.Remotely_Provisioned:
            print("provision_sts.Provisioned bit unset or provision_sts.Remotely_Provisioned bit unset.")
            print("Device must be remotely provisioned before attempting this app.")
            print("... exiting on error.")
            main_logger.critical("Device is not remotely provisioned.")
            exit(9)

        # Check that the SRK has previously been replaced (using replace_srk Option #0)
        if not device.provision_sts_struct.SRK_Replaced:
            print("The Micron server root key is already active.")
            print("Use replace_srk option #0 to change to active host server public keys.")
            print("... exiting on error.")
            main_logger.critical("The Micron server root key is already active.")
            exit(9)

        # Config
        # SigType: from command line
        # Command option: #1 (Config[1:0] = `01b)
        # Revert ECDSA: config[2] = '1b
        if sig_type == 'HMAC':
            config = b"\x00\x05"
        elif sig_type == 'ECDSA':
            config = b'\x80\x05'
        else:
            # shouldn't be able to reach here
            print('ERROR unrecognized signature type {}!'.format(sig_type))
            exit(1)

        if not key_mgmt_lib.contains_key(device.uid.hex(), fact_keyname_ecdsa):
            print('\n\nERROR no key with name \'{}\' found in the key repository.'.format(fact_keyname_ecdsa))
            print('To avoid locking the chip without the key, disallowing revert with output key\n\n')
            exit(1)

        if not key_mgmt_lib.contains_key(device.uid.hex(), fact_keyname_hmac):
            print('\n\nERROR no key with name \'{}\' found in the key repository.'.format(fact_keyname_hmac))
            print('To avoid locking the chip without the key disallowing revert with output key\n\n')
            exit(1)

        err, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b'\x00' * 64)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('replace_srk: {}  {}'.format(err.esr_and_response_success, err.error_msg))
            exit(1)

        print('\n\n')
        print("updating keys before replace_srk response validation")
        if not copy_key(device, fact_keyname_ecdsa, 'previous_ecdsa_prv', 'ecdsa_prv'):
            print("ERROR copying factory ECDSA private key to ecdsa_prv!")

        if not copy_key(device, fact_keyname_hmac, 'previous_ecdh_prv', 'ecdh_prv'):
            print("ERROR copying factory HMAC (ECDH) private key to ecdh_prv!")
        print('\n\n')

        ecdsa_resp_err = device.verify_replace_srk_response(config, response, sent_nonce, active_key_updated)
        if not ecdsa_resp_err.valid_response:
            main_logger.critical('replace_srk response: ' + ecdsa_resp_err.esr_bit_msg + ' - ' + ecdsa_resp_err.error_msg)
            exit(2)

        print("\nThe Host Server keys, (both ECDSA and ECDH/HMAC), have been reverted back to the Micron factory keys.")

    exit(0)


def copy_key(device, src_key_name, intermediate_key_name, dest_key_name):
    """
    copy dest_key_name to intermediate_key_name
    copy src_key_name to key_name
    
    delete existing intermediate_key_name if there is one
    """
    ecdsa_keys_changed = False
    if key_mgmt_lib.contains_key(device.uid.hex(), intermediate_key_name):
        if not key_mgmt_lib.delete_key(device.uid.hex(), intermediate_key_name):
            print("Failed to delete key '{}' for UID '{}' ".format(intermediate_key_name, device.uid.hex()))

    if key_mgmt_lib.contains_key(device.uid.hex(), dest_key_name):
        # a key with he dest_key_name exists, so back it up...
        if key_mgmt_lib.copy_key(device.uid.hex(), dest_key_name, intermediate_key_name):
            print("copied key '{}' to {} for UID '{}' ".format(dest_key_name, intermediate_key_name, device.uid.hex()))
        else:
            print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' (as a backup) ".format(dest_key_name, intermediate_key_name, device.uid.hex()))
            return False

        # ... and delete it
        if key_mgmt_lib.delete_key(device.uid.hex(), dest_key_name):
            print("deleted key '{}' for UID '{}' ".format(dest_key_name, device.uid.hex()))
        else:
            print("ERROR! Failed to delete key '{}' for UID '{}' ".format(dest_key_name, device.uid.hex()))

    # now copy the souece to the destination
    if key_mgmt_lib.copy_key(device.uid.hex(), src_key_name, dest_key_name):
        print("copied key '{}' to '{}' for UID '{}' ".format(src_key_name, dest_key_name, device.uid.hex()))
        ecdsa_keys_changed = True
    else:
        print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' ".format(src_key_name, dest_key_name, device.uid.hex()))

    return ecdsa_keys_changed

if __name__ == "__main__":
    main()
