#! usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers
import time

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = ''' DESCRIPTION:
\t A utility for 
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_secure_flash_boot.py <device alias> <sig_type>
     -- as in --
    $> app_secure_flash_boot.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and 'ECDSA' is the signature type

EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure
    9 : deprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Kacy Ebel

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    # The locations of MLO, u-boot, and kernel (address and size)
    # NOTE that measurement range sizes are in 4K-byte units
    MEAS_RANGE_SIZE_MLO = {
        "addr": 0x00000020,
        "size": 0x00000020
    }
    MEAS_RANGE_SIZE_UBOOT = {
        "addr": 0x00000060,
        "size": 0x000000C0
    }
    MEAS_RANGE_SIZE_KERNEL_EXT0 = {
        "addr": 0x00008C00,
        "size": 0x00000800
    }
    MEAS_RANGE_SIZE_KERNEL_EXT1 = {
        "addr": 0x00008804,
        "size": 0x000001F0
    }

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # call Request_UID to get device_config
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: ' + err.error_msg)
            exit(2)

        if sig_type == 'HMAC':
            config = b'\x00\x00'
        elif sig_type == 'ECDSA':
            config = b'\x80\x00'

        # Set measurement range
        # range_id, crtm_addr, crtm_size
        range_list = [(0, MEAS_RANGE_SIZE_MLO['addr'], MEAS_RANGE_SIZE_MLO['size']),
                      (1, MEAS_RANGE_SIZE_UBOOT['addr'], MEAS_RANGE_SIZE_UBOOT['size']),
                      (2, MEAS_RANGE_SIZE_KERNEL_EXT0['addr'], MEAS_RANGE_SIZE_KERNEL_EXT0['size']),
                      (3, MEAS_RANGE_SIZE_KERNEL_EXT1['addr'], MEAS_RANGE_SIZE_KERNEL_EXT1['size']),
                      (4, 0x0, 0x0)]

        if sig_type == "HMAC":
            config = b'\x00\x00'
        elif sig_type == "ECDSA":
            config = b'\x80\x00'

        ret = utils.set_measurement_range(device, config, range_list)
        if not ret:
            assert(False)

        print("\nset auto-measure ranges.")

        # read the current device config
        err, response = device.meta_crtm_register_read(config, CRTM_Registers.Device_Config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_read: ' + err.error_msg)
            exit(1)

        # call Update_Start Option #0 to indicate an update to the CRTM Registers
        err, _ = device.meta_acrtm_update_start(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_start: ' + err.error_msg)
            exit(1)
        
        # Set the Auto_Measure_Activate bit
        device_config = response[318:322]
        device_config_int = int.from_bytes(device_config, byteorder="big")
        data_int = device_config_int | 0x00000001
        data = data_int.to_bytes(4, byteorder="big")
        data = b'\x00'*92+data+b'\x00'*32
        mask = b'\x01'*92+b'\x00'*4+b'\x01'*32

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.Device_Config, data, mask)
        if not err.esr_success:
            device.meta_acrtm_update_finish(config)
            print("\n\n")
            main_logger.critical('register_write: ' + err.error_msg)
            exit(1)

        print("set auto-measure bit.")
        # NOTE MaxAutoMasureSize is designated in 4K-byte units (just as are measurment ranges sizes)


        # set the max_automeasure_size to 0x8E0 (x4K=0x8E0000) to a partial measurement
        # that covers the first two ranges: MLO_AND_UBOOT and KERNEL_EXT0
        data = b'\x00\x00\x08\xE0'
        mask = b'\x00'*4
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Max_AutoMeasure_Size, data, mask)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical(err.cmd_name + ': ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(config)
            exit(1)

        print("set CRTM_Max_AutoMeasure_Size to {} (NOTE this is in 4K-byte units).".format('0x8E0'))

        # call Update_Finish Option #0 to complete the writes to the CRTM registers
        err, _ = device.meta_acrtm_update_finish(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_start: ' + err.error_msg)
            exit(1)

        
        # Lock the data with APA_Modify
        # Config[15] = 0/1, SigType; Config[14] = 0, KeyRef = Host Server Key Only
        # Config[2] = 1, WriteAPA = write
        if sig_type == "HMAC":
            config = b'\x00\x04'
        elif sig_type == "ECDSA":
            config = b'\x80\x04'

        # Address to be locked
        # NOTE that typically the boot image block would be locked, 
        # but here we are locking a different block to present an 
        # example while ensuring we don't break the file system
        wpg_addr = 0xfd0000

        # AuthentaAccessRights[15] = 1, APAWriteProtect = Protected
        # AuthentaAccessRights[14] = 1, APAReadProtect = Protected
        # AuthentaAccessRights[9] = 0, EnableKTSKey = Host Server Key only
        # AuthentaAccessRights[2:0] = 000, PublicKeySlotPointer
        authenta_access_rights = b'\xC0\x00'

        apa_data = b'\x00'*72 + authenta_access_rights[::-1] + b'\x00'*6

        # key slot will be ignored but must be passed as an argument
        key_slot = 0x0

        err, _ = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: ' + err.error_msg)
            exit(1)

        print("locked flash memory block at address {:X} (using APA_Modify)".format(wpg_addr))
        print("    Note that typically the boot image block would be locked, ")
        print("    but here we are locking a different block to present an example")
        print("    while ensuring we don't break the file system")

        # Run Flash_Extend Option #1 to save Flash_DigestNV and Flash_DigestNV_Partial values
        # Config[15] = 0/1, SigType
        # Config[2]:Config[0] = 01, command option #1
        # Config[1] = 0, Start digest calculation at the beginning of the linked list
        if sig_type == "HMAC":
            config = b'\x00\x01'
        elif sig_type == "ECDSA":
            config = b'\x80\x01' 

        err, _ = device.meta_flash_extend(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('flash_extend option #1: ' + err.error_msg)
            exit(1)

        print('programmed NV_Flash_Digest and NV_Flash_Digest_Partial, (the auto-measure ')
        print('comparison measurements), to the values of the current flash contents')

    exit(0)

if __name__ == "__main__":
    main()
