#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers

import app_common_logging as app_logging

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for generating DICE key based on measurement in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_secure_update.py <device alias> <sig_type>
     -- as in --
    $> app_secure_update.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and ECDSA is the signature type


EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Tatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: {}'.format(err.error_msg))
            exit(2)

        print("\nDevice UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # Check if device is not provisioned. If not provisioned, exit the app.
        if not device.provision_sts_struct.Provisioned:
            print("provision_sts.Provisioned bit unset.")
            print("Device shall be provisioned prior to attempting this app.")
            print("... exiting on error.")
            main_logger.critical("Device is not provisioned.")
            exit(9)

        data_size = 0x1000
        test_data = b'\x11' * 512

        # wpg_addr should be boot code address
        # For demonstration purpose we will use another address that won't interfere with boot code
        wpg_addr = 0xf00000

        # Unlock address
        # APAWriteProtect = '0b
        # APAReadProtect = '0b
        apa_data = b'\x00'*80
        
        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign command with ECDSA
            config = b'\x80\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x04'

        err, response = device.meta_apa_modify(config, wpg_addr, apa_data, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to make 0x{:X} unprotected'.format(wpg_addr))
        
        # Erase previous boot code
        response = device.meta_write_multiple_block(wpg_addr, b"\x00"*512)
        if not response:
            print("\n\n")
            main_logger.critical('write_multiple_block: failed to write data')
            exit(1)

        print('Erased data in 0x{:X}'.format(wpg_addr))

        # Write data
        response = device.meta_write_multiple_block(wpg_addr, test_data)
        if not response:
            print("\n\n")
            main_logger.critical('write_multiple_block: failed to write data')
            exit(1)

        print('Written data to 0x{:X}'.format(wpg_addr))

        # Lock new data address
        # APAWriteProtect = '1b
        # APAReadProtect = '1b
        authentaAccessRightsData = b'\xC0\x00'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6
        err, response = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to make 0x{:X} read and write protected'.format(wpg_addr))

        # Update measurement ranges to not measure erased data and add measure new data
        # range_id, crtm_addr, crtm_size
        range_list = [(0, wpg_addr, data_size)]

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign command with ECDSA
            config = b'\x80\x00'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x00'

        ret = utils.set_measurement_range(device, config, range_list)
        if not ret:
            print("\n\n")
            main_logger.critical('set_measurement_range: failed')
            exit(1)

        print('Set up measurement range')

        # Set Auto_Measure_Activate to true
        ret = utils.enable_automeasure(device)
        if not ret:
            print("\n\n")
            main_logger.critical('enable_automeasure: failed')
            exit(1)

        print('Enabled automeasure')

        # Set CRTM_Max_AutoMeasure_Size to 0x1000
        ret = utils.set_CRTM_Max_AutoMeasure_Size(device, config, 0x1000)
        if not ret:
            print("\n\n")
            main_logger.critical('set_CRTM_Max_AutoMeasure_Size: failed')
            exit(1)

        print('Set CRTM_Max_AutoMeasure_Size')

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign command with ECDSA
            config = b'\x80\x01'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x01'

        # Update Flash_DigestNV and Flash_DigestNV_Partial
        err, response = device.meta_flash_extend(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('flash_extend: {}'.format(err.error_msg))
            exit(1)

        print('Programmed NV_Flash_Digest and NV_Flash_Digest_Partial, (the auto-measure ')
        print('comparison measurements), to the values of the current flash contents')

    exit(0)

if __name__ == "__main__":
    main()
