#! /usr/bin/env python3
import argparse as ap
import sys
import os
import json
import logging

# --- local imports ---
from authenta.indigo import utils
import app_common_logging as app_logging

from authenta.maps import Envr
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.indigo.utils import CRTM_Registers
from authenta import authenta_key_mgmt_interface

VERSION = 1.0
PROGRAM_DES = '''DESCRIPTION:
\t A utility for setting up 3 protected areas (Write Protection Groups) in a device with Micron Authenta Indigo
'''

EPILOG = '''

EXAMPLE CMDLNS:

    # commandline using default mapping files
    $> app_write_protection_groups.py <device alias> <sig_type>
     -- as in --
    $> app_write_protection_groups.py snor1 ECDSA
     - where 'snor1' is a device entry in the devices.json file
     - and ECDSA is the signature type

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR: Tatiane Troxell

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument('sig_type', metavar="<signature type>", choices=['ECDSA', 'HMAC'],
               help="Choose signature type: ECDSA or HMAC")

# command-line optional arguments
p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) enpoint OS device alias name.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script verion to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    device = IndigoDevice(ep)

    with device:
        sig_type = ARGS.sig_type

        # Call Request_UID to get Provision Status
        err, _ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: {}'.format(err.error_msg))
            exit(2)

        print("\nDevice UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # Check if device is not provisioned. If not provisioned, exit the app.
        if not device.provision_sts_struct.Provisioned:
            print("provision_sts.Provisioned bit unset.")
            print("Device shall be provisioned prior to attempting this app.")
            print("... exiting on error.")
            main_logger.critical("Device is not provisioned.")
            exit(9)

        # Unlock 3 addresses
        # AuthentaAccessRights[15]: APAWriteProtect = '0b
        # AuthentaAccessRights[14]: APAReadProtect = '0b
        apa_data = b'\x00'*80

        wpg_addr_1 = 0xf00000
        wpg_addr_2 = 0xf10000
        wpg_addr_3 = 0xf20000

        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x04'

        err, _= device.meta_apa_modify(config, wpg_addr_1, apa_data, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to make 0x{:X} not protected'.format(wpg_addr_1))

        err, _= device.meta_apa_modify(config, wpg_addr_2, apa_data, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to make 0x{:X} not protected'.format(wpg_addr_2))

        err, _= device.meta_apa_modify(config, wpg_addr_3, apa_data, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to make 0x{:X} not protected'.format(wpg_addr_3))

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x48'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x48'

        # Upload one ECDSA_KTS_PuK keys to slots 4
        # (TRD for 0.4) specifies in "3.4.11 Manage_KTS" that "The very first time a keyslot is 
        # populated with a key, the ECDSA_KTS_PuK must be uploaded first".
        byte_addr = 0x4
        key_name = "ecdsa_pub_slot_{}".format(byte_addr)
        ECDSA_KTS_PuK_key_slot_4 = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDSA_KTS_PuK_key_slot_4 is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, 0x4, ECDSA_KTS_PuK_key_slot_4[::-1])
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded a ECDSA public key to key table slot 4')

        # Upload one ECDSA_KTS_PuK keys to slots 5
        byte_addr = 0x5
        key_name = "ecdsa_pub_slot_{}".format(byte_addr)
        ECDSA_KTS_PuK_key_slot_5 = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDSA_KTS_PuK_key_slot_5 is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, 0x5, ECDSA_KTS_PuK_key_slot_5[::-1])
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded a ECDSA public key to key table slot 5')

        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\xA0'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\xA0'

        # Create a slave key in slot 4
        err, _ = device.meta_manage_kts(config, 0x4, b"\x00"*64)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Created a slave key in key table slot 4')

        # Create a slave key in slot 5
        err, _ = device.meta_manage_kts(config, 0x5, b"\x00"*64)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Created a slave key in key table slot 5')

        # Upload ECDH public key for hmac sigtype option
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x44'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x44'

        byte_addr = 0x4
        key_name = "ecdh_pub_slot_{}".format(byte_addr)
        ECDH_KTS_PuK_key_slot_4 = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDH_KTS_PuK_key_slot_4 is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, 0x4, ECDH_KTS_PuK_key_slot_4[::-1])
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded a ECDH public key to key table slot 4')
            
        byte_addr = 0x5
        key_name = "ecdh_pub_slot_{}".format(byte_addr)
        ECDH_KTS_PuK_key_slot_5 = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        if ECDH_KTS_PuK_key_slot_5 is None:
            #print("\nERROR No key for UID: {} and key_name {} found by the key library! Aborting.\n".format(device.uid.hex(), key_name))
            main_logger.critical("No key for UID {} and key_name '{}' found by the key library!".format(device.uid.hex(), key_name))
            exit(1)

        err, _ = device.meta_manage_kts(config, 0x5, ECDH_KTS_PuK_key_slot_5[::-1])
        if not err.esr_and_response_success:
            print("\n\n")
            main_logger.critical('manage_kts: {}'.format(err.error_msg))
            exit(1)

        print('Uploaded a ECDH public key to key table slot 5')

        # Enable modify APA data by slot key
        # AuthentaAccessRights[9]: EnableKTSKey = '1b
        # AuthentaAccessRights[2:0]: PublicKeySlotPointer = 0x1 (KTS slot 4)
        authentaAccessRightsData = b'\x02\x01'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6   

        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x04'

        # APA_Modify using server key to enable modify APA data by key slot 4
        err, _ = device.meta_apa_modify(config, wpg_addr_1, apa_data[::-1], 0x4)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to enable modifying APA data by key slot 4 for 0x{:X}'.format(wpg_addr_1))

        # Lock first address with slot key 4
        # AuthentaAccessRights[15]: APAWriteProtect = '1b
        # AuthentaAccessRights[14]: APAReadProtect = '1b
        # AuthentaAccessRights[9]: EnableKTSKey = '1b
        # AuthentaAccessRights[2:0]: PublicKeySlotPointer = 0x1 (KTS slot 4)
        authentaAccessRightsData = b'\xC2\x01'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\xC0\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x40\x04'

        # APA_Modify using kts slot key for signing
        err, _ = device.meta_apa_modify(config, wpg_addr_1, apa_data[::-1], 0x4)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to protect 0x{:X} using key slot 4 for sigining'.format(wpg_addr_1))

        # Enable modify APA data by slot key
        # AuthentaAccessRights[9]: EnableKTSKey = '1b
        # AuthentaAccessRights[2:0]: PublicKeySlotPointer = 0x2 (KTS slot 5)
        authentaAccessRightsData = b'\x02\x02'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6  

        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x04'

        # APA_Modify using server key to enable modify APA data by key slot 5
        err, _ = device.meta_apa_modify(config, wpg_addr_2, apa_data[::-1], 0x5)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to enable modifying APA data by key slot 5 for 0x{:X}'.format(wpg_addr_2))

        # Lock second address with slot key 5
        # AuthentaAccessRights[15]: APAWriteProtect = '1b
        # AuthentaAccessRights[14]: APAReadProtect = '1b
        # AuthentaAccessRights[9]: EnableKTSKey = '1b
        # AuthentaAccessRights[2:0]: PublicKeySlotPointer = 0x2 (KTS slot 5)
        authentaAccessRightsData = b'\xC2\x02'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6

        # Get config value based on command-line arguments
        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\xC0\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x40\x04'

        # APA_Modify using kts slot key for signing
        err, _ = device.meta_apa_modify(config, wpg_addr_2, apa_data[::-1], 0x5)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to protect 0x{:X} using key slot 5 for sigining'.format(wpg_addr_2))

        # Lock third address with server key
        # AuthentaAccessRights[15]: APAWriteProtect = '1b
        # AuthentaAccessRights[14]: APAReadProtect = '1b
        authentaAccessRightsData = b'\xC0\x00'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6

        if sig_type == 'ECDSA':
            # Sign with ECDSA
            config = b'\x80\x04'
        elif sig_type == 'HMAC':
            # Sign command with HMAC
            config = b'\x00\x04'

        # APA_Modify using kts slot key for signing
        err, _ = device.meta_apa_modify(config, wpg_addr_3, apa_data[::-1], 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('apa_modify: {}'.format(err.error_msg))
            exit(1)

        print('Issued APA_Modify to protect 0x{:X} using host server key for sigining'.format(wpg_addr_3))
       
    exit(0)

if __name__ == "__main__":
    main()
