#! /usr/bin/env python3

import argparse as ap
import json
import logging
import sys
import os
from sys import exit
from collections import namedtuple
from time import sleep

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.cmd_error_code import NaxosCmdErrorCode
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import CRTM_Registers, warn_on_response_verification_fail, change_word_order

import app_common_logging as app_logging

HARDCODED_GOLDEN_MEASUREMENT_1_5 = bytes.fromhex("b2de73cca0536c603c9848999e9ad96937ba281a038816cdf3fafe7fa6ef9eac")
HARDCODED_FDS_1_5 = bytes.fromhex("85100243cc8f142732ac9cc959a7fb219db2c3f5f479faec8eae966f041dabe7")

TEST_BLOCK_ADDR_1 = 0xFF3000
TEST_BLOCK_ADDR_2 = 0xFF5000
TEST_BLOCK_SIZE = 0x1000

VERSION=0.01

PROGRAM_DES = '''DESCRIPTION: 
\t A utility for demonstrating various security functions of a Micron Authenta
\t enabled SPI-NOR device.
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_demo.py  <endpoint alias> 
    -- as in --
   $> app_demo.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_demo.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_demo.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general error at request_uid
  2 : general error

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f",  dest='ifn',  metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device',  metavar="<device alias>",  help="endpoint-device alias found in devices.json file.")

p.add_argument("-t", dest='hextable', action='store_true', help="Hex table style output.")

p.add_argument("-v", "--verbosity",  action='store_true',
                help="REQUIRED: to print command's pass/fail message and exit status.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
                help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

# todo - remove "request_uid_status"

def main():
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        platform = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        platform = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #

    device = NaxosDevice(platform)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)

    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device is revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(platform)
        HARDCODED_GOLDEN_MEASUREMENT = HARDCODED_GOLDEN_MEASUREMENT_1_5
        HARDCODED_FDS = HARDCODED_FDS_1_5           
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(2)

    with device:
        err = None

        if device.provision_sts_struct is None:
            # perform request UID to retrieve AMTC0 and provison status
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print('Device UID:                       0x{}'.format(device.uid_as_str))
        print('Naxos Revision:                   {}'.format(hex(dev_rev)))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))
        print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))

        err, _ = device.read_extended_crtm_status()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed attempting to retrieve extended CRTM status from the device: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

#* Checks if the device is provisioned, and provision the device if not already provisioned
        if not device.provision_sts_struct.Provisioned:
            print("Naxos not provisioned, provision it")
            print("Attempting to provision device")
            main_logger.info("Device is not provisioned. Provisioning device.")
            config = b'\x05' # DICE enabled, remote
            err, *_ = device.meta_provision_crtm(config)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed attempting to provision device: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)

        # elif(provision_sts_struct.RIOT_Mode is not '1'):
        elif not device.provision_sts_struct.RIOT_Mode:
            print("Naxos provisioned, but not in RIOT mode, deprovision it")
            main_logger.info("Naxos provisioned, but not in RIOT mode, deprovisioning device.")
            err, _ = device.read_extended_crtm_status()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed attempting to retrieve extended CRTM status from the device (before deprovisioning): '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)

            config = b'\x02'
            err, *_ = device.meta_deprovision_crtm(config)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed attempting deprovision the device: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)

            print("\n\n")
            print("Authenta device deprovisioned.")
            print("Attempting to provision device.")
            main_logger.info("Device deprovisioned. Provisioning device.")
            config = b'\x05' # DICE enabled, remote
            err, *_ = device.meta_provision_crtm(config)

            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed attempting provision the device with DICE enabled: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)

            print("Authenta device provisioned with DICE enabled.")
            main_logger.info("Authenta device provisioned with DICE enabled.")

# set the auto-meaure registers

        # make sure boot_amtc0 is enabled
        if device.boot_amtc0 is None:
            err, _ = device.read_extended_crtm_status()

            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed attempting to retrieve extended CRTM status from the device (before calling AuthLock): '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)


        config = b'\x00'
        err, _ = device.meta_acrtm_update_start(config)

        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed calling ACRTM_Update_Start: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)


#smRC = SMDM_Dev_SetEEPROM_MemBlockAttributeSetting(devHndl, eepromMemRegBlkID_0, eepromMemRegBlkAttr_ModifyAndRead);
        err, response = device.meta_crtm_register_read(config, 0xD0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed reading EEPROM Attributes (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        eeprom_attrib_bytes = response[16:20]
        main_logger.info("EEPROM Attributes:{}".format(eeprom_attrib_bytes.hex()))

        eeprom_attrib_int = int(eeprom_attrib_bytes.hex(), 16)
        eeprom_attrib_int = eeprom_attrib_int | 0x03

        err, _ = device.meta_crtm_register_write(config, 0xD0, eeprom_attrib_int.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting EEPROM Attributes (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # Set MaxAutoMEasureSize to 0xB1000 (FSBL + DTB + UBoot + Test block 1)
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Max_AutoMeasure_Size, 0xB1000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting MaxAutoMeasureSize (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # FSBL: start = 0x0; size = 0x10000
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_0, 0x00.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 0 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_0, 0x10000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 0 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # DTB: start = 0x10000; size = 0x10000
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_1, 0x10000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 1 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_1, 0x10000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 1 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # UBoot: start = 0x20000; size = 0x90000
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_2, 0x20000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 2 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_2, 0x90000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 2 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # Test block 1
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_3, TEST_BLOCK_ADDR_1.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 3 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_3, TEST_BLOCK_SIZE.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 3 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # Kernel: start = 0xB0000; size = 0x400000
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_4, 0xB0000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 4 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_4, 0x400000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 4 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # Rootfs: start = 0xA00000; size = 0x300000
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_5, 0xA00000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 5 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_5, 0x300000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 5 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # Test block 2
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_6, TEST_BLOCK_ADDR_2.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 6 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_6, TEST_BLOCK_SIZE.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 6 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        # End of linked list
        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Start_Addr_7, 0x0000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Start Address 7 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, CRTM_Registers.CRTM_Range_Size_7, 0x0000.to_bytes(4, byteorder='big'), True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed setting CRTM_Range Size 7 (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

# Set auto-measure to True
#smRC = SMDM_Dev_SetDeviceConfigSetting(devHndl, deviceCnfgParam_AutoMeasure, true);
        ret = set_automeasure_bit(device, True)
        if not ret:
            print("\n\n")
            main_logger.critical('Failed attempting to enable Auto_Measure in command.')
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)

#8. Write the predetermined Golden Measurement to EEPROM
#HARDCODED_GOLDEN_MEASUREMENT
        byte_addr_arr   = [0,4,8,12,16,20,24,28]
        eeprom_addr_arr = [0x200,0x204,0x208,0x20c,0x210,0x214,0x218,0x21c]

        for index in range(0,8):
            main_logger.debug("eeprom write loop index {}".format(index))
            byte_addr  = byte_addr_arr[index]
            write_addr = eeprom_addr_arr[index]
            eeprom_write_data = int(HARDCODED_GOLDEN_MEASUREMENT[byte_addr:(byte_addr+4)].hex(),16)
            err, _ = device.meta_crtm_register_write(config, write_addr, eeprom_write_data.to_bytes(4, byteorder='big'), True)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Failed writing to EEPROM address { }: '.format(write_addr)+ err.esr_bit_msg + '; ' + err.error_msg)
                err, _ = device.meta_acrtm_update_finish(b'\x00')
                exit(2)
            warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(b'\x00')
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed calling ACRTM_Update_Finish after setting Auto-Measure ranges: '+ err.esr_bit_msg + '; ' + err.error_msg)
            err, _ = device.meta_acrtm_update_finish(b'\x00')
            exit(2)
        warn_on_response_verification_fail(err)

#9. Use Auth-Lock command to unlock the device

        err, _ = device.meta_auth_lock()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed cryptographically unlocking the device: (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

#10. run the following steps:

# unlock (volatile) block-lock at address TEST_BLOCK_ADDR_1, and TEST_BLOCK_ADDR_2
# erase data in those address;
# write the text "hello" at address TEST_BLOCK_ADDR_1;
# write the text "starting_demo" at TEST_BLOCK_ADDR_2;
# re-lock.

        err, _ = device.read_extended_crtm_status()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed reading Extended CRTM Status: (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

        print('Auth_Unlock before first write to 0x{:X} and 0x{:X}: {}'.format(TEST_BLOCK_ADDR_1, TEST_BLOCK_ADDR_2, device.extended_crtm_status_struct.Auth_Unlock))

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        sleep(1)
        
        ret = device.meta_page_program(TEST_BLOCK_ADDR_1, bytes("hello", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        sleep(1)

        ret = device.meta_page_program(TEST_BLOCK_ADDR_2, bytes("starting_demo", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

# OD: the sectors must actually be locked!
        ret = device.meta_ppb_erase()
        if not ret:
            main_logger.warning('Failed attempting to unlock flash memory.')
        sleep(1)
        ret = device.meta_ppb_program(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed attempting to lock address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        sleep(1)
        ret = device.meta_ppb_program(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed attempting to lock address 0x{:X}.'.format(TEST_BLOCK_ADDR_2))

        ret, ppb_read_data = device.meta_ppb_read(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed attempting to read lock bit at address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        else: 
            print("ppb at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_1, ppb_read_data.hex()))

        ret, ppb_read_data = device.meta_ppb_read(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed attempting to read lock bit address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        else:
            print("ppb at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_2, ppb_read_data.hex()))

        ret, ppb_read_data = device.meta_ppb_lock_bit_read()
        if not ret:
            main_logger.warning('Failed attempting to read PPB lock bit.')
        else:
            print("PPB lock bit: " + ppb_read_data.hex())

# PPB_Lock_Bit_Write(fd)

        ret = device.meta_ppb_lock_bit_write()
        if not ret:
            main_logger.warning('Failed attempting to write PPB lock bit.')
        else:
            print("PPB lock bit set.")

        ret, ppb_read_data = device.meta_ppb_lock_bit_read()
        if not ret:
            main_logger.warning('Failed attempting to read PPB lock bit.')
        else:
            print("PPB lock bit: " + ppb_read_data.hex())

        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X} after initial write: {}".format(TEST_BLOCK_ADDR_1, read_data[:32]))

        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_2))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X} after initial write: {}".format(TEST_BLOCK_ADDR_2, read_data[:32]))

#*       Note: This is the data written at those addresses that was used to create the Golden Measurement.
#              Later in the demo we will change the data there to demonstrate a change in the measured area.
#              The rest of the  measured area is the actual boot code which we don't want to change as part of a demo.

#11. Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
        config = b'\x01' # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list

        err, _ = device.meta_flash_extend(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed attempting to perform flash_extend option #1 command: (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

#12. Detach from the Endpoint and device, Instruct the user to reboot the Endpoint

 #context stuff

#<reboot>
    input("Please power cycle your endpoint, **soft reboot will not work** press [enter] once it is fully booted")

    device.test_only_set_boot_amtc0(None)
    with device:
        # perform request UID to retrieve AMTC0
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        print("\n\n")
        print("provision_sts_struct after 1 reboot")
        print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))

#*   15. Display Authenta device status (showing crtmMismatch false; and correct PCR0)
        err, _ = device.read_extended_crtm_status()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed attempting to retrieve extended CRTM status from the device: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

        print("\n\n")
        print('last_ACMD:',           device.extended_crtm_status_struct.LACMD)
        print('CRTM_Mismatch:',       device.extended_crtm_status_struct.CRTM_Mismatch)
        print('CRTM_Range_Invalid:',  device.extended_crtm_status_struct.CRTM_Range_Invalid)
        print('CRTM_PCR0_Invalid:',   device.extended_crtm_status_struct.CRTM_PCR0_Invalid)
        print('Auth_Unlock:',         device.extended_crtm_status_struct.Auth_Unlock)
        print('Recovery_Block:',      device.extended_crtm_status_struct.Recovery_Block)
        print('Partial_Measurement:', device.extended_crtm_status_struct.Partial_Measurement)
        print('Recovery_Mismatch:',   device.extended_crtm_status_struct.Recovery_Mismatch)

#*   16. Display actual and expected DICE Secret (FDS) (they should match)

# Call the Measure command with Option #1 to have the device compute the FDS and put it into the KTR

#if (MSECMEM_RC_OK != (smRC = Cmd_Measure(
#    msrCmdOpt_GenerateFDS, : config bits [1:0] = 01b
#    msrCmdAltSrc_None,     : config bits [9:8] = 00b
#    secMemEEPROM_MemRegBlockID_t eepromAltSrcAddr = 0,
#    msrCmdKeySrc_NoKey_Dgst, : config bits [5:3] = 000b
#    secMemEEPROM_MemRegBlockID_t = 0,
#    includeMetadata, = false,
#    inputDigest, = NULL,
#    lenInputDigest, = 0,
#    msrCmdDst_KTR,            : config bit 7 = 0b
#    eepromMsrDstAddr, 0,
#    bool verifyResultSig = true)))
#{
#    return smRC;
#}
#bit 2: reserved
#bit 6 = 0
#bit 10 = 0

        config =  b"\x00\x01"
        print("Call Measure to obtain FDS data")
        err, _ = device.meta_measure(config, 0x0, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
#            exit(1)
        warn_on_response_verification_fail(err)
        print("Measure completed")

        err, KTR_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.KTR_Base)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading least-significant-bits KTR register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        LSB_KTR = change_word_order(KTR_read_resp[16:32])
        
        err, KTR_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.KTR_Base + 16)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading most-significant-bits KTR register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        MSB_KTR = change_word_order(KTR_read_resp[16:32])

        print("\n\n")
        print("Actual FDS as extracted from KTR: 0x{}{}".format(MSB_KTR.hex(),LSB_KTR.hex()))
        print("Expected FDS: 0x{}".format(HARDCODED_FDS.hex()))
        #TODO calculate FDS in key driver

#*   17. Disable modification of DICE Secret (FDS)
#return Cmd_Measure(context,
#        msrCmdOpt_DisableFsAccess, config bits [1:0] = '10'
#        msrCmdAltSrc_None,:        config bits [9:8] = 00b
#        0,
#        msrCmdKeySrc_NoKey_Dgst,: config bits [5:3] = 000b
#        0,
#        false,
#        NULL,
#        0,
#        msrCmdDst_KTR, :              config bit 7 = 0b
#        0,
#        verifyResultSig);
#bit 2: reserved
#bit 6 = 0
#bit 10 = 0
        config =  b'\x00\x02'
        print("Call Measure to disable access to Fs(s)")
        err, _ = device.meta_measure(config, 0x0, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
        main_logger.info("Measure completed")

#*   18. Complete measurement

        print('Partial_Measurement:', device.extended_crtm_status_struct.Partial_Measurement)
        if(device.extended_crtm_status_struct.Partial_Measurement):
            print("The device has performed a partial measurement, now completing the measurement.")
            config = b'\x02' # bits 2 and 0 are '00' for option 0, bit 1 is '1' for start after max MaxAutoMEasureSize
            err, _ = device.meta_flash_extend(config)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Calling Flash_Extend to extend partial measurement: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)
            warn_on_response_verification_fail(err)

#*   19. Display actual and expected PCR0 values (they should match)

        # read the actual PCR0 value
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Flash_PCR0_Low_Addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading PCR Low register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        act_LSB_PCR0 = change_word_order(PCR0_read_resp[16:32])
        
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Flash_PCR0_High_Addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading PCR High register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        act_MSB_PCR0 = change_word_order(PCR0_read_resp[16:32])

        # read the expected PCR0 value from Customer EEPROM (where it was stored)
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Customer_EEPROM)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading Customer EEPROM for expected PCR0 Low value: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        exp_MSB_PCR0 = PCR0_read_resp[16:32]
        
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Customer_EEPROM + 0x10)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading Customer EEPROM for expected PCR0 High value: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        exp_LSB_PCR0 = PCR0_read_resp[16:32]

        print("Actual PCR0 as extracted from device register: 0x{}{}".format(act_MSB_PCR0.hex(),act_LSB_PCR0.hex()))
        print("Expected PCR0: 0x{}{}".format(exp_MSB_PCR0.hex(),exp_LSB_PCR0.hex()))

#*   20. attempt to write new data ("good-bye") to adress TEST_BLOCK_ADDR_1 and address TEST_BLOCK_ADDR_2 ("demo_complete") and show that it hasn't changed (because the flash was locked).

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        sleep(1)

        ret = device.meta_page_program(TEST_BLOCK_ADDR_1, bytes("good-bye", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        sleep(1)

        ret = device.meta_page_program(TEST_BLOCK_ADDR_2, bytes("demo_complete", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))


        # restore lock bit
        ret = device.meta_ppb_lock_bit_write()
        if not ret:
            main_logger.warning('Failed attempting to write PPB lock bit.')
        else:
            print("PPB lock bit set.")


        # read back and display data to see if the write worked
        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_1, read_data[:32]))

        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_2))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_2, read_data[:32]))

#*   21. Unlock the device by calling AuthLock command

        err, _ = device.meta_auth_lock()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed cryptographically unlocking the device: (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

#*   22. attempt to write new data ("good-bye") to adress TEST_BLOCK_ADDR_1 and address TEST_BLOCK_ADDR_2 ("demo_complete")... and this time show that it has changed (becuase the flash was unlocked).

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        sleep(1)

        ret = device.meta_page_program(TEST_BLOCK_ADDR_1, bytes("good-bye", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        sleep(1)

        ret = device.meta_page_program(TEST_BLOCK_ADDR_2, bytes("demo_complete", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))


        # restore lock bit
        ret = device.meta_ppb_lock_bit_write()
        if not ret:
            main_logger.warning('Failed attempting to write PPB lock bit.')
        else:
            print("PPB lock bit set.")


        # read back and display data to see if the write worked
        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_1))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_1, read_data[:32]))

        ret, read_data = device.meta_fast_read(TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to read data at address 0x{:X}.'.format(TEST_BLOCK_ADDR_2))
        else:
            #todo perhaps clip to first 0xff character?
            print("read content at address 0x{:X}: {}".format(TEST_BLOCK_ADDR_2, read_data[:32]))

#*   23. Detach from the Endpoint and device, Instruct the user to reboot the Endpoint
#*
        input("Please reboot your endpoint, press [enter] once it is fully booted")

#<reboot>

#*   26. Display Authenta device status (showing crtmMismatch true; and incorrect PCR0)

    device.test_only_set_boot_amtc0(None)
    with device:
        # perform request UID to retrieve AMTC0
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        print("provision_sts_struct after 2 reboot")
        print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))

        # Display Authenta device status (showing crtmMismatch false; and correct PCR0)
        err, _ = device.read_extended_crtm_status()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed attempting to retrieve extended CRTM status from the device: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

        print("\n\n")
        print('last_ACMD:',           device.extended_crtm_status_struct.LACMD)
        print('CRTM_Mismatch:',       device.extended_crtm_status_struct.CRTM_Mismatch)
        print('CRTM_Range_Invalid:',  device.extended_crtm_status_struct.CRTM_Range_Invalid)
        print('CRTM_PCR0_Invalid:',   device.extended_crtm_status_struct.CRTM_PCR0_Invalid)
        print('Auth_Unlock:',         device.extended_crtm_status_struct.Auth_Unlock)
        print('Recovery_Block:',      device.extended_crtm_status_struct.Recovery_Block)
        print('Partial_Measurement:', device.extended_crtm_status_struct.Partial_Measurement)
        print('Recovery_Mismatch:',   device.extended_crtm_status_struct.Recovery_Mismatch)

#*   27. Display actual and expected DICE Secret (FDS) (they should differ)


        config =  b'\x00\x01'
        print("Call Measure to obtain FDS data")
        err, _ = device.meta_measure(config, 0x0, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
        print("Measure completed")

        err, KTR_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.KTR_Base)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading least-significant-bits KTR register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        LSB_KTR = change_word_order(KTR_read_resp[16:32])
        
        err, KTR_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.KTR_Base + 16)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading most-significant-bits KTR register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        MSB_KTR = change_word_order(KTR_read_resp[16:32])

        print("Actual FDS as extracted from KTR: 0x{}{}".format(MSB_KTR.hex(),LSB_KTR.hex()))
        print("Expected FDS: 0x{}".format(HARDCODED_FDS.hex()))
        #TODO calculate FDS in key driver

#*   28. Disable modification of DICE Secret (FDS)
        config =  b'\x00\x02'
        print("Call Measure to disable access to Fs(s)")
        err, _ = device.meta_measure(config, 0x0, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
        print("Measure completed")

#*   29. Complete measurement

        print('Partial_Measurement:', device.extended_crtm_status_struct.Partial_Measurement)
        if(device.extended_crtm_status_struct.Partial_Measurement):
            print("The device has performed a partial measurement, now completing the measurement.")
            config = b'\x02' # bits 2 and 0 are '00' for option 0, bit 1 is '1' for start after max MaxAutoMEasureSize
            err, _ = device.meta_flash_extend(config)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('Calling Flash_Extend to extend partial measurement: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)
            warn_on_response_verification_fail(err)

#*   30. Display actual and expected PCR0 values (they should differ)

        # read the actual PCR0 value
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Flash_PCR0_Low_Addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading PCR Low register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        act_LSB_PCR0 = change_word_order(PCR0_read_resp[16:32])
        
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Flash_PCR0_High_Addr)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading PCR High register: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        act_MSB_PCR0 = change_word_order(PCR0_read_resp[16:32])

        # read the expected PCR0 value from Customer EEPROM (where it was stored)
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Customer_EEPROM)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading Customer EEPROM for expected PCR0 Low value: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)
    
        exp_MSB_PCR0 = PCR0_read_resp[16:32]
        
        err, PCR0_read_resp = device.meta_crtm_register_read(b'\00', CRTM_Registers.Customer_EEPROM + 0x10)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Reading Customer EEPROM for expected PCR0 High value: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
        warn_on_response_verification_fail(err)

        exp_LSB_PCR0 = PCR0_read_resp[16:32]

        print("Actual PCR0 as extracted from device register: 0x{}{}".format(act_MSB_PCR0.hex(),act_LSB_PCR0.hex()))
        print("Expected PCR0: 0x{}{}".format(exp_MSB_PCR0.hex(),exp_LSB_PCR0.hex()))

#*   31. Use Auth-Lock command to unlock the device

        err, _ = device.meta_auth_lock()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('Failed cryptographically unlocking the device: (' + err.cmd_name + '): '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)
        warn_on_response_verification_fail(err)

#*   32. run the following steps:
#*       unlock (volatile) block-lock at address TEST_BLOCK_ADDR_1, and TEST_BLOCK_ADDR_2
#*       erase data in those address;
#        write the text "hello" at address TEST_BLOCK_ADDR_1;
#*       write the text "starting_demo" at address TEST_BLOCK_ADDR_2;
#        re-lock.

        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))
        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))
        sleep(1)
        ret = device.meta_page_program(TEST_BLOCK_ADDR_1, bytes("hello", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))
        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_1)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_1))
        ret = device.meta_dyb_write(0, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to unlock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))
        ret = device.meta_subsector_erase(TEST_BLOCK_ADDR_2)
        sleep(1)
        if not ret:
            main_logger.warning('Failed to erase flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))
        ret = device.meta_page_program(TEST_BLOCK_ADDR_2, bytes("starting_demo", 'utf-8'))
        if not ret:
            main_logger.warning('Failed attempting to write data to flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))
        ret = device.meta_dyb_write(1, TEST_BLOCK_ADDR_2)
        if not ret:
            main_logger.warning('Failed to lock flash memory address 0x{:X} on the device.'.format(TEST_BLOCK_ADDR_2))

        # restore lock bit
        ret = device.meta_ppb_lock_bit_write()
        if not ret:
            main_logger.warning('Failed attempting to write PPB lock bit.')

#finished

    print(" --- <EOF/> --- ")
    exit(0)
    # <EOF/>


def set_automeasure_bit(device, auto_measure_flag: bool):
    """
    Enable or Disable automeasure

    param[in] device the device on which to set the auto_measure_bit
    param[in] aut_measure_flag bool True: enable aut_measure / False: disable auto_measure

    return True on success or Flase on failure
    """
    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    if not err.esr_success:
        print("\n\n")
        main_logger.critical('Failed calling crtm_reagister_read (attempting to set auto_measure bit): '+ err.esr_bit_msg + '; ' + err.error_msg)
        return False
    warn_on_response_verification_fail(err)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")
    if auto_measure_flag:
        data_int = device_config_int | 0x00000001
    else:
        data_int = device_config_int & 0xFFFFFFFE
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_success:
        print("\n\n")
        main_logger.critical('Failed calling crtm_reagister_write (attempting to set auto_measure bit): '+ err.esr_bit_msg + '; ' + err.error_msg)
        return False
    warn_on_response_verification_fail(err)

    return True


# program entry point
if __name__ == "__main__":
   # call main function
   # sys.exit() acts as a thin wrapper -
   # kind of global exception handler
   exit(main())

# <EOF/>
