#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
\t A utility for deprovisioning a Micron Authenta enabled SPI-NOR device
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_deprovision.py  <endpoint alias> 
    -- as in --
   $> app_deprovision.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_deprovision.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device 

KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<device json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    # (Note that when calling meta_request_id on the NaxosDevice
    # object signature verification on the response might fail
    # because the specifics of signing and byte order are
    # dependent on the revision).
    #

    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        # get the provision status from the device if we don't have it yet
        if device.provision_sts_struct is None or device.uid is None or device.revision is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)
            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        #
        # make sure the device is provisioned before deprovisioning
        #
        if device.provision_sts_struct.Provisioned:
            # deprovision device
            err, _ = device.read_extended_crtm_status()
            if not err.esr_success:
                main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

            #
            # deprovision the device
            #
            config = b'\x01' # bit 0 set: reset the EEPROM_Attributes register to default
            err, *_ = device.meta_deprovision_crtm(config)
            if not err.esr_success:
                main_logger.critical('meta_deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

            print("\n\n")
            print("The device has been deprovisioned.")
            main_logger.info("The device has been deprovisioned.")

        else:
            print("\n\n")
            print("request_UID.provision_sts[provisioned] bit unset.")
            print("Device must be provisioned prior to attempting this app. Exiting...")
            main_logger.info("Device is not provisioned.")
            exit(9)

        #
        # refetch the provision status from the device again and display it
        #
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)

        warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

    exit(0)


if __name__ == "__main__":
    main()
