#! /usr/bin/env python3
import argparse as ap
import json
import logging
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

import authenta.authenta_key_mgmt_interface as key_mgmt

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
\t A utility for retrieving current device state of Micron Authenta enabled SPI-NOR device
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_get_status.py <endpoint alias>
    -- as in --
   $> app_get_status.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_get_status.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<device json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main() -> None:
  app_logging.setup_app_logging(DEBUG)
  main_logger = logging.getLogger(__name__)

  envr = Envr(ARGS.device, ARGS.ifn)
  envr.setup()

  ip_params = envr.ep_ip_address.split(":")
  ip_addr = ip_params[0]

  if len(ip_params) > 1:
    port = ip_params[1]
    ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
  else:
    ep = GrpcPlatform(envr.dev_pth, ip_addr)

  # determine which revision of the Naxos device is connecting to
  #
  # (Note that when calling meta_request_id on the NaxosDevice
  # object signature verification on the response might fail
  # because the specifics of signing and byte order are
  # dependent on the revision).
  #
  device = NaxosDevice(ep)
  main_logger.info('fetching Naxos device revision')
  dev_rev = 0x000000
  dev_rev_bytes = device.get_device_revision()
  if dev_rev_bytes is None:
    print("\n\n")
    print("Failed to retrieve revision value from device; Exiting.")
    main_logger.critical('Failed to retrieve revision value from device')
    exit(1)

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
  if dev_rev == 0x010005:
    device = NaxosDevice_1_5(ep)
  else:
    main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
    exit(1)


  with device:
    if device.raw_uid is None or device.revision is None:
      err, *_ = device.meta_request_uid()
      if not err.esr_success:
        print("\n\n")
        main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        exit(2)

    dev_rev = int.from_bytes(device.revision, byteorder='big')

    print("\n\n")
    print("Device UID: {}".format(device.uid.hex()))
    print("Revision:   {}".format(device.revision.hex()))
    print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

    if not device.provision_sts_struct.Provisioned:
      print("\n\n")
      print("Device is not provisioned:... Nothing left to do. Exiting...")
      main_logger.info("Device is not provisioned.")
      exit(9)

    err, _ = device.read_extended_crtm_status()
    if not err.esr_success:
      main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
      exit(1)

    warn_on_response_verification_fail(err)

    # call a signed function for the case in which we don't have the 
    # correct SRK in the key library we should avoid locking the chip because 
    # of too many consecutive non-signed function calls
    err, _ = device.meta_acrtm_update_finish(b'\x00')
    if not err.valid_response:
      print("\n\n")
      main_logger.warning('Call to meta_acrtm_update_finish failed: {}; {}; (non-critical call - continuing...)'.format(err.esr_bit_msg, err.error_msg))

    # build bytes data table from previous block data
    err = device.register_read_all()
    if not err.esr_success:
      print("\n\n")
      main_logger.critical('register_read_all: '+ err.esr_bit_msg + '; ' + err.error_msg)
      exit(1)

    # call a signed function for the case in which we don't have the 
    # correct SRK in the key library we should avoid locking the chip because 
    # of too many consecutive non-signed function calls
    err, _ = device.meta_acrtm_update_finish(b'\x00')
    if not err.valid_response:
      print("\n\n")
      main_logger.warning('Call to meta_acrtm_update_finish failed: {}; {}; (non-critical call - continuing...)'.format(err.esr_bit_msg, err.error_msg))

    # build bytes data table from previous block data
    device.build_crtm_reg_data_byte_table()

    # map CRTM Register attributes to current device data
    device.build_crtm_reg_attr_table()

    # add device provision status bits
    device_attr = device.provision_sts_struct._asdict()

    # add device Extended CRTM Flag bits
    device_attr.update(device.extended_crtm_status_struct._asdict())

    # add all byte-wide CRTM Table Attributes
    device_attr.update(device.crtm_reg_data_table)

    # add CRTM Register Customer RAM Size Data to device status form
    device_attr['Customer_RAM_sz'] = len(device.crtm_reg_data_table['Customer_RAM'])

    device_attr.update(device.Device_CFG._asdict())

    # 2 - retrieve CRTM Register Device_CFG byte structure
    device_attr.update(device.Device_CFG_Volatile._asdict())

    device_attr['UID'] = device.uid.hex()
    device_attr['Revision'] = device.revision.hex()

    # print device status form to stdout
    print(device.DEVICE_STATUS_FORM.format(**device_attr))

  exit(0)

if __name__ == "__main__":
  main()
