#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail, CRTM_Registers

import authenta.naxos.utils as utils

import app_common_logging as app_logging

VERSION = 0.1
PROGRAM_DES = '''DESCRIPTION:
\t A sample app to demonstrate the Half Device Swap feature in Naxos 1.5
\t This app will (1) set Half_Device_Swap_Enable bit in Device_Config register;
\t (2) call Auth_Modify (option #3 with config[3] Swap_Unswap# set) to enter swapped state;
\t (3) call Auth_Modify (option #3 with config[3] Swap_Unswap# clear) to enter unswapped state;
\t (4) clear Half_Device_Swap_Enable bit in Device_Config regiser.
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_half_device_swap.py  <endpoint alias> 
    -- as in --
   $> app_half_device_swap.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_half_device_swap.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR:

COPYRIGHT: Micron Technology, Inc.; 2020

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument('-f', dest='ifn', metavar='<device json mapping file>',
               default='devices.json', type=ap.FileType('r'),
               help='JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.')

p.add_argument('device', metavar='<device alias>',
               help='endpoint-device alias found in devices.json file.')

p.add_argument('-v', '--verbosity', action='store_true',
               help='prints conditionals where var DEBUG is True.')

p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help='echo script version to stdout.')

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main() -> None:
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(':')
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    # (Note that when calling meta_request_id on the NaxosDevice
    # object signature verification on the response might fail
    # because the specifics of signing and byte order are
    # dependent on the revision).
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        main_logger.critical('Failed to retrieve revision value from device; Exiting...')
        exit(1)

    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    elif dev_rev == 0x010004:
        main_logger.critical('Half Device Swap feature is not available in Naxos revision {}!'.format(dev_rev))
        exit(1)
    else:
        main_logger.critical('Unrecognized Naxos revision {}!'.format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print('\n\n')
        print('Device UID: {}'.format(device.uid.hex()))
        print('Revision:   {}'.format(device.revision.hex()))
        print('provision_sts_struct.Provisioned: {}'.format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned:
            main_logger.critical('Device is not provisioned:... Nothing left to do. Exiting...')
            exit(9)
  
        # Check device swap state
        err, response = device.meta_crtm_register_read(b'\x00', CRTM_Registers.Device_Config)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        device_config_int = int.from_bytes(response[16:20], byteorder='big')
        swapped = (device_config_int & 0x00000020 != 0)
        
        if swapped:
            main_logger.critical('Device is in swapped state (Unexpected state). Exiting...')
            exit(1)
        
        print('Device is in unswapped state')

        # Set Half_Device_Swap_Enable bit in Device_Config register
        data_int = device_config_int | 0x00000040
        data = data_int.to_bytes(4, byteorder='big')

        err, _ = device.meta_acrtm_update_start(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(b'\x00', CRTM_Registers.Device_Config, data, True)
        if not err.esr_success:
            device.meta_acrtm_update_finish(b'\x00')
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        print('Set Half_Device_Swap_Enable bit in Device_Config register')

        half_addr = 0x800000
        half_size = 0x800000

        # Call measure to get digest of second half device
        ret = utils.set_crtm_addr_range_reg(device, b'\x01', 0, half_addr, half_size)
        if not ret:
            main_logger.critical('Failed to set CRTM_Addr0 and CRTM_Size0')
            exit(1)

        ret = utils.set_crtm_addr_range_reg(device, b'\x01', 1, 0x0, 0x0)
        if not ret:
            main_logger.critical('Failed to set CRTM_Addr1 and CRTM_Size1')
            exit(1)

        err, _ = device.meta_measure(b'\x00\x08', 0x0, 0x0, 0x0)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
        if digest is None:
            main_logger.critical('Failed to read Measure result from KTR')
            exit(1)

        print('Issued Measure to get the digest of the second half device')

        # Auth_Modify (option #3 with config[3]: Swap_Unswap# set): enter swapped state
        err, _ = device.meta_auth_modify(b'\x00\x0b', half_addr, 0x0, 0x0, source_content_digest=digest)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        print('Issued Auth_Modify (option 3 with config[3]: Swap_Unswap# set) to enter swapped state')

        # Read Device_Config to verify that device is in swapped state
        err, response = device.meta_crtm_register_read(b'\x00', CRTM_Registers.Device_Config)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        device_config_int = int.from_bytes(response[16:20], byteorder='big')
        if device_config_int & 0x00000020 != 0:
            print('Verified that Device_Config[5]: Half_Device_Swap_Active is set')
        else:
            main_logger.critical('Failed to verify that Device_Config[5]: Half_Device_Swap_Active is set')
            exit(1)

        # Auth_Modify (option #3 with config[3]: Swap_Unswap# clear): enter unswapped state
        err, _ = device.meta_auth_modify(b'\x00\x03', 0x0, 0x0, 0x0, source_content_digest=digest)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        print('Issued Auth_Modify (option 3 with config[3]: Swap_Unswap# clear) to enter unswapped state')

        # Read Device_Config to verify that device is in unswapped state
        err, response = device.meta_crtm_register_read(b'\x00', CRTM_Registers.Device_Config)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        device_config_int = int.from_bytes(response[16:20], byteorder='big')
        if device_config_int & 0x00000020 == 0:
            print('Verified that Device_Config[5]: Half_Device_Swap_Active is clear')
        else:
            main_logger.critical('Failed to verify that Device_Config[5]: Half_Device_Swap_Active is clear')
            exit(1)

        # Clear Half_Device_Swap_Enable bit in Device_Config register
        data_int = device_config_int & 0xFFFFFFBF
        data = data_int.to_bytes(4, byteorder='big')

        err, _ = device.meta_acrtm_update_start(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(b'\x00', CRTM_Registers.Device_Config, data, True)
        if not err.esr_success:
            device.meta_acrtm_update_finish(b'\x00')
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)
    
        warn_on_response_verification_fail(err)

        print('Cleared Half_Device_Swap_Enable bit in Device_Config register')   

    exit(0)


if __name__ == '__main__':
    main()
