#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail, CRTM_Registers

import app_common_logging as app_logging

VERSION = 0.1
PROGRAM_DES = '''DESCRIPTION:
\t A sample app to demonstrate the ID_Key_Select feature in Naxos 1.5
\t This app will (1) set the ID_Key_Select bit in Device_Config register;
\t (2) call request_uid, and verify the signature using the session key;
\t (3) clear the ID_Key_Select bit in Device_Config register.
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_id_key_select.py  <endpoint alias> 
    -- as in --
   $> app_id_key_select.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_id_key_select.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR:

COPYRIGHT: Micron Technology, Inc.; 2020

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument('-f', dest='ifn', metavar='<device json mapping file>',
               default='devices.json', type=ap.FileType('r'),
               help='JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.')

p.add_argument('device', metavar='<device alias>',
               help='endpoint-device alias found in devices.json file.')

p.add_argument('-v', '--verbosity', action='store_true',
               help='prints conditionals where var DEBUG is True.')

p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help='echo script version to stdout.')

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main() -> None:
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(':')
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    # (Note that when calling meta_request_id on the NaxosDevice
    # object signature verification on the response might fail
    # because the specifics of signing and byte order are
    # dependent on the revision).
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        main_logger.critical('Failed to retrieve revision value from device; Exiting...')
        exit(1)

    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    elif dev_rev == 0x010004:
        main_logger.critical('ID_Key_Select feature is not available in Naxos revision {}!'.format(dev_rev))
        exit(1)
    else:
        main_logger.critical('Unrecognized Naxos revision {}!'.format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print('\n\n')
        print('Device UID: {}'.format(device.uid.hex()))
        print('Revision:   {}'.format(device.revision.hex()))
        print('provision_sts_struct.Provisioned: {}'.format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned:
            main_logger.critical('Device is not provisioned:... Nothing left to do. Exiting...')
            exit(9)

        # Set the ID_Key_Select bit in Device_Config register
        err, response = device.meta_crtm_register_read(b'\x00', CRTM_Registers.Device_Config)
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        device_config = response[16:20]
        device_config_int = int.from_bytes(device_config, 'big')
        device_config_int_new = device_config_int | 0x00000080
        device_config_new = device_config_int_new.to_bytes(4, 'big') 
        
        err, _ = device.meta_acrtm_update_start(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, response = device.meta_crtm_register_write(b'\x00', CRTM_Registers.Device_Config, device_config_new, True)
        if not err.esr_success:
            device.meta_acrtm_update_finish(b'\x00')
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        print('Set ID_Key_Select bit in Device_Config')

        # Request_UID
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)

        print('Issued Request_UID')

        warn_on_response_verification_fail(err)

        if err.signature_matched:
            print('Request_UID response verified using session key')
        else:
            print('Request_UID response not verified using session key')

        # Clear the ID_Key_Select bit in Device_Config register
        device_config_int_new = device_config_int & 0xFFFFFF7F
        device_config_new = device_config_int_new.to_bytes(4, 'big') 
        
        err, _ = device.meta_acrtm_update_start(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, response = device.meta_crtm_register_write(b'\x00', CRTM_Registers.Device_Config, device_config_new, True)
        if not err.esr_success:
            device.meta_acrtm_update_finish(b'\x00')
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(b'\x00')
        if not err.esr_success:
            main_logger.critical(err.cmd_name + ': '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        print('Cleared ID_Key_Select bit in Device_Config')

    exit(0)


if __name__ == "__main__":
    main()
