#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 08:59:03 2020

@author: Marshall Shapiro
"""


import argparse as ap
import logging
from os import path

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device import NaxosDevice
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail, CRTM_Registers
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

import app_common_logging as app_logging

VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
\t A utility to connect to a Micron Authenta enabled SPI-NOR device,
\t get the UID, connect to Micron KMS, get the SRK, and save it into
\t the key library.
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_kms_confirm_device_identity.py  <endpoint alias> 
  -- as in --
   $> app_kms_confirm_device_identity.py snor1
  - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_kms_confirm_device_identity.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_kms_confirm_device_identity.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general error at request_uid
  2 : general error

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2020

'''


p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, 
                      description=PROGRAM_DES,
                      epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")


p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

p.add_argument('device', metavar="<device alias>", help="endpoint-device alias found in devices.json file.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

def confirm_device_identity_using_kms():
  """
  Confirm that you are accessing a valid Naxos device using the KMS Proxy for validation.
  
  1. connect to the Micron KMS and get a nonce, 
     USE certificate passed in command line, or default command-line cert to establish secure connection
  2. connect to input Naxos device (should not have be connected already when passed to this function)
  3. call Request_UID passing the KMS produced nonce
  4. Verify Request_UID response with the KMS (identity Verification)

  Return 0 of device identity confirmed, non-zero otherwise
  """

  envr = Envr(ARGS.device, ARGS.ifn)
  envr.setup()

  ip_params = envr.ep_ip_address.split(":")
  ip_addr = ip_params[0]

  if len(ip_params) > 1:
    port = ip_params[1]
    platform = GrpcPlatform(envr.dev_pth, ip_addr, port)
  else:
    platform = GrpcPlatform(envr.dev_pth, ip_addr)

  #
  # Set the password from command line inputs. If neither exists, ask user for password
  #
  if ARGS.tls_certificate_pw:
      cert_password = ARGS.tls_certificate_pw
  elif path.exists(ARGS.tls_certificate_pw_file):
      fp = open(ARGS.tls_certificate_pw_file)
      cert_password = fp.readline()
  else:
      cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")

  #
  # create a connection to the Micron KMS Proxy
  #
  connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)

  #
  # get a nonce from the Micron KMS Proxy
  #
  nonce, nonce_tag, _ = MicronKMSProxyAPI.getNonceWithMPN(connection, "MTANXOBD01")
  if nonce is None:
    # if nonce is None, error message will be in nonce_tag
    print("Failed to receive from KMS: {}".format(nonce_tag))
    return 1
  main_logger.info('nonce received from KMS')

  nonce_bytes = bytes.fromhex(nonce)
  nonce_bytes_len = len(nonce_bytes)
  if nonce_bytes_len > 4:
    print('Nonce retrieved from KMS is large to use in Naxos Request_UID command!')
    return 1
  padding = 4 - nonce_bytes_len
  if padding > 0:
    nonce_bytes = ('\x00' * padding) + nonce_bytes

  #
  # call Request_UID command on attached device using the nonce
  # retrieved from the Micron KMS Proxy
  #
  device = NaxosDevice(platform)
  err, response = get_request_uid_response(device, nonce_bytes)
  if not err.esr_success:
    print('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
    return 1

  print("\n")
  print('Device UID:                       {}'.format(device.uid_as_str))
  print('Naxos Revision:                   {}'.format(device.revision.hex()))
  print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))
  print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))
  print('A-MTC0:                           {}'.format(device._amtc0.hex()))
  print("\n")

  #
  # verify the request UID response from the chip
  #
  rc_int, *_ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, "MTANXOBD01", 'false')
  if rc_int == 0:
    print("\nDevice identity NOT VERIFIED! Verify Request_UID response from KMS failed!")
    return 1
 
  print("\nDevice identity VERIFIED!")

  return 0


def get_request_uid_response(device, nonce: bytes):
  """
    Make a device.meta_request_uid call outside a 'with' statement.
    I.e. if necessary, the platform.acquire_device() is done here.
  """
  try:
    acquired_device_here = False
    if not device.endpoint_platform.is_device_acquired():
      device.endpoint_platform.acquire_device(device.device_type)
      acquired_device_here = True

    err, response, *_ = device.meta_request_uid(nonce=nonce)
    return err, response
  finally:
    if acquired_device_here:
      device.endpoint_platform.release_device(device.device_type)



# program entry point
if __name__ == "__main__":
# call main function
   # sys.exit() acts as a thin wrapper -
   # kind of global exception handler
  rc = confirm_device_identity_using_kms()
  exit(rc)
