#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 08:59:03 2020

@author: Marshall Shapiro
"""


import argparse as ap
import logging
from os import path

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device import NaxosDevice
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI

import app_common_logging as app_logging

VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
\t KMS MPN Recover Device Example
\t A utility to connect to a Micron Authenta enabled SPI-NOR device,
\t get the UID, connect to Micron KMS, get the SRK, and save it into
\t the key library.
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_demo.py  <endpoint alias> 
  -- as in --
   $> app_demo.py snor1
  - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_demo.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_demo.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general error at request_uid
  2 : general error

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''


p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, 
                      description=PROGRAM_DES,
                      epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")


p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

p.add_argument('device', metavar="<device alias>", help="endpoint-device alias found in devices.json file.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

class static_vars:
    sig_connection = None

def get_deprovision_crtm_sig(device: NaxosDevice, command: bytes)->bytes:
    sig, rc_msg, status = MicronKMSProxyAPI.GenerateCmdSignature("signDeprovisionCRTMCommand", \
        command.hex(), "", static_vars.sig_connection, "MTANXOBG01", device.uid_as_str)
    if sig is not None:
        sigbytes = bytes.fromhex(sig)
    else:
        print("ERROR: Failed to fetch Deprovision_CRTM signature from the KMS Proxy! {}".format(rc_msg))
        print("Status code: {}\n".format(status))
        sigbytes = b''

    return sigbytes

def deprovision_using_kms_signing():
  """
  Perform Recover Device (MPN MTANXOBG01)

  Deprovision the atatched device using the KMS Proxy to sign the Deprovision_CTRM command.
  
  1.	connect to the Micron KMS and get a nonce, 
          USE certificate passed in command line, or default command-line cert to establish secure connection
  2.	connect to input Naxos device (should not have be connected already when passed to this function)
  3.  call Request_UID passing the KMS produced nonce
  5.	Verify Request_UID response with the KMS (identity Verification)
  6.	Get signature from for deprovision command from the KMS
  7.  Send signed deprovision command to the Naxos device
  7.	Verify response with KMS Proxy

  """
  envr = Envr(ARGS.device, ARGS.ifn)
  envr.setup()

  ip_params = envr.ep_ip_address.split(":")
  ip_addr = ip_params[0]

  if len(ip_params) > 1:
    port = ip_params[1]
    platform = GrpcPlatform(envr.dev_pth, ip_addr, port)
  else:
    platform = GrpcPlatform(envr.dev_pth, ip_addr)

  #
  # Set the password from command line inputs. If neither exists, ask user for password
  #
  if ARGS.tls_certificate_pw:
      cert_password = ARGS.tls_certificate_pw
  elif path.exists(ARGS.tls_certificate_pw_file):
      fp = open(ARGS.tls_certificate_pw_file)
      cert_password = fp.readline()
  else:
      cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")

  #
  # create a connection to the Micron KMS Proxy
  #
  connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)

  #
  # get a nonce from the Micron KMS Proxy
  #
  nonce, nonce_tag, status = MicronKMSProxyAPI.getNonceWithMPN(connection, "MTANXOBG01")
  if nonce is None:
        # if nonce is None, error message will be in nonce_tag
        print("Failed to receive from KMS: {}".format(nonce_tag))
        print("Status code: {}\n".format(status))
        exit(1)
  main_logger.info('nonce received from KMS')

  nonce_bytes = bytes.fromhex(nonce)
  nonce_bytes_len = len(nonce_bytes)
  if nonce_bytes_len > 4:
    print('Nonce retrieved from KMS is large to use in Naxos Request_UID command!')
    exit(1)
  padding = 4 - nonce_bytes_len
  if padding > 0:
    nonce_bytes = ('\x00' * padding) + nonce_bytes

  #
  # call Request_UID command on attached device using the nonce
  # retrieved from the Micron KMS Proxy
  #
  device = NaxosDevice(platform)
  err, response = get_request_uid_response(device, nonce_bytes)
  if not err.esr_success:
    print('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
    exit(1)

  print("\n")
  print('Device UID:                       {}'.format(device.uid_as_str))
  print('Naxos Revision:                   {}'.format(device.revision.hex()))
  print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))
  print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))
  print('A-MTC0:                           {}'.format(device._amtc0.hex()))
  print("\n")

  #
  # verify the request UID response from the chip
  #
#  print('\nNOTE this must be the first Authenta commands run on boot for command response verification to succeed.\n')
  rc_int, rc_msg, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, "MTANXOBG01", 'false')
  if rc_int == 0:
    print("Verify Request_UID response from KMS failed!")
    print("Error message: {}".format(rc_msg))
    print("Status code: {}\n".format(status))
    return 1

  dev_rev = int.from_bytes(device.revision, byteorder='big')
  main_logger.info('Found attached Naxos device revision {:X}'.format(dev_rev))
  if dev_rev == 0x010005:
    device = NaxosDevice_1_5(platform)
  else:
    main_logger.critical("Unrecognized Naxos revision {:X}!".format(dev_rev))
    return 1

  with device:
    if device.provision_sts_struct is None:
      err, *_ = device.meta_request_uid()
      if not err.esr_success:
        print("\n\n")
        main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        return 2

      warn_on_response_verification_fail(err)


    # provision the device
    if not device.provision_sts_struct.Provisioned:
      print("\nThe device is not provisioned.  Nothing to do.\n")
      return 0

    uid = device.uid_as_str; 

    # do a register read that includes the PowerOnAMTC value and
    # verify with KMS to supply the KMS with this value, (which it
    # needs to derive the session key).
    print("reading register 0xD0 in order to fetch the PowerOnAMTC0 from the Naxos device\n")
    err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
    if not err.esr_success:
      print("\n\n")
      main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
      return 1

    print("Verifying CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state\n")
    verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, "MTANXOBG01", uid)
    if verified is None:
      print('KMS verifyCRTMRegisterRWResponse Failed: {}; {} \n'.format(http_status, rc_msg))
      return 1

    device.kms_connection = connection

    print('Deprovisioning device...')
    # get the uid because it will be needed to verify the deprovision response, 
    # but device data will be cleared internally after successful deprovision
    static_vars.sig_connection = connection
    err, deprov_response = device.meta_deprovision_crtm(b'\x00', deprovision_signature_func=get_deprovision_crtm_sig)
    if not err.esr_success:
      print("\n\n")
      main_logger.critical('deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
      return 1

    print("need to verify the deprovision command signature")
    verified, rc_msg, status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyGenericACRTMResponse", deprov_response.hex(), connection, "MTANXOBG01", uid)
    if verified is None:
          print('KMS verifyGenericACRTMResponse Failed: {}'.format(rc_msg))
          print("Status code: {}\n".format(status))
          return 1

    if(verified):
      print("\nsuccessfully verified deprovision_crtm response with the KMS Proxy\n")
    else:
      print("\nSignature verification using the KMS Proxy of deprovision_crtm response FAILED\n")

  return 0


def get_request_uid_response(device, nonce: bytes):
  """
    Make a device.meta_request_uid call outside a 'with' statement.
    I.e. if necessary, the platform.acquire_device() is done here.
  """
  try:
    acquired_device_here = False
    if not device.endpoint_platform.is_device_acquired():
      device.endpoint_platform.acquire_device(device.device_type)
      acquired_device_here = True

    err, response, *_ = device.meta_request_uid(nonce=nonce)
    return err, response
  finally:
    if acquired_device_here:
      device.endpoint_platform.release_device(device.device_type)



# program entry point
if __name__ == "__main__":
# call main function
   # sys.exit() acts as a thin wrapper -
   # kind of global exception handler
  deprovision_using_kms_signing()
  exit(0)
