#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os
from sys import exit
from collections import namedtuple
from hashlib import sha256
from os import path

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
import authenta.authenta_key_mgmt_interface as key_mgmt

import app_common_logging as app_logging

VERSION=0.8
PROGRAM_DES = '''DESCRIPTION: 
KMS Replace Server Root Key (MPN MTANXSBC01) example
\t A utility for replacing the SRK secret in a Micron Authenta enabled SPI-NOR device.
\t If the device is Naxos 1.5, the app has an option "-d" to replace SRK with a derived SRK (encrypted).
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_replace_srk.py  <endpoint alias> <new-srk-name>
    -- as in --
   $> app_replace_srk.py snor1 new-srk
    - where 'snor1' is a device entry in the devices.json file
    - where new-srk-name, is the 'name' of a key entry defined in key_contents_prv.xml associated with the given UID

   # for test cases where the default devices.json file isn't what is needed
   $> app_replace_srk.py -f mydevices_mapping.json snor1  new-srk

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : device not in desired provision state

KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device',       metavar="<device alias>",           help="endpoint-device alias found in devices.json file.")
p.add_argument('new_key_name',  metavar="<replacement-key name>", help="key name of new SRK")

# command-line options
p.add_argument("-d", dest='derive_key', action='store_true', help=\
    "derive the new key when doing Replace_SRK (set Config[4]: Derive_Direct#) (only avaiable for Naxos 1.5)")

p.add_argument("-f",  dest='ifn',  metavar="<device json mapping file>", default="devices.json", type=ap.FileType('r'),
help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

p.add_argument("-v", "--verbosity",  action='store_true',
                help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
                help="echo script version to stdout.")

ARGS  = p.parse_args()
DEBUG = ARGS.verbosity

app_logging.setup_app_logging(DEBUG)
main_logger = logging.getLogger(__name__)


class static_vars:
    sig_connection = None
    new_srk_hash = None

def get_command_sig(operation: str, mpn: str, device: NaxosDevice, command: bytes, additional_data: str="")->bytes:
    """
    send the input command string to the Micron KMS Proxy to get the command signature
    and return the signature returned from the kms_connection

    This is a generic sign command. It will sign the command specified in operation argument
    to this function.  That string must be one of : “signReplaceSRKCommand,” “signProvisionCRTMCommand,”
    or “signDeprovisionCRTMCommand”

    NOTE For Replace_SRK the additional_data argument must contain the hash of the value enter
    into the KTR register (formatted as a hex string). For other commands additional_data should be then
    empty string.

    NOTE static_vars.sig_connection must be set to a valid connection to the KMS proxy before this 
    function is called!

    ARGS
        operation   Name of the operation; one of “signReplaceSRKCommand”, “signProvisionCRTMCommand”
                    or “signDeprovisionCRTMCommand”
        mpn         The MPN to use
        device      The Naxos Device instance
        command     THe command string to be signed
        additional_data  For Replace_SRK this must be the hash (SHA256) of the contents of KTR
    
    Return the signature as a bytes
    """
    sig, rc_msg, status = MicronKMSProxyAPI.GenerateCmdSignature(operation, \
        command.hex(), additional_data, static_vars.sig_connection, mpn, device.uid_as_str)
    if sig is not None:
        sigbytes = bytes.fromhex(sig)
    else:
        print("ERROR: Failed to fetch Deprovision_CRTM signature from the KMS Proxy! {}".format(rc_msg))
        print("Status code: {}\n".format(status))
        sigbytes = b''

    return sigbytes

def get_replace_srk_sig(device: NaxosDevice, command: bytes)->bytes:
    """
    NOTE static_vars.sig_connection must be set to a valid connection to the KMS proxy before this 
         function is called!

    NOTE static_vars.new_srk_hash must be set to the hash (SHA256) of the contents of the KTR_IDX_LST
         before calling this function
    """
    return get_command_sig("signReplaceSRKCommand", "MTANXSBC01", device, command, static_vars.new_srk_hash)

    
def get_provision_sig(device: NaxosDevice, command: bytes)->bytes:
    """
    NOTE static_vars.new_srk_hash must be set to the hash (SHA256) of the contents of the KTR_IDX_LST
         before calling this function
    """
    return get_command_sig("signProvisionCRTMCommand", "MTANXSBC01", device, command)


def main():
    new_srk_name = ARGS.new_key_name

    # get the hash of the new SRK - needed by KMS to sign Replace_SRK
    # static_vars.new_srk_hash = sha256(srk_new_bytes_lst).hexdigest()


    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # Set the password from command line inputs. If neither exists, ask user for password
    #
    if ARGS.tls_certificate_pw:
        cert_password = ARGS.tls_certificate_pw
    elif path.exists(ARGS.tls_certificate_pw_file):
        fp = open(ARGS.tls_certificate_pw_file)
        cert_password = fp.readline()
    else:
        cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")


    #
    # create a connection to the Micron KMS Proxy
    #
    connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)
    static_vars.sig_connection = connection

    nonce_bytes, nonce_tag = get_nonce_from_kms_proxy(connection, "MTANXSBC01")
    if nonce_bytes is None:
        # if nonce is None, error message will be in nonce_tag
        print("Failed to receive nonce from KMS: {}".format(nonce_tag))
        exit(1)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    err, response = get_request_uid_response(device, nonce_bytes)
    if not err.esr_success:
        print('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
        exit(1)


    #
    # verify the request UID response from the chip
    #

    # if the device is provisioned this function will not provision the device.
    # in that case the Boot A-MTC0 value will have to be set on the KMS in 
    # VerifyRequestUIDResponse. Therefore we will set it here, but that
    # creates a requirment that this Request_UID be the first command on boot.
    # (On the other hand if it is not provisioned we will provision in this 
    # function and the KMS will get the PowerOnAMTC0 value from verifying the 
    # provison.)
    # if device.provision_sts_struct is not None and device.provision_sts_struct.Provisioned:
    #     useAMTC0asPowerOnAMTC0 = 'true'
    # else:
    #     useAMTC0asPowerOnAMTC0 = 'false'
    useAMTC0asPowerOnAMTC0 = 'false'

    rc_int, rc_msg, status = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, "MTANXSBC01", useAMTC0asPowerOnAMTC0)
    if rc_int == 0:
        print("Verify Request_UID response from KMS failed!")
        print("Error message: {}".format(rc_msg))
        print("Status code: {}\n".format(status))
        return 1

    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(device.revision.hex()))

    replace_srk_config = b"\x00"

    if dev_rev == 0x010005:
        if ARGS.derive_key:
            replace_srk_config = b"\x10"
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                return 2

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        # get the uid because it will be needed in multiple places later
        uid = device.uid_as_str; 

        new_key = key_mgmt.export_priv_key_as_hex(uid, new_srk_name)
        if new_key is None:
            print(f'\nFailed to load new key from key repository: uid=\'{uid}\'; key_name=\'{new_srk_name}\'\n')
            exit(5)
            
        new_key_bytes = bytes.fromhex(new_key)
        static_vars.new_srk_hash = sha256(new_key_bytes).hexdigest()

        if device.provision_sts_struct.Provisioned and not device.provision_sts_struct.Remotely_Provisioned:
            print("\nThe device is Locally Provisioned.  Deprovision the device and try again.\n")
            main_logger.info("The device is Locally Provisioned. Aborting.")
            exit(9)

        if device.provision_sts_struct.Provisioned:
            # do a register read that includes the PowerOnAMTC value and
            # verify with KMS to supply the KMS with this value, (which it
            # needs to derive the session key).
            print("reading register 0xD0 in order to fetch the PowerOnAMTC0 from the Naxos device\n")
            err, rr_resp = device.meta_crtm_register_read(b'\x00', 0xD0)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
                return 1

            print("Verifying CRTM_Register_Read of address 0xD0 to add the PowerOnAMTC0 to the KMS session state\n")
            verified, rc_msg, http_status =  MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", rr_resp.hex(), connection, "MTANXOBG01", uid)
            if verified is None:
                print('KMS verifyCRTMRegisterRWResponse Failed: {}; {} \n'.format(http_status, rc_msg))
                return 1
        else:
            # note that pwrOnCount will be set with the provision and included in provision response
            print("\nThe device is not provisioned.  Provisioning...\n")
            main_logger.info("Device is not provisioned. Provisioning...")


            err, prov_response, sent_nonce = device.meta_provision_crtm(b'\x05', nonce=nonce_bytes, provision_signature_func=get_provision_sig)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('deprovision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
                return 1

            print("need to verify the provision command signature")
            verified, rc_msg, status = MicronKMSProxyAPI.VerifyResponseSignature("verifyProvisionCRTMResponse", prov_response.hex(), connection, "MTANXSBC01", uid)
            if verified is None:
                print('KMS verifyProvisionCRTMResponse Failed: {} \n'.format(rc_msg))
                print("Status code: {}\n".format(status))
                return 1

            if(verified):
                print("\nsuccessfully verified provision_crtm response with the KMS Proxy\n")
            else:
                print("\nSignature verification using the KMS Proxy of provision_crtm response Failed!\n")

        # 
        # update KTR register
        # 
        # for replace_srk option #0 (replace SRK with new customer SRK) KTR 
        # must be set to hold the new SRK
        #

        # Key Transfer Register - 32byte sub-register in CRTM_register
        # used to transfer SRK/UDS keys into Authenta SPI-NOR device.
        # CRTM Register writes are always in 4byte chunks.
        # KTR_IDX_LST is a time ordered list of starting indeces 
        # of those chunks.
        KTR_IDX_LST = [0xcc,0xc8,0xc4,0xc0,0xbc,0xb8,0xb4,0xb0]

        config = b'\x00'
        for reg_num, idx in enumerate(KTR_IDX_LST):
            data_blk = new_key_bytes[(reg_num*4):(reg_num+1)*4]

            err, reg_write_resp = device.meta_crtm_register_write(config, idx, data_blk, sign=False)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

            print("\nverify the register_write to the KTR for SRK tranch {} of 7".format(reg_num))
            verified, rc_msg, status = MicronKMSProxyAPI.VerifyResponseSignature("verifyCRTMRegisterRWResponse", reg_write_resp.hex(), connection, "MTANXSBC01", uid)
            if verified is None:
                print('KMS verifyCRTMRegisterRWResponse Failed: {} \n'.format(rc_msg))
                print("Status code: {}\n".format(status))
                return 1

            if(verified):
                print("successfully verified crtm_register_write response with the KMS Proxy\n")
            else:
                print("Signature verification using the KMS Proxy of crtm_register_write response Failed!\n")

        # 
        # Note, after successful completion the key library must be updated to associate
        # the new SRK with this UID.
        #
        new_key_hash = sha256(new_key_bytes).digest()
        err, repl_srk_resp, sent_nonce = device.meta_replace_srk(replace_srk_config, new_key_hash, nonce=nonce_bytes, replace_srk_signature_func=get_replace_srk_sig)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('replace_srk: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)
        if replace_srk_config == b"\x10":
            print("\nThe SRK has been replaced with a key derived from the key with name " + new_srk_name + "\n")
        else:
            print("\nThe SRK has been replaced with the key with name " + new_srk_name + "\n")

        # Create a copy of the original key in the key library
        key_mgmt.copy_key(uid, 'srk', 'srk_copy')

        # Remove original key from the key library
        key_mgmt.delete_key(uid, 'srk')

        # Add new (current) key to the key library
        key_mgmt.add_key(uid, 'srk', bytes.fromhex(new_key), None)


        print("Need to verify the replace srk command signature")
        verified, rc_msg, status = MicronKMSProxyAPI.VerifyResponseSignature("verifyReplaceSRKResponse", repl_srk_resp.hex(), connection, "MTANXSBC01", device.uid_as_str)
        if(verified):
            print("\nUnexpected success: verified replace SRK response with the KMS Proxy")
            print("Assume SRK was replaced with the same key\n")
            # Delete the original key
            key_mgmt.delete_key(uid, 'srk_copy')

        else:
            print("\nExpected failure: Signature verification using the KMS Proxy of replace_srk response Failed!")
            print("SRK has been replaced with a new key\n")
            # Delete the incorrect srk
            key_mgmt.delete_key(uid, 'srk')
            # Make new srk from the copy
            key_mgmt.copy_key(uid, 'srk_copy', 'srk')
            # Delete the copy
            key_mgmt.delete_key(uid, 'srk_copy')

    exit(0)

def get_request_uid_response(device, nonce: bytes):
    """
    Make a device.meta_request_uid call outside a 'with' statement.
    I.e. if necessary, the platform.acquire_device() is done here.
    """
    try:
        acquired_device_here = False
        if not device.endpoint_platform.is_device_acquired():
            device.endpoint_platform.acquire_device(device.device_type)
            acquired_device_here = True

        err, response, *_ = device.meta_request_uid(nonce=nonce)
        return err, response
    finally:
        if acquired_device_here:
            device.endpoint_platform.release_device(device.device_type)


def get_nonce_from_kms_proxy(connection, mpn):
    """
    get a nonce from the Micron KMS Proxy
    """

    nonce, nonce_tag, status = MicronKMSProxyAPI.getNonceWithMPN(connection, mpn)
    main_logger.info('nonce received from KMS')

    if nonce is None or nonce_tag is None:
        if nonce is None:
            # if nonce is None, error message will be in nonce_tag
            print("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
            print("Status code: {}\n".format(status))
        else:
            # something is wrong with what was returned by the KMS
            # this should not happen (nonce was returned from KMS and nonce_tag is None)
            print("Failed to receive valid nonce from KMS: missing nonce_tag\n")
            print("Status code: {}\n".format(status))

        return nonce, nonce_tag

    nonce_bytes = bytes.fromhex(nonce)
    nonce_bytes_len = len(nonce_bytes)
    if nonce_bytes_len > 4:
        print('Nonce retrieved from KMS is large to use in Naxos Request_UID command!')
        return None, None
    padding = 4 - nonce_bytes_len
    if padding > 0:
        nonce_bytes += '\x00' * padding

    return nonce_bytes, nonce_tag


if __name__ == "__main__":
    main()
