#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 08:59:03 2020

@author: Marshall Shapiro
"""


import argparse as ap
import logging
from os import path

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
import authenta.kms_proxy.MicronKMSProxyAPI as MicronKMSProxyAPI
import authenta.authenta_key_mgmt_interface as key_lib_key_mgmt

import authenta.kms_proxy.crypto_utils as cr
from   Cryptodome.PublicKey import RSA
from   Cryptodome.Cipher import AES
from   Cryptodome.Cipher import PKCS1_OAEP
from   Cryptodome.Hash import SHA256
from   Cryptodome.Signature import pss

import app_common_logging as app_logging

VERSION = 0.01

PROGRAM_DES = '''DESCRIPTION:
\t KMS Request Server Root Key (MPN MTANXSBB01) Example
\t A utility to connect to a Micron Authenta enabled SPI-NOR device,
\t get the UID, connect to Micron KMS, get the SRK, and save it into
\t the key library.
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_demo.py  <endpoint alias> 
  -- as in --
   $> app_demo.py snor1
  - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_demo.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_demo.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general error at request_uid
  2 : general error

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''


p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, 
                      description=PROGRAM_DES,
                      epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")


p.add_argument("-tls_cert", dest='tls_certificate', metavar="<TLS cert file>",
               default="../../kms_proxy_certs/active_kms_tls_client_cert.pem",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw", dest='tls_certificate_pw', metavar="<TLS cert password>",
               help="client certificate file for TLS session with the KMS server")

p.add_argument("-tls_cert_pw_file", dest='tls_certificate_pw_file', metavar="<TLS cert password file>",
                default="../../kms_proxy_certs/self_signed_authenta_key.txt",
                help="file containing client certificate password for TLS session with KMS server")

p.add_argument('device', metavar="<device alias>", help="endpoint-device alias found in devices.json file.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

def fetch_key_from_kms_store_in_local():
  """
  Request Server Root Key (MPN MTANXSBB01)
  
  Fetch an SRK from Micron KMS for the attached device
  and store it in the local key repository, as follows...
  1.	connect to the Micron KMS and get a nonce,
  2.	connect to input Naxos device (should not have be connected already when passed to this function)
  3.  call Request_UID passing the KMS produced nonce
  4.  generate (or get previously generated) RSA key
  5.	uses the Request_UID response and our generated public key to fetch the SRK from the KMS,
  6.	uses the matching private key to decrypt the returned SRK, and
  7.	Save that decrypted SRK into our local key store

  """
  envr = Envr(ARGS.device, ARGS.ifn)
  envr.setup()

  ip_params = envr.ep_ip_address.split(":")
  ip_addr = ip_params[0]

  if len(ip_params) > 1:
    port = ip_params[1]
    platform = GrpcPlatform(envr.dev_pth, ip_addr, port)
  else:
    platform = GrpcPlatform(envr.dev_pth, ip_addr)

  #
  # Set the password from command line inputs. If neither exists, ask user for password
  #
  if ARGS.tls_certificate_pw:
      cert_password = ARGS.tls_certificate_pw
  elif path.exists(ARGS.tls_certificate_pw_file):
      fp = open(ARGS.tls_certificate_pw_file)
      cert_password = fp.readline()
  else:
      cert_password = input("Please enter the password for the client certificate file for " + \
          "TLS session with the KMS server. (Press Enter if no password is required): ")

  #
  # create a connection to the Micron KMS Proxy
  #
  connection = MicronKMSProxyAPI.createMTTLS(ARGS.tls_certificate, cert_password)

  #
  # get a nonce from the Micron KMS Proxy
  #
  nonce, nonce_tag, status = MicronKMSProxyAPI.getNonceWithMPN(connection, "MTANXSBB01")
  if nonce is None:
        # if nonce is None, error message will be in nonce_tag
        print("Failed to receive nonce from KMS: {}\n".format(nonce_tag))
        print("Status code: {}\n".format(status))
        exit(1)
  if nonce_tag is None:
        # something is wrong with what was returned by the KMS
        # this should not happen (nonce was returned from KMS and nonce_tag is None)
        print("Failed to receive valid nonce from KMS: missing nonce_tag\n")
        print("Status code: {}\n".format(status))
        exit(1)

  main_logger.info('nonce received from KMS')

  nonce_bytes = bytes.fromhex(nonce)
  nonce_bytes_len = len(nonce_bytes)
  if nonce_bytes_len > 4:
    print('Nonce retrieved from KMS is large to use in Naxos Request_UID command!')
    exit(1)
  padding = 4 - nonce_bytes_len
  if padding > 0:
    nonce_bytes += '\x00' * padding

  #
  # call Request_UID command on attached device using the nonce
  # retrieved from the Micron KMS Proxy
  #
  device = NaxosDevice(platform)
  err, response = get_request_uid_response(device, nonce_bytes)
  if not err.esr_success:
    print('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
    exit(1)

  print("\n")
  print('Device UID:                       {}'.format(device.uid_as_str))
  print('Naxos Revision:                   {}'.format(device.revision.hex()))
  print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))
  print(device.PROVISION_STS_FORM.format(**device.provision_sts_struct._asdict()))
  print("\n")


  #
  # send the request UID response from the chip to the KMS to be verified
  #
  rc_int, *_ = MicronKMSProxyAPI.VerifyRequestUIDResponse(device.uid_as_str, response.hex(), connection, nonce_tag, "MTANXSBB01", 'false')
  if rc_int == 0:
    print("\nAborting: Device identity NOT VERIFIED! Verify Request_UID response from KMS failed!")
    main_logger.error("Device identity NOT VERIFIED! Verify Request_UID response from KMS failed!")
    return 1

  #
  # read the public RSA encryption key from file. This key will be sent to the KMS and
  # the KMS will encrypt the SRK with this key.
  #
  pubkeydata=[]
  try:
    main_logger.info("opening and reading public key file ./tmp/kms_rsa_encryption_key.pub")
    fp=open("./tmp/kms_rsa_encryption_key.pub", "r")
    for i in fp:
      pubkeydata.append(i.rstrip("\n"))
  except FileNotFoundError as err:
    main_logger.error("Missing public key file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
    main_logger.error("Please generate a public/private key pair for the KMS server to encrypt the SRK it sends back by running generate_rsa_encrypt_keys.py")
    return 1
  except OSError as err:
    main_logger.error("Failed to read public key from file ./tmp/kms_rsa_encryption_key.pub; {}: {}".format(type(err), err))
    return 1
  
  #
  # Read UID from KMS (passing the UID of the attached Naxos
  # device and the response buffer of the call to Request_UID)
  #
  transportKey_enc, iv, tag, SRK_enc, err_msg, status = MicronKMSProxyAPI.getKey(device.uid_as_str, connection, nonce, ''.join(pubkeydata[1:-1]))

  if transportKey_enc is None or iv is None or tag is None or SRK_enc is None:
    main_logger.info('Failed to retrieve SRK for device {} from KMS! {} {}'.format(device.uid_as_str, status, err_msg ))
    print('\nFailed to retrieve SRK for device {} from KMS! {} {}\n'.format(device.uid_as_str, status, err_msg))
    return 2

  #
  # decrypt the SRK
  #
  SRK = None
  try:
    main_logger.info("importing private key file ./tmp/kms_rsa_encryption_key.prv SRK decryption")
    privatekey = RSA.importKey(open('./tmp/kms_rsa_encryption_key.prv').read())
    cipher= PKCS1_OAEP.new(privatekey, hashAlgo=SHA256, mgfunc=lambda x,y: pss.MGF1(x,y, SHA256))
    transportkey = cipher.decrypt(bytes.fromhex(transportKey_enc))
    
    # Decrypt the SRK using the transport key    
    aesCipher = AES.new(transportkey, AES.MODE_GCM,bytes.fromhex(iv))
    SRK = aesCipher.decrypt_and_verify(bytes.fromhex(SRK_enc),bytes.fromhex(tag))
  except FileNotFoundError as err:
    main_logger.error("Missing private key file ./tmp/kms_rsa_encryption_key.prv; {}: {}".format(type(err), err))
    main_logger.error("Please generate a public/private key pair for the KMS server to encrypt the SRK it sends back by running generate_rsa_encrypt_keys.py")
  except OSError as err:
    main_logger.error("{}: {}".format(type(err), err))
  except ValueError as err:
    main_logger.error("{}: {}".format(type(err), err))
  except TypeError as err:
    main_logger.error("{}: {}".format(type(err), err))

  #
  # if an SRK was returned save it in the local key store
  # associated with the UID of the attached device
  #
  if SRK is not None:
    device.logger.info('retrieved SRK for device 0x{} from KMS. Adding to key library...'.format(device.uid_as_str))
    print('\nSRK successfully retrieved from KMS!')
    ret = key_lib_key_mgmt.add_key(device.uid_as_str, 'srk', SRK, None)
    if ret:
      device.logger.info('Added SRK to local keystore with device UID {}'.format(device.uid_as_str))
      print('Added SRK to local keystore with device UID {}\n'.format(device.uid_as_str))
    else:
      device.logger.info('Failed attempting to add retrieved SRK to local key store!')
      print('Failed attempting to add retrieved SRK to local key store!\n')
  else:
    device.logger.info('Failed to decrypt SRK for device {}!'.format(device.uid_as_str))
    print('\nFailed to decrypt SRK for device {}\n'.format(device.uid_as_str))



def get_request_uid_response(device, nonce: bytes):
  """
    Make a device.meta_request_uid call outside a 'with' statement.
    I.e. if necessary, the platform.acquire_device() is done here.
  """
  try:
    acquired_device_here = False
    if not device.endpoint_platform.is_device_acquired():
      device.endpoint_platform.acquire_device(device.device_type)
      acquired_device_here = True

    err, response, *_ = device.meta_request_uid(nonce=nonce)
    return err, response
  finally:
    if acquired_device_here:
      device.endpoint_platform.release_device(device.device_type)



# program entry point
if __name__ == "__main__":
# call main function
   # sys.exit() acts as a thin wrapper -
   # kind of global exception handler
  fetch_key_from_kms_store_in_local()
  exit(0)
