#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail, change_word_order

import app_common_logging as app_logging

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
    \t A utility for measuring a provisioned, Micron Authenta SPI-NOR's protected memory blocks.
    These memory blocks are defined, both starting index and size, in the 
    CRTM Register:CRTM_Addr[0-7] attributes. 
'''

EPILOG = '''

    EXAMPLE CMDLNS:
    # commandline using default mapping files
    $> app_measure.py  <endpoint alias> 
        -- as in --
    $> app_measure.py snor1
        - where 'snor1' is a device entry in the devices.json file

    # for test cases where the default devices.json file isn't what is needed
    $> app_measure.py -f mydevices_mapping.json snor1

    EXIT CODES:
    0 : success
    1 : general application error
    2 : Request_UID command failure

    KNOWN BUGS:

    TODOS:

    AUTHOR:
    \tAlan Haffner

    COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="REQUIRED: to print command's pass/fail message and exit status.")

p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

ARGS = p.parse_args()

def main():
    app_logging.setup_app_logging()
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned:
            print("\n\n")
            print("Naxos not provisioned, provision it")
            print("Attempting to provision device")
            main_logger.info("Device is not provisioned.")

            config = b'\x05'

            device.meta_provision_crtm(config)

        config = b'\x01'

        reg_write_arguments = [
            {'config': config, 'crtm_addr': 0x20,'data': (0x00000).to_bytes(4, 'big'), 'sign': False},
            {'config': config, 'crtm_addr': 0x24,'data': (0x20000).to_bytes(4, 'big'), 'sign': False},
            {'config': config, 'crtm_addr': 0x28,'data': (0x40000).to_bytes(4, 'big'), 'sign': False},
            {'config': config, 'crtm_addr': 0x2C,'data': (0x20000).to_bytes(4, 'big'), 'sign': False},
            {'config': config, 'crtm_addr': 0x30,'data': (0x00000).to_bytes(4, 'big'), 'sign': False},
            {'config': config, 'crtm_addr': 0x34,'data': (0x00000).to_bytes(4, 'big'), 'sign': False},
        ]

        for reg_args in reg_write_arguments:
            err, _ = device.meta_crtm_register_write(**reg_args)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('crtm_register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

        config = b"\x00\x08"
        err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('measure: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, LSB_KTR = device.meta_crtm_register_read(b"\x00", 0xB0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        LSB_KTR = change_word_order(LSB_KTR[0x10:0x20])

        err, MSB_KTR = device.meta_crtm_register_read(b"\x00", 0xC0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        MSB_KTR = change_word_order(MSB_KTR[0x10:0x20])

        print("Measure command result extracted from KTR, in Naxos device: 0x{}{}".format(MSB_KTR.hex(), LSB_KTR.hex()))

        err, eeprom_attrib_bytes = device.meta_crtm_register_read(b'\x00', 0xD0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        eeprom_attrib_bytes = eeprom_attrib_bytes[0x10:0x14]
        main_logger.info("bytes read from - CRTM_reg:eeprom_attrib_bytes:{}".format(eeprom_attrib_bytes.hex()))

        my_array = bytearray(eeprom_attrib_bytes)

        # reading one byte of a bytearray returns a type(int)
        # 'AND' the byte /w to keep upper nibble and discard bits from lower nibble
        # 'OR' the read byte with 0x03
        # finally - overwriting byte[3]
        my_array[3] = (my_array[3] & 0xf0) | 0x03
        main_logger.info("munged bytes from - CRTM_reg:eeprom_attrib_bytes:{}".format(my_array.hex()))

        eeprom_attrib_bytes = bytes(my_array)
        eeprom_attrib_int = int.from_bytes(eeprom_attrib_bytes, 'big')
        main_logger.info("CRTM_reg: modified to force last nibble to 0x3:{:08x}".format(eeprom_attrib_int))

        config = b'\x00'
        err, _ = device.meta_acrtm_update_start(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_start: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, _ = device.meta_crtm_register_write(config, 0xD0, eeprom_attrib_bytes, sign=True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_finish: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        config = b'\x00\x88'
        err, _ = device.meta_measure(config, 0x0, 0x0, 0x200)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('measure: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        config = b'\x00'
        err, BUFF_L = device.meta_crtm_register_read(config, 0x200)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        BUFF_L = change_word_order(BUFF_L[0x10:0x20])
        main_logger.info("measurement low 16B from eeprom 0x200:{}".format(BUFF_L.hex()))

        err, BUFF_H = device.meta_crtm_register_read(config, 0x210)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        BUFF_H = change_word_order(BUFF_H[0x10:0x20])
        main_logger.info("measurement high 16B from eeprom 0x200:{}".format(BUFF_H.hex()))
        print("Measure command result extracted from EEPROM, in Naxos device: 0x{}{}".format(BUFF_H.hex(), BUFF_L.hex()))

    exit(0)


if __name__ == "__main__":
    main()
