#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os
from sys import exit

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
import authenta.naxos.utils as utils

# import utils

import app_common_logging as app_logging

VERSION = 0.01
PROGRAM_DES = '''DESCRIPTION:
A sample application to demonstrate how an OTA update to a Naxos device can be accomplished
with the auth_modify command, using option #1 to program the chip and option #0 to erase it.

The sequence should be as follows

1. remote host configures a block as AM block, 
2. writes image version 1 to the configured AM block using auth_modify commands
3. perform a measure() on that image and print the result
4. attempt to modify the image using a page_program() call
5. perform a measure() on that image and print the result, show it is the same as in step 3
6. repeat step 2-5 using image version 2 and show the measurements concur, but are different from version 1.
'''

EPILOG = '''
EXAMPLE Stdout - applicatoin logs or error messages

EXAMPLE CMDLNS:
   app_ota_update.py  <endpoint alias>
    -- as in --
   app_ota_update.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   app_ota_update.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general error
  2 : general request uid error

KNOWN BUGS:

TODOS:

AUTHOR:

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, \
    description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'), \
    help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>", help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true', \
    help="REQUIRED: to print command's pass/fail message and exit status.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION), \
    help="echo script version to stdout.")

ARGS = p.parse_args()

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

def set_measurement_range(device, measure_addr, measure_size):
    """
    set measurment range for the image
    return True on success or False on failure
    """
    crtm_addrs = (0x20, 0x24, 0x28, 0x2c)
    data_to_write = (measure_addr, measure_size, 0x0, 0x0)

    for i in range(0, len(crtm_addrs)):
        err, _ = device.meta_crtm_register_write(b"\x01", crtm_addrs[i], \
            data_to_write[i].to_bytes(4, byteorder="big"), False)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to write data to CRTM register 0x%X" %(crtm_addrs[i]))
            main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
            return False

    return True

def main():
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    image_size = 0x1000
    images = []

    # content of image 1
    images.append(bytes.fromhex("11") * image_size)

    # content of image 2
    images.append(bytes.fromhex("22") * image_size)

    # address of the legacy block to write the image to
    image_addr = 0xe30000

    # address of block to be configured to AM
    am_block_addr = 0xe20000

    # AMBitmap register address for the block
    am_bitmap_addr = 0x19c

    # AMBitmap regsiter content to set
    am_bitmap_data = b"\xff\xff\xff\xfb"

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    # (Note that when calling meta_request_id on the NaxosDevice
    # object signature verification on the response might fail
    # because the specifics of signing and byte order are
    # dependent on the revision).
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            # Request_UID
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                return False

        # check provision status
        if device.provision_sts_struct.Provisioned == 0:
            print("Please provision the device before running this app")
            main_logger.info("Device is not provisioned.")
            return False

        # configure a block as AM
        err, _ = device.meta_acrtm_update_start(b"\x00")
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("acrtm_update_start: " + err.esr_bit_msg + "; " + err.error_msg)
            return False

        err, _ = device.meta_crtm_register_write(b"\x00", am_bitmap_addr, am_bitmap_data, True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to configure block 0x%08X as AM block" %(am_block_addr))
            main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
            device.meta_acrtm_update_finish(b"\x00")
            return False

        err, _ = device.meta_acrtm_update_finish(b"\x00")
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("acrtm_update_finish: " + err.esr_bit_msg + "; " + err.error_msg)
            return False

        main_logger.info("Configured block 0x%08X as AM block" %(am_block_addr))

        measurements = [None] * 2

        # perform same operations for image 1 and 2
        for i in range(0, len(images)):
            image = images[i]

            # write image to a legacy block
            main_logger.info("Writing image %d to legacy block 0x%08X" %(i + 1, image_addr))
            if not utils.write_data_to_block(device, image_addr, image):
                return False
            main_logger.info("Wrote image %d to legacy block 0x%08X" %(i + 1, image_addr))

            # device.meta_auth_modify(option #0): erase the AM block 
            err, _ = device.meta_auth_modify(b"\x00\x00", am_block_addr, 0x0, 0x0)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical("Failed to call Auth_Modify to erase AM block 0x%08X" %(am_block_addr))
                main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
                return False
            main_logger.info("Called Auth_Modify to erase AM block 0x%08X" %(am_block_addr))

            # write content of image to AM using auth_modify
            # device.meta_auth_modify(option #1): program the AM block
            err, _ = device.meta_auth_modify(b"\x00\x01", image_addr, am_block_addr, image_size, image)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical("Failed to call Auth_Modify to write image %d to AM block 0x%08X" %(i + 1, am_block_addr))
                main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
                return False
            main_logger.info("Called Auth_Modify to write image %d to AM block 0x%08X" %(i + 1, am_block_addr))

            # set measurement range
            ret = set_measurement_range(device, am_block_addr, image_size)
            if not ret:
                return False

            # measure the image in AM block
            err, _ = device.meta_measure(b"\x00\x00", 0x0000, 0x0000, 0x0000)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical("Failed to measure image %d in AM block" %(i + 1))
                main_logger.critical("measure: " + err.esr_bit_msg + "; " + err.error_msg)
                return False
            main_logger.info("Measured image %d in AM block" %(i + 1))

            # read results from KTR
            result = utils.get_flash_digest(device, 0xB0)
            if result is None:
                return False
            main_logger.info("Measurement result for image %d: %s" %(i + 1, result.hex()))
            measurements[i] = result

            # attempt to modify the image in AM block using page_program
            main_logger.info("Updating image %d in AM block using Page_Program" %(i + 1))
            if not utils.write_data_to_block(device, am_block_addr, bytes.fromhex("00000000")):
                return False
            main_logger.info("Updated image %d in AM block using Page_Program" %(i + 1))

            # measure the image in AM block
            err, _ = device.meta_measure(b"\x00\x00", 0x0000, 0x0000, 0x0000)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical("Failed to meausre image %d in AM block" %(i + 1))
                main_logger.critical("measure: " + err.esr_bit_msg + "; " + err.error_msg)
                return False
            main_logger.info("Measured image %d in AM block" %(i + 1))

            # read results from KTR
            result = utils.get_flash_digest(device, 0xB0)
            if result is None:
                return False
            main_logger.info("New measurement result for image %d: %s" %(i + 1, result.hex()))

            # compare measure results
            if measurements[i] == result:
                print("New measurement of image %d is the same with the old one" %(i + 1))
            else:
                print("New measurement of image %d is different from the old one" %(i + 1))

        # compare measurements of image 1 and image 2
        if measurements[0] == measurements[1]:
            print("Measurement of image 1 is the same with that of image 2")
        else:
            print("Measurement of image 1 is different from that of image 2")

        # configure AM block back to legacy block
        err, _ = device.meta_acrtm_update_start(b"\x00")
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("acrtm_update_start: " + err.esr_bit_msg + "; " + err.error_msg)
            return False

        err, _ = device.meta_crtm_register_write(b"\x00", am_bitmap_addr, b"\xff" * 4, True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to configure block 0x%08X back to legacy block" %(am_block_addr))
            main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
            device.meta_acrtm_update_finish(b"\x00")
            return False

        err, _ = device.meta_acrtm_update_finish(b"\x00")
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("acrtm_update_finish: " + err.esr_bit_msg + "; " + err.error_msg)
            return False

        main_logger.info("Configured block 0x%08X back to legacy block" %(am_block_addr))

    return True

# program entry point
if __name__ == "__main__":
    # call main function
    # sys.exit() acts as a thin wrapper -
    # kind of global excdevicetion handler
    exit(not main())

# <EOF/>
