#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
\t A utility for provisioning a Micron Authenta enabled SPI-NOR device
'''

EPILOG = '''

EXAMPLE CMDLNS:

   # commandline using default mapping files
   $> app_provision.py  <endpoint alias> 
    -- as in --
   $> app_provision.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where device automeasure is desired on power-on
   $> app_provision.py -a  <endpoint alias> 

   # for test cases where the default devices.json file isn't what is needed
   $> app_provision.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : provisioned device 


KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''
p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")


p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

p.add_argument("-a", dest='SetAutoMeasure', action='store_true',
               default="True", help="Set AutoMeasure True.")

ARGS  = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        if device.provision_sts_struct.Provisioned:
            print("\n\n")
            print("request_UID.provision_sts[provisioned] bit set.")
            print("Device is already provisioned. Exiting...")
            main_logger.info("Device has already been provisioned.")
            exit(9)

        print("\n\n")
        print('Provisioning device...')
        config = b'\x05'
        err, *_ = device.meta_provision_crtm(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('provision_crtm: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        if not ARGS.SetAutoMeasure:
            exit(0)

        # code block - update CRTM Measurment Blocks
        err, resp = device.meta_crtm_register_read(b"\x00", 0xD4)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        boot_amtc0 = resp[20:24]
        main_logger.info("CRTM_reg:PwrOnCount:{}".format(boot_amtc0.hex()))

        config = b'\x00'
        reg_write_arguments = [
            # set CRTM_Max_Auto_Measure_Size
            {'config': config, 'crtm_addr': 0xA8, 'data': (0x30000).to_bytes(4, 'big'), 'sign': True},

            # set start idx of CRTM_Addr0
            {'config': config, 'crtm_addr': 0x20, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},

            # set CRTM_Addr0 Size
            {'config': config, 'crtm_addr': 0x24, 'data': (0x20000).to_bytes(4, 'big'), 'sign': True},

            # set start idx of CRTM_Addr1
            {'config': config, 'crtm_addr': 0x28, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},

            # set CRTM_Addr1 Size. Zero length Address terminates linked list chain
            {'config': config, 'crtm_addr': 0x2C, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},
        ]

        for reg_args in reg_write_arguments:
            err, _ = device.meta_crtm_register_write(**reg_args)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('crtm_register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)
            main_logger.info("crtm_register_write(0x{:X}, 0x{})".format(reg_args['crtm_addr'], reg_args['data'].hex()))

        # code block: set Automeasure in CRTM Register
        # read CRTM Register offset 0xA0
        err, data_in = device.meta_crtm_register_read(config, 0xA0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # get CRTM_Register::Device_Config bytes
        device_cfg_bytes = data_in[0x10:0x14]

        # covert bytes to type int
        device_cfg_int = int(device_cfg_bytes.hex(),16)

        # set CRTM_Register::Device_Config to Automeasure
        device_cfg_int = device_cfg_int | 0x01

        print(".~" * 36)
        # write CRTM_Register::Device_Config - set Automeasure to On
        err, _ = device.meta_crtm_register_write(config, 0xA0, device_cfg_int.to_bytes(4, 'big'), sign=True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)
        main_logger.info("crtm_register_write(0x{:X}, 0x{:X})".format(0xA0, device_cfg_int))

        # code block: check Automeasure is set in CRTM Register
        # read CRTM Register offset 0xA0
        err, data_in = device.meta_crtm_register_read(config, 0xA0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # get CRTM_Register::Device_Config bytes
        device_cfg_bytes = data_in[0x10:0x14]

        automesure_bit = bool(int(device_cfg_bytes.hex(),16) % 2)
        if not automesure_bit:
            # automeasure not set in CRTM Register Device_Config register
            # exit /w error
            print("\n\n")
            main_logger.critical("Failed to set CRTM_Register:Device_Config:Automeasure bit")
            exit(1)

        # code block - dump Provision_sts data
        err, *_ = device.meta_request_uid()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(2)

        warn_on_response_verification_fail(err)
        print('provisioning complete.')
        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

    exit(0)


if __name__ == "__main__":
    main()
