#! /usr/bin/env python3
import argparse as ap
import sys
import os
import logging

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION:
\t SPI-NOR Flash Read-Write Test
'''
EPILOG = '''

EXAMPLE CMDLNS:

EXIT CODES:

KNOWN BUGS:

TODOS:

AUTHOR:

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")


p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

p.add_argument("-a", dest='SetAutoMeasure', action='store_true',
               default="True", help="Set AutoMeasure True.")

ARGS  = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:

        print("------- Flash Read Write Test -------")


        addr = 0x00F80000
        print("Address:{0:x}".format(addr))

        # Read
        err, data_read = device.meta_fast_read(addr)
        # print("Data read:{}".format(data_read.hex()))
        print(data_read)

        # Unlock
        err = device.meta_dyb_write(False, addr)

        # Erase
        err = device.meta_subsector_erase(addr)

        # Read
        err, data_read = device.meta_fast_read(addr)
        # print("Data read:{}".format(data_read.hex()))
        print(data_read)

        # Write
        # test_data = bytes.fromhex("0123456789ABCDEF")
        test_data = b"This is test data"
        err = device.meta_page_program(addr, test_data)

        # Read
        err, data_read = device.meta_fast_read(addr)
        # print("Data read:{}".format(data_read.hex()))
        print(data_read)

        # xxxxxxxxxx Erase / Write / Read
        err = device.meta_subsector_erase(addr)
        err, data_read = device.meta_fast_read(addr)
        err = device.meta_page_program(addr, test_data)
        err, data_read = device.meta_fast_read(addr)
        print(data_read)    

        # xxxxxxxxxx Erase / Write / Read
        err = device.meta_subsector_erase(addr)
        err = device.meta_page_program(addr, test_data)
        err, data_read = device.meta_fast_read(addr)
        print(data_read)

        # xxxxxxxxxx Erase / Write / Read
        err = device.meta_subsector_erase(addr)
        err = device.meta_page_program(addr, test_data)
        err, data_read = device.meta_fast_read(addr)
        print(data_read)

        # xxxxxxxxxx Erase / Write / Read
        err = device.meta_subsector_erase(addr)
        err = device.meta_page_program(addr, test_data)
        err, data_read = device.meta_fast_read(addr)
        print(data_read)                      

    exit(0)


if __name__ == "__main__":
    main()
