#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

VERSION = 0.9
PROGRAM_DES = '''DESCRIPTION: 
\t A utility for setting up device recovery mode in a Micron Authenta enabled SPI-NOR device
'''

EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_recovery.py  <endpoint alias> 
    -- as in --
   $> app_recovery.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_recovery.py -f mydevices_mapping.json snor1
  
   # for setting CRTM_Register:Device_Config:Recovery mode bit
   $> app_recovery.py -r snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  3 : user attempting unset a cleared Recovery Mode bit
  4 : user attempting to set an already set Recovery Mode bit
  9 : unprovisioned device

KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technoloty, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)


p.add_argument("-f",  dest='ifn',  metavar="<device json mapping file>", default="devices.json", type=ap.FileType('r'),
help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device',  metavar="<device alias>",  help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity",  action='store_true',
                help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
                help="echo script version to stdout.")

p.add_argument("-r", dest='recovery', action='store_true', help="set the CRTM_Register:Device_Config:Recovery mode bit.")


ARGS  = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned:
            print("\n\n")
            print("request_UID.provision_sts[provisioned] bit unset. Exiting...")
            main_logger.info("Device is not provisioned.")
            exit(9)

        err, _ = device.read_extended_crtm_status()
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # get CRTM_Register[device_config] bytes
        # set config to nonvolatile read
        config = b'\x00'
        err, cmd_response = device.meta_crtm_register_read(config, 0xA0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('crtm_register_read: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # get device_config bytes from CRTM_reg_read Response
        crtm_reg_device_cfg = cmd_response[0x10:0x14]

        # CRTM_Register.Device_Config[1] == recovery mode bit
        # bytes are kind-of strings in python...
        # unless you grab a one byte slice, then its an int type
        recovery_mode_enabled = True if crtm_reg_device_cfg[-1] & 0x2 else False

        main_logger.info("CRTM_Block_Read_Response:{}".format(recovery_mode_enabled))
        main_logger.info("crtm_reg_device_cfg:{}".format(crtm_reg_device_cfg.hex()))

        main_logger.info("ARGS.SetAutoMeasure is:{}".format(ARGS.recovery))
        main_logger.info("Device_Config.recovery_mode_enabled:{}".format(recovery_mode_enabled))

        if not recovery_mode_enabled and not ARGS.recovery:
            print("\n\n")
            print("Device recovery mode bit not set and the user has asked to clear it. Nothing to do. Exiting...")
            main_logger.critical("Recovery mode bit not set")
            exit(3)

        if recovery_mode_enabled and ARGS.recovery:
            print("\n\n")
            print("User asked to set recovery mode and its already set. Nothing to do. Exiting...")
            main_logger.critical("Recovery mode bit is already set")
            exit(4)

        # convert bytes type to int type
        crtm_reg_device_cfg_int = int(crtm_reg_device_cfg.hex(), 16)

        # clear out automeasure, recovery bits.  keep everything else
        crtm_reg_device_cfg_int &= 0xfffffff4

        if ARGS.recovery:
            # CRTM_Register.Device_Config[1] == recovery mode bit
            # CRTM_Register.Device_Config[0] == automeasure bit
            crtm_reg_device_cfg_int = 0x3 | crtm_reg_device_cfg_int
        else:
            # only bits 0-2 are defined in CRTM_Register.Device_Config
            # - all other bits are reserved
            crtm_reg_device_cfg_int = 0x1 | crtm_reg_device_cfg_int

        crtm_reg_device_cfg_hex = "{:08x}".format(crtm_reg_device_cfg_int)
        main_logger.info("__crtm_reg_device_cfg_hex__:{}".format(crtm_reg_device_cfg_hex))

        config = b'\x00'
        err, _ = device.meta_acrtm_update_start(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_start: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        config = b'\x00'
        reg_write_arguments = [
            {'config': config, 'crtm_addr': 0xA8, 'data': (0x30000).to_bytes(4, 'big'), 'sign': True},
            {'config': config, 'crtm_addr': 0x20, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},
            {'config': config, 'crtm_addr': 0x24, 'data': (0x20000).to_bytes(4, 'big'), 'sign': True},
            {'config': config, 'crtm_addr': 0x28, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},
            {'config': config, 'crtm_addr': 0x2C, 'data': (0x00000).to_bytes(4, 'big'), 'sign': True},
        ]

        for reg_args in reg_write_arguments:
            print(".~" * 36)
            err, _ = device.meta_crtm_register_write(**reg_args)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('crtm_register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)
            main_logger.info("crtm_register_write(0x{:X}, 0x{})".format(reg_args['crtm_addr'], reg_args['data'].hex()))

        # update CRTM_Register - Recovery_Base_Addr0
        err, _ = device.meta_crtm_register_write(config, 0xA4, 0x480000.to_bytes(4, 'big'), sign=True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # update CRTM_Register - Device_Config bytes
        err, _ = device.meta_crtm_register_write(config, 0xA0, crtm_reg_device_cfg_int.to_bytes(4, 'big'), sign=True)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        err, _ = device.meta_acrtm_update_finish(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('update_finish: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # Execute Flash_ext command /w option #1
        # - Start digest calculation at beginning of linked list
        # config[0] set
        # config[1] clear
        # config[2] clear
        # ========================
        config = b'\x01'
        err, _ = device.meta_flash_extend(config)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical('flash_extend: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

    exit(0)


if __name__ == "__main__":
    main()
