#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os
from sys import exit
from collections import namedtuple

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail
import authenta.authenta_key_mgmt_interface as key_mgmt

import app_common_logging as app_logging
from hashlib import sha256

VERSION=0.8
PROGRAM_DES = '''DESCRIPTION: 
\t A utility for replacing the SRK secret in a Micron Authenta enabled SPI-NOR device.
\t If the device is Naxos 1.5, the app has an option "-d" to replace SRK with a derived SRK (encrypted).
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_replace_srk.py  <endpoint alias> <new-srk-name>
    -- as in --
   $> app_replace_srk.py snor1 new-srk
    - where 'snor1' is a device entry in the devices.json file
    - where new-srk-name, is the 'name' of a key entry defined in key_contents_prv.xml associated with the given UID

   # for test cases where the default devices.json file isn't what is needed
   $> app_replace_srk.py -f mydevices_mapping.json snor1  new-srk

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : device not in desired provision state

KNOWN BUGS:

TODOS:

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

# command-line positional arguments
p.add_argument('device',       metavar="<device alias>",           help="endpoint-device alias found in devices.json file.")
p.add_argument('new_key_name',  metavar="<replacement-key name>", help="key name of new SRK")

# command-line options
p.add_argument("-d", dest='derive_key', action='store_true', help=\
    "derive the new key when doing Replace_SRK (set Config[4]: Derive_Direct#) (only avaiable for Naxos 1.5)")

p.add_argument("-f",  dest='ifn',  metavar="<device json mapping file>", default="devices.json", type=ap.FileType('r'),
help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument("-v", "--verbosity",  action='store_true',
                help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION),
                help="echo script version to stdout.")

ARGS  = p.parse_args()
DEBUG = ARGS.verbosity

def main():
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)
    
    new_srk_name = ARGS.new_key_name

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    
    replace_srk_config = b"\x00"

    if dev_rev == 0x010005:
        if ARGS.derive_key:
            replace_srk_config = b"\x10"
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned or not device.provision_sts_struct.Remotely_Provisioned:
            print("Device is not remotely provisioned. Nothing to do. Exiting...")
            main_logger.info("Device is not provisioned for remote use.")
            exit(9)

        # 
        # update KTR register
        # 
        # for replace_srk option #0 (replace SRK with new customer SRK) KTR 
        # must be set to hold the new SRK
        #

        # Key Transfer Register - 32byte sub-register in CRTM_register
        # used to transfer SRK/UDS keys into Authenta SPI-NOR device.
        # CRTM Register writes are always in 4byte chunks.
        # KTR_IDX_LST is a time ordered list of starting indeces 
        # of those chunks.
        KTR_IDX_LST = [0xcc,0xc8,0xc4,0xc0,0xbc,0xb8,0xb4,0xb0]
        new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, new_srk_name)
        new_key_bytes = bytes.fromhex(new_key)

        config = b'\x00'
        for reg_num, idx in enumerate(KTR_IDX_LST):
            data_blk = new_key_bytes[(reg_num*4):(reg_num+1)*4]

            err, _ = device.meta_crtm_register_write(config, idx, data_blk, sign=False)
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('register_write: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

        # 
        # For Naxos 1.5, replace the current SRK with a derived SRK (derived from srk_new)
        #
        # Note, after successful completion the key library must be updated to associate
        # the new SRK with this UID.
        #
        new_key_hash = sha256(new_key_bytes).digest()
        error_code, response_bytes, sent_nonce = device.meta_replace_srk(replace_srk_config, new_key_hash)
        if not error_code.esr_success:
            print("\n\n")
            main_logger.critical('replace_srk: '+ error_code.esr_bit_msg + '; ' + error_code.error_msg)
            exit(1)

        warn_on_response_verification_fail(error_code)
        if replace_srk_config == b"\x10":
            print("The SRK has been replaced with a key derived from the key named " + new_srk_name)
        else:
            print("The SRK has been replaced with the key named " + new_srk_name)

        # Create a copy of the original key in the key library
        key_mgmt.copy_key(device.uid_as_str, 'srk', 'srk_copy')

        # Remove original key from the key library
        key_mgmt.delete_key(device.uid_as_str, 'srk')

        # Add new (current) key to the key library
        key_mgmt.add_key(device.uid_as_str, 'srk', bytes.fromhex(new_key), None)


        # Update the boot_amtc0 and clear server_session_key
        device._boot_amtc0 = device._amtc0
        device._server_session_key = None

        # Verify response from meta_replace_srk
        device.verify_replace_srk_response(response_bytes, sent_nonce, error_code,  sign_with_session_key=True, force=True)

        # if verification checks pass, delete the copy of the original key
        if error_code.valid_response:
            key_mgmt.delete_key(device.uid_as_str, 'srk_copy')
        # if checks fail, switch the key in the xml file back to the original 
        # using the copied key
        else:
            # Delete the incorrect srk
            key_mgmt.delete_key(device.uid_as_str, 'srk')
            # Make new srk from the copy
            key_mgmt.copy_key(device.uid_as_str, 'srk_copy', 'srk')
            # Delete the copy
            key_mgmt.delete_key(device.uid_as_str, 'srk_copy')


    print(" --- <EOF/> --- ")
    exit(0)


if __name__ == "__main__":
    main()
