#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import warn_on_response_verification_fail

import app_common_logging as app_logging

VERSION=0.9
PROGRAM_DES = '''DESCRIPTION:
\t A utility for setting (or clearing) Auth_Lock, Unlock on a Micron Authenta enabled SPI-NOR device.
\t This will allow updating of locked flash memory blocks of flash memory.
\t Once set, the flag can be cleared again, (going back to the standard block-lock flag settings), by 
\t calling this function again using "-clear" or "-c" argument, or by hard reset.
'''
EPILOG = '''

EXAMPLE CMDLNS:
   # commandline using default mapping files
   $> app_set_auth_unlock.py  <endpoint alias> 
    -- as in --
   $> app_set_auth_unlock.py snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   $> app_set_auth_unlock.py -f mydevices_mapping.json snor1

EXIT CODES:
  0 : success
  1 : general application error
  2 : Request_UID command failure
  9 : unprovisioned device 

KNOWN BUGS:

TODOS:
1. add command-line switch and functionaliy that when call tests for and unsets 
   the Extended_CRTM_Status:auth_unlock_bit 

AUTHOR:
\tAlan Haffner

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter,
                      description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<device json mapping file>",
               default="devices.json", type=ap.FileType('r'),
               help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>",
               help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true',
               help="prints conditionals where var DEBUG is True.")

# version prints program version, then exits
p.add_argument('--version', action='version',
               version='%(prog)s {}'.format(VERSION),
               help="echo script version to stdout.")

p.add_argument("-c", "-clear", dest='clear', action='store_true', help="clear the Extended_CRTM_Status Register:Auth_Unlock bit.")

ARGS = p.parse_args()
DEBUG = ARGS.verbosity

def main() -> None:
    app_logging.setup_app_logging(DEBUG)
    main_logger = logging.getLogger(__name__)

    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    # (Note that when calling meta_request_id on the NaxosDevice
    # object signature verification on the response might fail
    # because the specifics of signing and byte order are
    # dependent on the revision).
    #

    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        if device.provision_sts_struct is None:
            err, *_ = device.meta_request_uid()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('request_uid: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(2)

            warn_on_response_verification_fail(err)

        print("\n\n")
        print("Device UID: {}".format(device.uid.hex()))
        print("Revision:   {}".format(device.revision.hex()))
        print("provision_sts_struct.Provisioned: {}".format(device.provision_sts_struct.Provisioned))

        if not device.provision_sts_struct.Provisioned:
            print("\n\n")
            print("Device is not provisioned:... Nothing left to do. Exiting...")
            main_logger.info("Device is not provisioned.")
            exit(9)

        err, Extended_CRTM_Status = device.read_extended_crtm_status()
        if not err.esr_success:
            main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
            exit(1)

        warn_on_response_verification_fail(err)

        # check to see if Auth_lock is already set in the Extended CRTM Status register
        auth_unlock_bit_status = bool(Extended_CRTM_Status.Auth_Unlock)

        if ARGS.clear:
            # clear Auth_Unlock bit
            if not auth_unlock_bit_status:
                print("\n\n")
                print("Auth_lock bit status is locked. Nothing left to do. Exiting...")
                main_logger.critical("Device Auth_lock bit status is already locked.")
                exit(1)

            ret = device.meta_ppb_lock_bit_write()
            if not ret:
                print("\n\n")
                main_logger.critical('meta_ppb_lock_bit_write failed')
                exit(1)

            err, Extended_CRTM_Status = device.read_extended_crtm_status()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

            auth_unlock_bit_status = bool(Extended_CRTM_Status.Auth_Unlock)

            if not auth_unlock_bit_status:
                print("\n\n")
                print("Extended_CRTM_Status:auth_unlock_bit:clear. Auth_lock state is: locked.")
                main_logger.info("Extended_CRTM_Status:auth_unlock_bit:clear. Auth_lock state is: locked.")
            else:
                print("\n\n")
                main_logger.critical("Failed to clear Auth_Unlock bit in Extended_CRTM_Status register.")
                exit(1)
        else:
            # set Auth_Unlock bit
            if auth_unlock_bit_status:
                print("\n\n")
                print("Auth_lock bit status is unlocked. Nothing left to do. Exiting...")
                main_logger.critical("Device Auth_lock bit status is already unlocked.")
                exit(1)

            err, _ = device.meta_auth_lock()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('meta_auth_lock: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            err, Extended_CRTM_Status = device.read_extended_crtm_status()
            if not err.esr_success:
                print("\n\n")
                main_logger.critical('read_extended_crtm_status: '+ err.esr_bit_msg + '; ' + err.error_msg)
                exit(1)

            warn_on_response_verification_fail(err)

            auth_unlock_bit_status = bool(Extended_CRTM_Status.Auth_Unlock)

            if auth_unlock_bit_status:
                print("\n\n")
                print("Extended_CRTM_Status:auth_unlock_bit:set. Auth_lock state is: unlocked.")
                main_logger.info("Extended_CRTM_Status:auth_unlock_bit:set. Auth_lock state is: unlocked.")
            else:
                print("\n\n")
                main_logger.critical("Failed to set Auth_Unlock bit in Extended_CRTM_Status register.")
                exit(1)

    exit(0)


if __name__ == "__main__":
    main()
