#! /usr/bin/env python3
import argparse as ap
import json
import logging
import sys
import os
from sys import exit

# --- Authenta imports ---
from authenta.maps import Envr
from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.naxos.device import NaxosDevice
from authenta.grpc_platform import GrpcPlatform
from authenta.naxos.utils import get_dev_info

import app_common_logging as app_logging

VERSION = 0.01
PROGRAM_DES = '''DESCRIPTION:
\t A sample application demonstrating how the Authenta Auth_Modify append
\t feature can be used to securely log information to memory.

\t This app will interact with the memory device alias currently listed in
\t devices.json.
'''

EPILOG = '''
EXAMPLE CMDLNS:
   app_smart_meter  <endpoint alias>
    -- as in --
   app_smart_meter snor1
    - where 'snor1' is a device entry in the devices.json file

   # for test cases where the default devices.json file isn't what is needed
   app_smart_meter -f mydevices_mapping.json snor1

EXIT CODES:
   0 : success
   1 : general error
   9 : provisioned device 

KNOWN BUGS:

TODOS:

AUTHOR:
\tMarshall Shapiro

COPYRIGHT: Micron Technology, Inc.; 2019

'''

p = ap.ArgumentParser(formatter_class=ap.RawDescriptionHelpFormatter, \
    description=PROGRAM_DES, epilog=EPILOG)

p.add_argument("-f", dest='ifn', metavar="<endpoint json mapping file>", default="devices.json", type=ap.FileType('r'), \
    help="JSON devices mapping file. Attributes include: 1) endpoint IP address; 2) endpoint OS device alias name.")

p.add_argument('device', metavar="<device alias>", help="endpoint-device alias found in devices.json file.")

p.add_argument("-v", "--verbosity", action='store_true', \
    help="REQUIRED: to printcommand's pass/fail message and exit status.")

# version prints program version, then exits
p.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION), \
    help="echo script version to stdout.")

ARGS = p.parse_args()

# addresses of registers controlling the Auth_Modify Append state of Naxos address space
# There are 256 bits (32 bytes) controlling the status of the 256 blocks of memeory
# on the chip. (Each block is 64K adding to the 16M of memory on the chip).
# the register read function is a 4-byte read.
Auth_Mod_Bitmap_1 = 0x180
Auth_Mod_Bitmap_2 = 0x190

# this address corresponds to the first bit of the second Auth_Mod_Bitmap register (0x190)
Smart_Meter_Log_Addr = 0xE30000
Log_Message_Len = 0x080 # 128 (bytes)

app_logging.setup_app_logging()
main_logger = logging.getLogger(__name__)

def main():
    envr = Envr(ARGS.device, ARGS.ifn)
    envr.setup()

    ip_params = envr.ep_ip_address.split(":")
    ip_addr = ip_params[0]

    if len(ip_params) > 1:
        port = ip_params[1]
        ep = GrpcPlatform(envr.dev_pth, ip_addr, port)
    else:
        ep = GrpcPlatform(envr.dev_pth, ip_addr)

    #
    # determine which revision of the Naxos device is connecting
    #
    device = NaxosDevice(ep)
    main_logger.info('fetching Naxos device revision')
    dev_rev = 0x000000
    dev_rev_bytes = device.get_device_revision()
    if dev_rev_bytes is None:
        print("\n\n")
        print("Failed to retrieve revision value from device; Exiting.")
        main_logger.critical('Failed to retrieve revision value from device')
        exit(1)
        
    dev_rev = int.from_bytes(device.revision, byteorder='big')
    main_logger.info('Found attached Naxos device revision {}'.format(dev_rev))
    if dev_rev == 0x010005:
        device = NaxosDevice_1_5(ep)
    else:
        main_logger.critical("Unrecognized Naxos revision {}!".format(dev_rev))
        exit(1)

    with device:
        # get device data
        ret = get_dev_info(device)
        if ret is not None:
            (_, _, _, provision_sts) = ret
        else:
            print("\n\n")
            main_logger.critical("Failed to retrieve device data")
            return False

        if  provision_sts.Provisioned != 1:
            print("\n\n")
            print("Device is not provisioned. Nothing to do. Exiting...")
            main_logger.critical("Authenta device is not provisioned.")           
            return False

        '''
         1 - configure a Naxos device in append mode for secure logs
        '''

        # set the log address to Auth_Modify mode (Option 0)
        ret = setAMBit(device, True)
        if not ret:
            print("\n\n")
            print("Failed to set AMBitmap register. Nothing to do. Exiting...")
            main_logger.critical("Failed to set AMBitmap register")
            return False

        # Auth_Modify(option #0): erase AM block to be configured as an AP block
        config = b"\x00\x10"
        err, _ = device.meta_auth_modify(config, Smart_Meter_Log_Addr, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to call Auth_Modify to setup/erase AP block 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        print("Called Auth_Modify (option #0) to setup/erase AP block 0x%08X" %(Smart_Meter_Log_Addr))

        '''
         2 - write logs with no knowledge of SRK emulating sensor data being written in append mode blocks
        '''

        # put log data into KTR register for transfer to append-only log 
        ret = log_to_KTR(device, b"[Log Entry 1] Test Data\n")
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to write log message to KTR register")
            return False
        main_logger.info("Wrote log message to KTR register")

        # use auth_modify option #2 to append data to the append log
        config = b"\x00\x02"
        err, _ = device.meta_auth_modify(config, 0x0, Smart_Meter_Log_Addr, Log_Message_Len)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to call Auth_Modify (option #2) to append log message to block 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        main_logger.info("Called Auth_Modify (option #2) to append log message to block 0x%08X" %(Smart_Meter_Log_Addr))

        # read the log that was just written
        ret, data = device.meta_fast_read(Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to read log from log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False
        main_logger.info("[LOG  0x{:08X}] {}... ".format(Smart_Meter_Log_Addr, data[:32]))
        print(data[:32])

        '''
         3 - attempt to modify said logs and show they are not modified
        '''
        
        # unlock the data block where log was written
        ret = device.meta_dyb_write(0, Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to set volatile lock bit to UNLOCKED (0) at log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False

        # write data to the log (standard write, not append and no signature)
        ret = device.meta_page_program(Smart_Meter_Log_Addr, bytes("THIS IS NOT A LOG MESSAGE! LOG HAS BEEN OVERWRITTEN!", "UTF-8"))
        if not ret:
            print("\n\n")
            main_logger.critical("page-program (write to flash) function failed at log address 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("Note, function should not fail even if the write is disallowed. Aborting.")
            return False
        print("Attempted to overwrite data in the append log")

        # lock the data block where log was written
        ret = device.meta_dyb_write(1, Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to set volatile lock bit to Locked (1) at log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False

        '''
         4 - remote user reads logs from chip and prints them to screen
        '''

        # read the log again to see if it has been overwritten - it should not have been
        ret, data = device.meta_fast_read(Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to read log from log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False
        main_logger.info("[LOG  0x{:08X}] {}... ".format(Smart_Meter_Log_Addr, data[:32]))
        print(data[:32])

        '''
         5 - remote user uses SRK to restore append mode block
        '''

        # Run Auth_Modify(option #0): again to erase the data in the append block (the log)
        config = b"\x00\x10"
        err, _ = device.meta_auth_modify(config, Smart_Meter_Log_Addr, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to call Auth_Modify to setup/erase AP block 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        main_logger.info("Called Auth_Modify (option #0) to setup/erase AP block 0x%08X" %(Smart_Meter_Log_Addr))

        '''
         6 - repeat steps 2 to 4 and show the sensor data has changed
        '''

        # Now write new log data (with no knowledge of SRK) to show the log be appened to again

        # put log data into KTR register for transfer to append-only log
        ret = log_to_KTR(device, b"[NEW Log Entry] Some other test data\n")
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to write log message to KTR register")
            return False
        print("Wrote log message to KTR register")

        # use auth_modify option #2 to append data to the append log
        config = b"\x00\x02"
        err, _ = device.meta_auth_modify(config, 0x0, Smart_Meter_Log_Addr, Log_Message_Len)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to call Auth_Modify (option #2) to append log message to block 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        main_logger.info("Called Auth_Modify (option #2) to append log message to block 0x%08X" %(Smart_Meter_Log_Addr))

        # read the new log that was just written - should read "[NEW Log..."
        ret, data = device.meta_fast_read(Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to read log from log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False
        main_logger.info("[LOG  0x{:08X}] {}... ".format(Smart_Meter_Log_Addr, data[:32]))
        print(data[:32])

        # attempt to modify the logs again and show they are not modified by standard page program
        
        # unlock the data block where log was written
        ret = device.meta_dyb_write(0, Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to set volatile lock bit to UNLOCKED (0) at log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False

        # write data to the log (standard write, not append and no signature)
        ret = device.meta_page_program(Smart_Meter_Log_Addr, bytes("THIS IS NOT A LOG MESSAGE! LOG HAS BEEN OVERWRITTEN!", "UTF-8"))
        if not ret:
            print("\n\n")
            main_logger.critical("page-program (write to flash) function failed at log address 0x%08X" %(Smart_Meter_Log_Addr))
            main_logger.critical("Note, function should not fail even if the write is disallowed. Aborting.")
            return False
        print("Attempted to overwrite data in the append log")

        # re-lock the data block where log was written
        ret = device.meta_dyb_write(1, Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to set volatile lock bit to Locked (1) at log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False

        # read the log again to if it has been overwritten - it should not have been
        ret, data = device.meta_fast_read(Smart_Meter_Log_Addr)
        if not ret:
            print("\n\n")
            main_logger.critical("Failed to read log from log address 0x%08X" %(Smart_Meter_Log_Addr))
            return False
        main_logger.info("[LOG  0x{:08X}] {}... ".format(Smart_Meter_Log_Addr, data[:32]))
        print(data[:32])

        # Auth_Modify(option #0): erase AP block to be configured as an AM block
        config = b"\x00\x00"
        err, _ = device.meta_auth_modify(config, Smart_Meter_Log_Addr, 0x0, 0x0)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("Failed to call Auth_Modify to erase AP block 0x%08X and configure it as AM block" %(Smart_Meter_Log_Addr))
            main_logger.critical("auth_modify: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        main_logger.info("Called Auth_Modify (option #0) to erase AP block 0x%08X and configure it as AM block" %(Smart_Meter_Log_Addr))

        # set the log address to legacy block
        ret = setAMBit(device, False)
        if not ret:
            print("\n\n")
            print("Failed to set AMBitmap register. Exiting on error.")
            main_logger.critical("Failed to set AMBitmap register")
            return False

        return True

def setAMBit(device, set_to_auth_mod):
    """
    set the AMBitmap bit designating an address range to be in Auth_Modify Append mode
    see Naxos IAS Mediterrabeo addendum section 3.3.16 AMBitmap
    """
    Smart_Meter_Log_Bit_Mask_am_on  = 0xFFFFFFF7 # = 1111 1111 1111 1111 1111 1111 1111 0111
    Smart_Meter_Log_Bit_Mask_am_off = 0x00000008 # = 0000 0000 0000 0000 0000 0000 0000 1000

    err, response = device.meta_crtm_register_read(b"\x00", Auth_Mod_Bitmap_2)
    if not err.esr_success:
        print("\n\n")
        main_logger.critical("crtm_register_read: " + err.esr_bit_msg + "; " + err.error_msg)
        return False
    ambitmap_data_read = response[28:32]

    # change bit requested value, 0 for AM on, 1 for off
    if set_to_auth_mod:
        ambitmap_data_write = int(ambitmap_data_read.hex(), 16) & Smart_Meter_Log_Bit_Mask_am_on
    else:
        ambitmap_data_write = int(ambitmap_data_read.hex(), 16) | Smart_Meter_Log_Bit_Mask_am_off

    ambitmap_data_write = ambitmap_data_write.to_bytes(4, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    if not err.esr_success:
        print("\n\n")
        main_logger.critical("acrtm_update_start: " + err.esr_bit_msg + "; " + err.error_msg)
        return False

    err, _ = device.meta_crtm_register_write(b"\x00", Auth_Mod_Bitmap_2 + 0xC, ambitmap_data_write, True)
    if not err.esr_success:
        print("\n\n")
        main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
        device.meta_acrtm_update_finish(b"\x00")
        return False

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    if not err.esr_success:
        print("\n\n")
        main_logger.critical("acrtm_update_finish: " + err.esr_bit_msg + "; " + err.error_msg)
        return False

    return True

def log_to_KTR(device, data):
    """
    Write data starting at the beginning of the KTR register
    data input must be a string formatted as a byte array
    """
    KTR_Base = 0xB0
    CUST_RAM_Base = 0x300
    data_len = len(data)

    # make the data 128 characters long with a new-line followed by at
    # least one null character at the end
    extra_nul_chars = 1 if data_len > 127 else 128 - data_len

    data = data[:127] + (b'\0' * extra_nul_chars)
    data_len = len(data)

    # the data has to be written in 4-byte blocks. Append 0s if necessary.

    extra_bytes = data_len % 4
    if extra_bytes != 0:
        data = b''.join((data, bytes('\0', "utf-8") * (4-extra_bytes)))
        data_len = len(data)

    # KTR is 32 bytes, write the rest to cust ram
    j = 0
    for i in range(0, 32, 4):
        # CRTM_Register_Write
        data_to_write = data[i:(i + 4)]
        data_to_write = data_to_write[::-1]
        addr_to_write = KTR_Base + j
        err, _ = device.meta_crtm_register_write(b"\x00", addr_to_write, data_to_write, False)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        j += 4

    # Customer RAM is 96 bytes, write the rest to KTR
    j = 0
    for i in range(32, 128, 4):
        # CRTM_Register_Write
        data_to_write = data[i:(i + 4)]
        data_to_write = data_to_write[::-1]
        addr_to_write = CUST_RAM_Base + j
        err, _ = device.meta_crtm_register_write(b"\x00", addr_to_write, data_to_write, False)
        if not err.esr_success:
            print("\n\n")
            main_logger.critical("crtm_register_write: " + err.esr_bit_msg + "; " + err.error_msg)
            return False
        j += 4

    return True

# program entry point
if __name__ == "__main__":
    # call main function
    # sys.exit() acts as a thin wrapper -
    # kind of global exception handler
    exit(not main())

# <EOF/>
