#!/usr/bin/env python3


import os, sys, time
import socket

VERSION = 0.751

# ----------------------------------------------------------
# File: iboot.py
# Author: Alan Haffner
# Date: 2018-10-5
# ----------------------------------------------------------
#  Usage: iboot [...] <iboot ip address> <AC power state> 
#
#  [...] - anything before the last two commandline arguments is a don't care
#          - allows tool to be compatible /w Wemo commandline interface
#  "ip address"  -  
#     - always second to last argument on commandline
#     - ipv4 only
#  "AC power states"  - 
#     - always last argument on commandline
#      <0|off> = off
#      <1|on> = on
#      cycle  = cycle from one state to other and back again
#      status = ask iboot what power state it is in
# ----------------------------------------------------------
#  configuration details
# ----------------------------------------------------------
NIC_VARS        = {'port':80, 'timeout':3, 'passwd':"PASS"} 
AC_SWITCH_VARS  = {'gap_time':1, 'pwr_off_delay':3,} 
# ----------------------------------------------------------
#```

class iBoot_IF(object):

  def __init__(self):
    # locals
    self._escape   = chr(0x1b)
    self._c_rtn    = chr(0x0d)
    self._passwd   = NIC_VARS['passwd']
    self._header   = ("{}{}{}".format(self._escape, self._passwd, self._escape))

  def _formatted_cmd(self, cmd_):
    # encde cmd into an immutable byte string.  
    return  bytes("{}{}{}".format(self._header, cmd_, self._c_rtn), 'ascii')

  @property
  def off(self):    
    cmd_ = 'f'
    return self._formatted_cmd(cmd_)

  @property
  def on(self):    
    cmd_ = 'n'
    return self._formatted_cmd(cmd_)

  @property
  def cycle(self):    
    cmd_ = 'c'
    return self._formatted_cmd(cmd_)

  @property
  def status(self):    
    # status = 'q'
    cmd_ = 'q'
    return self._formatted_cmd(cmd_)
#```

class TCP_handler(object):

  def __init__(self, ip_addr_tpl=None):

    # connect to dataport TCP server
    self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self._socket.settimeout(NIC_VARS['timeout'])
    self.ip_addr_tpl = (ip_addr_tpl, NIC_VARS['port'])

  def __enter__(self):

    try:
      self._socket.connect(self.ip_addr_tpl)
      print("Establish endpoint connection to iBoot at - addr:{}, port:{}".format(*self.ip_addr_tpl))

    except Exception:
      print("\n\n\n")
      print("error:TCP Connection to iBoot Failed at - addr:{}, port:{}".format(*self.ip_addr_tpl))
      print("\n\n\n")
      return False

    # give the host a little time to complete Socket setup    
    time.sleep(AC_SWITCH_VARS['gap_time']/2)

    return self

  def __exit__(self, *discard):
    # close TCP connection to dataprobe server
    self._socket.close()
    return True

  def send_cmd(self, command_):
    rc = True
    read_buff = None
    cmd_, direction_ = command_

    # send the commmand 
    try:
      self._socket.sendall(cmd_)

    except Exception:
      print("error:Socket.send Cmd Failure")
      rc = False

    if rc:
      # give the device a bit of time to process the cmd
      time.sleep(AC_SWITCH_VARS['gap_time'])

      if(direction_ == 'r'):
        read_buff = self._socket.recv(8)

    return read_buff, rc
#```

def token2cmd(request):
  rc     = True
  cmd_   = None
  switch = iBoot_IF()

  try: 
    cmd_ = {
      'on'     : (switch.on,     'w'),
      '1'      : (switch.on,     'w'),
      'off'    : (switch.off,    'w'),
      '0'      : (switch.off,    'w'),
      'cycle'  : (switch.cycle,  'w'),
      'status' : (switch.status, 'r'),
    }[request.lower()]

  except KeyError as err: 
    print("error:request key not found:{0}".format(err))
    rc = False

  return cmd_, rc
#```

def run(ip_addr, request):
  rc        = True
  exit_code = 0
  cmd_      = None
  read_buf  = None
  pwr_off_delay = AC_SWITCH_VARS['pwr_off_delay']

  if ('off' in request) or ('0' in request):
    time.sleep(pwr_off_delay)

  if rc:
    cmd_, rc =  token2cmd(request)

  if rc:
    with TCP_handler(ip_addr) as conn:
  
      read_buf, rc = conn.send_cmd(cmd_)

  if rc:
    if read_buf:
      print(read_buf.decode('utf-8'))
     
  if not rc:
    exit_code = -1
    
  return exit_code

def main():
  # assign all but last two argv tokens to the blank identifier -
  # as only the last two tokens in it are useful
  *_, ip_addr, request = sys.argv

  return run(ip_addr, request)
   
### program entry point  
if __name__ == "__main__":
   sys.exit(main()) 

#<EOF/>
