#! /usr/bin/env python3

"""
Test setup.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import json
import time

import pytest
from authenta.indigo.device import IndigoDevice
from authenta.grpc_platform import GrpcPlatform

device_json_file_name = 'devices.json'
ep_alias = 'snor1'

def get_dev_info_from_json(filename, alias):
    """
    Get device info from device json file.
    return (device path, endpoint IP address, endpoint port number, iboot IP address)
    """
    try:
        with open(filename) as f:
            json_data = json.load(f)

        json_data = json_data[alias]

        ip_params = json_data['endpoint_ip'].split(':')
        ep_ip = ip_params[0]
        ep_port = None
        if len(ip_params) > 1:
            ep_port = ip_params[1]

        dev_path = json_data['endpoint_fn']

        iboot_ip = json_data.get('iboot_ip', '')
        iboot_ip = iboot_ip if iboot_ip != '' else None

    except FileNotFoundError:
        print('Error: Failed to open file:{}'.format(filename))
        sys.exit(1)
    except json.decoder.JSONDecodeError:
        print('Error: Failed to parse json file:{}'.format(filename))
        sys.exit(1)
    except:
        print('Error: Failed to get device info from json file:{}'.format(filename))
        sys.exit(1)

    return dev_path, ep_ip, ep_port, iboot_ip

@pytest.fixture(scope = 'session', autouse = True)
def run_setup(request):
    """
    Run test setup before all tests
    """
    dev_path, ep_ip, ep_port, iboot_ip = get_dev_info_from_json(
        device_json_file_name, ep_alias)

    if ep_port is None:
        ep = GrpcPlatform(dev_path, ep_ip)
    else:
        ep = GrpcPlatform(dev_path, ep_ip, ep_port)

    # save shared data in pytest cache
    request.config.cache.set('dev_path', dev_path)
    request.config.cache.set('ep_ip', ep_ip)
    request.config.cache.set('ep_port', ep_port)
    request.config.cache.set('iboot_ip', iboot_ip)

def get_device(request):
    """
    Create IndigoDevice object from pytest cache data
    """
    dev_path = request.config.cache.get("dev_path", None)
    ep_ip = request.config.cache.get("ep_ip", None)
    ep_port = request.config.cache.get("ep_port", None)
    
    if ep_port is None:
        ep = GrpcPlatform(dev_path, ep_ip)
    else:
        ep = GrpcPlatform(dev_path, ep_ip, ep_port)

    device = IndigoDevice(ep)

    return device