import sys
import os
import conftest
import pytest

from authenta.indigo import utils
from authenta.indigo.utils import CRTM_Registers

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_acrtm_update_start_setup(request):
    global device
    device = conftest.get_device(request)

    with device:
        device.meta_request_uid()
        if device.provision_sts_struct.Provisioned and \
            not device.provision_sts_struct.Remotely_Provisioned:
            rc, _ = device.meta_crtm_register_read(b'\00\00', 0x00)
            if rc.valid_response:
                if device.device_config_struct is not None:
                    if device.provision_sts_struct.HSM_Present:
                        if not device.device_config_struct.CRK_Read_Disable:
                            device.set_CRK_Read_Disable_bit(True)
                    else:
                        if not device.device_config_volatile_struct.CSK_Read_Disable:
                            device.set_CSK_Read_Disable_bit(True)

    yield

def test_acrtm_update_start_option0_ecdsa_positive():
    """
    SigType: ECDSA signature
    Option #0
    """
    with device:
        error_code, response_bytes = device.meta_acrtm_update_start(b"\x80\x00")
        assert(error_code.esr_and_response_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b"\x80\x00")
        assert(err.esr_and_response_success)

def test_acrtm_update_start_option0_hmac_positive():
    """
    SigType: HMAC signature
    Option #0
    """
    with device:
        error_code, response_bytes = device.meta_acrtm_update_start(b'\x00\x00')
        assert(error_code.esr_and_response_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b'\x00\x00')
        assert(err.esr_and_response_success)

def test_acrtm_update_start_option1_ecdsa_positive():
    """
    SigType: ECDSA signature
    Option #1
    """
    with device:
        ret = utils.enable_automeasure(device)
        assert(ret)
        ret = utils.enable_recovery_partition(device)
        assert(ret)

        error_code, response_bytes = device.meta_acrtm_update_start(b'\x80\x01')
        assert(error_code.esr_and_response_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b'\x80\x01')
        assert(err.esr_and_response_success)

        ret = utils.disable_automeasure_and_recovery_partition(device)
        assert(ret)

def test_acrtm_update_start_option1_hmac_positive():
    """
    SigType: HMAC signature
    Option #1
    """
    with device:
        ret = utils.enable_automeasure(device)
        assert(ret)
        ret = utils.enable_recovery_partition(device)
        assert(ret)

        error_code, response_bytes = device.meta_acrtm_update_start(b'\x00\x01')
        assert(error_code.esr_and_response_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b'\x00\x01')
        assert(err.esr_and_response_success)

        ret = utils.disable_automeasure_and_recovery_partition(device)
        assert(ret)

@pytest.mark.parametrize("config, expectation", [(0x10, pytest.raises(TypeError)), (b"\x80\x00", does_not_raise())])
def test_meta_acrtm_update_start_config_type(config, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_acrtm_update_start(config)
            assert(error_code.esr_and_response_success)

            error_code, _ = device.meta_acrtm_update_finish(config)
            assert(error_code.esr_and_response_success)
