import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode
from authenta import authenta_key_mgmt_interface

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_apa_modify_setup(request):
    global device
    global wpg_addr
    global key_slot
    global apa_data
    device = conftest.get_device(request)
    wpg_addr = 0xf00000
    key_slot = 0x3
    apa_data = b'\x00'*80

    with device:
        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # get ECDSA_KTS_PuK_key
        key_name = "ecdsa_pub_slot_{}".format(key_slot)
        ECDSA_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDSA_KTS_PuK_key)

        # replace slot public ECDSA public key
        err, _ = device.meta_manage_kts(b"\x80\x48", key_slot, ECDSA_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # get ECDH_KTS_PuK_key
        key_name = "ecdh_pub_slot_{}".format(key_slot)
        ECDH_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDH_KTS_PuK_key)

        # replace slot public ECDH key and recalculate shared secret
        err, _ = device.meta_manage_kts(b"\x80\x44", key_slot, ECDH_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

    yield

def test_apa_modify_ECDSA_sig_positive():
    with device:

        error_code, response = device.meta_apa_modify(b'\x80\x00', wpg_addr, apa_data[::-1], key_slot)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_apa_modify_HMAC_sig_positive():
    with device:

        error_code, response = device.meta_apa_modify(b'\x00\x00', wpg_addr, apa_data[::-1], key_slot)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)      

def test_apa_modify_ECDSA_sign_with_kts_slot_key():
    with device:

        apa_data = b'\x00'*80

        err, _ = device.meta_apa_modify(b'\x80\x04', wpg_addr, apa_data[::-1], 0x3)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # enable modify APA data by slot key (EnableKTSKey = 1b and kts_slot_3 = 0)
        authentaAccessRightsData = b'\x02\x00'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6

        # APA_Modify using server key
        error_code, _ = device.meta_apa_modify(b'\x80\x04', wpg_addr, apa_data[::-1], 0x3)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # APAWriteProtect = '1b
        # APAReadProtect = '1b
        authentaAccessRightsData = b'\xC0\x00'
        apa_data = b'\x00'*72 + authentaAccessRightsData[::-1] + b'\x00'*6

        # APA_Modify using kts slot key for signing
        error_code, _ = device.meta_apa_modify(b'\xC0\x04', wpg_addr, apa_data[::-1], 0x3)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

@pytest.mark.parametrize("config, expectation", [(0x0, pytest.raises(TypeError)), (b"\x80\x00", does_not_raise())])
def test_meta_apa_modify_config_type(config, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_apa_modify(config, wpg_addr, apa_data[::-1], key_slot)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("wpg_addr, expectation", [(b"\x00\x00", pytest.raises(TypeError)), (0x0, does_not_raise())])
def test_meta_apa_modify_wpg_addr_type(wpg_addr, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_apa_modify(b'\x80\x00', wpg_addr, apa_data[::-1], key_slot)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("apa_data, expectation", [(0x0, pytest.raises(TypeError)), (b"\x00"*80, does_not_raise())])
def test_meta_apa_modify_apa_data_type(apa_data, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_apa_modify(b'\x80\x00', wpg_addr, apa_data[::-1], key_slot)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
