import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode
from authenta import authenta_key_mgmt_interface

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_create_verify_signature_setup(request):
    global device
    global byte_addr
    global partitionMsgAddr
    device = conftest.get_device(request)
    byte_addr = 0x3
    partitionMsgAddr = 0xf10000

    with device:
        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # get ECDSA_KTS_PuK_key
        key_name = "ecdsa_pub_slot_{}".format(byte_addr)
        ECDSA_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDSA_KTS_PuK_key)

        # replace slot public ECDSA public key
        err, _ = device.meta_manage_kts(b"\x80\x48", byte_addr, ECDSA_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # get ECDSA_KTS_Prv_key
        key_name = "ecdsa_prv_slot_{}".format(byte_addr)
        ECDSA_KTS_Prv_key = authenta_key_mgmt_interface.export_private_key_in_the_clear(device.uid.hex(), key_name)
        assert(ECDSA_KTS_Prv_key)

        # replace slot slave ECDSA private key
        err, _ = device.meta_manage_kts(b"\x80\x40", byte_addr, ECDSA_KTS_Prv_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # get ECDH_KTS_PuK_key
        key_name = "ecdh_pub_slot_{}".format(byte_addr)
        ECDH_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDH_KTS_PuK_key)

        # replace slot public ECDH key and recalculate shared secret
        err, _ = device.meta_manage_kts(b"\x80\x44", byte_addr, ECDH_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        if device.provision_sts_struct.Provisioned and \
            not device.provision_sts_struct.Remotely_Provisioned:
            rc, _ = device.meta_crtm_register_read(b'\00\00', 0x00)
            if rc.valid_response:
                if device.device_config_struct is not None:
                    if device.provision_sts_struct.HSM_Present:
                        if not device.device_config_struct.CRK_Read_Disable:
                            device.set_CRK_Read_Disable_bit(True)
                    else:
                        if not device.device_config_volatile_struct.CSK_Read_Disable:
                            device.set_CSK_Read_Disable_bit(True)

    yield

def test_create_verify_signature_ecdsa_sign_input():
    with device:

        # use active host server public or session key for signatures
        # sign input digest content 
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_hmac_sign_input():
    with device:

        # use active host server public or session key for signatures
        # sign input digest content 
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x01\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x01\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_ecdsa_kts_sign_input():
    with device:

        # use KTS slot key for signatures
        # sign input digest content 
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\xC1\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\xC1\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_hmac_kts_sign_input():
    with device:

        # use KTS slot key for signatures
        # sign input digest content 
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x41\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x41\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)


def test_create_verify_signature_ecdsa_sign_partition_content():
    with device:

        # use active host server public or session key for signatures
        # sign the first 32 bytes of partition content
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x83\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x83\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_hmac_sign_partition_content():
    with device:

        # use active host server public or session key for signatures
        # sign the first 32 bytes of partition content
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x03\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x03\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_ecdsa_kts_sign_partition_content():
    with device:

        # use KTS slot key for signatures
        # sign the first 32 bytes of partition content
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\xC3\x10", byte_addr, 0x0, 0x0, partitionMsgAddr, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\xC3\x10", byte_addr, 0x0, 0x0, partitionMsgAddr, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_create_verify_signature_hmac_kts_sign_partition_content():
    with device:

        # use KTS slot key for signatures
        # sign the first 32 bytes of partition content
        # use KTS slave private key as the private signing key
        err, response = device.meta_create_signature(b"\x43\x10", byte_addr, 0x0, 0x0, partitionMsgAddr, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Create Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        err, _ = device.meta_verify_signature(b"\x43\x10", byte_addr, 0x0, 0x0, partitionMsgAddr, b"\x00"*32)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Verify Signature Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

@pytest.mark.parametrize("config, expectation", [(0x10, pytest.raises(TypeError)), (b"\x81\x10", does_not_raise())])
def test_meta_create_signature_config_type(config, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(config, byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)

@pytest.mark.parametrize("byte_addr, expectation", [(b"\x00", pytest.raises(TypeError)), (0x3, does_not_raise())])
def test_meta_create_signature_byte_addr_type(byte_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_dest_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0x0, does_not_raise())])
def test_meta_create_signature_partition_src_addr_type(partition_dest_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, partition_dest_addr, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_key_src_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0x0, does_not_raise())])
def test_meta_create_signature_partition_key_src_addr_type(partition_key_src_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, partition_key_src_addr, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_msg_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0xf10000, does_not_raise())])
def test_meta_create_signature_partition_msg_addr_type(partition_msg_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, 0x0, partition_msg_addr, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("input_digest, expectation",[(0x0, pytest.raises(TypeError)),(b"\x00"*32, does_not_raise())])
def test_meta_create_signature_input_digest_type(input_digest, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, input_digest)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("input_digest, expectation",[(b"\x00", pytest.raises(ValueError)),(b"\x00"*32, does_not_raise())])
def test_meta_create_signature_input_digest_size(input_digest, expectation):
    with device:
        with expectation:
            err, _ = device.meta_create_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, input_digest)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("config, expectation", [(0x10, pytest.raises(TypeError)), (b"\x81\x10", does_not_raise())])
def test_meta_verify_signature_config_type(config, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(config, byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("byte_addr, expectation", [(b"\x00", pytest.raises(TypeError)), (0x6, does_not_raise())])
def test_meta_verify_signature_byte_addr_type(byte_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_src_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0x0, does_not_raise())])
def test_meta_verify_signature_partition_src_addr_type(partition_src_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, partition_src_addr, 0x0, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_key_src_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0x0, does_not_raise())])
def test_meta_verify_signature_partition_key_src_addr_type(partition_key_src_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, partition_key_src_addr, 0x0, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("partition_msg_addr, expectation",[(b"\x00", pytest.raises(TypeError)),(0x0, does_not_raise())])
def test_meta_verify_signature_partition_msg_addr_type(partition_msg_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, 0x0, partition_msg_addr, b"\x00"*32)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("input_digest, expectation",[(0x0, pytest.raises(TypeError)),(b"\x00"*32, does_not_raise())])
def test_meta_verify_signature_input_digest_type(input_digest, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, input_digest)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("input_digest, expectation",[(b"\x00", pytest.raises(ValueError)),(b"\x00"*32, does_not_raise())])
def test_meta_verify_signature_input_digest_size(input_digest, expectation):
    with device:
        with expectation:
            err, _ = device.meta_verify_signature(b"\x81\x10", byte_addr, 0x0, 0x0, 0x0, input_digest)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)
