import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_pcr_extend_setup(request):
    global device
    global provision_stats
    device = conftest.get_device(request)
    yield

def test_crtm_register_write_positive_nonvolatile_ecdsa():
    """
    SigType: ECDSA signature
    Nonvolatile write
    """
    with device:

        data =b'\x00\x00\x00\x02'
        mask = b'\x00'*4
        err, _ = device.meta_acrtm_update_start(b'\x80\x00')
        assert(err.esr_and_response_success)

        error_code, response_bytes = device.meta_crtm_register_write(b'\x80\x00', 0x20, data, mask)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b'\x80\x00')
        assert(err.esr_and_response_success)


def test_crtm_register_write_positive_nonvolatile_hmac():
    """
    SigType: HMAC signature
    Nonvolatile write
    """
    with device:
        data = b'\x00\x00\x00\x01'
        mask = b'\x00'*4
        err, _ = device.meta_acrtm_update_start(b'\x00\x00')
        assert(err.esr_and_response_success)

        error_code, response_bytes = device.meta_crtm_register_write(b'\x00\x00', 0x20, data, mask)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        err, _ = device.meta_acrtm_update_finish(b'\x00\x00')
        assert(err.esr_and_response_success)


def test_crtm_register_write_volatile_ecdsa_positive():
    """
    SigType: ECDSA signature
    Volatile write
    """
    with device:
        error_code, response_bytes = device.meta_crtm_register_write(b'\x80\x01', 0x34, b'\x00\x00\x00\x01', b'\x00'*4)

        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        # Rewrite the register so the test will run correctly next time
        err, _ = device.meta_crtm_register_write(b'\x80\x01', 0x34, b'\x00\x00\x00\x00', b'\x00'*4)
        assert(err.esr_and_response_success)

def test_crtm_register_write_volatile_hmac_positive():
    """
    SigType: HMAC signature
    Volatile write
    """
    with device:
        error_code, response_bytes = device.meta_crtm_register_write(b'\x00\x01', 0x34, b'\x00\x00\x00\x02', b'\x00'*4)

        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response_bytes, bytes)

        # Rewrite the register so the test will run correctly next time
        err, _ = device.meta_crtm_register_write(b'\x00\x01', 0x34, b'\x00\x00\x00\x03', b'\x00'*4)
        assert(err.esr_and_response_success)

@pytest.mark.parametrize("config, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00", does_not_raise())])
def test_meta_crtm_register_write_config_type(config, expectation):
    with device:
        with expectation:
            device.meta_acrtm_update_start(b'\x00\x00')
            error_code, _ = device.meta_crtm_register_write(config,
                                            crtm_addr = 0x0,
                                            data = b'\x00'*88+b'\x00\x00\x00\x02'+b'\x00'*36,
                                            mask = b'\x00'*128)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
            device.meta_acrtm_update_finish(b'\x80\x00')

@pytest.mark.parametrize("crtm_addr, expectation", [(b"\x00\x00", pytest.raises(TypeError)), (0x15, does_not_raise())])
def test_meta_crtm_register_write_crtm_addr_type(crtm_addr, expectation):
    with device:
        with expectation:
            device.meta_acrtm_update_start(b'\x00\x00')
            error_code, _ = device.meta_crtm_register_write(config = b"\x00\x00",
                                            crtm_addr = crtm_addr,
                                            data = b'\x00'*88+b'\x00\x00\x00\x02'+b'\x00'*36,
                                            mask = b'\x00'*128)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
            device.meta_acrtm_update_finish(b'\x80\x00')

@pytest.mark.parametrize("data, expectation", [(43, pytest.raises(TypeError)), (b'\x00'*88+b'\x00\x00\x00\x02'+b'\x00'*36, does_not_raise())])
def test_meta_crtm_register_write_data_type(data, expectation):
    with device:
        with expectation:
            device.meta_acrtm_update_start(b'\x00\x00')
            error_code, _ = device.meta_crtm_register_write(config = b"\x00\x00",
                                            crtm_addr = 0x0,
                                            data = data,
                                            mask = b'\x00'*128)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
            device.meta_acrtm_update_finish(b'\x80\x00')

@pytest.mark.parametrize("mask, expectation", [(43, pytest.raises(TypeError)), (b'\x00'*128, does_not_raise())])
def test_meta_crtm_register_write_mask_type(mask, expectation):
    with device:
        with expectation:
            device.meta_acrtm_update_start(b'\x00\x00')
            error_code, _ = device.meta_crtm_register_write(config = b"\x00\x00",
                                            crtm_addr = 0x0,
                                            data = b'\x00'*88+b'\x00\x00\x00\x02'+b'\x00'*36,
                                            mask = mask)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
            device.meta_acrtm_update_finish(b'\x80\x00')
