import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_deprovision_crtm_setup(request):
    global device
    device = conftest.get_device(request)
    yield

def test_deprovision_crtm_HMAC_signature_positive():
    with device:

        error_code, _ = device.meta_request_uid()
        assert(error_code.esr_and_response_success)

        if not device.provision_sts_struct.Provisioned:
            error_code, response = device.meta_provision_crtm(b'\x80\x05')
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

            # Assert Return Types
            assert isinstance(error_code, IndigoCmdErrorCode)
            assert isinstance(response, bytes)
     
        # Deprovision Device
        error_code, response = device.meta_deprovision_crtm(b'\x00\x00')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        # Provision Device
        error_code, response = device.meta_provision_crtm(b'\x80\x05')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)


def test_deprovision_crtm_ECDSA_signature_positive():
    with device:

        error_code, _ = device.meta_request_uid()
        assert(error_code.esr_and_response_success)

        if not device.provision_sts_struct.Provisioned:
            error_code, response = device.meta_provision_crtm(b'\x80\x05')
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

            # Assert Return Types
            assert isinstance(error_code, IndigoCmdErrorCode)
            assert isinstance(response, bytes)

        # Deprovision Device
        error_code, response = device.meta_deprovision_crtm(b'\x80\x00')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        # Provision Device
        error_code, response = device.meta_provision_crtm(b'\x80\x05')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

def test_deprovision_crtm_ECDSA_signature_allow_srk():
    with device:

        error_code, _ = device.meta_request_uid()
        assert(error_code.esr_and_response_success)
      
        if not device.provision_sts_struct.Server_Deprovision:

            if device.provision_sts_struct.Provisioned:
                error_code, response = device.meta_deprovision_crtm(b'\x80\x00')
                assert(error_code.esr_and_response_success)
                assert(error_code._sig_matched)

            # Locally provision device
            # Set Server_Deprovision
            error_code, response = device.meta_provision_crtm(b'\x80\x16')
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

            # Assert Return Types
            assert isinstance(error_code, IndigoCmdErrorCode)
            assert isinstance(response, bytes)
     
        # Deprovision Device, allow_srk
        error_code, response = device.meta_deprovision_crtm(b'\x80\x02')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        # Provision Device
        error_code, response = device.meta_provision_crtm(b'\x80\x05')
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

@pytest.mark.parametrize("config, expectation", [(0x0, pytest.raises(TypeError)), (b"\x80\x00", does_not_raise())])
def test_deprovision_crtm_config_type(config, expectation):
    with device:
        with expectation:
            err, _ = device.meta_request_uid()
            assert(err.esr_and_response_success)

            if not device.provision_sts_struct.Provisioned:
                error_code, response = device.meta_provision_crtm(b'\x80\x05')
                assert(error_code.esr_and_response_success)
                assert(error_code._sig_matched)

            # Deprovision Device
            error_code, _ = device.meta_deprovision_crtm(config)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

            # Provision Device
            error_code, response = device.meta_provision_crtm(b'\x80\x05')
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
