import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode
from authenta import authenta_key_mgmt_interface

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_manage_kts_setup(request):
    global device
    global byte_addr
    device = conftest.get_device(request)
    byte_addr = 0x3
    with device:
        err, _ = device.meta_request_uid()
        assert(err.esr_success)

    yield

def test_manage_kts_ecdsa_public_key():
    with device:

        # get ECDSA_KTS_PuK_key
        key_name = "ecdsa_pub_slot_{}".format(byte_addr)
        ECDSA_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDSA_KTS_PuK_key)

        # replace slot #3 ECDSA public key
        err, response = device.meta_manage_kts(b"\x80\x48", byte_addr, ECDSA_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        # Assert Return Types
        assert isinstance(err, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_manage_kts_ecdsa_private_key():
    with device:

        # get ECDSA_KTS_Prv_key
        key_name = "ecdsa_prv_slot_{}".format(byte_addr)
        ECDSA_KTS_Prv_key = authenta_key_mgmt_interface.export_private_key_in_the_clear(device.uid.hex(), key_name)
        assert(ECDSA_KTS_Prv_key)

        # replace slot #3 slave ECDSA private key
        err, response = device.meta_manage_kts(b"\x80\xC0", byte_addr, ECDSA_KTS_Prv_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        assert isinstance(response, bytes)

def test_manage_kts_ecdh_public_key():
    with device:

        # get ECDH_KTS_PuK_key
        key_name = "ecdh_pub_slot_{}".format(byte_addr)
        ECDH_KTS_PuK_key = authenta_key_mgmt_interface.read_public_key(device.uid.hex(), key_name)
        assert(ECDH_KTS_PuK_key)

        # replace slot #3 ECDH public key
        err, response = device.meta_manage_kts(b"\x80\x84", byte_addr, ECDH_KTS_PuK_key[::-1])
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

        assert isinstance(response, bytes)

def test_manage_kts_ecdsa_private_puf():
    with device:

        # replace slot #3
        err, _ = device.meta_manage_kts(b"\x80\x00", byte_addr, b"\x00\x00")
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

def test_manage_kts_ecdsa_private_puf_entropy():
    with device:

        # replace slot #3
        err, _ = device.meta_manage_kts(b"\x80\x20", byte_addr, b"\x00\x00")
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

def test_manage_kts_hmac_private_puf():
    with device:

        # replace slot #3
        err, _ = device.meta_manage_kts(b"\x00\x00", byte_addr, b"\x00\x00")
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

def test_manage_kts_option_1():
    with device:

        # Read the public key associated with a key table slot slave private key, ECDSA_KTSS_PrKx.
        err, _ = device.meta_manage_kts(b"\x80\x6D", byte_addr, b"\x00\x00"*64)
        assert(err.esr_and_response_success)
        assert(err._sig_matched)

@pytest.mark.parametrize("config, expectation", [(0x10, pytest.raises(TypeError)), (b"\x80\x00", does_not_raise())])
def test_manage_kts_config_type(config, expectation):
    with device:
        with expectation:
            err, _ = device.meta_manage_kts(config, 0x3, b"\x00\x00"*64)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("byte_addr, expectation", [(b"\x00\x00", pytest.raises(TypeError)), (0x3, does_not_raise())])
def test_manage_kts_byte_addr_type(byte_addr, expectation):
    with device:
        with expectation:
            err, _ = device.meta_manage_kts(b"\x80\x00", byte_addr, b"\x00\x00"*64)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)

@pytest.mark.parametrize("input_key, expectation", [(0x0, pytest.raises(TypeError)), (b"\x00\x00"*64, does_not_raise())])
def test_manage_kts_input_key_type_negative(input_key, expectation):
    with device:
        with expectation:
            err, _ = device.meta_manage_kts(b"\x80\x00", byte_addr, input_key)
            assert(err.esr_and_response_success)
            assert(err._sig_matched)
