import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_measure_setup(request):
    global device
    global provision_stats
    global input_digest
    device = conftest.get_device(request)
    input_digest = b"\x00"*32

    with device:
        device.meta_request_uid()
        if device.provision_sts_struct.Provisioned and \
            not device.provision_sts_struct.Remotely_Provisioned:
            rc, _ = device.meta_crtm_register_read(b'\00\00', 0x00)
            if rc.valid_response:
                if device.device_config_struct is not None:
                    if device.provision_sts_struct.HSM_Present:
                        if not device.device_config_struct.CRK_Read_Disable:
                            device.set_CRK_Read_Disable_bit(True)
                    else:
                        if not device.device_config_volatile_struct.CSK_Read_Disable:
                            device.set_CSK_Read_Disable_bit(True)

    yield

def test_measure_ecdsa_option_0():
    with device:
    
        # Create digest or HMAC
        # MeasureDestination is not set, no signature
        error_code, response = device.meta_measure(b"\x82\x28", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_hmac_option_0():
    with device:

        # Create digest or HMAC
        # MeasureDestination is not set, no signature
        error_code, response = device.meta_measure(b"\x02\x28", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_ecdsa_option_1():
    with device:

        # generate HMAC_CDI & related keys
        # MeasureDestination is not set, no signature
        error_code, response = device.meta_measure(b"\x82\x29", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_hmac_option_1():
    with device:

        # generate HMAC_CDI & related keys
        # MeasureDestination is not set, no signature
        error_code, response = device.meta_measure(b"\x02\x29", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_ecdsa_option_0_measure_destination():
    with device:

        # Create digest or HMAC
        # MeasureDestination is set, sign command
        error_code, response = device.meta_measure(b"\x82\xA8", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_hmac_option_0_measure_destination():
    with device:

        # Create digest or HMAC
        # MeasureDestination is set, sign command
        error_code, response = device.meta_measure(b"\x02\xA8", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_ecdsa_option_1_measure_destination():
    with device:

        # generate HMAC_CDI & related keys
        # MeasureDestination is set, sign command
        error_code, response = device.meta_measure(b"\x82\xA9", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_measure_hmac_option_1_measure_destination():
    with device:

        # generate HMAC_CDI & related keys
        # MeasureDestination is set, sign command
        error_code, response = device.meta_measure(b"\x02\xA9", 0x0, 0x0, 0x0, input_digest)
        assert(error_code.esr_and_response_success)
        assert(error_code._sig_matched)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

@pytest.mark.parametrize("config, expectation", [(0x0, pytest.raises(TypeError)), (b"\x80\x80", does_not_raise())])
def test_meta_measure_config_type(config, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(config, 0x0, 0x0, 0x0, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("partition_alt_src_addr, expectation", [(b"\x00", pytest.raises(TypeError)), (0x0, does_not_raise())])
def test_meta_measure_partition_alt_src_addr_type(partition_alt_src_addr, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(b"\x80\x80", partition_alt_src_addr, 0x0, 0x0, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("partition_key_src_addr, expectation", [(b"\x00", pytest.raises(TypeError)), (0x0, does_not_raise())])
def test_meta_measure_partition_key_src_addr_type(partition_key_src_addr, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(b"\x80\x80", 0x0, partition_key_src_addr, 0x0, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("partition_dest_addr, expectation", [(b"\x00", pytest.raises(TypeError)), (0x0, does_not_raise())])
def test_meta_measure_partition_dest_addr_type(partition_dest_addr, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(b"\x80\x80", 0x0, 0x0, partition_dest_addr, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("input_digest, expectation", [(0x0, pytest.raises(TypeError)), (b"\x00"*32, does_not_raise())])
def test_meta_measure_input_digest_type(input_digest, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(b"\x80\x80", 0x0, 0x0, 0x0, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)

@pytest.mark.parametrize("input_digest, expectation", [(b"\x00", pytest.raises(ValueError)), (b"\x00"*32, does_not_raise())])
def test_meta_measure_input_digest_size(input_digest, expectation):
    with device:
        with expectation:
            error_code, _ = device.meta_measure(b"\x80\x80", 0x0, 0x0, 0x0, input_digest)
            assert(error_code.esr_and_response_success)
            assert(error_code._sig_matched)
