import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

import authenta.authenta_key_mgmt_interface as key_mgmt_lib

@contextmanager
def does_not_raise():
    yield

@pytest.fixture(scope = "module", autouse = True)
def test_request_uid_setup(request):
    global device
    device = conftest.get_device(request)
    yield

def copy_key(device, key_name, intermediate_key_name, new_key_name):
    ecdsa_keys_changed = False
    if not key_mgmt_lib.delete_key(device.uid.hex(), intermediate_key_name):
        print("Failed to delete key '{}' for UID '{}' ".format(intermediate_key_name, device.uid.hex()))

    if key_mgmt_lib.copy_key(device.uid.hex(), key_name, intermediate_key_name):
        print("copied key '{}' to {} for UID '{}' ".format(key_name, intermediate_key_name, device.uid.hex()))
        if key_mgmt_lib.delete_key(device.uid.hex(), key_name):
            print("deleted key '{}' for UID '{}' ".format(key_name, device.uid.hex()))
            if key_mgmt_lib.copy_key(device.uid.hex(), new_key_name, key_name):
                print("copied key '{}' to '{}' for UID '{}' ".format(new_key_name, key_name, device.uid.hex()))
                ecdsa_keys_changed = True
            else:
                print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' ".format(new_key_name, key_name, device.uid.hex()))
        else:
            print("ERROR! Failed to delete key '{}' for UID '{}' ".format(key_name, device.uid.hex()))
    else:
        print("ERROR! Failed to copy key '{}' to '{}' for UID '{}' ".format(key_name, intermediate_key_name, device.uid.hex()))

    return ecdsa_keys_changed

def verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated):
    ecdsa_keys_changed = False
    if active_key_updated is not None:
        # the active ecdsa key has been changed
        # updating keys before replace_srk response validation
        ecdsa_keys_changed = copy_key(device, 'ecdsa_prv', 'previous_ecdsa_prv', 'new_ecdsa_prv')
        if not ecdsa_keys_changed:
            print("ERROR changing keys. Continuing to attempt validation, but expecting error!")

    ecdsa_resp_err = device.verify_replace_srk_response(config, response, sent_nonce, active_key_updated)
    if not ecdsa_resp_err.valid_response:
        if ecdsa_keys_changed:
            # changing back to old ECDSA private key due to replace_srk response validation failure
            ecdsa_keys_changed = copy_key(device, 'previous_ecdsa_prv', 'tmp_ecdsa_prv', 'ecdsa_prv')

    return ecdsa_keys_changed, ecdsa_resp_err

def verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated):
    ecdh_keys_changed = False
    if active_key_updated is not None:
        # updating keys before (ECDH) replace_srk response validation
        ecdh_keys_changed = copy_key(device, 'ecdh_prv', 'previous_ecdh_prv', 'new_ecdh_prv')
        if not ecdh_keys_changed:
            print("ERROR changing keys. Comtinuing to attempt validation, but expecting error!")
        
    ecdh_resp_err = device.verify_replace_srk_response(config, response, sent_nonce, active_key_updated)
    if not ecdh_resp_err.valid_response:
        if ecdh_keys_changed:
            # changing back to old ECDH private key due to replace_srk response validation failure
            ecdh_keys_changed = copy_key(device, 'previous_ecdh_prv', 'tmp_ecdh_prv', 'ecdh_prv')
    
    return ecdh_keys_changed, ecdh_resp_err

def change_ecdsa_key(active_key_updated, ecdsa_keys_changed):
    if active_key_updated is not None and active_key_updated and not ecdsa_keys_changed:
        # changing ECDSA key in key library
        ecdsa_keys_changed = copy_key(device, 'ecdsa_prv', 'previous_ecdsa_prv', 'new_ecdsa_prv')
        if not ecdsa_keys_changed:
            print("ERROR changing ECDSA key. Comtinuing to attempt validation, but expecting error!")
        
    return ecdsa_keys_changed

@pytest.mark.parametrize("config, expectation", [(0x0, pytest.raises(TypeError)), (b"\x80\x05", does_not_raise())])
def test_meta_replace_srk_config_type(config, expectation):
    with device:
        with expectation:
            device.meta_replace_srk(config, b"\x00\x00")

@pytest.mark.parametrize("public_key, expectation", [(0x0, pytest.raises(TypeError)), (b'\x00'*64, does_not_raise())])
def test_meta_replace_srk_public_key_type(public_key, expectation):
    with device:
        with expectation:
            device.meta_replace_srk(b'\x80\x04', public_key)

def test_replace_srk_ecdsa_option_1_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # revert to Micron host server public keys in Slot #1 & make slot #1 the active keys
        config = b"\x80\x01"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_hmac_option_1_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # revert to Micron host server public keys in Slot #1 & make slot #1 the active keys
        config = b"\x00\x01"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_ecdsa_option_3_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # revert to Micron PUF_UDS
        config = b"\x80\x07"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_hmac_option_3_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # revert to Micron PUF_UDS
        config = b"\x00\x07"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_ecdsa_option_2_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # replace active UDS (either PUF_UDS or customer UDS) with customer key
        config = b"\x80\x06"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_hmac_option_2_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        # replace active UDS (either PUF_UDS or customer UDS) with customer key
        config = b"\x00\x06"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, b"\x00\x00")
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

def test_replace_srk_ecdsa_option_0_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        new_ecdsa_pub_key_name = 'new_ecdsa_pub'
        new_ecdsa_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdsa_pub_key_name)
        assert(new_ecdsa_pub_key)

        # replace Customer host server public key in Slot #2 & make Slot #2 the active keys if both keys exist
        # replace ECDSA public key
        config = b"\x80\x04"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdsa_pub_key[::-1])
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        new_ecdh_pub_key_name = 'new_ecdh_pub'
        new_ecdh_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdh_pub_key_name)
        assert(new_ecdh_pub_key)

        # replace ECDH public key
        config = b"\x80\x00"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdh_pub_key[::-1])
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

        # at this point, if the ECDSA key has not already been changed,
        # change them now.
        ecdsa_keys_changed = change_ecdsa_key(active_key_updated, ecdsa_keys_changed)

    # clear keys by doing with device again
    with device:
        config = b'\x80\x00'
        # run update start
        error_code, _ = device.meta_acrtm_update_start(config)
        assert(error_code.esr_success)

        # run update finish
        error_code, _ = device.meta_acrtm_update_finish(config)
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

def test_replace_srk_hmac_option_0_positive():
    with device:

        err, _ = device.meta_request_uid()
        assert(err.esr_success)

        new_ecdsa_pub_key_name = 'new_ecdsa_pub'
        new_ecdsa_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdsa_pub_key_name)
        assert(new_ecdsa_pub_key)

        # replace Customer host server public key in Slot #2 & make Slot #2 the active keys if both keys exist
        # replace ECDSA public key
        config = b"\x00\x04"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdsa_pub_key[::-1])
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdsa_keys_changed, ecdsa_resp_err = verify_replace_srk_response_ecdsa(config, response, sent_nonce, active_key_updated)
        assert(ecdsa_resp_err.esr_and_response_success)

        new_ecdh_pub_key_name = 'new_ecdh_pub'
        new_ecdh_pub_key = key_mgmt_lib.read_public_key(device.uid.hex(), new_ecdh_pub_key_name)
        assert(new_ecdh_pub_key)

        # replace ECDH public key
        config = b"\x00\x00"
        error_code, response, sent_nonce, active_key_updated = device.meta_replace_srk(config, new_ecdh_pub_key[::-1])
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)

        ecdh_keys_changed, ecdh_resp_err = verify_replace_srk_response_ecdh(config, response, sent_nonce, active_key_updated)
        assert(ecdh_resp_err.esr_and_response_success)

        # at this point, if the ECDSA key has not already been changed,
        # change them now.
        ecdsa_keys_changed = change_ecdsa_key(active_key_updated, ecdsa_keys_changed)

    # clear keys by doing with device again
    with device:
        config = b'\x00\x00'
        # run update start
        error_code, _ = device.meta_acrtm_update_start(config)
        assert(error_code.esr_success)

        # run update finish
        error_code, _ = device.meta_acrtm_update_finish(config)
        assert(error_code.esr_success)

        # Assert Return Types
        assert isinstance(error_code, IndigoCmdErrorCode)
        assert isinstance(response, bytes)
