import sys
import os
import conftest
import pytest

from contextlib import contextmanager
from authenta.indigo.device import IndigoDevice, PlatformReturnType
from authenta.authenta_device import AuthentaDevice
from authenta.indigo.cmd_error_code import IndigoCmdErrorCode

@pytest.fixture(scope = "module", autouse = True)
def test_utility_methods_setup(request):
    global device
    device = conftest.get_device(request)
    yield

@contextmanager
def does_not_raise():
    yield

def test_generic_pack_command_length_positive():
    with device:
        cmd_str = device._generic_pack_command(b'\x9B\x3E', b"\x00\x00")
        assert len(cmd_str) == 512

def test_generic_pack_command_length_negative():
    with device:
        with pytest.raises(ValueError):
            device._generic_pack_command(b'\x9B\x3E', b"\x00")

@pytest.mark.parametrize("response_bytes, expectation", [(43, pytest.raises(TypeError)), (b'\x12\x34'*256, does_not_raise())])
def test_parse_response_bytes_type(response_bytes, expectation):
    with device:
        with expectation:
            device._parse_response_bytes(response_bytes)

@pytest.mark.parametrize("response_bytes, expectation", [(b'\x12\x34'*56, pytest.raises(ValueError)), (b'\x12\x34'*256, does_not_raise())])
def test_parse_response_bytes_length(response_bytes, expectation):
    with device:
        with expectation:
            device._parse_response_bytes(response_bytes)

def test_generic_pack_return_type():
    with device:
        cmd_str = device._generic_pack_command(b'\x9B\x3E', b"\x00\x00")
        assert isinstance(cmd_str, bytes)

@pytest.mark.parametrize("req,expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00", does_not_raise())])
def test_generic_pack_request_input_type(req, expectation):
    with device:
        with expectation:
            device._generic_pack_command(req, b"\x00\x00")

@pytest.mark.parametrize("config, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00", does_not_raise())])
def test_generic_pack_config_input_type(config, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", config)

@pytest.mark.parametrize("address, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*2, does_not_raise())])
def test_generic_pack_address_input_type(address, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", address = address)

@pytest.mark.parametrize("monotonic_counter, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*4, does_not_raise())])
def test_generic_pack_monotonic_counter_input_type(monotonic_counter, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", monotonic_counter = monotonic_counter)

@pytest.mark.parametrize("nonce, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*8, does_not_raise())])
def test_generic_pack_nonce_input_type(nonce, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", a_nonce = nonce)

@pytest.mark.parametrize("data, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*128, does_not_raise())])
def test_generic_pack_data_input_type(data, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", data = data)

@pytest.mark.parametrize("mac_ecdsa_s, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*16, does_not_raise())])
def test_generic_pack_mac_ecdsa_s_input_type(mac_ecdsa_s, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", mac_ecdsa_s = mac_ecdsa_s)

@pytest.mark.parametrize("ecdsa_r, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*16, does_not_raise())])
def test_generic_pack_ecdsa_r_input_type(ecdsa_r, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", ecdsa_r = ecdsa_r)

@pytest.mark.parametrize("stuff_bytes, expectation", [(43, pytest.raises(TypeError)), (b"\x00\x00"*80, does_not_raise())])
def test_generic_pack_stuff_bytes_input_type(stuff_bytes, expectation):
    with device:
        with expectation:
            device._generic_pack_command(b"\x00\x00", b"\x00\x00", stuff_bytes = stuff_bytes)

@pytest.mark.parametrize("command_name, expectation", [(43, pytest.raises(TypeError)), ('request_uid', does_not_raise())])
def test_meta_handle_atomic_command_response_command_name_type(command_name, expectation):
    with device:
        with expectation:
            device._meta_handle_atomic_command_response(command_name, b"\x00\x00", b"\x12"*512)

@pytest.mark.parametrize("response,expectation", [(43, pytest.raises(TypeError)), (b"\x12"*512, does_not_raise())])
def test_meta_handle_atomic_command_response_response_type(response, expectation):
    with device:
        with expectation:
            device._meta_handle_atomic_command_response('request_uid', b"\x00\x00", response)

@pytest.mark.parametrize("do_response_verification,expectation", [(43, pytest.raises(TypeError)), (True, does_not_raise())])
def test_meta_handle_atomic_command_response_do_response_verification_type(do_response_verification, expectation):
    with device:
        with expectation:
            device._meta_handle_atomic_command_response('request_uid', b"\x00\x00", b"\x12"*512, do_response_verification)

@pytest.mark.parametrize("caller_name,expectation", [(43, pytest.raises(TypeError)), ("caller name", does_not_raise())])
def test_get_amtc0_input_type(caller_name, expectation):
    with device:
        with expectation:
            device._get_amtc0_if_not_set(caller_name)

@pytest.mark.parametrize("caller_name,expectation", [(43, pytest.raises(TypeError)), ("caller name", does_not_raise())])
def test_get_boot_amtc0_input_type(caller_name, expectation):
    with device:
        with expectation:
            device._get_boot_amtc0_if_not_set(caller_name)

def test_get_boot_amtc0_return_type():
    with device:
        error_code = device._get_boot_amtc0_if_not_set('meta_acrtm_update_start')
        assert isinstance(error_code, IndigoCmdErrorCode)

def test_get_amtc0_return_type():
    with device:
        error_code = device._get_amtc0_if_not_set('meta_acrtm_update_start', b'\x56'*6)
        assert isinstance(error_code, IndigoCmdErrorCode)
