#! /usr/bin/env python3

"""
Test setup.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import json
import time
import pytest
import os
import sys

from authenta.naxos.device_1_5 import NaxosDevice_1_5
from authenta.grpc_platform import GrpcPlatform
import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers
import authenta.authenta_key_mgmt_interface as key_mgmt
import authenta.authenta_key_lib_interface as key_lib_interface
from hashlib import sha256

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import iboot
#import wemo as iboot

device_json_file_name = 'devices.json'
key_xml_file_name = 'key_contents_prv.xml'
ep_alias = 'snor1'

# These keys should be put in the key xml file before tests
srk_handle = '9001'
uds_handle = '9002'
test_key_handle = '9003'
micron_srk_handle = '9004'
micron_uds_handle = '9005'

def get_dev_info_from_json(filename, alias):
  """
  Get device info from device json file.
  return (device path, endpoint IP address, endpoint port number, iboot IP address)
  """
  try:
    with open(filename) as f:
      json_data = json.load(f)

    json_data = json_data[alias]

    ip_params = json_data['endpoint_ip'].split(':')
    ep_ip = ip_params[0]
    ep_port = None
    if len(ip_params) > 1:
      ep_port = ip_params[1]

    dev_path = json_data['endpoint_fn']

    iboot_ip = json_data.get('iboot_ip', '')
    iboot_ip = iboot_ip if iboot_ip != '' else None

  except FileNotFoundError:
    print('Error: Failed to open file:{}'.format(filename))
    pytest.exit(1)
  except json.decoder.JSONDecodeError:
    print('Error: Failed to parse json file:{}'.format(filename))
    pytest.exit(1)
  except:
    print('Error: Failed to get device info from json file:{}'.format(filename))
    pytest.exit(1)

  return dev_path, ep_ip, ep_port, iboot_ip

def check_provision_state(device):
  """
  Check device provion state (for now, test does not support local provisioned device)
  If not provisioned, ask the user if they want to provision the device
  If user answers yes, provision the device for remote use with RIoT turned on
  return True if operation is successful or False if not
  """
  err, *_ = device.meta_request_uid()
  if not err.esr_and_response_success:
    return False

  provision_sts = device.provision_sts_struct

  if provision_sts.Provisioned == 0:
    # not proviisoned
    # ask user whether to provison device
    ans = input("\nThe device is not provisioned.\n" \
      "If the device is not provisioned, a subset of the tests including provisoin and deprovision will be run.\n" \
      "If the device is provisioned, most tests excluding provision and deprovision will be run.\n" \
      "Provison device? [y or n]: ")
    if ans.lower() == 'y':
      # provision device for remote use with RIoT turned on
      config = b'\x05'
      err, *_ = device.meta_provision_crtm(config)
      if not err.esr_and_response_success:
        print('Error: Failed to provision device')
        return False
      
      # update provision status
      ret = utils.get_dev_info(device)
      if ret is None:
        return False

  elif provision_sts.Remotely_Provisioned == 0:
    print('Device is not provisioned for remote use')
    print('Error: Test doesn\'t support current provision state')
    return False
  else:
    print('Device has already been provisioned for remote use')

  return True

def force_clear_invalid_bit(device):
  """
  Force to clear CRTM_Range_Invalid bit by calling ACRTM_Update_Finish opton #0
  return True if operation is successful or False if not
  """
  # ACRTM_Update_Start
  config = b'\x00'
  err, _ = device.meta_acrtm_update_start(config)
  if not err.esr_and_response_success:
    return False

  # ACRTM_Update_Finish
  err, _ = device.meta_acrtm_update_finish(config)
  if not err.esr_and_response_success:
    return False

  return True

def write_to_flash(device):
  """
  Write custom data to flash
  return true if operation is successful or false if failed
  """
  for addr in range(0xe00000, 0xe20000, 0x1000):
    # unlock blocks
    ret = device.meta_dyb_write(False, addr)
    if not ret:
      return False

    # erase the data in specific blocks
    ret = device.meta_subsector_erase(addr)
    if not ret:
      device.meta_dyb_write(True, addr)
      return False

    # write data to specific positions
    data = bytes('This_is_a_test_' + format((addr - 0xe00000), 'x'), 'utf8')
    ret = device.meta_page_program(addr, data)
    if not ret:
      device.meta_dyb_write(True, addr)
      return False

    # lock blocks
    ret = device.meta_dyb_write(True, addr)
    if not ret:
      return False

  return True

@pytest.fixture(scope='session', autouse=True)
def run_setup(request):
  """
  Run test setup before all tests
  """
  dev_path, ep_ip, ep_port, iboot_ip = get_dev_info_from_json(
    device_json_file_name, ep_alias)

  if ep_port is None:
    ep = GrpcPlatform(dev_path, ep_ip)
  else:
    ep = GrpcPlatform(dev_path, ep_ip, ep_port)

  device = NaxosDevice_1_5(ep)

  with device:
    if device.provision_sts_struct.Provisioned and not device.provision_sts_struct.Remotely_Provisioned:
      # locally provisioned get the CRK or CSK from the KTR register and set in key driver
      err, crk = device.read_ktr(verify_sig=False)
      if not err.esr_success:
        print('Error: Failed to read CRK or CSK from the device KTR register')
        pytest.exit(1)

      if device.provision_sts_struct.TPM_Supported:
        # KTR is holding the CRK
        if not key_mgmt.add_key(device.uid_as_str, key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
          print('Error: Failed to add the CRK to the key repository')
#          pytest.exit(1)
      else:
        # KTR is holding the CSK
        if not key_mgmt.add_key(device.uid_as_str, key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
          print('Error: Failed to add the CRK to the key repository')
          pytest.exit(1)
            
    # Make sure srk is correct by checking the signature of Request_UID
    ret = utils.check_srk(device)
    if not ret:
      print('Error: Failed in checking SRK')
      pytest.exit(1)

    # Check device provion state (for now, test does not support local provisioned device)
    # If not provisioned, ask the user if they want to provision the device
    # If user answers yes, provision the device for remote use with RIoT turned on
    ret = check_provision_state(device)
    if not ret:
      print('Error: Failed in checking provision state')
      pytest.exit(1)

    if device.provision_sts_struct.Provisioned == 1:
      # force to clear CRTM_Range_Invalid bit by calling ACRTM_Update_Finish opton #0
      ret = force_clear_invalid_bit(device)
      if not ret:
        print('Error: Failed to clear CRTM_Range_Invalid bit')
        pytest.exit(1)

      # write test data to flash
      print('Writing test data to flash ...')
      ret = write_to_flash(device)
      if not ret:
        print('Error: Failed to write data to flash for test setup')
        pytest.exit(1)

  # save shared data in pytest cache
  request.config.cache.set('dev_path', dev_path)
  request.config.cache.set('ep_ip', ep_ip)
  request.config.cache.set('ep_port', ep_port)
  request.config.cache.set('iboot_ip', iboot_ip)

def reboot(device, iboot_ip):
  """
  Reboot device and verify the connection to RPC server
  return True on success or False on failure
  """
  print()
  print('Rebooting device ...')

  if iboot_ip:
    # automatically reboot
    iboot.run(iboot_ip, 'off')
    time.sleep(5)
    iboot.run(iboot_ip, 'on')
  else:
    # manual reboot
    input('Please unplug all cables from device, reboot the device (and RPC server), and Press Enter:')

  flag = False
  retry_times = 0
  time_out = 20
  max_retry_times = 10

  while not flag and retry_times < max_retry_times:
    try:
      with device:
        err, *_ = device.meta_request_uid()
        if err.esr_and_response_success:
          flag = True
        else:
          print('Retry connecting to RPC server ...')
          time.sleep(time_out)
          retry_times += 1
    except:
      print('Retry connecting to RPC server ...')
      time.sleep(time_out)
      retry_times += 1

  if not flag:
    print('Error: Failed to connect to RPC server')
    return False

  return True


def replace_srk_util(device, uid, config, new_key_name=None, old_key_name=None):
  """
  Replace_SRK with different options (including the step to write key data to KTR)
  and update key data in the key xml file
  return True on success or Flase on failure
  """
  config_int = int.from_bytes(config, byteorder='big')    
  if config_int & 0x03 == 0x00 or config_int & 0x03 == 0x02:
    # replace with third-party key
    # get new key data
    if new_key_name is None:
      print('ERROR: new_key_name can\'t be None for config 0x00 and 0x02')
      return False

    if config_int & 0x03 == 0x02 and old_key_name is None:
      print('ERROR: old_key_name can\'t be None for config 0x02')
      return False

    key_data_to_ktr = key_mgmt.export_priv_key_as_hex(device.uid_as_str, new_key_name)
    if key_data_to_ktr is None:
      print('ERROR: Failed to get key data by name: {}'.format(new_key_name))
      return False
    key_data_to_ktr = bytes.fromhex(key_data_to_ktr)
    if key_data_to_ktr is None:
      print('ERROR: Failed to get key data by name: {}'.format(new_key_name))
      return False
  else:
    if config_int & 0x03 == 0x01 and new_key_name is None:
      print('ERROR: new_key_name can\'t be None for config 0x01')
      return False

    # revert back to micron key
    key_data_to_ktr = bytes([0x00] * 32)

  # write key data to KTR
  ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
    key_data_to_ktr, False)
  if not ret:
    print('ERROR: Failed to write key data to KTR')
    return False

  # Replace_SRK
  new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, new_key_name)
  new_key = bytes.fromhex(new_key)
  if old_key_name is not None:
      old_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, old_key_name)
      old_key = bytes.fromhex(old_key)
      data_to_hash = b"".join((old_key, new_key))
  else:
      data_to_hash = new_key
  hashed_data = sha256(data_to_hash).digest()
  err, response, sent_nonce = device.meta_replace_srk(config, additional_data=hashed_data)

  if not err.esr_success:
    print('ERROR: meta_replace_srk Failed: {} {}'.format(err.esr_bit_msg, err.error_msg))
    return False
  
  session_key_replaced_only = (config_int & 0x03 == 0x00) and (config_int & 0x20 != 0x00)
  derive_key = (config_int & 0x03 == 0x00 or config_int & 0x03 == 0x02) and (config_int & 0x10 != 0x00)

  if config_int & 0x03 == 0x00 or config_int & 0x03 == 0x01:
    key_name = 'srk'
    copy_name = 'srk_copy'
  else:
    key_name = 'uds'
    copy_name = 'uds_copy'

  tmp_boot_amtc0 = device._boot_amtc0
  tmp_session_key = device._server_session_key

  if config_int & 0x03 == 0x00 or config_int & 0x03 == 0x01:
    # replace srk or session key
    if session_key_replaced_only:
      # only replace session key; srk is not replaced
      if derive_key:
        # derive new session key from key in KTR
        # Ksn(s)(new) = HMAC(KTR)Ksn(s)(existing)
        old_session_key = None
        if device._server_session_key is not None:
          old_session_key = device._server_session_key
        else:
          # generate old session key from srk
          key_lib_interface.set_srk_as_hmac_root_signing_key(device._keylib_context)
          device.get_boot_amtc0_if_not_set("replace_srk")
          old_session_key = key_lib_interface.derive_hmac_session_key(device._keylib_context, device._boot_amtc0, True)
          if old_session_key is None:
            print('Failed attempting to generate old session key after replace_srk')
            return
                
          device._server_session_key = hmac.new(old_session_key, key_data_to_ktr, sha256).digest()
          key_lib_interface.set_hmac_session_key(device._keylib_context, device._server_session_key)
      else:
        # replace session key with the key in the KTR
        if key_lib_interface.set_hmac_session_key(device._keylib_context, key_data_to_ktr):
          device.set_server_session_key(key_data_to_ktr)
    else:
      # replace srk
      if derive_key:
        # derive new srk from key in KTR
        # Kr(s)(new) = HMAC(KTR)Kr(s)(existing)
        key_lib_interface.set_srk_as_hmac_root_signing_key(device._keylib_context)
        new_key = key_lib_interface.gen_root_sha256_hmac(device._keylib_context, key_data_to_ktr)
        if new_key is None:
          print('Failed attempting to derive new srk after replace_srk')
          return

      # make a copy of the old SRK in key XML
      key_mgmt.copy_key(device.uid_as_str, key_name, copy_name)

      # delete original 'srk' key from key XML
      key_mgmt.delete_key(device.uid_as_str, key_name)

      # add new key to key XML
      key_mgmt.add_key(device.uid_as_str, key_name, new_key, None)

      # if replacing SRK succeeded the boot_amtc0 is no longer valid and
      # will need to be updated (= current amtc0) to verify the command response
      # (since the session key will be based on the new SRK and new boot_amtc0)
      device._boot_amtc0 = device._amtc0

      # clear session key
      device._server_session_key = None


  # Verify response from meta_replace_srk
  device.verify_replace_srk_response(response, sent_nonce, err, sign_with_session_key=True, force=True)

  if err.esr_and_response_success:
    if not session_key_replaced_only:
      if err.esr_and_response_success:
        # delete copy of original key from XML file
        key_mgmt.delete_key(device.uid_as_str, copy_name)
      else:
        # replace_srk failed, revert keys in XML file
        key_mgmt.delete_key(device.uid_as_str, key_name)
        key_mgmt.copy_key(device.uid_as_str, copy_name, key_name)
        key_mgmt.delete_key(device.uid_as_str, copy_name)

  return True

def auth_modify_write_data_to_src(device, addr, data):
  """
  For Auth_Modify option #1 (DataSource set to `1b: KTR || Customer_RAM) and #2,
  wirte data to specied source address (KTR or Customer_RAM) in the way required by spec
  return True on success or False on failure
  """
  data_len = len(data)
  assert(data_len % 4 == 0)

  for i in range(0, data_len, 4):
    # CRTM_Register_Write
    data_to_write = data[i:(i + 4)]
    data_to_write = data_to_write[::-1]
    addr_to_write = addr + i

    err, _ = device.meta_crtm_register_write(b"\x00", addr_to_write, data_to_write, False)
    if not err.esr_and_response_success:
      return False

  return True

def get_device(request):
  """
  Create NaxosDevice_1_5 object from pytest cache data
  """
  dev_path = request.config.cache.get("dev_path", None)
  ep_ip = request.config.cache.get("ep_ip", None)
  ep_port = request.config.cache.get("ep_port", None)
  
  if ep_port is None:
    ep = GrpcPlatform(dev_path, ep_ip)
  else:
    ep = GrpcPlatform(dev_path, ep_ip, ep_port)

  device = NaxosDevice_1_5(ep)

  return device
