#! /usr/bin/env python3

"""
API Test for ACRTM_Upate_Start.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_acrtm_update_start_setup(request):
  global device
  global provision_stats

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

  yield

def test_acrtm_update_start_Positive_CrtmRegs():
  """
  Option #0
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    err, _ = device.meta_acrtm_update_start(config)
    assert(err.esr_and_response_success)

    # call API that needs a Update_Start
    data = bytes.fromhex("00000000")
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(config)
      assert(False)

    err, _ = device.meta_acrtm_update_finish(config)
    assert(err.esr_and_response_success)

def test_acrtm_update_start_Positive_RecoveryBlock():
  """
  Option #1
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # enable automeasure
    ret = utils.enable_automeasure(device)
    assert(ret)

    # recovery bolock update
    config = b"\x01"

    err, _ = device.meta_acrtm_update_start(config)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_finish(config)
    assert(err.esr_and_response_success)

    # disable automeasure
    ret = utils.disable_automeasure(device)
    assert(ret)

def test_acrtm_update_start_Positive_MultiCalls():
  """
  Call ACRTM_Update_Start multiple times
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    err, _ = device.meta_acrtm_update_start(config)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_start(config)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(config)
      assert(False)

    err, _ = device.meta_acrtm_update_finish(config)
    assert(err.esr_and_response_success)

def test_acrtm_update_start_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"

      err, _ = device.meta_acrtm_update_start(config)
      assert(err.esr == b"\x00")

def test_acrtm_update_start_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_acrtm_update_start(config)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)

    device.test_only_set_boot_amtc0(None)

def test_acrtm_update_start_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_acrtm_update_start(config)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_acrtm_update_start_Negative_RangeInvalid():
  """
  command option #1 & CRTM_Range_Invalid set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # set CRTM_Range_Invalid bit
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_start(b"\x01")
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_acrtm_update_start_Negative_RecoveryBlock():
  """
  command option #0 & Recovery_Block set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    ret = utils.enable_automeasure(device)
    assert(ret)

    # set Recovery_Block bit
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_start(b"\x00")
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    ret = utils.disable_automeasure(device)
    assert(ret)

def test_acrtm_update_start_Negative_AutoMeasureActive():
  """
  command option #1 & Auto_Measure_Activate not set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    ret = utils.disable_automeasure(device)
    assert(ret)

    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr == b"\x20")
