#! /usr/bin/env python3

"""
API Test for Auth_Modify.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers
import authenta.authenta_key_mgmt_interface as key_mgmt
from hashlib import sha256

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_auth_modify_setup(request):
  global device
  global iboot_ip
  global provision_stats
  # address for block to test
  global block_addr
  block_addr = 0xe20000

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      # configure as AM block
      ret = utils.configure_am(device, block_addr, True)
      assert(ret)

  yield

  if provision_stats.Provisioned == 1:
    with device:
      # configure as legacy block
      ret = utils.configure_am(device, block_addr, False)
      assert(ret)

def test_auth_modify_Positive_Erase_AM():
  """
  option #0: erase
  erase AM block and keep it as AM block
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, block_addr, dest_addr, size)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Erase_AP():
  """
  option #0: erase
  erase AM block and make it AP block
  erase AP block and make it AP block
  erase AP block and make it AM block
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    dest_addr = 0x0
    size = 0x0

    # Auth_Modify(option #0): erase AM block and make it AP block
    config = b"\x00\x10"

    err, _ = device.meta_auth_modify(config, block_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase AP block and make it AP block
    config = b"\x00\x10"

    err, _ = device.meta_auth_modify(config, block_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase AP block and make it AM block
    config = b"\x00\x00"

    err, _ = device.meta_auth_modify(config, block_addr, dest_addr, size)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Program_SrcFlash_1():
  """
  option #1: program content
  source: flash main array
  size: 0x80
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x80
    source_content = bytes.fromhex("11") * size

    # write source content to legacy block
    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    ret = utils.write_data_to_block(device, source_addr, source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Program_SrcFlash_2():
  """
  option #1: program content
  source: flash main array
  size: 0x100
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x100
    source_content = bytes.fromhex("11") * size

    # write source content to legacy block
    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    ret = utils.write_data_to_block(device, source_addr, source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Program_SrcFlash_3():
  """
  option #1: program content
  source: flash main array
  size: 0x10000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x10000
    source_content = bytes.fromhex("11") * size

    # write source content to legacy block
    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    block_size = 0x1000
    for start in range(0, size, block_size):
      addr = source_addr + start
      ret = utils.write_data_to_block(device, addr, source_content[start:(start + block_size)])
      assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Program_SrcKTR_1():
  """
  option #1: program content
  source: KTR || Customer_RAM
  size: 0x20
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x05"
    source_addr = 0x0
    size = 0x20
    source_content = bytes.fromhex("11") * size

    # write source content to KTR
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Program_SrcKTR_2():
  """
  option #1: program content
  source: KTR || Customer_RAM
  size: 0x80
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x05"
    source_addr = 0x0
    size = 0x80
    source_content = bytes.fromhex("11") * size

    # write source content to Custom_RAM
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.Customer_RAM, \
      source_content[0:0x60])
    assert(ret)

    # write source content to KTR
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content[0x60:0x80])
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Append_1():
  """
  option #2: append content
  size: 0x20
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # Auth_Modify(option #0): erase AM block to be configured as an AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    config = b"\x00\x02"
    source_addr = 0x0
    size = 0x20

    # Auth_Modify(option #2)
    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase AP block to be configured as an AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_Append_2():
  """
  option #2: append content
  size: 0x80
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # Auth_Modify(option #0): erase AM block to be configured as an AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    config = b"\x00\x02"
    source_addr = 0x0
    size = 0x80

    # Auth_Modify(option #2)
    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase AP block to be configured as an AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_HalfDeviceSwap_SourceAddr_0():
  """
  option #3: half device swap
  SourceAddr = 0x0
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # check device swap state
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config_int = int.from_bytes(response[16:20], byteorder="big")
    swapped = (device_config_int & 0x00000020 != 0)
    
    # Set Half_Device_Swap_Enable bit in device config
    data_int = device_config_int | 0x00000040
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    half_addr = 0x800000
    half_size = 0x800000
    source_addr = 0x0

    # call measure to get digest of first half device
    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, half_size)
    assert(ret)

    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
    assert(ret)

    err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest is not None)

    if not swapped:
      # Auth_Modify(option #3): enter swapped state
      err, _ = device.meta_auth_modify(b"\x00\x0b", source_addr, 0x0, 0x0, source_content_digest=digest)
      assert(err.esr_and_response_success)

      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      device_config_int = int.from_bytes(response[16:20], byteorder="big")
      assert(device_config_int & 0x00000020 != 0)

      # call measure to get digest of second half device
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, half_addr, half_size)
      assert(ret)

      err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
      assert(err.esr_and_response_success)

      digest1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest1 is not None)
      assert(digest1 == digest)

      # call measure to get digest of first half device
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, half_size)
      assert(ret)

      err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
      assert(err.esr_and_response_success)

      digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest is not None)

    # Auth_Modify(option #3): enter unswapped state
    err, _ = device.meta_auth_modify(b"\x00\x03", source_addr, 0x0, 0x0, source_content_digest=digest)
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config_int = int.from_bytes(response[16:20], byteorder="big")
    assert(device_config_int & 0x00000020 == 0)

    # call measure to get digest of second half device
    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, half_addr, half_size)
    assert(ret)

    err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    digest1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest1 is not None)
    assert(digest1 == digest)

    # disable Half_Device_Swap_Enable bit in device config
    data_int = device_config_int & 0xFFFFFFBF
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_HalfDeviceSwap_SourceAddr_Middle():
  """
  option #3: half device swap
  SourceAddr = 0x800000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # check device swap state
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config_int = int.from_bytes(response[16:20], byteorder="big")
    swapped = (device_config_int & 0x00000020 != 0)
    
    # Set Half_Device_Swap_Enable bit in device config
    data_int = device_config_int | 0x00000040
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    half_addr = 0x800000
    half_size = 0x800000
    source_addr = half_addr

    # call measure to get digest of second half device
    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, half_addr, half_size)
    assert(ret)

    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
    assert(ret)

    err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest is not None)

    if not swapped:
      # Auth_Modify(option #3): enter swapped state
      err, _ = device.meta_auth_modify(b"\x00\x0b", source_addr, 0x0, 0x0, source_content_digest=digest)
      assert(err.esr_and_response_success)

      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      device_config_int = int.from_bytes(response[16:20], byteorder="big")
      assert(device_config_int & 0x00000020 != 0)

      # call measure to get digest of first half device
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, half_size)
      assert(ret)

      err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
      assert(err.esr_and_response_success)

      digest1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest1 is not None)
      assert(digest1 == digest)

      # call measure to get digest of second half device
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, half_addr, half_size)
      assert(ret)

      err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
      assert(err.esr_and_response_success)

      digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest is not None)

    # Auth_Modify(option #3): enter unswapped state
    err, _ = device.meta_auth_modify(b"\x00\x03", source_addr, 0x0, 0x0, source_content_digest=digest)
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config_int = int.from_bytes(response[16:20], byteorder="big")
    assert(device_config_int & 0x00000020 == 0)

    # call measure to get digest of first half device
    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, half_size)
    assert(ret)

    err, _ = device.meta_measure(b"\x00\x08", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    digest1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest1 is not None)
    assert(digest1 == digest)

    # disable Half_Device_Swap_Enable bit in device config
    data_int = device_config_int & 0xFFFFFFBF
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_ClearKTRFlags():
  """
  All KTR flags shall be cleared by Auth_Modify command.
  """    
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  with device:
    key_data_to_ktr = b"\x00" * 32

    # set KTR flags by writing data to KTR
    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      key_data_to_ktr, False)
    assert(ret)

    # option 0
    config = b"\x00\x00"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, block_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # after auth_modify command, KTR flags should be cleared
    # and replace_srk command should return ESR 5
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    # set KTR flags by writing data to KTR
    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      key_data_to_ktr, False)
    assert(ret)

    # option 1
    config = b"\x00\x05"
    source_addr = 0x0
    size = 0x20

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=key_data_to_ktr)
    assert(err.esr_and_response_success)

    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    # option 2
    # Auth_Modify(option #0): erase AM block to be configured as an AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # set KTR flags by writing data to KTR
    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      key_data_to_ktr, False)
    assert(ret)

    config = b"\x00\x02"
    source_addr = 0x0
    size = 0x20

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size)
    assert(err.esr_and_response_success)

    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    if err.esr != b"\x20":
      # Auth_Modify(option #0): erase AP block to be configured as an AM block
      err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
      assert(err.esr_and_response_success)

      assert(False)

    # Auth_Modify(option #0): erase AP block to be configured as an AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

def test_auth_modify_Positive_IncreaseAppendAddr():
  """
  AppendConfig AppendAddr can only be incremented with Auth Modify option #2.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # Auth_Modify(option #0): erase AM block to be configured as an AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # read AppendConig
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)

    append_config_addr = None
    index = -1
    append_config = response[16:32]
    for i in range(0, len(append_config), 4):
      data = append_config[i:(i + 4)]
      if data[0] == 0x00 and data[1] * 0x10000 == block_addr:
        append_config_addr = int.from_bytes(data[2:], byteorder="big")
        index = i

    assert(index != -1)

    config = b"\x00\x02"
    source_addr = 0x0
    size = 0x80

    # Auth_Modify(option #2)
    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size)
    assert(err.esr_and_response_success)

    # read AppendConig
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)

    append_config = response[16:32]
    data = append_config[index:(index + 4)]
    append_config_addr_1 = int.from_bytes(data[2:], byteorder="big")

    assert(append_config_addr_1 == append_config_addr + size)

    # try option #1 - should fail and not change AppendConfig
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x80
    source_content = bytes.fromhex("11") * size

    # write source content to legacy block
    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    ret = utils.write_data_to_block(device, source_addr, source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, \
      source_content=source_content)
    assert(err.esr != 0x80)

    # read AppendConig
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)

    append_config_1 = response[16:32]
    assert(append_config_1 == append_config)

    # Auth_Modify(option #0): erase AP block to be configured as an AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # read AppendConig
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)

    append_config = response[16:32]
    data = append_config[index:(index + 4)]
    assert(data == b"\xff" * 4)

def test_auth_modify_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00\x00"
      source_addr = 0xe30000
      dest_addr = 0x0
      size = 0x0

      err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
      assert(err.esr == b"\x00")

def test_auth_modify_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    source_addr = 0xe30000
    dest_addr = 0x0
    size = 0x0

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)

    device.test_only_set_boot_amtc0(None)

def test_auth_modify_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    source_addr = 0xe30000
    dest_addr = 0x0
    size = 0x0

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_auth_modify_Negative_ModifyNonAMBlock():
  """
  Modify non-AM blocks using Auth_Modify
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 0
    test_addr = 0xe30000

    # configure as legacy block
    ret = utils.configure_am(device, test_addr, False)
    assert(ret)

    # option 0
    config = b"\x00\x00"
    source_addr = test_addr
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    assert(err.esr == b"\x35")

    # option 1
    config = b"\x00\x01"
    source_addr = 0xe20000
    dest_addr = test_addr
    size = 0x80

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, source_content=source_content)
    assert(err.esr == b"\x35")

    # option 2
    config = b"\x00\x02"
    source_addr = 0x0
    dest_addr = test_addr
    size = 0x20

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    assert(err.esr == b"\x36")

def test_auth_modify_Negative_ModifyAPBlock():
  """
  Option #1 Dest_Addr Try on AM block that is also an AP block. Should set ESR bit 2.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 0 - erase and create AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # option 1
    config = b"\x00\x01"
    source_addr = 0xe30000
    dest_addr = block_addr
    size = 0x80

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, source_content=source_content)
    assert(err.esr == b"\x04")

    # convert AP block to AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)


def test_auth_modify_Negative_InvalidParams_SourceAddr():
  """
  Source_Addr is invalid
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 0
    config = b"\x00\x00"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, 0xe31000, dest_addr, size)
    assert(err.esr == b"\x04")
    
    err, _ = device.meta_auth_modify(config, 0xFF000000, dest_addr, size)
    assert(err.esr == b"\x04")

    # option 1
    # source from main flash array
    config = b"\x00\x01"
    source_addr = 0xe30010
    dest_addr = block_addr
    size = 0x80

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, \
      source_content=source_content)
    assert(err.esr == b"\x04")

    # source from KTR || Customer_RAM
    config = b"\x00\x05"
    source_addr = 0xe30000
    dest_addr = block_addr
    size = 0x20
    source_content = b"\x00" * size

    # write data to KTR
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, \
      source_content=source_content)
    assert(err.esr == b"\x04")

    # option 2
    config = b"\x00\x02"

    err, _ = device.meta_auth_modify(config, 0xe30000, block_addr, 0x20)
    assert(err.esr == b"\x04")

def test_auth_modify_Negative_InvalidParams_DestAddr():
  """
  Dest_Addr is invalid
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 0
    config = b"\x00\x00"

    err, _ = device.meta_auth_modify(config, block_addr, 0xe30000, 0x0)
    assert(err.esr == b"\x04")

    # option 1
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x80

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, 0xe30010, size, source_content=source_content)
    assert(err.esr == b"\x04")

    err, _ = device.meta_auth_modify(config, source_addr, 0xFF000000, size, source_content=source_content)
    assert(err.esr == b"\x04")

    # option 2
    config = b"\x00\x02"
    source_addr = 0x0
    size = 0x20

    err, _ = device.meta_auth_modify(config, source_addr, 0xe31000, size)
    assert(err.esr == b"\x04")

    err, _ = device.meta_auth_modify(config, source_addr, 0xFF000000, size)
    assert(err.esr == b"\x04")

def test_auth_modify_Negative_InvalidParams_Size():
  """
  Size is invalid
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 0
    config = b"\x00\x00"

    err, _ = device.meta_auth_modify(config, block_addr, 0x0, 0x80)
    assert(err.esr == b"\x04")

    # option 1
    # source from main flash array
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x40

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, block_addr, size, source_content=source_content)
    assert(err.esr == b"\x04")

    # source from KTR || Customer_RAM
    config = b"\x00\x05"
    source_addr = 0xe30000
    size = 0x10
    source_content = b"\x00" * size

    # write data to KTR
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, source_content)
    assert(ret)

    err, _ = device.meta_auth_modify(config, 0x0, block_addr, size, \
      source_content=source_content)
    assert(err.esr == b"\x04")

    # option 2
    config = b"\x00\x02"

    err, _ = device.meta_auth_modify(config, 0x0, block_addr, 0x10)
    assert(err.esr == b"\x04")

def test_auth_modify_Negative_InvalidAddr():
  """
  option #1 && [Dest_Addr + Size - 1] exceeds block boundry
  option #1 && [Source_Addr + Size - 1] exceeds the main array address space of the device
  option #2 && [APAppendAddr + Size - 1] exceeds block boundry
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option 1
    # source from main flash array
    config = b"\x00\x01"
    source_addr = 0xe30000
    size = 0x1000

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(config, source_addr, block_addr + 0xff00, size, \
      source_content=source_content)
    assert(err.esr == b"\x04")

    err, _ = device.meta_auth_modify(config, 0xfff000, block_addr, 0x10000, \
      source_content=source_content)
    assert(err.esr == b"\x04")

    # option 2
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # convert test AM block to AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # append until the block is full        
    config = b"\x00\x02"
    size = 0x80

    # 256 non-authenticaed cmds will cause ESR 2
    # so add Update_Finish in the middle of cmds
    for i in range(0, int(0x10000 / size)):
      if i % 32 == 0:
        err, _ = device.meta_acrtm_update_finish(b"\x00")
        assert(err.esr_and_response_success)

      err, _ = device.meta_auth_modify(config, 0x0, block_addr, size)
      if not err.esr_and_response_success:
        device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
        assert(False)

    # When APAppendAddr + Size reach the boundry, 
    # should set APAppendAddr to 0xFFFF indicating block full
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)
    addr = None

    append_config = response[16:32]
    for i in range(0, len(append_config), 4):
      data = append_config[i:(i + 4)]
      if data[1] * 0x10000 == block_addr:
        addr = data[2:4]
        break
    assert(addr == b"\xff\xff")

    # one more append will exceed block boundry - should fail
    err, _ = device.meta_auth_modify(config, 0x0, block_addr, size)
    assert(err.esr == b"\x04")

    # convert test AP block to AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

def test_auth_modify_Negative_RecoveyMode():
  """
  Device is in Recovery Mode
  """
  if provision_stats.Provisioned == 0:
    return

  ap_block_addr = 0xe40000

  with device:
    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # turn on automeasure and recovery mode
    data_int = device_config_int | 0x00000003
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # set recovery base address
    recovery_base_addr = 0xf00000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # set CRTM range to meaure
    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xe00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # CRTM_Max_AutoMeasure_Size
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x02\x00\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # reprogram Flash_DigestNV and Flash_DigestNV_Partial
    err, _ = device.meta_flash_extend(b"\x01")
    assert(err.esr_and_response_success)

    # change measurement range
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # change CRTM range to measure
    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xe00000, 0x1000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # create test AP block
    ret = utils.configure_am(device, ap_block_addr, True)
    assert(ret)

    err, _ = device.meta_auth_modify(b"\x00\x10", ap_block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

  # reboot device to enter recovery mode
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # option 0
    err, _ = device.meta_auth_modify(b"\x00\x00", block_addr, 0x0, 0x0)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # option 1
    source_addr = 0xe30000
    size = 0x80
    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response is not None)
    source_content = response[:size]

    err, _ = device.meta_auth_modify(b"\x00\x01", source_addr, block_addr, size, \
      source_content=source_content)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # option 2
    err, _ = device.meta_auth_modify(b"\x00\x02", 0x0, ap_block_addr, 0x20)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # finish recovery mode
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # turn off automeasure and recovery mode
    data_int = device_config_int & 0xFFFFFFFC
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # convert test AP block back to leagcy block
    err, _ = device.meta_auth_modify(b"\x00\x00", ap_block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    ret = utils.configure_am(device, ap_block_addr, False)
    assert(ret)

def test_auth_modify_Negative_NoMoreAPBlocks():
  """
  option #0 && CreateAPBlock set && no more AP blocks may be created
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # try to create 4 AP blocks
    addrs = [0xe30000, 0xe40000, 0xe50000, 0xe60000]

    for addr in addrs:
      # configure as AM block
      ret = utils.configure_am(device, addr, True)
      assert(ret)

      # option 0 and CreateAPBlock
      config = b"\x00\x10"
      source_addr = 0xe30000
      size = 0x1000

      device.meta_auth_modify(config, addr, 0x0, 0x0)

    # try to create another AP block
    flag = True
    addr = 0xe70000
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    err, _ = device.meta_auth_modify(config, addr, 0x0, 0x0)
    if err.esr != b"\x37":
      flag = False

    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # convert 4 AP blocks back to legacy block
    for addr in addrs:
      # conver to AM blocks first
      config = b"\x00\x00"
      err, _ = device.meta_auth_modify(config, addr, 0x0, 0x0)
      assert(err.esr_and_response_success)

      # configure as legacy block
      ret = utils.configure_am(device, addr, False)
      assert(ret)

    assert(flag)

def test_auth_modify_Negative_RecoveryUpdateCreateAP():
  """
  option #0 && device is in Recovery Block update process
  && Source_Addr == Recovery_Base && Recovery Block is configured as AM Block
  && CreateAPBlock set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # enale automeasure and clear AMRecovery bit in Device_Config register
    data = b"\x00\x00\x00\x01"
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # set recovery block base address
    recovery_base_addr = block_addr
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # start recovery block update process
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase block and make it AP block
    config = b"\x00\x10"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, recovery_base_addr, dest_addr, size)
    if err.esr != b"\x3B":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # finish recovery block update process
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # disable automeasure
    ret = utils.disable_automeasure(device)
    assert(ret)

def test_auth_modify_Negative_RecoveryUpdateAMRecovery():
  """
  option #0 && device is in Recovery Block update process
  && Source_Addr == Recovery_Base && AMRecovery bit is 1
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # enale automeasure and set AMRecovery bit in Device_Config register
    data = b"\x00\x00\x00\x09"
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # set recovery block base address
    recovery_base_addr = block_addr
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # start recovery block update process
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase block
    config = b"\x00\x00"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, recovery_base_addr, dest_addr, size)
    if err.esr != b"\x35":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # finish recovery block update process
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # disable automeasure
    ret = utils.disable_automeasure(device)
    assert(ret)

def test_auth_modify_Negative_NonAMCreateAP():
  """
  Option #0 CreateAPBlock = 1. Try on a non-AM block. Should fail with ESR = 0x35.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    leagcy_addr = 0xe30000

    ret = utils.configure_am(device, leagcy_addr, False)
    assert(ret)

    # Auth_Modify(option #0): erase block and make it AP block
    config = b"\x00\x10"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, leagcy_addr, dest_addr, size)
    assert(err.esr == b"\x35")
