#! /usr/bin/env python3

"""
API Test for CRTM_Regsiter_Read.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_crtm_register_read_setup(request):
  global device
  global provision_stats

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

  yield

def test_crtm_register_read_Positive_All():
  """
  Read all CRTM registers with config set to 0x00
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # set all EEPROM to readable
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data = b"\x22" * 4
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    config = b"\x00"

    for addr in range(0x0, 0x360, 16):
      err, _ = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)

def test_crtm_register_read_Positive_DualTypeVolatile():
  """
  Read all dual-type CRTM registers with config set to 0x01
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"
    
    start = CRTM_Registers.CRTM_Range_Start_Addr_0
    end = start + 0x40
    for addr in range(start, end, 16):
      err, _ = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)

def test_crtm_register_read_Positive_EEPROM():
  """
  Read EEPROM registers one at a time with corresponding EEPROM_Attributes bit set for readable
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    # EEPROM_Attributes readonly for each EEPROM
    attr_list = (
      b"\x00\x00\x00\x02", b"\x00\x00\x00\x20", b"\x00\x00\x02\x00", b"\x00\x00\x20\x00",
      b"\x00\x02\x00\x00", b"\x00\x20\x00\x00", b"\x02\x00\x00\x00", b"\x20\x00\x00\x00"
    )

    for i in range(0, 8):
      data = attr_list[i]
      addr = CRTM_Registers.Customer_EEPROM + i * 32
      
      # set only one EEPROM to readable
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_read(config, addr + 16)
      assert(err.esr_and_response_success)

def test_crtm_register_read_Positive_VerifyAddrInResponse():
  """
  Verify the CRTM_Register_Address in response
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = CRTM_Registers.Device_Config
    err, response = device.meta_crtm_register_read(b"\x00", addr)
    assert(err.esr_and_response_success)
    addr1 = int.from_bytes(response[4:6], byteorder="big")
    assert(addr1 == addr)

def test_acrtm_crtm_register_read_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"
      addr = CRTM_Registers.CRTM_Range_Start_Addr_0

      err, _ = device.meta_crtm_register_read(config, addr)
      assert(err.esr != b"\x80")

def test_acrtm_crtm_register_read_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = CRTM_Registers.CRTM_Range_Start_Addr_0

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_crtm_register_read(config, addr)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_acrtm_crtm_register_read_Negative_InvalidAddr():
  """
  CRTM_Register_Address out of range (> 0xFFF)
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = 0xFFFF

    err, _ = device.meta_crtm_register_read(config, addr)
    assert(err.esr == b"\x04")

def test_acrtm_crtm_register_read_Negative_InvalidConfig():
  """
  Set config=`1b for registers other than CRTM_Addr_Size
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"
    addr = CRTM_Registers.EEPROM_Attributes

    err, _ = device.meta_crtm_register_read(config, addr)
    assert(err.esr == b"\x04")

def test_acrtm_crtm_register_read_Negative_EEPROMNotReadable():
  """
  EEPROM_Read bit clear && read EEPROM register
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = CRTM_Registers.Customer_EEPROM

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # set EEPROM to modifiable and not readable
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, b"\x11" * 4, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # write test data to EEPROM
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Customer_EEPROM, b"\x00\x11\x22\x33", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # should not read correct data
    err, response = device.meta_crtm_register_read(b"\x00", addr)
    assert(err.esr_and_response_success)
    assert(response[16:20] == b"\x00" * 4)
