#! /usr/bin/env python3

"""
API Test for CRTM_Regsiter_Write.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_crtm_register_write_setup(request):
  global device
  global provision_stats

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

  yield

def test_crtm_register_write_Positive_Volatile():
  """
  Write to volatile CRTM registers with config set to 0x00
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    sign = False
    test_data = b"\x00\x11\22\33"

    start = CRTM_Registers.KTR_Base
    end = start + 0x20
    ktr_addrs = list(range(start, end, 4))

    start = CRTM_Registers.Customer_RAM
    end = start + 0x60
    custom_ram_addrs = list(range(start, end, 4))

    # CRTM_Registers.Device_Config_Volatile can be only written when provision local without TPM
    # not include it in current test; it is tested in provision/deproviison tests
    addrs = ktr_addrs + custom_ram_addrs

    for addr in addrs:
      # read original data
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      index = 16 + addr % 16
      old_data = response[index:(index + 4)]

      # write test data
      err, _ = device.meta_crtm_register_write(config, addr, test_data, sign)
      assert(err.esr_and_response_success)

      # read data back to verify
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      assert(response[index:(index + 4)] == test_data)

      # write original data back
      err, _ = device.meta_crtm_register_write(config, addr, old_data, sign)
      assert(err.esr_and_response_success)

def test_crtm_register_write_Positive_NonVolatile():
  """
  Write to nonvolatile CRTM registers (except EEPROM) with config set to 0x00
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    sign = True
    test_data_common = b"\x00\x11\22\33"

    start = CRTM_Registers.CRTM_Range_Start_Addr_0
    end = start + 0x40
    crtm_range_addrs = list(range(start, end, 4))

    start = CRTM_Registers.Auth_Mod_Block_Config
    end = start + 0x20
    ambitmap_addrs = list(range(start, end, 4))

    addrs = crtm_range_addrs + [CRTM_Registers.Device_Config, CRTM_Registers.Recovery_Block_Base_Addr, \
      CRTM_Registers.CRTM_Max_AutoMeasure_Size, CRTM_Registers.EEPROM_Attributes] + ambitmap_addrs

    for addr in addrs:
      # read original data
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      index = 16 + addr % 16
      old_data = response[index:(index + 4)]

      # set test data based on CRTM regsiter types, not all registers can be written with arbitrary data
      if addr == CRTM_Registers.Device_Config:
        test_data = b"\x00\x00\x00\x01"
      elif addr == CRTM_Registers.Recovery_Block_Base_Addr \
        or addr == CRTM_Registers.CRTM_Max_AutoMeasure_Size:
        test_data = b"\x00\x01\x00\x00"
      else:
        test_data = test_data_common

      # write test data
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)
            
      err, _ = device.meta_crtm_register_write(config, addr, test_data, sign)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      # read data back to verify
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      assert(response[index:(index + 4)] == test_data)

      # write original data back
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(config, addr, old_data, sign)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

def test_crtm_register_write_Positive_DualTypeVolatile():
  """
  Write to dual-type CRTM registers with config set to 0x01
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"
    sign = False
    test_data = b"\x00\x11\22\33"

    start = CRTM_Registers.CRTM_Range_Start_Addr_0
    end = start + 0x40
    addrs = list(range(start, end, 4))

    for addr in addrs:
      # read original data
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      index = 16 + addr % 16
      old_data = response[index:(index + 4)]

      # write test data
      err, _ = device.meta_crtm_register_write(config, addr, test_data, sign)
      assert(err.esr_and_response_success)

      # read data back to verify
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      assert(response[index:(index + 4)] == test_data)

      # write original data back
      err, _ = device.meta_crtm_register_write(config, addr, old_data, sign)
      assert(err.esr_and_response_success)

def test_crtm_register_write_Positive_EEPROM():
  """
  Write to EEPROM registers one at a time with corresponding EEPROM_Attributes bit set for modifiable
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    sign = True
    test_data = b"\x00\x11\22\33"

    # EEPROM_Attributes (read and modify) for each EEPROM
    attr_list = (
      b"\x00\x00\x00\x03", b"\x00\x00\x00\x30", b"\x00\x00\x03\x00", b"\x00\x00\x30\x00",
      b"\x00\x03\x00\x00", b"\x00\x30\x00\x00", b"\x03\x00\x00\x00", b"\x30\x00\x00\x00"
    )

    for i in range(0, 8): 
      # set EEPROM_Attributes
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, attr_list[i], True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)
       
      addr = CRTM_Registers.Customer_EEPROM + i * 32

      # read original data
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      index = 16 + addr % 16
      old_data = response[index:(index + 4)]

      # write test data
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)
            
      err, _ = device.meta_crtm_register_write(config, addr, test_data, sign)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      # read data back to verify
      err, response = device.meta_crtm_register_read(config, addr)
      assert(err.esr_and_response_success)
      assert(response[index:(index + 4)] == test_data)

      # write original data back
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(config, addr, old_data, sign)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

def test_crtm_register_write_Positive_VerifyAddrInResponse():
  """
  Verify the CRTM_Register_Address in response
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = CRTM_Registers.Customer_RAM
    data = b"\x00" * 4
    err, response = device.meta_crtm_register_write(b"\x00", addr, data, False)
    assert(err.esr_and_response_success)
    addr1 = int.from_bytes(response[4:6], byteorder="big")
    assert(addr1 == addr)

def test_crtm_register_write_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"
      addr = CRTM_Registers.KTR_Base
      data = b"\x00" * 4

      err, _ = device.meta_crtm_register_write(config, addr, data, False)
      assert(err.esr == b"\x00")

def test_crtm_register_write_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = CRTM_Registers.CRTM_Range_Start_Addr_0
    data = b"\x00" * 4

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_crtm_register_write(config, addr, data, True)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)

    device.test_only_set_boot_amtc0(None)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_crtm_register_write_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = CRTM_Registers.KTR_Base
    data = b"\x00" * 4

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_crtm_register_write(config, addr, data, False)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_crtm_register_write_Negative_InvalidAddr():
  """
  CRTM_Register_Address out of range (> 0xFFF)
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = 0xFFFF
    data = b"\x00" * 4

    err, _ = device.meta_crtm_register_write(config, addr, data, False)
    assert(err.esr == b"\x04")

def test_crtm_register_write_Negative_InvalidConfig():
  """
  Set config=`1b for registers other than CRTM_Addr_Size
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"
    addr = CRTM_Registers.KTR_Base
    data = b"\x00" * 4

    err, _ = device.meta_crtm_register_write(config, addr, data, False)
    assert(err.esr == b"\x04")

def test_crtm_register_write_Negative_RangeInvalid():
  """
  Write to non-volatile register without Update_Start
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = CRTM_Registers.CRTM_Range_Start_Addr_0
    data = b"\x00" * 4

    err, _ = device.meta_crtm_register_write(config, addr, data, True)
    assert(err.esr == b"\x04")

def test_crtm_register_write_Negative_VolatileWithSig():
  """
  Write to volatile register with signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"
    addr = CRTM_Registers.KTR_Base
    data = b"\x00" * 4

    err, _ = device.meta_crtm_register_write(config, addr, data, True)
    assert(err.esr == b"\x04")

def test_crtm_register_write_Negative_EEPROMNotModifiable():
  """
  EEPROM_Modify bit clear && write to EEPROM register
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # set EEPROM to not modifiable
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data = b"\x00\x00\x00\x00"

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    config = b"\x00"
    addr = CRTM_Registers.Customer_EEPROM

    err, _ = device.meta_crtm_register_write(config, addr, data, True)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)
    
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_crtm_register_write_Negative_AppendConfigNotModifiable():
  """
  AppendConfig can not be modified using register write.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data_write = b"\x11" * 16

    for i in range(0, len(data_write), 4):
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Append_Config_0 + i, \
        data_write[i:(i + 4)], True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Append_Config_0)
    assert(err.esr_and_response_success)

    assert(response[16:32] != data_write)