#! /usr/bin/env python3

"""
API Test for Deprovision_CRTM.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

from test.naxos.internal_no_boot_trutiphitest import response
import pytest

import authenta.naxos.utils as utils
import authenta.authenta_key_mgmt_interface as key_lib_key_mgmt
from authenta.naxos.utils import CRTM_Registers

import conftest

def write_test_data_to_eeprom():
  """
  Write test data to EEPROM_Attributes register and EEPROM
  """
  err, _ = device.meta_acrtm_update_start(b"\x00")
  assert(err.esr_and_response_success)

  data = b"\x11" * 4
  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  # write data to EEPROM
  data = b"\x00" * 4
  for addr in range(CRTM_Registers.Customer_EEPROM, CRTM_Registers.Customer_EEPROM + 0x100, 4):
    err, _ = device.meta_crtm_register_write(b"\x00", addr, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

  # write data to EEPROM_Attributes register
  data = b"\x00" * 4
  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  err, _ = device.meta_acrtm_update_finish(b"\x00")
  assert(err.esr_and_response_success)

@pytest.fixture(scope="module", autouse=True)
def test_deprovision_crtm_setup(request):
  global device
  global iboot_ip
  global uid
  global provision_stats

  device = device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (uid, _, _, provision_stats) = ret

  yield

"""
some deprovision tests are included in provision tests.
Here only add more deprovision tests
"""

def test_deprovision_crtm_Positive_AllowSRK():
  """
  Deprovision with Allow_SRK bit set:
  use active server root key as signing key
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    # provision for local use with TPM
    config = b"\x02"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, crk = device.read_ktr(verify_signature=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x02")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 1)

    # deprovision usins SRK as signing key
    err, _ = device.meta_deprovision_crtm(b"\x02")
    assert(err.esr_and_response_success)

def test_deprovision_crtm_Positive_DestroyEEPROMContent():
  """
  Deprovision with Destroy_EEPROM_Content bit set
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    # provision for remote use
    config = b"\x01"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 1)
    
    write_test_data_to_eeprom()

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x01")
    assert(err.esr_and_response_success)

    # provision for remote use
    config = b"\x01"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify EEPROM_Attributes set to `1b
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.EEPROM_Attributes)
    assert(err.esr_and_response_success)
    assert(response[16:20] == b"\x33" * 4)

    # verify EEPROM set to `1b
    for addr in range(CRTM_Registers.Customer_EEPROM, CRTM_Registers.Customer_EEPROM + 0x100, 16):
      err, response = device.meta_crtm_register_read(b"\x00", addr)
      assert(err.esr_and_response_success)
      assert(response[16:32] == b"\xFF" * 16)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x01")
    assert(err.esr_and_response_success)

def test_deprovision_crtm_Positive_NotDestroyEEPROMContent():
  """
  Deprovision with Destroy_EEPROM_Content bit clear
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    # provision for remote use
    config = b"\x01"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 1)
    
    write_test_data_to_eeprom()

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

    # provision for remote use
    config = b"\x01"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify EEPROM_Attributes not changed
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.EEPROM_Attributes)
    assert(err.esr_and_response_success)
    assert(response[16:20] == b"\x00" * 4)

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data = b"\x22" * 4
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # verify EEPROM not changed
    for addr in range(CRTM_Registers.Customer_EEPROM, CRTM_Registers.Customer_EEPROM + 0x100, 16):
      err, response = device.meta_crtm_register_read(b"\x00", addr)
      assert(err.esr_and_response_success)
      assert(response[16:32] == b"\x00" * 16)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x01")
    assert(err.esr_and_response_success)

def test_deprovision_crtm_Negative_DisableDeprovision():
  """
  Local provision without TPM Present
  Set Disable_Deprovision bit and try to deprovision
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x00"
    err, response, sent_nonce = device.meta_provision_crtm(config)
    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_signature=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)
    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # get Device_Config_Volatile
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    device_config = response[24:28]
    
    device_config_int = int.from_bytes(device_config, byteorder="big")
    data_int = device_config_int | 0x00000008
    data = data_int.to_bytes(4, byteorder="big")

    # set Disable_Deprovision bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)

    # try to deprovision
    # ESR check fails on 1.4, don't exit
    flag = True
    err, _ = device.meta_deprovision_crtm(b"\x00")
    if err.esr != b"\x20":
      flag = False

  # poweron to clear CSK_Read_Disable bit
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

  assert(flag)

def test_deprovision_crtm_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"

      err, _ = device.meta_deprovision_crtm(config)
      assert(err.esr == b"\x00")

def test_deprovision_crtm_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  # change SRK in the key xml file to get incorrect signature
  ret = key_lib_key_mgmt.set_new_srk_key_data_for_uid(uid.hex(), conftest.test_key_handle)
  assert(ret)
  
  with device:

    config = b"\x00"

    err, _ = device.meta_deprovision_crtm(config)
    if err.esr != b"\x04":
      key_lib_key_mgmt.set_new_srk_key_data_for_uid(uid.hex(), conftest.srk_handle)
      assert(False)

    # change SRK back
    ret = key_lib_key_mgmt.set_new_srk_key_data_for_uid(uid.hex(), conftest.srk_handle)
    assert(ret)

def test_deprovision_crtm_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  with device:
    config = b"\x00"

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc1(b"\x00\x00\x00\x00")

    err, _ = device.meta_deprovision_crtm(config)
    if err.esr != b"\x10":
      device.test_only_set_amtc1(None)
      assert(False)

    device.test_only_set_amtc1(None)