#! /usr/bin/env python3

"""
API Test for Flash_Extend.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_flash_extend_setup(request):
  global device
  global provision_stats

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret
     
    if provision_stats.Provisioned == 1:
      # read current device config
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      device_config = response[16:20]
      device_config_int = int.from_bytes(device_config, byteorder="big")

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # turn off recovery mode
      data_int = device_config_int & 0xFFFFFFFD
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      if err.esr != b"\x80":
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set measurement range
      ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xe00000, 0x2000)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set max automeasure size
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, 0x1000.to_bytes(4, 'big'), True)
      if err.esr != b"\x80":
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

  yield

def test_flash_extend_Positive_Measure_LinkedList():
  """
  Option #0
  Start digest calculation at the beginning of the linked list
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\00"

    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    status = device.extended_crtm_status_struct
    assert(status.Partial_Measurement == 0)

def test_flash_extend_Positive_Measure_CrtmMaxAutoMeasSize():
  """
  Option #0
  Start digest calculation at CRTM_Max_AutoMeasure_Size
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\02"

    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    status = device.extended_crtm_status_struct
    assert(status.Partial_Measurement == 0)


def test_flash_extend_Positive_ReprgmNvDgsts():
  """
  Option #1
  Reprogram the both Flash_DigestNV and Flash_DigestNV_Partial registers
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\01"

    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    status = device.extended_crtm_status_struct
    assert(status.Partial_Measurement == 0)
    assert(status.CRTM_Mismatch == 0)

def test_flash_extend_Positive_CkRcvryBlk():
  """
  Option #2
  Check the integrity of the recovery block contents
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # reprogram Recovery_DigestNV register
    config = b"\05"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    # check integrity of the recovery block contents
    config = b"\04"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    status = device.extended_crtm_status_struct
    assert(status.Recovery_Mismatch == 0)

def test_flash_extend_Positive_ReprgmRcvryBlkNvDgst():
  """
  Option #3
  Check the integrity of the recovery block contents
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\04"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    status = device.extended_crtm_status_struct
    assert(status.Recovery_Mismatch == 0)

def test_flash_extend_Positive_IgnoreVolatileRegs():
  """
  Flash_Extend will ignore volatile registers state.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # set volatile CRTM Addr0 and Size0 registers to 0
    # if Flash_Extend uses the volatile registers, it will fail
    ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, 0x0)
    assert(ret)

    # option 0
    config = b"\00"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    config = b"\02"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    # option 1
    config = b"\01"
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)


def test_flash_extend_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"

      err, _ = device.meta_flash_extend(config)
      assert(err.esr != b"\x80")

def test_flash_extend_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_flash_extend(config)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)  

    device.test_only_set_boot_amtc0(None)

def test_flash_extend_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_flash_extend(config)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_flash_extend_Negative_RangeInvalid():
  """
  CRTM_Range_Invalid is set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    # set CRTM_Range_Invalid bit
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_flash_extend(config)
    if err.esr != b"\x10":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_flash_extend_Negative_RecoveryBlock():
  """
  Recovery_Block bit is set
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    # set Recovery_Block bit
    ret = utils.enable_automeasure(device)
    assert(ret)

    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    err, _ = device.meta_flash_extend(config)
    if err.esr != b"\x10":
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    ret = utils.disable_automeasure(device)
    assert(ret)

def test_flash_extend_Negative_InvalidAddrSize():
  """
  a size of 0x00000000 is read in the CRTM_Size[0] register
  || CRTM_Addr[0] is outside of the main array address space
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # read current CRTM_Addr0 and CRTM_Size0
    err, response = device.meta_crtm_register_read(b"\00", CRTM_Registers.CRTM_Range_Start_Addr_0)
    assert(err.esr_and_response_success)
    crtm_addr0 = response[16:20]
    crtm_size0 = response[20:24]

    # set CRTM_Addr0 to invalid address
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_write(b"\00", CRTM_Registers.CRTM_Range_Start_Addr_0, \
      b"\xFF" * 4, True)
    if err.esr != b"\x80":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_flash_extend(b"\x00")
    assert err.esr == b"\x10"

    # revert CRTM_Addr0
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_write(b"\00", CRTM_Registers.CRTM_Range_Start_Addr_0, \
      crtm_addr0, True)
    if err.esr != b"\x80":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # set CRTM_Size0 to 0
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_write(b"\00", CRTM_Registers.CRTM_Range_Size_0, \
      b"\x00" * 4, True)
    if err.esr != b"\x80":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_flash_extend(b"\x00")
    assert err.esr == b"\x10"

    # revert CRTM_Addr0
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_write(b"\00", CRTM_Registers.CRTM_Range_Size_0, \
      crtm_size0, True)
    if err.esr != b"\x80":
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
