#! /usr/bin/env python3

"""
API Test for Measure.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers
import authenta.authenta_key_mgmt_interface as key_mgmt
from hashlib import sha256

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_measure_setup(request):
  global device
  global provision_stats
  global iboot_ip
  global uds_disabled

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      # read current device config
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      # read uds_disable bit
      uds_disabled = device.extended_crtm_status_struct.UDS_Disabled

      device_config = response[16:20]
      device_config_int = int.from_bytes(device_config, byteorder="big")

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # turn off recovery mode
      data_int = device_config_int & 0xFFFFFFFD
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set measure range (CRTM Addr and Size volatile registers)
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, 0x100000)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set EEPROM attribute to ModifyAndRead
      data = b"\x33" * 4
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

  yield

def test_measure_Positive_Dest_KTR():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest only
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_Dest_EEPROM():
  """
  Option #0
  MeasureDestination: EEPROM
  KeySource: create a digest only
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # try all 8 addresses
    config = b"\x00\x80"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
 
    for i in range(0, 8):
      eeprom_dest_addr = CRTM_Registers.Customer_EEPROM + 32 * i
      err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

def test_measure_Positive_AltSrc_KTR():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest only
  AltSource source: KTR
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x02\x00"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_AltSrc_EEPROM():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest only
  AltSource source: EEPROM
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # try all 8 addresses
    config = b"\x03\x00"
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    for i in range(0, 8):
      eeprom_alt_src_addr = CRTM_Registers.Customer_EEPROM + 32 * i
      err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

def test_measure_Positive_KeySrc_NoConstants():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest (no HMAC) without constants in calculation
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x08"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_KeySrc_KTR():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create an HMAC using the contents of the KTR as the secret key
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x20"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_KeySrc_Ksn():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create an HMAC using the currently active session key as the secret key
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x28"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_KeySrc_Uds():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create an HMAC using UDS as the secret key
  """
  if provision_stats.Provisioned == 0 or provision_stats.RIOT_Mode == 0 or uds_disabled == 1:
    return

  with device:
    config = b"\x00\x30"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_KeySrc_EEPROM():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create an HMAC using the contents of the specified EEPROM as the secret key
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # try all 8 addresses
    config = b"\x00\x38"
    eeprom_alt_src_addr = 0x0
    eeprom_dest_addr = 0x0

    for i in range(0, 8):
      eeprom_key_src_addr = CRTM_Registers.Customer_EEPROM + 32 * i
      err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

def test_measure_Positive_IncludeMetadata():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest only
  IncludeMetadata: `1b
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x04\x00"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_InputDigest():
  """
  Option #0
  MeasureDestination: KTR
  KeySource: create a digest only
  DigestProvided: `1b
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x40"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    in_digest = b"\x01" * 32
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr, \
      in_digest=in_digest)
    assert(err.esr_and_response_success)

def test_measure_Positive_GetFDS():
  """
  Option #1
  """
  if provision_stats.Provisioned == 0 or uds_disabled == 1 \
    or provision_stats.RIOT_Mode == 0:
    return

  with device:
    config = b"\x00\x01"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

def test_measure_Positive_DisableUDSAccess():
  """
  Option #2
  """
  if provision_stats.Provisioned == 0 or provision_stats.RIOT_Mode == 0:
    return

  # 17.B2 bug
  failed = False

  with device:
    config = b"\x00\x02"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    # verify status
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)
    assert(device.extended_crtm_status_struct.UDS_Disabled == 1)

    # try to get FDS
    err, _ = device.meta_measure(b"\x00\x01", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      failed = True

      # read KTR - make sure it is 0
      digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest == b"\x00" * 32)

    # try option #2 multiple times - should be done without error
    for i in range(0, 10):
      err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      err, _ = device.meta_measure(b"\x00\x01", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      if err.esr != b"\x20":
        failed = True

        # read KTR - make sure it is 0
        digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
        assert(digest == b"\x00" * 32)

  # reboot to enable UDS access
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  assert(not failed)

  with device:
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)
    assert(device.extended_crtm_status_struct.UDS_Disabled == 0)

def test_measure_Positive_FDSOnlyAtPoweron():
  """
  FDS_Metadata_Digest is only calculated at poweron and not immediately
  after ProvisioningConfig[1:0] == `11b
  This test will provision the device and get FDS immedeately
  The result should be different than the FDS obtained after poweron
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    # provision device
    config = b"\x05"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # get FDS
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    config = b"\x00\x01"

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    fds = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(fds)

  # reboot
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)
  
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # get FDS
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    fds1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(fds1)

    assert(fds != fds1)

    # deprovision device
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_measure_Positive_IgnoreNVRegs():
  """
  The Measure command shall always operate on the registers’ current volatile state, not the NV state.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    v_result = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(v_result)

    # set NV CRTM_Addr and CRTM_Size to different values than volatile ones
    # should get same measure results

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0x0, 0x1000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    nv_result = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(nv_result)
    assert(nv_result == v_result)

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0x0, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    nv_result = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(nv_result)
    assert(nv_result == v_result)

def test_measure_Positive_ClearKTRFlags():
  """
  All KTR flags shall be cleared by Measure command.
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  # 17.FF bug
  failed = False

  with device:
    key_data_to_ktr = b"\x00" * 32

    # set KTR flags by writing data to KTR
    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      key_data_to_ktr, False)
    assert(ret)

    # option 0
    config = b"\x00\x80"

    err, _ = device.meta_measure(config, 0x0, 0x0, CRTM_Registers.Customer_EEPROM)
    assert(err.esr_and_response_success)

    # after measure command, KTR flags should be cleared
    # and replace_srk command should return ESR 5 after this
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    if provision_stats.RIOT_Mode == 1:
      # set KTR flags by writing data to KTR
      ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
        key_data_to_ktr, False)
      assert(ret)

      # option 2
      config = b"\x00\x02"

      err, _ = device.meta_measure(config, 0x0, 0x0, 0x0)
      assert(err.esr_and_response_success)

      new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      new_key_bytes = bytes.fromhex(new_key)
      new_key_hash = sha256(new_key_bytes).digest()
      err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
      if err.esr != b"\x20":
        failed = True

  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  assert(not failed)

def test_measure_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00\x00"
      eeprom_alt_src_addr = 0x0
      eeprom_key_src_addr = 0x0
      eeprom_dest_addr = 0x0

      err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr != b"\x80")

def test_measure_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x80"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = CRTM_Registers.Customer_EEPROM

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)

    device.test_only_set_boot_amtc0(None)

def test_measure_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00\x00"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_measure_Negative_InvalidCmdSize():
  """
  For option #0, Config is set to include optional param (Input_Digest) but command does not.
  For option #1 and #2, optional param exists
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    # option #0
    config = b"\x00\x40"

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr != b"\x80")

    # option #1
    config = b"\x00\x41"
    in_digest = b"\x00" * 32

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr, in_digest=in_digest)
    assert(err.esr == b"\x04")

    # option #2
    config = b"\x00\x42"
    in_digest = b"\x00" * 32
    flag = True

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr, in_digest=in_digest)
    assert(err.esr == b"\x04")

def test_measure_Negative_RIoTModeDisabled():
  """
  Provision device with RIoT mode disabled and call Measure
  option #0 (use UDS as KeySource) and option #1.
  """
  if provision_stats.Provisioned == 0:
    # 1.5 B2 Bug
    flag = False

    with device:
      # provision device with RIoT mode disabled
      err, *_ = device.meta_provision_crtm(b"\x01")
      assert(err.esr_and_response_success)

      eeprom_alt_src_addr = 0x0
      eeprom_key_src_addr = 0x0
      eeprom_dest_addr = 0x0

      # deprovision device on failure
      err, _ = device.meta_measure(b"\x00\x30", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      if err.esr != b"\x20":
        flag = True
        
        # read KTR - make sure it is 0
        digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
        assert(digest == b"\x00" * 32)

      err, _ = device.meta_measure(b"\x00\x01", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      if err.esr != b"\x20":
        flag = True
        
        # read KTR - make sure it is 0
        digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
        assert(digest == b"\x00" * 32)

      err, _ = device.meta_measure(b"\x00\x02", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      if err.esr != b"\x20":
        flag = True

      if not flag:
        # deprovision
        err, _ = device.meta_deprovision_crtm(b"\x00")
        assert(err.esr_and_response_success)

    if flag:
      # 1.5 B2 bug causes chip in bad state
      # need to reboot to run remaining tests
      ret = conftest.reboot(device, iboot_ip)
      assert(ret)

      with device:
        ret = utils.get_dev_info(device)
        assert(ret != None)

        err, _ = device.meta_deprovision_crtm(b"\x00")
        assert(err.esr_and_response_success)
  
    assert(not flag)

def test_measure_Negative_RecoveryBlock():
  """
  RecoveryBlock bit is set.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    ret = utils.enable_automeasure(device)
    assert(ret)
    
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    err, _ = device.meta_measure(b"\x00\x00", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")

    err, _ = device.meta_measure(b"\x00\x01", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")

    err, _ = device.meta_measure(b"\x00\x02", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      device.meta_acrtm_update_finish(b"\x01")

    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    ret = utils.disable_automeasure(device)
    assert(ret)

def test_measure_Negative_UDSDisabled():
  """
  For option #0 (use UDS as KeySource) and option #1, UDS is disabled.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    # 17.B2 bug
    failed = False

    # disable UDS
    err, _ = device.meta_measure(b"\x00\x02", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr_and_response_success)

    err, _ = device.meta_measure(b"\x00\x30", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      failed = True
    
      # read KTR - make sure it is 0
      digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest == b"\x00" * 32)

    err, _ = device.meta_measure(b"\x00\x01", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    if err.esr != b"\x20":
      failed = True

      # read KTR - make sure it is 0
      digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(digest == b"\x00" * 32)

  # reboot to enable UDS access
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

  assert(not failed)

def test_measure_Negative_InvalidEEPROMAddr_AltSource():
  """
  Option #0 and invalid EEPROM_AltSource_Addr
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0

    err, _ = device.meta_measure(b"\x03\x00", 0xF4, eeprom_key_src_addr, eeprom_dest_addr)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x03\x00", 0xF0, eeprom_key_src_addr, eeprom_dest_addr)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x00\x00", 0xF4, eeprom_key_src_addr, eeprom_dest_addr)
    assert err.esr == b"\x04"

def test_measure_Negative_InvalidEEPROMAddr_KeySource():
  """
  Option #0 and invalid EEPROM_KeySource_Addr
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_dest_addr = 0x0

    err, _ = device.meta_measure(b"\x00\x38", eeprom_alt_src_addr, 0xF4, eeprom_dest_addr)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x00\x38", eeprom_alt_src_addr, 0xF0, eeprom_dest_addr)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x00\x00", eeprom_alt_src_addr, 0xF4, eeprom_dest_addr)
    assert err.esr == b"\x04"

def test_measure_Negative_InvalidEEPROMAddr_Dest():
  """
  Option #0 and invalid EEPROM_Dest_Addr
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0

    err, _ = device.meta_measure(b"\x00\x80", eeprom_alt_src_addr, eeprom_key_src_addr, 0xF4)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x00\x80", eeprom_alt_src_addr, eeprom_key_src_addr, 0xF0)
    assert err.esr == b"\x04"

    err, _ = device.meta_measure(b"\x00\x00", eeprom_alt_src_addr, eeprom_key_src_addr, 0xF4)
    assert err.esr == b"\x04"

def test_measure_Negative_InvalidConfig():
  """
  Config[1:0] == `11b
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = 0x0
    config = b"\x00\x03"

    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert err.esr == b"\x20"

def test_measure_Negative_EEPROMNotModifiable():
  """
  EEPROM_Modify bit clear && write measurement result to EEPROM register
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # read current EEPROM attribute
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.EEPROM_Attributes)
    assert(err.esr_and_response_success)
    assert(response is not None)
    cur_eeprom_attr = response[16:20]

    # set EEPROM to not modifiable
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data = b"\x00\x00\x00\x00"

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # Measure with destination set to EEPROM
    config = b"\x00\x80"
    eeprom_alt_src_addr = 0x0
    eeprom_key_src_addr = 0x0
    eeprom_dest_addr = CRTM_Registers.Customer_EEPROM
    
    err, _ = device.meta_measure(config, eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
    assert(err.esr == b"\x20")
    
    # revert EEPROM_Attributes back
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, cur_eeprom_attr, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_measure_Negative_PriorToCSKReadDisable():
  """
  Provision device for local use with no TPM 
  && set destination to KTR 
  && issue Measure command prior to setting CSK_Read_Disable bit
  """
  if provision_stats.Provisioned == 0:
    with device:
      # provision device for local use with no TPM
      err, response, sent_nonce = device.meta_provision_crtm(b"\x00")

      # for local provisioning provision verification must be done seperately
      # after adding the CRK or CSK in this case to the key repository
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err, csk = device.read_ktr(verify_signature=False)
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      # now make the csk available to the key driver
      if not key_mgmt.add_key(device.uid_as_str, key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

      assert(err.esr_and_response_success)

      eeprom_alt_src_addr = 0x0
      eeprom_key_src_addr = 0x0
      eeprom_dest_addr = 0x0

      err, _ = device.meta_measure(b"\x00\x00", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr == b"\x04")

      # deprovision
      err, _ = device.meta_deprovision_crtm(b"\x00")
      assert(err.esr_and_response_success)

def test_measure_Negative_PriorToCRKReadDisable():
  """
  Provision device for local use with TPM 
  && set destination to KTR 
  && issue Measure command prior to setting CRK_Read_Disable bit
  """
  if provision_stats.Provisioned == 0:
    with device:
      # provision device for local use with no TPM
      err, response, sent_nonce  = device.meta_provision_crtm(b"\x02")

      # for local provisioning provision verification must be done seperately
      # after adding the CRK or CSK in this case to the key repository
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err, csk = device.read_ktr(verify_signature=False)
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      # now make the csk available to the key driver
      if not key_mgmt.add_key(device.uid_as_str, key_mgmt.KeyMgmt_KeyName_CSK, csk, None):
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

      assert(err.esr_and_response_success)

      eeprom_alt_src_addr = 0x0
      eeprom_key_src_addr = 0x0
      eeprom_dest_addr = 0x0

      err, _ = device.meta_measure(b"\x00\x00", eeprom_alt_src_addr, eeprom_key_src_addr, eeprom_dest_addr)
      assert(err.esr == b"\x04")

      # deprovision
      err, _ = device.meta_deprovision_crtm(b"\x00")
      assert(err.esr_and_response_success)
