#! /usr/bin/env python3

"""
API Test for Provision_CRTM.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers
import authenta.authenta_key_mgmt_interface as key_lib_key_mgmt
import authenta.authenta_key_lib_interface as key_lib

import conftest

def get_flash_digest_no_sig_check(device, addr):
  """
  Get flash digest value (no signature check)
  return digest on success or None on failure
  """
  digest = None
  config = b"\x00"

  err, response = device.meta_crtm_register_read(config, addr)
  if err.esr != b"\x80":
    print("ERROR: Failed to read data from CRTM register 0x%X" %(addr))
    return digest

  digest_l = utils.change_word_order(response[16:32])

  addr += 16
  err, response = device.meta_crtm_register_read(config, addr)
  if err.esr != b"\x80":
    print("ERROR: Failed to read data from CRTM register 0x%X" %(addr))
    return digest

  digest_h = utils.change_word_order(response[16:32])

  digest = b"".join((digest_h, digest_l))
  assert(len(digest) == 32)

  return digest

@pytest.fixture(scope="module", autouse=True)
def test_provision_crtm_setup(request):
  global device
  global iboot_ip
  global uid
  global provision_stats

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (uid, _, _, provision_stats) = ret

  yield

def test_provision_crtm_Positive_Remote():
  """
  Remote provision
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x01"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 1)

    # verify extended CRTM status
    extended_status = device.extended_crtm_status_struct
    assert(extended_status.CRTM_Mismatch == 1)
    assert(extended_status.CRTM_Range_Invalid == 1)
    assert(extended_status.CRTM_PCR0_Invalid == 1)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Remote_EnableRiot():
  """
  Remote provision with RIoT mode Enabled
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x05"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 1)
    assert(stats.RIOT_Mode == 1)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Remote_PoweronOnlyFlashPcr0Reset():
  """
  Remote provision with Flash_PCR0_Reset bit set
  Only a poweron reset shall reset Flash_PCR0;
  pin reset can't
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x09"
    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr_and_response_success)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 1)

    # Now not able to verify that the pin reset can't reset FlashPcr0
    # because pin reset is not working

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_TPM():
  """
  Local provision with TPM Present
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x02"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, crk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      client_root_key = get_flash_digest_no_sig_check(device, CRTM_Registers.KTR_Base)
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 1)

    # get client root key
    client_root_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_root_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_TPM_EnableRiot():
  """
  Local provision with TPM Present and RIoT mode Enabled
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x06"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, crk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      client_root_key = get_flash_digest_no_sig_check(device, CRTM_Registers.KTR_Base)
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 1)
    assert(stats.RIOT_Mode == 1)

    # get client root key
    client_root_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_root_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_TPM_PoweronOnlyFlashPcr0Reset():
  """
  Local provision with TPM Present and Flash_PCR0_Reset bit set
  Only a poweron reset shall reset Flash_PCR0;
  pin reset can't
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x0a"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, crk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CRK, crk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      client_root_key = get_flash_digest_no_sig_check(device, CRTM_Registers.KTR_Base)
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 1)

    # Now not able to verify that the pin reset can't reset FlashPcr0
    # because pin reset is not working

    # get client root key
    client_root_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_root_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_NoTPM():
  """
  Local provision without TPM Present
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x00"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_NoTPM_EnableRiot():
  """
  Local provision without TPM Present and with RIoT mode Enabled
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x04"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)
    assert(stats.RIOT_Mode == 1)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_NoTPM_PoweronOnlyFlashPcr0Reset():
  """
  Local provision without TPM Present and with Flash_PCR0_Reset bit set
  Only a poweron reset shall reset Flash_PCR0;
  pin reset can't
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x08"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)

    # Now not able to verify that the pin reset can't reset FlashPcr0
    # because pin reset is not working

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Positive_Local_NoTPM_ReadSessionKeyAfterPoweron():
  """
  Local provision without TPM Present
  Read client session key after poweron
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x00"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)

  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    ret = utils.get_dev_info(device)
    assert(ret != None)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Negative_Local_NoTPM_DisableCSKRead():
  """
  Local provision without TPM Present
  Set CSK_Read_Disable and try to read client session key
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x00"
    err, response, sent_nonce = device.meta_provision_crtm(config)

    # for local provisioning provision verification must be done seperately
    # after adding the CRK or CSK in this case to the key repository
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err, csk = device.read_ktr(verify_sig=False)
    if not err.esr_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    # now make the csk available to the key driver
    if not key_lib_key_mgmt.add_key(device.uid_as_str, key_lib_key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
      device.meta_deprovision_crtm(b"\x00")
      assert(False)
    err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

    if not err.esr_and_response_success:
      device.meta_deprovision_crtm(b"\x00")
      assert(False)

    # verify provision state
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.Provisioned == 1)
    assert(stats.Remotely_Provisioned == 0)
    assert(stats.TPM_Supported == 0)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != None)

    # get Device_Config_Volatile
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    device_config = response[24:28]
    
    device_config_int = int.from_bytes(device_config, byteorder="big")
    data_int = device_config_int | 0x00000004
    data = data_int.to_bytes(4, byteorder="big")

    # set CSK_Read_Disable bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)

    # try to get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key == b"\x00" * 32)

  # poweron to clear CSK_Read_Disable bit
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # get client session key
    client_session_key = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(client_session_key != b"\x00" * 32)

    # deprovision
    err, _ = device.meta_deprovision_crtm(b"\x00")
    assert(err.esr_and_response_success)

def test_provision_crtm_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    # change SRK in the key xml file to get incorrect signature
    test_key = key_lib_key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    assert(test_key)
    ret = key_lib.set_hmac_root_signing_key(device._keylib_context, bytes.fromhex(test_key))
    assert(ret)
    ret = key_lib_key_mgmt.copy_key(device.uid_as_str, 'srk', 'srk_copy')
    assert(ret)
    ret = key_lib_key_mgmt.delete_key(device.uid_as_str, 'srk')
    assert(ret)
    ret = key_lib_key_mgmt.add_key(device.uid_as_str, 'srk', bytes.fromhex(test_key), None)
    assert(ret)

    config = b"\x01"

    err, *_ = device.meta_provision_crtm(config)
    if err.esr != b"\x04":
      ret = key_lib_key_mgmt.delete_key(device.uid_as_str, 'srk')
      assert(ret)
      ret = key_lib_key_mgmt.copy_key(device.uid_as_str, 'srk_copy', 'srk')
      assert(ret)
      ret = key_lib_key_mgmt.delete_key(device.uid_as_str, 'srk_copy')
      assert(ret)
      get_srk = key_lib_key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'srk')
      assert(get_srk)
      ret = key_lib.set_hmac_root_signing_key(device._keylib_context, bytes.fromhex(get_srk))
      assert(ret)
      assert(False)

    # change SRK back
    ret = key_lib_key_mgmt.delete_key(device.uid_as_str, 'srk')
    assert(ret)
    ret = key_lib_key_mgmt.copy_key(device.uid_as_str, 'srk_copy', 'srk')
    assert(ret)
    ret = key_lib_key_mgmt.delete_key(device.uid_as_str, 'srk_copy')
    assert(ret)
    get_srk = key_lib_key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'srk')
    assert(get_srk)
    ret = key_lib.set_hmac_root_signing_key(device._keylib_context, bytes.fromhex(get_srk))
    assert(ret)

def test_provision_crtm_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 1:
    return

  with device:
    config = b"\x01"

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    err, *_ = device.meta_provision_crtm(config)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_provision_crtm_Negative_AlreadyProvisioned():
  """
  Already provisioned
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x01"

    err, *_ = device.meta_provision_crtm(config)
    assert(err.esr == b"\x02")
