#! /usr/bin/env python3

"""
API Test for Replace_SRK.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest
import hmac
import hashlib

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers
import authenta.authenta_key_mgmt_interface as key_mgmt

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_replace_srk_setup(request):
  global device
  global iboot_ip
  global uid
  global provision_stats
  global uds_disabled

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (uid, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      # read uds_disable bit
      uds_disabled = device.extended_crtm_status_struct.UDS_Disabled

  yield

def test_replace_srk_Positive_SrkIsMicronKey():
  """
  When SRK is micron default SRK,
  run option #0 to replace SRK with a thrid-party one
  and run option #1 to revert it back
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.SRK_Replaced == 1:
    return

  with device:
    # option #0
    config = b"\x00"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.SRK_Replaced == 1)

    # verify SRK
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)
 
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # option #1
    config = b"\x01"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='micron_srk_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.SRK_Replaced == 0)

    # verify SRK
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)
 
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)


def test_replace_srk_Positive_SrkIsThirdPartyKey():
  """
  When SRK is third-party SRK,
  run option #0 to replace SRK with another thrid-party one
  and run option #0 to revert it back
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.SRK_Replaced == 0:
    return

  with device:
    # option #0
    config = b"\x00"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.SRK_Replaced == 1)

    # verify SRK
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)
 
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # option #0
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='srk_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.SRK_Replaced == 1)

    # verify SRK
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)
 
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_replace_srk_Positive_UdsIsMicronKey():
  """
  When UDS is micron default UDS
  run option #2 to replace UDS with a thrid-party one
  and run option #3 to revert it back
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or provision_stats.UDS_Replaced == 1 \
    or uds_disabled == 1:
    return

  with device:
    # option #2
    config = b"\x02"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle', \
      old_key_name='micron_uds_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.UDS_Replaced == 1)

    # option #3
    config = b"\x03"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name="micron_uds_handle")
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.UDS_Replaced == 0)

def test_replace_srk_Positive_UdsIsThirdPartyKey():
  """
  When UDS is third-party UDS
  run option #2 to replace UDS with another thrid-party one
  and run option #2 to revert it back
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or provision_stats.UDS_Replaced == 0 \
    or uds_disabled == 1:
    return

  with device:
    # option #2
    config = b"\x02"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle', \
      old_key_name='uds_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.UDS_Replaced == 1)

    # option #2
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='uds_handle', \
      old_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.UDS_Replaced == 1)

def test_replace_srk_Positive_DeriveKey():
  """
  Derive_Direct# is set
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or uds_disabled == 1:
    return
  
  with device:
    # option #0
    config = b"\x10"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.SRK_Replaced == 1)

    if provision_stats.SRK_Replaced == 1:
      config = b"\x00"
      ret = conftest.replace_srk_util(device, uid, config, new_key_name='srk_handle')
      assert(ret)
    else:
      config = b"\x01"
      ret = conftest.replace_srk_util(device, uid, config, new_key_name='micron_srk_handle')
      assert(ret)

    # option #2
    config = b"\x12"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle', \
      old_key_name='uds_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)
    stats = device.provision_sts_struct
    assert(stats.UDS_Replaced == 1)

    # derive uds
    # UDS(new) = HMAC(KTR)UDS(existing)
    test_key_data = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    assert(test_key_data is not None)

    uds_data = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'uds_handle')
    assert(uds_data is not None)

    derived_uds_data = hmac.new(bytes.fromhex(uds_data), bytes.fromhex(test_key_data), hashlib.sha256).digest()
    assert(derived_uds_data is not None)

    # save derived uds to XML
    derived_ret = key_mgmt.add_key(device.uid_as_str, "new_uds", derived_uds_data, None)
    assert(derived_ret)

    config = b"\x02"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='uds_handle', \
      old_key_name='new_uds')
    if not ret:
      key_mgmt.delete_key(device.uid_as_str, 'new_uds')
      assert(ret)

    # delete derived uds from XML
    ret = key_mgmt.delete_key(device.uid_as_str, 'new_uds')
    assert(ret)

    if provision_stats.UDS_Replaced == 0:
      config = b"\x03"
      ret = conftest.replace_srk_util(device, uid, config, new_key_name="micron_uds_handle")
      assert(ret)

def test_replace_srk_Positive_SessionKey():
  """
  Session_Root# is set
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  with device:
    # not derive key
    config = b"\x20"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    # derive key
    config = b"\x30"
    ret = conftest.replace_srk_util(device, uid, config, new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

def test_replace_srk_Positive_KTRCleared():
  """
  Option #0, #1, #2 and #3: Check that the KTR is cleared after the command.
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or provision_stats.SRK_Replaced == 1 \
    or provision_stats.UDS_Replaced == 1 or uds_disabled == 1:
    return

  with device:
    # option #0
    ret = conftest.replace_srk_util(device, uid, b"\x00", new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest == b"\x00" * 32)

    # option #1
    ret = conftest.replace_srk_util(device, uid, b"\x01", new_key_name='micron_srk_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest == b"\x00" * 32)

    # option #2
    ret = conftest.replace_srk_util(device, uid, b"\x02", \
      new_key_name='test_key_handle', \
      old_key_name='micron_uds_handle')
    assert(ret)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest == b"\x00" * 32)

    # option #3
    ret = conftest.replace_srk_util(device, uid, b"\x03", new_key_name="micron_uds_handle")
    assert(ret)

    digest = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(digest == b"\x00" * 32)

def test_replace_srk_Positive_KTRWriteFlagCleared():
  """
  Option #0, #1, #2, #3: Check that the KTR register write flags are cleared after the command.
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or provision_stats.SRK_Replaced == 1 \
    or provision_stats.UDS_Replaced == 1 or uds_disabled == 1:
    return

  with device:
    # option #0
    ret = conftest.replace_srk_util(device, uid, b"\x00", new_key_name='test_key_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    # call replace_srk again should return ESR 5 because KTR is cleared
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    # option #1
    ret = conftest.replace_srk_util(device, uid, b"\x01", new_key_name='micron_srk_handle')
    assert(ret)

    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    # call replace_srk again should return ESR 5 because KTR is cleared
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    # option #2
    ret = conftest.replace_srk_util(device, uid, b"\x02", \
      new_key_name='test_key_handle', \
      old_key_name='micron_uds_handle')
    assert(ret)

    # call replace_srk again should return ESR 5 because KTR is cleared
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

    # option #3
    ret = conftest.replace_srk_util(device, uid, b"\x03", new_key_name='micron_uds_handle')
    assert(ret)

    # call replace_srk again should return ESR 5 because KTR is cleared
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(b"\x00", additional_data=new_key_hash)
    assert(err.esr == b"\x20")

def test_replace_srk_Negative_LocalProvision():
  """
  Provision device for local use and call Replace_SRK
  """
  if provision_stats.Provisioned == 0:
    with device:
      # provision device for local use
      err, response, sent_nonce = device.meta_provision_crtm(b"\x00")

      # for local provisioning provision verification must be done seperately
      # after adding the CRK or CSK in this case to the key repository
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err, csk = device.read_ktr(verify_signature=False)
      if not err.esr_success:
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      # now make the csk available to the key driver
      if not key_mgmt.add_key(device.uid_as_str, key_mgmt.KeyMgmt_KeyName_CSK, csk, None, True, True):
        device.meta_deprovision_crtm(b"\x00")
        assert(False)
      err = device.verify_provision_crtm_response_for_local_provision(response, sent_nonce)

      assert(err.esr_and_response_success)

      # option #0
      config = b"\x00"
      new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      new_key_bytes = bytes.fromhex(new_key)
      new_key_hash = hashlib.sha256(new_key_bytes).digest()
      err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
      assert(err.esr == b"\x04")

      # deprovision
      err, _ = device.meta_deprovision_crtm(b"\x00")
      assert(err.esr_and_response_success)

def test_replace_srk_Negative_KTRCleared():
  """
  Verify KTR is cleared when command fails
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0:
    return

  with device:
    key_data_to_ktr = b"\xFF" * 32

    # write data to KTR that does not match the new key
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == key_data_to_ktr)

    # option #0
    config = b"\x00"
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    assert(err.esr != b"\x80")

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == b"\x00" * 32)

    # write data to KTR that is not 0
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == key_data_to_ktr)

    # option #1
    config = b"\x01"
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'micron_srk_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    assert(err.esr != b"\x80")

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == b"\x00" * 32)

    # write data to KTR that does not match the new key
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == key_data_to_ktr)

    # option #2
    config = b"\x02"
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    old_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'uds_handle')
    data_to_hash = "".join((old_key, new_key))
    data_bytes = bytes.fromhex(data_to_hash)
    data_hash = hashlib.sha256(data_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=data_hash)
    assert(err.esr != b"\x80")

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == b"\x00" * 32)

    # write data to KTR that is not 0
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == key_data_to_ktr)

    # option #3
    config = b"\x03"
    err, *_ = device.meta_replace_srk(config)
    assert(err.esr != b"\x80")

    # read KTR
    ktr_data_read = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(ktr_data_read == b"\x00" * 32)

def test_replace_srk_Negative_NotProvisioned():
  """
  Not provisioned
  """
  if provision_stats.Provisioned == 0:
    with device:
      config = b"\x00"

      new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      new_key_bytes = bytes.fromhex(new_key)
      new_key_hash = hashlib.sha256(new_key_bytes).digest()
      err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
      assert(err.esr == b"\x00")

def test_replace_srk_Negative_InvalidSig():
  """
  Invalid signature
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    key_data_to_ktr = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    assert(key_data_to_ktr != None)

    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      bytes.fromhex(key_data_to_ktr), False)
    assert(ret)

    device.test_only_set_boot_amtc0(b"\x00\x00\x00\x00")

    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    if err.esr != b"\x04":
      device.test_only_set_boot_amtc0(None)
      assert(False)

    device.test_only_set_boot_amtc0(None)

def test_replace_srk_Negative_InvalidCounter():
  """
  Invalid counter data
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    key_data_to_ktr = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    assert(key_data_to_ktr != None)

    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      bytes.fromhex(key_data_to_ktr), False)
    assert(ret)

    err, _ = device.read_extended_crtm_status()
    assert(err.esr_and_response_success)

    device.test_only_set_amtc0(b"\x00\x00\x00\x00")

    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    if err.esr != b"\x10":
      device.test_only_set_amtc0(None)
      assert(False)

    device.test_only_set_amtc0(None)

def test_replace_srk_Negative_KTRNotAllWritten():
  """
  Not all eight KTR registers have been written
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    config = b"\x00"

    # clear KTR by a failed replace_srk command
    key_data_to_ktr = b"\xFF" * 32

    ret = utils.write_data_reverse(device, b'\x00', CRTM_Registers.KTR_Base, \
      key_data_to_ktr, False)
    assert(ret)

    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    assert(err.esr != b"\x80")

    # replace_srk with cleared KTR
    new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    new_key_bytes = bytes.fromhex(new_key)
    new_key_hash = hashlib.sha256(new_key_bytes).digest()
    err, *_ = device.meta_replace_srk(config, additional_data=new_key_hash)
    assert(err.esr == b"\x20")

def test_replace_srk_Negative_KTRNotZero():
  """
  Option #1/#3 and KTR != 0x00..00
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # write non zero data to KTR
    key_data_to_ktr = b"\xFF" * 32
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    err, *_ = device.meta_replace_srk(b"\x01")
    assert(err.esr != b"\x80")

    key_data_to_ktr = b"\xFF" * 32
    ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
    assert(ret)

    err, *_ = device.meta_replace_srk(b"\x03")
    assert(err.esr != b"\x80")

def test_replace_srk_Negative_UDSAccessDisabled():
  """
  (option #2 || #3) && UDS is not allowed to access
  """
  if provision_stats.Provisioned == 0:
    return

  flag = True

  with device:
    # disable UDS
    err, _ = device.meta_measure(b"\x00\x02", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # option #2
    key_data_to_ktr = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
    if key_data_to_ktr is None:
      flag = False

    if flag:
      ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, bytes.fromhex(key_data_to_ktr), False)
      if not ret:
        flag = False

    if flag:
      new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      old_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'uds_handle')
      data_to_hash = "".join((old_key, new_key))
      data_bytes = bytes.fromhex(data_to_hash)
      data_hash = hashlib.sha256(data_bytes).digest()
      err, *_ = device.meta_replace_srk(b"\x02", additional_data=data_hash)
      if err.esr != b"\x04":
        flag = False

  # reboot to enable UDS access
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)
  assert(flag)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    uds_replaced = ret[3].UDS_Replaced
    
    # replace UDS with a third-party UDS if UDS is micron UDS
    # need this step because option #3 will return ESR 1 if UDS is already micron UDS
    if not uds_replaced:
      ret = conftest.replace_srk_util(device, uid, b"\x02", new_key_name='test_key_handle', \
        old_key_name='uds_handle')
      assert(ret)

    # disable UDS
    err, _ = device.meta_measure(b"\x00\x02", 0x0, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # option #3
    if flag:
      key_data_to_ktr = b"\x00" * 32
      ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, key_data_to_ktr, False)
      if not ret:
        flag = False
    
    if flag:
      err, *_ = device.meta_replace_srk(b"\x03")
      if err.esr != b"\x04":
        flag = False

  # reboot to enable UDS access
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  # revert UDS to micron UDS if necessary
  if not uds_replaced:
    with device:      
      ret = conftest.replace_srk_util(device, uid, b"\x03")
      assert(ret)

  assert(flag)

def test_replace_srk_Negative_RIoTModeDisabled():
  """
  Provision device with RIoT Mode disabled and call Replace_SRK (option #2 || #3)
  """
  if provision_stats.Provisioned == 0:
    with device:
      # provision device with RIoT mode disabled
      err, *_ = device.meta_provision_crtm(b"\x01")
      assert(err.esr_and_response_success)

      ret = utils.get_dev_info(device)
      assert(ret != None)

      # option #2
      key_data_to_ktr = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      assert(key_data_to_ktr is not None)

      ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, bytes.fromhex(key_data_to_ktr), False)
      assert(ret)

      new_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'test_key_handle')
      old_key = key_mgmt.export_priv_key_as_hex(device.uid_as_str, 'uds_handle')
      data_to_hash = "".join((old_key, new_key))
      data_bytes = bytes.fromhex(data_to_hash)
      data_hash = hashlib.sha256(data_bytes).digest()
      err, *_ = device.meta_replace_srk(b"\x02", additional_data=data_hash)
      assert(err.esr == b"\x04")

      # option #3
      # replace UDS with a third-party UDS if UDS is micron UDS
      # need this step because option #3 will return ESR 1 if UDS is already micron UDS
      if provision_stats.UDS_Replaced == 0:
        err, _ = device.meta_deprovision_crtm(b"\x00")
        assert(err.esr_and_response_success)

        err, *_ = device.meta_provision_crtm(b"\x05")
        assert(err.esr_and_response_success)    
      
        ret = conftest.replace_srk_util(device, uid, b"\x02", new_key_name='test_key_handle', \
          old_key_name='uds_handle')
        assert(ret)

        err, _ = device.meta_deprovision_crtm(b"\x00")
        assert(err.esr_and_response_success)

        # provision device with RIoT mode disabled
        err, *_ = device.meta_provision_crtm(b"\x01")
        assert(err.esr_and_response_success)
      
      key_data_to_ktr = b"\x00" * 32
      ret = utils.write_data_reverse(device, b"\x00", CRTM_Registers.KTR_Base, bytes.fromhex(key_data_to_ktr), False)
      assert(ret)
      
      err, *_ = device.meta_replace_srk(b"\x03")
      assert(err.esr == b"\x04")

      # revert UDS to micron UDS if necessary
      if provision_stats.UDS_Replaced == 0:
        err, _ = device.meta_deprovision_crtm(b"\x00")
        assert(err.esr_and_response_success)

        err, *_ = device.meta_provision_crtm(b"\x05")
        assert(err.esr_and_response_success)

        ret = conftest.replace_srk_util(device, uid, b"\x03")
        assert(ret)

      # deprovision
      err, _ = device.meta_deprovision_crtm(b"\x00")
      assert(err.esr_and_response_success)
