#! /usr/bin/env python3

"""
API Test for Request_UID.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import pytest
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_request_uid_setup(request):
  global device
  device = conftest.get_device(request)

  yield

def test_request_uid_Positive():
  with device:
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

def test_request_uid_Positive_IDKeySelect():
  """
  ID_Key_Select is set in device config
  """
  with device:
    err, *_ = device.meta_request_uid()
    assert(err.esr_and_response_success)

    if device._provision_sts_struct.Provisioned == 1:
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)
      device_config_int = int.from_bytes(response[16:20], byteorder="big")
      
      # set ID_Key_Select bit
      data_int = device_config_int | 0x00000080
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      assert(err.esr_and_response_success)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      err, response, *_ = device.meta_request_uid()
      assert(err.esr_and_response_success)
      assert(device.provision_sts_struct.ID_Key_Select == 1)

      # clear ID_Key_Select bit
      data_int = device_config_int & 0xFFFFFF7F
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      assert(err.esr_and_response_success)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      err, response, *_ = device.meta_request_uid()
      assert(err.esr_and_response_success)
      assert(device.provision_sts_struct.ID_Key_Select == 0)

def test_request_uid_Negative_NotInTestMode():
  """
  For internal testing purposes only, a Config value of 0x01 shall indicate that the Request_UID
  signature field in the command’s response shall be replaced by a digest of all keys in the device as
  follows: digest = SHA256[Micron Kr(s) || UDS || TMK || Customer Kr(s) || SUDS || UID]. Once
  internal testing has completed, this test option shall be disabled until device zeroization.
  Since the device is not in test mode, the signature should not be replaced.
  """
  with device:
    err, *_ = device.meta_request_uid(config=b"\x01")

    # verify the signature is not replaced
    # SDK internally verify the signature using the formula: sig = HMAC(“GPR contents”)Kr(s)
    # If it passes the signature verification, that means the signature is not replaced
    assert(err.esr_and_response_success)

def test_request_uid_Negative_NotDualDie():
  """
  Confirm current Naxos device is not dual-die stack
  Request_UID with Config[7] set or not should return same UID
  """
  with device:
    err, *_ = device.meta_request_uid(config=b"\x00")
    assert(err.esr_and_response_success)
    uid = device.uid

    err, *_ = device.meta_request_uid(config=b"\x80")
    assert(err.esr_and_response_success)
    uid1 = device.uid

    assert(uid == uid1)
