#! /usr/bin/env python3

"""
System test: auth_lock
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_system_auth_lock_setup(request):
  global device
  global provision_stats
  global iboot_ip

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

  yield

def test_system_auth_lock():
  """
  When device is provisioned,
  (1) call Auth_Lock and try to modify PPB bits - should work;
  (2) call PPB_Lock_Bit_Write and try to modify PPB bits - should not work;
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # test block addr
    addr = 0xe40000

    # Auth_Lock
    err, _ = device.meta_auth_lock()
    assert(err.esr_and_response_success)

    # erase PPB bits
    ret = device.meta_ppb_erase()
    assert(ret)

    # read PPB bits
    ret, response = device.meta_ppb_read(addr)
    assert(ret)
    assert(response == b"\xFF")

    # program PPB bits
    ret = device.meta_ppb_program(addr)
    assert(ret)

    # read PPB bits
    ret, response = device.meta_ppb_read(addr)
    assert(ret)
    assert(response == b"\x00")

    # erase PPB bits
    ret = device.meta_ppb_erase()
    assert(ret)

    # read PPB bits
    ret, response = device.meta_ppb_read(addr)
    assert(ret)
    assert(response == b"\xFF")

    # call Read_Status_Register
    ret, response = device.meta_read_status_register()
    assert(ret)
    assert(response is not None)

    # call PPB_Lock_Bit_Write
    ret = device.meta_ppb_lock_bit_write()
    assert(ret)

    # PPB_Lock_Bit_Read
    ret, response = device.meta_ppb_lock_bit_read()
    assert(ret)
    assert(response == b"\x00")
  
    # program PPB bits
    ret = device.meta_ppb_program(addr)
    assert(ret)

    # read PPB bits
    ret, response = device.meta_ppb_read(addr)
    assert(ret)

    if response == b"\x00":
      # test fail - erase PPB bits
      ret = device.meta_ppb_erase()
      assert(False)

def test_system_auth_lock_Powercycle():
  """
  Auth_Unlock bit is cleared after Powercycle.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_auth_lock()
    assert(err.esr_and_response_success)

  conftest.reboot(device, iboot_ip)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    extended_status = device.extended_crtm_status_struct
    assert(extended_status.Auth_Unlock == 0)
