#! /usr/bin/env python3

"""
System test: auth_modify
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

def change_byte_order_source_content(data):
  """
  Change byte order of source_content for auth_modify option #1 
  to pass signature check
  """
  data_len = len(data)
  assert(data_len % 4 == 0)
  data_ret = bytearray(data)

  for i in range(0, data_len, 4):
    data_ret[i:(i + 4)] = data[i:(i + 4)][::-1]

  return data_ret

@pytest.fixture(scope="module", autouse=True)
def test_system_auth_modify_setup(request):
  global device
  global provision_stats
  global iboot_ip

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

  yield

def test_system_auth_modify_Erase_APToLegacy():
  """
  option #0: erase
  set AMBitmap to make a test block AM block
  erase the test block and make it AP block (Auth_Modify)
  set AMBitmap to make the test block legacy block
  try to modify the test block using legacy method - should fail
  set AMBitmap to make the test blcok AM block
  erase the test block and make it AM block (Auth_Modify)
  set AMBitmap to make the test block legacy block
  try to modify the test block using legacy method - should work
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    source_addr = 0xe20000
    dest_addr = 0x0
    size = 0x0

    # unlock test block
    ret = device.meta_dyb_write(False, source_addr)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, source_addr, True)
    assert(ret)

    # Auth_Modify(option #0): erase block and make it AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", source_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # set AMBitmap to make the test block legacy block - should fail
    ret = utils.configure_am(device, source_addr, False)
    assert(not ret)

    # try to modify the test block using legacy method - should fail
    ret = device.meta_page_program(source_addr, b"\x11" * 4)
    assert(ret)

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    if response[0:4] != b"\xFF" * 4:
      utils.configure_am(device, source_addr, True)
      device.meta_auth_modify(b"\x00\x00", source_addr, dest_addr, size)
      utils.configure_am(device, source_addr, False)
      assert(False)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, source_addr, True)
    assert(ret)

    # Auth_Modify(option #0): erase block and make it AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", source_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # set AMBitmap to make the test block legacy block
    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    # try to modify the test block using legacy method - should work
    ret = device.meta_page_program(source_addr, b"\x11" * 4)
    assert(ret)

    ret, response = device.meta_fast_read(source_addr)
    assert(ret)
    assert(response[0:4] == b"\x11" * 4)

    # erase test block
    ret = device.meta_subsector_erase(source_addr)
    assert(ret)

    # lock test block
    ret = device.meta_dyb_write(True, source_addr)
    assert(ret)

def test_system_auth_modify_Erase_Recovery():
  """
  option #0: erase
  in recovery block update process, erase recovery AM block and keep it as AM block
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # set recovery block base address
    recovery_base_addr = 0xe20000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # enable automeausre and clear AMRecovery bit
    data_int = device_config_int | 0x00000001
    data_int = data_int & 0xFFFFFFF7
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # start recovery block update process
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase block and make it AM block
    config = b"\x00\x00"
    dest_addr = 0x0
    size = 0x0

    err, _ = device.meta_auth_modify(config, recovery_base_addr, dest_addr, size)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # finish recovery block update process
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)
    
    # revert Device_Config back
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, device_config, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_system_auth_modify_Program_SourceFlash():
  """
  option #1: program content
  source: flash main array
  verify data written
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # write source content to legacy block
    size = 0x80
    source_addr = 0xe20000
    source_content = bytes.fromhex("11") * size

    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    ret = utils.write_data_to_block(device, source_addr, source_content)
    assert(ret)

    # configure to be AM block
    dest_addr = 0xe30000
    ret = utils.configure_am(device, dest_addr, True)
    assert(ret)

    # Auth_Modify(option #0): erase AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #1): data from flash main array
    err, _ = device.meta_auth_modify(b"\x00\x01", source_addr, dest_addr, size, \
      source_content=source_content)
    assert(err.esr_and_response_success)

    # read programed data back to verify
    ret, response = device.meta_fast_read(dest_addr)
    assert(ret)
    assert(response[0:size] == source_content)

    # convert back to legacy block
    ret = utils.configure_am(device, dest_addr, False)
    assert(ret)

def test_system_auth_modify_Program_SourceKTR_1():
  """
  option #1: program content
  source: KTR || Customer_RAM
  size: 0x20
  verify data written
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # write source content to KTR
    size = 0x20
    source_content = bytes(range(0, size))

    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content)
    assert(ret)

    # change byte order of source_content to pass signature check
    source_content_1 = change_byte_order_source_content(source_content)

    # configure AM block
    dest_addr = 0xe30000
    ret = utils.configure_am(device, dest_addr, True)
    assert(ret)

    # Auth_Modify(option #0): erase AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #1): data from KTR || Customer_RAM
    config = b"\x00\x05"
    source_addr = 0x0

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, \
      source_content=source_content_1)
    assert(err.esr_and_response_success)

    # read programed data back to verify
    ret, response = device.meta_fast_read(dest_addr)
    assert(ret)
    assert(response[0:size] == source_content)

    # convert back to legacy block
    ret = utils.configure_am(device, dest_addr, False)
    assert(ret)

def test_system_auth_modify_Program_SourceKTR_2():
  """
  option #1: program content
  source: KTR || Customer_RAM
  size: 0x80
  verify data written
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # write source content to KTR and Custom_RAM
    size = 0x80
    source_content = bytes(range(0, size))

    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content[:0x20])
    assert(ret)

    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.Customer_RAM, \
      source_content[0x20:])
    assert(ret)

    # change byte order of source_content to pass signature check
    source_content_1 = change_byte_order_source_content(source_content)

    # configure AM block
    dest_addr = 0xe30000
    ret = utils.configure_am(device, dest_addr, True)
    assert(ret)

    # Auth_Modify(option #0): erase AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # Auth_Modify(option #1): data from KTR || Customer_RAM
    config = b"\x00\x05"
    source_addr = 0x0

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size, \
      source_content=source_content_1)
    assert(err.esr_and_response_success)

    # read programed data back to verify
    ret, response = device.meta_fast_read(dest_addr)
    assert(ret)
    assert(response[0:size] == source_content)

    # convert back to legacy block
    ret = utils.configure_am(device, dest_addr, False)
    assert(ret)

def test_system_auth_modify_Program_Recovery():
  """
  option #1: program content
  dest: recovery block in recovery update process
  verify data written
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # write source content to legacy block
    size = 0x80
    source_addr = 0xe20000
    source_content = bytes.fromhex("11") * size

    ret = utils.configure_am(device, source_addr, False)
    assert(ret)

    ret = utils.write_data_to_block(device, source_addr, source_content)
    assert(ret)

    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # set recovery block base address
    recovery_base_addr = 0xe30000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # enable automeausre and clear AMRecovery bit
    data_int = device_config_int | 0x00000001
    data_int = data_int & 0xFFFFFFF7
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # start recovery block update process
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Auth_Modify(option #0): erase block and make it AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", recovery_base_addr, 0x0, 0x0)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # Auth_Modify(option #1): data from flash main array
    err, _ = device.meta_auth_modify(b"\x00\x01", source_addr, recovery_base_addr, size, \
      source_content=source_content)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # read programed data back to verify
    ret, response = device.meta_fast_read(recovery_base_addr)
    if not ret:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)
    assert(response[0:size] == source_content)

    # finish recovery block update process
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)
    
    # revert Device_Config back
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, device_config, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

def test_system_auth_modify_Append_1():
  """
  option #2: append content
  verify data written
  size: 0x20
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    dest_addr = 0xe30000

    # configure the block to AM block
    ret = utils.configure_am(device, dest_addr, True)
    assert(ret)

    # erase the AM block and convert to AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # write append data to KTR
    size = 0x20
    source_content = bytes(range(0, size))
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content)
    assert(ret)

    # Auth_Modify(option #2)
    config = b"\x00\x02"
    source_addr = 0x0

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # read appended data back to verify
    ret, response = device.meta_fast_read(dest_addr)
    assert(ret)
    assert(response[0:size] == source_content)

    # erase the AP block and convert to AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # convert back to legacy block
    ret = utils.configure_am(device, dest_addr, False)
    assert(ret)

def test_system_auth_modify_Append_2():
  """
  option #2: append content
  verify data written
  size: 0x80
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    dest_addr = 0xe30000

    # configure the block to AM block
    ret = utils.configure_am(device, dest_addr, True)
    assert(ret)

    # erase the AM block and convert to AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # write append data to KTR and Custom RAM
    size = 0x80
    source_content = bytes(range(0, size))
    
    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.KTR_Base, \
      source_content[:0x20])
    assert(ret)

    ret = conftest.auth_modify_write_data_to_src(device, CRTM_Registers.Customer_RAM, \
      source_content[0x20:])
    assert(ret)

    # Auth_Modify(option #2)
    config = b"\x00\x02"
    source_addr = 0x0

    err, _ = device.meta_auth_modify(config, source_addr, dest_addr, size)
    assert(err.esr_and_response_success)

    # read appended data back to verify
    ret, response = device.meta_fast_read(dest_addr)
    assert(ret)
    if response[0:size] != source_content:
      device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
      utils.configure_am(device, dest_addr, False)
      assert(False)

    # erase the AP block and convert to AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", dest_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # convert back to legacy block
    ret = utils.configure_am(device, dest_addr, False)
    assert(ret)

def test_system_auth_modify_AMBlock_LegacyModify():
  """
  AM blocks may only be programmed and erased using the Auth_Modify command.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    am_block_addr = 0xe20000
    legacy_block_addr = 0xe30000

    # write data to legacy block
    ret = utils.configure_am(device, legacy_block_addr, False)
    assert(ret)

    size = 0x80
    data = b"\x11" * size
    ret = utils.write_data_to_block(device, legacy_block_addr, data)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, am_block_addr, True)
    assert(ret)

    # Auth_Modify option #0
    err, _ = device.meta_auth_modify(b"\x00\x00", am_block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response == b"\xff" * len(response))

    # Auth_Modify option #1
    err, _ = device.meta_auth_modify(b"\x00\x01", legacy_block_addr, am_block_addr, size, \
      source_content=data)
    assert(err.esr_and_response_success)

    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response[0:size] == data)

    # try to modify AM block using legacy block operation - should not change AM block
    data_1 = b"\x22" * size
    ret = utils.write_data_to_block(device, am_block_addr, data)
    assert(ret)

    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response[0:size] == data)

    # convert test block back to legacy block
    ret = utils.configure_am(device, am_block_addr, False)
    assert(ret)

def test_system_auth_modify_AMBlock_OtherLocking():
  """
  AM blocks shall no longer be affected by any other block locking.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    am_block_addr = 0xe20000
    legacy_block_addr = 0xe30000

    # write data to legacy block
    ret = utils.configure_am(device, legacy_block_addr, False)
    assert(ret)

    size = 0x80
    data = b"\x11" * size
    ret = utils.write_data_to_block(device, legacy_block_addr, data)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, am_block_addr, True)
    assert(ret)

    # Lock DYB lock - should not affect AM operation
    ret = device.meta_dyb_write(True, am_block_addr)
    assert(ret)

    # Auth_Modify option #0
    err, _ = device.meta_auth_modify(b"\x00\x00", am_block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)
    
    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response == b"\xff" * len(response))

    # Auth_Modify option #1
    err, _ = device.meta_auth_modify(b"\x00\x01", legacy_block_addr, am_block_addr, size, \
      source_content=data)
    assert(err.esr_and_response_success)

    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response[0:size] == data)
     
    ret = device.meta_dyb_write(False, am_block_addr)
    assert(ret)

    # Lock PPB lock - should not affect AM operation
    # Auth_Lock
    err, _ = device.meta_auth_lock()
    assert(err.esr_and_response_success)

    # program PPB bits
    ret = device.meta_ppb_program(am_block_addr)
    assert(ret)

    # Auth_Modify option #0
    err, _ = device.meta_auth_modify(b"\x00\x00", am_block_addr, 0x0, 0x0)
    assert(err.esr_and_response_success)
    
    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response == b"\xff" * len(response))

    # Auth_Modify option #1
    err, _ = device.meta_auth_modify(b"\x00\x01", legacy_block_addr, am_block_addr, size, \
      source_content=data)
    assert(err.esr_and_response_success)

    # read block data
    err, response = device.meta_fast_read(am_block_addr)
    assert(response[0:size] == data)

    # erase PPB bits
    ret = device.meta_ppb_erase()
    assert(ret)

    # call PPB_Lock_Bit_Write
    ret = device.meta_ppb_lock_bit_write()
    assert(ret)

    # convert test block back to legacy block
    ret = utils.configure_am(device, am_block_addr, False)
    assert(ret)

def test_system_auth_modify_AMBlock_BackToLegacy():
  """
  When a block is changed back to a legacy block by setting its AMBitmap bit, 
  any block lock protection shall again be restored.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = 0xe20000

    # Lock DYB
    ret = device.meta_dyb_write(True, addr)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # convert AM block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # check DYB
    ret, response = device.meta_dyb_read(addr)
    assert(response == b"\x01")

    # Unlock DYB
    ret = device.meta_dyb_write(False, addr)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # convert AM block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # check DYB
    ret, response = device.meta_dyb_read(addr)
    assert(response == b"\x00")

    # Auth_Lock
    err, _ = device.meta_auth_lock()
    assert(err.esr_and_response_success)

    # Lock PPB
    ret = device.meta_ppb_program(addr)
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # convert AM block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # check PPB
    ret, response = device.meta_ppb_read(addr)
    assert(response == b"\x00")

    # Unlock PPB
    ret = device.meta_ppb_erase()
    assert(ret)

    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # convert AM block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # check PPB
    ret, response = device.meta_ppb_read(addr)
    assert(response == b"\xff")

    # PPB_Lock_Bit_Write
    ret = device.meta_ppb_lock_bit_write()
    assert(ret)

def test_system_auth_modify_TamperDetect_AMBlock():
  """
  Tamper_Detect should be set when trying to modify AM block via legacy commands
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = 0xe20000

    # convert test block to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # unlock DYB
    ret = device.meta_dyb_write(False, addr)
    assert(ret)
    
    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    device_config_volatile_int = int.from_bytes(response[24:28], byteorder="big")        

    data_int = device_config_volatile_int | 0x00000080
    data = data_int.to_bytes(4, byteorder="big")

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify AM block via erase
    ret = device.meta_subsector_erase(addr)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify AM block via program
    ret = device.meta_page_program(addr, b"\x11" * 4)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # convert test block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

def test_system_auth_modify_TamperDetect_APBlock():
  """
  Tamper_Detect should be set when trying to modify AP block via legacy commands
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = 0xe20000

    # convert test block to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # unlock DYB
    ret = device.meta_dyb_write(False, addr)
    assert(ret)
    
    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # convert test block to AP block
    err, _ = device.meta_auth_modify(b"\x00\x10", addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    device_config_volatile_int = int.from_bytes(response[24:28], byteorder="big")        

    data_int = device_config_volatile_int | 0x00000080
    data = data_int.to_bytes(4, byteorder="big")

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify AP block via erase
    ret = device.meta_subsector_erase(addr)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify AP block via program
    ret = device.meta_page_program(addr, b"\x11" * 4)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # convert test block to AM block
    err, _ = device.meta_auth_modify(b"\x00\x00", addr, 0x0, 0x0)
    assert(err.esr_and_response_success)

    # convert test block back to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

def test_system_auth_modify_TamperDetect_RecoveryBlock():
  """
  Tamper_Detect should be set when trying to modify Recovery Block via legacy commands
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = 0xe20000

    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # set recovery block base address
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # enable automeausre and clear AMRecovery bit
    data_int = device_config_int | 0x00000001
    data_int = data_int & 0xFFFFFFF7
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # start recovery block update process
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success) 

    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    device_config_volatile_int = int.from_bytes(response[24:28], byteorder="big")        

    data_int = device_config_volatile_int | 0x00000080
    data = data_int.to_bytes(4, byteorder="big")

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify Recvoery block via erase
    ret = device.meta_subsector_erase(addr)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # try to modify AM block via program
    ret = device.meta_page_program(addr, b"\x11" * 4)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)

    # clear Tamper_Detect bit
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config_Volatile, data, False)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # finish recover block update process
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

def test_system_auth_modify_TamperDetect_Poweron():
  """
  Tamper_Detect should be cleared after poweron
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    addr = 0xe20000        
    
    # convert test block to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)

    # unlock DYB
    ret = device.meta_dyb_write(False, addr)
    assert(ret)
    
    # set AMBitmap to make a test block AM block
    ret = utils.configure_am(device, addr, True)
    assert(ret)

    # try to modify AM block via erase
    ret = device.meta_subsector_erase(addr)
    assert(ret)
    
    # Tamper_Detect should be set
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 1)
  
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    # Tamper_Detect should be clear
    err, _ = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config_Volatile)
    assert(err.esr_and_response_success)
    assert(device._extended_crtm_status_struct.Tamper_Detect == 0)

    # convert test block to legacy block
    ret = utils.configure_am(device, addr, False)
    assert(ret)