#! /usr/bin/env python3

"""
System test: automeasure.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_system_automeasure_setup(request):
  global device
  global iboot_ip
  global provision_stats

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      ret = utils.enable_automeasure(device)
      assert(ret)

  yield

  if provision_stats.Provisioned == 1:
    with device:
      ret = utils.get_dev_info(device)
      assert(ret != None)

      ret = utils.disable_automeasure(device)
      assert(ret)

def test_system_automeasure_MaxSizeNotExceeded_1():
  """
  max auto measure size (0x20000) is not exceeded
  1 crtm addr range register is used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x20000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x02\x00\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 0)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([
      0xc6, 0x9d, 0xbc, 0xef, 0x72, 0xe0, 0xcc, 0x4b,
      0x1b, 0x11, 0x9e, 0x9c, 0x41, 0x0b, 0xd1, 0xb7,
      0xde, 0xdf, 0xa5, 0x50, 0x18, 0x5d, 0x2b, 0x81,
      0xa6, 0xf7, 0xa7, 0xed, 0xda, 0x13, 0x9f, 0x2c
    ])

    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeNotExceeded_2():
  """
  max auto measure size (0x20000) is not exceeded
  2 crtm addr range registers are used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x20000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x02\x00\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0xE04000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 2, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
    
    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 0)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0xf9, 0xe5, 0xd8, 0x20, 0xc0, 0xe7, 0xbe, 0x6b,
      0x34, 0x6a, 0xcf, 0x45, 0x07, 0x5b, 0xa4, 0x21,
      0xa5, 0x3a, 0xb4, 0x09, 0x15, 0x06, 0x40, 0x30,
      0x5e, 0x5d, 0xae, 0x31, 0xf1, 0x68, 0xce, 0x56
    ])

    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeNotExceeded_3():
  """
  max auto measure size (0x20000) is not exceeded
  all crtm addr range registers are used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x20000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x02\x00\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0xE04000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 2, 0xE08000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 3, 0xE0C000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 4, 0xE10000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 5, 0xE14000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)
 
    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 6, 0xE18000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)                           

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 7, 0xE1C000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
    
    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 0)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([
      0xcd, 0x49, 0xda, 0x38, 0xd5, 0xfb, 0x82, 0x1a,
      0xa1, 0xa9, 0xb8, 0x13, 0xf5, 0x36, 0x77, 0x64,
      0x77, 0x34, 0x67, 0x33, 0x42, 0xab, 0x48, 0xba,
      0x41, 0x65, 0xf8, 0x02, 0x07, 0xe4, 0x76, 0x27
    ])

    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeNotExceeded_4():
  """
  max auto measure size (0x20000) is not exceeded
  1 crtm addr range register is used
  block size 0x4000
  """
  if provision_stats.Provisioned == 0:
    return
    
  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x20000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x02\x00\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x4000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
    
    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 0)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0x43, 0x83, 0x11, 0x69, 0x80, 0xdb, 0x3d, 0xe0,
      0x59, 0x43, 0x4f, 0x2b, 0x3c, 0xda, 0x4f, 0xc1,
      0x5e, 0x9c, 0x11, 0x4b, 0xc8, 0xdd, 0x31, 0xd7,
      0x00, 0x8e, 0xfd, 0x18, 0x88, 0xf0, 0xf3, 0x0f
    ])

    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeExceeded_1():
  """
  max auto measure size (0x1000) is exceeded
  1 crtm addr range register is used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x1000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x00\x10\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
    
    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 1)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0x45, 0x2e, 0xd7, 0x46, 0xf2, 0xc3, 0xe6, 0xe4,
      0x07, 0x73, 0x3e, 0xf6, 0x91, 0x5c, 0xdc, 0x0e,
      0x21, 0x35, 0x10, 0x61, 0xf0, 0xe1, 0x6d, 0xad,
      0xce, 0x0f, 0x33, 0x20, 0xca, 0x80, 0x84, 0x3c
    ])

    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeExceeded_2():
  """
  max auto measure size (0x3000) is exceeded
  2 crtm addr range register is used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x3000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x00\x30\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0xE04000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 2, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 1)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0x4e, 0xef, 0x72, 0x7b, 0x3c, 0x60, 0x72, 0x4f,
      0x49, 0xeb, 0x53, 0x25, 0x55, 0x7e, 0x84, 0x0c,
      0xba, 0x40, 0x60, 0x90, 0x9f, 0x60, 0xb6, 0xbc,
      0xa3, 0xab, 0xdb, 0x47, 0x39, 0xb6, 0xc3, 0xf5
    ])
    
    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeExceeded_3():
  """
  max auto measure size (0x8000) is exceeded
  all crtm addr range registers are used
  block size 0x2000
  """
  if provision_stats.Provisioned == 0:
    return
    
  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x8000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x00\x80\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0xE04000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 2, 0xE08000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 3, 0xE0C000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 4, 0xE10000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 5, 0xE14000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)
 
    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 6, 0xE18000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)                           

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 7, 0xE1C000, 0x2000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)
    
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 1)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0x75, 0x66, 0x97, 0xd6, 0x7d, 0xb5, 0x54, 0xd0,
      0xe7, 0x3e, 0x6a, 0xb7, 0x25, 0xc0, 0x20, 0x3a,
      0x62, 0xb0, 0xf2, 0x12, 0x11, 0x98, 0xe1, 0x99,
      0x50, 0xc5, 0x00, 0x87, 0x3c, 0x88, 0xa8, 0xaf
    ])
    
    assert(dgst == expected_dgst)

def test_system_automeasure_MaxSizeExceeded_4():
  """
  max auto measure size (0x2000) is not exceeded
  1 crtm addr range register is used
  block size 0x400s0
  """
  if provision_stats.Provisioned == 0:
    return
    
  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # Set MaxAutoMeasureSize to 0x2000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      b"\x00\x00\x20\x00", True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x4000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # Do a flash-extend option 1, to save nv and nv-partial values (setting what value auto-measure will compare to)
    config = b"\x01" # bits 2 and 0 are '01' for option 1, bit 1 is '0' for start at beginning of linked list
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

  # Reboot and reconnect
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.CRTM_PCR0_Invalid == 0)
    assert(extended_crtm_status.Partial_Measurement == 1)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # compare to expected measurement value
    expected_dgst = bytes([ 
      0xc6, 0x9d, 0xbc, 0xef, 0x72, 0xe0, 0xcc, 0x4b,
      0x1b, 0x11, 0x9e, 0x9c, 0x41, 0x0b, 0xd1, 0xb7,
      0xde, 0xdf, 0xa5, 0x50, 0x18, 0x5d, 0x2b, 0x81,
      0xa6, 0xf7, 0xa7, 0xed, 0xda, 0x13, 0x9f, 0x2c
    ])

    assert(dgst == expected_dgst)
