#! /usr/bin/env python3

"""
System test: crtm_register_write.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_system_crtm_register_write_setup(request):
  global device
  global provision_stats
  global addr
  global data
  addr = utils.CRTM_Registers.CRTM_Range_Start_Addr_0
  data = b"\x31\x41\x52\x65"

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret
  yield

def test_system_crtm_register_write_Volatile():
  """ Test volatile write behavior

    1. Send a CRTM_Register_Write command to Naxos with non-zero data
       content, in volatile mode.
    2. Read the following GPRR register.

    Verify that bytes 0x10 to 0x13 of the returned GPRR do contain the
    CRTM register data transferred in step 1.

    The idea of this test is to verify that volatile writes conducted
    without 'update start' and 'update finish' will succeed, but
    non-volatile writes conducted without 'update start' and 'update finish'
    will fail.
  """
  global data

  if provision_stats.Provisioned == 0:
    return

  config = b"\x01"

  with device:
    err, response = device.meta_crtm_register_read(config, addr)
    assert(err.esr_and_response_success)

    prior = response[16:20]

    if prior == data:
      data_int = int.from_bytes(data, byteorder="big")
      data_int += 1
      data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(config, addr, data, False)
    assert(err.esr_and_response_success)

    err, response = device.meta_crtm_register_read(config, addr)
    assert(err.esr_and_response_success)

    value = response[16:20]

  assert value == data

def test_system_crtm_register_write_NonVolatile():
  """ Test non-volatile GPRR write behavior

    1. Send a CRTM_Register_Write command to Naxos with non-zero data
       content, in non-volatile mode.
    2. Read the following GPRR register.

    Verify that bytes 0x10 to 0x13 of the returned GPRR do not contain the
    CRTM register data transferred in step 1.

    The idea of this test is to verify that volatile writes conducted
    without 'update start' and 'update finish' will succeed, but
    non-volatile writes conducted without 'update start' and 'update finish'
    will fail.
  """
  global data

  if provision_stats.Provisioned == 0:
    return

  config = b"\x00"

  with device:
    err, response = device.meta_crtm_register_read(config, addr)
    assert(err.esr_and_response_success)

    prior = response[16:20]

    if prior == data:
      data_int = int.from_bytes(data, byteorder="big")
      data_int += 1
      data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(config, addr, data, False)
    assert(err.esr == b"\x04")

    err, response = device.meta_crtm_register_read(config, addr)
    assert(err.esr_and_response_success)

    value = response[16:20]

  assert value == prior
