#! /usr/bin/env python3

"""
System test: flash_extend.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest
import hashlib

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

def set_measure_range():
  """
  Set measurement range and make sure device config is right
  """
  # read current device config
  err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
  assert(err.esr_and_response_success)

  device_config = response[16:20]
  device_config_int = int.from_bytes(device_config, byteorder="big")

  err, _ = device.meta_acrtm_update_start(b"\x00")
  assert(err.esr_and_response_success)

  # turn off automeasure and recovery mode
  data_int = device_config_int & 0xFFFFFFFC
  data = data_int.to_bytes(4, byteorder="big")

  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  # set measurement range
  ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xE00000, 0x2000)
  if not ret:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0xE04000, 0x2000)
  if not ret:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  ret = utils.set_crtm_addr_range_reg(device, b"\x00", 2, 0xE08000, 0x2000)
  if not ret:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  ret = utils.set_crtm_addr_range_reg(device, b"\x00", 3, 0xE0C000, 0x2000)
  if not ret:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  ret = utils.set_crtm_addr_range_reg(device, b"\x00", 4, 0x0, 0x0)
  if not ret:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  # set max automeasure size
  err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
    b"\x00\x00\x40\x00", True)
  if not err.esr_and_response_success:
    device.meta_acrtm_update_finish(b"\x00")
    assert(False)

  err, _ = device.meta_acrtm_update_finish(b"\x00")
  assert(err.esr_and_response_success)

def reset_flash_pcr0():
  """
  Reset Flash_PCR0 to 00 .. 00 and make sure extended status is right
  """
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # get Flash_PCR0 value
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)

    # asserted poweron Flash_PCR0 value
    std_dgst = bytes([
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ])
    assert(dgst == std_dgst)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Range_Invalid == 0)
    assert(extended_crtm_status.Recovery_Block == 0)

def flash_extend_option0_test(config, std_dgst):
  """
  commen code for test_system_flash_extend_Measure_*
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    set_measure_range()
  
  reset_flash_pcr0()

  with device:
    # reprogram both Flash_DigestNV and Flash_DigestNV_Partial registers
    err, _ = device.meta_flash_extend(b"\x01")
    assert(err.esr_and_response_success)

    # Flash_Extend
    err, _ = device.meta_flash_extend(config)
    assert(err.esr_and_response_success)

    # Flash_PCR0
    dgst = utils.get_flash_digest(device, CRTM_Registers.Flash_PCR0_Low_Addr)
    assert(dgst)
    
    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)    
    assert(extended_crtm_status.CRTM_Range_Invalid == 0)        
    assert(extended_crtm_status.Partial_Measurement == 0)

    # verify results
    assert(std_dgst == dgst)

@pytest.fixture(scope="module", autouse=True)
def test_system_flash_extend_setup(request):
  global device
  global iboot_ip
  global provision_stats

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      err, _ = device.read_extended_crtm_status()
      assert(err.esr == b'\x80')

      ret = utils.enable_automeasure(device)
      assert(ret)

  yield

  if provision_stats.Provisioned == 1:
    with device:
      ret = utils.disable_automeasure(device)
      assert(ret)

def test_system_flash_extend_Measure_LinkedList():
  """
  option #0
  flashExtndStrt_CrtmMaxAutoMeasSize
  """
  config = b"\x00"

  std_dgst = bytes([
    0x75, 0x66, 0x97, 0xd6, 0x7d, 0xb5, 0x54, 0xd0,
    0xe7, 0x3e, 0x6a, 0xb7, 0x25, 0xc0, 0x20, 0x3a,
    0x62, 0xb0, 0xf2, 0x12, 0x11, 0x98, 0xe1, 0x99,
    0x50, 0xc5, 0x00, 0x87, 0x3c, 0x88, 0xa8, 0xaf
  ])

  flash_extend_option0_test(config, std_dgst)

def test_system_flash_extend_Measure_CrtmMaxAutoMeasSize():
  """
  option #0
  flashExtndStrt_CrtmMaxAutoMeasSize
  """
  config = b"\x02"

  std_dgst = bytes([
    0xcc, 0xbd, 0xbe, 0xa0, 0xb6, 0x88, 0x84, 0x1b,
    0x7c, 0x1a, 0x07, 0xa0, 0x3b, 0x76, 0x9f, 0xc2,
    0xce, 0x68, 0x43, 0xc9, 0x17, 0xe3, 0x8a, 0x31,
    0xc8, 0x1d, 0x87, 0x2b, 0xfe, 0xd1, 0xf8, 0x17
  ])

  flash_extend_option0_test(config, std_dgst)

def test_system_flash_extend_ReprgmNvDgsts():
  """
  option #1
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # option #1
    # reprogram both Flash_DigestNV and Flash_DigestNV_Partial registers
    err, _ = device.meta_flash_extend(b"\x01")
    assert(err.esr_and_response_success)

    # Flash_DigestNV_Partial
    dgst_nv_partial = utils.get_flash_digest(device, CRTM_Registers.Flash_DigestNV_Partial_L)
    assert(dgst_nv_partial)

    # Flash_DigestNV
    dgst_nv = utils.get_flash_digest(device, CRTM_Registers.Flash_DigestNV_L)
    assert(dgst_nv)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 0)
    assert(extended_crtm_status.Partial_Measurement == 0)
    
    # verify results
    std_dgst_nv_partial = bytes([
      0xbb, 0x18, 0x11, 0x42, 0x49, 0xe2, 0xed, 0x43,
      0x16, 0x64, 0x44, 0xcc, 0x81, 0xbd, 0xa5, 0x6e,
      0x18, 0x91, 0x58, 0x75, 0x77, 0xc2, 0xf3, 0xa5,
      0xf5, 0x83, 0x25, 0x6a, 0x0b, 0x46, 0x88, 0xad
    ])
    assert(std_dgst_nv_partial == dgst_nv_partial)

    std_dgst_nv = bytes([
      0x5c, 0x3d, 0x57, 0xd0, 0xe3, 0xa3, 0x38, 0x38,
      0x0d, 0x3e, 0x6a, 0x4d, 0xaf, 0x34, 0x02, 0xf2,
      0x28, 0x56, 0x4f, 0xac, 0x0d, 0xc2, 0x30, 0x18,
      0x97, 0xd2, 0x03, 0x73, 0x6d, 0x68, 0x20, 0xad
    ])
    assert(std_dgst_nv == dgst_nv)

def test_system_flash_extend_CkRcvryBlk_ReprgmRcvryBlkNvDgst():
  """
  option #2 and #3
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # read current device config
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)

    device_config = response[16:20]
    device_config_int = int.from_bytes(device_config, byteorder="big")

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # turn on automeasure, turn off recovery mode, and set AMRecovery bit
    data_int = device_config_int | 0x00000009
    data_int = (data_int & 0xFFFFFFFD)
    data = data_int.to_bytes(4, byteorder="big")

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # set recovery base address
    recovery_base_addr = 0xf00000
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
      recovery_base_addr.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

    # erase recovery block

    # ACRTM_Update_Start
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Subsector_Erase
    ret = device.meta_subsector_erase(recovery_base_addr)
    if not ret:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # ACRTM_Update_Finish
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # reprogram Recovery_DigestNV register: Flash_Extend option #3
    err, _ = device.meta_flash_extend(b"\x05")
    assert(err.esr_and_response_success)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.Recovery_Mismatch == 0)

    # update recovery block

    # ACRTM_Update_Start
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Page_Program
    data = b"\x11" * 256
    ret = device.meta_page_program(recovery_base_addr, data)
    if not ret:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # ACRTM_Update_Finish
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # check recovery block: Flash_Extend option #2
    err, response = device.meta_flash_extend(b"\x04")
    assert(err.esr_and_response_success)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.Recovery_Mismatch == 1)

    # reset the recovery block

    # ACRTM_Update_Start
    err, _ = device.meta_acrtm_update_start(b"\x01")
    assert(err.esr_and_response_success)

    # Subsector_Erase
    ret = device.meta_subsector_erase(recovery_base_addr)
    if not ret:
      device.meta_acrtm_update_finish(b"\x01")
      assert(False)

    # ACRTM_Update_Finish
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # check recovery block: Flash_Extend option #2
    err, _ = device.meta_flash_extend(b"\x04")
    assert(err.esr_and_response_success)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.Recovery_Mismatch == 0)

    ret = utils.disable_automeasure(device)
    assert(ret)
