#! /usr/bin/env python3

"""
System test: get_FDS
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import hmac
import hashlib
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_system_get_fds_setup(request):
  global device
  global iboot_ip
  global provision_stats
  global uid
  global uds_disabled

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (uid, _, _, provision_stats) = ret

    if provision_stats.Provisioned == 1:
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      # read uds_disable bit
      uds_disabled = device.extended_crtm_status_struct.UDS_Disabled

      if provision_stats.Remotely_Provisioned == 1 \
        and provision_stats.RIOT_Mode == 1 and uds_disabled == 0:

        # read current device config
        err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
        assert(err.esr_and_response_success)

        device_config = response[16:20]
        device_config_int = int.from_bytes(device_config, byteorder="big")

        err, _ = device.meta_acrtm_update_start(b"\x00")
        assert(err.esr_and_response_success)

        # turn on automeasure and turn off recovery mode
        data_int = device_config_int | 0x00000001
        data_int = data_int & 0xFFFFFFFD
        data = data_int.to_bytes(4, byteorder="big")

        err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
        if not err.esr_and_response_success:
          device.meta_acrtm_update_finish(b"\x00")
          assert(False)

        err, _ = device.meta_acrtm_update_finish(b"\x00")
        assert(err.esr_and_response_success)

        # replace UDS with a known value
        ret = conftest.replace_srk_util(device, uid, b"\x02", new_key_name='test_key_handle', old_key_name='uds_handle')
        assert(ret)

  yield

  if provision_stats.Provisioned != 0 and provision_stats.Remotely_Provisioned == 1 \
    and provision_stats.RIOT_Mode == 1 and uds_disabled == 0:
    
    with device:
      # turn off automeasure
      ret = utils.disable_automeasure(device)
      assert(ret)

      # revert uds back
      # set config based on whether the old UDS is micron default UDS
      if provision_stats.UDS_Replaced == 1:
        config = b"\x02"
        new_key_name = 'uds_handle'
        old_key_name = 'test_key_handle'
      else:
        config = b"\x03"
        new_key_name = None
        old_key_name = None

      ret = conftest.replace_srk_util(device, uid, config, new_key_name, old_key_name)
      assert(ret)

def get_fds_test(addrs, sizes, max_automeasure_size, std_dgst):
  """
  commen code for test_system_get_fds_*
  """
  if provision_stats.Provisioned != 0 and provision_stats.Remotely_Provisioned == 1 \
    and provision_stats.RIOT_Mode == 1 and uds_disabled == 0:
    
    assert(len(addrs) == 8 and len(sizes) == 8)

    with device:
      # ACRTM_Update_Start
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # set CRTM ranges
      for i in range(0, 8):
        ret  = utils.set_crtm_addr_range_reg(device, b"\x00", i, addrs[i], sizes[i])
        if not ret:
          device.meta_acrtm_update_finish(b"\x00")
          assert(False)

      # CRTM_Max_AutoMeasure_Size
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size,
        max_automeasure_size.to_bytes(4, byteorder="big"), True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

    # reboot
    ret = conftest.reboot(device, iboot_ip)
    assert(ret)

    with device:
      ret = utils.get_dev_info(device)
      assert(ret != None)

      # Measure option #1
      err, _ = device.meta_measure(b"\x00\x01", 0x0000, 0x0000, 0x0000)
      assert(err.esr_and_response_success)

      # read KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      assert(dgst == std_dgst)


def test_system_get_fds_1():
  """
  CRTM Measure command Option # 1
  Flash_PCR0 set by automeasure
  1 addr range register is used
  block size 0x2000
  max auto measure size 0x20000
  """
  # crtm addrs and sizes
  addrs = (0xe00000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  sizes = (0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  max_automeasure_size = 0x20000

  std_dgst = bytes([
    0x77, 0x96, 0x5f, 0xb6, 0xc7, 0x5b, 0x60, 0x26,
    0xac, 0x01, 0x89, 0xd2, 0x66, 0xd6, 0xeb, 0xbb,
    0x42, 0xb4, 0x64, 0x98, 0x06, 0xf1, 0xb6, 0x32,
    0x21, 0x88, 0xc8, 0xf0, 0xd7, 0x90, 0x22, 0x5f
  ])

  get_fds_test(addrs, sizes, max_automeasure_size, std_dgst)

def test_system_get_fds_2():
  """
  CRTM Measure command Option # 1
  Flash_PCR0 set by automeasure
  2 addr range registers are used
  block size 0x2000
  max auto measure size 0x20000
  """
  # crtm addrs and sizes
  addrs = (0xe00000, 0xe04000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  sizes = (0x2000, 0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  max_automeasure_size = 0x20000

  std_dgst = bytes([
    0x88, 0xfa, 0xff, 0x82, 0xab, 0x48, 0xc3, 0x2e,
    0x66, 0x34, 0x9e, 0x03, 0x20, 0x25, 0x96, 0xb4,
    0xda, 0xed, 0x67, 0xe9, 0x4c, 0x07, 0x41, 0xe4,
    0xc9, 0x64, 0x69, 0x3a, 0x46, 0x42, 0xd4, 0xb9
  ])

  get_fds_test(addrs, sizes, max_automeasure_size, std_dgst)

def test_system_get_fds_3():
  """
  CRTM Measure command Option # 1
  Flash_PCR0 set by automeasure
  all addr range registers are used
  block size 0x2000
  max auto measure size 0x20000
  """
  # crtm addrs and sizes
  addrs = (0xe00000, 0xe04000, 0xe08000, 0xe0c000, 0xe10000, 0xe14000, 0xe18000, 0xe1c000)
  sizes = (0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000)
  max_automeasure_size = 0x20000

  std_dgst = bytes([
    0x99, 0x0e, 0xe3, 0xb3, 0x74, 0x2f, 0x5b, 0xfd,
    0xdd, 0x35, 0x6c, 0xc6, 0x72, 0xc8, 0x63, 0x84,
    0x33, 0xe2, 0xb0, 0x94, 0xfa, 0x50, 0x3f, 0x90,
    0x0c, 0x37, 0x06, 0x3f, 0xfe, 0x8e, 0xac, 0x3b
  ])

  get_fds_test(addrs, sizes, max_automeasure_size, std_dgst)

def test_system_get_fds_4():
  """
  CRTM Measure command Option # 1
  Flash_PCR0 set by automeasure
  1 addr range register is used
  block size 0x4000
  max auto measure size 0x20000
  """
  # crtm addrs and sizes
  addrs = (0xe00000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  sizes = (0x4000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  max_automeasure_size = 0x20000

  std_dgst = bytes([
    0x3d, 0x6f, 0x2c, 0xf7, 0x42, 0xb9, 0x88, 0xe6,
    0x44, 0x01, 0xcb, 0x7e, 0x04, 0x72, 0x78, 0x28,
    0xe7, 0x66, 0x6e, 0xe7, 0x07, 0xed, 0x48, 0xcd,
    0xe4, 0x6d, 0x38, 0x4d, 0x0f, 0x7f, 0x3e, 0xb2
  ])

  get_fds_test(addrs, sizes, max_automeasure_size, std_dgst)

def test_system_get_fds_5():
  """
  CRTM Measure command Option # 1
  Flash_PCR0 set by automeasure
  1 addr range register is used
  block size 0x2000
  max auto measure size 0x1000
  """
  # crtm addrs and sizes
  addrs = (0xe00000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  sizes = (0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  max_automeasure_size = 0x1000

  std_dgst = bytes([
    0x24, 0xdb, 0x04, 0xcd, 0x22, 0xe3, 0xe8, 0x8e,
    0x73, 0x1e, 0xac, 0xe0, 0x76, 0x41, 0x95, 0x9e,
    0x13, 0xc8, 0x5a, 0xcb, 0xed, 0xe4, 0xc1, 0x27,
    0x8d, 0xbb, 0x06, 0x94, 0xad, 0x08, 0x7e, 0xbe
  ])

  get_fds_test(addrs, sizes, max_automeasure_size, std_dgst)


def test_system_get_fds_FDSMetaDigest():
  """
  Test that this is calculated during the first command that could affect its contents.
  Should match with powerup.
  """
  if provision_stats.Provisioned == 0 or provision_stats.Remotely_Provisioned == 0 \
    or provision_stats.RIOT_Mode == 0 or uds_disabled == 1:
    return

  # set CRTM settings and get FDS at powerup
  addrs = (0xe00000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  sizes = (0x2000, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0)
  max_automeasure_size = 0x20000
  fds = None

  with device:
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    for i in range(0, 8):
      ret  = utils.set_crtm_addr_range_reg(device, b"\x00", i, addrs[i], sizes[i])
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size,
      max_automeasure_size.to_bytes(4, byteorder="big"), True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

  # reboot
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # measure option #1
    err, _ = device.meta_measure(b"\x00\x01", 0x0000, 0x0000, 0x0000)
    assert(err.esr_and_response_success)

    fds = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(fds is not None)

  # reboot
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret is not None)

    # change CRTM Automeasure size - FDS_Meta_Digest shouldn't change after this
    err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
    assert(err.esr_and_response_success)
    max_automeasure_size = int.from_bytes(response[24:28], byteorder="big")

    max_automeasure_size += 0x1000

    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    data = max_automeasure_size.to_bytes(4, byteorder="big")
    err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
      data, True)
    if not err.esr_and_response_success:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
    
    # measure option 1
    err, _ = device.meta_measure(b"\x00\x01", 0x0000, 0x0000, 0x0000)
    assert(err.esr_and_response_success)

    fds_1 = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
    assert(fds is not None)

    # verify that FDS_Meta_Digest doesn't change
    assert(fds == fds_1)
