#! /usr/bin/env python3

"""
System test: measure.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

def write_measurement_data(addr, data, sign):
  """
  write measurement src/keySrc data to KTR or EEPROM
  """
  data_len = len(data)
  assert(data_len == 32)

  if sign:
    # EEPROM
    # ACRTM_Update_Start
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    ret = utils.write_data_reverse(device, b"\x00", addr, data, True)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # ACRTM_Update_Finish
    ret = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)
  else:
    # KTR
    ret = utils.write_data_reverse(device, b"\x00", addr, data, False)
    assert(ret)

@pytest.fixture(scope="module", autouse=True)
def test_system_measure_setup(request):
  global device
  global provision_stats
  global uid
  global uds_disabled

  device = conftest.get_device(request)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (uid, _, _, provision_stats) = ret

    if provision_stats.Provisioned != 0:
      # read current device config
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      device_config = response[16:20]
      device_config_int = int.from_bytes(device_config, byteorder="big")

      # read uds_disable bit
      uds_disabled = device.extended_crtm_status_struct.UDS_Disabled

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # turn off recovery mode
      data_int = device_config_int & 0xFFFFFFFD
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set EEPROM attribute: modify and read for the hightest 32-byte EEPROM register
      # TODO - temporary fix for 1.4 eeprom attribute settings: eeprom_attr = 0x30000000 doesn't work
      eeprom_attr = b"\x33" * 4

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.EEPROM_Attributes, eeprom_attr, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      # get EEPROM attribute: verify results
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.EEPROM_Attributes)
      assert(err.esr_and_response_success)
      eeprom_attr1 = response[16:20]
      assert(eeprom_attr1 == eeprom_attr)

  yield

def test_system_measure_Range_1():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xc9, 0x61, 0x2a, 0xd7, 0xfe, 0x54, 0x75, 0x66,
        0xf8, 0x52, 0x56, 0x0f, 0x72, 0xc7, 0x35, 0x77,
        0xee, 0x74, 0x76, 0x1f, 0x84, 0xdc, 0xe3, 0xa3,
        0xd2, 0xa5, 0x4e, 0xc7, 0x05, 0x25, 0x99, 0x86
      ])
      assert(dgst == std_dgst)

def test_system_measure_Range_2():
  """
  CRTM Measure command Option # 0
  2 addr Range registers are used
  block size 0x2000
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0xe04000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 2, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x8a, 0xa2, 0x60, 0x40, 0xb7, 0x78, 0x54, 0x1b,
        0xeb, 0x09, 0x32, 0xa5, 0x3f, 0x93, 0x3d, 0x47,
        0xe0, 0x9a, 0x69, 0x49, 0x7c, 0x0d, 0x5d, 0x97,
        0x04, 0x37, 0x00, 0x15, 0x04, 0xa3, 0xf0, 0x08
      ])
      assert(dgst == std_dgst)

def test_system_measure_Range_3():
  """
  CRTM Measure command Option # 0
  all addr Range registers are used
  block size 0x2000
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0xe04000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 2, 0xe08000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 3, 0xe0c000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 4, 0xe10000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 5, 0xe14000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 6, 0xe18000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 7, 0xe1c000, 0x2000)
      assert(ret)

      # Measure
      config = b"\x00\00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x2c, 0x9b, 0x52, 0x86, 0xb9, 0xac, 0xde, 0x89,
        0x6f, 0xd5, 0xdd, 0xb2, 0x78, 0x4f, 0xa1, 0x22,
        0x6a, 0xd4, 0x0f, 0xbb, 0x7b, 0xa7, 0x7e, 0x8c,
        0x5e, 0x1f, 0xd0, 0x19, 0x37, 0xf6, 0xa3, 0x1a
      ])
      assert(dgst == std_dgst)

def test_system_measure_Range_4():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x4000
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x4000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xc6, 0xad, 0x68, 0x24, 0xa1, 0x97, 0xa4, 0xed,
        0xc7, 0xcb, 0x8a, 0xdb, 0xce, 0xc7, 0x0a, 0x39,
        0x5d, 0xfd, 0xea, 0x7d, 0xb9, 0xef, 0x73, 0x07,
        0x15, 0x9e, 0xf2, 0xee, 0xcd, 0xd4, 0xdb, 0xaa
      ])
      assert(dgst == std_dgst)

def test_system_measure_Dest_EEPROM():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  destination EEPROM
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\x80"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      # hightest 32-byte EEPROM register
      eeprom_dest_addr = CRTM_Registers.Customer_EEPROM + 0xe0
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from EEPROM
      dgst = utils.get_flash_digest(device, eeprom_dest_addr)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xc9, 0x61, 0x2a, 0xd7, 0xfe, 0x54, 0x75, 0x66,
        0xf8, 0x52, 0x56, 0x0f, 0x72, 0xc7, 0x35, 0x77,
        0xee, 0x74, 0x76, 0x1f, 0x84, 0xdc, 0xe3, 0xa3,
        0xd2, 0xa5, 0x4e, 0xc7, 0x05, 0x25, 0x99, 0x86
      ])
      assert(dgst == std_dgst)

def test_system_measure_AltSrc_EEPROM():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  altSrc EEPROM
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # set EEPROM to known value
      data = bytes([
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
      ])
      # hightest 32-byte EEPROM register
      eeprom_src_addr = CRTM_Registers.Customer_EEPROM + 0xe0
      write_measurement_data(eeprom_src_addr, data, True)

      # Measure
      config = b"\x03\x00"
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x3d, 0x8c, 0xcf, 0x46, 0x3d, 0x79, 0x91, 0xdd,
        0x53, 0x6a, 0xa8, 0x13, 0xed, 0x45, 0x6f, 0xe6,
        0x24, 0x17, 0x0d, 0x0a, 0xa4, 0xab, 0xcb, 0x87,
        0x3d, 0x1d, 0x44, 0xa6, 0x87, 0x32, 0xd8, 0xbc
      ])
      assert(dgst == std_dgst)

def test_system_measure_InputDigest():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  include known input digest
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      in_digest = bytes([
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
      ])

      # Measure
      config = b"\x00\x40"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr, \
        in_digest=in_digest)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x50, 0x72, 0x5f, 0x52, 0x7a, 0xb2, 0xf4, 0x0d,
        0xbb, 0x0c, 0x8b, 0x40, 0x8c, 0xb2, 0x86, 0x2e,
        0x26, 0x51, 0x20, 0xe2, 0xc4, 0xf6, 0x9b, 0x16,
        0xae, 0x69, 0x79, 0xb5, 0x3b, 0x5c, 0x65, 0x9d
      ])
      assert(dgst == std_dgst)

def test_system_measure_KeySrc_NoConstants():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  keySrc: create a digest (no HMAC) without constants in calculation
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\x08"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x49, 0x96, 0xe8, 0x62, 0x13, 0xce, 0xaf, 0xbb,
        0x43, 0xa3, 0xc6, 0xe9, 0x54, 0xae, 0xc3, 0xd1,
        0xbf, 0xc2, 0x31, 0x5b, 0x11, 0x46, 0xbb, 0x13,
        0x1f, 0x85, 0x70, 0x62, 0x2e, 0x05, 0xbd, 0x20
      ])
      assert(dgst == std_dgst)

def test_system_measure_KeySrc_KTR():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  keySrc: KTR
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set KTR to known value
      data = bytes([
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
      ])
      write_measurement_data(CRTM_Registers.KTR_Base, data, False)

      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)            

      # Measure
      config = b"\x00\x20"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x38, 0x97, 0xb5, 0x82, 0xee, 0x29, 0xca, 0x05,
        0x5e, 0x7b, 0x53, 0xbb, 0xa5, 0xee, 0x99, 0xa0,
        0xdd, 0xd6, 0x87, 0x6c, 0x88, 0x35, 0x70, 0x73,
        0x81, 0x4e, 0x09, 0x4e, 0xd1, 0x18, 0xc2, 0xc0
      ])
      assert(dgst == std_dgst)

def test_system_measure_KeySrc_EEPROM():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  keySrc: EEPROM
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set EEPROM to known value
      data = bytes([
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
        0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
      ])
      # hightest 32-byte EEPROM register
      eeprom_key_addr = CRTM_Registers.Customer_EEPROM + 0xe0
      write_measurement_data(eeprom_key_addr, data, True)

      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)            

      # Measure
      config = b"\x00\x38"
      eeprom_src_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xab, 0xc3, 0x60, 0x4c, 0x2c, 0x51, 0x58, 0xcd,
        0x0b, 0xed, 0x61, 0x4c, 0xfd, 0xbf, 0x4e, 0x22,
        0x49, 0xdd, 0x7e, 0xb8, 0x15, 0x56, 0x9a, 0xf1,
        0xcb, 0x4a, 0x7c, 0xcd, 0x1f, 0xc0, 0x5b, 0xaf
      ])
      assert(dgst == std_dgst)

def test_system_measure_KeySrc_UDS():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  keySrc: UDS
  """
  if provision_stats.Provisioned != 0 and provision_stats.Remotely_Provisioned == 1 \
    and provision_stats.RIOT_Mode == 1 and uds_disabled == 0:
    with device:
      # replace UDS with a known value
      ret = conftest.replace_srk_util(device, uid, b"\x02", new_key_name='test_key_handle', old_key_name='uds_handle')
      assert(ret)

      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 1, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\x30"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0x38, 0x97, 0xb5, 0x82, 0xee, 0x29, 0xca, 0x05,
        0x5e, 0x7b, 0x53, 0xbb, 0xa5, 0xee, 0x99, 0xa0,
        0xdd, 0xd6, 0x87, 0x6c, 0x88, 0x35, 0x70, 0x73,
        0x81, 0x4e, 0x09, 0x4e, 0xd1, 0x18, 0xc2, 0xc0
      ])

      # revert UDS to the old one
      # set config based on whether the old UDS is micron default UDS
      if provision_stats.UDS_Replaced == 1:
        config = b"\x02"
        new_key_name = 'uds_handle'
        old_key_name = 'test_key_handle'
      else:
        config = b"\x03"
        new_key_name = None
        old_key_name = None

      ret = conftest.replace_srk_util(device, uid, config, new_key_name, old_key_name)
      assert(ret)

      assert(dgst == std_dgst)

def test_system_measure_IncludeMetadata():
  """
  CRTM Measure command Option # 0
  1 addr Range register is used
  block size 0x2000
  inlcude CTRM metadata
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0xe00000, 0x2000)
      assert(ret)

      for i in range(1, 8):
        ret = utils.set_crtm_addr_range_reg(device, b"\x01", i, 0x0, 0x0)
        assert(ret)

      # ACRTM_Update_Start
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # ACRTM_Update_Finish
      ret = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      # Measure
      config = b"\x04\x00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xbc, 0x91, 0x93, 0x88, 0xbd, 0xf2, 0xc3, 0xdc,
        0x2d, 0x42, 0x30, 0xae, 0xc3, 0xb1, 0xea, 0x93,
        0xff, 0x81, 0x5d, 0x0d, 0x20, 0x0d, 0x21, 0xbd,
        0xa5, 0xbf, 0x96, 0x8a, 0x0f, 0xb4, 0xa7, 0xdf
      ])
      assert(dgst == std_dgst)

def test_system_measure_NoInput():
  """
  CRTM Measure command Option # 0
  CRTM linked list pointed to by the CRTM_Addr
  and CRTM_Size registers results in nothing to measure
  """
  if provision_stats.Provisioned != 0:
    with device:
      # set volatile register Range
      ret = utils.set_crtm_addr_range_reg(device, b"\x01", 0, 0x0, 0x0)
      assert(ret)

      # Measure
      config = b"\x00\x00"
      eeprom_src_addr = 0x0000
      eeprom_key_addr = 0x0000
      eeprom_dest_addr = 0x0000
      err, _ = device.meta_measure(config, eeprom_src_addr, eeprom_key_addr, eeprom_dest_addr)
      assert(err.esr_and_response_success)

      # read results from KTR
      dgst = utils.get_flash_digest(device, CRTM_Registers.KTR_Base)
      assert(dgst)

      # verify results
      std_dgst = bytes([
        0xa1, 0xff, 0x90, 0x96, 0x80, 0x25, 0xb4, 0x28,
        0x7f, 0x54, 0x4a, 0xce, 0x9e, 0x6a, 0xb6, 0x67,
        0x18, 0x7c, 0xf8, 0x8e, 0x0a, 0x37, 0xd6, 0x50,
        0xda, 0xae, 0x90, 0x23, 0xd4, 0x4d, 0x1c, 0x7a
      ])
      assert(dgst == std_dgst)

