#! /usr/bin/env python3

"""
System test: recovery.
(C) 2019 Micron Technology, Inc. All rights reserved.
"""

import sys
import os
import pytest

import authenta.naxos.utils as utils
from authenta.naxos.utils import CRTM_Registers

import conftest

@pytest.fixture(scope="module", autouse=True)
def test_system_recovery_setup(request):
  global device
  global iboot_ip
  global provision_stats
  global device_config_int

  device = conftest.get_device(request)
  iboot_ip = request.config.cache.get("iboot_ip", None)
  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)
    (_, _, _, provision_stats) = ret

    if provision_stats.Provisioned != 0:
      # read current device config
      err, response = device.meta_crtm_register_read(b"\x00", CRTM_Registers.Device_Config)
      assert(err.esr_and_response_success)

      device_config = response[16:20]
      device_config_int = int.from_bytes(device_config, byteorder="big")

      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # turn on automeasure and recovery mode
      data_int = device_config_int | 0x00000003
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set recovery base address
      recovery_base_addr = 0xf00000
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Recovery_Block_Base_Addr, \
        recovery_base_addr.to_bytes(4, byteorder="big"), True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # set CRTM range to meaure
      ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xe00000, 0x2000)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
      if not ret:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      # CRTM_Max_AutoMeasure_Size
      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.CRTM_Max_AutoMeasure_Size, \
        b"\x00\x02\x00\x00", True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

      # reprogram both the Flash_DigestNV_Partial and Flash_DigestNV
      err, _ = device.meta_flash_extend(b"\x01")
      assert(err.esr_and_response_success)

  yield

  if provision_stats.Provisioned != 0:
    with device:
      err, _ = device.meta_acrtm_update_start(b"\x00")
      assert(err.esr_and_response_success)

      # turn off automeasure and recovery mode
      data_int = device_config_int & 0xFFFFFFFC
      data = data_int.to_bytes(4, byteorder="big")

      err, _ = device.meta_crtm_register_write(b"\x00", CRTM_Registers.Device_Config, data, True)
      if not err.esr_and_response_success:
        device.meta_acrtm_update_finish(b"\x00")
        assert(False)

      err, _ = device.meta_acrtm_update_finish(b"\x00")
      assert(err.esr_and_response_success)

def test_system_recovery():
  """
  Change measurement range and reboot to make device go to recovery mode.
  Finish recovery mode by calling ACRTM_Update_Finish.
  """
  if provision_stats.Provisioned == 0:
    return

  with device:
    # ACRTM_Update_Start
    err, _ = device.meta_acrtm_update_start(b"\x00")
    assert(err.esr_and_response_success)

    # change CRTM range to measure
    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 0, 0xe00000, 0x1000)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    ret = utils.set_crtm_addr_range_reg(device, b"\x00", 1, 0x0, 0x0)
    if not ret:
      device.meta_acrtm_update_finish(b"\x00")
      assert(False)

    # ACRTM_Update_Finish
    err, _ = device.meta_acrtm_update_finish(b"\x00")
    assert(err.esr_and_response_success)

  # reboot and automeasure
  # should cause CRTM mismatch and enter recovery mode
  ret = conftest.reboot(device, iboot_ip)
  assert(ret)

  with device:
    ret = utils.get_dev_info(device)
    assert(ret != None)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.CRTM_Mismatch == 1)
    assert(extended_crtm_status.CRTM_Range_Invalid == 0)
    assert(extended_crtm_status.Recovery_Block == 1)

    # finish recovery mode        
    # ACRTM_Update_Finish option #1
    err, _ = device.meta_acrtm_update_finish(b"\x01")
    assert(err.esr_and_response_success)

    # Extended_CRTM_Status
    extended_crtm_status = device.extended_crtm_status_struct
    assert(extended_crtm_status.Recovery_Block == 0)
